//
// Compaction.cpp contains routines dealing with the compaction technique
// for reducing trace size.
//
// The point of the compaction technique is to not record too many message-
// check calls by a thread when that thread is busy-waiting. So if the
// thread is in a tight PeekMessage loop, we don't want to generate tons
// of events for each such call.
//
// The way we accomplish this is to keep track, for each thread, of how
// many message-checks that thread has performed since the last time it
// showed some signs of activity. If a message-check occurs which causes
// this count to become some integer which is not a power of 2, we don't
// record it. This means that if a thread is in some tight PeekMessage loop,
// we record the 1st, 2nd, 4th, 8th, 16th, etc. calls to PeekMessage, but
// no more. As soon as the thread shows some signs of activity, we set the
// count back to 0 so that the next 1st, 2nd, 4th, etc. calls to PeekMessage
// can be recorded.
//
// The array of such counts is kept in kernelUserSharedState, an array in
// memory that is shared by the kernel and the user-level processes. It
// is shared in this way because both the kernel and the user-level loggers
// need fast, shared access to it. Both need to be able to indicate activity
// when they detect activity by clearing the count to 0. The user-level
// logger needs to increment and check the count whenever it detects and
// is considering logging a message-check.
//
// One complication is that while the process's cursor is an hourglass,
// we consider all threads of the process to be continually active. So we
// keep track of the current cursor, and if we detect a message-check while
// the cursor is an hourglass, we set the count to 1 (only this one message-
// check since the last activity).
//
// Another complication is that the array of counts only contains one byte
// per thread. So if the count goes above 254, we need to keep track of
// what the real count is. Per-thread state is allocated and used whenever
// this becomes necessary, and this state reflects the count. So, if the
// kernel-user-shared count is 255, this means the real count is in the
// per-thread state. Only the user-level code needs access to this state,
// fortunately. (This is because the kernel only needs to zero the count,
// and never needs to increment or query it.)
//

#include <windows.h>
#include "../Common/VTrace.h"
#include "Globals.h"

////////////////////////////////////////////////////////////////////////////
//                     INITIALIZATION / TERMINATION
////////////////////////////////////////////////////////////////////////////

// InitCompactionTechniquePerProcesVariables initializes all the
// per-process values for needed for the compaction technique.

void InitCompactionTechniquePerProcessVariables (void)
{
  compactionProcessVars.waitCursor = NULL;
  compactionProcessVars.appStartingCursor = NULL;
  compactionProcessVars.cursorIsHourglass = FALSE;
}

// InitCompactionTechniquePerProcesVariables initializes all the
// per-thread values needed for the compaction technique.  These values
// are in the structure pointed to by the parameter "vars".

void InitCompactionTechniquePerThreadVariables
                                         (COMPACTION_PER_THREAD_VARS *vars)
{
  vars->msgChecksSinceActivity = 255;
}

// FreeCompactionTechniquePerThreadVariables frees the space for all the
// per-thread values needed for the compaction technique.

void FreeCompactionTechniquePerThreadVariables (void)
{
  COMPACTION_PER_THREAD_VARS *compactionThreadVarsPtr;

  compactionThreadVarsPtr = (COMPACTION_PER_THREAD_VARS *)
                            TlsGetValue(compactionThreadVarsPtrTlsIndex);
  if (compactionThreadVarsPtr == NULL) return;

  // If GlobalFree succeeds, it returns NULL and this sets the thread-local
  // storage value to NULL (so it never gets deallocated again).  If
  // GlobalFree fails, it returns the original pointer, which is put in
  // thread-local storage where it can eventually be freed again.

  TlsSetValue(compactionThreadVarsPtrTlsIndex,
              GlobalFree(compactionThreadVarsPtr));
}

////////////////////////////////////////////////////////////////////////////
//                          THREAD-LOCAL STORAGE
////////////////////////////////////////////////////////////////////////////

// GetCompactionThreadVarsPtr gets the pointer to the compaction per-thread
// variables for the current thread. If space for those variables has not
// yet been allocated, it allocates and initializes that space. If it fails,
// it returns NULL.

COMPACTION_PER_THREAD_VARS *GetCompactionThreadVarsPtr (void)
{
  LPVOID currentValue;

  currentValue = TlsGetValue(compactionThreadVarsPtrTlsIndex);
  if (currentValue == NULL) {
    currentValue = (LPVOID) GlobalAlloc(0, sizeof(COMPACTION_PER_THREAD_VARS));
    if (currentValue == NULL)
      return NULL;
    else {
      if (TlsSetValue(compactionThreadVarsPtrTlsIndex, currentValue)) {
        InitCompactionTechniquePerThreadVariables(
                       (COMPACTION_PER_THREAD_VARS *) currentValue);
      }
      else {
        GlobalFree((HGLOBAL) currentValue);
        return NULL;
      }
    }
  }
  return (COMPACTION_PER_THREAD_VARS *) currentValue;
}

////////////////////////////////////////////////////////////////////////////
//                           THREAD ACTIVITY
////////////////////////////////////////////////////////////////////////////

// IndicateThreadActivity indicates that the current thread has been active.

void IndicateThreadActivity (void)
{
  unsigned long threadId = GetCurrentThreadId();

  if (threadId < kernelUserSharedState->maxThreads)
    kernelUserSharedState->msgChecksSinceActivity[threadId] = '\0';
}

// WasThreadActiveBeforeNow indicates that the current thread has been active,
// and returns a boolean indicating if this was already known.

BOOL WasThreadActiveBeforeNow (void)
{
  unsigned long threadId = GetCurrentThreadId();

  if (threadId >= kernelUserSharedState->maxThreads)
    return TRUE;

  if (kernelUserSharedState->msgChecksSinceActivity[threadId] != '\0') {
    kernelUserSharedState->msgChecksSinceActivity[threadId] = '\0';
    return FALSE;
  }
  else
    return TRUE;
}

// IsMessageCheckWorthLogging is called when the message queue is checked.
// It returns a boolean indicating whether this check is worth logging.
// The parameters passed are a boolean indicating if the message queue check
// found a message (messageFound), and a pointer in which to place the
// encoding of the number of consecutive message checks without activity
// this represents. If the number is 0 (i.e. if this message check found a
// message and thus generated activity), its encoding is 0; otherwise, the
// encoding is one plus the log base 2 of the number (which is guaranteed to
// be an integer since only integer powers of 2 are worth logging).
//
// The way this routine works is it computes the number of message checks
// since the last activity, sets the per-thread state to reflect this, then,
// if this number of message checks is 0 or an integer power of 2, returns
// TRUE and the proper encoding of this number. The per-thread state is
// recorded in the following manner. Usually, we simply record the number
// of message checks without activity in the byte at kernelUserSharedState->
// msgChecksSinceActivity[GetCurrentThreadId()]. However, if the number
// is 255 or greater, then kernelUserSharedState->msgChecksSinceActivity
// [GetCurrentThreadId()] will be 255, and the number will be found in
// GetCompactionThreadVarsPtr()->msgChecksSinceActivity.

BOOL IsMessageCheckWorthLogging (BOOL messageFound, char *checksEncodingPtr)
{
  DWORD threadId = GetCurrentThreadId();
  COMPACTION_PER_THREAD_VARS *compactionThreadVarsPtr;
  DWORD msgChecksSinceActivity;

  // If this thread ID is more than the number we can watch, we have no
  // choice but to just assume there has been no activity since the last
  // check.

  if (threadId >= kernelUserSharedState->maxThreads) {
    *checksEncodingPtr = messageFound ? '\0' : '\1';
    return TRUE;
  }

  // If a message was found, then this is worth logging and the encoding is 0.
  // Furthermore, we should set the global number of checks since the last
  // activity to 0.

  if (messageFound) {
    kernelUserSharedState->msgChecksSinceActivity[threadId] = '\0';
    *checksEncodingPtr = '\0';
    return TRUE;
  }

  // If the cursor is an hourglass, then the thread is considered to have
  // been active. We indicate that by setting the number of message checks
  // since the last activity to 1 (since this is a message check) and
  // encoding that as 1.

  if (compactionProcessVars.cursorIsHourglass) {
    kernelUserSharedState->msgChecksSinceActivity[threadId] = '\1';
    *checksEncodingPtr = '\1';
    return TRUE;
  }

  // Increment the number of message checks since the last activity, saving
  // the new count of message checks in msgChecksSinceActivity. This is
  // a little complicated because, as described above, the message checks
  // counter reaches a maximum at 255, beyond which point thread local storage
  // is used to contain the counter value.

  msgChecksSinceActivity =
    kernelUserSharedState->msgChecksSinceActivity[threadId];
  if (msgChecksSinceActivity < 255) {
    msgChecksSinceActivity++;
    kernelUserSharedState->msgChecksSinceActivity[threadId] =
      (unsigned char) msgChecksSinceActivity;
    if (msgChecksSinceActivity == 255) {
      compactionThreadVarsPtr = GetCompactionThreadVarsPtr();
      if (compactionThreadVarsPtr)
        compactionThreadVarsPtr->msgChecksSinceActivity = 255;
    }
  }
  else {
    compactionThreadVarsPtr = GetCompactionThreadVarsPtr();
    if (compactionThreadVarsPtr) {
      compactionThreadVarsPtr->msgChecksSinceActivity++;
      msgChecksSinceActivity = compactionThreadVarsPtr->msgChecksSinceActivity;
    }
  }

  // If the message checks count is an integer power of two, we indicate
  // that this message check is worth logging and encode the count as
  // the base-2 logarithm of the count plus 1.

  if (!(msgChecksSinceActivity & (msgChecksSinceActivity - 1))) {
    *checksEncodingPtr = '\0';
    do {
      msgChecksSinceActivity >>= 1;
      (*checksEncodingPtr)++;
    } while (msgChecksSinceActivity);
    return TRUE;
  }
  else
    return FALSE;
}
