//
// HookFuncs.cpp
//
// Code that implements "new" versions of certain functions in USER32.DLL
// and KERNEL32.DLL.  These "new" versions essentially just trace the
// parameters and return values while calling the "old" versions, found
// in the hook info structure.  However, some of these functions have other
// uses.  For instance, some of the message hook functions implement trace
// compaction in order to cut down on message-queue-check trace entries.
// Also, the load-library hook functions wait until the new libraries are
// loaded, then install all the hook functions into these new DLL's.
//

#include <windows.h>
#include <winioctl.h>
#include <winsvc.h>
#include "../Common/VTrace.h"
#include "Globals.h"
#include "Compaction.h"
#include "LogEvent.h"
#include "HookManage.h"
#include "VTraceDLL.h"

// We undefine TranslateAccelerator so that it is not synonymous with
// TranslateAcceleratorA. TranslateAccelerator and TranslateAcceleratorA
// are actually different functions.

#undef TranslateAccelerator

////////////////////////////////////////////////////////////////////////////
//                              MACROS
////////////////////////////////////////////////////////////////////////////

#define MIN(a,b)     (((a)>(b))?(b):(a))

/////////////////////////////////////////////////////////////////////////
//                         HELPER FUNCTIONS
/////////////////////////////////////////////////////////////////////////

// MyCopyUnicodeToCStr is a helper function that copies a Unicode string
// to a C string, but not more than maxLength characters.  It returns
// the number of characters copied.

UINT MyCopyUnicodeToCStr (char *cStr, LPCWSTR unicodeStr, UINT maxLength)
{
  UINT i;

  for (i = 0; (*unicodeStr != L'\0') && (i < maxLength); i++) {
    cStr[i] = * ((char *) unicodeStr);
    unicodeStr++;
  }

  return i;
}

// EncryptAscii encrypts the WPARAM and LPARAM fields of a message
// indicating the arrival of a character.  It is used for WM_CHAR
// messages, which represent the character in ASCII.  The encryption
// is done by converting any letters to x's and any numbers to 9's.
// The parameters to this function are pointers to the WPARAM and
// LPARAM values to be encrypted.

void EncryptAscii (ULONG *wParam, ULONG *lParam)
{
  if (*wParam >= 'a' && *wParam <= 'z') {      // If it is a lowercase letter,
    *wParam = 'x';                             //   convert to lowercase x and
    *lParam &= 0xFF00FFFFUL;                   //   change the scan code to be
    *lParam |= 0x002D0000UL;                   //   0x2d (x)
  }
  else if (*wParam >= 'A' && *wParam <= 'Z') { // If it is an uppercase letter,
    *wParam = 'X';                             //   convert it to capital X and
    *lParam &= 0xFF00FFFFUL;                   //   change the scan code to be
    *lParam |= 0x002D0000UL;                   //   0x2d (x)
  }
  else if (*wParam >= '0' && *wParam <= '9') { // If it is a number,
    *wParam = '9';                             //   convert it to a 9 and...
    if ((*lParam & 0x00FF0000UL) >= 0x00110000UL) {  // If it is from keypad
      *lParam &= 0xFF00FFFFUL;                       //   change scan code to
      *lParam |= 0x00490000UL;                       //   0x49 (keypad 9)
    }
    else {                                           // Otherwise,
      *lParam &= 0xFF00FFFFUL;                       //   change scan code to
      *lParam |= 0x000A0000UL;                       //   0x0a (nonkeypad 9)
    }
  }
}

// EncryptVirtual encrypts the WPARAM and LPARAM fields of a message
// indicating the arrival of a character.  It is used for WM_KEYDOWN,
// WM_KEYUP, WM_SYSKEYDOWN, and WM_SYSKEYUP messages, which represent
// the character as a virtual key code.  The encryption is done by
// converting any letters to x's and any numbers to 9's.  The parameters
// to this function are pointers to the WPARAM and LPARAM values to be
// encrypted.

void EncryptVirtual (ULONG *wParam, ULONG *lParam)
{
  if (*wParam >= 'A' && *wParam <= 'Z') {         // If it is a letter,
    *wParam = 'X';                                //   change it to an X
    *lParam &= 0xFF00FFFFUL;                      //   and change scan code to
    *lParam |= 0x002D0000UL;                      //   0x2d (X)
  }
  else if (*wParam >= '0' && *wParam <= '9') {    // If it is a normal digit,
    *wParam = '9';                                //   change it to a normal 9
    *lParam &= 0xFF00FFFFUL;                      //   and change scan code to
    *lParam |= 0x000A0000UL;                      //   0x0a (normal 9)
  }
  else if (*wParam >= VK_NUMPAD0 && *wParam <= VK_NUMPAD9) {
                                                  // If it is a keypad digit,
    *wParam = VK_NUMPAD9;                         //    change it to a keypad 9
    *lParam &= 0xFF00FFFFUL;                      //    and change scan code to
    *lParam |= 0x00490000UL;                      //    0x49 (keypad 9)
  }
}

// EnterMessage traces the receipt of a message. The parameters to this
// function are as follows.
//   messageReceived : was a message received?
//   lpMsg : a pointer to the message received
//   eventType : the event type that caused this message to be received
//
// Entering the message means copying its relevant fields to the
// trace entry buffer and putting the ultimate length of this buffer at
// entry[1]. The job is a bit complicated by two things.  First, if
// this is a virtual key code, we have to check the keyboard state to
// get important information about the state of the modifier keys (which
// we then put in the trace entry buffer). Second, if this is an unmodified
// character, we have to encrypt it before storing it in the entry buffer.

void EnterMessage (BOOL messageReceived, LPMSG lpMsg, char eventType)
{
  UCHAR keyState = 0;
  HWND hwnd = lpMsg->hwnd;
  UINT message = lpMsg->message;
  ULONG wParam = lpMsg->wParam;
  ULONG lParam = lpMsg->lParam;
  ULONG bytesReturned;
  char eventBuffer[23];

  // If no message was received, then enter "no message".

  if (!messageReceived) {
    eventBuffer[0] = eventType;
    eventBuffer[1] = 2;
    LogEvent(eventBuffer, 2);
    return;
  }

  // If this is a virtual-key message, find out the "current" state of
  // the modifier keys.  This is actually the state that the keyboard was
  // in when this message was put in the queue.  Then, if the Ctrl or
  // Alt key was not on, and this is not a system message, encrypt the
  // message.

  // Otherwise, if this is a character message, encrypt it no matter
  // what.

  if (message == WM_KEYDOWN || message == WM_KEYUP ||
      message == WM_SYSKEYDOWN || message == WM_SYSKEYUP) {
    if (GetKeyState(VK_SHIFT) & 0x8000) keyState |= SHIFT_MODIFIER_ON;
    if (GetKeyState(VK_CONTROL) & 0x8000) keyState |= CTRL_MODIFIER_ON;
    if (GetKeyState(VK_MENU) & 0x8000) keyState |= ALT_MODIFIER_ON;
    if ((message <= WM_KEYUP) &&
        !(keyState & (CTRL_MODIFIER_ON | ALT_MODIFIER_ON)))
      EncryptVirtual(&wParam, &lParam);
  }
  else if (message == WM_CHAR)
    EncryptAscii(&wParam, &lParam);

  // If this is a representation of user mouse activity other than mouse
  // movement, indicate user activity.  (We don't indicate activity for
  // keypresses, because VTrcKbd should already do that.  Also, we don't
  // indicate activity for mouse movement, because some screen savers,
  // including the logon screen saver, generate bogus mouse movement
  // events.)

  if (message > WM_MOUSEMOVE && message <= WM_MOUSELAST) {
    DeviceIoControl(vtrcInitHandle, VTRACE_CTRL_INDICATE_USER_ACTIVITY,
                    NULL, 0, NULL, 0, &bytesReturned, NULL);
  }

  // Only copy the key state field if it is nonzero (i.e. if there are
  // any modifiers down worth reporting).  Based on whether the key
  // state field is entered or not, the entry length may be 22 or 23.

  if (keyState) {
    eventBuffer[0] = eventType;
    eventBuffer[1] = (char) 23;
    eventBuffer[22] = keyState;
  }
  else {
    eventBuffer[0] = eventType;
    eventBuffer[1] = (char) 22;
  }

  // Copy all the fields to the event buffer if logging is on.

  memcpy(&eventBuffer[2], &lpMsg, 4);
  memcpy(&eventBuffer[6], &message, 4);
  memcpy(&eventBuffer[10], &hwnd, 4);
  memcpy(&eventBuffer[14], &wParam, 4);
  memcpy(&eventBuffer[18], &lParam, 4);
  LogEvent(eventBuffer, eventBuffer[1]);
}

////////////////////////////////////////////////////////////////////////////
//                              HOOK FUNCTIONS...
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
//                         MESSAGE QUEUE FUNCTIONS                        //
////////////////////////////////////////////////////////////////////////////

// VTraceHookGetMessageA is the hook for GetMessageA.  It logs the call
// and return.

BOOL
WINAPI
VTraceHookGetMessageA(
    LPMSG lpMsg,
    HWND hWnd,
    UINT wMsgFilterMin,
    UINT wMsgFilterMax)
{
  BOOL retVal;
  char eventBuffer[13];

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the call, indicating whether a filter was provided or not

  if (wMsgFilterMin || wMsgFilterMax || hWnd) {
    eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_GET_FILTER_CALL;
    memcpy(&eventBuffer[1], &wMsgFilterMin, 4);
    memcpy(&eventBuffer[5], &wMsgFilterMax, 4);
    memcpy(&eventBuffer[9], &hWnd, 4);
    LogEvent(eventBuffer, 13);
  }
  else {
    eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_GET_NOFILTER_CALL;
    LogEvent(eventBuffer, 1);
  }

  // Call the original function, retaining the return value

  retVal = (* (GetMessage_PROC)
              hookInfo[GetMessageA_INDEX].origFunctionAddress)
              (lpMsg, hWnd, wMsgFilterMin, wMsgFilterMax);

  // If a message was found, then record the message, otherwise record no
  // message.

  EnterMessage(retVal != -1, lpMsg, (char) ENTRY_TYPE_MESSAGE_GET_RETURN);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookGetMessageW is the hook for GetMessageW.  It logs the call
// and return.

BOOL
WINAPI
VTraceHookGetMessageW(
    LPMSG lpMsg,
    HWND hWnd ,
    UINT wMsgFilterMin,
    UINT wMsgFilterMax)
{
  BOOL retVal;
  char eventBuffer[13];

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the call, indicating whether a filter was provided or not

  if (wMsgFilterMin || wMsgFilterMax || hWnd) {
    eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_GET_FILTER_CALL;
    memcpy(&eventBuffer[1], &wMsgFilterMin, 4);
    memcpy(&eventBuffer[5], &wMsgFilterMax, 4);
    memcpy(&eventBuffer[9], &hWnd, 4);
    LogEvent(eventBuffer, 13);
  }
  else {
    eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_GET_NOFILTER_CALL;
    LogEvent(eventBuffer, 1);
  }

  // Call the original function, retaining the return value

  retVal = (* (GetMessage_PROC)
              hookInfo[GetMessageW_INDEX].origFunctionAddress)
              (lpMsg, hWnd, wMsgFilterMin, wMsgFilterMax);

  // If a message was found, then record the message, otherwise record no
  // message.

  EnterMessage(retVal != -1, lpMsg, (char) ENTRY_TYPE_MESSAGE_GET_RETURN);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookPeekMessageA is the hook for PeekMessageA.  It logs the call
// and return, and also performs processing related to the compaction
// technique.

BOOL
WINAPI
VTraceHookPeekMessageA(
    LPMSG lpMsg,
    HWND hWnd ,
    UINT wMsgFilterMin,
    UINT wMsgFilterMax,
    UINT wRemoveMsg)
{
  BOOL retVal;
  char checksEncoding;
  char eventBuffer[14];

  // Call the original function, retaining the return value

  retVal = (* (PeekMessage_PROC)
              hookInfo[PeekMessageA_INDEX].origFunctionAddress)
              (lpMsg, hWnd, wMsgFilterMin, wMsgFilterMax, wRemoveMsg);

  // Find out if this message check is worth recording and, if so, what
  // encoding of the number of consecutive message checks to use.
  // If it is worth recording, record both the call and return.  Note that
  // the logging of the call in this case is actually done _after_ the call,
  // because we needed to know whether a message was actually returned when
  // deciding whether it was worth recording.

  if (IsMessageCheckWorthLogging(retVal, &checksEncoding)) {
    // Log the call, indicating whether a filter was provided and whether
    // this call is a "remove" call, i.e. one that will remove any event
    // that is found.

    if (wMsgFilterMin || wMsgFilterMax || hWnd) {
      eventBuffer[0] = (char) ((wRemoveMsg & PM_REMOVE) ?
                               ENTRY_TYPE_MESSAGE_PEEK_REMOVE_FILTER_CALL :
                               ENTRY_TYPE_MESSAGE_PEEK_NOREMOVE_FILTER_CALL);
      memcpy(&eventBuffer[1], &wMsgFilterMin, 4);
      memcpy(&eventBuffer[5], &wMsgFilterMax, 4);
      memcpy(&eventBuffer[9], &hWnd, 4);
      eventBuffer[13] = checksEncoding;
      LogEvent(eventBuffer, 14);
    }
    else {
      eventBuffer[0] = (char) ((wRemoveMsg & PM_REMOVE) ?
                               ENTRY_TYPE_MESSAGE_PEEK_REMOVE_NOFILTER_CALL :
                               ENTRY_TYPE_MESSAGE_PEEK_NOREMOVE_NOFILTER_CALL);
      eventBuffer[1] = checksEncoding;
      LogEvent(eventBuffer, 2);
    }

    // If a message was found, then record the message, otherwise record no
    // message.

    EnterMessage(retVal, lpMsg, (char) ENTRY_TYPE_MESSAGE_PEEK_RETURN);
  }

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookPeekMessageW is the hook for PeekMessageW.  It logs the call
// and return, and also performs processing related to the compaction
// technique.

BOOL
WINAPI
VTraceHookPeekMessageW(
    LPMSG lpMsg,
    HWND hWnd ,
    UINT wMsgFilterMin,
    UINT wMsgFilterMax,
    UINT wRemoveMsg)
{
  BOOL retVal;
  char checksEncoding;
  char eventBuffer[14];

  // Call the original function, retaining the return value

  retVal = (* (PeekMessage_PROC)
              hookInfo[PeekMessageW_INDEX].origFunctionAddress)
              (lpMsg, hWnd, wMsgFilterMin, wMsgFilterMax, wRemoveMsg);

  // Find out if this message check is worth recording and, if so, what
  // encoding of the number of consecutive message checks to use.
  // If it is worth recording, record both the call and return.  Note that
  // the logging of the call in this case is actually done _after_ the call,
  // because we needed to know whether a message was actually returned when
  // deciding whether it was worth recording.

  if (IsMessageCheckWorthLogging(retVal, &checksEncoding)) {
    // Log the call, indicating whether a filter was provided and whether
    // this call is a "remove" call, i.e. one that will remove any event
    // that is found.
    if (wMsgFilterMin || wMsgFilterMax || hWnd) {
      eventBuffer[0] = (char) ((wRemoveMsg & PM_REMOVE) ?
                               ENTRY_TYPE_MESSAGE_PEEK_REMOVE_FILTER_CALL :
                               ENTRY_TYPE_MESSAGE_PEEK_NOREMOVE_FILTER_CALL);
      memcpy(&eventBuffer[1], &wMsgFilterMin, 4);
      memcpy(&eventBuffer[5], &wMsgFilterMax, 4);
      memcpy(&eventBuffer[9], &hWnd, 4);
      eventBuffer[13] = checksEncoding;
      LogEvent(eventBuffer, 14);
    }
    else {
      eventBuffer[0] = (char) ((wRemoveMsg & PM_REMOVE) ?
                               ENTRY_TYPE_MESSAGE_PEEK_REMOVE_NOFILTER_CALL :
                               ENTRY_TYPE_MESSAGE_PEEK_NOREMOVE_NOFILTER_CALL);
      eventBuffer[1] = checksEncoding;
      LogEvent(eventBuffer, 2);
    }

    // If a message was found, then record the message, otherwise record no
    // message.

    EnterMessage(retVal, lpMsg, (char) ENTRY_TYPE_MESSAGE_PEEK_RETURN);
  }

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookMsgWaitForMultipleObjects is the hook for
// MsgWaitForMultipleObjects.  It logs the call and return.

DWORD
WINAPI
VTraceHookMsgWaitForMultipleObjects(
    DWORD nCount,
    LPHANDLE pHandles,
    BOOL fWaitAll,
    DWORD dwMilliseconds,
    DWORD dwWakeMask)
{
  DWORD retVal;
  char eventBuffer[MAX_WAITABLE_OBJECTS_RECORDED * 4 + 10];
  DWORD i, objectsToRecord;

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Trace the beginning of the wait, including all the handles being waited
  // on (up to MAX_WAITABLE_OBJECTS_RECORDED of them).

  objectsToRecord = MIN(nCount, MAX_WAITABLE_OBJECTS_RECORDED);
  eventBuffer[0] = (char) ENTRY_TYPE_MSG_AND_WAITOBJ_WAIT_CALL;
  eventBuffer[1] = (char) (objectsToRecord * 4 + 10);
  memcpy(&eventBuffer[2], &dwMilliseconds, 4);
  memcpy(&eventBuffer[6], &dwWakeMask, 4);
  for (i = 0; i < objectsToRecord; i++)
    memcpy(&eventBuffer[i * 4 + 10], &pHandles[i], 4);
  LogEvent(eventBuffer, objectsToRecord * 4 + 10);

  // Call the original function, retaining the return value

  retVal = (* (MsgWaitForMultipleObjects_PROC)
              hookInfo[MsgWaitForMultipleObjects_INDEX].origFunctionAddress)
              (nCount, pHandles, fWaitAll, dwMilliseconds, dwWakeMask);

  // Trace the end of the wait

  eventBuffer[0] = (char) ENTRY_TYPE_MSG_AND_WAITOBJ_WAIT_RETURN;
  memcpy(&eventBuffer[1], &retVal, 4);
  LogEvent(eventBuffer, 5);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookMsgWaitForMultipleObjectsEx is the hook for
// MsgWaitForMultipleObjectsEx.  It logs the call and return.

DWORD
WINAPI
VTraceHookMsgWaitForMultipleObjectsEx(
    DWORD nCount,
    LPHANDLE pHandles,
    DWORD dwMilliseconds,
    DWORD dwWakeMask,
    DWORD dwFlags)
{
  DWORD retVal;
  char eventBuffer[MAX_WAITABLE_OBJECTS_RECORDED * 4 + 10];
  DWORD i, objectsToRecord;

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Trace the beginning of the wait, including all the handles being waited
  // for (up to MAX_WAITABLE_OBJECTS_RECORDED of them).

  objectsToRecord = MIN(nCount, MAX_WAITABLE_OBJECTS_RECORDED);
  eventBuffer[0] = (char) ENTRY_TYPE_MSG_AND_WAITOBJ_WAIT_CALL;
  eventBuffer[1] = (char) (objectsToRecord * 4 + 10);
  memcpy(&eventBuffer[2], &dwMilliseconds, 4);
  memcpy(&eventBuffer[6], &dwWakeMask, 4);
  for (i = 0; i < objectsToRecord; i++)
    memcpy(&eventBuffer[i * 4 + 10], &pHandles[i], 4);
  LogEvent(eventBuffer, objectsToRecord * 4 + 10);

  // Call the original function, retaining the return value

  retVal = (* (MsgWaitForMultipleObjectsEx_PROC)
              hookInfo[MsgWaitForMultipleObjectsEx_INDEX].origFunctionAddress)
              (nCount, pHandles, dwMilliseconds, dwWakeMask, dwFlags);

  // Trace the end of the wait

  eventBuffer[0] = (char) ENTRY_TYPE_MSG_AND_WAITOBJ_WAIT_RETURN;
  memcpy(&eventBuffer[1], &retVal, 4);
  LogEvent(eventBuffer, 5);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookWaitMessage is the hook for WaitMessage.  It logs the call and
// return.

BOOL
WINAPI
VTraceHookWaitMessage(
    VOID)
{
  BOOL retVal;
  char eventBuffer[1];

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the call.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_WAIT_CALL;
  LogEvent(eventBuffer, 1);

  // Call the original function, retaining the return value

  retVal = (* (WaitMessage_PROC)
             hookInfo[WaitMessage_INDEX].origFunctionAddress)
             ();

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_WAIT_RETURN;
  LogEvent(eventBuffer, 1);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookGetInputState is the hook for GetInputState.  It logs the return.
// Also, since this is essentially a check of the message queue, this function
// also helps to implement the compaction technique.

BOOL
WINAPI
VTraceHookGetInputState(
    VOID)
{
  BOOL retVal;
  char checksEncoding;
  char eventBuffer[2];

  // Call the original function, retaining the return value

  retVal = (* (GetInputState_PROC)
             hookInfo[GetInputState_INDEX].origFunctionAddress)
             ();

  // Find out if this message check is worth recording and, if so, what
  // encoding of the number of consecutive message checks to use

  if (IsMessageCheckWorthLogging(retVal, &checksEncoding)) {
    eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_GET_INPUT_STATE;
    eventBuffer[1] = checksEncoding;
    LogEvent(eventBuffer, 2);
  }

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookGetQueueStatus is the hook for GetQueueStatus.  It logs the
// return.  Also, since this is essentially a check of the message queue,
// this function also helps to implement the compaction technique.

DWORD
WINAPI
VTraceHookGetQueueStatus(
    UINT flags)
{
  DWORD retVal;
  char checksEncoding;
  char eventBuffer[5];

  // Call the original function, retaining the return value

  retVal = (* (GetQueueStatus_PROC)
             hookInfo[GetQueueStatus_INDEX].origFunctionAddress)
             (flags);

  // Find out if this message check is worth recording and, if so, what
  // encoding of the number of consecutive message checks to use

  if (IsMessageCheckWorthLogging(retVal != 0, &checksEncoding)) {
    eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_GET_QUEUE_STATUS;
    memcpy(&eventBuffer[1], &retVal, 4);
    // If the message queue is empty, encode the number of consecutive
    // message checks in the first (least significant) byte.  The resultant
    // number can be distinguished from a non-empty message queue since it
    // can never be the case that retVal's most-significant two bytes are 0
    // while its least-significant two bytes are non-zero (since the least-
    // significant two bytes are a bitfield representing a subset of the set
    // represented by the most-significant two bytes as a bitfield).
    if (!retVal)
      eventBuffer[1] = checksEncoding;
    LogEvent(eventBuffer, 5);
  }

  return retVal;           // Return the return value obtained earlier
}

////////////////////////////////////////////////////////////////////////////
//                            MESSAGE FUNCTIONS                           //
////////////////////////////////////////////////////////////////////////////

// VTraceHookDispatchMessageA is the hook for DispatchMessageA.  It
// logs the call and return.

LONG
WINAPI
VTraceHookDispatchMessageA(
    CONST MSG *lpMsg)
{
  LONG retVal;
  char eventBuffer[5];

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the call.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_DISPATCH_CALL;
  memcpy(&eventBuffer[1], &lpMsg, 4);
  LogEvent(eventBuffer, 5);

  // Call the original function, retaining the return value

  retVal = (* (DispatchMessage_PROC)
              hookInfo[DispatchMessageA_INDEX].origFunctionAddress)
              (lpMsg);

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_DISPATCH_RETURN;
  LogEvent(eventBuffer, 1);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookDispatchMessageW is the hook for DispatchMessageW.  It logs
// the call and return.

LONG
WINAPI
VTraceHookDispatchMessageW(
    CONST MSG *lpMsg)
{
  LONG retVal;
  char eventBuffer[5];

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the call.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_DISPATCH_CALL;
  memcpy(&eventBuffer[1], &lpMsg, 4);
  LogEvent(eventBuffer, 5);

  // Call the original function, retaining the return value

  retVal = (* (DispatchMessage_PROC)
              hookInfo[DispatchMessageW_INDEX].origFunctionAddress)
              (lpMsg);

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_DISPATCH_RETURN;
  LogEvent(eventBuffer, 1);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookTranslateAccelerator is the hook for TranslateAccelerator.
// It logs the call and return

int
WINAPI
VTraceHookTranslateAccelerator(
    HWND hWnd,
    HACCEL hAccTable,
    LPMSG lpMsg)
{
  int retVal;
  char eventBuffer[5];

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the call.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_TRANSLATE_ACCEL_CALL;
  memcpy(&eventBuffer[1], &lpMsg, 4);
  LogEvent(eventBuffer, 5);

  // Call the original function, retaining the return value

  retVal = (* (TranslateAccelerator_PROC)
              hookInfo[TranslateAccelerator_INDEX].origFunctionAddress)
              (hWnd, hAccTable, lpMsg);

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_TRANSLATE_ACCEL_RETURN;
  eventBuffer[1] = (char) (retVal ? 1 : 0);
  LogEvent(eventBuffer, 2);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookTranslateAcceleratorA is the hook for TranslateAcceleratorA.
// It logs the call and return.

int
WINAPI
VTraceHookTranslateAcceleratorA(
    HWND hWnd,
    HACCEL hAccTable,
    LPMSG lpMsg)
{
  int retVal;
  char eventBuffer[5];

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the call.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_TRANSLATE_ACCEL_CALL;
  memcpy(&eventBuffer[1], &lpMsg, 4);
  LogEvent(eventBuffer, 5);

  // Call the original function, retaining the return value

  retVal = (* (TranslateAccelerator_PROC)
              hookInfo[TranslateAcceleratorA_INDEX].origFunctionAddress)
              (hWnd, hAccTable, lpMsg);

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_TRANSLATE_ACCEL_RETURN;
  eventBuffer[1] = (char) (retVal ? 1 : 0);
  LogEvent(eventBuffer, 2);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookTranslateAcceleratorW is the hook for TranslateAcceleratorW.
// It logs the call and return.

int
WINAPI
VTraceHookTranslateAcceleratorW(
    HWND hWnd,
    HACCEL hAccTable,
    LPMSG lpMsg)
{
  int retVal;
  char eventBuffer[5];

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the call.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_TRANSLATE_ACCEL_CALL;
  memcpy(&eventBuffer[1], &lpMsg, 4);
  LogEvent(eventBuffer, 5);

  // Call the original function, retaining the return value

  retVal = (* (TranslateAccelerator_PROC)
              hookInfo[TranslateAcceleratorW_INDEX].origFunctionAddress)
              (hWnd, hAccTable, lpMsg);

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_TRANSLATE_ACCEL_RETURN;
  eventBuffer[1] = (char) (retVal ? 1 : 0);
  LogEvent(eventBuffer, 2);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookTranslateMessage is the hook for TranslateMessage.
// It logs the call and return.

BOOL
WINAPI
VTraceHookTranslateMessage(
    CONST MSG *lpMsg)
{
  BOOL retVal;
  char eventBuffer[5];

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the call.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_TRANSLATE_CALL;
  memcpy(&eventBuffer[1], &lpMsg, 4);
  LogEvent(eventBuffer, 5);

  // Call the original function, retaining the return value

  retVal = (* (TranslateMessage_PROC)
              hookInfo[TranslateMessage_INDEX].origFunctionAddress)
              (lpMsg);

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_TRANSLATE_RETURN;
  eventBuffer[1] = (char) (retVal ? 1 : 0);
  LogEvent(eventBuffer, 2);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookTranslateMDISysAccel is the hook for TranslateMDISysAccel.
// It logs the call and return.

BOOL
WINAPI
VTraceHookTranslateMDISysAccel(
    HWND hWndClient,
    LPMSG lpMsg)
{
  BOOL retVal;
  char eventBuffer[5];

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the call.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_TRANSLATE_MDI_CALL;
  memcpy(&eventBuffer[1], &lpMsg, 4);
  LogEvent(eventBuffer, 5);

  // Call the original function, retaining the return value

  retVal = (* (TranslateMDISysAccel_PROC)
              hookInfo[TranslateMDISysAccel_INDEX].origFunctionAddress)
              (hWndClient, lpMsg);

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_TRANSLATE_MDI_RETURN;
  eventBuffer[1] = (char) (retVal ? 1 : 0);
  LogEvent(eventBuffer, 2);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookSendMessageA is the hook for SendMessageA.  It logs the call.

LRESULT
WINAPI
VTraceHookSendMessageA(
    HWND hWnd,
    UINT Msg,
    WPARAM wParam,
    LPARAM lParam)
{
  char eventBuffer[1];

  // Indicate that the current thread is active, and log this indication
  // of thread activity only if it is important in the determination that
  // this thread is active.

  if (!WasThreadActiveBeforeNow()) {
    eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_SEND;
    LogEvent(eventBuffer, 1);
  }

  // Call the original function.

  return (* (SendMessage_PROC)
            hookInfo[SendMessageA_INDEX].origFunctionAddress)
            (hWnd, Msg, wParam, lParam);
}

// VTraceHookSendMessageW is the hook for SendMessageW.  It logs the call.

LRESULT
WINAPI
VTraceHookSendMessageW(
    HWND hWnd,
    UINT Msg,
    WPARAM wParam,
    LPARAM lParam)
{
  char eventBuffer[1];

  // Indicate that the current thread is active, and log this indication
  // of thread activity only if it is important in the determination that
  // this thread is active.

  if (!WasThreadActiveBeforeNow()) {
    eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_SEND;
    LogEvent(eventBuffer, 1);
  }

  // Call the original function.

  return (* (SendMessage_PROC)
            hookInfo[SendMessageW_INDEX].origFunctionAddress)
            (hWnd, Msg, wParam, lParam);
}

// VTraceHookSendMessageTimeoutA is the hook for SendMessageTimeoutA.  It logs
// the call.

LRESULT
WINAPI
VTraceHookSendMessageTimeoutA(
    HWND hWnd,
    UINT Msg,
    WPARAM wParam,
    LPARAM lParam,
    UINT fuFlags,
    UINT uTimeout,
    LPDWORD lpdwResult)
{
  char eventBuffer[1];

  // Indicate that the current thread is active, and log this indication
  // of thread activity only if it is important in the determination that
  // this thread is active.

  if (!WasThreadActiveBeforeNow()) {
    eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_SEND;
    LogEvent(eventBuffer, 1);
  }

  // Call the original function.

  return (* (SendMessageTimeout_PROC)
            hookInfo[SendMessageTimeoutA_INDEX].origFunctionAddress)
            (hWnd, Msg, wParam, lParam, fuFlags, uTimeout, lpdwResult);
}

// VTraceHookSendMessageTimeoutW is the hook for SendMessageTimeoutW.  It logs
// the call.

LRESULT
WINAPI
VTraceHookSendMessageTimeoutW(
    HWND hWnd,
    UINT Msg,
    WPARAM wParam,
    LPARAM lParam,
    UINT fuFlags,
    UINT uTimeout,
    LPDWORD lpdwResult)
{
  char eventBuffer[1];

  // Indicate that the current thread is active, and log this indication
  // of thread activity only if it is important in the determination that
  // this thread is active.

  if (!WasThreadActiveBeforeNow()) {
    eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_SEND;
    LogEvent(eventBuffer, 1);
  }

  // Call the original function.

  return (* (SendMessageTimeout_PROC)
            hookInfo[SendMessageTimeoutW_INDEX].origFunctionAddress)
            (hWnd, Msg, wParam, lParam, fuFlags, uTimeout, lpdwResult);
}

// VTraceHookSendMessageCallbackA is the hook for SendMessageCallbackA.  It
// logs the call.

BOOL
WINAPI
VTraceHookSendMessageCallbackA(
    HWND hWnd,
    UINT Msg,
    WPARAM wParam,
    LPARAM lParam,
    SENDASYNCPROC lpResultCallBack,
    DWORD dwData)
{
  char eventBuffer[1];

  // Indicate that the current thread is active, and log this indication
  // of thread activity only if it is important in the determination that
  // this thread is active.

  if (!WasThreadActiveBeforeNow()) {
    eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_SEND;
    LogEvent(eventBuffer, 1);
  }

  // Call the original function.

  return (* (SendMessageCallback_PROC)
            hookInfo[SendMessageCallbackA_INDEX].origFunctionAddress)
            (hWnd, Msg, wParam, lParam, lpResultCallBack, dwData);
}

// VTraceHookSendMessageCallbackW is the hook for SendMessageCallbackW.  It
// logs the call.

BOOL
WINAPI
VTraceHookSendMessageCallbackW(
    HWND hWnd,
    UINT Msg,
    WPARAM wParam,
    LPARAM lParam,
    SENDASYNCPROC lpResultCallBack,
    DWORD dwData)
{
  char eventBuffer[1];

  // Indicate that the current thread is active, and log this indication
  // of thread activity only if it is important in the determination that
  // this thread is active.

  if (!WasThreadActiveBeforeNow()) {
    eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_SEND;
    LogEvent(eventBuffer, 1);
  }

  // Call the original function.

  return (* (SendMessageCallback_PROC)
            hookInfo[SendMessageCallbackW_INDEX].origFunctionAddress)
            (hWnd, Msg, wParam, lParam, lpResultCallBack, dwData);
}

// VTraceHookPostMessageA is the hook for PostMessageA.  It logs the call.

BOOL
WINAPI
VTraceHookPostMessageA(
    HWND hWnd,
    UINT Msg,
    WPARAM wParam,
    LPARAM lParam)
{
  char eventBuffer[17];

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the call.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_POST_WINDOW;
  memcpy(&eventBuffer[1], &hWnd, 4);
  memcpy(&eventBuffer[5], &Msg, 4);
  memcpy(&eventBuffer[9], &wParam, 4);
  memcpy(&eventBuffer[13], &lParam, 4);
  LogEvent(eventBuffer, 17);

  return (* (PostMessage_PROC)
            hookInfo[PostMessageA_INDEX].origFunctionAddress)
            (hWnd, Msg, wParam, lParam);
}

// VTraceHookPostMessageW is the hook for PostMessageW.  It logs the call.

BOOL
WINAPI
VTraceHookPostMessageW(
    HWND hWnd,
    UINT Msg,
    WPARAM wParam,
    LPARAM lParam)
{
  char eventBuffer[17];

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the call.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_POST_WINDOW;
  memcpy(&eventBuffer[1], &hWnd, 4);
  memcpy(&eventBuffer[5], &Msg, 4);
  memcpy(&eventBuffer[9], &wParam, 4);
  memcpy(&eventBuffer[13], &lParam, 4);
  LogEvent(eventBuffer, 17);

  return (* (PostMessage_PROC)
            hookInfo[PostMessageW_INDEX].origFunctionAddress)
            (hWnd, Msg, wParam, lParam);
}

// VTraceHookPostThreadMessageA is the hook for PostThreadMessageA.  It logs
// the call.

BOOL
WINAPI
VTraceHookPostThreadMessageA(
    DWORD idThread,
    UINT Msg,
    WPARAM wParam,
    LPARAM lParam)
{
  char eventBuffer[17];

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the call.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_POST_THREAD;
  memcpy(&eventBuffer[1], &idThread, 4);
  memcpy(&eventBuffer[5], &Msg, 4);
  memcpy(&eventBuffer[9], &wParam, 4);
  memcpy(&eventBuffer[13], &lParam, 4);
  LogEvent(eventBuffer, 17);

  return (* (PostThreadMessage_PROC)
            hookInfo[PostThreadMessageA_INDEX].origFunctionAddress)
            (idThread, Msg, wParam, lParam);
}

// VTraceHookPostThreadMessageW is the hook for PostThreadMessageW.  It logs
// the call.

BOOL
WINAPI
VTraceHookPostThreadMessageW(
    DWORD idThread,
    UINT Msg,
    WPARAM wParam,
    LPARAM lParam)
{
  char eventBuffer[17];

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the call.

  eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_POST_THREAD;
  memcpy(&eventBuffer[1], &idThread, 4);
  memcpy(&eventBuffer[5], &Msg, 4);
  memcpy(&eventBuffer[9], &wParam, 4);
  memcpy(&eventBuffer[13], &lParam, 4);
  LogEvent(eventBuffer, 17);

  return (* (PostThreadMessage_PROC)
            hookInfo[PostThreadMessageW_INDEX].origFunctionAddress)
            (idThread, Msg, wParam, lParam);
}

// VTraceHookSetTimer is the hook for SetTimer.  It logs the return if the
// call sets up a timer message.

UINT
WINAPI
VTraceHookSetTimer( 
    HWND hWnd, 
    UINT nIDEvent, 
    UINT uElapse, 
    TIMERPROC lpTimerFunc 
    )
{
  UINT retVal;
  char eventBuffer[17];

  // Call the original function, retaining the return value

  retVal = (* (SetTimer_PROC)
              hookInfo[SetTimer_INDEX].origFunctionAddress)
              (hWnd, nIDEvent, uElapse, lpTimerFunc);

  // If this is supposed to send a timer message to some window, log it

  if (retVal && (hWnd != NULL)) {
    eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_SET_TIMER;
    memcpy(&eventBuffer[1], &hWnd, 4);
    memcpy(&eventBuffer[5], &nIDEvent, 4);
    memcpy(&eventBuffer[9], &uElapse, 4);
    memcpy(&eventBuffer[13], &lpTimerFunc, 4);
    LogEvent(eventBuffer, 17);
  }

  return retVal;         // Return the return value obtained earlier
}

// VTraceHookKillTimer is the hook function for KillTimer.  It logs the return.

BOOL
WINAPI
VTraceHookKillTimer( 
    HWND hWnd, 
    UINT uIDEvent 
    )
{
  BOOL retVal;
  char eventBuffer[9];

  // Call the original function, retaining the return value

  retVal = (* (KillTimer_PROC)
              hookInfo[KillTimer_INDEX].origFunctionAddress)
              (hWnd, uIDEvent);

  // If this corresponds to cancelling a timer message to some window, log it

  if (retVal && (hWnd != NULL)) {
    eventBuffer[0] = (char) ENTRY_TYPE_MESSAGE_CANCEL_TIMER;
    memcpy(&eventBuffer[1], &hWnd, 4);
    memcpy(&eventBuffer[5], &uIDEvent, 4);
    LogEvent(eventBuffer, 9);
  }

  return retVal;         // Return the return value obtained earlier
}


////////////////////////////////////////////////////////////////////////////
//                           CURSOR FUNCTIONS                             //
////////////////////////////////////////////////////////////////////////////

// VTraceHookLoadCursorA is the hook for LoadCursorA.  It logs the return
// if the cursor type can be determined (i.e. if the MAKEINTRESOURCE
// technique was used to specify the cursor).  Also, if the cursor type
// can be determined, it makes sure that the waitCursor and appStartingCursor
// fields of the compaction per-process variables are always appropriately
// assigned to the current handles of those types of cursor.  It has to
// make some assumptions, however, such as that two handles are never
// opened to the same cursor, and that these cursors are only opened
// by the MAKEINTRESOURCE technique.

HCURSOR
WINAPI
VTraceHookLoadCursorA(
    HINSTANCE hInstance,
    LPCSTR lpCursorName)
{
  HCURSOR retVal;
  USHORT cursorType;
  char eventBuffer[7];

  // Call the original function, retaining the return value

  retVal = (* (LoadCursorA_PROC)
              hookInfo[LoadCursorA_INDEX].origFunctionAddress)
              (hInstance, lpCursorName);

  if ((hInstance == NULL) && retVal) {   // If the cursor was loaded, and it
                                         //   was specified by MAKEINTRESOURCE
    cursorType = LOWORD(lpCursorName);   //   then cursor type is the low word

    // Log the return.

    eventBuffer[0] = (char) ENTRY_TYPE_CURSOR_LOAD;
    memcpy(&eventBuffer[1], &retVal, 4);
    memcpy(&eventBuffer[5], &cursorType, 2);
    LogEvent(eventBuffer, 7);

    // If this is the IDC_WAIT cursor, then set the per-process variable
    // indicating what handle the wait cursor is to the returned handle.
    // Otherwise, make sure that the returned handle is not recorded in
    // this variable (as might happen if this handle had previously been
    // the wait cursor, and now is not).  Then, do the same thing for
    // IDC_APPSTARTING.

    if (cursorType == 32514)
      compactionProcessVars.waitCursor = retVal;
    else if (compactionProcessVars.waitCursor == retVal)
      compactionProcessVars.waitCursor = NULL;

    if (cursorType == 32650)
      compactionProcessVars.appStartingCursor = retVal;
    else if (compactionProcessVars.appStartingCursor == retVal)
      compactionProcessVars.appStartingCursor = NULL;
  }

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookLoadCursorW is the hook for LoadCursorW.  It logs the return
// if the cursor type can be determined (i.e. if the MAKEINTRESOURCE
// technique was used to specify the cursor).  Also, if the cursor type
// can be determined, it makes sure that the waitCursor and appStartingCursor
// fields of the compaction per-process variables are always appropriately
// assigned to the current handles of those types of cursor.  It has to
// make some assumptions, however, such as that two handles are never
// opened to the same cursor, and that these cursors are only opened
// by the MAKEINTRESOURCE technique.

HCURSOR
WINAPI
VTraceHookLoadCursorW(
    HINSTANCE hInstance,
    LPCWSTR lpCursorName)
{
  HCURSOR retVal;
  USHORT cursorType;
  char eventBuffer[7];

  // Call the original function, retaining the return value

  retVal = (* (LoadCursorW_PROC)
              hookInfo[LoadCursorW_INDEX].origFunctionAddress)
              (hInstance, lpCursorName);

  if ((hInstance == NULL) && retVal) {    // If the cursor was loaded, and was
                                          //   specified with MAKEINTRESOURCE,
    cursorType = LOWORD(lpCursorName);    //   then the low word is the type

    // Log the return.

    eventBuffer[0] = (char) ENTRY_TYPE_CURSOR_LOAD;
    memcpy(&eventBuffer[1], &retVal, 4);
    memcpy(&eventBuffer[5], &cursorType, 2);
    LogEvent(eventBuffer, 7);

    // If this is the IDC_WAIT cursor, then set the per-process variable
    // indicating what handle the wait cursor is to the returned handle.
    // Otherwise, make sure that the returned handle is not recorded in
    // this variable (as might happen if this handle had previously been
    // the wait cursor, and now is not).  Then, do the same thing for
    // IDC_APPSTARTING.

    if (cursorType == 32514)
      compactionProcessVars.waitCursor = retVal;
    else if (compactionProcessVars.waitCursor == retVal)
      compactionProcessVars.waitCursor = NULL;

    if (cursorType == 32650)
      compactionProcessVars.appStartingCursor = retVal;
    else if (compactionProcessVars.appStartingCursor == retVal)
      compactionProcessVars.appStartingCursor = NULL;
  }

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookLoadCursorFromFileA is the hook for LoadCursorFromFileA.
// It logs the return if the cursor type can be determined (i.e. if the
// MAKEINTRESOURCE technique was used to specify the cursor).  Also, if the
// cursor type can be determined, it makes sure that the waitCursor and
// appStartingCursor fields of the compaction per-process variables are
// always appropriately assigned to the current handles of those types of
// cursor.  It has to make some assumptions, however, such as that two
// handles are never opened to the same cursor, and that these cursors
// are only opened by the MAKEINTRESOURCE technique.

HCURSOR
WINAPI
VTraceHookLoadCursorFromFileA(
    LPCSTR lpFileName)
{
  HCURSOR retVal;
  USHORT cursorType;
  char eventBuffer[7];

  // Call the original function, retaining the return value

  retVal = (* (LoadCursorFromFileA_PROC)
              hookInfo[LoadCursorFromFileA_INDEX].origFunctionAddress)
              (lpFileName);

  if ((HIWORD(lpFileName) == 0) && retVal) {// If the cursor was loaded and was
                                            //   specified by MAKEINTRESOURCE,
    cursorType = LOWORD(lpFileName);        //   then the low word is the type

    // Log the return.

    eventBuffer[0] = (char) ENTRY_TYPE_CURSOR_LOAD;
    memcpy(&eventBuffer[1], &retVal, 4);
    memcpy(&eventBuffer[5], &cursorType, 2);
    LogEvent(eventBuffer, 7);

    // If this is the IDC_WAIT cursor, then set the per-process variable
    // indicating what handle the wait cursor is to the returned handle.
    // Otherwise, make sure that the returned handle is not recorded in
    // this variable (as might happen if this handle had previously been
    // the wait cursor, and now is not).  Then, do the same thing for
    // IDC_APPSTARTING.

    if (cursorType == 32514)
      compactionProcessVars.waitCursor = retVal;
    else if (compactionProcessVars.waitCursor == retVal)
      compactionProcessVars.waitCursor = NULL;

    if (cursorType == 32650)
      compactionProcessVars.appStartingCursor = retVal;
    else if (compactionProcessVars.appStartingCursor == retVal)
      compactionProcessVars.appStartingCursor = NULL;
  }

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookLoadCursorFromFileW is the hook for LoadCursorFromFileW.
// It logs the return if the cursor type can be determined (i.e. if the
// MAKEINTRESOURCE technique was used to specify the cursor).  Also, if the
// cursor type can be determined, it makes sure that the waitCursor and
// appStartingCursor fields of the compaction per-process variables are
// always appropriately assigned to the current handles of those types of
// cursor.  It has to make some assumptions, however, such as that two
// handles are never opened to the same cursor, and that these cursors
// are only opened by the MAKEINTRESOURCE technique.

HCURSOR
WINAPI
VTraceHookLoadCursorFromFileW(
    LPCWSTR lpFileName)
{
  HCURSOR retVal;
  USHORT cursorType;
  char eventBuffer[7];

  // Call the original function, retaining the return value

  retVal = (* (LoadCursorFromFileW_PROC)
              hookInfo[LoadCursorFromFileW_INDEX].origFunctionAddress)
              (lpFileName);

  if ((HIWORD(lpFileName) == 0) && retVal) {// If the cursor was loaded and was
                                            //   specified by MAKEINTRESOURCE,
    cursorType = LOWORD(lpFileName);        //   then the low word is the type

    // Log the return.

    eventBuffer[0] = (char) ENTRY_TYPE_CURSOR_LOAD;
    memcpy(&eventBuffer[1], &retVal, 4);
    memcpy(&eventBuffer[5], &cursorType, 2);
    LogEvent(eventBuffer, 7);

    // If this is the IDC_WAIT cursor, then set the per-process variable
    // indicating what handle the wait cursor is to the returned handle.
    // Otherwise, make sure that the returned handle is not recorded in
    // this variable (as might happen if this handle had previously been
    // the wait cursor, and now is not).  Then, do the same thing for
    // IDC_APPSTARTING.

    if (cursorType == 32514)
      compactionProcessVars.waitCursor = retVal;
    else if (compactionProcessVars.waitCursor == retVal)
      compactionProcessVars.waitCursor = NULL;

    if (cursorType == 32650)
      compactionProcessVars.appStartingCursor = retVal;
    else if (compactionProcessVars.appStartingCursor == retVal)
      compactionProcessVars.appStartingCursor = NULL;
  }

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookSetCursor is the hook for SetCursor.  It logs the return if
// the cursor was changed.  Also, if the cursor was changed, it ensures
// that the per-process variable cursorIsHourglass is set appropriately,
// according to whether the handle of the cursor set was one of the two
// pre-defined hourglass cursors.

HCURSOR
WINAPI
VTraceHookSetCursor(
    HCURSOR hCursor)
{
  HCURSOR retVal;
  char eventBuffer[5];

  // Call the original function, retaining the return value

  retVal = (* (SetCursor_PROC)
              hookInfo[SetCursor_INDEX].origFunctionAddress)
              (hCursor);

  if (retVal != hCursor) {            // If the cursor was changed, i.e. if
                                      //   the cursor set is different from
                                      //   the old cursor (which is in retVal)

    // Indicate that the current thread is active.

    IndicateThreadActivity();

    // Log the return.

    eventBuffer[0] = (char) ENTRY_TYPE_CURSOR_SET;
    memcpy(&eventBuffer[1], &hCursor, 4);
    LogEvent(eventBuffer, 5);

    // Set the cursorIsHourglass field according to whether the handle of
    // the cursor set is the wait cursor or the app-starting cursor

    compactionProcessVars.cursorIsHourglass =
       (hCursor == compactionProcessVars.waitCursor ||
        hCursor == compactionProcessVars.appStartingCursor) &&
       hCursor != NULL && hCursor != INVALID_HANDLE_VALUE;
  }

  return retVal;           // Return the return value obtained earlier
}

////////////////////////////////////////////////////////////////////////////
//                            WINDOW FUNCTIONS                            //
////////////////////////////////////////////////////////////////////////////

// VTraceHookCreateWindowExA is the hook for CreateWindowExA.  It logs
// the return.

HWND
WINAPI
VTraceHookCreateWindowExA(
    DWORD dwExStyle,
    LPCSTR lpClassName,
    LPCSTR lpWindowName,
    DWORD dwStyle,
    int X,
    int Y,
    int nWidth,
    int nHeight,
    HWND hWndParent,
    HMENU hMenu,
    HINSTANCE hInstance,
    LPVOID lpParam)
{
  HWND retVal;
  char eventBuffer[10];

  // Call the original function, retaining the return value

  retVal = (* (CreateWindowExA_PROC)
             hookInfo[CreateWindowExA_INDEX].origFunctionAddress)
             (dwExStyle, lpClassName, lpWindowName, dwStyle, X, Y, nWidth,
              nHeight, hWndParent, hMenu, hInstance, lpParam);

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_WINDOW_CREATE;
  eventBuffer[1] = (char) 10;
  memcpy(&eventBuffer[2], &retVal, 4);
  memcpy(&eventBuffer[6], &hWndParent, 4);
  LogEvent(eventBuffer, 10);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookCreateWindowExW is the hook for CreateWindowExW.  It logs
// the return.

HWND
WINAPI
VTraceHookCreateWindowExW(
    DWORD dwExStyle,
    LPCWSTR lpClassName,
    LPCWSTR lpWindowName,
    DWORD dwStyle,
    int X,
    int Y,
    int nWidth,
    int nHeight,
    HWND hWndParent ,
    HMENU hMenu,
    HINSTANCE hInstance,
    LPVOID lpParam)
{
  HWND retVal;
  char eventBuffer[10];

  // Call the original function, retaining the return value

  retVal = (* (CreateWindowExW_PROC)
             hookInfo[CreateWindowExW_INDEX].origFunctionAddress)
             (dwExStyle, lpClassName, lpWindowName, dwStyle, X, Y, nWidth,
              nHeight, hWndParent, hMenu, hInstance, lpParam);

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_WINDOW_CREATE;
  eventBuffer[1] = (char) 10;
  memcpy(&eventBuffer[2], &retVal, 4);
  memcpy(&eventBuffer[6], &hWndParent, 4);
  LogEvent(eventBuffer, 10);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookCreateDialogParamA is the hook for CreateDialogParamA.  It
// logs the return.

HWND
WINAPI
VTraceHookCreateDialogParamA(
    HINSTANCE hInstance,
    LPCSTR lpTemplateName,
    HWND hWndParent ,
    DLGPROC lpDialogFunc,
    LPARAM dwInitParam)
{
  HWND retVal;
  char eventBuffer[10];

  // Call the original function, retaining the return value

  retVal = (* (CreateDialogParamA_PROC)
             hookInfo[CreateDialogParamA_INDEX].origFunctionAddress)
             (hInstance, lpTemplateName, hWndParent, lpDialogFunc,
              dwInitParam);

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_DIALOG_CREATE;
  memcpy(&eventBuffer[1], &retVal, 4);
  memcpy(&eventBuffer[5], &hWndParent, 4);
  eventBuffer[9] = '\1';                    // Modeless
  LogEvent(eventBuffer, 10);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookCreateDialogParamW is the hook for CreateDialogParamW.  It
// logs the return.

HWND
WINAPI
VTraceHookCreateDialogParamW(
    HINSTANCE hInstance,
    LPCWSTR lpTemplateName,
    HWND hWndParent ,
    DLGPROC lpDialogFunc,
    LPARAM dwInitParam)
{
  HWND retVal;
  char eventBuffer[10];

  // Call the original function, retaining the return value

  retVal = (* (CreateDialogParamW_PROC)
             hookInfo[CreateDialogParamW_INDEX].origFunctionAddress)
             (hInstance, lpTemplateName, hWndParent, lpDialogFunc,
              dwInitParam);

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_DIALOG_CREATE;
  memcpy(&eventBuffer[1], &retVal, 4);
  memcpy(&eventBuffer[5], &hWndParent, 4);
  eventBuffer[9] = '\1';                    // Modeless
  LogEvent(eventBuffer, 10);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookCreateDialogIndirectParamA is the hook for
// CreateDialogIndirectParamA.  It logs the return.

HWND
WINAPI
VTraceHookCreateDialogIndirectParamA(
    HINSTANCE hInstance,
    LPCDLGTEMPLATEA lpTemplate,
    HWND hWndParent,
    DLGPROC lpDialogFunc,
    LPARAM dwInitParam)
{
  HWND retVal;
  char eventBuffer[10];

  // Call the original function, retaining the return value

  retVal = (* (CreateDialogIndirectParamA_PROC)
             hookInfo[CreateDialogIndirectParamA_INDEX].origFunctionAddress)
             (hInstance, lpTemplate, hWndParent, lpDialogFunc, dwInitParam);

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_DIALOG_CREATE;
  memcpy(&eventBuffer[1], &retVal, 4);
  memcpy(&eventBuffer[5], &hWndParent, 4);
  eventBuffer[9] = '\1';                    // Modeless
  LogEvent(eventBuffer, 10);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookCreateDialogIndirectParamW is the hook for
// CreateDialogIndirectParamW.  It logs the return.

HWND
WINAPI
VTraceHookCreateDialogIndirectParamW(
    HINSTANCE hInstance,
    LPCDLGTEMPLATEW lpTemplate,
    HWND hWndParent,
    DLGPROC lpDialogFunc,
    LPARAM dwInitParam)
{
  HWND retVal;
  char eventBuffer[10];

  // Call the original function, retaining the return value

  retVal = (* (CreateDialogIndirectParamW_PROC)
             hookInfo[CreateDialogIndirectParamW_INDEX].origFunctionAddress)
             (hInstance, lpTemplate, hWndParent, lpDialogFunc, dwInitParam);

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_DIALOG_CREATE;
  memcpy(&eventBuffer[1], &retVal, 4);
  memcpy(&eventBuffer[5], &hWndParent, 4);
  eventBuffer[9] = '\1';                    // Modeless
  LogEvent(eventBuffer, 10);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookCreateDialogIndirectParamAorW is the hook for
// CreateDialogIndirectParamAorW.  It logs the return.

HWND
WINAPI
VTraceHookCreateDialogIndirectParamAorW(
    HINSTANCE hInstance,
    LPCDLGTEMPLATE lpTemplate,
    HWND hWndParent,
    DLGPROC lpDialogFunc,
    LPARAM dwInitParam,
    DWORD unknownParameter)
{
  HWND retVal;
  char eventBuffer[10];

  // Call the original function, retaining the return value

  retVal = (* (CreateDialogIndirectParamAorW_PROC)
             hookInfo[CreateDialogIndirectParamAorW_INDEX].origFunctionAddress)
             (hInstance, lpTemplate, hWndParent, lpDialogFunc, dwInitParam,
              unknownParameter);

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_DIALOG_CREATE;
  memcpy(&eventBuffer[1], &retVal, 4);
  memcpy(&eventBuffer[5], &hWndParent, 4);
  eventBuffer[9] = '\1';                    // Modeless
  LogEvent(eventBuffer, 10);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookDialogBoxParamA is the hook for DialogBoxParamA.  It logs the
// return.

int
WINAPI
VTraceHookDialogBoxParamA(
    HINSTANCE hInstance,
    LPCSTR lpTemplateName,
    HWND hWndParent,
    DLGPROC lpDialogFunc,
    LPARAM dwInitParam)
{
  int retVal;
  char eventBuffer[10];

  // Call the original function, retaining the return value

  retVal = (* (DialogBoxParamA_PROC)
             hookInfo[DialogBoxParamA_INDEX].origFunctionAddress)
             (hInstance, lpTemplateName, hWndParent, lpDialogFunc,
              dwInitParam);

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_DIALOG_CREATE;
  memcpy(&eventBuffer[1], &retVal, 4);
  memcpy(&eventBuffer[5], &hWndParent, 4);
  eventBuffer[9] = '\0';                    // Modal
  LogEvent(eventBuffer, 10);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookDialogBoxParamW is the hook for DialogBoxParamW.  It logs the
// return.

int
WINAPI
VTraceHookDialogBoxParamW(
    HINSTANCE hInstance,
    LPCWSTR lpTemplateName,
    HWND hWndParent,
    DLGPROC lpDialogFunc,
    LPARAM dwInitParam)
{
  int retVal;
  char eventBuffer[10];

  // Call the original function, retaining the return value

  retVal = (* (DialogBoxParamW_PROC)
             hookInfo[DialogBoxParamW_INDEX].origFunctionAddress)
             (hInstance, lpTemplateName, hWndParent, lpDialogFunc,
              dwInitParam);

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_DIALOG_CREATE;
  memcpy(&eventBuffer[1], &retVal, 4);
  memcpy(&eventBuffer[5], &hWndParent, 4);
  eventBuffer[9] = '\0';                    // Modal
  LogEvent(eventBuffer, 10);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookDialogBoxIndirectParamA is the hook for DialogBoxIndirectParamA.
// It logs the return.

int
WINAPI
VTraceHookDialogBoxIndirectParamA(
    HINSTANCE hInstance,
    LPCDLGTEMPLATEA hDialogTemplate,
    HWND hWndParent,
    DLGPROC lpDialogFunc,
    LPARAM dwInitParam)
{
  int retVal;
  char eventBuffer[10];

  // Call the original function, retaining the return value

  retVal = (* (DialogBoxIndirectParamA_PROC)
             hookInfo[DialogBoxIndirectParamA_INDEX].origFunctionAddress)
             (hInstance, hDialogTemplate, hWndParent, lpDialogFunc,
              dwInitParam);

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_DIALOG_CREATE;
  memcpy(&eventBuffer[1], &retVal, 4);
  memcpy(&eventBuffer[5], &hWndParent, 4);
  eventBuffer[9] = '\0';                    // Modal
  LogEvent(eventBuffer, 10);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookDialogBoxIndirectParamW is the hook for DialogBoxIndirectParamW.
// It logs the return.

int
WINAPI
VTraceHookDialogBoxIndirectParamW(
    HINSTANCE hInstance,
    LPCDLGTEMPLATEW hDialogTemplate,
    HWND hWndParent,
    DLGPROC lpDialogFunc,
    LPARAM dwInitParam)
{
  int retVal;
  char eventBuffer[10];

  // Call the original function, retaining the return value

  retVal = (* (DialogBoxIndirectParamW_PROC)
             hookInfo[DialogBoxIndirectParamW_INDEX].origFunctionAddress)
             (hInstance, hDialogTemplate, hWndParent, lpDialogFunc,
              dwInitParam);

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_DIALOG_CREATE;
  memcpy(&eventBuffer[1], &retVal, 4);
  memcpy(&eventBuffer[5], &hWndParent, 4);
  eventBuffer[9] = '\0';                    // Modal
  LogEvent(eventBuffer, 10);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookDialogBoxIndirectParamAorW is the hook for
// DialogBoxIndirectParamAorW.  It logs the return.

int
WINAPI
VTraceHookDialogBoxIndirectParamAorW(
    HINSTANCE hInstance,
    LPCDLGTEMPLATEA hDialogTemplate,
    HWND hWndParent,
    DLGPROC lpDialogFunc,
    LPARAM dwInitParam,
    DWORD unknownParameter)
{
  int retVal;
  char eventBuffer[10];

  // Call the original function, retaining the return value

  retVal = (* (DialogBoxIndirectParamAorW_PROC)
             hookInfo[DialogBoxIndirectParamAorW_INDEX].origFunctionAddress)
             (hInstance, hDialogTemplate, hWndParent, lpDialogFunc,
              dwInitParam, unknownParameter);

  // Indicate that the current thread is active.

  IndicateThreadActivity();

  // Log the return.

  eventBuffer[0] = (char) ENTRY_TYPE_DIALOG_CREATE;
  memcpy(&eventBuffer[1], &retVal, 4);
  memcpy(&eventBuffer[5], &hWndParent, 4);
  eventBuffer[9] = '\0';                    // Modal
  LogEvent(eventBuffer, 10);

  return retVal;           // Return the return value obtained earlier
}

////////////////////////////////////////////////////////////////////////////
//                        LIBRARY-LOAD FUNCTIONS                          //
////////////////////////////////////////////////////////////////////////////

// VTraceHookLoadLibraryA is the hook for LoadLibraryA.  After the return,
// if the library load was successful, it installs all the hooks into all
// loaded modules.

HMODULE
WINAPI
VTraceHookLoadLibraryA(
    LPCSTR lpLibFileName)
{
  HMODULE retVal;

  // Call the original function, retaining the return value

  retVal = (* (LoadLibraryA_PROC)
              hookInfo[LoadLibraryA_INDEX].origFunctionAddress)
              (lpLibFileName);

  // If the call is successful, call InstallHooksInModule for each module
  // loaded in the current process.

  if (retVal != NULL)
    EnumerateLoadedModules(GetCurrentProcess(), &InstallHooksInModule, NULL);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookLoadLibraryW is the hook for LoadLibraryW.  After the return,
// if the library load was successful, it installs all the hooks into all
// loaded modules.

HMODULE
WINAPI
VTraceHookLoadLibraryW(
    LPCWSTR lpLibFileName)
{
  HMODULE retVal;

  // Call the original function, retaining the return value

  retVal = (* (LoadLibraryW_PROC)
              hookInfo[LoadLibraryW_INDEX].origFunctionAddress)
              (lpLibFileName);

  // If the call is successful, call InstallHooksInModule for each module
  // loaded in the current process.

  if (retVal != NULL)
    EnumerateLoadedModules(GetCurrentProcess(), &InstallHooksInModule, NULL);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookLoadLibraryExA is the hook for LoadLibraryExA.  After the return,
// if the library load was successful, it installs all the hooks into all
// loaded modules.

HMODULE
WINAPI
VTraceHookLoadLibraryExA(
    LPCSTR lpLibFileName,
    HANDLE hFile,
    DWORD dwFlags
    )
{
  HMODULE retVal;

  // Call the original function, retaining the return value

  retVal = (* (LoadLibraryExA_PROC)
              hookInfo[LoadLibraryExA_INDEX].origFunctionAddress)
              (lpLibFileName, hFile, dwFlags);

  // If the call is successful, call InstallHooksInModule for each module
  // loaded in the current process.

  if (retVal != NULL)
    EnumerateLoadedModules(GetCurrentProcess(), &InstallHooksInModule, NULL);

  return retVal;           // Return the return value obtained earlier
}

// VTraceHookLoadLibraryExW is the hook for LoadLibraryExW.  After the return,
// if the library load was successful, it installs all the hooks into all
// loaded modules.

HMODULE
WINAPI
VTraceHookLoadLibraryExW(
    LPCWSTR lpLibFileName,
    HANDLE hFile,
    DWORD dwFlags
    )
{
  HMODULE retVal;

  // Call the original function, retaining the return value

  retVal = (* (LoadLibraryExW_PROC)
              hookInfo[LoadLibraryExW_INDEX].origFunctionAddress)
              (lpLibFileName, hFile, dwFlags);

  // If the call is successful, call InstallHooksInModule for each module
  // loaded in the current process.

  if (retVal != NULL)
    EnumerateLoadedModules(GetCurrentProcess(), &InstallHooksInModule, NULL);

  return retVal;           // Return the return value obtained earlier
}
