//
// HookManage.cpp contains the functions for managing the initialization
// and insertion of DLL hook functions.
//

#include <windows.h>
#include "../Common/VTrace.h"
#include "Debug.h"
#include "Globals.h"
#include "HookFuncs.h"
#include "HookManage.h"

////////////////////////////////////////////////////////////////////////////
//                               MACROS
////////////////////////////////////////////////////////////////////////////

// The MAX_PROTECTION_PAGES macro is the maximum number of consecutive
// pages we allow ourselves to change the protections on at a time

#define MAX_PROTECTION_PAGES             4

// MakePtr is a macro that allows you to easily add to values (including
// pointers) together without dealing with C's pointer arithmetic.  It
// essentially treats the last two parameters as DWORDs.  The first
// parameter is used to typecast the result to the appropriate pointer type.

#define MakePtr(cast, ptr, addValue) (cast)((DWORD)(ptr)+(DWORD)(addValue))

////////////////////////////////////////////////////////////////////////////
//                              STRUCTURES
////////////////////////////////////////////////////////////////////////////

// The protection info structure contains data about the original
// protections of a set of pages that have had their protections
// temporarily changed by this program

typedef struct _protection_info {
  int pagesChanged;                               // # of pages changed
  PVOID pagePtr[MAX_PROTECTION_PAGES];            // their addresses
  DWORD pageOrigProtection[MAX_PROTECTION_PAGES]; // their original protections
} PROTECTION_INFO;

////////////////////////////////////////////////////////////////////////////
//                       PAGE PROTECTION FUNCTIONS
////////////////////////////////////////////////////////////////////////////

// RestoreProtections restores protections to as they were "before",
// where "before" is as described by the protection info structure
// passed.

BOOL RestoreProtections (PROTECTION_INFO *info)
{
  int i;
  BOOL retVal = TRUE;
  DWORD oldProtections;

  for (i = 0; i < info->pagesChanged; i++) {
    if (!VirtualProtect(info->pagePtr[i], pageSize,
                        info->pageOrigProtection[i], &oldProtections))
      retVal = FALSE;
  }

  return retVal;
}

// SetRWXProtection sets the virtual memory protection on a set of
// pages to read/write/execute.  The set of pages is the smallest set
// containing the "size" bytes pointed to by "ptr".  The passed
// protection info structure pointer "info" is used to record which
// pages were re-protected and what their original protections were.
// (Thus, it can be used in a later call to RestoreProtections to
// undo everything that this function did.)  This function returns FALSE
// if it cannot perform the requested function; if it does so, it undoes
// whatever protection changes it already made.  This function will not
// change the protections on more than MAX_PROTECTION_PAGES pages.

BOOL SetRWXProtection (PVOID ptr, DWORD size, PROTECTION_INFO *info)
{
  DWORD ptrAsDword = (DWORD) ptr;
  PVOID curPtr = (PVOID) (ptrAsDword - (ptrAsDword % pageSize));
  PVOID endPtr = MakePtr(PVOID, ptr, size);
  DWORD oldProtections;

  // Change protections on pages, one at a time, starting with the one
  // containing ptr, and going until either MAX_PROTECTION_PAGES have
  // been changed or ptr+size-1 has been changed

  info->pagesChanged = 0;
  while (info->pagesChanged < MAX_PROTECTION_PAGES && curPtr < endPtr) {
    if (VirtualProtect(curPtr, pageSize, PAGE_EXECUTE_READWRITE,
                       &oldProtections)) {
      info->pagePtr[info->pagesChanged] = curPtr;
      info->pageOrigProtection[info->pagesChanged] = oldProtections;
      info->pagesChanged++;
      curPtr = MakePtr(PVOID, curPtr, pageSize);
    }
    else
      break;
  }

  // If ptr+size-1 has been changed, return success; otherwise, undo any
  // protection changes we've made and return failure

  if (curPtr >= endPtr)
    return TRUE;
  else {
    RestoreProtections(info);
    return FALSE;
  }
}

////////////////////////////////////////////////////////////////////////////
//                       HOOK INFO HASH TABLE FUNCTIONS
////////////////////////////////////////////////////////////////////////////

// EmptyOrigHashTable empties the original-address hash table.

void EmptyOrigHashTable (void)
{
  int i;

  for (i = 0; i < NUMBER_OF_HOOK_HASH_BUCKETS; i++)
    hookOrigHashBucket[i] = NULL;
}

// AddToOrigHashTable adds a hook info structure to the original-address hash
// table.

void AddToOrigHashTable (HOOKED_FUNCTION_INFO *info)
{
  int bucket;
  
  bucket = ((DWORD) info->origFunctionAddress) % NUMBER_OF_HOOK_HASH_BUCKETS;
  info->nextInOrigHash = hookOrigHashBucket[bucket];
  hookOrigHashBucket[bucket] = info;
}

// FindInOrigHashTable looks for an original-address hash table entry whose
// original-function address element matches the given search address.
// If it does so, it returns a pointer to that entry; otherwise, it
// returns NULL.

HOOKED_FUNCTION_INFO *FindInOrigHashTable (PDWORD searchAddress)
{
  int bucket;
  HOOKED_FUNCTION_INFO *curPtr;
  
  bucket = ((DWORD) searchAddress) % NUMBER_OF_HOOK_HASH_BUCKETS;
  for (curPtr = hookOrigHashBucket[bucket];
       curPtr != NULL && curPtr->origFunctionAddress != searchAddress;
       curPtr = curPtr->nextInOrigHash)
    ;

  return curPtr;
}

// EmptyNewHashTable empties the new-address hash table.

void EmptyNewHashTable (void)
{
  int i;

  for (i = 0; i < NUMBER_OF_HOOK_HASH_BUCKETS; i++)
    hookNewHashBucket[i] = NULL;
}

// AddToNewHashTable adds a hook info structure to the new-address hash table.

void AddToNewHashTable (HOOKED_FUNCTION_INFO *info)
{
  int bucket;
  
  bucket = ((DWORD) info->newFunctionAddress) % NUMBER_OF_HOOK_HASH_BUCKETS;
  info->nextInNewHash = hookNewHashBucket[bucket];
  hookNewHashBucket[bucket] = info;
}

// FindInNewHashTable looks for a new-address hash table entry whose
// hook-function address element matches the given search address.
// If it does so, it returns a pointer to that entry; otherwise, it
// returns NULL.

HOOKED_FUNCTION_INFO *FindInNewHashTable (PDWORD searchAddress)
{
  int bucket;
  HOOKED_FUNCTION_INFO *curPtr;
  
  bucket = ((DWORD) searchAddress) % NUMBER_OF_HOOK_HASH_BUCKETS;
  for (curPtr = hookNewHashBucket[bucket];
       curPtr != NULL && curPtr->newFunctionAddress != searchAddress;
       curPtr = curPtr->nextInNewHash)
    ;

  return curPtr;
}

////////////////////////////////////////////////////////////////////////////
//                        HOOK INFORMATION FUNCTIONS
////////////////////////////////////////////////////////////////////////////

// InitStaticHookInfo initializes the information in hookInfo that can be
// determined statically (e.g. at compile time).

BOOL InitStaticHookInfo (void)
{
  hookInfo[GetMessageA_INDEX].name = "GetMessageA";
  hookInfo[GetMessageA_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookGetMessageA;

  hookInfo[GetMessageW_INDEX].name = "GetMessageW";
  hookInfo[GetMessageW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookGetMessageW;

  hookInfo[PeekMessageA_INDEX].name = "PeekMessageA";
  hookInfo[PeekMessageA_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookPeekMessageA;

  hookInfo[PeekMessageW_INDEX].name = "PeekMessageW";
  hookInfo[PeekMessageW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookPeekMessageW;

  hookInfo[MsgWaitForMultipleObjects_INDEX].name = "MsgWaitForMultipleObjects";
  hookInfo[MsgWaitForMultipleObjects_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookMsgWaitForMultipleObjects;

  hookInfo[MsgWaitForMultipleObjectsEx_INDEX].name =
    "MsgWaitForMultipleObjectsEx";
  hookInfo[MsgWaitForMultipleObjectsEx_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookMsgWaitForMultipleObjectsEx;

  hookInfo[WaitMessage_INDEX].name = "WaitMessage";
  hookInfo[WaitMessage_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookWaitMessage;

  hookInfo[GetInputState_INDEX].name = "GetInputState";
  hookInfo[GetInputState_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookGetInputState;

  hookInfo[GetQueueStatus_INDEX].name = "GetQueueStatus";
  hookInfo[GetQueueStatus_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookGetQueueStatus;

  hookInfo[DispatchMessageA_INDEX].name = "DispatchMessageA";
  hookInfo[DispatchMessageA_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookDispatchMessageA;

  hookInfo[DispatchMessageW_INDEX].name = "DispatchMessageW";
  hookInfo[DispatchMessageW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookDispatchMessageW;

  hookInfo[TranslateAccelerator_INDEX].name = "TranslateAccelerator";
  hookInfo[TranslateAccelerator_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookTranslateAccelerator;

  hookInfo[TranslateAcceleratorA_INDEX].name = "TranslateAcceleratorA";
  hookInfo[TranslateAcceleratorA_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookTranslateAcceleratorA;

  hookInfo[TranslateAcceleratorW_INDEX].name = "TranslateAcceleratorW";
  hookInfo[TranslateAcceleratorW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookTranslateAcceleratorW;

  hookInfo[TranslateMessage_INDEX].name = "TranslateMessage";
  hookInfo[TranslateMessage_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookTranslateMessage;

  hookInfo[TranslateMDISysAccel_INDEX].name = "TranslateMDISysAccel";
  hookInfo[TranslateMDISysAccel_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookTranslateMDISysAccel;

  hookInfo[SendMessageA_INDEX].name = "SendMessageA";
  hookInfo[SendMessageA_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookSendMessageA;

  hookInfo[SendMessageW_INDEX].name = "SendMessageW";
  hookInfo[SendMessageW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookSendMessageW;

  hookInfo[SendMessageTimeoutA_INDEX].name = "SendMessageTimeoutA";
  hookInfo[SendMessageTimeoutA_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookSendMessageTimeoutA;

  hookInfo[SendMessageTimeoutW_INDEX].name = "SendMessageTimeoutW";
  hookInfo[SendMessageTimeoutW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookSendMessageTimeoutW;

  hookInfo[SendMessageCallbackA_INDEX].name = "SendMessageCallbackA";
  hookInfo[SendMessageCallbackA_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookSendMessageCallbackA;

  hookInfo[SendMessageCallbackW_INDEX].name = "SendMessageCallbackW";
  hookInfo[SendMessageCallbackW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookSendMessageCallbackW;

  hookInfo[PostMessageA_INDEX].name = "PostMessageA";
  hookInfo[PostMessageA_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookPostMessageA;

  hookInfo[PostMessageW_INDEX].name = "PostMessageW";
  hookInfo[PostMessageW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookPostMessageW;

  hookInfo[PostThreadMessageA_INDEX].name = "PostThreadMessageA";
  hookInfo[PostThreadMessageA_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookPostThreadMessageA;

  hookInfo[PostThreadMessageW_INDEX].name = "PostThreadMessageW";
  hookInfo[PostThreadMessageW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookPostThreadMessageW;

  hookInfo[SetTimer_INDEX].name = "SetTimer";
  hookInfo[SetTimer_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookSetTimer;

  hookInfo[KillTimer_INDEX].name = "KillTimer";
  hookInfo[KillTimer_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookKillTimer;

  hookInfo[LoadCursorA_INDEX].name = "LoadCursorA";
  hookInfo[LoadCursorA_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookLoadCursorA;

  hookInfo[LoadCursorW_INDEX].name = "LoadCursorW";
  hookInfo[LoadCursorW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookLoadCursorW;

  hookInfo[LoadCursorFromFileA_INDEX].name = "LoadCursorFromFileA";
  hookInfo[LoadCursorFromFileA_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookLoadCursorFromFileA;

  hookInfo[LoadCursorFromFileW_INDEX].name = "LoadCursorFromFileW";
  hookInfo[LoadCursorFromFileW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookLoadCursorFromFileW;

  hookInfo[SetCursor_INDEX].name = "SetCursor";
  hookInfo[SetCursor_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookSetCursor;

  hookInfo[CreateWindowExA_INDEX].name = "CreateWindowExA";
  hookInfo[CreateWindowExA_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookCreateWindowExA;

  hookInfo[CreateWindowExW_INDEX].name = "CreateWindowExW";
  hookInfo[CreateWindowExW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookCreateWindowExW;

  hookInfo[CreateDialogParamA_INDEX].name = "CreateDialogParamA";
  hookInfo[CreateDialogParamA_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookCreateDialogParamA;

  hookInfo[CreateDialogParamW_INDEX].name = "CreateDialogParamW";
  hookInfo[CreateDialogParamW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookCreateDialogParamW;

  hookInfo[CreateDialogIndirectParamA_INDEX].name =
    "CreateDialogIndirectParamA";
  hookInfo[CreateDialogIndirectParamA_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookCreateDialogIndirectParamA;

  hookInfo[CreateDialogIndirectParamW_INDEX].name =
    "CreateDialogIndirectParamW";
  hookInfo[CreateDialogIndirectParamW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookCreateDialogIndirectParamW;

  hookInfo[CreateDialogIndirectParamAorW_INDEX].name =
    "CreateDialogIndirectParamAorW";
  hookInfo[CreateDialogIndirectParamAorW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookCreateDialogIndirectParamAorW;

  hookInfo[DialogBoxParamA_INDEX].name = "DialogBoxParamA";
  hookInfo[DialogBoxParamA_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookDialogBoxParamA;

  hookInfo[DialogBoxParamW_INDEX].name = "DialogBoxParamW";
  hookInfo[DialogBoxParamW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookDialogBoxParamW;

  hookInfo[DialogBoxIndirectParamA_INDEX].name = "DialogBoxIndirectParamA";
  hookInfo[DialogBoxIndirectParamA_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookDialogBoxIndirectParamA;

  hookInfo[DialogBoxIndirectParamW_INDEX].name = "DialogBoxIndirectParamW";
  hookInfo[DialogBoxIndirectParamW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookDialogBoxIndirectParamW;

  hookInfo[DialogBoxIndirectParamAorW_INDEX].name =
    "DialogBoxIndirectParamAorW";
  hookInfo[DialogBoxIndirectParamAorW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookDialogBoxIndirectParamAorW;

  hookInfo[LoadLibraryA_INDEX].name = "LoadLibraryA";
  hookInfo[LoadLibraryA_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookLoadLibraryA;

  hookInfo[LoadLibraryW_INDEX].name = "LoadLibraryW";
  hookInfo[LoadLibraryW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookLoadLibraryW;

  hookInfo[LoadLibraryExA_INDEX].name = "LoadLibraryExA";
  hookInfo[LoadLibraryExA_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookLoadLibraryExA;

  hookInfo[LoadLibraryExW_INDEX].name = "LoadLibraryExW";
  hookInfo[LoadLibraryExW_INDEX].newFunctionAddress =
    (PDWORD) &VTraceHookLoadLibraryExW;

  return TRUE;
}

// InitHookInfo initializes all the hook information.  In other words,
// it fills out the hookInfo array with all the names, original addresses,
// and hooked addresses of the hooked functions.  It also puts all the
// hookInfo structures into the hash tables and records the page size in
// the global variable pageSize.  It returns a boolean indicating whether
// it was successful.

BOOL InitHookInfo (void)
{
  int i;
  HMODULE hModUser32Library, hModKernel32Library;
  SYSTEM_INFO systemInfo;

  GetSystemInfo(&systemInfo);
  pageSize = systemInfo.dwPageSize;
  if (pageSize == 0) return FALSE;

  hModUser32Library = GetModuleHandle("USER32.DLL");
  hModKernel32Library = GetModuleHandle("KERNEL32.DLL");
  if (hModUser32Library != NULL && hModKernel32Library != NULL) {
    InitStaticHookInfo();
    EmptyOrigHashTable();
    EmptyNewHashTable();
    for (i = 0; i < NUMBER_OF_HOOKED_FUNCTIONS; i++) {
      if (i < NUMBER_OF_USER32_HOOKED_FUNCTIONS)
        hookInfo[i].origFunctionAddress =
          (PDWORD) GetProcAddress(hModUser32Library, hookInfo[i].name);
      else
        hookInfo[i].origFunctionAddress =
          (PDWORD) GetProcAddress(hModKernel32Library, hookInfo[i].name);
      AddToOrigHashTable(&hookInfo[i]);
      AddToNewHashTable(&hookInfo[i]);
    }
    return TRUE;
  }

  return FALSE;
}

// InstallHooksInModule installs the hook functions in the given module
// (at base memory location ModuleBase).  It has the function prototype
// it does to make it useful as a callback in EnumerateLoadedModules.
// It always returns TRUE to ensure that EnumerateLoadedModules does not
// stop calling it for subsequently found modules.  (It's not clear that
// returning FALSE would make it stop; this is just paranoia.)

BOOL CALLBACK InstallHooksInModule (LPSTR ModuleName, ULONG ModuleBase,
                                    ULONG ModuleSize, PVOID UserContext)
{
  PIMAGE_DOS_HEADER pDOSHeader = (PIMAGE_DOS_HEADER) ModuleBase;
  PIMAGE_NT_HEADERS pNTHeader;
  PIMAGE_IMPORT_DESCRIPTOR pImportDesc;
  PIMAGE_THUNK_DATA pFirstThunk, pThunk;
  PIMAGE_DATA_DIRECTORY pImportDataDirectory;
  PCHAR pImportedDLLName;
  PHOOKED_FUNCTION_INFO info;
  PROTECTION_INFO protectionInfo;
  int numberOfThunks;
  BOOL hookLoadLibraryA;

  // For some reason, Age of Empires II can't stand to have LoadLibraryA
  // hooked in its main module. Presumably, this is because they perform
  // safety checks to make sure that their copy-protection library isn't
  // circumvented. So, if this is empires2.exe, we don't hook LoadLibraryA.

  hookLoadLibraryA = (ModuleName == NULL ||
                      strcmp(ModuleName, "empires2.exe"));

  PrintDebugInfo("Installing hooks in module %s\n", ModuleName, 0);

  // Ensure we can read at least the DOS header of this module

  if (IsBadReadPtr((PVOID) ModuleBase, sizeof(IMAGE_DOS_HEADER)))
    return TRUE;

  // If this is not a DOS signature, bail out

  if (pDOSHeader->e_magic != IMAGE_DOS_SIGNATURE)
    return TRUE;

  // Get the NT header, and make sure its signature is correct

  pNTHeader = MakePtr(PIMAGE_NT_HEADERS, pDOSHeader, pDOSHeader->e_lfanew);
  if (pNTHeader->Signature != IMAGE_NT_SIGNATURE)
    return TRUE;

  // Find the data directory for imports

  pImportDataDirectory = &pNTHeader->OptionalHeader.
                          DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT];

  // Bail out if there is no imports section

  if (pImportDataDirectory->VirtualAddress == 0)
    return TRUE;

  // Loop through all the import descriptors, each of which describes
  // the imports to this module from a given other DLL
    
  for (pImportDesc = MakePtr(PIMAGE_IMPORT_DESCRIPTOR, ModuleBase, 
                             pImportDataDirectory->VirtualAddress);
       pImportDesc < MakePtr(PIMAGE_IMPORT_DESCRIPTOR, ModuleBase,
                             ModuleSize) &&
       pImportDesc->Name; pImportDesc++) {

    // Get the name of the DLL this descriptor is describing, then
    // skip to the next descriptor if it the name is not user32.dll
    // or kernel32.dll (doing a case-insensitive match)

    pImportedDLLName = MakePtr(PCHAR, ModuleBase, pImportDesc->Name);
    if (_stricmp(pImportedDLLName, "user32.dll") &&
        _stricmp(pImportedDLLName, "kernel32.dll"))
      continue;

    // Get the first "thunk" describing the first function imported
    // from this DLL

    pFirstThunk = MakePtr(PIMAGE_THUNK_DATA,
                          ModuleBase, pImportDesc->FirstThunk);

    // Loop through all the thunks to see how many there are, so that
    // we can set the page protection on all those thunks to read/
    // write/execute.

    for (pThunk = pFirstThunk, numberOfThunks = 0; pThunk->u1.Function;
         pThunk++, numberOfThunks++)
      ;

    // Set the page protection on the numberOfThunks thunks to
    // read/write/execute.  If we cannot, skip to the next import
    // descriptor.

    if (!SetRWXProtection(pFirstThunk,
                          numberOfThunks * sizeof(IMAGE_THUNK_DATA),
                          &protectionInfo))
      continue;

    // Loop through all the thunks, checking each function to see
    // if we have a matching hook information structure in the hash
    // table.  If we do, we replace the function pointer with the
    // hook function pointer.

    for (pThunk = pFirstThunk; pThunk->u1.Function; pThunk++) {
      info = FindInOrigHashTable(pThunk->u1.Function);
      if (info != NULL) {
        if (hookLoadLibraryA || strcmp(info->name, "LoadLibraryA")) {
          pThunk->u1.Function = info->newFunctionAddress;
        }
      }
    }

    // Restore virtual memory protections on the thunk pages to
    // what they were originally

    RestoreProtections(&protectionInfo);
  }

  PrintDebugInfo("Successfully installed hooks in module %s\n", ModuleName, 0);

  return TRUE;
}

// UninstallHooksInModule uninstalls the hook functions in the given module
// (at base memory location ModuleBase).  It has the function prototype
// it does to make it useful as a callback in EnumerateLoadedModules.
// It always returns TRUE to ensure that EnumerateLoadedModules does not
// stop calling it for subsequently found modules.  (It's not clear that
// returning FALSE would make it stop; this is just paranoia.)

BOOL CALLBACK UninstallHooksInModule (LPSTR ModuleName, ULONG ModuleBase,
                                      ULONG ModuleSize, PVOID UserContext)
{
  PIMAGE_DOS_HEADER pDOSHeader = (PIMAGE_DOS_HEADER) ModuleBase;
  PIMAGE_NT_HEADERS pNTHeader;
  PIMAGE_IMPORT_DESCRIPTOR pImportDesc;
  PIMAGE_THUNK_DATA pFirstThunk, pThunk;
  PIMAGE_DATA_DIRECTORY pImportDataDirectory;
  PCHAR pImportedDLLName;
  PHOOKED_FUNCTION_INFO info;
  PROTECTION_INFO protectionInfo;
  int numberOfThunks;

  // Ensure we can read at least the DOS header of this module

  if (IsBadReadPtr((PVOID) ModuleBase, sizeof(IMAGE_DOS_HEADER)))
    return TRUE;

  // If this is not a DOS signature, bail out
   
  if (pDOSHeader->e_magic != IMAGE_DOS_SIGNATURE)
    return TRUE;

  // Get the NT header, and make sure its signature is correct

  pNTHeader = MakePtr(PIMAGE_NT_HEADERS, pDOSHeader, pDOSHeader->e_lfanew);
  if (pNTHeader->Signature != IMAGE_NT_SIGNATURE)
    return TRUE;

  // Find the data directory for imports

  pImportDataDirectory = &pNTHeader->OptionalHeader.
                          DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT];

  // Bail out if there is no imports section

  if (pImportDataDirectory->VirtualAddress == 0)
    return TRUE;

  // Loop through all the import descriptors, each of which describes
  // the imports to this module from a given other DLL
    
  for (pImportDesc = MakePtr(PIMAGE_IMPORT_DESCRIPTOR, ModuleBase, 
                             pImportDataDirectory->VirtualAddress);
       pImportDesc->Name; pImportDesc++) {

    // Get the name of the DLL this descriptor is describing, then
    // skip to the next descriptor if it the name is not user32.dll
    // or kernel32.dll (doing a case-insensitive match)

    pImportedDLLName = MakePtr(PCHAR, ModuleBase, pImportDesc->Name);
    if (_stricmp(pImportedDLLName, "user32.dll") &&
        _stricmp(pImportedDLLName, "kernel32.dll"))
      continue;

    // Get the first "thunk" describing the first function imported
    // from this DLL

    pFirstThunk = MakePtr(PIMAGE_THUNK_DATA,
                          ModuleBase, pImportDesc->FirstThunk);

    // Loop through all the thunks to see how many there are, so that
    // we can set the page protection on all those thunks to read/
    // write/execute.

    for (pThunk = pFirstThunk, numberOfThunks = 0; pThunk->u1.Function;
         pThunk++, numberOfThunks++)
      ;

    // Set the page protection on the numberOfThunks thunks to
    // read/write/execute.  If we cannot, skip to the next import
    // descriptor.

    if (!SetRWXProtection(pFirstThunk,
                          numberOfThunks * sizeof(IMAGE_THUNK_DATA),
                          &protectionInfo))
      continue;

    // Loop through all the thunks, checking each function to see
    // if we have a matching hook information structure in the hash
    // table.  If we do, we replace the function pointer with the
    // hook function pointer.

    for (pThunk = pFirstThunk; pThunk->u1.Function; pThunk++) {
      info = FindInNewHashTable(pThunk->u1.Function);
      if (info != NULL)
        pThunk->u1.Function = info->origFunctionAddress;
    }

    // Restore virtual memory protections on the thunk pages to
    // what they were originally

    RestoreProtections(&protectionInfo);
  }

  return TRUE;
}
