//
// VTraceDLL.cpp contains initialization and cleanup routines for the
// VTrace DLL.
//

#include <windows.h>
#include <winioctl.h>
#include <stdio.h>
#include "../Common/VTrace.h"
#include "Globals.h"
#include "Compaction.h"
#include "HookManage.h"
#include "LogEvent.h"
#include "VTraceDLL.h"

////////////////////////////////////////////////////////////////////////////
//                          FUNCTION PROTOTYPES
////////////////////////////////////////////////////////////////////////////

BOOL InitializeDLL (void);
void ReleaseResources (void);

////////////////////////////////////////////////////////////////////////////
//                           DLL MAIN ROUTINE
////////////////////////////////////////////////////////////////////////////

// DllMain does either initialization or completion for this DLL
// when it is first attached to or detached from a process or
// thread.

BOOL WINAPI DllMain (HANDLE hInstance, DWORD what, LPVOID notUsed)
{
  char eventBuffer[5];
  long priority;

  switch (what) {
  case DLL_PROCESS_ATTACH :            // On attachment to a process...
    // Initialize all the resources we need access to. If we can't, set
    // initializedOK to FALSE so that DllMain will always do nothing when
    // it is called for this process. Even when DllMain does nothing, it
    // should still return TRUE, to avoid a messy error message that the
    // user has to click away.

    initializedOK = InitializeDLL();
    if (!initializedOK) return TRUE;

    // Log the priority of this process.

    priority = GetPriorityClass(GetCurrentProcess());

    eventBuffer[0] = (char) ENTRY_TYPE_PROCESS_SET_PRIORITY;
    CopyMemory(&eventBuffer[1], &priority, 4);
    LogEvent(eventBuffer, 5);

    // Since this also serves as an attachment to the first thread of
    // the process, do what we need to do on thread attachment: initialize
    // the number of message checks since last activity for this thread to 0.

    IndicateThreadActivity();

    // Install the hook functions in each module of this process.

    EnumerateLoadedModules(GetCurrentProcess(), &InstallHooksInModule, NULL);
    break;

  case DLL_PROCESS_DETACH :      // On detachment from a process...
    if (!initializedOK) return TRUE;

    // Uninstall the hook functions from each module of this process

    EnumerateLoadedModules(GetCurrentProcess(),
                           &UninstallHooksInModule, NULL);

    // Free the storage for the per-thread compaction variables of this
    // main thread

    FreeCompactionTechniquePerThreadVariables();

    // Free all resources obtained at process-attach by InitializeDLL.

    ReleaseResources();
    break;

  case DLL_THREAD_ATTACH :         // On attachment to a thread...
    if (!initializedOK) return TRUE;

    // Log the thread priority.

    priority = GetThreadPriority(GetCurrentThread());

    eventBuffer[0] = (char) ENTRY_TYPE_THREAD_SET_PRIORITY;
    RtlCopyMemory(&eventBuffer[1], &priority, 4);
    LogEvent(eventBuffer, 5);

    // Initialize the number of message-checks since last activity for this
    // thread to 0.

    IndicateThreadActivity();
    break;

  case DLL_THREAD_DETACH :         // On detachment from a thread...
    if (!initializedOK) return TRUE;
    FreeCompactionTechniquePerThreadVariables(); // Deallocate per-thread vars
    break;
  }

  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
//                       INITIALIZING / TERMINATING
////////////////////////////////////////////////////////////////////////////

// InitializeDLL attains all the resources needed to do the VTrace stuff.
// It returns a boolean indicating if it was successful. If it returns TRUE,
// the caller must subsequently call ReleaseResources to release all the
// resources it acquired. If it returns FALSE, it will have already released
// any resources it obtained.

BOOL InitializeDLL (void)
{
  DWORD returnSize;

  // Obtain a handle to VTrcInit.

  vtrcInitHandle = CreateFile("\\\\.\\VTrcInit", GENERIC_READ | GENERIC_WRITE,
                              0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,
                              NULL);
  if (vtrcInitHandle == INVALID_HANDLE_VALUE)
    return FALSE;

  // Get a pointer to the kernel-user shared state from VTrcInit.

  if (!DeviceIoControl(vtrcInitHandle,
                       VTRACE_CTRL_GET_KERNEL_USER_SHARED_STATE,
                       NULL, 0, &kernelUserSharedState, sizeof(PVOID),
                       &returnSize, NULL) ||
      returnSize != sizeof(PVOID) || kernelUserSharedState == NULL) {
    CloseHandle(vtrcInitHandle);
    return FALSE;
  }

  // Initialize the hook information: where each hooked function is and
  // where each corresponding hook is.

  if (!InitHookInfo()) {
    CloseHandle(vtrcInitHandle);
    return FALSE;
  }

  // Initialize thread-local storage for per-thread compaction technique
  // variables.  Store the index into that storage in compactionThread-
  // VarsPtrTlsIndex.

  compactionThreadVarsPtrTlsIndex = TlsAlloc();
  if (compactionThreadVarsPtrTlsIndex == 0xFFFFFFFF) {
    CloseHandle(vtrcInitHandle);
    return FALSE;
  }

  // Initialize per-process compaction technique variables.

  InitCompactionTechniquePerProcessVariables();

  // Return success.

  return TRUE;
}

// ReleaseResources releases any resources obtained in a successful call
// to InitializeDLL.

void ReleaseResources (void)
{
  TlsFree(compactionThreadVarsPtrTlsIndex);
  CloseHandle(vtrcInitHandle);
}
