//
// InstallRoutines.cpp contains routines for installing VTrace in stages.
//
// For 79-character width, truncate the window after this sentence is finished.

#include "stdafx.h"
#include "Install.h"
#include "InstallDlg.h"
#include "InstallRoutines.h"
#include "RefTime.h"
#include "../Common/RegConstraints.h"
#include "../Common/Version.h"
#include <winioctl.h>
#include <winsvc.h>
#include <shlobj.h>

////////////////////////////////////////////////////////////////////////////
//                                MACROS
////////////////////////////////////////////////////////////////////////////

#define INITIAL_GUESS_SERVICE_CONFIG_SIZE       256

////////////////////////////////////////////////////////////////////////////
//                          FORWARD DEFINITIONS
////////////////////////////////////////////////////////////////////////////

void DoInstall (HWND hWnd, CString& mainDirectory, CString& personalName,
                DWORD previousVersion);

BOOL IsItAGoodNTVersion (void);
BOOL DoNeededSystemFilesExist (char *systemRoot);
BOOL IsItAGoodProcessor (void);
BOOL DoesAnyConflictingFileExist (char *systemRoot);
BOOL CreateMainDirectoryAndTracesSubdirectory (CString& mainDirectory);
BOOL CreateSystemConfigurationFile (CString& mainDirectory,
                                    CString& personalName, REF_TIME& refTime);
BOOL CreateVTraceRegistryKeys (CString& mainDirectory, CString& personalName);
BOOL InstallDriver (char *driverName, char *dependencies, DWORD startTime,
                    char *group, BOOL isFileSystem, DWORD *tagPtr,
                    SC_HANDLE schSCManager, char *systemRoot);
BOOL InstallDLL (char *systemRoot);
BOOL InstallVTrcSrvc (char *systemRoot, SC_HANDLE schSCManager);
BOOL InstallUtilities (CString& mainDirectory, DWORD previousVersion);
void AdvancedConfiguration (void);
void RemoveActivateVTraceNextTimeKey (void);
BOOL MyRegSetValue (HKEY hKey, char *name, BYTE *valuePtr, DWORD type,
                    DWORD length, char *errorDescription);
BOOL GetTagAndGroup (SC_HANDLE schService, DWORD *tagPtr, char *groupName,
                     DWORD maxGroupSize);
BOOL RemoveTagFromGroup (DWORD tag, char *groupName);
BOOL AddTagsToGroup (char *groupName, DWORD *tags, DWORD numNewTags);

#if TARGET_NT_VERSION == 5

BOOL AddUpperFilter (char *driverName, char *classGuid);
BOOL DisableKernelOverwrite (void);

#endif

////////////////////////////////////////////////////////////////////////////
//                            ERROR DISPLAY
////////////////////////////////////////////////////////////////////////////

// DisplayError displays, in a message box, the error message given
// in the string generalReason as well as the error message implied
// by the system error state.

void DisplayError (char *generalReason)
{
  char *lpMsgBuf;
  CString str;

  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                NULL, GetLastError(),
                MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &lpMsgBuf,
                0, NULL);
  str.Format("%s because %s", generalReason, lpMsgBuf);
  MessageBox(NULL, str, "Error", MB_OK | MB_ICONEXCLAMATION);

  LocalFree(lpMsgBuf);
}

////////////////////////////////////////////////////////////////////////////
//                            FREE DISK SPACE
////////////////////////////////////////////////////////////////////////////

// GetFreeMB gets the number of MB free in the given directory (directoryName).
// If the directory doesn't exist, it tries to remove subdirectory names from
// the end of its name until it reaches one that does exist.  It puts the
// number of free MB in the given pointer (freeMBPtr) and returns a boolean
// indicating whether it was successful.

BOOL GetFreeMB (CString& directoryName, UINT *freeMBPtr)
{
  ULARGE_INTEGER freeBytes, dummy1, dummy2;

  // Check how much free space is in the directory

  if (GetDiskFreeSpaceEx(directoryName, &freeBytes, &dummy1, &dummy2)) {
    *freeMBPtr = (UINT) (freeBytes.QuadPart / 1024 / 1024);
    return TRUE;
  }

  // If we didn't succeed, try subdirectories

  char smallerDirectoryName[MAX_PATH];
  int endPos;

  // First, copy the directory name into smallerDirectoryName so we can play
  // around with it, and put its end position in endPos

  strcpy(smallerDirectoryName, directoryName);
  endPos = strlen(smallerDirectoryName);

  // Keep looping as long as the last call to GetDiskFreeSpaceEx returned
  // a path (or file) not found error

  while (GetLastError() == ERROR_PATH_NOT_FOUND ||
         GetLastError() == ERROR_FILE_NOT_FOUND) {
    // Each time through the loop, scan backward from the end for the last \.
    // If there isn't one, break out of the loop and fail.  If there is such
    // a \, replace it with a null character and try getting the disk space
    // again.
    for (endPos--; endPos > 1 && smallerDirectoryName[endPos] != '\\';
         endPos--)
      ;
    if (endPos < 2) break;
    smallerDirectoryName[endPos] = '\0';
    if (GetDiskFreeSpaceEx(smallerDirectoryName, &freeBytes, &dummy1,
                           &dummy2)) {
      *freeMBPtr = (UINT) (freeBytes.QuadPart / 1024 / 1024);
      return TRUE;
    }
  }

  // We failed, so just return failure

  return FALSE;
}

////////////////////////////////////////////////////////////////////////////
//                      DISK AND PARTITION NUMBERS
////////////////////////////////////////////////////////////////////////////

// GetDiskNumbers finds the disk, partition, and volume numbers for a
// certain drive, such as C:, given the drive letter (e.g, 'C'). It
// returns these numbers in the appropriate pointer locations
// passed. It returns a boolean indicating if it was successful. (If
// it fails to get a volume number, it may still return success if it
// succeeds at getting a disk and partition number; in this case, it
// returns a bogus volume number of 0xFFFFFFFFUL). There is a
// different version of this function for NT 4.0 and Windows 2000.

#if TARGET_NT_VERSION == 4

// The NT 4.0 version of GetDiskNumbers uses QueryDosDevice

BOOL GetDiskNumbers (char driveLetter, DWORD *diskNumberPtr,
                     DWORD *partitionNumberPtr, DWORD *volumeNumberPtr)
{
  char driveName[3] = "A:";
  char diskInfo[64], sample[64], *p;

  // Use QueryDosDevice to get the raw device name.

  driveName[0] = driveLetter;
  if (!QueryDosDevice(driveName, diskInfo, 64))
    return FALSE;

  // If it doesn't begin with "\Device\Harddisk", fail.

  if (strncmp(diskInfo, "\\Device\\Harddisk", 16))
    return FALSE;

  // Extract the hard disk number and partition number.

  diskInfo[63] = '\0';
  *diskNumberPtr = atoi(&diskInfo[16]);
  p = strchr(&diskInfo[17], 'n');
  if (p == NULL)
    return FALSE;
  *partitionNumberPtr = atoi(p+1);
  *volumeNumberPtr = 0xFFFFFFFFUL;  // Volume number is meaningless on NT 4

  // Make sure the disk number and partition number extracted really
  // match, and if so, return TRUE.

  sprintf(sample, "\\Device\\Harddisk%ld\\Partition%ld", *diskNumberPtr,
          *partitionNumberPtr);
  return !strcmp(diskInfo, sample);
}

#else

// The following declarations for getting volume numbers are borrowed from
// ntddvol.h in the W2K DDK:

typedef struct _VOLUME_NUMBER {
    ULONG   VolumeNumber;
    WCHAR   VolumeManagerName[8];
} VOLUME_NUMBER, *PVOLUME_NUMBER;

#define IOCTL_VOLUME_BASE   ((ULONG) 'V')
#define IOCTL_VOLUME_QUERY_VOLUME_NUMBER  \
       CTL_CODE(IOCTL_VOLUME_BASE, 7, METHOD_BUFFERED, FILE_ANY_ACCESS)

// The Windows 2000 version of GetDiskNumbers uses the superior,
// but unavailable on NT 4.0, IOCTL_STORAGE_GET_DEVICE_NUMBER and
// IOCTL_VOLUME_QUERY_VOLUME_NUMBER methods.

BOOL GetDiskNumbers (char driveLetter, DWORD *diskNumberPtr,
                     DWORD *partitionNumberPtr, DWORD *volumeNumberPtr)
{
  HANDLE fileHandle;
  STORAGE_DEVICE_NUMBER deviceInfo;
  VOLUME_NUMBER volumeInfo;
  DWORD bytesReturned;
  char rawDiskName[] = "\\\\.\\A:";

  // Open the raw disk

  rawDiskName[4] = driveLetter;
  fileHandle = CreateFile(rawDiskName, GENERIC_READ,
                          FILE_SHARE_READ | FILE_SHARE_WRITE,
                          NULL, OPEN_EXISTING,
                          FILE_ATTRIBUTE_NORMAL, NULL);
  if (fileHandle == INVALID_HANDLE_VALUE) return FALSE;

  // Get disk and partition number information using a device control
  // request.

  if (!DeviceIoControl(fileHandle, IOCTL_STORAGE_GET_DEVICE_NUMBER, NULL,
                       0, &deviceInfo, sizeof(deviceInfo), &bytesReturned,
                       NULL))
    return FALSE;
  if (bytesReturned != sizeof(deviceInfo))
    return FALSE;

  // Put the values into the passed pointers.

  *diskNumberPtr = deviceInfo.DeviceNumber;
  *partitionNumberPtr = deviceInfo.PartitionNumber;

  // Get the volume information using another device control request.

  if (DeviceIoControl(fileHandle, IOCTL_VOLUME_QUERY_VOLUME_NUMBER, NULL,
                      0, &volumeInfo, sizeof(volumeInfo), &bytesReturned,
                      NULL) &&
      bytesReturned == sizeof(volumeInfo))
    *volumeNumberPtr = volumeInfo.VolumeNumber;
  else
    *volumeNumberPtr = 0xFFFFFFFFUL;

  // Close the handle to the raw disk and return success

  CloseHandle(fileHandle);

  return TRUE;
}

#endif

////////////////////////////////////////////////////////////////////////////
//                    HIGH-LEVEL INSTALLATION ROUTINES
////////////////////////////////////////////////////////////////////////////

// InstallAsynchronously installs VTrace asynchronously.  It takes as its
// single parameter an LPVOID which can be cast into a pointer to an
// INSTALL_PARAMETERS structure containing all the information needed
// for the install.  Its strange definition style is so that it can be
// called as a thread routine.

UINT InstallAsynchronously (LPVOID param)
{
  INSTALL_PARAMETERS *installParameters = (INSTALL_PARAMETERS *) param;

  DoInstall(installParameters->hWnd, installParameters->mainDirectory,
            installParameters->personalName,
            installParameters->previousVersion);
  PostMessage(installParameters->hWnd, WM_INSTALL_COMPLETE, 0, 0);
  return 0;
}

// DoInstall installs VTrace.  It takes as parameters a window handle (hWnd)
// that it uses for changing the status message, a directory name
// (mainDirectory) that it uses as the place to install VTrace, a
// personal name (personalName) that the user entered, a DWORD
// indicating the currently installed version (previousVersion, 0 if none
// installed).

void DoInstall (HWND hWnd, CString& mainDirectory, CString& personalName,
                DWORD previousVersion)
{
  SC_HANDLE schSCManager;
  char systemRoot[MAX_PATH];
  int response;
  REF_TIME refTime;
  DWORD tag[NUM_TAGS_NEEDED];

  // Get a time reference for later use in determining the clock speed

  GetCurrentRefTime(refTime);

  // Make sure this version of NT can run VTrace

  PostMessage(hWnd, WM_CHANGE_STATUS,
              (WPARAM) "Checking that we can run on this version of Windows",
              0);
  if (!IsItAGoodNTVersion()) return;
  if (GetEnvironmentVariable("SystemRoot", systemRoot, 128) == 0) {
    DisplayError("Could not obtain the system root directory environment "
                 "variable");
    return;
  }
  if (!DoNeededSystemFilesExist(systemRoot)) return;
  if (!IsItAGoodProcessor()) return;
  if (DoesAnyConflictingFileExist(systemRoot)) return;

  // Create main VTrace directory and its traces subdirectory

  PostMessage(hWnd, WM_CHANGE_STATUS, (WPARAM) "Creating VTrace directories",
              0);
  if (!CreateMainDirectoryAndTracesSubdirectory(mainDirectory)) return;

  // Create and set the registry information for VTrace

  PostMessage(hWnd, WM_CHANGE_STATUS,
              (WPARAM) "Setting VTrace registry information", 0);
  if (!CreateVTraceRegistryKeys(mainDirectory, personalName)) return;

  PostMessage(hWnd, WM_CHANGE_STATUS, (WPARAM) "Installing drivers", 0);

#if TARGET_NT_VERSION == 5

  // We no longer need to enable kernel overwriting in this version,
  // so disable it if an existing version enabled it.

  DisableKernelOverwrite();

#endif

  // Get handle to service control manager

  schSCManager = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
  if (schSCManager == NULL) {
    DisplayError("Could not open the service control manager");
    return;
  }

  // Install drivers. The differences between NT and 2000 are as
  // follows. First, in NT, we use the "Filter" group, while in 2000
  // we use the "System Bus Extender" group (the same group as the
  // disks) so that VTrcRawD doesn't get "moved up" ahead of VTrcInit
  // in the actual load order. Second, in NT, VTrcKbd uses
  // SERVICE_AUTO_START as its start value, while in 2000 it uses
  // SERVICE_DEMAND_START as its start value.  This is what's
  // recommended by the DDK for devices started by the operating
  // system as upper filters that don't have to start at boot.

#if TARGET_NT_VERSION == 4

  if (!InstallDriver("VTrcInit", "", SERVICE_BOOT_START,
                     "Filter", FALSE, &tag[0], schSCManager,
                     systemRoot) ||
      !InstallDriver("VTrcLog", "", SERVICE_AUTO_START, "", FALSE,
                     NULL, schSCManager, systemRoot) ||
      !InstallDriver("VTrcFS", "", SERVICE_BOOT_START,
                     "Filter", TRUE, &tag[1], schSCManager,
                     systemRoot) ||
      !InstallDriver("VTrcFsrc", "", SERVICE_BOOT_START,
                     "Filter", TRUE, &tag[2], schSCManager,
                     systemRoot) ||
      !InstallDriver("VTrcKbd", "", SERVICE_AUTO_START, "", FALSE,
                     NULL, schSCManager, systemRoot) ||
      !InstallDriver("VTrcNet", "", SERVICE_AUTO_START, "", FALSE,
                     NULL, schSCManager, systemRoot) ||
      !InstallDriver("VTrcRawD", "", SERVICE_BOOT_START,
                     "Filter", FALSE, &tag[3], schSCManager,
                     systemRoot)) {
    CloseServiceHandle(schSCManager);
    return;
  }

  // Add tags to the Filter group.

  if (!AddTagsToGroup("Filter", tag, NUM_TAGS_NEEDED)) {
    CloseServiceHandle(schSCManager);
    return;
  }

#else

  if (!InstallDriver("VTrcInit", "", SERVICE_BOOT_START,
                     "System Bus Extender", FALSE, &tag[0], schSCManager,
                     systemRoot) ||
      !InstallDriver("VTrcLog", "", SERVICE_AUTO_START, "", FALSE,
                     NULL, schSCManager, systemRoot) ||
      !InstallDriver("VTrcFS", "", SERVICE_BOOT_START,
                     "System Bus Extender", TRUE, &tag[1], schSCManager,
                     systemRoot) ||
      !InstallDriver("VTrcFsrc", "", SERVICE_BOOT_START,
                     "System Bus Extender", TRUE, &tag[2], schSCManager,
                     systemRoot) ||
      !InstallDriver("VTrcKbd", "", SERVICE_DEMAND_START, "", FALSE,
                     NULL, schSCManager, systemRoot) ||
      !InstallDriver("VTrcNet", "", SERVICE_AUTO_START, "", FALSE,
                     NULL, schSCManager, systemRoot) ||
      !InstallDriver("VTrcRawD", "", SERVICE_BOOT_START,
                     "System Bus Extender", FALSE, &tag[3], schSCManager,
                     systemRoot)) {
    CloseServiceHandle(schSCManager);
    return;
  }

  // Add tags to the System Bus Extender group.

  if (!AddTagsToGroup("System Bus Extender", tag, NUM_TAGS_NEEDED)) {
    CloseServiceHandle(schSCManager);
    return;
  }

#endif

#if TARGET_NT_VERSION == 5

  // Set VTrcKbd as an upper filter on the keyboard class and VTrcRawD as
  // an upper filter on the storage volume class

  if (!AddUpperFilter("VTrcKbd", KEYBOARD_CLASS_GUID) ||
      !AddUpperFilter("VTrcRawD", STORAGE_VOLUME_CLASS_GUID)) {
    CloseServiceHandle(schSCManager);
    return;
  }

#endif

  // Install user-mode service

  PostMessage(hWnd, WM_CHANGE_STATUS, (WPARAM) "Installing user-mode service",
              0);
  if (!InstallVTrcSrvc(systemRoot, schSCManager)) {
    CloseServiceHandle(schSCManager);
    return;
  }

  // Release service control manager handle
  
  CloseServiceHandle(schSCManager);

  // Install VTrace DLL

  PostMessage(hWnd, WM_CHANGE_STATUS, (WPARAM) "Installing DLL", 0);
  if (!InstallDLL(systemRoot)) return;

  // Install utilities in main VTrace directory and put shortcuts to them
  // in the Start menu under a new directory called VTrace

  PostMessage(hWnd, WM_CHANGE_STATUS,
              (WPARAM) "Installing utilities in Start menu under VTrace", 0);
  if (!InstallUtilities(mainDirectory, previousVersion)) return;

  // Create the system configuration file in the traces subdirectory

  PostMessage(hWnd, WM_CHANGE_STATUS,
              (WPARAM) "Getting information about this machine", 0);
  if (!CreateSystemConfigurationFile(mainDirectory, personalName, refTime))
    return;

  // Remove obsolete key to activate VTrace at next boot-up

  RemoveActivateVTraceNextTimeKey();

  // Display success message

  PostMessage(hWnd, WM_CHANGE_STATUS, (WPARAM) "Installation complete", 0);
  response = MessageBox(NULL,
     "Install performed successfully.\n"
     "You must reboot the system for some aspects of the installation "
     "to be complete.\n"
     "\n"
     "Do you want to perform advanced configuration of VTrace now?",
     "Success", MB_YESNO | MB_DEFBUTTON1);
  if (response == IDYES)
    AdvancedConfiguration();
}

////////////////////////////////////////////////////////////////////////////
//                    FILE/DIRECTORY BASIC OPERATIONS
////////////////////////////////////////////////////////////////////////////

// RecursiveCreateDirectory creates a directory with the given name (name)
// and security attributes (security).  Unlike CreateDirectory, which just
// fails if a subdirectory of the desired directory does not exist, this
// routine actually creates any necessary subdirectories.  For example,
// if the desired name was C:\FOO\BAR\ZIP, and C:\FOO did not exist, then
// this routine would create C:\FOO, then C:\FOO\BAR, then C:\FOO\BAR\ZIP.

BOOL RecursiveCreateDirectory (char *name, LPSECURITY_ATTRIBUTES security)
{
  char *p;
  BOOL result;
  DWORD err;

  // First, try to create the directory straight on with CreateDirectory.
  // Usually, this will succeed.  (That's why we do it first -- make the
  // common case fast.)

  result = CreateDirectory(name, security);
  if (result) return TRUE;
  err = GetLastError();
  if (err == ERROR_ALREADY_EXISTS) return TRUE;

  // If it failed because the path didn't exist, call RecursiveCreateDirectory
  // on the subdirectory of the given directory, then try CreateDirectory
  // again.

  if (err == ERROR_PATH_NOT_FOUND) {
    p = strrchr(name, '\\');
    if (p != NULL) {
      *p = '\0';
      result = RecursiveCreateDirectory(name, security);
      *p = '\\';
      if (result)
        result = CreateDirectory(name, security);
    }
  }

  return result;
}

// CreateShortcut creates a shortcut with a given name (linkName) pointing
// to an existing file or directory of another given name (targetName).  It
// assumes that CoInitialize() has already been called to initialize COM,
// which is used by this routine.

BOOL CreateShortcut (char *linkName, char *targetName)
{
  HRESULT hres;
  IShellLink* psl;
  IPersistFile* ppf;
  WORD linkNameAsUnicode[MAX_PATH];

  // Get a pointer to the IShellLink interface
  hres = CoCreateInstance(CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER,
                          IID_IShellLink, (void **) &psl);
  if (!SUCCEEDED(hres))
    return FALSE;

  // Query IShellLink for the IPersistFile interface for saving the
  // shell link in persistent storage
  hres = psl->QueryInterface(IID_IPersistFile, (void **) &ppf);
  if (!SUCCEEDED(hres)) {
    psl->Release();
    return FALSE;
  }

  // Set the path to the shell link target
  hres = psl->SetPath(targetName);
  if (!SUCCEEDED(hres)) {
    ppf->Release();
    psl->Release();
    return FALSE;
  }

  // Convert the link name ANSI string to Unicode
  if (MultiByteToWideChar(CP_ACP, 0, linkName, -1, linkNameAsUnicode,
                          MAX_PATH) == 0) {
    ppf->Release();
    psl->Release();
    return FALSE;
  }

  // Save the link via the IPersistFile::Save method
  hres = ppf->Save(linkNameAsUnicode, TRUE);
    
  // Release pointers to IPersistFile and IShellLink
  ppf->Release();
  psl->Release();

  return SUCCEEDED(hres);
}

////////////////////////////////////////////////////////////////////////////
//                            STRING OPERATIONS
////////////////////////////////////////////////////////////////////////////

// ContainsVTraceDLL searches for the VTrace DLL name in a string which is
// a space-separated list of DLL names.  The VTrace DLL name must be in all
// caps for it to match here.  It returns a boolean indicating whether the
// name was found.

BOOL ContainsVTraceDLL (char *str)
{
  char *p;

  // Search starting at the beginning of the string.  Each time through the
  // loop, look for the first occurrence of the substring "VTRACE.DLL".  If it
  // isn't found, break out of the loop and return FALSE.  Otherwise, if the
  // substring represents the full name of a DLL (i.e., it is bounded on the
  // left by a space or beginning-of-string, and bounded on the right by a
  // space or end-of-string), then return TRUE.  Otherwise, advance the search
  // start point by 10, since the next occurrence of VTRACE.DLL must be at
  // least 10 characters further in the string.

  for (p = str; 1; p+=10) {
    p = strstr(p, "VTRACE.DLL");
    if (p == NULL) break;
    if ((p == str || *(p-1) == ' ') && (*(p+10) == ' ' || *(p+10) == '\0'))
      return TRUE;
  }
  return FALSE;
}

// ConvertVersionNumberToString converts a VTrace version number (version) to
// a string (str).  For instance, it converts 0x010203 to "1.2.3".

void ConvertVersionNumberToString (CString& str, DWORD version)
{
  if (version & 0xFF)
    str.Format("%u.%u.%u", version >> 16, (version >> 8) & 0xFF,
               version & 0xFF);
  else
    str.Format("%u.%u", version >> 16, (version >> 8) & 0xFF);
}

// FindStringInMultiSz finds the position of the given full string in
// the given REG_MULTI_SZ value of the given length. If the string
// isn't found, it returns -1.

int FindStringInMultiSz (char *multiSzVal, int multiSzLength,
                         char *searchString)
{
  for (int possibleMatchPos = 0;
       possibleMatchPos < multiSzLength && multiSzVal[possibleMatchPos];
       possibleMatchPos += (strlen(&multiSzVal[possibleMatchPos]) + 1))
    if (!stricmp(&multiSzVal[possibleMatchPos], searchString))
      return possibleMatchPos;

  return -1;
}

////////////////////////////////////////////////////////////////////////////
//                        LOAD ORDER GROUP TAGS
////////////////////////////////////////////////////////////////////////////

// GetTagAndGroup determines the tag and group of a given service.
// If such a tag and group exists, it puts them in tagPtr and groupName and
// returns TRUE; otherwise, it returns FALSE.

BOOL GetTagAndGroup (SC_HANDLE schService, DWORD *tagPtr, char *groupName,
                     DWORD maxGroupSize)
{
  QUERY_SERVICE_CONFIG *serviceConfigInfo;
  DWORD sizeNeeded;

  // Read the configuration information for the service. The first guess
  // as to the needed size for this information could be wrong, so try
  // again if the return value indicates a bigger size is needed.

  serviceConfigInfo = (QUERY_SERVICE_CONFIG *)
    malloc(INITIAL_GUESS_SERVICE_CONFIG_SIZE);
  if (serviceConfigInfo == NULL) return FALSE;
  if (!QueryServiceConfig(schService, serviceConfigInfo,
                          INITIAL_GUESS_SERVICE_CONFIG_SIZE, &sizeNeeded)) {
    if (GetLastError() != ERROR_INSUFFICIENT_BUFFER) {
      free(serviceConfigInfo);
      return FALSE;
    }
    else {
      free(serviceConfigInfo);
      serviceConfigInfo = (QUERY_SERVICE_CONFIG *) malloc(sizeNeeded);
      if (serviceConfigInfo == NULL) return FALSE;
      if (!QueryServiceConfig(schService, serviceConfigInfo, sizeNeeded,
                              &sizeNeeded)) {
        free(serviceConfigInfo);
        return FALSE;
      }
    }
  }

  // Extract the tag and group information if it exists.

  if (serviceConfigInfo->lpLoadOrderGroup &&
      serviceConfigInfo->lpLoadOrderGroup[0] &&
      strlen(serviceConfigInfo->lpLoadOrderGroup) < maxGroupSize) {
    strcpy(groupName, serviceConfigInfo->lpLoadOrderGroup);
    *tagPtr = serviceConfigInfo->dwTagId;
    free(serviceConfigInfo);
    return TRUE;
  }
  else {
    free(serviceConfigInfo);
    return FALSE;
  }
}

// RemoveTagFromGroup removes the given tag number from the given
// group. Actually, it removes the tag from the GroupOrderList
// registry entry that indicates the order in which that group
// services should load.

BOOL RemoveTagFromGroup (DWORD tag, char *groupName)
{
  DWORD err, type, length, lengthNeeded, i, j, *groupOrderList;
  HKEY hKey;

  // Open the group order list registry key.

  err = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                     "System\\CurrentControlSet\\Control\\GroupOrderList",
                     NULL, KEY_READ | KEY_WRITE, &hKey);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not open registry key for service group order list");
    return FALSE;
  }

  // Read the value named after the group to get the actual group order
  // list. First we read the value with no buffer to put it into, to find
  // out how long a buffer we need, if any. If there is no such key, we're
  // done. Otherwise, we read the value from the registry. Note that the
  // structure for the value is a sequence of DWORD's, the first of which
  // is the number of tags n, and the next n of which are those tags in the
  // order in which they should be started.

  lengthNeeded = 0;
  err = RegQueryValueEx(hKey, groupName, NULL, &type, NULL, &lengthNeeded);
  if (err == ERROR_FILE_NOT_FOUND) {
    RegCloseKey(hKey);
    return TRUE;
  }
  else if (err == ERROR_SUCCESS) {
    groupOrderList = (DWORD *) malloc(lengthNeeded);
    length = lengthNeeded;
    err = RegQueryValueEx(hKey, groupName, NULL, &type,
                          (LPBYTE) groupOrderList, &length);
    if (err != ERROR_SUCCESS) {
      SetLastError(err);
      DisplayError("Could not read group order list value");
      RegCloseKey(hKey);
      return FALSE;
    }
    if (type != REG_BINARY || lengthNeeded != length ||
        (groupOrderList[0]+1) * sizeof(DWORD) != length) {
      MessageBox(NULL, "Read invalid group order list value", "Error",
                 MB_OK | MB_ICONEXCLAMATION);
      RegCloseKey(hKey);
      return FALSE;
    }
  }
  else {
    SetLastError(err);
    DisplayError("Could not get size of group order list value");
    RegCloseKey(hKey);
    return FALSE;
  }

  // Search for the tag in the group order list. If it's there, remove it.

  for (i = 1; i <= groupOrderList[0]; i++) {
    if (groupOrderList[i] == tag) {
      for (j = i; j < groupOrderList[0]; j++)
        groupOrderList[j] = groupOrderList[j+1];
      groupOrderList[0]--;
      break;
    }
  }

  // Set the new group order list in the registry.

  err = RegSetValueEx(hKey, groupName, 0, REG_BINARY, (LPBYTE) groupOrderList,
                      (groupOrderList[0] + 1) * sizeof(DWORD));
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not set the group order list in the registry");
    RegCloseKey(hKey);
    return FALSE;
  }

  // Release resources used and return a status value.

  free(groupOrderList);
  RegCloseKey(hKey);

  return TRUE;
}

// AddTagsToGroup adds a given set of tags to the load order of a given
// service group. The name of the group is given in groupName; the set of
// tags is given in the array tags of length numNewTags. The order the tags
// appear in the array match the order in which they should apppear in the
// group order list. If a tag already exists in the group order list, it
// should not be added.

BOOL AddTagsToGroup (char *groupName, DWORD *tags, DWORD numNewTags)
{
  DWORD err, type, length, lengthNeeded, i, j, *groupOrderList;
  BOOL found, retVal;
  HKEY hKey;

  // Open the group order list registry key.

  err = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                     "System\\CurrentControlSet\\Control\\GroupOrderList",
                     NULL, KEY_READ | KEY_WRITE, &hKey);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not open registry key for service group order list");
    return FALSE;
  }

  // Read the value named after the group to get the actual group order
  // list. First we read the value with no buffer to put it into, to find
  // out how long a buffer we need, if any. If there is no such key, the
  // group does not exist, so we create our own value with no tags. Otherwise,
  // we read the value from the registry. In either case, we allocate extra
  // space for the value to accommodate the extra numNewTags tags. Note that
  // the structure for the value is a sequence of DWORD's, the first of which
  // is the number of tags n, and the next n of which are those tags in the
  // order in which they should be started.

  lengthNeeded = 0;
  err = RegQueryValueEx(hKey, groupName, NULL, &type, NULL, &lengthNeeded);
  if (err == ERROR_FILE_NOT_FOUND) {
    groupOrderList = (DWORD *) malloc((numNewTags+1) * sizeof(DWORD));
    groupOrderList[0] = 0;
  }
  else if (err == ERROR_SUCCESS) {
    groupOrderList = (DWORD *) malloc(lengthNeeded +
                                      numNewTags * sizeof(DWORD));
    length = lengthNeeded;
    err = RegQueryValueEx(hKey, groupName, NULL, &type,
                          (LPBYTE) groupOrderList, &length);
    if (err != ERROR_SUCCESS) {
      SetLastError(err);
      DisplayError("Could not read group order list value");
      RegCloseKey(hKey);
      return FALSE;
    }
    if (type != REG_BINARY || lengthNeeded != length ||
        (groupOrderList[0]+1) * sizeof(DWORD) != length) {
      MessageBox(NULL, "Read invalid group order list value", "Error",
                 MB_OK | MB_ICONEXCLAMATION);
      RegCloseKey(hKey);
      return FALSE;
    }
  }
  else {
    SetLastError(err);
    DisplayError("Could not read size of group order list value");
    RegCloseKey(hKey);
    return FALSE;
  }

  // Add the array of tags to the list of tags in 'groupOrderList'. Ignore any
  // tags that are already found in the array.

  for (i = 0; i < numNewTags; i++) {
    found = FALSE;
    for (j = 1; j <= groupOrderList[0]; j++) {
      if (groupOrderList[j] == tags[i]) {
        found = TRUE;
        break;
      }
    }
    if (!found) {
      groupOrderList[0]++;
      groupOrderList[groupOrderList[0]] = tags[i];
    }
  }

  // Change the group order list to the new group order list.

  retVal = MyRegSetValue(hKey, groupName, (LPBYTE) groupOrderList, REG_BINARY,
                         (groupOrderList[0] + 1) * sizeof(DWORD),
                         "new group order list");

  // Release resources used and return a status value.

  free(groupOrderList);
  RegCloseKey(hKey);

  return retVal;
}

////////////////////////////////////////////////////////////////////////////
//                          REGISTRY OPERATIONS
////////////////////////////////////////////////////////////////////////////

// MyRegSetValue is like RegSetValueEx, except it handles error conditions.
// It takes as parameters the key to change (hKey), the name of the value
// to change (name), a pointer to the new value (valuePtr), the new type
// (type), the length of the new value (length), and a description of the
// registry value to be used in generating error messages (errorDescription).

BOOL MyRegSetValue (HKEY hKey, char *name, BYTE *valuePtr, DWORD type,
                    DWORD length, char *errorDescription)
{
  LONG status;

  status = RegSetValueEx(hKey, name, 0, type, valuePtr, length);
  if (status != ERROR_SUCCESS) {
    char errorString[256];
    sprintf(errorString, "Could not set the %s in the registry",
            errorDescription);
    SetLastError(status);
    DisplayError(errorString);
    RegCloseKey(hKey);
    return FALSE;
  }
  return TRUE;
}

// GetLocalCPUSpeed computes the local CPU speed in Hz and puts it in the
// given pointer location (cpuSpeedPtr).  It returns a boolean indicating
// whether it succeeded.  It first tries to use a reference time established
// earlier (refTime) to compute the CPU speed; if this is not possible, it
// uses the less precise method of reading the processor's MHz from the
// registry.

BOOL GetLocalCPUSpeed (DWORD *cpuSpeedPtr, REF_TIME& refTime)
{
  LONG err;
  DWORD type, length;
  HKEY hKey;

  if (DeduceCpuSpeedFromOldRefTime(refTime, cpuSpeedPtr))
    return TRUE;

  err = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                     "HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\0",
                     NULL, KEY_READ, &hKey);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not open processor registry key to get CPU speed "
                 "info");
    return FALSE;
  }

  length = 4;
  err = RegQueryValueEx(hKey, "~MHz", NULL, &type, (LPBYTE) cpuSpeedPtr,
                        &length);
  RegCloseKey(hKey);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not read CPU speed from registry");
    return FALSE;
  }
  if (type != REG_DWORD || length != 4) {
    MessageBox(NULL, "Found invalid format for CPU speed in registry", "Error",
               MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }

  *cpuSpeedPtr *= 1000000;
  return TRUE;
}

#if TARGET_NT_VERSION == 5

// AddUpperFilter adds the given driver name as an upper filter on the
// devices belonging to the class of the given GUID. It returns a boolean
// indicating if it was successful.

BOOL AddUpperFilter (char *driverName, char *classGuid)
{
  char buf[512];
  DWORD bufSize = 512, type, length, bytesToAdd;
  LONG status;
  HKEY hKey;

  sprintf(buf, "System\\CurrentControlSet\\Control\\Class\\{%s}", classGuid);
  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, KEY_QUERY_VALUE |
                        KEY_SET_VALUE, &hKey);
  if (status != ERROR_SUCCESS) {
    SetLastError(status);
    DisplayError("Could not open keyboard class registry key");
    return FALSE;
  }

  length = bufSize;
  status = RegQueryValueEx(hKey, "UpperFilters", NULL, &type,
                           (BYTE *) buf, &length);
  if (status == ERROR_FILE_NOT_FOUND) {
    buf[0] = '\0';
    length = 1;
  }
  else if (status != ERROR_SUCCESS) {
    RegCloseKey(hKey);
    SetLastError(status);
    DisplayError("Could not open UpperFilters value of keyboard class");
    return FALSE;
  }
  else if (type != REG_MULTI_SZ) {
    RegCloseKey(hKey);
    MessageBox(NULL, "Read invalid UpperFilters value in keyboard class",
               "Error", MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }
  else if (buf[length-1] != '\0') {
    RegCloseKey(hKey);
    MessageBox(NULL, "Found invalid UpperFilters value in keyboard class",
               "Error", MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }
  else if (FindStringInMultiSz(buf, length, driverName) >= 0) {
    RegCloseKey(hKey);
    return TRUE;
  }

  bytesToAdd = strlen(driverName) + 1;
  if (length + bytesToAdd > bufSize) {
    RegCloseKey(hKey);
    MessageBox(NULL,
               "UpperFilters list on keyboard class too long to add another",
               "Error", MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }
  strcpy(&buf[length-1], driverName);
  length += bytesToAdd;
  buf[length-1] = '\0';

  status = RegSetValueEx(hKey, "UpperFilters", 0, REG_MULTI_SZ,
                         (BYTE *) buf, length);
  if (status != ERROR_SUCCESS) {
    RegCloseKey(hKey);
    SetLastError(status);
    DisplayError("Could not set new UpperFilters value in keyboard class");
    return FALSE;
  }

  RegCloseKey(hKey);
  return TRUE;
}

#endif

////////////////////////////////////////////////////////////////////////////
//                      LOW-LEVEL INSTALLATION ROUTINES
////////////////////////////////////////////////////////////////////////////

// IsItAGoodNTVersion returns whether this is a version of Windows NT that
// VTrace can run on.

BOOL IsItAGoodNTVersion (void)
{
  OSVERSIONINFO versionInfo;

  // Read the version information

  versionInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
  if (!GetVersionEx(&versionInfo)) {
    DisplayError("Could not get Windows NT version information");
    return FALSE;
  }

  // Check to see if it's the right version of Windows NT

  if (versionInfo.dwPlatformId != VER_PLATFORM_WIN32_NT || 
      versionInfo.dwMajorVersion != TARGET_NT_VERSION) {
#if TARGET_NT_VERSION == 4
      MessageBox(NULL, "This is not Windows NT 4.0. Cannot install.", "Error",
                 MB_OK | MB_ICONEXCLAMATION);
#else
      MessageBox(NULL, "This is not Windows 2000. Cannot install.", "Error",
                 MB_OK | MB_ICONEXCLAMATION);
#endif
    return FALSE;
  }

  return TRUE;
}

// IsItAGoodProcessor checks whether it is running on a uniprocessor x86 that
// is Pentium or better.  It returns a boolean indicating if that is the case.

BOOL IsItAGoodProcessor (void)
{
  SYSTEM_INFO sysInfo;

  GetSystemInfo(&sysInfo);
  if (sysInfo.dwNumberOfProcessors > 1) {
    MessageBox(NULL,
               "VTrace does not work on a multiprocessor.  Cannot install.",
               "Error", MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }
  if (sysInfo.wProcessorArchitecture != PROCESSOR_ARCHITECTURE_INTEL) {
    MessageBox(NULL,
               "VTrace can only work on an Intel processor.  Cannot install.",
               "Error", MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }
  if (sysInfo.wProcessorLevel < 5) {
    MessageBox(NULL,
               "VTrace needs a Pentium or better processor.  Cannot install.",
               "Error", MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }

  return TRUE;
}

// DoNeededSystemFilesExist checks whether system files we need exist in the
// system directory, specified by the given string (systemRoot).

BOOL DoNeededSystemFilesExist (char *systemRoot)
{
  char fileName[MAX_PATH];
  HANDLE hfile;

  // Check for IMAGEHLP.DLL, which is needed for the function
  // EnumLoadedModules, used in VTRACE.DLL to enumerate all the modules
  // of a program and hook itself into each of them

  sprintf(fileName, "%s\\SYSTEM32\\IMAGEHLP.DLL", systemRoot);
  hfile = CreateFile(fileName, GENERIC_READ, FILE_SHARE_READ, NULL,
                     OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (hfile == INVALID_HANDLE_VALUE) {
    DisplayError("The system file IMAGEHLP.DLL is missing");
    return FALSE;
  }
  CloseHandle(hfile);

  // Check for OLE32.DLL, which is needed for creating shortcuts to
  // utilities in the Start menu.

  sprintf(fileName, "%s\\SYSTEM32\\OLE32.DLL", systemRoot);
  hfile = CreateFile(fileName, GENERIC_READ, FILE_SHARE_READ, NULL,
                     OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (hfile == INVALID_HANDLE_VALUE) {
    DisplayError("The system file OLE32.DLL is missing");
    return FALSE;
  }
  CloseHandle(hfile);

  return TRUE;
}

// DoesConflictingFileExist returns a BOOL indicating whether a file
// in the system directory exists.  The parameters are the system root
// directory (systemRoot), and the file name (fileName).  The file name
// may contain backslashes.

BOOL DoesConflictingFileExist (char *systemRoot, char *fileName)
{
  char fullPath[MAX_PATH];
  HANDLE fileHandle;

  // Put the full path of the file in fullPath

  sprintf(fullPath, "%s\\%s", systemRoot, fileName);

  // Try opening the file.  If it succeeds, we know the file exists.

  fileHandle = CreateFile(fullPath, GENERIC_READ, FILE_SHARE_READ |
                          FILE_SHARE_WRITE | FILE_SHARE_DELETE, NULL,
                          OPEN_EXISTING, 0, NULL);
  if (fileHandle != INVALID_HANDLE_VALUE) {
    char errorString[MAX_PATH + 240];
    CloseHandle(fileHandle);
    sprintf(errorString,
            "VTrace cannot be installed.\n"
            "Apparently, some other software created a file with the same "
            "name as one that VTrace needs to create.\n"
            "In case you think this file is unnecessary and want to try "
            "deleting it,\n"
            "its name is \"%s\".", fullPath);
    MessageBox(NULL, errorString, "Error", MB_OK | MB_ICONEXCLAMATION);
    return TRUE;
  }

  return FALSE;
}

// DoesAnyConflictingFileExist returns a BOOL indicating whether a
// conficting file exists, like a VTRACE.DLL that some other program is
// using.  It takes as a parameter the name of the system root directory.

BOOL DoesAnyConflictingFileExist (char *systemRoot)
{
  HKEY hKey;
  LONG status;
  DWORD version, length, type;

  // Check if the VTrace registry key exists.  If so, we've already installed
  // VTrace, so any conflicting files we'd find are from VTrace itself and thus
  // there are no conflicting files.

  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                        "Software\\Jay Lorch\\VTrace\\CurrentVersion",
                        0, KEY_QUERY_VALUE, &hKey);
  if (status != ERROR_SUCCESS)
    version = 0;
  else {
    length = sizeof(version);
    status = RegQueryValueEx(hKey, "Version", NULL, &type, (LPBYTE) &version,
                             &length);
    if (status != ERROR_SUCCESS || type != REG_DWORD || length != 4)
      version = 0;
  }
  RegCloseKey(hKey);

  if (version == 0 &&
      (DoesConflictingFileExist(systemRoot, "System32\\VTrace.dll") ||
       DoesConflictingFileExist(systemRoot, "System32\\VTrcSrvc.exe") ||
       DoesConflictingFileExist(systemRoot, "System32\\Drivers\\VTrcFS.sys") ||
       DoesConflictingFileExist(systemRoot,
                                "System32\\Drivers\\VTrcKbd.sys") ||
       DoesConflictingFileExist(systemRoot,
                                "System32\\Drivers\\VTrcLog.sys") ||
       DoesConflictingFileExist(systemRoot,
                                "System32\\Drivers\\VTrcNet.sys") ||
       DoesConflictingFileExist(systemRoot,
                                "System32\\Drivers\\VTrcRawD.sys")))
    return TRUE;
  if (version < VTRACE_VERSION_WITH_NEW_FILES &&
      (DoesConflictingFileExist(systemRoot,
                                "System32\\Drivers\\VTrcFsrc.sys") ||
       DoesConflictingFileExist(systemRoot,
                                "System32\\Drivers\\VTrcInit.sys")))
    return TRUE;

  return FALSE;
}

// CreateMainDirectoryAndTracesSubdirectory creates the main VTrace directory,
// named by the parameter mainDirectory, as well as a subdirectory of that
// named Traces.

BOOL CreateMainDirectoryAndTracesSubdirectory (CString& mainDirectory)
{
  char dirName[MAX_PATH];

  // Create the main directory

  strcpy(dirName, mainDirectory);
  if (!RecursiveCreateDirectory(dirName, NULL)) {
    DisplayError("Could not create main directory");
    return FALSE;
  }

  // Create the Traces subdirectory

  sprintf(dirName, "%s\\Traces", mainDirectory);
  if (!CreateDirectory(dirName, NULL) &&
      GetLastError() != ERROR_ALREADY_EXISTS) {
    DisplayError("Could not create Traces subdirectory in VTrace main "
                 "directory");
    return FALSE;
  }

  return TRUE;
}

// CreateVTraceRegistryKeys creates the VTrace registry key.  Some of the
// pieces of data it must put there are the location of its main directory
// and the personal name given by the user; those are given as parameters to
// this function.

BOOL CreateVTraceRegistryKeys (CString& mainDirectory, CString& personalName)
{
  HKEY hKey, oldKey;
  LONG err;
  DWORD n, type;
  SYSTEMTIME currentTimeST;
  FILETIME currentTimeFT;

  // Create a subkey "Jay Lorch" under software in the registry.  This
  // is the "company" of which VTrace is a "product".

  err = RegCreateKey(HKEY_LOCAL_MACHINE, "Software\\Jay Lorch", &hKey);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not create a registry key for this software");
    return FALSE;
  }

  // Create a subkey "VTrace" under "Jay Lorch" in the registry.  This is
  // where we put all the registry information for VTrace.

  oldKey = hKey;
  err = RegCreateKey(oldKey, "VTrace", &hKey);
  RegCloseKey(oldKey);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not create a registry key for this software");
    return FALSE;
  }

  // Create a subkey "CurrentVersion" under "VTrace" in the registry.  This is
  // where we hold all the information for the current version of VTrace.

  oldKey = hKey;
  err = RegCreateKey(oldKey, "CurrentVersion", &hKey);
  RegCloseKey(oldKey);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not create a registry key for this version of the "
                 "software");
    return FALSE;
  }

  // If an earlier version created an "UploadURL" value, delete it

  RegDeleteValue(hKey, "UploadURL");

  // Set various parameters in the registry

  n = VTRACE_VERSION;
  if (!MyRegSetValue(hKey, "Version", (BYTE *) &n, REG_DWORD, 4,
                     "version number"))
    return FALSE;

  if (!MyRegSetValue(hKey, "MainDirectory", (BYTE *) (LPCTSTR) mainDirectory,
                     REG_SZ, mainDirectory.GetLength() + 1,
                     "VTrace main directory"))
    return FALSE;

  if (!MyRegSetValue(hKey, "PersonalName", (BYTE *) (LPCTSTR) personalName,
                     REG_SZ, personalName.GetLength() + 1, "e-mail address"))
    return FALSE;

  // Set the user number in the registry only if it has not already
  // been set

  err = RegQueryValueEx(hKey, "UserNumber", NULL, &type, NULL, NULL);
  if (err != ERROR_SUCCESS || type != REG_DWORD) {
    srand(time(0));
    rand();                  // ignore the first random number
    n = rand();              // get another random number
    n = (n << 16) + rand();  // shift it left 16 bits and add another random #
    if (!MyRegSetValue(hKey, "UserNumber", (BYTE *) &n, REG_DWORD, 4,
                       "user number"))
      return FALSE;
  }

  // If metadata has never been recorded, get the current time as a FILETIME
  // structure and record one year before that as the last time the metadata
  // was recorded

  err = RegQueryValueEx(hKey, "TimeMetadataRecorded", NULL, &type, NULL, NULL);
  if (err != ERROR_SUCCESS) {
    GetLocalTime(&currentTimeST);
    currentTimeST.wYear--;
    if (SystemTimeToFileTime(&currentTimeST, &currentTimeFT))
      RegSetValueEx(hKey, "TimeMetadataRecorded", 0, REG_BINARY,
                    (BYTE *) &currentTimeFT, 8);
  }

  // Close the registry key

  RegCloseKey(hKey);

  return TRUE;
}

#if TARGET_NT_VERSION == 5

// DisableKernelOverwrite disables kernel overwriting if set by VTrace.

BOOL DisableKernelOverwrite (void)
{
  DWORD status, length, type, n;
  HKEY hKeyVTrace, hKeySystem;

  // Open the VTrace registry key

  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                        "Software\\Jay Lorch\\VTrace\\CurrentVersion",
                        0, KEY_QUERY_VALUE | KEY_SET_VALUE, &hKeyVTrace);
  if (status != ERROR_SUCCESS) {
    SetLastError(status);
    DisplayError("Cannot open VTrace registry key to determine if "
                 "kernel overwrite was enabled");
    return FALSE;
  }

  // Get the EnabledKernelOverwrite value from there. If it's not there or
  // is 0, we're done.

  length = 4;
  status = RegQueryValueEx(hKeyVTrace, "EnabledKernelOverwrite", NULL, &type,
                           (BYTE *) &n, &length);
  if (status != ERROR_SUCCESS) {
    RegCloseKey(hKeyVTrace);
    return TRUE;
  }

  if (type != REG_DWORD || length != 4) {
    RegCloseKey(hKeyVTrace);
    MessageBox(NULL,
               "EnabledKernelOverwrite value in VTrace registry key is not "
               "the right type.", "Error", MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }

  if (n == 0) {
    RegCloseKey(hKeyVTrace);
    return TRUE;
  }

  // Open the memory management registry key

  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
     "System\\CurrentControlSet\\Control\\Session Manager\\Memory Management",
                        0, KEY_QUERY_VALUE | KEY_SET_VALUE, &hKeySystem);
  if (status != ERROR_SUCCESS) {
    RegCloseKey(hKeyVTrace);
    SetLastError(status);
    DisplayError("Could not open system registry key to reestablish kernel "
                 "memory protection");
    return FALSE;
  }

  // Delete the EnforceWriteProtection value from the memory management
  // registry key

  status = RegDeleteValue(hKeySystem, "EnforceWriteProtection");
  if (status != ERROR_SUCCESS) {
    RegCloseKey(hKeyVTrace);
    RegCloseKey(hKeySystem);
    SetLastError(status);
    DisplayError("Could not adjust registry to reestablish kernel "
                 "memory protection");
    return FALSE;
  }

  // Delete the EnabledKernelOverwrite value from the VTrace registry key so we
  // don't try to disable kernel overwrite again.

  status = RegDeleteValue(hKeyVTrace, "EnabledKernelOverwrite");
  if (status != ERROR_SUCCESS) {
    RegCloseKey(hKeyVTrace);
    RegCloseKey(hKeySystem);
    SetLastError(status);
    DisplayError("Could not indicate in VTrace registry key that kernel "
                 "memory protections were restored");
    return FALSE;
  }

  // Close all open keys

  RegCloseKey(hKeyVTrace);
  RegCloseKey(hKeySystem);
  return TRUE;
}

#endif

// InstallDriver installs a given driver (named driverName), dependent on zero
// or more other drivers (specified in dependencies), that starts at a certain
// time (given by startTime), in a certain group (named in group).  To do this,
// it needs an open handle to the service control manager (schSCManager) and
// the name of the system root directory (systemRoot).  It returns a boolean
// indicating whether it was successful.

BOOL InstallDriver (char *driverName, char *dependencies, DWORD startTime,
                    char *group, BOOL isFileSystem, DWORD *tagPtr,
                    SC_HANDLE schSCManager, char *systemRoot)
{
  SC_HANDLE schService;
  char targetFileLocation[MAX_PATH], fileToInstall[MAX_PATH],
       errString[MAX_PATH];
  char oldGroupName[128];
  DWORD maxGroupSize = 128, oldTag;

  // Copy the driver file (ending in .SYS) to the drivers subdirectory of
  // the system32 subdirectory of the system root directory.  This is a good
  // place for it, because it is guaranteed to be available for reading at the
  // time that drivers are loaded (as opposed to the VTrace main directory,
  // which might be on a remote file system).

  sprintf(fileToInstall, "%s.SYS", driverName);
  sprintf(targetFileLocation, "%s\\SYSTEM32\\DRIVERS\\%s.SYS", systemRoot,
          driverName);
  if (!CopyFile(fileToInstall, targetFileLocation, FALSE)) {
    sprintf(errString, "Could not copy %s.sys to the drivers directory",
            driverName);
    DisplayError(errString);
    return FALSE;
  }

  // Set targetFileLocation to contain the name of the new driver location, in
  // general terms (using %systemroot% to denote the system root)

  sprintf(targetFileLocation, "%%systemroot%%\\system32\\drivers\\%s.sys",
          driverName);

  // Create the service for this driver.

  schService = OpenService(schSCManager, driverName, SERVICE_ALL_ACCESS);
  if (schService == NULL)
    schService = CreateService(schSCManager, driverName, driverName,
                               SERVICE_ALL_ACCESS,
                               isFileSystem ? SERVICE_FILE_SYSTEM_DRIVER
                                            : SERVICE_KERNEL_DRIVER,
                               startTime,
                               SERVICE_ERROR_NORMAL, targetFileLocation,
                               group[0] ? group : NULL,
                               tagPtr,
                               dependencies[0] ? dependencies : NULL,
                               NULL, NULL);
  else {
    // If the driver belongs to a different group than we now are adding
    // it to, then remove its tag from that group.

    if (GetTagAndGroup(schService, &oldTag, oldGroupName, maxGroupSize) &&
        strcmp(oldGroupName, group))
      RemoveTagFromGroup(oldTag, oldGroupName);

    // Change the service parameters to the appropriate new ones.

    if (!ChangeServiceConfig(schService,
                             isFileSystem ? SERVICE_FILE_SYSTEM_DRIVER
                                          : SERVICE_KERNEL_DRIVER,
                             startTime, SERVICE_ERROR_NORMAL, NULL, group,
                             tagPtr, dependencies, NULL, NULL, NULL)) {
      sprintf(errString, "Could not change configuration of %s service",
              driverName);
      DisplayError(errString);
      return FALSE;
    }
  }

  if (schService == NULL && GetLastError() != ERROR_SERVICE_EXISTS) {
    sprintf(errString, "Could not create %s service", driverName);
    DisplayError(errString);
    return FALSE;
  }

  // Creating the service opens a handle to it, which we don't need, so close
  // that handle

  if (schService != NULL)
    CloseServiceHandle(schService);

  return TRUE;
}

// InstallDLL installs the VTrace DLL.  To do this, it needs the name of the
// system root directory (given in systemRoot).

BOOL InstallDLL (char *systemRoot)
{
  LONG err;
  HKEY hKey;
  DWORD type, length;
  char appInitDLLs[MAX_PATH];
  char realTargetFileName[MAX_PATH], temporaryTargetFileName[MAX_PATH];

  // Copy VTRACE.DLL to the system32 subdirectory of the system root directory.
  // Because VTRACE.DLL might already exist and be open (e.g. if this is
  // an upgrade to a running version of VTrace), we cannot simply overwrite it.
  // Instead, we copy VTRACE.DLL to a temporary location in the System32
  // directory and arrange for it to be moved to the proper location at
  // reboot.  We put the temporary location in the System32 directory,
  // since that location will always be available at reboot.

  sprintf(realTargetFileName, "%s\\SYSTEM32\\VTRACE.DLL", systemRoot);
  sprintf(temporaryTargetFileName, "%s\\SYSTEM32\\TempVTraceFileVTRACE.DLL",
          systemRoot);
  if (!CopyFile("VTRACE.DLL", temporaryTargetFileName, FALSE)) {
    DisplayError("Could not copy VTRACE.DLL to the System32 directory");
    return FALSE;
  }
  if (!MoveFileEx(temporaryTargetFileName, realTargetFileName,
                  MOVEFILE_DELAY_UNTIL_REBOOT | MOVEFILE_REPLACE_EXISTING)) {
    DisplayError("Could not install VTRACE.DLL in the System32 directory");
    DeleteFile(temporaryTargetFileName);
    return FALSE;
  }

  // Open the Windows software registry key

  err = RegOpenKeyEx
          (HKEY_LOCAL_MACHINE,
           "Software\\Microsoft\\Windows NT\\CurrentVersion\\Windows", 0,
           KEY_QUERY_VALUE | KEY_SET_VALUE, &hKey);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not open registry key for Windows software");
    return FALSE;
  }

  // Query the value of "AppInit_DLLs", which returns a space-separated list
  // of all the DLL's that are automatically loaded into each application

  length = sizeof(appInitDLLs);
  err = RegQueryValueEx(hKey, "AppInit_DLLs", 0, &type, (BYTE *) appInitDLLs,
                        &length);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not get AppInit_DLLs registry value");
    RegCloseKey(hKey);
    return FALSE;
  }
  if (type != REG_SZ) {
    MessageBox(NULL, "AppInit_DLLs registry value is not a string", "Error",
               MB_OK | MB_ICONEXCLAMATION);
    RegCloseKey(hKey);
    return FALSE;
  }

  // If VTRACE.DLL does not already appear on that list, append it to the end
  // of the list and change the "AppInit_DLLs" value to be the new list

  if (!ContainsVTraceDLL(appInitDLLs)) {
    if (appInitDLLs[0] == '\0')
      strcpy(appInitDLLs, "VTRACE.DLL");
    else
      strcat(appInitDLLs, " VTRACE.DLL");
    err = RegSetValueEx(hKey, "AppInit_DLLs", 0, REG_SZ, (BYTE *) appInitDLLs,
                        strlen(appInitDLLs) + 1 /* includes trailing null */);
    if (err != ERROR_SUCCESS) {
      SetLastError(err);
      DisplayError("Could not set AppInit_DLLs registry value");
      RegCloseKey(hKey);
      return FALSE;
    }
  }

  // Close the Windows software registry key

  RegCloseKey(hKey);

  return TRUE;
}

// InstallVTrcSrvc installs the user-mode service VTrcSrvc.  To do this, it
// needs the name of the system root directory (systemRoot) and an open handle
// to the service control manager (schSCManager).  It returns a boolean
// indicating whether it was successful.

BOOL InstallVTrcSrvc (char *systemRoot, SC_HANDLE schSCManager)
{
  SC_HANDLE schService;
  char realTargetFileName[MAX_PATH], temporaryTargetFileName[MAX_PATH];

  // We want to put VTrcSrvc.EXE in the system32 directory.  This is a good
  // place for it, as it is guaranteed to be readable at system boot (unlike,
  // say, a remote drive, which the main VTrace directory might be).

  // Because VTrcSrvc.EXE might already exist and be open (e.g. if this is
  // an upgrade to a running version of VTrace), we cannot simply overwrite it.
  // Instead, we copy VTrcSrvc.EXE to a temporary location in the System32
  // directory and arrange for it to be moved to the proper location at
  // reboot.  We put the temporary location in the System32 directory,
  // since that location will always be available at reboot.

  sprintf(realTargetFileName, "%s\\SYSTEM32\\VTrcSrvc.EXE", systemRoot);
  sprintf(temporaryTargetFileName, "%s\\SYSTEM32\\TempVTraceFileVTrcSrvc.EXE",
          systemRoot);
  if (!CopyFile("VTrcSrvc.EXE", temporaryTargetFileName, FALSE)) {
    DisplayError("Could not copy VTrcSrvc.exe to the System32 directory");
    return FALSE;
  }
  if (!MoveFileEx(temporaryTargetFileName, realTargetFileName,
                  MOVEFILE_DELAY_UNTIL_REBOOT | MOVEFILE_REPLACE_EXISTING)) {
    DisplayError("Could not install VTrcSrvc.exe in the System32 directory");
    DeleteFile(temporaryTargetFileName);
    return FALSE;
  }

  // Create the service, starting at boot time and dependent on VTrcLog.
  // (It needs to load after VTrcLog so it can get a handle to it and
  // communicate with it.)

  schService = CreateService(schSCManager, "VTrcSrvc", "VTrcSrvc",
                             SERVICE_ALL_ACCESS, SERVICE_WIN32_OWN_PROCESS,
                             SERVICE_AUTO_START, SERVICE_ERROR_NORMAL,
                             "%systemroot%\\system32\\VTrcSrvc.exe",
                             NULL, NULL, "VTrcLog\0\0", NULL, NULL);

  if (schService == NULL && GetLastError() != ERROR_SERVICE_EXISTS) {
    DisplayError("Could not create VTrcSrvc service");
    return FALSE;
  }

  // CreateService, if successful, opens a handle to the new service, which we
  // don't need.  So, close that handle.

  if (schService != NULL)
    CloseServiceHandle(schService);

  return TRUE;
}

// CreateSystemConfigurationFile creates the system configuration file
// SysInfo.txt in the Traces subdirectory, containing information about
// the system.  This is done at installation because that's a convenient
// time for it.  It returns a boolean indicating whether it was successful.
// It takes three parameters:  mainDirectory, which indicates where the VTrace
// main directory is (and therefore where to find the Traces subdirectory);
// personalName, which indicates the name the user entered; and refTime,
// which is a reference time established earlier which we use to compute the
// clock speed.

BOOL CreateSystemConfigurationFile (CString& mainDirectory,
                                    CString& personalName,
                                    REF_TIME& refTime)
{
  char fileName[MAX_PATH];
  HANDLE hfile;
  SYSTEM_INFO sysInfo;
  CString str;
  DWORD bytesWritten, cpuSpeed;
  OSVERSIONINFO versionInfo;
  MEMORYSTATUS memoryStatus;
  char driveName[3] = "A:";
  char rootFileName[4] = "A:\\";
  char driveLetter;
  DWORD diskNumber, partitionNumber, volumeNumber, bytesPerSector,
        sectorsPerCluster, freeClusters, totalClusters;
  SYSTEMTIME localTime;
  TIME_ZONE_INFORMATION timeZoneInfo;

  // Get the CPU speed and other system information

  GetSystemInfo(&sysInfo);
  versionInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
  if (!GetVersionEx(&versionInfo)) return FALSE;
  if (!GetLocalCPUSpeed(&cpuSpeed, refTime)) return FALSE;

  // Create the file SysInfo.txt in the Traces subdirectory, and get a handle
  // to it

  sprintf(fileName, "%s\\Traces\\SysInfo.txt", mainDirectory);
  hfile = CreateFile(fileName, GENERIC_READ | GENERIC_WRITE, 0, NULL,
                     CREATE_ALWAYS,
                     FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL);
  if (hfile == INVALID_HANDLE_VALUE) {
    DisplayError("Cannot create system information file");
    return FALSE;
  }

  // Output information about the version of VTrace

  str.Format("VTrace version: %d.%d.%d\r\n", VTRACE_VERSION >> 16,
             (VTRACE_VERSION >> 8) % 256, VTRACE_VERSION % 256);
  WriteFile(hfile, (LPCSTR) str, str.GetLength(), &bytesWritten, NULL);

  // Output information about the processor type

  str.Format("Processor: Architecture %d, level %d, revision 0x%X\r\n",
             sysInfo.wProcessorArchitecture, sysInfo.wProcessorLevel,
             sysInfo.wProcessorRevision);
  WriteFile(hfile, (LPCSTR) str, str.GetLength(), &bytesWritten, NULL);

  // Output information about the processor speed

  str.Format("Processor speed: %d Hz\r\n", cpuSpeed);
  WriteFile(hfile, (LPCSTR) str, str.GetLength(), &bytesWritten, NULL);

  // Output information about the number of processors

  str.Format("Number of processors: %d\r\n", sysInfo.dwNumberOfProcessors);
  WriteFile(hfile, (LPCSTR) str, str.GetLength(), &bytesWritten, NULL);

  // Output personal name of user

  str.Format("User e-mail address: %s\r\n", personalName);
  WriteFile(hfile, (LPCSTR) str, str.GetLength(), &bytesWritten, NULL);

  // Output the OS version information

  str.Format("NT version: %d.%d build %d\r\n", versionInfo.dwMajorVersion,
             versionInfo.dwMinorVersion, versionInfo.dwBuildNumber);
  WriteFile(hfile, (LPCSTR) str, str.GetLength(), &bytesWritten, NULL);

  // Output the service pack information

  str.Format("Last service pack: %s\r\n", versionInfo.szCSDVersion);
  WriteFile(hfile, (LPCSTR) str, str.GetLength(), &bytesWritten, NULL);

  // Output the memory size

  GlobalMemoryStatus(&memoryStatus);
  str.Format("Memory bytes: %lu\r\n", memoryStatus.dwTotalPhys);
  WriteFile(hfile, (LPCSTR) str, str.GetLength(), &bytesWritten, NULL);

  // Output sizes of fixed disks

  for (driveLetter = 'A'; driveLetter <= 'Z'; driveLetter++) {
    // Check if it is a fixed drive.  If so, get the disk information and
    // make sure that begins with \Device\Harddisk
    driveName[0] = driveLetter;
    rootFileName[0] = driveLetter;
    if (GetDriveType(driveName) == DRIVE_FIXED &&
        GetDiskNumbers(driveLetter, &diskNumber, &partitionNumber,
                       &volumeNumber) &&
        GetDiskFreeSpace(rootFileName, &sectorsPerCluster, &bytesPerSector,
                         &freeClusters, &totalClusters)) {
      if (volumeNumber == 0xFFFFFFFFUL)
        str.Format("Size of drive %c (disk %ld, partition %ld): "
                   "%ld x %ld x %ld\r\n",
                   driveLetter, diskNumber, partitionNumber,
                   bytesPerSector, sectorsPerCluster, totalClusters);
      else
        str.Format("Size of drive %c (disk %ld, partition %ld, volume %ld): "
                   "%ld x %ld x %ld\r\n",
                   driveLetter, diskNumber, partitionNumber, volumeNumber,
                   bytesPerSector, sectorsPerCluster, totalClusters);
      WriteFile(hfile, (LPCSTR) str, str.GetLength(), &bytesWritten, NULL);
    }
  }

  // Output the VTrace directory information

  str.Format("VTrace directory: %s\r\n", mainDirectory);
  WriteFile(hfile, (LPCSTR) str, str.GetLength(), &bytesWritten, NULL);

  // Output today's date

  GetLocalTime(&localTime);
  str.Format("Date of install: %d/%d/%d\r\n", localTime.wMonth, localTime.wDay,
             localTime.wYear);
  WriteFile(hfile, (LPCSTR) str, str.GetLength(), &bytesWritten, NULL);

  // Output time zone information

  GetTimeZoneInformation(&timeZoneInfo);
  str.Format("UTC offset bias: %ld\r\n", timeZoneInfo.Bias);
  WriteFile(hfile, (LPCSTR) str, str.GetLength(), &bytesWritten, NULL);
  str.Format("Daylight savings time bias: %ld\r\n",
             timeZoneInfo.DaylightDate.wMonth ? timeZoneInfo.DaylightBias : 0);
  WriteFile(hfile, (LPCSTR) str, str.GetLength(), &bytesWritten, NULL);

  // Close the open file handle

  CloseHandle(hfile);

  return TRUE;
}

// InstallUtilities installs various utility programs in the VTrace main
// directory (mainDirectory).  For those of them that the user might want
// to run, it puts shortcuts to them in the Start menu, under a new
// subdirectory VTrace <version>.  It returns a boolean indicating whether
// it was successful.  It takes an additional parameter, previousVersion,
// that indicates the previous version of VTrace that was installed, so
// that if this value is non-zero the old program folder for that version
// can be deleted.

BOOL InstallUtilities (CString& mainDirectory, DWORD previousVersion)
{
  char sourceFile[MAX_PATH], destFile[MAX_PATH],
       topLevelProgramFolder[MAX_PATH];
  CString programFolder, versionAsString;
  LPITEMIDLIST idList;

  // Find the program group directory name

  if (SHGetSpecialFolderLocation(NULL, CSIDL_COMMON_PROGRAMS, &idList) !=
      NO_ERROR) {
    DisplayError("Could not find Start menu directory");
    return FALSE;
  }
  if (!SHGetPathFromIDList(idList, topLevelProgramFolder)) {
    DisplayError("Could not get Start menu directory name");
    return FALSE;
  }

  // Add \VTrace {version number} to the end of the directory name to get the
  // program folder name for VTrace.

  ConvertVersionNumberToString(versionAsString, VTRACE_VERSION);
  programFolder = topLevelProgramFolder;
  programFolder += "\\VTrace ";
  programFolder += versionAsString;

  // If there was a previous version installed, try to rename the old program
  // folder to the new name

  if (previousVersion != 0 && previousVersion != VTRACE_VERSION) {
    CString oldProgramFolder;
    ConvertVersionNumberToString(versionAsString, previousVersion);
    oldProgramFolder = topLevelProgramFolder;
    oldProgramFolder += "\\VTrace ";
    oldProgramFolder += versionAsString;
    MoveFile(oldProgramFolder, programFolder);
  }

  // Create the new program folder for VTrace

  if (!CreateDirectory(programFolder, NULL) &&
      GetLastError() != ERROR_ALREADY_EXISTS) {
    DisplayError("Could not create VTrace subdirectory in Start menu");
    return FALSE;
  }

  // Initialize COM, to be used for creating shortcuts

  if (!SUCCEEDED(CoInitialize(NULL))) {
    DisplayError("Could not initialize the Component Object Model (COM)");
    return FALSE;
  }

  // Copy the Readme file to the main VTrace directory, and create a shortcut
  // to it in the VTrace program folder

  sprintf(sourceFile, "Readme.txt");
  sprintf(destFile, "%s\\Readme.txt", mainDirectory);
  if (!CopyFile(sourceFile, destFile, FALSE)) {
    DisplayError("Could not copy Readme text file to VTrace directory");
    CoUninitialize();
    return FALSE;
  }
  sprintf(sourceFile, "%s\\Readme file.lnk", programFolder);
  if (!CreateShortcut(sourceFile, destFile)) {
    DisplayError("Could not create shortcut to Readme file in the Start menu");
    CoUninitialize();
    return FALSE;
  }

  // Copy the Override program to the main VTrace directory, and create a
  // shortcut to it in the VTrace program folder

  sprintf(sourceFile, "Override.exe");
  sprintf(destFile, "%s\\Override.exe", mainDirectory);
  if (!CopyFile(sourceFile, destFile, FALSE)) {
    DisplayError("Could not copy Override program to VTrace directory");
    CoUninitialize();
    return FALSE;
  }
  sprintf(sourceFile, "%s\\Temporarily turn off VTrace.lnk", programFolder);
  if (!CreateShortcut(sourceFile, destFile)) {
    DisplayError("Could not create shortcut to Override program in the "
                 "Start menu");
    CoUninitialize();
    return FALSE;
  }

  // Copy the Upload program to the main VTrace directory, and create a
  // shortcut to it in the VTrace program folder

  sprintf(sourceFile, "Upload.exe");
  sprintf(destFile, "%s\\Upload.exe", mainDirectory);
  if (!CopyFile(sourceFile, destFile, FALSE)) {
    DisplayError("Could not copy Upload program to VTrace directory");
    CoUninitialize();
    return FALSE;
  }
  sprintf(sourceFile, "%s\\Upload files via the Internet.lnk", programFolder);
  if (!CreateShortcut(sourceFile, destFile)) {
    DisplayError("Could not create shortcut to Upload program in the Start "
                 "menu");
    CoUninitialize();
    return FALSE;
  }

  // Copy the Uninstall program to the main VTrace directory, and create a
  // shortcut to it in the VTrace program folder

  sprintf(sourceFile, "Uninstal.exe");
  sprintf(destFile, "%s\\Uninstal.exe", mainDirectory);
  if (!CopyFile(sourceFile, destFile, FALSE)) {
    DisplayError("Could not copy Uninstall program to VTrace directory");
    CoUninitialize();
    return FALSE;
  }
  sprintf(sourceFile, "%s\\Uninstall VTrace.lnk", programFolder);
  if (!CreateShortcut(sourceFile, destFile)) {
    DisplayError("Could not create shortcut to Uninstall program in the Start "
                 "menu");
    CoUninitialize();
    return FALSE;
  }

  // Copy the Config program to the main VTrace directory, and create a
  // shortcut to it in the VTrace program folder

  sprintf(sourceFile, "Config.exe");
  sprintf(destFile, "%s\\Config.exe", mainDirectory);
  if (!CopyFile(sourceFile, destFile, FALSE)) {
    DisplayError("Could not copy Config program to VTrace directory");
    CoUninitialize();
    return FALSE;
  }
  sprintf(sourceFile, "%s\\Advanced configuration for VTrace.lnk",
          programFolder);
  if (!CreateShortcut(sourceFile, destFile)) {
    DisplayError("Could not create shortcut to Config program in the Start "
                 "menu");
    CoUninitialize();
    return FALSE;
  }

  // Close COM now that we're done with it

  CoUninitialize();

  // Copy the Metadata program to the main VTrace directory, where VTrcSrvc
  // can easily find it

  sprintf(sourceFile, "Metadata.exe");
  sprintf(destFile, "%s\\Metadata.exe", mainDirectory);
  if (!CopyFile(sourceFile, destFile, FALSE)) {
    DisplayError("Could not copy Metadata program to VTrace directory");
    return FALSE;
  }

  // Copy the Gzip program to the main VTrace directory, where our programs
  // can easily find it

  sprintf(sourceFile, "Gzip.exe");
  sprintf(destFile, "%s\\Gzip.exe", mainDirectory);
  if (!CopyFile(sourceFile, destFile, FALSE)) {
    DisplayError("Could not copy Gzip program to VTrace directory");
    return FALSE;
  }

  // Copy the GNU DLL to the main VTrace directory, where it can be used by
  // Gzip.exe.

  sprintf(sourceFile, "Win32Gnu.DLL");
  sprintf(destFile, "%s\\Win32Gnu.DLL", mainDirectory);
  if (!CopyFile(sourceFile, destFile, FALSE)) {
    DisplayError("Could not copy GNU DLL to VTrace directory");
    return FALSE;
  }

  return TRUE;
}

// AdvancedConfiguration launches the Config.exe program in the current
// directory to perform advanced configuration of VTrace.

void AdvancedConfiguration (void)
{
  STARTUPINFO startupInfo;
  PROCESS_INFORMATION processInfo;

  GetStartupInfo(&startupInfo);
  CreateProcess("Config.exe", NULL, NULL, NULL, FALSE, DETACHED_PROCESS,
                NULL, NULL, &startupInfo, &processInfo);
}

// RemoveActivateVTraceNextTimeKey removes the obsolete key that activates
// VTrace at the next boot-up.

void RemoveActivateVTraceNextTimeKey (void)
{
  HKEY hKey;
  LONG err;

  err = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                     "System\\CurrentControlSet\\Services\\VTrcRawD", 0,
                     KEY_WRITE, &hKey);
  if (err != ERROR_SUCCESS)
    return;

  RegDeleteValue(hKey, "ActivateNextTime");
  RegDeleteValue(hKey, "ActivateThisTime");
  RegCloseKey(hKey);
}
