//
// RefTime.cpp contains routines for getting and using reference times.
//

#include "stdafx.h"
#include "RefTime.h"

// GetCurrentRefTime fills in the given reference time structure with the
// current time.  It returns a boolean indicating if it was successful.

BOOL GetCurrentRefTime (REF_TIME& refTime)
{
  LARGE_INTEGER perfCounter;
  unsigned long *bufPtr = (unsigned long *) &refTime.cycleCount;

  __asm {
    _emit 0x0F
    _emit 0x31
    mov ebx, bufPtr
    mov [ebx], eax
    mov [ebx+4], edx
  }

  // For some reason, if we pass &refTime.performanceCounter into
  // QueryPerformanceCounter instead of a pointer to a local variable, it
  // fails, so we keep it happy in this way

  if (QueryPerformanceCounter(&perfCounter)) {
    refTime.performanceCounter = perfCounter;
    return TRUE;
  }
  else
    return FALSE;
}

// DeduceCpuSpeedFromOldRefTime deduces the local machine's CPU speed in Hz
// from a reference time taken in the past.  That old reference time is given
// by oldRefTime, and this function places its result in the location pointed
// to by cpuSpeedPtr.  It returns a boolean indicating if it was successful.

BOOL DeduceCpuSpeedFromOldRefTime (REF_TIME& oldRefTime, DWORD *cpuSpeedPtr)
{
  REF_TIME newRefTime;
  LARGE_INTEGER performanceFrequency, diffCycles;
  double diffSec;

  // Get the performance counter frequency

  if (!QueryPerformanceFrequency(&performanceFrequency))
    return FALSE;

  // Get the current reference time

  if (!GetCurrentRefTime(newRefTime))
    return FALSE;

  // Compute the difference in seconds between the reference times

  diffSec = (newRefTime.performanceCounter.QuadPart -
             oldRefTime.performanceCounter.QuadPart) * 1.0 /
             performanceFrequency.QuadPart;
  if (diffSec < 0.0)
    return FALSE;

  // If there were less than 5 seconds between the reference times,
  // wait long enough to make there be at least 5 seconds and take a new
  // current reference time (and compute the difference in seconds again)

  if (diffSec < 5.0) {
    Sleep((unsigned long) ((5.5 - diffSec) * 1000));
    if (!GetCurrentRefTime(newRefTime))
      return FALSE;
    diffSec = (newRefTime.performanceCounter.QuadPart -
               oldRefTime.performanceCounter.QuadPart) * 1.0 /
               performanceFrequency.QuadPart;
    if (diffSec < 5.0)
      return FALSE;
  }

  // Compute the CPU speed in Hz by dividing the difference in cycles
  // between the two reference times by the difference in seconds between
  // them

  diffCycles.QuadPart = newRefTime.cycleCount.QuadPart -
                        oldRefTime.cycleCount.QuadPart;
  *cpuSpeedPtr = (DWORD) (diffCycles.QuadPart / diffSec + 0.5);
  return TRUE;
}
