//
// UninstallRoutines.cpp contains routines for uninstalling VTrace in stages.
//

#include "stdafx.h"
#include <shlobj.h>
#include <winioctl.h>
#include <winsvc.h>
#include "../Common/VTrace.h"

////////////////////////////////////////////////////////////////////////////
//                                MACROS
////////////////////////////////////////////////////////////////////////////

#define INITIAL_GUESS_SERVICE_CONFIG_SIZE       256

////////////////////////////////////////////////////////////////////////////
//                             ERROR DISPLAY
////////////////////////////////////////////////////////////////////////////

// DisplayError displays, in a message box, the error message given
// in the string generalReason as well as the error message implied
// by the system error state.

void DisplayError (char *generalReason)
{
  char *lpMsgBuf;
  CString str;

  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                NULL, GetLastError(),
                MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &lpMsgBuf,
                0, NULL);
  str.Format("%s because %s", generalReason, lpMsgBuf);
  MessageBox(NULL, str, "Error", MB_OK | MB_ICONEXCLAMATION);

  LocalFree(lpMsgBuf);
}

////////////////////////////////////////////////////////////////////////////
//                            REGISTRY ACCESS
////////////////////////////////////////////////////////////////////////////

// GetMainDirectory gets the main directory name from the registry and puts
// it in the buffer given (mainDirectory, with size bufferSize).  It returns
// a boolean indicating whether it was successful.

BOOL GetMainDirectory (char *mainDirectory, int bufferSize)
{
  HKEY hKey;
  LONG status;
  DWORD type, length;

  // Open the VTrace registry key

  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                        "Software\\Jay Lorch\\VTrace\\CurrentVersion",
                        0, KEY_QUERY_VALUE, &hKey);
  if (status != ERROR_SUCCESS) {
    SetLastError(status);
    DisplayError("Cannot open VTrace registry key to obtain main directory");
    return FALSE;
  }

  // Read the MainDirectory value from the key

  length = bufferSize;
  status = RegQueryValueEx(hKey, "MainDirectory", NULL, &type,
                           (BYTE *) mainDirectory, &length);
  RegCloseKey(hKey);
  if (status != ERROR_SUCCESS) {
    SetLastError(status);
    DisplayError("Cannot read VTrace registry key to obtain main directory");
    return FALSE;
  }
  if (type != REG_SZ) {
    MessageBox(NULL, "Main directory in VTrace registry key not a string",
               "Error", MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }

  return TRUE;
}

// GetVTraceVersion reads VTrace version number from the registry and puts it
// in the given pointer location (vtraceVersionPtr).  It returns a boolean
// indicating whether it was successful.

BOOL GetVTraceVersion (DWORD *vtraceVersionPtr)
{
  DWORD status, length, type;
  HKEY hKey;

  // Open the VTrace registry key

  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                        "Software\\Jay Lorch\\VTrace\\CurrentVersion",
                        0, KEY_QUERY_VALUE, &hKey);
  if (status != ERROR_SUCCESS) {
    SetLastError(status);
    DisplayError("Cannot open VTrace registry key to obtain version number");
    return FALSE;
  }

  // Read the Version value from the key

  length = sizeof(DWORD);
  status = RegQueryValueEx(hKey, "Version", NULL, &type,
                           (BYTE *) vtraceVersionPtr, &length);
  RegCloseKey(hKey);
  if (status != ERROR_SUCCESS) {
    SetLastError(status);
    DisplayError("Cannot read VTrace registry key to obtain VTrace version "
                 "number");
    return FALSE;
  }
  if (type != REG_DWORD || length != sizeof(DWORD)) {
    MessageBox(NULL,
               "Version number in VTrace registry key is not the right type.",
               "Error", MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }

  return TRUE;
}

// DisableKernelOverwrite disables kernel overwriting if set by VTrace.

BOOLEAN DisableKernelOverwrite (void)
{
  DWORD status, length, type, n;
  HKEY hKeyVTrace, hKeySystem;

  // Open the VTrace registry key

  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                        "Software\\Jay Lorch\\VTrace\\CurrentVersion",
                        0, KEY_QUERY_VALUE | KEY_SET_VALUE, &hKeyVTrace);
  if (status != ERROR_SUCCESS) {
    SetLastError(status);
    DisplayError("Cannot open VTrace registry key to determine if "
                 "kernel overwrite was enabled");
    return FALSE;
  }

  // Get the EnabledKernelOverwrite value from there. If it's not there or
  // is 0, we're done.

  length = 4;
  status = RegQueryValueEx(hKeyVTrace, "EnabledKernelOverwrite", NULL, &type,
                           (BYTE *) &n, &length);
  if (status != ERROR_SUCCESS) {
    RegCloseKey(hKeyVTrace);
    return TRUE;
  }

  if (type != REG_DWORD || length != 4) {
    RegCloseKey(hKeyVTrace);
    MessageBox(NULL,
               "EnabledKernelOverwrite value in VTrace registry key is not "
               "the right type.", "Error", MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }

  if (n == 0) {
    RegCloseKey(hKeyVTrace);
    return TRUE;
  }

  // Open the memory management registry key

  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
     "System\\CurrentControlSet\\Control\\Session Manager\\Memory Management",
                        0, KEY_QUERY_VALUE | KEY_SET_VALUE, &hKeySystem);
  if (status != ERROR_SUCCESS) {
    RegCloseKey(hKeyVTrace);
    SetLastError(status);
    DisplayError("Could not open system registry key to reestablish kernel "
                 "memory protection");
    return FALSE;
  }

  // Delete the EnforceWriteProtection value from the memory management
  // registry key

  status = RegDeleteValue(hKeySystem, "EnforceWriteProtection");
  if (status != ERROR_SUCCESS) {
    RegCloseKey(hKeyVTrace);
    RegCloseKey(hKeySystem);
    SetLastError(status);
    DisplayError("Could not adjust registry to reestablish kernel "
                 "memory protection");
    return FALSE;
  }

  // Delete the EnabledKernelOverwrite value from the VTrace registry key so we
  // don't try to disable kernel overwrite again.

  status = RegDeleteValue(hKeyVTrace, "EnabledKernelOverwrite");
  if (status != ERROR_SUCCESS) {
    RegCloseKey(hKeyVTrace);
    RegCloseKey(hKeySystem);
    SetLastError(status);
    DisplayError("Could not indicate in VTrace registry key that kernel "
                 "memory protections were restored");
    return FALSE;
  }

  // Close all open keys

  RegCloseKey(hKeyVTrace);
  RegCloseKey(hKeySystem);
  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
//                            STRING ROUTINES
////////////////////////////////////////////////////////////////////////////

// RemoveVTraceDLL removes VTRACE.DLL from the given string (str) listing a
// space-separated list of dynamic link library names.  It returns a boolean
// indicating whether it found such a substring to remove.

BOOL RemoveVTraceDLL (char *str)
{
  char *p, *q;

  // Loop, each time finding the next instance of the substring VTRACE.DLL
  // and checking to see if it is surrounded on the left and right by
  // spaces and/or beginning-or-end-of-string.  If such a bounded substring
  // is found, it is extracted.

  for (p = str; 1; p += 10) {
    // Find the next VTRACE.DLL at 'p'.  If not found, return FALSE
    p = strstr(p, "VTRACE.DLL");
    if (p == NULL) return FALSE;

    if (p == str) {            // Case 1: <start-of-str>VTRACE.DLL
      if (*(p+10) == '\0') {   // Case 1a: <start-of-str>VTRACE.DLL<end-of-str>
        *p = '\0';             //  ... turn str into null string
        return TRUE;
      }
      if (*(p+10) == ' ') {    // Case 1b: <start-of-str>VTRACE.DLL<space>
        for (q = p; 1; q++) {  //   ... copy stuff after space to start-of-str
          *q = *(q+11);
          if (*q == '\0') break;
        }
        return TRUE;
      }
    }
    else if (*(p-1) == ' ') {  // Case 2: <space>VTRACE.DLL
      if (*(p+10) == '\0') {   // Case 2a: <space>VTRACE.DLL<end-of-str>
        *(p-1) = '\0';         //  ... terminate string where space is
        return TRUE;
      }
      if (*(p+10) == ' ') {    // Case 2b: <space>VTRACE.DLL<space>
        for (q = p; 1; q++) {  //  ... copy stuff after end space to where V is
          *q = *(q+11);
          if (*q == '\0') break;
        }
        return TRUE;
      }
    }
  }

  return FALSE;
}

////////////////////////////////////////////////////////////////////////////
//                       DIRECTORY DELETION ROUTINES
////////////////////////////////////////////////////////////////////////////

// RemoveContentsOfDirectory removes the contents of the directory named in
// the given string (directory).  It returns a boolean indicating whether it
// was successful.

BOOLEAN RemoveContentsOfDirectory (char *directory)
{
  char searchPattern[MAX_PATH], otherString[MAX_PATH];
  WIN32_FIND_DATA findData;
  HANDLE findHandle;

  sprintf(searchPattern, "%s\\*.*", directory);

  // Loop, each time deleting the first file found matching the pattern that
  // is not a directory (like ".").  (We don't just loop through them all using
  // FindNextFile, since DeleteFile may mess up the search.)

  while (1) {
    // Search for the first file that isn't a directory.  If one cannot be
    // found, return TRUE since the deletion is complete.

    findHandle = FindFirstFile(searchPattern, &findData);
    if (findHandle == INVALID_HANDLE_VALUE)
      return TRUE;

    while (findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
      if (!FindNextFile(findHandle, &findData)) {
        FindClose(findHandle);
        return TRUE;
      }
    FindClose(findHandle);

    // Delete the file found

    sprintf(otherString, "%s\\%s", directory, findData.cFileName);
    if (!DeleteFile(otherString)) {
      char errString[MAX_PATH+30];
      sprintf(errString, "Could not delete file %s", findData.cFileName);
      DisplayError(errString);
      return FALSE;
    }
  }

  return TRUE;
}

// RemoveContentsOfDirectoryAtReboot arranges for all the files in the
// given directory (name given in directory) to be deleted at reboot.
// It returns a boolean indicating whether it was successful.

BOOLEAN RemoveContentsOfDirectoryAtReboot (char *directory)
{
  char searchPattern[MAX_PATH], fileName[MAX_PATH];
  WIN32_FIND_DATA findData;
  HANDLE findHandle;
  BOOLEAN retval = TRUE;

  sprintf(searchPattern, "%s\\*.*", directory);

  // Find the first file in the directory.  If it doesn't exist, we have
  // succeeded trivially.

  findHandle = FindFirstFile(searchPattern, &findData);
  if (findHandle == INVALID_HANDLE_VALUE)
    return TRUE;

  // If a file was found, loop through each file in the directory, arranging
  // for each to be deleted at reboot if it is not a directory (like ".")

  do {
    if (!(findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
      sprintf(fileName, "%s\\%s", directory, findData.cFileName);
      if (!MoveFileEx(fileName, NULL, MOVEFILE_DELAY_UNTIL_REBOOT)) {
        char errString[MAX_PATH+30];
        sprintf(errString, "Could not delete file %s", findData.cFileName);
        DisplayError(errString);
        retval = FALSE;
      }
    }
  } while (FindNextFile(findHandle, &findData));
  FindClose(findHandle);

  return retval;
}

////////////////////////////////////////////////////////////////////////////
//                        PROGRAM FOLDER ROUTINES
////////////////////////////////////////////////////////////////////////////

// GetVTraceProgramFolder gets the name of the VTrace subfolder of the main
// program folder, whichever version it is.  It returns that subdirectory
// name in programFolder, and returns a boolean indicating if it was
// successful.

BOOLEAN GetVTraceProgramFolder (char *programFolder)
{
  LPITEMIDLIST idList;
  DWORD vtraceVersion;

  // Get the version number from the registry

  if (!GetVTraceVersion(&vtraceVersion)) return FALSE;

  // Find the main program folder

  if (SHGetSpecialFolderLocation(NULL, CSIDL_COMMON_PROGRAMS, &idList) !=
      NO_ERROR) {
    DisplayError("Could not find Start menu location");
    return FALSE;
  }
  if (!SHGetPathFromIDList(idList, programFolder)) {
    DisplayError("Could not get Start menu directory name");
    return FALSE;
  }

  // Append "VTrace" and the version number to the program folder name

  if (vtraceVersion % 256 == 0)
    sprintf(&programFolder[strlen(programFolder)], "\\VTrace %d.%d",
            vtraceVersion >> 16, (vtraceVersion >> 8) % 256);
  else
    sprintf(&programFolder[strlen(programFolder)], "\\VTrace %d.%d.%d",
            vtraceVersion >> 16, (vtraceVersion >> 8) % 256,
            vtraceVersion % 256);
  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
//                            DRIVER COMMUNICATION
////////////////////////////////////////////////////////////////////////////

// StopLogging tells VTrcLog to stop logging.

BOOLEAN StopLogging (void)
{
  HANDLE hDevice;
  DWORD stopReason = BECAUSE_UNINSTALLING;
  DWORD length;

  hDevice = CreateFile("\\\\.\\VTrcLog", GENERIC_READ | GENERIC_WRITE, 0, NULL,
                       OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (hDevice == ((HANDLE) -1)) {
    DisplayError("Could not connect to VTrcLog driver to tell it to stop "
                 "logging");
    return FALSE;
  }

  DeviceIoControl(hDevice, VTRACE_CTRL_ADD_REASONS_TO_STOP, &stopReason, 4,
                  NULL, 0, &length, NULL);
  CloseHandle(hDevice);

  Sleep(PAUSE_FOR_FLUSHING_MS);

  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
//                          FILTER GROUP TAGS
////////////////////////////////////////////////////////////////////////////

// GetTagInFilterGroup determines the tag of a given service in the Filter
// group. If such a tag exists, it puts it in tagPtr and returns TRUE;
// otherwise, it returns FALSE.

BOOLEAN GetTagInFilterGroup (SC_HANDLE schService, DWORD *tagPtr)
{
  QUERY_SERVICE_CONFIG *serviceConfigInfo;
  DWORD sizeNeeded;

  // Read the configuration information for the service. The first guess
  // as to the needed size for this information could be wrong, so try
  // again if the return value indicates a bigger size is needed.

  serviceConfigInfo = (QUERY_SERVICE_CONFIG *)
    malloc(INITIAL_GUESS_SERVICE_CONFIG_SIZE);
  if (serviceConfigInfo == NULL) return FALSE;
  if (!QueryServiceConfig(schService, serviceConfigInfo,
                          INITIAL_GUESS_SERVICE_CONFIG_SIZE, &sizeNeeded)) {
    if (GetLastError() != ERROR_INSUFFICIENT_BUFFER) {
      free(serviceConfigInfo);
      return FALSE;
    }
    else {
      free(serviceConfigInfo);
      serviceConfigInfo = (QUERY_SERVICE_CONFIG *) malloc(sizeNeeded);
      if (serviceConfigInfo == NULL) return FALSE;
      if (!QueryServiceConfig(schService, serviceConfigInfo, sizeNeeded,
                              &sizeNeeded)) {
        free(serviceConfigInfo);
        return FALSE;
      }
    }
  }

  // Extract the tag information if the group it belongs to is "Filter".

  if (serviceConfigInfo->lpLoadOrderGroup &&
      !strcmp(serviceConfigInfo->lpLoadOrderGroup, "Filter")) {
    *tagPtr = serviceConfigInfo->dwTagId;
    free(serviceConfigInfo);
    return TRUE;
  }
  else {
    free(serviceConfigInfo);
    return FALSE;
  }
}

// RemoveTagFromFilterGroup removes the given tag number from the Filter
// group. Actually, it removes the tag from the GroupOrderList registry
// entry that indicates the order in which Filter group services should
// load.

BOOLEAN RemoveTagFromFilterGroup (DWORD tag)
{
  DWORD err, type, length, lengthNeeded, i, j, *groupOrderList;
  HKEY hKey;

  // Open the group order list registry key.

  err = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                     "System\\CurrentControlSet\\Control\\GroupOrderList",
                     NULL, KEY_READ | KEY_WRITE, &hKey);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not open registry key for service group order list");
    return FALSE;
  }

  // Read the value named after the group to get the actual group order
  // list. First we read the value with no buffer to put it into, to find
  // out how long a buffer we need, if any. If there is no such key, we're
  // done. Otherwise, we read the value from the registry. Note that the
  // structure for the value is a sequence of DWORD's, the first of which
  // is the number of tags n, and the next n of which are those tags in the
  // order in which they should be started.

  lengthNeeded = 0;
  err = RegQueryValueEx(hKey, "Filter", NULL, &type, NULL, &lengthNeeded);
  if (err == ERROR_FILE_NOT_FOUND) {
    RegCloseKey(hKey);
    return TRUE;
  }
  else if (err == ERROR_SUCCESS) {
    groupOrderList = (DWORD *) malloc(lengthNeeded);
    length = lengthNeeded;
    err = RegQueryValueEx(hKey, "Filter", NULL, &type,
                          (LPBYTE) groupOrderList, &length);
    if (err != ERROR_SUCCESS) {
      SetLastError(err);
      DisplayError("Could not read group order list value");
      RegCloseKey(hKey);
      return FALSE;
    }
    if (type != REG_BINARY || lengthNeeded != length ||
        (groupOrderList[0]+1) * sizeof(DWORD) != length) {
      MessageBox(NULL, "Read invalid group order list value", "Error",
                 MB_OK | MB_ICONEXCLAMATION);
      RegCloseKey(hKey);
      return FALSE;
    }
  }
  else {
    SetLastError(err);
    DisplayError("Could not get size of group order list value");
    RegCloseKey(hKey);
    return FALSE;
  }

  // Search for the tag in the group order list. If it's there, remove it.

  for (i = 1; i <= groupOrderList[0]; i++) {
    if (groupOrderList[i] == tag) {
      for (j = i; j < groupOrderList[0]; j++)
        groupOrderList[j] = groupOrderList[j+1];
      groupOrderList[0]--;
      break;
    }
  }

  // Set the new group order list in the registry.

  err = RegSetValueEx(hKey, "Filter", 0, REG_BINARY, (LPBYTE) groupOrderList,
                      (groupOrderList[0] + 1) * sizeof(DWORD));
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not set the Filter group order list in the registry");
    RegCloseKey(hKey);
    return FALSE;
  }

  // Release resources used and return a status value.

  free(groupOrderList);
  RegCloseKey(hKey);

  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
//                       UNINSTALLATION SUBROUTINES
////////////////////////////////////////////////////////////////////////////

// DeleteDLL deletes the VTrace DLL VTRACE.DLL.  For this, it needs to know
// the name of the system root directory (given in systemRoot).  It returns
// a boolean indicating whether it succeeded.

BOOLEAN DeleteDLL (char *systemRoot)
{
  char fileName[MAX_PATH], appInitDLLs[MAX_PATH];
  DWORD status, length, type;
  HKEY hKey;

  // Remove VTRACE.DLL from the list of DLL's automatically loaded into
  // each application.  We do this by updating the registry, specifically by
  // removing VTRACE.DLL from the list of DLL's in the key AppInit_DLLs in
  // the Windows registry key.

  status = RegOpenKeyEx
             (HKEY_LOCAL_MACHINE,
              "Software\\Microsoft\\Windows NT\\CurrentVersion\\Windows", 0,
              KEY_QUERY_VALUE | KEY_SET_VALUE, &hKey);
  if (status != ERROR_SUCCESS) {
    SetLastError(status);
    DisplayError("Could not open Windows key");
    return FALSE;
  }

  length = sizeof(appInitDLLs);
  status = RegQueryValueEx(hKey, "AppInit_DLLs", 0, &type,
                           (BYTE *) appInitDLLs, &length);
  if (status != ERROR_SUCCESS) {
    RegCloseKey(hKey);
    SetLastError(status);
    DisplayError("Could not get AppInit_DLLs registry value");
    return FALSE;
  }

  if (type != REG_SZ) {
    RegCloseKey(hKey);
    MessageBox(NULL, "AppInit_DLLs registry value is not a string", "Error",
               MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }

  if (RemoveVTraceDLL(appInitDLLs)) {
    status = RegSetValueEx(hKey, "AppInit_DLLs", 0, REG_SZ,
                           (BYTE *) appInitDLLs, strlen(appInitDLLs) + 1);
    if (status != ERROR_SUCCESS) {
      RegCloseKey(hKey);
      SetLastError(status);
      DisplayError("Could not set AppInit_DLLs registry value");
      return FALSE;
    }
  }

  RegCloseKey(hKey);

  // Delete VTRACE.DLL from the SYSTEM32 subdirectory of the system root
  // directory at reboot (not now, because it's in use).

  sprintf(fileName, "%s\\SYSTEM32\\VTRACE.DLL", systemRoot);
  if (!MoveFileEx(fileName, NULL, MOVEFILE_DELAY_UNTIL_REBOOT) &&
      GetLastError() != ERROR_FILE_NOT_FOUND) {
    DisplayError("Could not delete VTRACE.DLL from system directory");
    return FALSE;
  }

  return TRUE;
}

// DeleteDriver deletes the driver of the given name (driverName).  It must be
// passed an open handle to the service control manager (schSCManager) and a
// string containing the system root directory name (systemRoot).  It returns
// a boolean indicating whether it succeeded.

BOOLEAN DeleteDriver (char *driverName, SC_HANDLE schSCManager,
                      char *systemRoot)
{
  char fileName[MAX_PATH], otherString[MAX_PATH];
  SC_HANDLE schService;
  SERVICE_STATUS serviceStatus;
  BOOLEAN hasTagInFilterGroup;
  DWORD tagInFilterGroup;

  // Open the service.  If successful, stop the service then delete it.
  // If the service belongs to the Filter group, determine its tag value.

  schService = OpenService(schSCManager, driverName, SERVICE_ALL_ACCESS);
  if (schService == NULL) {
    if (GetLastError() != ERROR_SERVICE_DOES_NOT_EXIST) {
      sprintf(otherString, "Could not open the %s service to delete it",
              driverName);
      DisplayError(otherString);
      return FALSE;
    }
  }
  else {
    hasTagInFilterGroup = GetTagInFilterGroup(schService, &tagInFilterGroup);
    ControlService(schService, SERVICE_CONTROL_STOP, &serviceStatus);
    DeleteService(schService);
    CloseServiceHandle(schService);
  }

  // Delete the no-longer-needed SYS file from the drivers directory

  sprintf(fileName, "%s\\SYSTEM32\\DRIVERS\\%s.SYS", systemRoot, driverName);
  if (!DeleteFile(fileName) && GetLastError() != ERROR_FILE_NOT_FOUND) {
    sprintf(otherString, "Could not delete %s.sys from drivers directory",
            driverName);
    DisplayError(otherString);
    return FALSE;
  }

  // Remove the no-longer-needed tag from the Filter value in the
  // GroupOrderList.

  if (hasTagInFilterGroup)
    RemoveTagFromFilterGroup(tagInFilterGroup);

  return TRUE;
}

// DeleteVTrcSrvc removes the VTrcSrvc service.  It must be passed an open
// handle to the service control manager (schSCManager), and a string
// containing the name of the system root directory (systemRoot).  It returns
// a boolean indicating whether it was successful.

BOOLEAN DeleteVTrcSrvc (SC_HANDLE schSCManager, char *systemRoot)
{
  char fileName[MAX_PATH];
  SC_HANDLE schService;
  SERVICE_STATUS serviceStatus;

  // Open the service.  If successful, stop it and then delete it.

  schService = OpenService(schSCManager, "VTrcSrvc", SERVICE_ALL_ACCESS);
  if (schService == NULL) {
    if (GetLastError() != ERROR_SERVICE_DOES_NOT_EXIST) {
      DisplayError("Could not open the VTrcSrvc service to delete it");
      return FALSE;
    }
  }
  else {
    ControlService(schService, SERVICE_CONTROL_STOP, &serviceStatus);
    DeleteService(schService);
    CloseServiceHandle(schService);
  }

  // Remove the now-unneeded VTrcSrvc.exe file from the SYSTEM32 directory.

  sprintf(fileName, "%s\\SYSTEM32\\VTrcSrvc.exe", systemRoot);
  if (!MoveFileEx(fileName, NULL, MOVEFILE_DELAY_UNTIL_REBOOT) &&
      GetLastError() != ERROR_FILE_NOT_FOUND) {
    DisplayError("Could not delete VTrcSrvc.exe from System32 directory");
    return FALSE;
  }

  return TRUE;
}

// DeleteMainDirectory removes the main VTrace directory and all its contents
// (including the Traces subdirectory and its contents).  It returns a boolean
// indicating whether it succeeded.  It takes as parameter the name of the
// main directory.

BOOLEAN DeleteMainDirectory (char *mainDirectory)
{
  char traceDirectory[MAX_PATH];

  // Derive the name of the traces subdirectory from the main directory

  sprintf(traceDirectory, "%s\\Traces", mainDirectory);

  // Remove the contents of the trace directory, then the trace directory
  // itself. We can't remove the trace directory itself immediately since
  // VTrcLog has an open handle to it (and even if we called it now to close
  // that handle, we would be calling it in a process context in which that
  // handle was invalid and therefore impossible and dangerous to close).
  // So we delay the removal of the trace directory until reboot.

  if (!RemoveContentsOfDirectory(traceDirectory))
    return FALSE;

  if (!MoveFileEx(traceDirectory, NULL, MOVEFILE_DELAY_UNTIL_REBOOT) &&
      GetLastError() != ERROR_FILE_NOT_FOUND &&
      GetLastError() != ERROR_PATH_NOT_FOUND) {
    DisplayError("Could not remove trace subdirectory");
    return FALSE;
  }

  // Arrange for the contents of the main VTrace directory to be deleted at
  // reboot (not now, since the uninstall utility is there and obviously
  // running), as well as for the main VTrace directory itself to be deleted
  // at reboot.

  if (!RemoveContentsOfDirectoryAtReboot(mainDirectory))
    return FALSE;

  if (!MoveFileEx(mainDirectory, NULL, MOVEFILE_DELAY_UNTIL_REBOOT) &&
      GetLastError() != ERROR_FILE_NOT_FOUND &&
      GetLastError() != ERROR_PATH_NOT_FOUND) {
    DisplayError("Could not remove main VTrace directory");
    return FALSE;
  }

  return TRUE;
}

// DeleteStartMenuShortcuts removes all the VTrace utility shortcuts from the
// Start menu, as well as the VTrace subdirectory that they were in.

BOOLEAN DeleteStartMenuShortcuts (void)
{
  char programFolder[MAX_PATH];

  if (!GetVTraceProgramFolder(programFolder)) return FALSE;

  // Remove the contents of the VTrace program folder, and finally that folder
  // itself

  if (!RemoveContentsOfDirectory(programFolder))
    return FALSE;
  if (!RemoveDirectory(programFolder) &&
      GetLastError() != ERROR_FILE_NOT_FOUND &&
      GetLastError() != ERROR_PATH_NOT_FOUND) {
    DisplayError("Could not remove VTrace program folder");
    return FALSE;
  }

  return TRUE;
}

// DeleteRegistryInfo deletes all the information about VTrace in the registry
// that is not important to keep from one installation to another.

BOOLEAN DeleteRegistryInfo (void)
{
  DWORD status;
  HKEY hKey;

  // Open the VTrace registry key

  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                        "Software\\Jay Lorch\\VTrace\\CurrentVersion",
                        0, KEY_QUERY_VALUE | KEY_SET_VALUE, &hKey);
  if (status != ERROR_SUCCESS) {
    SetLastError(status);
    DisplayError("Cannot open VTrace registry key to delete registry "
                 "information");
    return FALSE;
  }

  // We don't check the status on any of the following since success is not
  // that important (and failure is pretty unlikely anyway)

  RegDeleteValue(hKey, "AlwaysConnected");
  RegDeleteValue(hKey, "ProxyServer");
  RegDeleteValue(hKey, "ProxyPort");
  RegDeleteValue(hKey, "MaxTransferKB");
  RegDeleteValue(hKey, "UploadUrlList");
  RegDeleteValue(hKey, "FreeMBNeeded");
  RegDeleteValue(hKey, "LogBufferKB");
  RegDeleteValue(hKey, "PersonalName");
  RegDeleteValue(hKey, "MainDirectory");
  RegDeleteValue(hKey, "MaxTracePieceMB");

  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
//                 ENSURING TRACE FILES GET UPLOADED FIRST
////////////////////////////////////////////////////////////////////////////

// TraceMBRemaining returns the number of MB of trace files remaining in the
// traces subdirectory of the given main directory (mainDirectory).

DWORD TraceMBRemaining (char *mainDirectory)
{
  char searchPattern[MAX_PATH];
  WIN32_FIND_DATA findData;
  HANDLE findHandle;
  ULARGE_INTEGER totalFileSize, currentFileSize;

  // Derive the search pattern for files in the traces subdirectory from
  // the main directory name

  sprintf(searchPattern, "%s\\Traces\\*.*", mainDirectory);

  // Loop through all the files matching the pattern and sum up their sizes
  // in totalFileSize

  totalFileSize.QuadPart = 0;
  findHandle = FindFirstFile(searchPattern, &findData);
  while (findHandle != INVALID_HANDLE_VALUE) {
    if (!(findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
      currentFileSize.HighPart = findData.nFileSizeHigh;
      currentFileSize.LowPart = findData.nFileSizeLow;
      totalFileSize.QuadPart += currentFileSize.QuadPart;
    }
    if (!FindNextFile(findHandle, &findData)) {
      FindClose(findHandle);
      findHandle = INVALID_HANDLE_VALUE;
    }
  }

  // Convert bytes to MB and return the value

  return (UINT) (totalFileSize.QuadPart / 1024 / 1024);
}

// LaunchUploadUtility launches the Upload.exe program in the main directory
// (mainDirectory) to upload trace files.

void LaunchUploadUtility (char *mainDirectory)
{
  STARTUPINFO startupInfo;
  PROCESS_INFORMATION processInfo;
  char executableName[MAX_PATH];

  sprintf(executableName, "%s\\%s", mainDirectory, "Upload.exe");
  GetStartupInfo(&startupInfo);
  CreateProcess(executableName, NULL, NULL, NULL, FALSE, DETACHED_PROCESS,
                NULL, NULL, &startupInfo, &processInfo);
}
