//
// metadata.c
//
// Recording metadata.  Most code was written by Jacob Lorch, but some is
// borrowed from the Linux NTFS code which is....
//
//  Copyright (C) 1995-1997 Martin von Lwis
//  Copyright (C) 1997 Rgis Duchesne
//

#include "ntfstypes.h"
#include "struct.h"
#include "util.h"
#include "super.h"
#include "nttools.h"
#include "inode.h"
#include "macros.h"
#include "dir.h"
#include "support.h"
#include "nonntfs.h"
#include <stdio.h>
#include <stdlib.h>
#include <winioctl.h>
#include "../Common/RegConstraints.h"
#include "../Common/Version.h"

////////////////////////////////////////////////////////////////////////////
//                              STRUCTURES
////////////////////////////////////////////////////////////////////////////

typedef struct _attribute_info {
  int size;
  int resident;
  int found;
  union {
    char *data;
    struct {
      ntfs_runlist *runlist;
      int len;
    } r;
  } d;
} attribute_info;

typedef struct _inode_info {
  int inodeNumber;
  ntfs_volume *vol;
  int isDirectory;
  attribute_info standardInfo;
  attribute_info fileName;
  attribute_info security;
  attribute_info data;
  attribute_info indexRoot;
  attribute_info indexAllocation;
  attribute_info bitmap;
  int dosNameLength;
  char *dosName;
} inode_info;

////////////////////////////////////////////////////////////////////////////
//                         FORWARD DEFINITIONS
////////////////////////////////////////////////////////////////////////////

void DoAllFixedDiskPartitions (char *mainDirectory, int outputFileNumber,
                               DWORD vtraceVersion);

BOOL AccessRegistry (char *mainDirectory, DWORD *freeMBNeededPtr,
                     DWORD *vtraceVersionPtr);
int GetMetadataFileCount (void);

BOOL GetDiskAndPartitionNumbers (char driveLetter, DWORD *diskNumberPtr,
                                 DWORD *partitionNumberPtr);

BOOL IsSpaceAvailable (char *mainDirectory, DWORD extraBytesNeeded);

void InitInodeInfo (inode_info *info, ntfs_volume *vol, int inodeNumber);
void GetAttributeInfo (ntfs_attribute *attr, attribute_info *info);
int GetInodeInfo (ntfs_volume *vol, int inodeNumber, ntfs_inode *inode,
                  inode_info *info);

void OutputPartitionDescription (int diskNumber, int partitionNumber,
                                 char driveLetter, char *mainDirectory,
                                 int outputFileNumber, DWORD vtraceVersion);
void OutputInodeDescription (ntfs_volume *vol, int inodeNumber, FILE *fp,
                             int numInodes, char *whichInodesProcessed);
void OutputTimeInfo (FILE *fp);
void OutputVolumeInfo (ntfs_volume *vol, char driveLetter, FILE *fp);
void OutputAttributeInfo (attribute_info *info, FILE *fp, ntfs_volume *vol);
void OutputInodeInfo (inode_info *info, FILE *fp);
void OutputRunlist (ntfs_runlist *runlist, int len, FILE *fp);
void OutputIndexAllocationBlocks (attribute_info *info, ntfs_volume *vol,
                                  FILE *fp);

void TraverseDirectory (ntfs_inode *inode, inode_info *info, FILE *fp,
                        int numInodes, char *whichInodesProcessed);
void ProcessDirectoryEntry (ntfs_inode *inode, char *entry, FILE *fp,
                            int numInodes, char *whichInodesProcessed);
void ProcessDirectoryRecord (ntfs_inode *inode, int nextblock, FILE *fp,
                             int numInodes, char *whichInodesProcessed);

////////////////////////////////////////////////////////////////////////////
//                                MAIN
////////////////////////////////////////////////////////////////////////////

// The main routine outputs all the metadata on all partitions on local
// fixed disks.

int main (int argc,char *argv[])
{
  char mainDirectory[MAX_PATH];
  DWORD freeMBNeeded, vtraceVersion;

  if (!AccessRegistry(mainDirectory, &freeMBNeeded, &vtraceVersion)) return -1;

  if (IsSpaceAvailable(mainDirectory, (freeMBNeeded + 10) * 1024 * 1024))
    DoAllFixedDiskPartitions(mainDirectory, GetMetadataFileCount(),
                             vtraceVersion);

  return 0;
}

// DoAllFixedDiskPartitions takes an input the name of the VTrace main
// directory.  It outputs to appropriate files in that directory a
// description of the metadata of each partition on local fixed disks.

void DoAllFixedDiskPartitions (char *mainDirectory, int outputFileNumber,
                               DWORD vtraceVersion)
{
  char diskName[3] = "A:";
  DWORD diskNumber, partitionNumber;

  // For each disk, see if it is a fixed disk partition and if so do it.

  for (diskName[0] = 'A'; diskName[0] <= 'Z'; diskName[0]++) {
    if (GetDriveType(diskName) == DRIVE_FIXED &&
        GetDiskAndPartitionNumbers(diskName[0], &diskNumber,
                                   &partitionNumber)) {
      OutputPartitionDescription(diskNumber, partitionNumber, diskName[0],
                                 mainDirectory, outputFileNumber,
                                vtraceVersion);
    }
  }
}

////////////////////////////////////////////////////////////////////////////
//                      DISK AND PARTITION NUMBERS
////////////////////////////////////////////////////////////////////////////

// GetDiskAndPartitionNumbers finds the disk and partition number for a
// certain drive, such as C:, given the drive letter (e.g, 'C'). It
// returns the disk number in diskNumberPtr and partitionNumberPtr. It
// returns a boolean indicating if it was successful. There is a different
// version of this function for NT 4.0 and Windows 2000.

#if TARGET_NT_VERSION == 4

// The NT 4.0 version of GetDiskAndPartitionNumbers uses QueryDosDevice

BOOL GetDiskAndPartitionNumbers (char driveLetter, DWORD *diskNumberPtr,
                                 DWORD *partitionNumberPtr)
{
  char driveName[3] = "A:";
  char diskInfo[64], sample[64], *p;

  // Use QueryDosDevice to get the raw device name.

  driveName[0] = driveLetter;
  if (!QueryDosDevice(driveName, diskInfo, 64))
    return FALSE;

  // If it doesn't begin with "\Device\Harddisk", fail.

  if (strncmp(diskInfo, "\\Device\\Harddisk", 16))
    return FALSE;

  // Extract the hard disk number and partition number.

  diskInfo[63] = '\0';
  *diskNumberPtr = atoi(&diskInfo[16]);
  p = strchr(&diskInfo[17], 'n');
  if (p == NULL)
    return FALSE;
  *partitionNumberPtr = atoi(p+1);

  // Make sure the disk number and partition number extracted really
  // match, and if so, return TRUE.

  sprintf(sample, "\\Device\\Harddisk%ld\\Partition%ld", *diskNumberPtr,
          *partitionNumberPtr);
  return !strcmp(diskInfo, sample);
}

#else

// The Windows 2000 version of GetDiskAndPartitionNumbers uses the superior,
// but unavailable on NT 4.0, IOCTL_STORAGE_GET_DEVICE_NUMBER method.

BOOL GetDiskAndPartitionNumbers (char driveLetter, DWORD *diskNumberPtr,
                                 DWORD *partitionNumberPtr)
{
  HANDLE fileHandle;
  STORAGE_DEVICE_NUMBER deviceInfo;
  DWORD bytesReturned;
  char rawDiskName[] = "\\\\.\\A:";

  // Open the raw disk

  rawDiskName[4] = driveLetter;
  fileHandle = CreateFile(rawDiskName, GENERIC_READ,
                          FILE_SHARE_READ | FILE_SHARE_WRITE,
                          NULL, OPEN_EXISTING,
                          FILE_ATTRIBUTE_NORMAL, NULL);
  if (fileHandle == INVALID_HANDLE_VALUE) return FALSE;

  // Get disk and partition number information using a device control
  // request

  if (!DeviceIoControl(fileHandle, IOCTL_STORAGE_GET_DEVICE_NUMBER, NULL,
                       0, &deviceInfo, sizeof(deviceInfo), &bytesReturned,
                       NULL))
    return FALSE;
  if (bytesReturned != sizeof(deviceInfo))
    return FALSE;

  // Put values into the passed pointers

  *diskNumberPtr = deviceInfo.DeviceNumber;
  *partitionNumberPtr = deviceInfo.PartitionNumber;
  if (*partitionNumberPtr < 0 || *partitionNumberPtr > 255)
    *partitionNumberPtr = 0;

  // Close the handle to the raw disk and return success

  CloseHandle(fileHandle);

  return TRUE;
}

#endif

////////////////////////////////////////////////////////////////////////////
//                          REGISTRY ACCESS
////////////////////////////////////////////////////////////////////////////

// AccessRegistry accesses the VTrace key of the registry to set the time of
// the most recent metadata recording to the current time, as well as to get
// various registry parameters: the VTrace main directory and the number of
// free MB needed.  If it succeeds, it puts the main directory in the given
// buffer (mainDirectory), the number of free MB needed into the given
// pointer location (freeMBNeededPtr), the version of VTrace into the given
// pointer location (vtraceVersionPtr), and returns TRUE.

BOOL AccessRegistry (char *mainDirectory, DWORD *freeMBNeededPtr,
                     DWORD *vtraceVersionPtr)
{
  LONG status;
  HKEY hKey;
  DWORD type, length;
  SYSTEMTIME currentTimeST;
  FILETIME currentTimeFT;

  // Open the registry key for VTrace

  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                        "Software\\Jay Lorch\\VTrace\\CurrentVersion", 0,
                        KEY_QUERY_VALUE | KEY_SET_VALUE, &hKey);
  if (status != ERROR_SUCCESS) return FALSE;

  // Get the current time as a FILETIME structure and record that as the last
  // time the metadata was recorded

  GetLocalTime(&currentTimeST);
  if (!SystemTimeToFileTime(&currentTimeST, &currentTimeFT)) {
    RegCloseKey(hKey);
    return FALSE;
  }
  status = RegSetValueEx(hKey, "TimeMetadataRecorded", 0, REG_BINARY,
                         (BYTE *) &currentTimeFT, 8);
  if (status != ERROR_SUCCESS) {
    RegCloseKey(hKey);
    return FALSE;
  }

  // Obtain the version of VTrace this is

  length = 4;
  status = RegQueryValueEx(hKey, "Version", 0, &type,
                           (BYTE *) vtraceVersionPtr, &length);
  if (status != ERROR_SUCCESS || type != REG_DWORD || length != 4) {
    RegCloseKey(hKey);
    return FALSE;
  }

  // Get the name of the main directory

  length = MAX_PATH;
  status = RegQueryValueEx(hKey, "MainDirectory", 0, &type,
                           (BYTE *) mainDirectory, &length);
  if (status != ERROR_SUCCESS || type != REG_SZ) {
    RegCloseKey(hKey);
    return FALSE;
  }

  // Get the number of free MB required to remain on disk during logging

  length = 4;
  status = RegQueryValueEx(hKey, "FreeMBNeeded", 0, &type,
                           (BYTE *) freeMBNeededPtr, &length);
  if (status != ERROR_SUCCESS || type != REG_DWORD || length != 4)
    *freeMBNeededPtr = DEFAULT_FREE_SPACE;

  // Close the VTrace registry key, then return an indication of success

  RegCloseKey(hKey);
  return TRUE;
}

// GetMetadataFileCount reads the VTrace key to find out how many metadata
// files have already been written.  It increments that number and returns
// the new count.

int GetMetadataFileCount (void)
{
  LONG status;
  HKEY hKey;
  DWORD type, length, count;

  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                        "Software\\Jay Lorch\\VTrace\\CurrentVersion", 0,
                        KEY_QUERY_VALUE | KEY_SET_VALUE, &hKey);
  if (status != ERROR_SUCCESS) return 0;

  length = 4;
  status = RegQueryValueEx(hKey, "MetadataFileCount", 0, &type,
                           (BYTE *) &count, &length);
  if (status != ERROR_SUCCESS || type != REG_DWORD || length != 4)
    count = 1;
  else
    count++;

  status = RegSetValueEx(hKey, "MetadataFileCount", 0, REG_DWORD,
                         (BYTE *) &count, 4);
  RegCloseKey(hKey);

  if (status == ERROR_SUCCESS)
    return count;
  else
    return 0;
}

////////////////////////////////////////////////////////////////////////////
//                          DISK SPACE ROUTINES
////////////////////////////////////////////////////////////////////////////

BOOL IsSpaceAvailable (char *mainDirectory, DWORD extraBytesNeeded)
{
  ULARGE_INTEGER bytesAvail, dummy1, dummy2;

  if (!GetDiskFreeSpaceEx(mainDirectory, &bytesAvail, &dummy1, &dummy2))
    return FALSE;
  return (bytesAvail.HighPart > 0) || (bytesAvail.LowPart >= extraBytesNeeded);
}

////////////////////////////////////////////////////////////////////////////
//                            INODE INFO ROUTINES
////////////////////////////////////////////////////////////////////////////

// InitInodeInfo initializes an inode_info structure.

void InitInodeInfo (inode_info *info, ntfs_volume *vol, int inodeNumber)
{
  info->inodeNumber = inodeNumber;
  info->vol = vol;
  info->standardInfo.found = 0;
  info->fileName.found = 0;
  info->security.found = 0;
  info->data.found = 0;
  info->indexRoot.found = 0;
  info->indexAllocation.found = 0;
  info->bitmap.found = 0;
  info->dosNameLength = 0;
}

// GetAttributeInfo fills in an attribute_info structure pointed to by info
// with information gleaned from a passed ntfs_attribute structure.

void GetAttributeInfo (ntfs_attribute *attr, attribute_info *info)
{
  if (info->found)
    return;
  info->found = 1;
  info->size = attr->size;
  info->resident = attr->resident;
  if (info->resident)
    info->d.data = attr->d.data;
  else {
    info->d.r.runlist = attr->d.r.runlist;
    info->d.r.len = attr->d.r.len;
  }
}

// GetFileAttributeInfo fills in an inode_info structure pointed to be
// info with file information gleaned from a passed ntfs_attribute structure.

void GetFileAttributeInfo (ntfs_attribute *attr, inode_info *info)
{
  int newNameType, oldNameType;

  if (!attr->resident || attr->size < 0x42)
    return;

  if (info->fileName.found) {
    newNameType = attr->d.data[0x41];
    oldNameType = info->fileName.d.data[0x41];
    if (oldNameType == 3 /* Unicode/DOS */) return;
    if (oldNameType == 1 /* Unicode */ && newNameType != 3 /* Unicode/DOS */)
      return;
    if (oldNameType == 2 /* DOS */) {
      if (newNameType == 0 /* POSIX */)
        return;
      else {
        info->dosNameLength =
          (int) (unsigned char) info->fileName.d.data[0x40];
        info->dosName = &info->fileName.d.data[0x42];
      }
    }
  }

  info->fileName.found = 1;
  info->fileName.size = attr->size;
  info->fileName.resident = attr->resident;
  info->fileName.d.data = attr->d.data;
}

// GetInodeInfo reads enough information about an inode from a volume to fill
// in an ntfs_inode structure (pointed to by inode) and an inode_info structure
// (pointed to by info).  The inode is specified as an inode number
// (inodeNumber) within a given volume (vol).  This function returns nonzero
// on success.

int GetInodeInfo (ntfs_volume *vol, int inodeNumber, ntfs_inode *inode,
                  inode_info *info)
{
  ntfs_attribute *attr;
  int i;

  InitInodeInfo(info, vol, inodeNumber);

  if (ntfs_init_inode(inode, vol, inodeNumber) != 0) return 0;

  for (i = 0; i < inode->attr_count; i++) {
    attr = inode->attrs + i;
    switch (attr->type) {
    case 0x10 : GetAttributeInfo(attr, &info->standardInfo); break;
    case 0x30 : GetFileAttributeInfo(attr, info); break;
    case 0x50 : GetAttributeInfo(attr, &info->security); break;
    case 0x80 : GetAttributeInfo(attr, &info->data); break;
    case 0x90 : GetAttributeInfo(attr, &info->indexRoot); break;
    case 0xA0 : GetAttributeInfo(attr, &info->indexAllocation); break;
    case 0xB0 : GetAttributeInfo(attr, &info->bitmap); break;
    }
  }

  return 1;
}

////////////////////////////////////////////////////////////////////////////
//                          OUTPUT ROUTINES
////////////////////////////////////////////////////////////////////////////

// OutputPartitionDescription outputs to an appropriately named file in
// the directory mainDirectory a description of the metadata for disk number
// diskNumber, partition number partitionNumber.

void OutputPartitionDescription (int diskNumber, int partitionNumber,
                                 char driveLetter, char *mainDirectory,
                                 int outputFileNumber, DWORD vtraceVersion)
{
  ntfs_volume *vol;
  char fileName[MAX_PATH];
  FILE *fp;
  ntfs_inode inode;
  inode_info info;
  int numInodes;
  char *whichInodesProcessed;

  // First, open the volume.  If the volume is invalid (e.g. if it is not
  // an NTFS partition), try reading it as a non-NTFS partition.

  vol = ntfs_open_volume(driveLetter, 0, 1, 0);
  if (vol == NULL) {
    OutputNonNTFSPartitionDescription(diskNumber, partitionNumber, driveLetter,
                                      mainDirectory, outputFileNumber,
                                      vtraceVersion);
    return;
  }

  // Get the information for inode 0, the MFT record.  From this information,
  // we deduce the size of the MFT, from which we derive the number of inode
  // records in the MFT.  We use this to decide the size of the array
  // whichInodesProcessed, each element of which is used to record whether
  // the associated inode number has been processed yet.

  if (GetInodeInfo(vol, 0, &inode, &info) == 0) {
    ntfs_release_volume(vol);
    return;
  }
  numInodes = info.data.size / vol->mft_recordsize;
  whichInodesProcessed = ntfs_malloc(numInodes);
  ntfs_bzero(whichInodesProcessed, numInodes);

  // Open an output file in the Traces subdirectory of the VTrace main
  // directory into which we will output all the descriptions.

  sprintf(fileName, "%s\\Traces\\M%dD%dP%d.met", mainDirectory,
          outputFileNumber, diskNumber, partitionNumber);
  fp = fopen(fileName, "wb");

  // If the open succeeds, first output the description of inode 0 (the MFT).
  // Then, output the description of inode 5, the root directory.  Since the
  // routine for outputting descriptions recursively descends directories,
  // this will wind up outputting the descriptions of every file in the
  // partition.
  
  if (fp != NULL) {
    fwrite(&vtraceVersion, 1, 4, fp);
    OutputTimeInfo(fp);
    OutputVolumeInfo(vol, driveLetter, fp);
    OutputInodeInfo(&info, fp);
    ntfs_clear_inode(&inode);
    whichInodesProcessed[0] = 1;
    OutputInodeDescription(vol, 5, fp, numInodes, whichInodesProcessed);
    fclose(fp);
  }
  else
    ntfs_clear_inode(&inode);

  // Release the memory associated with this volume and the array we
  // allocated for determining which inodes have been processed.

  ntfs_release_volume(vol);
  ntfs_free(whichInodesProcessed);
}

// OutputInodeDescriptions obtains information about a specified inode, and
// outputs that information to the file pointed to by fp.  The inode is
// specified by volume (vol) and inode number (inodeNumber).  The array
// whichInodesProcessed, of size numInodes, indicates which inodes have
// already been processed.  We use this in two ways: (1) if this inode
// has already been processed, we skip it, and (2) if we do process this
// inode, we set its element of the array to make sure it isn't processed
// again.

// This routine is recursive, in that if it is called with an inode that
// turns out to correspond to a directory, it calls itself for each item
// in the subdirectory (after outputting the description of the directory
// itself).

void OutputInodeDescription (ntfs_volume *vol, int inodeNumber, FILE *fp,
                             int numInodes, char *whichInodesProcessed)
{
  ntfs_inode inode;
  inode_info info;

  // Make sure this inode hasn't been processed yet

  if (inodeNumber < 0 || inodeNumber >= numInodes) return;
  if (whichInodesProcessed[inodeNumber]) return;
  whichInodesProcessed[inodeNumber] = 1;

  // Process the inode (and, recursively, its descendants)

  if (GetInodeInfo(vol, inodeNumber, &inode, &info)) {
    OutputInodeInfo(&info, fp);
    if (NTFS_GETU32(info.fileName.d.data+0x38) & 0x10000000)
      TraverseDirectory(&inode, &info, fp, numInodes, whichInodesProcessed);
    ntfs_clear_inode(&inode);
  }
}

// OutputTimeInfo outputs information about the current time

void OutputTimeInfo (FILE *fp)
{
  SYSTEMTIME systemTime;

  GetLocalTime(&systemTime);
  fwrite(&systemTime, 1, sizeof(SYSTEMTIME), fp);
}

// OutputVolumeInfo outputs information about a volume as a whole to an open
// file, pointed to by fp.  The volume is passed as vol, and its drive letter,
// i.e. 'D' for D: is passed as driveLetter.

void OutputVolumeInfo (ntfs_volume *vol, char driveLetter, FILE *fp)
{
  char outbuf[23];

  *outbuf = FILESYSTEM_STATISTICS_TYPE_NTFS;
  outbuf[1] = PARTITION_IFS;
  outbuf[2] = driveLetter;
  * (ULONG *) &outbuf[3] = vol->blocksize;
  * (ULONG *) &outbuf[7] = vol->clusterfactor;
  * (ULONG *) &outbuf[11] = vol->mft_clusters_per_record;
  * (ULONG *) &outbuf[15] = vol->index_clusters_per_record;
  * (ULONG *) &outbuf[19] = vol->num_clusters;
  fwrite(outbuf, 1, 23, fp);
}

// OutputAttributeInfo outputs an attribute info structure (pointed to by info)
// to an open file pointer (pointed to by fp).

void OutputAttributeInfo (attribute_info *info, FILE *fp, ntfs_volume *vol)
{
  char *cluster;
  int totalBytesRemaining;
  ntfs_cluster_t i, j, clusterNum;
  ULONG dummy[2] = {0, 0}, bytesToCopy;
  ntfs_io io;

  // If it's not found, output a size of 0 and a runlist length of 0

  if (!info->found) {
    fwrite(dummy, 1, 8, fp);
    return;
  }

  // Output the size

  fwrite(&info->size, 1, 4, fp);

  // If it's resident, output a runlist length of 0 and then copy the data

  if (info->resident) {
    fwrite(dummy, 1, 4, fp);
    fwrite(info->d.data, 1, info->size, fp);
    return;
  }

  // At this point, we know it isn't resident, so we output the runlist length
  // and the runlist

  OutputRunlist(info->d.r.runlist, info->d.r.len, fp);

  // Now we read the data, one cluster at a time (guided by the runlist),
  // and output it

  cluster = ntfs_malloc(vol->clustersize);
  totalBytesRemaining = info->size;
  for (i = 0; i < info->d.r.len; i++) {
    for (j = 0, clusterNum = info->d.r.runlist[i].cluster;
         j < info->d.r.runlist[i].len && totalBytesRemaining > 0;
         j++, clusterNum++) {
      io.do_read=1;
      io.param=cluster;
      io.size=vol->clustersize;
      if (ntfs_getput_clusters(vol,(int)clusterNum,0,&io))
        continue;
      bytesToCopy = min(totalBytesRemaining, vol->clustersize);
      fwrite(cluster, 1, bytesToCopy, fp);
      totalBytesRemaining -= bytesToCopy;
    }
  }

  // If we weren't able to read the whole thing, pad the end with zeros

  if (totalBytesRemaining) {
    ntfs_bzero(cluster, vol->clustersize);
    do {
      bytesToCopy = min(totalBytesRemaining, vol->clustersize);
      fwrite(cluster, 1, bytesToCopy, fp);
      totalBytesRemaining -= bytesToCopy;
    } while (totalBytesRemaining);
  }

  // Free the memory we allocated for this routine to read data

  ntfs_free(cluster);
}

// OutputInodeInfo outputs to an open file pointed to by fp the contents of an
// inode_info structure pointed to by info.

void OutputInodeInfo (inode_info *info, FILE *fp)
{
  if (!info->fileName.found || !info->fileName.resident ||
      info->fileName.size < 0x3C)
    return;

  fwrite("VtSG", 1, 4, fp);
  fwrite(&info->inodeNumber, 1, 4, fp);

  OutputAttributeInfo(&info->standardInfo, fp, info->vol);
  OutputAttributeInfo(&info->security, fp, info->vol);
  fwrite(&info->dosNameLength, 1, 1, fp);
  if (info->dosNameLength)
    fwrite(info->dosName, 1, info->dosNameLength*2, fp);
  OutputAttributeInfo(&info->fileName, fp, info->vol);
  if (NTFS_GETU32(info->fileName.d.data+0x38) & 0x10000000) {
    // it's a directory
    OutputAttributeInfo(&info->indexRoot, fp, info->vol);
    OutputAttributeInfo(&info->bitmap, fp, info->vol);
    OutputIndexAllocationBlocks(&info->indexAllocation, info->vol, fp);
  }
  else {
    // it's a file
    fwrite(&info->data.size, 1, 4, fp);
    if (info->data.resident)
      OutputRunlist(NULL, 0, fp);
    else
      OutputRunlist(info->data.d.r.runlist, info->data.d.r.len, fp);
  }
}

// OutputRunlist outputs to an open file pointed to by fp a run list of length
// len pointed to by runlist.

void OutputRunlist (ntfs_runlist *runlist, int len, FILE *fp)
{
  int i;
  
  fwrite(&len, 1, 4, fp);
  for (i = 0; i < len; i++) {
    fwrite(&runlist[i].cluster, 1, 4, fp);
    fwrite(&runlist[i].len, 1, 4, fp);
  }
}

// OutputIndexAllocationBlocks outputs to an open file pointed to by fp the
// location and contents of index allocation blocks described in the
// attribute_info structure pointed to by info.  It takes a volume vol as
// parameter so that it can search that volume for the blocks.

void OutputIndexAllocationBlocks (attribute_info *info, ntfs_volume *vol,
                                  FILE *fp)
{
  ULONG dummy[2] = {0xFFFFFFFFUL, 0}, bytesToCopy;
  char *cluster;
  ntfs_cluster_t i, j, clusterNum;
  int usedBytesRemainingInCurrentBlock, totalBytesRemainingInCurrentBlock;
  ntfs_io io;

  // If it's not found, output a dummy cluster 0xFFFFFFFF and its length, 0

  if (!info->found) {
    fwrite(dummy, 1, 8, fp);
    return;
  }

  // If it's resident, output a dummy cluster 0xFFFFFFFF and its length, the
  // size of the data, followed by that data

  if (info->resident) {
    fwrite(dummy, 1, 4, fp);
    fwrite(&info->size, 1, 4, fp);
    fwrite(info->d.data, 1, info->size, fp);
  }

  // At this point we know it isn't resident, so we use the runlist to read
  // each cluster and output it.  However, to save space, we parse those
  // clusters to see which bytes are actually in use.  The clusters form
  // blocks, each of which starts with a header than describes the size of
  // the block and how many bytes of the block are actually in use.

  cluster = ntfs_malloc(vol->clustersize);
  totalBytesRemainingInCurrentBlock = 0;
  usedBytesRemainingInCurrentBlock = 0;
  for (i = 0; i < info->d.r.len; i++) {
    for (j = 0, clusterNum = info->d.r.runlist[i].cluster;
         j < (int) info->d.r.runlist[i].len;
         j++, clusterNum++) {
      if (totalBytesRemainingInCurrentBlock <= 0) {
        // If we're done with the previous block, this is the first cluster of
        // the block.  We read it and parse it to find out the length of the
        // block (which we put in totalBytesRemainingInCurrentBlock), and the
        // length of the in-use part of the block (which we put in usedBytes-
        // remainingInCurrentBlock).

        io.do_read=1;
        io.param=cluster;
        io.size=vol->clustersize;
        if (ntfs_getput_clusters(vol,(int)clusterNum,0,&io) ||
            !IS_INDEX_RECORD(cluster)) {
          ntfs_free(cluster);
          fwrite(dummy, 1, 8, fp);
          return;
        }
        usedBytesRemainingInCurrentBlock = NTFS_GETU32(cluster+0x1C) + 0x18;
        totalBytesRemainingInCurrentBlock = NTFS_GETU32(cluster+0x20) + 0x18;
        bytesToCopy = min(usedBytesRemainingInCurrentBlock, vol->clustersize);
        fwrite(&clusterNum, 1, 4, fp);
        fwrite(&bytesToCopy, 1, 4, fp);
        fwrite(cluster, 1, bytesToCopy, fp);
        usedBytesRemainingInCurrentBlock -= bytesToCopy;
        totalBytesRemainingInCurrentBlock -= vol->clustersize;
      }
      else {
        // This is not the first cluster in a block, so just write as much of
        // it as is in use

        if (usedBytesRemainingInCurrentBlock > 0) {
          io.do_read=1;
          io.param=cluster;
          io.size=vol->clustersize;
          if (ntfs_getput_clusters(vol,(int)clusterNum,0,&io)) {
            ntfs_free(cluster);
            fwrite(dummy, 1, 8, fp);
            return;
          }
          bytesToCopy = min(usedBytesRemainingInCurrentBlock, vol->clustersize);
          fwrite(&clusterNum, 1, 4, fp);
          fwrite(&bytesToCopy, 1, 4, fp);
          fwrite(cluster, 1, bytesToCopy, fp);
          usedBytesRemainingInCurrentBlock -= bytesToCopy;
        }
        totalBytesRemainingInCurrentBlock -= vol->clustersize;
      }
    }
  }

  // Free the memory we allocated for this routine to read a cluster

  ntfs_free(cluster);

  // Write a dummy final cluster 0xFFFFFFFF and its length, 0

  fwrite(dummy, 1, 8, fp);
}

//////////////////////////////////////////////////////////////////////////////
//                       DIRECTORY TRAVERSAL ROUTINES
//////////////////////////////////////////////////////////////////////////////

// TraverseDirectory outputs inode descriptions for each file and directory
// below the given inode in the filesystem hierarchy.  The output is done to
// the open file pointed to by fp.  The inode is specified by a pointer to
// an ntfs_inode structure, inode.  Also passed is an array
// whichInodesProcessed used to keep track of which inodes have been already
// processed so they don't get processed again; numInodes is the size of this
// array, and represents one more than the highest possible inode number in
// this partition.  Also passed is an inode_info structure corresponding to
// the inode.

void TraverseDirectory (ntfs_inode *inode, inode_info *info, FILE *fp,
                        int numInodes, char *whichInodesProcessed)
{
  char *buf = info->indexRoot.d.data;
  int length;

  inode->u.index.recordsize=NTFS_GETU32(buf+0x8);
  inode->u.index.clusters_per_record=NTFS_GETU32(buf+0xC);
  buf+=0x20;
  while(1)
  {
    length=NTFS_GETU16(buf+8);
    ProcessDirectoryEntry(inode,buf,fp,numInodes,whichInodesProcessed);
    if(NTFS_GETU8(buf+12)&2)break;
    buf+=length;
    if(!length)  // Error!
      break;
  }
}

// ProcessDirectoryEntry outputs inode descriptions for a given directory
// entry and all entries in its associated subblock.  (Since the routine it
// uses to output inode descriptions recursively descends directories, it also
// outputs descriptions for items below those in the filesystem hierarchy.)
// This routine is passed a parent inode (inode), a pointer to the directory
// entry (entry), an open file pointer to which to send output (fp), and an
// array of which inode numbers have already been processed
// (whichInodesProcessed, and its size, numInodes).

void ProcessDirectoryEntry (ntfs_inode *inode, char *entry, FILE *fp,
                            int numInodes, char *whichInodesProcessed)
{
  int length=NTFS_GETU16(entry+8);
  int used=(NTFS_GETU8(entry+12)&2)==0;
  if(used)
    OutputInodeDescription(inode->vol, NTFS_GETU32(entry), fp,
                           numInodes,whichInodesProcessed);
  if((int)NTFS_GETU8(entry+12)&1){
    int nextblock=(int)NTFS_GETU64(entry+length-8);
    ProcessDirectoryRecord(inode,nextblock,fp,numInodes,whichInodesProcessed);
  }
}

// ProcessDirectoryRecord outputs inode descriptions for all inodes described
// in a given allocation block.  (Since the routine it uses to output inode
// descriptions recursively descends directories, it also outputs descriptions
// for items below those in the filesystem hierarchy.)  This routine is passed
// a parent inode (inode), the allocation block number within that inode, an
// open file pointer to which to send output (fp), and an array of which inode
// numbers have already been processed (whichInodesProcessed, and its size,
// numInodes).

void ProcessDirectoryRecord (ntfs_inode *inode, int nextblock, FILE *fp,
                             int numInodes, char *whichInodesProcessed)
{
  int length,error;
  char record[8192];
  char *offset;
  ntfs_io io;

  io.fn_put=ntfs_put;
  io.fn_get=0;
  io.param=record;
  io.size=inode->u.index.recordsize;
  error=ntfs_read_attr(inode,inode->vol->at_index_allocation,"$I30",
                       nextblock*inode->vol->clustersize,&io);
  if(error || io.size!=inode->u.index.recordsize) // Error!
    return;
  if(!ntfs_check_index_record(inode,record)) // Error!
    return;
  offset=record+NTFS_GETU16(record+0x18)+0x18;
  do{
    ProcessDirectoryEntry(inode,offset,fp,numInodes,whichInodesProcessed);
    if(*(offset+0xC)&2)break;
    length=NTFS_GETU16(offset+8);
    if(!length)break;
    offset+=length;
  }while(1);
}
