//
// nonntfs.c
//
// Recording metadata of non-NTFS partitions.  Written by Jacob Lorch,
//

#include "ntfstypes.h"
#include "struct.h"
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <winbase.h>
#include <winioctl.h>
#include "metadata.h"
#include "../Common/RegConstraints.h"

////////////////////////////////////////////////////////////////////////////
//                        FORWARD DEFINITIONS
////////////////////////////////////////////////////////////////////////////

void OutputNonNTFSVolumeInfo (char driveLetter, FILE *fp);
void OutputContentsOfDirectory (FILE *fp, char *directoryName,
                                DWORD directoryNameLength,
                                DWORD directoryNameBufferSize,
                                SECURITY_INFORMATION *securityBuffer,
                                DWORD securityBufferLength);
void OutputNonNTFSInode (FILE *fp, WIN32_FIND_DATA *findData,
                         char *parentDirectoryName,
                         DWORD parentDirectoryNameLength,
                         DWORD parentDirectoryNameBufferSize,
                         SECURITY_INFORMATION *securityBuffer,
                         DWORD securityBufferLength);

////////////////////////////////////////////////////////////////////////////
//                          OUTPUT ROUTINES
////////////////////////////////////////////////////////////////////////////

// OutputNonNTFSPartitionDescription outputs to an appropriately named file
// in the directory mainDirectory a description of the metadata for disk
// number diskNumber, partition number partitionNumber, which is not NTFS.

void OutputNonNTFSPartitionDescription (int diskNumber, int partitionNumber,
                                        char driveLetter, char *mainDirectory,
                                        int outputFileNumber,
                                        DWORD vtraceVersion)
{
  FILE *fp;
  char fileName[1024];
  char securityInfo[4096];

  // Open an output file in the Traces subdirectory of the VTrace main
  // directory into which we will output all the descriptions.

  sprintf(fileName, "%s\\Traces\\M%dD%dP%d.met", mainDirectory,
          outputFileNumber, diskNumber, partitionNumber);
  fp = fopen(fileName, "wb");
  if (fp == NULL)
    return;

  // If the open succeeds, first output the version of VTrace, then
  // the time this data is being generated, then information about the
  // volume, and finally information about all the inodes. Do the
  // inodes by calling OutputContentsOfDirectory to do all the
  // subdirectories of the root directory.

  fwrite(&vtraceVersion, 1, 4, fp);
  OutputTimeInfo(fp);
  OutputNonNTFSVolumeInfo(driveLetter, fp);
  sprintf(fileName, "%c:\\", driveLetter);
  OutputContentsOfDirectory(fp, fileName, 3, 1024, securityInfo, 4096);

  // Close the output file.

  fclose(fp);
}

// OutputNonNTFSVolumeInfo outputs information about the volume with drive
// letter 'driveLetter' to the open output file pointed to by 'fp'.

void OutputNonNTFSVolumeInfo (char driveLetter, FILE *fp)
{
  char volumeFileName[] = "\\\\.\\A:";
  char rootFileName[] = "A:\\";
  HANDLE fileHandle;
  DWORD bytesReturned, sectorsPerCluster, bytesPerSector, numFreeClusters,
        numTotalClusters;
  PARTITION_INFORMATION partitionInfo;
  char partitionType;

  // Use DeviceIoControl with a device I/O control code of
  // IOCTL_DISK_GET_PARTITION_INFO in order to find out the partition type.

  partitionType = PARTITION_ENTRY_UNUSED;
  volumeFileName[4] = driveLetter;
  fileHandle = CreateFile(volumeFileName, GENERIC_READ,
                          FILE_SHARE_READ | FILE_SHARE_WRITE,
                          NULL, OPEN_EXISTING,
                          FILE_ATTRIBUTE_NORMAL, NULL);
  if (fileHandle != INVALID_HANDLE_VALUE) {
    if (DeviceIoControl(fileHandle, IOCTL_DISK_GET_PARTITION_INFO, NULL, 0,
                        &partitionInfo, sizeof(partitionInfo), &bytesReturned,
                        NULL) &&
        (bytesReturned >= sizeof(partitionInfo)))
      partitionType = partitionInfo.PartitionType;
    CloseHandle(fileHandle);
  }

  // Get information about the disk, including free space, by calling
  // GetDiskFreeSpace.

  rootFileName[0] = driveLetter;
  if (!GetDiskFreeSpace(rootFileName, &sectorsPerCluster, &bytesPerSector,
                        &numFreeClusters, &numTotalClusters)) {
    sectorsPerCluster = 0;
    bytesPerSector = 0;
    numFreeClusters = 0;
    numTotalClusters = 0;
  }

  // Write the information we've collected to the output file.

  fputc(FILESYSTEM_STATISTICS_TYPE_FAT, fp);
  fputc(partitionType, fp);
  fputc(driveLetter, fp);
  fwrite(&sectorsPerCluster, 1, 4, fp);
  fwrite(&bytesPerSector, 1, 4, fp);
  fwrite(&numFreeClusters, 1, 4, fp);
  fwrite(&numTotalClusters, 1, 4, fp);
}

// OutputContentsOfDirectory outputs information about all the subfiles and
// subdirectories of the given directory (named 'directoryName') to the
// given output file (pointed to by 'fp').

void OutputContentsOfDirectory (FILE *fp, char *directoryName,
                                DWORD directoryNameLength,
                                DWORD directoryNameBufferSize,
                                SECURITY_INFORMATION *securityBuffer,
                                DWORD securityBufferLength)
{
  WIN32_FIND_DATA findData;
  HANDLE findHandle;

  // Loop through all the subfiles and subdirectories other than "." and "..".
  // For each one, output an 'I' followed by the inode description (which may
  // include the descriptions of further subfiles and subdirectories). Then,
  // output an 'E' to indicate the end of the list.

  directoryName[directoryNameLength] = '*';
  directoryName[directoryNameLength+1] = '\0';
  findHandle = FindFirstFile(directoryName, &findData);
  if (findHandle != INVALID_HANDLE_VALUE) {
    do {
      if (!(findData.cFileName[0] == '.' &&
            (findData.cFileName[1] == '\0' ||
             (findData.cFileName[1] == '.' &&
              findData.cFileName[2] == '\0')))) {
        fputc('I', fp);
        OutputNonNTFSInode(fp, &findData, directoryName, directoryNameLength,
                           directoryNameBufferSize, securityBuffer,
                           securityBufferLength);
      }
    } while (FindNextFile(findHandle, &findData));
    FindClose(findHandle);
  }

  fputc('E', fp);
}

// OutputNonNTFSInode outputs information about a given inode to a given
// output file (pointed to by 'fp'). The inode is described by its parent's
// name (parentDirectoryName) and by a WIN32_FIND_DATA structure (pointed to
// by findData). If this inode is a directory, we recursively output
// information about all the children of this inode by calling
// OutputContentsOfDirectory.

void OutputNonNTFSInode (FILE *fp, WIN32_FIND_DATA *findData,
                         char *parentDirectoryName,
                         DWORD parentDirectoryNameLength,
                         DWORD parentDirectoryNameBufferSize,
                         SECURITY_INFORMATION *securityBuffer,
                         DWORD securityBufferLength)
{
  int fileNameLen, alternateFileNameLen;
  DWORD securityLen;
  DWORD returnSize;

  // First, write the signature that precedes each inode.

  fwrite("VtSG", 1, 4, fp);

  // Write the first 36 bytes of the WIN32_FIND_DATA structure, which contain
  // information about attributes, change times, file size, etc.

  fwrite(findData, 1, 36, fp);

  // Output the length of the file name as a 2-byte number, followed by the
  // file name.

  fileNameLen = strlen(findData->cFileName);
  fwrite(&fileNameLen, 1, 2, fp);
  fwrite(findData->cFileName, 1, fileNameLen, fp);

  // Output the length of the alternate file name (which can't be more than
  // 14 bytes long) as a 1-byte number, followed by the file name.

  alternateFileNameLen = strlen(findData->cAlternateFileName);
  fwrite(&alternateFileNameLen, 1, 1, fp);
  fwrite(findData->cAlternateFileName, 1, alternateFileNameLen, fp);

  // If there isn't room to put the full path of this file in the buffer,
  // along with a backslash, asterisk, and terminating null character,
  // enter no security information and exit now so that the directory
  // isn't traversed.

  if (parentDirectoryNameLength + fileNameLen + 3 >
      parentDirectoryNameBufferSize) {
    securityLen = 0;
    fwrite(&securityLen, 1, 2, fp);
    if (findData->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
      fputc('E', fp);
    return;
  }
  
  // Get the security information and enter it in the file.

  strcpy(&parentDirectoryName[parentDirectoryNameLength],
         findData->cFileName);
  GetFileSecurity(parentDirectoryName,
                  OWNER_SECURITY_INFORMATION |
                  GROUP_SECURITY_INFORMATION |
                  DACL_SECURITY_INFORMATION,
                  securityBuffer, 0, &securityLen);
  if (securityLen > securityBufferLength ||
      !GetFileSecurity(parentDirectoryName,
                       OWNER_SECURITY_INFORMATION |
                       GROUP_SECURITY_INFORMATION |
                       DACL_SECURITY_INFORMATION,
                       securityBuffer, securityLen, &returnSize))
    securityLen = 0;
  fwrite(&securityLen, 1, 2, fp);
  fwrite(securityBuffer, 1, securityLen, fp);

  // If this inode is a directory, recursively output information about its
  // children by calling OutputContentsOfDirectory with the appropriate
  // parameters.

  if (findData->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
    parentDirectoryName[parentDirectoryNameLength + fileNameLen] = '\\';
    parentDirectoryName[parentDirectoryNameLength + fileNameLen + 1] = '\0';
    OutputContentsOfDirectory(fp, parentDirectoryName,
                              parentDirectoryNameLength + fileNameLen + 1,
                              parentDirectoryNameBufferSize,
                              securityBuffer, securityBufferLength);
  }
}
