/*
 *  nttools.c
 *  Helper functions for the tools
 *
 *  Copyright (C) 1995-1997, 1999 Martin von Lwis
 *  Copyright (C) 1997 Rgis Duchesne
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "ntfstypes.h"
#include "struct.h"
#include "nttools.h"

#include <stdio.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_IO_H
#include <io.h>
#endif
#include <errno.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/stat.h>
#include <sys/types.h>
#include <stdarg.h>
#include <time.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "super.h"
#include "inode.h"
#include "dir.h"
#include "macros.h"
#include "support.h"
#include "util.h"
#include "volsize.h"
#include <winioctl.h>

void ntfs_release_cluster(void *data)
{
	free(data);
}

// OpenPartition opens in read-only mode a file handle for the given drive
// letter.  It sets the fileHandle parameter of the given ntfs_volume structure
// to that file handle.  This function is not part of the original Linux NTFS
// distribution; we added it to use file handles in accessing partitions and
// to access them by drive letter.

int OpenPartition (ntfs_volume *vol, char driveLetter)
{
  char fileName[] = "\\\\.\\A:";

  fileName[4] = driveLetter;
  vol->fileHandle = CreateFile(fileName, GENERIC_READ,
                               FILE_SHARE_READ | FILE_SHARE_WRITE,
                               NULL, OPEN_EXISTING,
                               FILE_ATTRIBUTE_NORMAL, NULL);
  if (vol->fileHandle == INVALID_HANDLE_VALUE) return FALSE;
  vol->num_clusters = GetVolumeSize(vol->fileHandle);
  return TRUE;
}

ntfs_volume *ntfs_open_volume(char driveLetter, int bias, int silent, int no_inodes)
{
	char cluster0[512];
	ntfs_io io;
	ntfs_volume *newvol;

	io.fn_put=0;
	io.fn_get=0; /* we don't need copy functions here */
	newvol=malloc(sizeof(ntfs_volume));
  if (!OpenPartition(newvol, driveLetter)) return NULL;
  newvol->partition_bias=bias;
	/* read the boot sector */
  newvol->clustersize = 512;
  if (!ReadCluster(newvol,0,cluster0) || !IS_NTFS_VOLUME(cluster0)) {
    CloseHandle(newvol->fileHandle);
    ntfs_free(newvol);
    return NULL;
  }
	ntfs_init_volume(newvol,cluster0);
	/* read the first mft record */
	newvol->mft=malloc(newvol->mft_recordsize);
	io.do_read=1;
	io.param=newvol->mft;
	io.size=newvol->mft_recordsize;
	ntfs_getput_clusters(newvol,newvol->mft_cluster,0,&io);
	/* fix it */
	if(!ntfs_check_mft_record(newvol,newvol->mft)){
		ntfs_release_volume(newvol);
		return 0;
	}
	if(!no_inodes)
		ntfs_load_special_files(newvol);
	return newvol;
}

/* print a unicode string */
void uniprint(char *first, int length)
{
	while(length--){
		putchar(*first++);
		/* character above 255 or not a valid unicode string */
		if(*first++){
			printf("!!!!Error printing file name\n");
			return;
		}
	}
}

void uniprint_lower(char *first,int length)
{
	while(length--){
		putchar(tolower(*first));
		first++;
		/* character above 255 or not a valid unicode string */
		if(*first++){
			printf("!!!!Error printing file name\n");
			return;
		}
	}
}

/* unit of 100ns since 1.1.1601 */
void print_time(ntfs_time64_t t)
{
	ntfs_time64_t sec;
	ntfs_time_t unix_utc;
	char *str;
	sec=t/10000000;
	unix_utc=sec-((ntfs_time64_t)369*365+89)*24*3600;
	str=ctime(&unix_utc);
	/* remove \n */
	str[strlen(str)-1]='\0';
	printf("%s",str);
}

/* answer the MFT record number for file name in directory ino */
int ntfs_find_file(ntfs_inode *ino, char *name)
{
	char item[1000];
	ntfs_iterate_s walk;
	int error=ntfs_decodeuni(ino->vol,name,strlen(name),
				 &walk.name,&walk.namelen);
	if(error)
		return -1;
	walk.type=BY_NAME;
	walk.dir=ino;
	walk.result=item;
	if(!ntfs_getdir_byname(&walk))return -1;
	free(walk.name);
	return *(int*)item;
}

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
