/*
 *  support.c
 *  Specific support functions
 *
 *  Copyright (C) 1999 Martin von Lwis
 *  Copyright (C) 1997 Rgis Duchesne
 */

#include "ntfstypes.h"
#include "struct.h"
#include "support.h"
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
/* Solaris defines bzero() here */
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_IO_H
#include <io.h>
#endif
#include <stdarg.h>
#include <stdio.h>
#include <errno.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <time.h>
#include "inode.h"
#include "nttools.h"
#include "macros.h"
#include "util.h"

int ntdebug;

void ntfs_debug(int mask, const char *fmt, ...)
{
	va_list args;

	if (!(mask & ntdebug))
		return;
	va_start( args, fmt );
	vfprintf( stderr, fmt, args );
	va_end (args);
}

void *ntfs_malloc(int size)
{
    return malloc(size);
}

void ntfs_free(void *block)
{
    free(block);
}

void ntfs_bzero(void *s, int n)
{
#ifdef HAVE_BZERO
    bzero(s, n);
#else
// Jay Lorch took the ifdef, etc out below since memset exists but HAVE_MEMSET
// is not set:
//#ifdef HAVE_MEMSET
        memset(s,'\0',n);
//#else
//#error no bzero implementation
//#endif
#endif
}

void ntfs_memcpy(void *dest, const void *src, ntfs_size_t n)
{
	memcpy(dest, src, n);
}

void ntfs_memmove(void *dest, const void *src, ntfs_size_t n)
{
	memmove(dest, src, n);
}

/* Warn that an error occured */
void ntfs_error(const char *fmt,...)
{
        va_list ap;

        va_start(ap, fmt);
        vfprintf(stderr, fmt, ap);
        va_end(ap);
        fputs("", stderr);
}

int ntfs_read_mft_record(ntfs_volume *vol, int mftno, char *buf)
{
	ntfs_io io;
	int error;

	io.fn_put=0;
	io.fn_get=0;
	/* record 0 of file 0 is always in memory */
	if(mftno==0)
	{
		memcpy(buf,vol->mft,vol->mft_recordsize);
		return 0;
	}
	if(!vol->mft_ino)
	{
		fprintf(stderr,"Cannot load mft %x without mft 0\n",mftno);
		return EINVAL;
	}
	io.param=buf;
	io.size=vol->mft_recordsize;
	error=ntfs_read_attr(vol->mft_ino,vol->at_data,NULL,
			   mftno*vol->mft_recordsize,&io);
	if(error)return error;
	if(io.size!=vol->mft_recordsize)return EINVAL;
	if(!ntfs_check_mft_record(vol,buf))
	{
		fprintf(stderr,"Inode not found\n");
		return EINVAL;
	}
	return 0;
}

// ReadCluster reads cluster number clusterNum from volume vol into the
// buffer buf.  This function is not part of the original Linux NTFS code;
// we wrote it to do the reading via a file handle.

int ReadCluster (ntfs_volume *vol, int clusterNum, char *buf)
{
  LARGE_INTEGER offset;
  int bytesRead;

  offset.QuadPart = clusterNum;
  offset.QuadPart *= vol->clustersize;

  if (SetFilePointer(vol->fileHandle, offset.LowPart, &offset.HighPart,
                     FILE_BEGIN) == 0xFFFFFFFF)
    return 0;
  return ReadFile(vol->fileHandle, buf, vol->clustersize, &bytesRead, NULL) &&
         (bytesRead == vol->clustersize);
}

int ntfs_getput_clusters(ntfs_volume *pvol, int cluster, ntfs_size_t offs,
        ntfs_io *buf)
{
  int bytesToCopy, bytesToRead = buf->size;
  char *p = buf->param, *clusterContents;

  if (!buf->do_read)
    return EINVAL;

  clusterContents = ntfs_malloc(pvol->clustersize);
  while (bytesToRead > 0) {
    if (!ReadCluster(pvol, cluster, clusterContents)) {
      ntfs_free(clusterContents);
      return EIO;
    }
    bytesToCopy = min((int)(pvol->clustersize - offs), bytesToRead);
    memcpy(p, clusterContents + offs, bytesToCopy);
    if (bytesToRead -= bytesToCopy) {
      offs = 0;
      cluster++;
      p += bytesToCopy;
    }
  }
  ntfs_free(clusterContents);

  ((char *)buf->param) += buf->size;
  return 0;
}

ntfs_time64_t ntfs_now(void)
{
	return ntfs_unixutc2ntutc(time(0));
}

int ntfs_dupuni2map(ntfs_volume *vol, ntfs_u16 *in, int in_len, char **out,
        int *out_len)
{
	/* Not supported here */
	return EINVAL;
}

int ntfs_dupmap2uni(ntfs_volume *vol, char* in, int in_len, ntfs_u16 **out,
        int *out_len)
{
	/* Not supported here */
	return EINVAL;
}

/*
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
