//
// volsize.c contains a routine for determining the number of clusters
// in a volume.  It contains code written by Mark Russinovitch, 1997.
//

#include "ntfstypes.h"
#include <stdio.h>
#include <windows.h>

//
// File System Control command for getting NTFS information
//
#define FSCTL_GET_VOLUME_INFORMATION     0x90064

// return code type

typedef UINT NTSTATUS;

// Error codes returned by NtFsControlFile (see NTSTATUS.H)

#define STATUS_SUCCESS                   ((NTSTATUS)0x00000000L)
#define STATUS_BUFFER_OVERFLOW           ((NTSTATUS)0x80000005L)
#define STATUS_INVALID_PARAMETER         ((NTSTATUS)0xC000000DL)
#define STATUS_BUFFER_TOO_SMALL          ((NTSTATUS)0xC0000023L)
#define STATUS_ALREADY_COMMITTED         ((NTSTATUS)0xC0000021L)
#define STATUS_INVALID_DEVICE_REQUEST    ((NTSTATUS)0xC0000010L)

// Io Status block (see NTDDK.H)

typedef struct _IO_STATUS_BLOCK {
    NTSTATUS Status;
    ULONG Information;
} IO_STATUS_BLOCK, *PIO_STATUS_BLOCK;

// Apc Routine (see NTDDK.H)

typedef VOID (*PIO_APC_ROUTINE) (PVOID ApcContext,
                                 PIO_STATUS_BLOCK IoStatusBlock,
                                 ULONG Reserved);

// NTFS volume information

typedef struct {
  LARGE_INTEGER  SerialNumber;
  LARGE_INTEGER  NumberOfSectors;
  LARGE_INTEGER  TotalClusters;
  LARGE_INTEGER  FreeClusters;
  LARGE_INTEGER  Reserved;
  ULONG          BytesPerSector;
  ULONG          BytesPerCluster;
  ULONG          BytesPerMFTRecord;
  ULONG          ClustersPerMFTRecord;
  LARGE_INTEGER  MFTLength;
  LARGE_INTEGER  MFTStart;
  LARGE_INTEGER  MFTMirrorStart;
  LARGE_INTEGER  MFTZoneStart;
  LARGE_INTEGER  MFTZoneEnd;
} NTFS_VOLUME_DATA_BUFFER, *PNTFS_VOLUME_DATA_BUFFER;

// The undocumented NtFsControlFile
//
// This function is used to send File System Control (FSCTL)
// commands into file system drivers. Its definition is 
// in ntdll.dll (ntdll.lib), a file shipped with the NTDDK.

NTSTATUS (__stdcall *NtFsControlFile)(HANDLE FileHandle,
                                      HANDLE Event,                // optional
                                      PIO_APC_ROUTINE ApcRoutine,  // optional
                                      PVOID ApcContext,            // optional
                                      PIO_STATUS_BLOCK IoStatusBlock,
                                      ULONG FsControlCode,
                                      PVOID InputBuffer,           // optional
                                      ULONG InputBufferLength,
                                      PVOID OutputBuffer,          // optional
                                      ULONG OutputBufferLength) = NULL;

// GetVolumeSize returns the number of clusters in the given volume (open
// handle fileHandle), or 0 if an error occurs trying to get that information.

int GetVolumeSize (HANDLE fileHandle)
{
  NTSTATUS status;
  NTFS_VOLUME_DATA_BUFFER volumeInfo;
  IO_STATUS_BLOCK ioStatus;

  // Query the volume information

  if (NtFsControlFile == NULL)
    NtFsControlFile = (void *) GetProcAddress(GetModuleHandle("ntdll.dll"),
                                              "NtFsControlFile");
  if (NtFsControlFile == NULL) return 0;
  status = NtFsControlFile(fileHandle, NULL, NULL, 0, &ioStatus,
                           FSCTL_GET_VOLUME_INFORMATION, NULL, 0,
                           &volumeInfo, sizeof(NTFS_VOLUME_DATA_BUFFER));
  if (status != STATUS_SUCCESS) return 0;
  return volumeInfo.TotalClusters.LowPart;
}
