//
// Compression.cpp contains routines related to file compression.
//

#include "StdAfx.h"
#include "Structs.h"
#include "Upload.h"
#include "UploadDlg.h"
#include "LogDriver.h"
#include "Helpers.h"
#include <windows.h>

////////////////////////////////////////////////////////////////////////////
//                        FILE SELECTION ROUTINES
////////////////////////////////////////////////////////////////////////////

// WorthCompressing takes as input a data structure describing a file
// (findData) and returns a boolean indicating whether the file is
// worth compressing.  It also places the number of bytes to compress
// in bytesToCompress if it returns TRUE.

BOOL WorthCompressing (WIN32_FIND_DATA& findData, DWORD& bytesToCompress)
{
  DWORD length;

  if (findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) return FALSE;
  length = strlen(findData.cFileName);
  if (length < 3) return TRUE;
  if (!strcmp(&findData.cFileName[length-3], ".gz")) return FALSE;
  if (!strcmp(&findData.cFileName[length-3], "txt")) return FALSE;

  if (findData.nFileSizeHigh)
    bytesToCompress = 0xFFFFFFFFUL;
  else
    bytesToCompress = findData.nFileSizeLow;

  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
//                           FILE COMPRESSION
////////////////////////////////////////////////////////////////////////////

// CompressFile compresses the file described in fileToCompressData, and
// updates the progress information in progressInfo as it goes along.
// The other parameter it is passed is the regInfo structure containing
// information from the registry, including the name of the main directory,
// where GZIP.EXE will be found, and the name of the trace directory, where
// the described file is.

void CompressFile (WIN32_FIND_DATA& fileToCompressData, REGISTRY_INFO& regInfo,
                   PROGRESS_INFO& progressInfo)
{
  DWORD fileSize;
  char gzipFullPath[MAX_PATH], commandLine[MAX_PATH+8],
       compressedFileName[MAX_PATH];
  STARTUPINFO startupInfo;
  PROCESS_INFORMATION processInfo;
  WIN32_FIND_DATA compressedFileData;
  double expectedCompressedSize, fractionCompressed;
  HANDLE findHandle;
  PROGRESS_INFO tempProgressInfo;

  if (!WorthCompressing(fileToCompressData, fileSize) || fileSize == 0)
    return;

  // Launch a GZIP process to compress the file

  sprintf(gzipFullPath, "%s\\GZIP.EXE", regInfo.mainDirectory);
  sprintf(commandLine, "GZIP \"%s\\%s\"", regInfo.traceDirectory,
          fileToCompressData.cFileName);
  GetStartupInfo(&startupInfo);
  if (!CreateProcess(gzipFullPath, commandLine, NULL, NULL, FALSE,
                     DETACHED_PROCESS, NULL, NULL, &startupInfo, &processInfo))
    return;

  // Figure out the expected size of the compressed file, using the
  // heuristic of 14% for metadata files and 60% for VTrace trace files.
  // Also, figure out what the name of the compressed file will be.

  expectedCompressedSize =
    (fileToCompressData.cFileName[0] == 'M' ? 0.15 : 0.56) * fileSize;
  if (expectedCompressedSize < 1.0) expectedCompressedSize = 1.0;
  sprintf(compressedFileName, "%s\\%s.gz", regInfo.traceDirectory,
          fileToCompressData.cFileName);

  // Loop until the compression process is finished.  Each time through the
  // loop, wait one second and then see how big the compressed file is.
  // Update the progress meter accordingly.

  while (WaitForSingleObject(processInfo.hProcess, 1000) == WAIT_TIMEOUT) {
    findHandle = FindFirstFile(compressedFileName, &compressedFileData);
    if (findHandle != INVALID_HANDLE_VALUE) {
      fractionCompressed = compressedFileData.nFileSizeLow /
                           expectedCompressedSize;
      if (fractionCompressed > 0.95) fractionCompressed = 0.95;
      tempProgressInfo = progressInfo;
      UpdateProgressBar(tempProgressInfo,
                        (DWORD) (fractionCompressed * fileSize));
      FindClose(findHandle);
    }
  }

  // Now that we've compressed a file, update the progress bar and indicate
  // there's more disk space available for tracing

  UpdateProgressBar(progressInfo, fileSize);
  theApp.logDriver.IndicateReducedDiskUsage();
}

// CompressAll compresses all the uncompressed files in the Traces directory.
// It updates the status message and progress bars of the given window (hWnd)
// as it goes along.  It is also passed as a parameter a registry info
// structure containing, among other things, the names of the main and trace
// directories.

void CompressAll (HWND hWnd, REGISTRY_INFO& regInfo)
{
  char searchPattern[MAX_PATH];
  HANDLE findHandle;
  WIN32_FIND_DATA findData;
  PROGRESS_INFO progressInfo;

  // Update status message

  PostMessage(hWnd, WM_CHANGE_STATUS, (WPARAM) "Preparing to compress", 0);

  // Initialize progress-related stuff

  if (!InitProgress(progressInfo, regInfo.traceDirectory, hWnd,
                    &WorthCompressing))
    return;

  // Update status message

  DisplayByteActionMessage(hWnd, "Compressing",
                           progressInfo.totalBytesToProcess);

  // Compress the files

  sprintf(searchPattern, "%s\\*", regInfo.traceDirectory);
  findHandle = FindFirstFile(searchPattern, &findData);
  if (findHandle != INVALID_HANDLE_VALUE) {
    do {
      CompressFile(findData, regInfo, progressInfo);
    } while (FindNextFile(findHandle, &findData));
    FindClose(findHandle);
  }
}
