//
// Helpers.cpp contains functions to help out the Upload program.
//

#include "StdAfx.h"
#include "Structs.h"
#include "Upload.h"
#include "UploadDlg.h"
#include <windows.h>

////////////////////////////////////////////////////////////////////////////
//                             ERROR DISPLAY
////////////////////////////////////////////////////////////////////////////

// DisplayError displays, in a message box, the error message given in the
// string generalReason as well as the error message implied by the system
// error state.

void DisplayError (char *generalReason)
{
  char *lpMsgBuf;
  CString str;

  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                NULL, GetLastError(),
                MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &lpMsgBuf,
                0, NULL);
  str.Format("%s because %s", generalReason, lpMsgBuf);
  MessageBox(NULL, str, "Error", MB_OK | MB_ICONEXCLAMATION);

  LocalFree(lpMsgBuf);
}

////////////////////////////////////////////////////////////////////////////
//                        PROGRESS DISPLAY ROUTINES
////////////////////////////////////////////////////////////////////////////

// UpdateProgressBar updates the progress bar to reflect the additional
// bytes (additionalBytes) processed since the last time.

void UpdateProgressBar (PROGRESS_INFO& progressInfo, DWORD additionalBytes)
{
  DWORD percentDone;

  progressInfo.totalBytesProcessed += additionalBytes;
  percentDone = (DWORD) (progressInfo.totalBytesProcessed * 100.0 /
                         progressInfo.totalBytesToProcess);
  PostMessage(progressInfo.hWnd, WM_CHANGE_PROGRESS, (WPARAM) percentDone, 0);
}

// InitProgress initializes progress-related stuff, namely the progress
// bar and the progress information structure.  Parameters:
//
//   progressInfo -- the progress info structure
//   directory -- the directory files to be processed are in
//   hWnd -- the window in which the progress bar exists
//   selectionFunction -- a function that, given data about a file, returns
//                        a boolean indicating whether the file should be
//                        processed and a DWORD indicating how many bytes
//                        of the file should be processed

BOOL InitProgress (PROGRESS_INFO& progressInfo, char *directory, HWND hWnd,
                   BOOL (*selectionFunction) (WIN32_FIND_DATA&, DWORD&))
{
  char searchPattern[MAX_PATH];
  WIN32_FIND_DATA findData;
  HANDLE findHandle;
  DWORD bytesToProcess = 0;

  // Set progress bar to 0

  PostMessage(hWnd, WM_CHANGE_PROGRESS, (WPARAM) 0, 0);

  // Initialize progressInfo structure

  progressInfo.hWnd = hWnd;
  progressInfo.totalBytesProcessed = 0;
  progressInfo.totalBytesToProcess = 0;

  // Figure out how many bytes need to be processed.  If it overflows,
  // just set it to 0xFFFFFFFFUL.

  sprintf(searchPattern, "%s\\*.*", directory);
  findHandle = FindFirstFile(searchPattern, &findData);
  if (findHandle != INVALID_HANDLE_VALUE) {
    do {
      if ((*selectionFunction)(findData, bytesToProcess)) {
        if (progressInfo.totalBytesToProcess + bytesToProcess <
            progressInfo.totalBytesToProcess) {
          progressInfo.totalBytesToProcess = 0xFFFFFFFFUL;
          break;
        }
        else
          progressInfo.totalBytesToProcess += bytesToProcess;
      }
    } while (FindNextFile(findHandle, &findData));
    FindClose(findHandle);
  }

  return (progressInfo.totalBytesToProcess != 0);
}

////////////////////////////////////////////////////////////////////////////
//                            STATUS MESSAGES
////////////////////////////////////////////////////////////////////////////

// DisplayByteActionMessage displays a status message about the current
// action on a set of bytes.  The parameters passed are a handle to the
// window to display to (hWnd), the action being taken [e.g., "Uploading"]
// (action), and the number of bytes (bytes).

void DisplayByteActionMessage (HWND hWnd, char *action, DWORD bytes)
{
  // The message must go into a static buffer, since the message will
  // not in general be received by the window procedure while this
  // function is still active
  static char buf[100];

  if (bytes < 1024 /* 1 KB */)
    sprintf(buf, "%s %lu bytes", action, bytes);
  else if (bytes < 1048576 /* 1 MB */)
    sprintf(buf, "%s %.2lf KB", action, (double) (bytes / 1024.0));
  else if (bytes < 1073741824 /* 1 GB */)
    sprintf(buf, "%s %.2lf MB", action, (double) (bytes / 1048576.0));
  else
    sprintf(buf, "%s %.2lf GB", action, (double) (bytes / 1073741824.0));

  PostMessage(hWnd, WM_CHANGE_STATUS, (WPARAM) buf, 0);
}

////////////////////////////////////////////////////////////////////////////
//                            REGISTRY ACCESS
////////////////////////////////////////////////////////////////////////////

// ReadRegistryString reads a string value from a registry key and returns
// a boolean indicating whether it succeeded.
//
// Parameters:
//   hKey -- the registry key to read from
//   name -- the name of the registry value
//   valuePtr -- the place to put the string value read from the key
//   maxLength -- the maximum number of bytes to read into valuePtr
//   errorDescription -- a description of this value, used in error messages
//
// If the errorDescription field is NULL, then failure should not be considered
// an error, i.e., the registry key should not be closed and no error message
// should be displayed if there is an error.

BOOL ReadRegistryString (HKEY hKey, char *name, char *valuePtr,
                         DWORD maxLength, char *errorDescription)
{
  DWORD type;
  LONG status;
  char errorString[256];

  status = RegQueryValueEx(hKey, name, 0, &type, (BYTE*) valuePtr, &maxLength);

  if (errorDescription == NULL)
    return (status == ERROR_SUCCESS && type == REG_SZ);

  if (status != ERROR_SUCCESS) {
    RegCloseKey(hKey);
    sprintf(errorString, "Could not get %s from registry", errorDescription);
    SetLastError(status);
    DisplayError(errorString);
    return FALSE;
  }
  if (type != REG_SZ) {
    RegCloseKey(hKey);
    sprintf(errorString, "The %s registry entry is invalid.",
            errorDescription);
    MessageBox(NULL, errorString, "Error", MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }

  return TRUE;
}

// ReadRegistryDword reads a DWORD value from a registry key and returns
// a boolean indicating whether it succeeded.
//
// Parameters:
//   hKey -- the registry key to read from
//   name -- the name of the registry value
//   valuePtr -- the place to put the DWORD value read from the key
//   errorDescription -- a description of this value, used in error messages
//
// If the errorDescription field is NULL, then failure should not be considered
// an error, i.e., the registry key should not be closed and no error message
// should be displayed if there is an error.

BOOL ReadRegistryDword (HKEY hKey, char *name, DWORD *valuePtr,
                        char *errorDescription)
{
  DWORD type, length;
  LONG status;
  char errorString[256];

  length = 4;
  status = RegQueryValueEx(hKey, name, 0, &type, (BYTE*) valuePtr, &length);

  if (errorDescription == NULL)
    return (status == ERROR_SUCCESS && type == REG_DWORD);

  if (status != ERROR_SUCCESS) {
    RegCloseKey(hKey);
    sprintf(errorString, "Could not get %s from registry", errorDescription);
    SetLastError(status);
    DisplayError(errorString);
    return FALSE;
  }
  if (type != REG_DWORD || length != 4) {
    RegCloseKey(hKey);
    sprintf(errorString, "The %s registry entry is invalid.",
            errorDescription);
    MessageBox(NULL, errorString, "Error", MB_OK | MB_ICONEXCLAMATION);
    return FALSE;
  }

  return TRUE;
}

// ReadRegistry reads information from the registry, displaying an error
// message and returning FALSE if it is unsuccessful, and returning TRUE if
// it is successful.  It puts the information in the structure regInfo.

BOOL ReadRegistry (REGISTRY_INFO& regInfo)
{
  HKEY hKey;
  LONG status;

  // Open registry key

  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                        "Software\\Jay Lorch\\VTrace\\CurrentVersion", 0,
                        KEY_QUERY_VALUE, &hKey);
  if (status != ERROR_SUCCESS) {
    SetLastError(status);
    DisplayError("Could not open the registry key for VTrace");
    return FALSE;
  }

  // Read various parameters from the registry that we must fail if we
  // don't find

  if (!ReadRegistryDword(hKey, "Version", &regInfo.vtraceVersion,
                         "version number"))
    return FALSE;

  if (!ReadRegistryString(hKey, "MainDirectory", regInfo.mainDirectory,
                          MAX_PATH, "VTrace main directory"))
    return FALSE;

  if (!ReadRegistryDword(hKey, "UserNumber", &regInfo.userNumber,
                         "user number"))
    return FALSE;

  // Deduce the trace directory from the main directory

  sprintf(regInfo.traceDirectory, "%s\\Traces", regInfo.mainDirectory);

  // Read some parameters from the registry that we can set to their
  // defaults if we don't find

  if (!ReadRegistryString(hKey, "ProxyServer", regInfo.proxyServer,
                          MAX_SERVER_LENGTH, NULL))
    regInfo.proxyServer[0] = '\0';

  if (!ReadRegistryDword(hKey, "ProxyPort", &regInfo.proxyPort, NULL))
    regInfo.proxyPort = DEFAULT_PROXY_PORT;

    // Make sure that maxTransferSize is within the proper range

  if (!ReadRegistryDword(hKey, "MaxTransferKB", &regInfo.maxTransferSize,
                         NULL) ||
      regInfo.maxTransferSize < MIN_TRANSFER_SIZE_LIMIT ||
      regInfo.maxTransferSize > MAX_TRANSFER_SIZE_LIMIT)
    regInfo.maxTransferSize = DEFAULT_TRANSFER_SIZE_LIMIT * 1024;
  else
    regInfo.maxTransferSize *= 1024;

  if (!ReadRegistryString(hKey, "UploadUrlList", regInfo.uploadUrlList,
                          MAX_UPLOAD_URL_LIST_LENGTH, NULL))
    strcpy(regInfo.uploadUrlList, DEFAULT_UPLOAD_URL_LIST);

  RegCloseKey(hKey);
  return TRUE;
}

// GetCurrentTraceFileNumber obtains the current trace file number and puts
// it in currentTraceFileNumberPtr. It returns a boolean indicating if it
// was successful.

BOOL GetCurrentTraceFileNumber (DWORD *currentTraceFileNumberPtr)
{
  HKEY hKey;
  LONG status;

  // Open registry key

  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                        "Software\\Jay Lorch\\VTrace\\CurrentVersion", 0,
                        KEY_QUERY_VALUE, &hKey);
  if (status != ERROR_SUCCESS) {
    SetLastError(status);
    DisplayError("Could not open the registry key for VTrace");
    return FALSE;
  }

  // Read various parameters from the registry that we must fail if we
  // don't find

  if (!ReadRegistryDword(hKey, "LastTraceNumber", currentTraceFileNumberPtr,
                         "last trace number"))
    return FALSE;

  RegCloseKey(hKey);
  return TRUE;
}
