//
// FastIo.c
//
// Code for fast I/O routines in the VTrcFS driver.
//
// (Uses some code by Mark Russinovich and Bryce Cogswell)
//

#include "NTDDK.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "VTrcFS.h"
#include "FastIo.h"
#include "FileName.h"
#include "GetState.h"
#include "Hash.h"

///////////////////////////////////////////////////////////////////////////
//                                 MACROS
///////////////////////////////////////////////////////////////////////////

// Print macro that only turns on when checked builds are on

#if DBG
#define DebugPrint(arg) DbgPrint arg
#else
#define DebugPrint(arg)
#endif

////////////////////////////////////////////////////////////////////////////
//                           FORWARD DEFINITIONS
////////////////////////////////////////////////////////////////////////////

BOOLEAN VTrcFSFastIoCheckIfPossible (IN PFILE_OBJECT FileObject,
                                     IN PLARGE_INTEGER FileOffset,
                                     IN ULONG Length, IN BOOLEAN Wait,
                                     IN ULONG LockKey,
                                     IN BOOLEAN CheckForReadOperation,
                                     OUT PIO_STATUS_BLOCK IoStatus,
                                     IN PDEVICE_OBJECT DeviceObject);
BOOLEAN VTrcFSFastIoRead (IN PFILE_OBJECT FileObject,
                          IN PLARGE_INTEGER FileOffset, IN ULONG Length,
                          IN BOOLEAN Wait, IN ULONG LockKey, OUT PVOID Buffer,
                          OUT PIO_STATUS_BLOCK IoStatus,
                          IN PDEVICE_OBJECT DeviceObject);
BOOLEAN VTrcFSFastIoWrite (IN PFILE_OBJECT FileObject,
                           IN PLARGE_INTEGER FileOffset,
                           IN ULONG Length, IN BOOLEAN Wait, IN ULONG LockKey,
                           IN PVOID Buffer, OUT PIO_STATUS_BLOCK IoStatus,
                           IN PDEVICE_OBJECT DeviceObject);
BOOLEAN VTrcFSFastIoQueryBasicInfo (IN PFILE_OBJECT FileObject,
                                    IN BOOLEAN Wait,
                                    OUT PFILE_BASIC_INFORMATION Buffer,
                                    OUT PIO_STATUS_BLOCK IoStatus,
                                    IN PDEVICE_OBJECT DeviceObject);
BOOLEAN VTrcFSFastIoQueryStandardInfo (IN PFILE_OBJECT FileObject,
                                       IN BOOLEAN Wait,
                                       OUT PFILE_STANDARD_INFORMATION Buffer,
                                       OUT PIO_STATUS_BLOCK IoStatus,
                                       IN PDEVICE_OBJECT DeviceObject);
BOOLEAN VTrcFSFastIoLock (IN PFILE_OBJECT FileObject,
                          IN PLARGE_INTEGER FileOffset,
                          IN PLARGE_INTEGER Length, PEPROCESS ProcessId,
                          ULONG Key, BOOLEAN FailImmediately,
                          BOOLEAN ExclusiveLock, OUT PIO_STATUS_BLOCK IoStatus,
                          IN PDEVICE_OBJECT DeviceObject);
BOOLEAN VTrcFSFastIoUnlockSingle (IN PFILE_OBJECT FileObject,
                                  IN PLARGE_INTEGER FileOffset,
                                  IN PLARGE_INTEGER Length,
                                  PEPROCESS ProcessId, ULONG Key,
                                  OUT PIO_STATUS_BLOCK IoStatus,
                                  IN PDEVICE_OBJECT DeviceObject);
BOOLEAN VTrcFSFastIoUnlockAll (IN PFILE_OBJECT FileObject, PEPROCESS ProcessId,
                               OUT PIO_STATUS_BLOCK IoStatus,
                               IN PDEVICE_OBJECT DeviceObject);
BOOLEAN VTrcFSFastIoUnlockAllByKey (IN PFILE_OBJECT FileObject,
                                    PEPROCESS ProcessId, ULONG Key,
                                    OUT PIO_STATUS_BLOCK IoStatus,
                                    IN PDEVICE_OBJECT DeviceObject);
BOOLEAN VTrcFSFastIoDeviceControl (IN PFILE_OBJECT FileObject, IN BOOLEAN Wait,
                                   IN PVOID InputBuffer,
                                   IN ULONG InputBufferLength,
                                   OUT PVOID OutputBuffer,
                                   IN ULONG OutputBufferLength,
                                   IN ULONG IoControlCode,
                                   OUT PIO_STATUS_BLOCK IoStatus,
                                   IN PDEVICE_OBJECT DeviceObject);
VOID VTrcFSAcquireFileForNtCreateSection (PFILE_OBJECT FileObject);
VOID VTrcFSReleaseFileForNtCreateSection (PFILE_OBJECT FileObject);
VOID VTrcFSFastIoDetachDevice (PDEVICE_OBJECT SourceDevice,
                               PDEVICE_OBJECT TargetDevice);
BOOLEAN VTrcFSFastIoQueryNetworkOpenInfo
  (IN PFILE_OBJECT FileObject, IN BOOLEAN Wait,
   OUT struct _FILE_NETWORK_OPEN_INFORMATION *Buffer,
   OUT struct _IO_STATUS_BLOCK *IoStatus, IN PDEVICE_OBJECT DeviceObject);
NTSTATUS VTrcFSAcquireForModWrite (IN PFILE_OBJECT FileObject,
                                   IN PLARGE_INTEGER EndingOffset,
                                   OUT struct _ERESOURCE **ResourceToRelease,
                                   IN PDEVICE_OBJECT DeviceObject);
BOOLEAN VTrcFSMdlRead (IN PFILE_OBJECT FileObject,
                       IN PLARGE_INTEGER FileOffset, IN ULONG Length,
                       IN ULONG LockKey, OUT PMDL *MdlChain,
                       OUT PIO_STATUS_BLOCK IoStatus,
                       IN PDEVICE_OBJECT DeviceObject);
BOOLEAN VTrcFSMdlReadComplete (IN PFILE_OBJECT FileObject, IN PMDL MdlChain,
                               IN PDEVICE_OBJECT DeviceObject);
BOOLEAN VTrcFSPrepareMdlWrite (IN PFILE_OBJECT FileObject,
                               IN PLARGE_INTEGER FileOffset, IN ULONG Length,
                               IN ULONG LockKey, OUT PMDL *MdlChain,
                               OUT PIO_STATUS_BLOCK IoStatus,
                               IN PDEVICE_OBJECT DeviceObject);
BOOLEAN VTrcFSMdlWriteComplete (IN PFILE_OBJECT FileObject,
                                IN PLARGE_INTEGER FileOffset, IN PMDL MdlChain,
                                IN PDEVICE_OBJECT DeviceObject);
BOOLEAN VTrcFSFastIoReadCompressed
  (IN PFILE_OBJECT FileObject, IN PLARGE_INTEGER FileOffset,
   IN ULONG Length, IN ULONG LockKey, OUT PVOID Buffer, OUT PMDL *MdlChain,
   OUT PIO_STATUS_BLOCK IoStatus,
   OUT struct _COMPRESSED_DATA_INFO *CompressedDataInfo,
   IN ULONG CompressedDataInfoLength, IN PDEVICE_OBJECT DeviceObject);
BOOLEAN VTrcFSFastIoWriteCompressed
  (IN PFILE_OBJECT FileObject, IN PLARGE_INTEGER FileOffset,
   IN ULONG Length, IN ULONG LockKey, IN PVOID Buffer, OUT PMDL *MdlChain,
   OUT PIO_STATUS_BLOCK IoStatus,
   IN struct _COMPRESSED_DATA_INFO *CompressedDataInfo,
   IN ULONG CompressedDataInfoLength, IN PDEVICE_OBJECT DeviceObject);
BOOLEAN VTrcFSMdlReadCompleteCompressed (IN PFILE_OBJECT FileObject,
                                         IN PMDL MdlChain,
                                         IN PDEVICE_OBJECT DeviceObject);
BOOLEAN VTrcFSMdlWriteCompleteCompressed (IN PFILE_OBJECT FileObject,
                                          IN PLARGE_INTEGER FileOffset,
                                          IN PMDL MdlChain,
                                          IN PDEVICE_OBJECT DeviceObject);
BOOLEAN VTrcFSFastIoQueryOpen
  (IN struct _IRP *Irp, OUT PFILE_NETWORK_OPEN_INFORMATION NetworkInformation,
   IN PDEVICE_OBJECT DeviceObject);
NTSTATUS VTrcFSReleaseForModWrite (IN PFILE_OBJECT FileObject,
                                   IN struct _ERESOURCE *ResourceToRelease,
                                   IN PDEVICE_OBJECT DeviceObject);
NTSTATUS VTrcFSAcquireForCcFlush (IN PFILE_OBJECT FileObject,
                                  IN PDEVICE_OBJECT DeviceObject);
NTSTATUS VTrcFSReleaseForCcFlush (IN PFILE_OBJECT FileObject,
                                  IN PDEVICE_OBJECT DeviceObject);

///////////////////////////////////////////////////////////////////////////
//                              GLOBALS
///////////////////////////////////////////////////////////////////////////

// This is VTrcFS's Fast I/O dispatch table. Note that NT assumes that
// file system drivers support some Fast I/O calls, so this table must
// be present for a file system filter driver.

FAST_IO_DISPATCH VTrcFSFastIoDispatchTable = {
    sizeof(FAST_IO_DISPATCH),
    VTrcFSFastIoCheckIfPossible,
    VTrcFSFastIoRead,
    VTrcFSFastIoWrite,
    VTrcFSFastIoQueryBasicInfo,
    VTrcFSFastIoQueryStandardInfo,
    VTrcFSFastIoLock,
    VTrcFSFastIoUnlockSingle,
    VTrcFSFastIoUnlockAll,
    VTrcFSFastIoUnlockAllByKey,
    VTrcFSFastIoDeviceControl,
    VTrcFSAcquireFileForNtCreateSection,
    VTrcFSReleaseFileForNtCreateSection,
    VTrcFSFastIoDetachDevice,
    VTrcFSFastIoQueryNetworkOpenInfo,      // From here on is new for NT 4.0
    VTrcFSAcquireForModWrite,
    VTrcFSMdlRead,
    VTrcFSMdlReadComplete,
    VTrcFSPrepareMdlWrite,
    VTrcFSMdlWriteComplete,
    VTrcFSFastIoReadCompressed,
    VTrcFSFastIoWriteCompressed,
    VTrcFSMdlReadCompleteCompressed,
    VTrcFSMdlWriteCompleteCompressed,
    VTrcFSFastIoQueryOpen,
    VTrcFSReleaseForModWrite,
    VTrcFSAcquireForCcFlush,
    VTrcFSReleaseForCcFlush
};

///////////////////////////////////////////////////////////////////////////
//                       FAST I/O DISPATCH ROUTINES
///////////////////////////////////////////////////////////////////////////

// VTrcFSFastIoCheckIfPossible returns true if the real fast I/O routine
// exists and returns true

BOOLEAN VTrcFSFastIoCheckIfPossible (IN PFILE_OBJECT FileObject,
                                     IN PLARGE_INTEGER FileOffset,
                                     IN ULONG Length, IN BOOLEAN Wait,
                                     IN ULONG LockKey,
                                     IN BOOLEAN CheckForReadOperation,
                                     OUT PIO_STATUS_BLOCK IoStatus,
                                     IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 8 ||
      origFastIoTable->FastIoCheckIfPossible == NULL)
    return FALSE;

  // Call the original driver's fast I/O dispatch routine

  return origFastIoTable->FastIoCheckIfPossible(FileObject, FileOffset, Length,
                                                Wait, LockKey,
                                                CheckForReadOperation,
                                                IoStatus, origDevice);
}

// VTrcFSFastIoRead traces incoming parameters and return values.

BOOLEAN VTrcFSFastIoRead (IN PFILE_OBJECT FileObject,
                          IN PLARGE_INTEGER FileOffset, IN ULONG Length,
                          IN BOOLEAN Wait, IN ULONG LockKey, OUT PVOID Buffer,
                          OUT PIO_STATUS_BLOCK IoStatus,
                          IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;
  BOOLEAN retval;
  LONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer = NULL;
  ULONG threadId;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 12 ||
      origFastIoTable->FastIoRead == NULL)
    return FALSE;

  // If there is a file name, record this call

  if (FileObject->FileName.Buffer != NULL) {
    // Indicate that the current thread is active

    threadId = (ULONG) PsGetCurrentThreadId();
    if (threadId < sharedState->kernelUserSharedState.maxThreads)
      sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] =
        '\0';

    // Record information for the file if we haven't already

    if (Wait) NameFileIfUnopened(FileObject, origDevice, TRUE);

    // Log this call, describing the input parameters

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_READ, 24);
    if (eventBuffer) {
      // Get a new sequence number to keep track of this request
      seq = InterlockedIncrement(&sequenceNumber);
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &FileObject, 4);
      RtlCopyMemory(&eventBuffer[9], FileOffset, 5);
      RtlCopyMemory(&eventBuffer[14], &Length, 4);
      // no IRP flags, obviously
      RtlZeroMemory(&eventBuffer[18], 4);
      eventBuffer[22] = IRP_MN_NORMAL;
      eventBuffer[23] = '\0';
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Call the real fast I/O routine, recording the return value

  retval = origFastIoTable->FastIoRead(FileObject, FileOffset, Length,
                                       Wait, LockKey, Buffer, IoStatus,
                                       origDevice);

  if (eventBuffer) {
    // Log the return, including return values

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_OPERATION, 14);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &IoStatus->Status, 4);
      RtlCopyMemory(&eventBuffer[9], &IoStatus->Information, 4);
      if (retval)
        eventBuffer[13] = FLAG_COMPLETED_SUCCESSFULLY | FLAG_FAST_IO;
      else
        eventBuffer[13] = FLAG_FAST_IO;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Return the original routine's return value

  return retval;
}

// VTrcFSFastIoWrite traces incoming parameters and return values.

BOOLEAN VTrcFSFastIoWrite (IN PFILE_OBJECT FileObject,
                           IN PLARGE_INTEGER FileOffset,
                           IN ULONG Length, IN BOOLEAN Wait, IN ULONG LockKey,
                           IN PVOID Buffer, OUT PIO_STATUS_BLOCK IoStatus,
                           IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;
  BOOLEAN retval;
  LONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer = NULL;
  ULONG threadId;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 16 ||
      origFastIoTable->FastIoWrite == NULL)
    return FALSE;

  // If there is a file name, record this call

  if (FileObject->FileName.Buffer != NULL) {
    // Indicate that the current thread is active

    threadId = (ULONG) PsGetCurrentThreadId();
    if (threadId < sharedState->kernelUserSharedState.maxThreads)
      sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] =
        '\0';

    // Record information for the file if we haven't already

    if (Wait) NameFileIfUnopened(FileObject, origDevice, TRUE);

    // Log this call, describing the input parameters

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_WRITE, 24);
    if (eventBuffer) {
      // Get a new sequence number to keep track of this request
      seq = InterlockedIncrement(&sequenceNumber);
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &FileObject, 4);
      RtlCopyMemory(&eventBuffer[9], FileOffset, 5);
      RtlCopyMemory(&eventBuffer[14], &Length, 4);
      RtlZeroMemory(&eventBuffer[18], 4);
      eventBuffer[22] = IRP_MN_NORMAL;
      eventBuffer[23] = '\0';
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Call the real fast I/O routine, recording the return value

  retval = origFastIoTable->FastIoWrite(FileObject, FileOffset, Length, Wait,
                                        LockKey, Buffer, IoStatus, origDevice);

  // Log the return, including return values

  if (eventBuffer) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_OPERATION, 14);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &IoStatus->Status, 4);
      RtlCopyMemory(&eventBuffer[9], &IoStatus->Information, 4);
      if (retval)
        eventBuffer[13] = FLAG_COMPLETED_SUCCESSFULLY | FLAG_FAST_IO;
      else
        eventBuffer[13] = FLAG_FAST_IO;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Return the original routine's return value

  return retval;
}

// VTrcFSFastIoQueryBasicinfo

BOOLEAN VTrcFSFastIoQueryBasicInfo (IN PFILE_OBJECT FileObject,
                                    IN BOOLEAN Wait,
                                    OUT PFILE_BASIC_INFORMATION Buffer,
                                    OUT PIO_STATUS_BLOCK IoStatus,
                                    IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;
  BOOLEAN retval;
  LONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer = NULL;
  ULONG threadId;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 20 ||
      origFastIoTable->FastIoQueryBasicInfo == NULL)
    return FALSE;

  // If there is a file name, log this call

  if (FileObject->FileName.Buffer != NULL) {
    // Indicate that the current thread is active

    threadId = (ULONG) PsGetCurrentThreadId();
    if (threadId < sharedState->kernelUserSharedState.maxThreads)
      sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] =
        '\0';

    // Record information for the file if we haven't already

    if (Wait) NameFileIfUnopened(FileObject, origDevice, TRUE);

    // Log this call, describing the input parameters

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_QUERY_INFO, 10);
    if (eventBuffer) {
      // Get a new sequence number to keep track of this request
      seq = InterlockedIncrement(&sequenceNumber);
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &FileObject, 4);
      eventBuffer[9] = (char) FileBasicInformation;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Call the real fast I/O routine, recording the return value

  retval = origFastIoTable->FastIoQueryBasicInfo(FileObject, Wait, Buffer,
                                                 IoStatus, origDevice);

  // Log the return, including return values

  if (eventBuffer) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_OPERATION, 14);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &IoStatus->Status, 4);
      RtlCopyMemory(&eventBuffer[9], &IoStatus->Information, 4);
      if (retval)
        eventBuffer[13] = FLAG_COMPLETED_SUCCESSFULLY | FLAG_FAST_IO;
      else
        eventBuffer[13] = FLAG_FAST_IO;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Return the original routine's return value

  return retval;
}

// VTrcFSFastIoQueryStandardInfo

BOOLEAN VTrcFSFastIoQueryStandardInfo (IN PFILE_OBJECT FileObject,
                                       IN BOOLEAN Wait,
                                       OUT PFILE_STANDARD_INFORMATION Buffer,
                                       OUT PIO_STATUS_BLOCK IoStatus,
                                       IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;
  BOOLEAN retval;
  LONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer = NULL;
  ULONG threadId;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 24 ||
      origFastIoTable->FastIoQueryStandardInfo == NULL)
    return FALSE;

  // If there is a file name, log this call

  if (FileObject->FileName.Buffer != NULL) {
    // Indicate that the current thread is active

    threadId = (ULONG) PsGetCurrentThreadId();
    if (threadId < sharedState->kernelUserSharedState.maxThreads)
      sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] =
        '\0';

    // Record information for the file if we haven't already

    if (Wait) NameFileIfUnopened(FileObject, origDevice, TRUE);

    // Log this call, describing the input parameters
 
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_QUERY_INFO, 10);
    if (eventBuffer) {
      // Get a new sequence number to keep track of this request
      seq = InterlockedIncrement(&sequenceNumber);
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &FileObject, 4);
      eventBuffer[9] = (char) FileStandardInformation;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Call the real fast I/O routine, recording the return value

  retval = origFastIoTable->FastIoQueryStandardInfo(FileObject, Wait, Buffer,
                                                    IoStatus, origDevice);

  // Log the return, including return values

  if (eventBuffer) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_OPERATION, 14);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &IoStatus->Status, 4);
      RtlCopyMemory(&eventBuffer[9], &IoStatus->Information, 4);
      if (retval)
        eventBuffer[13] = FLAG_COMPLETED_SUCCESSFULLY | FLAG_FAST_IO;
      else
        eventBuffer[13] = FLAG_FAST_IO;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }
  
  // Return the original routine's return value

  return retval;
}

// VTrcFSFastIoLock

BOOLEAN VTrcFSFastIoLock (IN PFILE_OBJECT FileObject,
                          IN PLARGE_INTEGER FileOffset,
                          IN PLARGE_INTEGER Length, PEPROCESS ProcessId,
                          ULONG Key, BOOLEAN FailImmediately,
                          BOOLEAN ExclusiveLock, OUT PIO_STATUS_BLOCK IoStatus,
                          IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;
  BOOLEAN retval;
  LONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer = NULL;
  ULONG threadId;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 28 ||
      origFastIoTable->FastIoLock == NULL)
    return FALSE;

  // If there is a file name, log this call

  if (FileObject->FileName.Buffer != NULL) {
    // Indicate that the current thread is active

    threadId = (ULONG) PsGetCurrentThreadId();
    if (threadId < sharedState->kernelUserSharedState.maxThreads)
      sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] =
        '\0';

    // Record information for the file if we haven't already

    NameFileIfUnopened(FileObject, origDevice, TRUE);

    // Log this call, describing the input parameters

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_LOCK, 24);
    if (eventBuffer) {
      // Get a new sequence number to keep track of this request
      seq = InterlockedIncrement(&sequenceNumber);
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &FileObject, 4);
      eventBuffer[9] = (FailImmediately ? SL_FAIL_IMMEDIATELY : 0) |
                       (ExclusiveLock ? SL_EXCLUSIVE_LOCK : 0);
      RtlCopyMemory(&eventBuffer[10], Length, 5);
      RtlCopyMemory(&eventBuffer[15], &Key, 4);
      RtlCopyMemory(&eventBuffer[19], FileOffset, 5);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Call the original driver's fast I/O routine

  retval = origFastIoTable->FastIoLock(FileObject, FileOffset, Length,
                                       ProcessId, Key, FailImmediately,
                                       ExclusiveLock, IoStatus, origDevice);

  // Log the return, including return values

  if (eventBuffer) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_OPERATION, 14);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &IoStatus->Status, 4);
      RtlCopyMemory(&eventBuffer[9], &IoStatus->Information, 4);
      if (retval)
        eventBuffer[13] = FLAG_COMPLETED_SUCCESSFULLY | FLAG_FAST_IO;
      else
        eventBuffer[13] = FLAG_FAST_IO;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Return the original routine's return value

  return retval;
}

// VTrcFSFastIoUnlockSingle

BOOLEAN VTrcFSFastIoUnlockSingle (IN PFILE_OBJECT FileObject,
                                  IN PLARGE_INTEGER FileOffset,
                                  IN PLARGE_INTEGER Length,
                                  PEPROCESS ProcessId, ULONG Key,
                                  OUT PIO_STATUS_BLOCK IoStatus,
                                  IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;
  BOOLEAN retval;
  LONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer = NULL;
  ULONG threadId;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 32 ||
      origFastIoTable->FastIoUnlockSingle == NULL)
    return FALSE;

  // If there is a file name, log this call

  if (FileObject->FileName.Buffer != NULL) {
    // Indicate that the current thread is active

    threadId = (ULONG) PsGetCurrentThreadId();
    if (threadId < sharedState->kernelUserSharedState.maxThreads)
      sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] =
        '\0';

    // Record information for the file if we haven't already

    NameFileIfUnopened(FileObject, origDevice, TRUE);

    // Log this call, describing the input parameters

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_UNLOCK, 25);
    if (eventBuffer) {
      // Get a new sequence number to keep track of this request
      seq = InterlockedIncrement(&sequenceNumber);
      eventBuffer[1] = (char) 25;
      eventBuffer[2] = IRP_MN_UNLOCK_SINGLE;
      RtlCopyMemory(&eventBuffer[3], &seq, 4);
      RtlCopyMemory(&eventBuffer[7], &FileObject, 4);
      if (Length == NULL)
        RtlZeroMemory(&eventBuffer[11], 5);
      else
        RtlCopyMemory(&eventBuffer[11], Length, 5);
      RtlCopyMemory(&eventBuffer[16], &Key, 4);
      if (FileOffset == NULL)
        RtlZeroMemory(&eventBuffer[20], 5);
      else
        RtlCopyMemory(&eventBuffer[20], FileOffset, 5);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Call the original driver's fast I/O routine

  retval = origFastIoTable->FastIoUnlockSingle(FileObject, FileOffset, Length,
                                               ProcessId, Key, IoStatus,
                                               origDevice);

  // Log the return, including return values

  if (eventBuffer) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_OPERATION, 14);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &IoStatus->Status, 4);
      RtlCopyMemory(&eventBuffer[9], &IoStatus->Information, 4);
      if (retval)
        eventBuffer[13] = FLAG_COMPLETED_SUCCESSFULLY | FLAG_FAST_IO;
      else
        eventBuffer[13] = FLAG_FAST_IO;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Return the original routine's return value

  return retval;
}

// VTrcFSFastIoUnlockAll

BOOLEAN VTrcFSFastIoUnlockAll (IN PFILE_OBJECT FileObject, PEPROCESS ProcessId,
                               OUT PIO_STATUS_BLOCK IoStatus,
                               IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;
  BOOLEAN retval;
  LONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer = NULL;
  ULONG threadId;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 36 ||
      origFastIoTable->FastIoUnlockAll == NULL)
    return FALSE;

  // If there is a file name, log this call

  if (FileObject->FileName.Buffer != NULL) {
    // Indicate that the current thread is active

    threadId = (ULONG) PsGetCurrentThreadId();
    if (threadId < sharedState->kernelUserSharedState.maxThreads)
      sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] =
        '\0';

    // Record information for the file if we haven't already

    NameFileIfUnopened(FileObject, origDevice, TRUE);

    // Log this call, describing the input parameters

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_UNLOCK, 11);
    if (eventBuffer) {
      // Get a new sequence number to keep track of this request
      seq = InterlockedIncrement(&sequenceNumber);
      eventBuffer[1] = (char) 11;
      eventBuffer[2] = IRP_MN_UNLOCK_ALL;
      RtlCopyMemory(&eventBuffer[3], &seq, 4);
      RtlCopyMemory(&eventBuffer[7], &FileObject, 4);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Call the original driver's fast I/O routine

  retval = origFastIoTable->FastIoUnlockAll(FileObject, ProcessId, IoStatus,
                                            origDevice);

  // Log the return, including return values

  if (eventBuffer) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_OPERATION, 14);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &IoStatus->Status, 4);
      RtlCopyMemory(&eventBuffer[9], &IoStatus->Information, 4);
      if (retval)
        eventBuffer[13] = FLAG_COMPLETED_SUCCESSFULLY | FLAG_FAST_IO;
      else
        eventBuffer[13] = FLAG_FAST_IO;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Return the original routine's return value

  return retval;
}

// VTrcFSFastIoUnlockAllByKey

BOOLEAN VTrcFSFastIoUnlockAllByKey (IN PFILE_OBJECT FileObject,
                                    PEPROCESS ProcessId, ULONG Key,
                                    OUT PIO_STATUS_BLOCK IoStatus,
                                    IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;
  BOOLEAN retval;
  LONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer = NULL;
  ULONG threadId;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 40 ||
      origFastIoTable->FastIoUnlockAllByKey == NULL)
    return FALSE;

  // If there is a file name, log this call

  if (FileObject->FileName.Buffer != NULL) {
    // Indicate that the current thread is active

    threadId = (ULONG) PsGetCurrentThreadId();
    if (threadId < sharedState->kernelUserSharedState.maxThreads)
      sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] =
        '\0';

    // Record information for the file if we haven't already

    NameFileIfUnopened(FileObject, origDevice, TRUE);

    // Log this call, describing the input parameters

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_UNLOCK, 15);
    if (eventBuffer) {
      // Get a new sequence number to keep track of this request
      seq = InterlockedIncrement(&sequenceNumber);
      eventBuffer[1] = (char) 15;
      eventBuffer[2] = IRP_MN_UNLOCK_ALL_BY_KEY;
      RtlCopyMemory(&eventBuffer[3], &seq, 4);
      RtlCopyMemory(&eventBuffer[7], &FileObject, 4);
      RtlCopyMemory(&eventBuffer[11], &Key, 4);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Call the original driver's fast I/O routine

  retval = origFastIoTable->FastIoUnlockAllByKey(FileObject, ProcessId, Key,
                                                 IoStatus, origDevice);

  // Log the return, including return values

  if (eventBuffer) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_OPERATION, 14);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &IoStatus->Status, 4);
      RtlCopyMemory(&eventBuffer[9], &IoStatus->Information, 4);
      if (retval)
        eventBuffer[13] = FLAG_COMPLETED_SUCCESSFULLY | FLAG_FAST_IO;
      else
        eventBuffer[13] = FLAG_FAST_IO;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Return the original routine's return value

  return retval;
}

// VTrcFSFastIoDeviceControl

BOOLEAN VTrcFSFastIoDeviceControl (IN PFILE_OBJECT FileObject, IN BOOLEAN Wait,
                                   IN PVOID InputBuffer,
                                   IN ULONG InputBufferLength,
                                   OUT PVOID OutputBuffer,
                                   IN ULONG OutputBufferLength,
                                   IN ULONG IoControlCode,
                                   OUT PIO_STATUS_BLOCK IoStatus,
                                   IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 44 ||
      origFastIoTable->FastIoDeviceControl == NULL)
    return FALSE;

  // Call the original driver's fast I/O routine

  return origFastIoTable->FastIoDeviceControl(FileObject, Wait, InputBuffer,
                                              InputBufferLength, OutputBuffer,
                                              OutputBufferLength,
                                              IoControlCode, IoStatus,
                                              origDevice);
}

// VTrcFSAcquireFileForNtCreateSection

VOID VTrcFSAcquireFileForNtCreateSection (PFILE_OBJECT FileObject)
{
  PDEVICE_OBJECT deviceObject = IoGetRelatedDeviceObject(FileObject);
  PFS_HOOK_EXTENSION hookExt = deviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return

  if (origFastIoTable->SizeOfFastIoDispatch < 48 ||
      origFastIoTable->AcquireFileForNtCreateSection == NULL)
    return;

  // Call the original driver's fast I/O routine

  origFastIoTable->AcquireFileForNtCreateSection(FileObject);
}

// VTrcFSReleaseFileForNtCreateSection

VOID VTrcFSReleaseFileForNtCreateSection (PFILE_OBJECT FileObject)
{
  PDEVICE_OBJECT deviceObject = IoGetRelatedDeviceObject(FileObject);
  PFS_HOOK_EXTENSION hookExt = deviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 52 ||
      origFastIoTable->ReleaseFileForNtCreateSection == NULL)
    return;

  // Call the original driver's fast I/O routine

  origFastIoTable->ReleaseFileForNtCreateSection(FileObject);
}

// VTrcFSFastIoDetachDevice is called when a device that we have hooked
// is being deleted. This happens when, for example, a floppy is formatted.
// We have to detach from it and delete our device.

VOID VTrcFSFastIoDetachDevice (PDEVICE_OBJECT SourceDevice,
                               PDEVICE_OBJECT TargetDevice)
{
  PFS_HOOK_EXTENSION hookExt = SourceDevice->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  KIRQL oldirql;
  PCHAR eventBuffer;

  DebugPrint(("VTrcFS: Detaching %x from drive %x.\n",
              SourceDevice, TargetDevice));

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_DEVICE_UNHOOK, 5);
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[1], &origDevice, 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  IoDetachDevice(TargetDevice);
  IoDeleteDevice(SourceDevice);

  // No need to call original detach device routine, since it will be called
  // automatically by the I/O manager
}

// VTrcFSFastIoQueryNetworkOpenInfo

BOOLEAN VTrcFSFastIoQueryNetworkOpenInfo
          (IN PFILE_OBJECT FileObject, IN BOOLEAN Wait,
           OUT struct _FILE_NETWORK_OPEN_INFORMATION *Buffer,
           OUT struct _IO_STATUS_BLOCK *IoStatus,
           IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;
  BOOLEAN retval;
  LONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer = NULL;
  ULONG threadId;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 60 ||
      origFastIoTable->FastIoQueryNetworkOpenInfo == NULL)
    return FALSE;

  // If there is a file name, log this call

  if (FileObject->FileName.Buffer != NULL) {
    // Indicate that the current thread is active

    threadId = (ULONG) PsGetCurrentThreadId();
    if (threadId < sharedState->kernelUserSharedState.maxThreads)
      sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] =
        '\0';

    // Record information for the file if we haven't already

    if (Wait) NameFileIfUnopened(FileObject, origDevice, TRUE);

    // Log this call, describing the input parameters

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_QUERY_INFO, 10);
    if (eventBuffer) {
      // Get a new sequence number to keep track of this request
      seq = InterlockedIncrement(&sequenceNumber);
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &FileObject, 4);
      eventBuffer[9] = (char) FileNetworkOpenInformation;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Call the real fast I/O routine, recording the return value

  retval = origFastIoTable->FastIoQueryNetworkOpenInfo(FileObject, Wait,
                                                       Buffer, IoStatus,
                                                       origDevice);

  // Log the return, including return values

  if (eventBuffer) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_OPERATION, 14);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &IoStatus->Status, 4);
      RtlCopyMemory(&eventBuffer[9], &IoStatus->Information, 4);
      if (retval)
        eventBuffer[13] = FLAG_COMPLETED_SUCCESSFULLY | FLAG_FAST_IO;
      else
        eventBuffer[13] = FLAG_FAST_IO;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Return the original routine's return value

  return retval;
}

// VTrcFSAcquireForModWrite

NTSTATUS VTrcFSAcquireForModWrite (IN PFILE_OBJECT FileObject,
                                   IN PLARGE_INTEGER EndingOffset,
                                   OUT struct _ERESOURCE **ResourceToRelease,
                                   IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 64 ||
      origFastIoTable->AcquireForModWrite == NULL)
    return FALSE;

  // Call the original driver's fast I/O routine

  return origFastIoTable->AcquireForModWrite(FileObject, EndingOffset,
                                             ResourceToRelease, origDevice);
}

// VTrcFSMdlRead traces incoming parameters and return values.

BOOLEAN VTrcFSMdlRead (IN PFILE_OBJECT FileObject,
                       IN PLARGE_INTEGER FileOffset, IN ULONG Length,
                       IN ULONG LockKey, OUT PMDL *MdlChain,
                       OUT PIO_STATUS_BLOCK IoStatus,
                       IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;
  BOOLEAN retval;
  LONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer = NULL;
  ULONG threadId;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 68 ||
      origFastIoTable->MdlRead == NULL)
    return FALSE;

  // If there is a file name, record this call

  if (FileObject->FileName.Buffer != NULL) {
    // Indicate that the current thread is active

    threadId = (ULONG) PsGetCurrentThreadId();
    if (threadId < sharedState->kernelUserSharedState.maxThreads)
      sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] =
        '\0';

    // Record information for the file if we haven't already

    NameFileIfUnopened(FileObject, origDevice, TRUE);

    // Log this call, describing the input parameters

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_READ, 24);
    if (eventBuffer) {
      // Get a new sequence number to keep track of this request
      seq = InterlockedIncrement(&sequenceNumber);
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &FileObject, 4);
      RtlCopyMemory(&eventBuffer[9], FileOffset, 5);
      RtlCopyMemory(&eventBuffer[14], &Length, 4);
      RtlZeroMemory(&eventBuffer[18], 4); // no IRP flags, of course
      eventBuffer[22] = IRP_MN_MDL;
      eventBuffer[23] = '\0';
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Call the real fast I/O routine, recording the return value

  retval = origFastIoTable->MdlRead(FileObject, FileOffset, Length, LockKey,
                                    MdlChain, IoStatus, origDevice);

  // Log the return, including return values

  if (eventBuffer) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_MDL_OPERATION, 18);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &IoStatus->Status, 4);
      RtlCopyMemory(&eventBuffer[9], &IoStatus->Information, 4);
      if (retval)
        eventBuffer[13] = FLAG_COMPLETED_SUCCESSFULLY | FLAG_FAST_IO;
      else
        eventBuffer[13] = FLAG_FAST_IO;
      if (MdlChain != NULL)
        RtlCopyMemory(&eventBuffer[14], MdlChain, 4);
      else
        RtlZeroMemory(&eventBuffer[14], 4);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Return the original routine's return value

  return retval;
}

// VTrcFSMdlReadComplete

BOOLEAN VTrcFSMdlReadComplete (IN PFILE_OBJECT FileObject, IN PMDL MdlChain,
                               IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;
  BOOLEAN retval;
  LONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer = NULL;
  ULONG threadId;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 72 ||
      origFastIoTable->MdlReadComplete == NULL)
    return FALSE;

  // If there is a file name, record this call

  if (FileObject->FileName.Buffer != NULL) {
    // Indicate that the current thread is active

    threadId = (ULONG) PsGetCurrentThreadId();
    if (threadId < sharedState->kernelUserSharedState.maxThreads)
      sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] =
        '\0';

    // Record information for the file if we haven't already

    NameFileIfUnopened(FileObject, origDevice, TRUE);

    // Log this call, describing the input parameters

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_READ, 24);
    if (eventBuffer) {
      // Get a new sequence number to keep track of this request
      seq = InterlockedIncrement(&sequenceNumber);
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &FileObject, 4);
      RtlZeroMemory(&eventBuffer[9], 13);
      eventBuffer[22] = IRP_MN_COMPLETE_MDL;
      eventBuffer[23] = '\0';
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Call the original driver's fast I/O routine

  retval = origFastIoTable->MdlReadComplete(FileObject, MdlChain, origDevice);

  // Log the return, including return values

  if (eventBuffer) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_MDL_OPERATION, 18);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlZeroMemory(&eventBuffer[5], 8);
      if (retval)
        eventBuffer[13] = FLAG_COMPLETED_SUCCESSFULLY | FLAG_FAST_IO;
      else
        eventBuffer[13] = FLAG_FAST_IO;
      RtlCopyMemory(&eventBuffer[14], &MdlChain, 4);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Return the original routine's return value

  return retval;
}

// VTrcFSFastIoPrepareMdlWrite

BOOLEAN VTrcFSPrepareMdlWrite (IN PFILE_OBJECT FileObject,
                               IN PLARGE_INTEGER FileOffset, IN ULONG Length,
                               IN ULONG LockKey, OUT PMDL *MdlChain,
                               OUT PIO_STATUS_BLOCK IoStatus,
                               IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;
  BOOLEAN retval;
  LONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer = NULL;
  ULONG threadId;

  // For some reason, NtFilemon had the following code, and we figure it can't
  // hurt

  IoStatus->Status = STATUS_NOT_SUPPORTED;
  IoStatus->Information = 0;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 76 ||
      origFastIoTable->PrepareMdlWrite == NULL)
    return FALSE;

  // If there is a file name, record this call

  if (FileObject->FileName.Buffer != NULL) {
    // Indicate that the current thread is active

    threadId = (ULONG) PsGetCurrentThreadId();
    if (threadId < sharedState->kernelUserSharedState.maxThreads)
      sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] =
        '\0';

    // Record information for the file if we haven't already

    NameFileIfUnopened(FileObject, origDevice, TRUE);

    // Log this call, describing the input parameters

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_WRITE, 24);
    if (eventBuffer) {
      // Get a new sequence number to keep track of this request
      seq = InterlockedIncrement(&sequenceNumber);
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &FileObject, 4);
      RtlCopyMemory(&eventBuffer[9], FileOffset, 5);
      RtlCopyMemory(&eventBuffer[14], &Length, 4);
      RtlZeroMemory(&eventBuffer[18], 4);
      eventBuffer[22] = IRP_MN_MDL;
      eventBuffer[23] = '\0';
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Call the original driver's fast I/O routine

  retval = origFastIoTable->PrepareMdlWrite(FileObject, FileOffset, Length,
                                            LockKey, MdlChain, IoStatus,
                                            origDevice);

  // Log the return, including return values

  if (eventBuffer) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_MDL_OPERATION, 18);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &IoStatus->Status, 4);
      RtlCopyMemory(&eventBuffer[9], &IoStatus->Information, 4);
      if (retval)
        eventBuffer[13] = FLAG_COMPLETED_SUCCESSFULLY | FLAG_FAST_IO;
      else
        eventBuffer[13] = FLAG_FAST_IO;
      if (MdlChain != NULL)
        RtlCopyMemory(&eventBuffer[14], MdlChain, 4);
      else
        RtlZeroMemory(&eventBuffer[14], 4);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Return the original routine's return value

  return retval;
}

// VTrcFSMdlWriteComplete

BOOLEAN VTrcFSMdlWriteComplete (IN PFILE_OBJECT FileObject,
                                IN PLARGE_INTEGER FileOffset,
                                IN PMDL MdlChain,
                                IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;
  BOOLEAN retval;
  LONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer = NULL;
  ULONG threadId;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 80 ||
      origFastIoTable->MdlWriteComplete == NULL)
    return FALSE;

  // If there is a file name, record this call

  if (FileObject->FileName.Buffer != NULL) {
    // Indicate that the current thread is active

    threadId = (ULONG) PsGetCurrentThreadId();
    if (threadId < sharedState->kernelUserSharedState.maxThreads)
      sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] =
        '\0';

    // Record information for the file if we haven't already

    NameFileIfUnopened(FileObject, origDevice, TRUE);

    // Log this call, describing the input parameters

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_WRITE, 24);
    if (eventBuffer) {
      // Get a new sequence number to keep track of this request
      seq = InterlockedIncrement(&sequenceNumber);
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &FileObject, 4);
      if (FileOffset != NULL)
        RtlCopyMemory(&eventBuffer[9], FileOffset, 5);
      else
        RtlZeroMemory(&eventBuffer[9], 5);
      RtlZeroMemory(&eventBuffer[14], 8);
      eventBuffer[22] = IRP_MN_COMPLETE_MDL;
      eventBuffer[23] = '\0';
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Call the original driver's fast I/O routine

  retval = origFastIoTable->MdlWriteComplete(FileObject, FileOffset, MdlChain,
                                             origDevice);

  // Log the return, including return values

  if (eventBuffer) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_MDL_OPERATION, 18);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlZeroMemory(&eventBuffer[5], 8);
      if (retval)
        eventBuffer[13] = FLAG_COMPLETED_SUCCESSFULLY | FLAG_FAST_IO;
      else
        eventBuffer[13] = FLAG_FAST_IO;
      RtlCopyMemory(&eventBuffer[14], &MdlChain, 4);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Return the original routine's return value

  return retval;
}

// VTrcFSFastIoReadCompressed traces incoming parameters and return values.

BOOLEAN VTrcFSFastIoReadCompressed
          (IN PFILE_OBJECT FileObject, IN PLARGE_INTEGER FileOffset,
           IN ULONG Length, IN ULONG LockKey, OUT PVOID Buffer,
           OUT PMDL *MdlChain, OUT PIO_STATUS_BLOCK IoStatus,
           OUT struct _COMPRESSED_DATA_INFO *CompressedDataInfo,
           IN ULONG CompressedDataInfoLength, IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;
  BOOLEAN retval;
  LONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer = NULL;
  ULONG threadId;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 84 ||
      origFastIoTable->FastIoReadCompressed == NULL)
    return FALSE;

  // If there is a file name, record this call

  if (FileObject->FileName.Buffer != NULL) {
    // Indicate that the current thread is active

    threadId = (ULONG) PsGetCurrentThreadId();
    if (threadId < sharedState->kernelUserSharedState.maxThreads)
      sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] =
        '\0';

    // Record information for the file if we haven't already

    NameFileIfUnopened(FileObject, origDevice, TRUE);

    // Log this call, describing the input parameters

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_READ, 24);
    if (eventBuffer) {
      // Get a new sequence number to keep track of this request
      seq = InterlockedIncrement(&sequenceNumber);
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &FileObject, 4);
      RtlCopyMemory(&eventBuffer[9], FileOffset, 5);
      RtlCopyMemory(&eventBuffer[14], &Length, 4);
      RtlZeroMemory(&eventBuffer[18], 4); // no IRP flags, of course
      eventBuffer[22] = MdlChain ? (IRP_MN_COMPRESSED | IRP_MN_MDL)
                                 : IRP_MN_COMPRESSED;
      eventBuffer[23] = '\0';
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Call the real fast I/O routine, recording the return value

  retval = origFastIoTable->FastIoReadCompressed(FileObject, FileOffset,
                                                 Length, LockKey, Buffer,
                                                 MdlChain, IoStatus,
                                                 CompressedDataInfo,
                                                 CompressedDataInfoLength,
                                                 origDevice);

  // Log the return, including return values

  if (eventBuffer) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    if (MdlChain) {
      eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_MDL_OPERATION,
                              18);
      if (eventBuffer) {
        RtlCopyMemory(&eventBuffer[14], MdlChain, 4);
      }
    }
    else {
      eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_OPERATION, 14);
    }
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &IoStatus->Status, 4);
      RtlCopyMemory(&eventBuffer[9], &IoStatus->Information, 4);
      if (retval)
        eventBuffer[13] = FLAG_COMPLETED_SUCCESSFULLY | FLAG_FAST_IO;
      else
        eventBuffer[13] = FLAG_FAST_IO;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Return the original routine's return value

  return retval;
}

// VTrcFSFastIoWriteCompressed traces incoming parameters and return values.

BOOLEAN VTrcFSFastIoWriteCompressed
          (IN PFILE_OBJECT FileObject, IN PLARGE_INTEGER FileOffset,
           IN ULONG Length, IN ULONG LockKey, IN PVOID Buffer,
           OUT PMDL *MdlChain, OUT PIO_STATUS_BLOCK IoStatus,
           IN struct _COMPRESSED_DATA_INFO *CompressedDataInfo,
           IN ULONG CompressedDataInfoLength, IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;
  BOOLEAN retval;
  LONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer = NULL;
  ULONG threadId;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 88 ||
      origFastIoTable->FastIoWriteCompressed == NULL)
    return FALSE;

  // If there is a file name, record this call

  if (FileObject->FileName.Buffer != NULL) {
    // Indicate that the current thread is active

    threadId = (ULONG) PsGetCurrentThreadId();
    if (threadId < sharedState->kernelUserSharedState.maxThreads)
      sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] =
        '\0';

    // Record information for the file if we haven't already

    NameFileIfUnopened(FileObject, origDevice, TRUE);

    // Log this call, describing the input parameters

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_WRITE, 24);
    if (eventBuffer) {
      // Get a new sequence number to keep track of this request
      seq = InterlockedIncrement(&sequenceNumber);
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &FileObject, 4);
      RtlCopyMemory(&eventBuffer[9], FileOffset, 5);
      RtlCopyMemory(&eventBuffer[14], &Length, 4);
      RtlZeroMemory(&eventBuffer[18], 4); // no IRP flags, of course
      eventBuffer[22] = MdlChain ? (IRP_MN_COMPRESSED | IRP_MN_MDL)
                                 : IRP_MN_COMPRESSED;
      eventBuffer[23] = '\0';
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Call the real fast I/O routine, recording the return value

  retval = origFastIoTable->FastIoWriteCompressed(FileObject, FileOffset,
                                                  Length, LockKey, Buffer,
                                                  MdlChain, IoStatus,
                                                  CompressedDataInfo,
                                                  CompressedDataInfoLength,
                                                  origDevice);

  // Log the return, including return values

  if (eventBuffer) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    if (MdlChain) {
      eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_MDL_OPERATION,
                              18);
      if (eventBuffer) {
        RtlCopyMemory(&eventBuffer[14], MdlChain, 4);
      }
    }
    else {
      eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_OPERATION, 14);
    }
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &IoStatus->Status, 4);
      RtlCopyMemory(&eventBuffer[9], &IoStatus->Information, 4);
      if (retval)
        eventBuffer[13] = FLAG_COMPLETED_SUCCESSFULLY | FLAG_FAST_IO;
      else
        eventBuffer[13] = FLAG_FAST_IO;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Return the original routine's return value

  return retval;
}

// VTrcFSMdlReadCompleteCompressed

BOOLEAN VTrcFSMdlReadCompleteCompressed (IN PFILE_OBJECT FileObject,
                                         IN PMDL MdlChain,
                                         IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;
  BOOLEAN retval;
  LONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer = NULL;
  ULONG threadId;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 92 ||
      origFastIoTable->MdlReadCompleteCompressed == NULL)
    return FALSE;

  // If there is a file name, record this call

  if (FileObject->FileName.Buffer != NULL) {
    // Indicate that the current thread is active

    threadId = (ULONG) PsGetCurrentThreadId();
    if (threadId < sharedState->kernelUserSharedState.maxThreads)
      sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] =
        '\0';

    // Record information for the file if we haven't already

    NameFileIfUnopened(FileObject, origDevice, TRUE);

    // Log this call, describing the input parameters

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_READ, 24);
    if (eventBuffer) {
      // Get a new sequence number to keep track of this request
      seq = InterlockedIncrement(&sequenceNumber);
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &FileObject, 4);
      RtlZeroMemory(&eventBuffer[9], 13);
      eventBuffer[22] = MdlChain ? (IRP_MN_COMPRESSED | IRP_MN_COMPLETE_MDL) :
                                   (IRP_MN_COMPRESSED | IRP_MN_COMPLETE);
      eventBuffer[23] = '\0';
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Call the original driver's fast I/O routine

  retval = origFastIoTable->MdlReadCompleteCompressed(FileObject, MdlChain,
                                                      origDevice);

  // Log the return, including return values

  if (eventBuffer) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    if (MdlChain) {
      eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_MDL_OPERATION,
                              18);
      if (eventBuffer) {
        RtlCopyMemory(&eventBuffer[14], &MdlChain, 4);
      }
    }
    else {
      eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_OPERATION, 14);
    }
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlZeroMemory(&eventBuffer[5], 8);
      if (retval)
        eventBuffer[13] = FLAG_COMPLETED_SUCCESSFULLY | FLAG_FAST_IO;
      else
        eventBuffer[13] = FLAG_FAST_IO;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Return the original routine's return value

  return retval;
}

// VTrcFSMdlWriteCompleteCompressed

BOOLEAN VTrcFSMdlWriteCompleteCompressed (IN PFILE_OBJECT FileObject,
                                          IN PLARGE_INTEGER FileOffset,
                                          IN PMDL MdlChain,
                                          IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;
  BOOLEAN retval;
  LONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer = NULL;
  ULONG threadId;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 96 ||
      origFastIoTable->MdlWriteCompleteCompressed == NULL)
    return FALSE;

  // If there is a file name, record this call

  if (FileObject->FileName.Buffer != NULL) {
    // Indicate that the current thread is active

    threadId = (ULONG) PsGetCurrentThreadId();
    if (threadId < sharedState->kernelUserSharedState.maxThreads)
      sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] =
        '\0';

    // Record information for the file if we haven't already

    NameFileIfUnopened(FileObject, origDevice, TRUE);

    // Log this call, describing the input parameters

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_WRITE, 24);
    if (eventBuffer) {
      // Get a new sequence number to keep track of this request
      seq = InterlockedIncrement(&sequenceNumber);
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &FileObject, 4);
      if (FileOffset != NULL)
        RtlCopyMemory(&eventBuffer[9], FileOffset, 5);
      else
        RtlZeroMemory(&eventBuffer[9], 5);
      RtlZeroMemory(&eventBuffer[14], 8);
      eventBuffer[22] = MdlChain ? (IRP_MN_COMPRESSED | IRP_MN_COMPLETE_MDL) :
                                   (IRP_MN_COMPRESSED | IRP_MN_COMPLETE);
      eventBuffer[23] = '\0';
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Call the original driver's fast I/O routine

  retval = origFastIoTable->MdlWriteCompleteCompressed(FileObject, FileOffset,
                                                       MdlChain, origDevice);

  // Log the return, including return values

  if (eventBuffer) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    if (MdlChain) {
      eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_MDL_OPERATION,
                              18);
      if (eventBuffer) {
        RtlCopyMemory(&eventBuffer[14], &MdlChain, 4);
      }
    }
    else {
      eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_OPERATION, 14);
    }
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlZeroMemory(&eventBuffer[5], 8);
      if (retval)
        eventBuffer[13] = FLAG_COMPLETED_SUCCESSFULLY | FLAG_FAST_IO;
      else
        eventBuffer[13] = FLAG_FAST_IO;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Return the original routine's return value

  return retval;
}

// VTrcFSFastIoQueryOpen (This call actually passes an IRP!)

BOOLEAN VTrcFSFastIoQueryOpen
          (IN struct _IRP *Irp,
           OUT PFILE_NETWORK_OPEN_INFORMATION NetworkInformation,
           IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;
  BOOLEAN retval;
  PIO_STACK_LOCATION currentIrpStack;
  PIO_STACK_LOCATION nextIrpStack;
  LONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer = NULL;
  PFILE_OBJECT fileObject;
  ULONG threadId;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 100 ||
      origFastIoTable->FastIoQueryOpen == NULL)
    return FALSE;

  // Get the Irp stack location pointers

  currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
  nextIrpStack    = IoGetNextIrpStackLocation(Irp);
  fileObject      = currentIrpStack->FileObject;

  // If there is a file name, record this call

  if (fileObject->FileName.Buffer != NULL) {
    // Indicate that the current thread is active

    threadId = (ULONG) PsGetCurrentThreadId();
    if (threadId < sharedState->kernelUserSharedState.maxThreads)
      sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] =
        '\0';

    // Remove this file number from the hash since if it was open before, it's
    // different now.

    RemoveFileObjectFromHash(fileObject);

    // Log this call, describing the input parameters

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_OPEN_QUERY_CLOSE, 13);
    if (eventBuffer) {
      // Get a new sequence number to keep track of this request
      seq = InterlockedIncrement(&sequenceNumber);
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &fileObject, 4);
      RtlCopyMemory(&eventBuffer[9], &origDevice, 4);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Copy parameters down to next level in the stack, but don't copy
  // completion routine.

  *nextIrpStack = *currentIrpStack;
  nextIrpStack->DeviceObject = origDevice;
  IoSetCompletionRoutine(Irp, NULL, NULL, FALSE, FALSE, FALSE);
  IoSetNextIrpStackLocation(Irp);

  // Call the real fast I/O routine, recording the return value

  retval = origFastIoTable->FastIoQueryOpen(Irp, NetworkInformation,
                                            origDevice);

  if (eventBuffer) {
    char name[249];
    ULONG fileNameLength;

    // Get the name of the file, and the length of that name, before we
    // acquire the spin lock, since this is a potentially long operation.

    fileNameLength = GetFullPath(fileObject, name, 249);

    // Log the return, including return values

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_OPERATION, 14);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &Irp->IoStatus.Status, 4);
      RtlCopyMemory(&eventBuffer[9], &Irp->IoStatus.Information, 4);
      if (!retval || Irp->IoStatus.Status != STATUS_SUCCESS)
        eventBuffer[13] = FLAG_FAST_IO;
      else {
        eventBuffer[13] = FLAG_COMPLETED_SUCCESSFULLY | FLAG_FAST_IO;
        eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_INFORMATION, 73);
        if (eventBuffer) {
          RtlCopyMemory(&eventBuffer[1], &fileObject, 4);
          RtlCopyMemory(&eventBuffer[5], &origDevice, 4);
          RtlCopyMemory(&eventBuffer[9], &fileObject->Flags, 4);
          RtlCopyMemory(&eventBuffer[13], NetworkInformation, 32);
          RtlCopyMemory(&eventBuffer[45],
                        &NetworkInformation->FileAttributes, 4);
          RtlCopyMemory(&eventBuffer[49],
                        &NetworkInformation->AllocationSize, 5);
          RtlCopyMemory(&eventBuffer[54],
                        &NetworkInformation->EndOfFile, 5);
          RtlZeroMemory(&eventBuffer[59], 9);
          RtlCopyMemory(&eventBuffer[68], &fileObject->FsContext, 4);
          eventBuffer[72] = GetCompressedAccessMask(fileObject);

          eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_NAME,
                                  fileNameLength + 6);
          if (eventBuffer) {
            eventBuffer[1] = (char) (fileNameLength + 6);
            RtlCopyMemory(&eventBuffer[2], &fileObject, 4);
            RtlCopyMemory(&eventBuffer[6], name, fileNameLength);
          }
        }
      }
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Return the original routine's return value

  return retval;
}

// VTrcFSReleaseForModWrite

NTSTATUS VTrcFSReleaseForModWrite (IN PFILE_OBJECT FileObject,
                                   IN struct _ERESOURCE *ResourceToRelease,
                                   IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return false

  if (origFastIoTable->SizeOfFastIoDispatch < 104 ||
      origFastIoTable->ReleaseForModWrite == NULL)
    return FALSE;

  // Call the original driver's fast I/O routine

  return origFastIoTable->ReleaseForModWrite(FileObject, ResourceToRelease,
                                             origDevice);
}

// VTrcFSAcquireForCcFlush

NTSTATUS VTrcFSAcquireForCcFlush (IN PFILE_OBJECT FileObject,
                                  IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return STATUS_NOT_SUPPORTED

  if (origFastIoTable->SizeOfFastIoDispatch < 108 ||
      origFastIoTable->AcquireForCcFlush == NULL)
    return STATUS_NOT_SUPPORTED;

  // Call the original driver's fast I/O routine

  return origFastIoTable->AcquireForCcFlush(FileObject, origDevice);
}

// VTrcFSFastIoReleaseForCcFlush

NTSTATUS VTrcFSReleaseForCcFlush (IN PFILE_OBJECT FileObject,
                                  IN PDEVICE_OBJECT DeviceObject)
{
  PFS_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PDEVICE_OBJECT origDevice = hookExt->attachedDevice;
  PFAST_IO_DISPATCH origFastIoTable = origDevice->DriverObject->FastIoDispatch;

  // If this routine doesn't exist in the original driver's fast I/O dispatch
  // table, return STATUS_NOT_SUPPORTED

  if (origFastIoTable->SizeOfFastIoDispatch < 112 ||
      origFastIoTable->ReleaseForCcFlush == NULL)
    return STATUS_NOT_SUPPORTED;

  // Call the original driver's fast I/O routine

  return origFastIoTable->ReleaseForCcFlush(FileObject, origDevice);
}
