//
// FileName.c
//
// Code for routines having to do with getting file names for VTrcFS.
//
// (Uses some code by Mark Russinovich and Bryce Cogswell)
//

#include "NTDDK.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "VTrcFS.h"
#include "GetState.h"
#include "Hash.h"

///////////////////////////////////////////////////////////////////////////
//                              MACROS
///////////////////////////////////////////////////////////////////////////

// Print macro that only turns on when checked builds are on

#if DBG
#define DebugPrint(arg) DbgPrint arg
#else
#define DebugPrint(arg)
#endif

///////////////////////////////////////////////////////////////////////////
//                              STRUCTURES
///////////////////////////////////////////////////////////////////////////

// These are undocumented structures (found in Rajeev Nagar's book) that
// are returned by query-info IRP's.

typedef struct _FILE_INTERNAL_INFORMATION {
  LARGE_INTEGER IndexNumber;
} FILE_INTERNAL_INFORMATION;

typedef struct _FILE_EA_INFORMATION {
  ULONG EaSize;
} FILE_EA_INFORMATION;

typedef struct _FILE_ACCESS_INFORMATION {
  ACCESS_MASK AccessFlags;
} FILE_ACCESS_INFORMATION;

typedef struct _FILE_MODE_INFORMATION {
  ULONG Mode;
} FILE_MODE_INFORMATION;

typedef struct _EXPANDED_FILE_NAME_INFORMATION {
  ULONG FileNameLength;
  WCHAR FileName[300];
} EXPANDED_FILE_NAME_INFORMATION;

typedef struct _FILE_ALL_INFORMATION {
  FILE_BASIC_INFORMATION BasicInformation;
  FILE_STANDARD_INFORMATION StandardInformation;
  FILE_INTERNAL_INFORMATION InternalInformation;
  FILE_EA_INFORMATION EaInformation;
  FILE_ACCESS_INFORMATION AccessInformation;
  FILE_POSITION_INFORMATION PositionInformation;
  FILE_MODE_INFORMATION ModeInformation;
  FILE_ALIGNMENT_INFORMATION AlignmentInformation;
  EXPANDED_FILE_NAME_INFORMATION NameInformation;
} FILE_ALL_INFORMATION;

typedef struct _query_info_context {
  PFILE_OBJECT fileObject;
  PDEVICE_OBJECT deviceObject;
  KEVENT event;
  FILE_ALL_INFORMATION fileInfo;
  BOOLEAN informationLoggedSuccessfully;
} QUERY_INFO_CONTEXT;

///////////////////////////////////////////////////////////////////////////
//                   GETTTING FILE INFO WITHOUT QUERY
///////////////////////////////////////////////////////////////////////////

// GetFullPath takes a fileObject and returns a canonical path,
// nicely formatted, in fullPathName.  No more than maxPathLen characters
// are copied; if more would be required, the first character is * and
// as many of the final characters are copied as possible.  So, for
// example, if the full path were C:\FOO\BAR\ReallyLongFileName, and
// maxPathLen were only 22, we would copy *AR\ReallyLongFileName to
// fullPathName.  (We cut off beginning characters rather than trailing
// ones because the most important part of the name is the leaf name.)
// The function returns the length of the file name.  No trailing null
// character is put on the file name.

LONG GetFullPath (PFILE_OBJECT fileObject, PCHAR fullPathName, LONG maxPathLen)
{
  LONG                pathLen, i, remainingCharsToCopy;
  PFILE_OBJECT        relatedFileObject;
  UNICODE_STRING      *fileName;
  BOOLEAN             overflow = FALSE;   // is path name too long?

  // Figure out how long the file name really is before copying any
  // characters, putting the file name length in pathLen.  This is
  // accomplished by looping through related file objects, i.e. going
  // up levels all the way to the root.

  pathLen = 0;
  relatedFileObject = fileObject;
  while (relatedFileObject != NULL) {
    fileName = &relatedFileObject->FileName;
    if (fileName->Length < 2 || !MmIsAddressValid(fileName->Buffer))
      break;
    pathLen += (fileName->Length / 2);
    if (fileObject != relatedFileObject &&
        fileName->Buffer[fileName->Length / 2 - 1] != L'\\')
      pathLen++;                     // account for trailing backslash
    if (pathLen > maxPathLen) {
      pathLen = maxPathLen;
      overflow = TRUE;
      break;
    }
    if (fileName->Buffer[0] == L'\\')
      break;
    relatedFileObject = relatedFileObject->RelatedFileObject;
  }

  // Copy pathLen characters from the file name to fullPathName,
  // starting at the end and working toward the beginning.  Note
  // that we're copying Unicode characters, so we assume they have
  // ASCII equivalents so that each WCHAR consists of one ASCII
  // character followed by an unneeded character which we don't copy.
  // Nothing too bad happens if the characters don't correspond to
  // ASCII characters; we just get a garbled file name.

  remainingCharsToCopy = pathLen;
  relatedFileObject = fileObject;
  while (remainingCharsToCopy > 0) {
    fileName = &relatedFileObject->FileName;
    if (fileObject != relatedFileObject &&
      fileName->Buffer[fileName->Length / 2 - 1] != L'\\') {
      fullPathName[--remainingCharsToCopy] = '\\';
    }
    for (i = fileName->Length / 2 - 1; i >= 0 && remainingCharsToCopy > 0; i--)
      fullPathName[--remainingCharsToCopy] = * (char *) &fileName->Buffer[i];
    relatedFileObject = relatedFileObject->RelatedFileObject;
  }

  // If we couldn't copy the whole name, put a '*' as the first
  // character.

  if (overflow) fullPathName[0] = '*';

  return pathLen;
}

// GetCompressedAccessMask returns a single character representing
// the compressed access mask deduced from information in the given
// file object structure, fileObject.

char GetCompressedAccessMask (PFILE_OBJECT fileObject)
{
  char compressedAccessMask = '\0';

  if (fileObject->ReadAccess)
    compressedAccessMask |= COMPRESSED_ACCESS_MASK_READ;
  if (fileObject->WriteAccess)
    compressedAccessMask |= COMPRESSED_ACCESS_MASK_WRITE;
  if (fileObject->DeleteAccess)
    compressedAccessMask |= COMPRESSED_ACCESS_MASK_DELETE;

  return compressedAccessMask;
}

// RecordFileInformationWithoutQuery records information about a given file
// object, with given associated device object, in the case that a query
// either cannot be done or failed. It does this by just examining fields
// in the file object structure. It returns a boolean indicating whether
// it actually put file information in the log.

BOOLEAN RecordFileInformationWithoutQuery (PFILE_OBJECT fileObject,
                                           PDEVICE_OBJECT deviceObject)
{
  KIRQL oldirql;
  PCHAR eventBuffer;
  char name[249];
  LONG nameLength;

  // First, obtain the name directly from the buffer pointed to by the
  // file object. Do this before acquiring the spin lock because it takes
  // a while.

  nameLength = GetFullPath(fileObject, name, 249);

  // Then, acquire the spin lock and generate the log entries. Use zeroes
  // for any information we don't know.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_INFORMATION, 73);
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[1], &fileObject, 4);
    RtlCopyMemory(&eventBuffer[5], &deviceObject, 4);
    RtlCopyMemory(&eventBuffer[9], &fileObject->Flags, 4);
    RtlZeroMemory(&eventBuffer[13], 55);
    RtlCopyMemory(&eventBuffer[68], &fileObject->FsContext, 4);
    eventBuffer[72] = GetCompressedAccessMask(fileObject);

    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_NAME, nameLength + 6);
    if (eventBuffer) {
      eventBuffer[1] = (char) (nameLength + 6);
      RtlCopyMemory(&eventBuffer[2], &fileObject, 4);
      RtlCopyMemory(&eventBuffer[6], name, nameLength);
    }
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Return true only if we successfully generated the log entries.

  return (eventBuffer != NULL);
}

///////////////////////////////////////////////////////////////////////////
//                         QUERY INFORMATION FUNCTIONS
///////////////////////////////////////////////////////////////////////////

// QueryInfoCompletionRoutine is the completion routine for our query-info
// IRP.  Its main responsibilities are to signal the event in the IRP and
// free the IRP.

NTSTATUS QueryInfoCompletionRoutine (PDEVICE_OBJECT DeviceObject,
                                     PIRP Irp, PVOID Context)
{
  QUERY_INFO_CONTEXT *queryInfoContext = (QUERY_INFO_CONTEXT *) Context;
  PFILE_OBJECT fileObject = queryInfoContext->fileObject;
  FILE_ALL_INFORMATION *fileInfo = &queryInfoContext->fileInfo;
  ULONG fileNameLength, i;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Status is returned in ioStatusBlock, so check it to see how the request
  // went

  if (Irp->IoStatus.Status == STATUS_SUCCESS) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_INFORMATION, 73);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &fileObject, 4);
      RtlCopyMemory(&eventBuffer[5], &queryInfoContext->deviceObject, 4);
      RtlCopyMemory(&eventBuffer[9], &fileObject->Flags, 4);
      RtlCopyMemory(&eventBuffer[13], &fileInfo->BasicInformation, 36);
      RtlCopyMemory(&eventBuffer[49],
                    &fileInfo->StandardInformation.AllocationSize, 5);
      RtlCopyMemory(&eventBuffer[54], &fileInfo->StandardInformation.EndOfFile,
                    5);
      eventBuffer[59] =
                 (unsigned char) fileInfo->StandardInformation.NumberOfLinks;
      RtlCopyMemory(&eventBuffer[60], &fileInfo->InternalInformation, 8);
      RtlCopyMemory(&eventBuffer[68], &fileObject->FsContext, 4);
      eventBuffer[72] = GetCompressedAccessMask(fileObject);

      fileNameLength = fileInfo->NameInformation.FileNameLength / 2;
      if (fileNameLength > 249) fileNameLength = 249;
      eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_NAME, fileNameLength + 6);
      if (eventBuffer) {
        eventBuffer[1] = (char) (fileNameLength + 6);
        RtlCopyMemory(&eventBuffer[2], &fileObject, 4);
        for (i = 0; i < fileNameLength; i++)
          eventBuffer[i+6] =
            * (char *) &fileInfo->NameInformation.FileName[i];
        queryInfoContext->informationLoggedSuccessfully = TRUE;
      }
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // As in (almost) any completion routine, we have to mark Irp as pending if
  // necessary.  Note the braces around the call to IoMarkIrpPending --- they
  // are important because IoMarkIrpPending is a macro that expands to multiple
  // statements.

  if (Irp->PendingReturned) {
    IoMarkIrpPending(Irp);
  }

  // Set event so that the routine waiting on this IRP can continue

  KeSetEvent(&queryInfoContext->event, 0, FALSE);

  // Free the IRP

  IoFreeIrp(Irp);

  // Return STATUS_MORE_PROCESSING_REQUIRED to indicate that this driver
  // is taking control of the IRP.  This is necessary because the IRP has
  // just been freed, so we don't want anyone else trying to work on it.
  // Indeed, we have just finished processing the IRP by freeing it.

  return STATUS_MORE_PROCESSING_REQUIRED;
}

// RecordFileInformation records information about a given file object, with
// given associated device object.  It does this by passing a query-info
// IRP down to that device object and entering the resulting information
// in the log.  If this query fails or cannot be done, it tries calling
// RecordFileInformationWithoutQuery.  It returns a boolean indicating if
// information was recorded for this file.

BOOLEAN RecordFileInformation (PFILE_OBJECT fileObject,
                               PDEVICE_OBJECT deviceObject)
{
  PIRP irp;
  PIO_STACK_LOCATION irpStack;
  QUERY_INFO_CONTEXT queryInfoContext;

  // Allocate an IRP for the device request. If an IRP cannot be allocated,
  // we call RecordFileInformationWithoutQuery.

  irp = IoAllocateIrp(deviceObject->StackSize, FALSE);
  if (irp == NULL)
    return RecordFileInformationWithoutQuery(fileObject, deviceObject);

  ///////////////////////////////////////
  // Initialize the fields in the IRP. //
  ///////////////////////////////////////

  irp->AssociatedIrp.SystemBuffer = &queryInfoContext.fileInfo;

  // If the synchronous flag is not set, a bug(?) in FASTFAT can cause the
  // completion routine to be called when the status is still STATUS_PENDING,
  // ultimately causing a blue-screen

  irp->Flags = IRP_SYNCHRONOUS_API | IRP_OB_QUERY_NAME;

  // Set other miscellaneous IRP info

  irp->Tail.Overlay.Thread = PsGetCurrentThread();
  irp->Tail.Overlay.OriginalFileObject = fileObject;
  irp->RequestorMode = KernelMode;

  // Set information in the first stack location indicating what type
  // of request this is, and other info

  irpStack = IoGetNextIrpStackLocation(irp);
  irpStack->MajorFunction = IRP_MJ_QUERY_INFORMATION;
  irpStack->MinorFunction = 0;
  irpStack->Parameters.QueryFile.FileInformationClass = FileAllInformation;
  irpStack->Parameters.QueryFile.Length = sizeof(queryInfoContext.fileInfo);
  irpStack->DeviceObject = deviceObject;
  irpStack->FileObject = fileObject;

  // Initialize an event so that we can know when the IRP is completed.

  KeInitializeEvent(&queryInfoContext.event, SynchronizationEvent, FALSE);

  // Initialize the "was information logged successfully" field of the
  // context so that we will be able to tell if information was actually
  // recorded in the log by the completion routine.

  queryInfoContext.informationLoggedSuccessfully = FALSE;

  // Set our completion routine so that we're called when the IRP completes

  queryInfoContext.fileObject = fileObject;
  queryInfoContext.deviceObject = deviceObject;
  IoSetCompletionRoutine(irp, QueryInfoCompletionRoutine, &queryInfoContext,
                         TRUE, TRUE, TRUE);

  // Pass on the IRP.

  IoCallDriver(deviceObject, irp);

  // Wait for the event set by the completion routine so that the query
  // info context does not go away prematurely.

  while (KeWaitForSingleObject(&queryInfoContext.event, Executive,
                               KernelMode, TRUE, NULL) != STATUS_SUCCESS)
    ;

  // If we were successful, return TRUE, otherwise try calling
  // RecordFileInformationWithoutQuery.

  if (queryInfoContext.informationLoggedSuccessfully)
    return TRUE;
  else
    return RecordFileInformationWithoutQuery(fileObject, deviceObject);
}

// RecordInformationForFileWithCommonFcb records information about a
// given file object, with given associated device object, given that
// an open file with the same FCB was already recorded in the log. In
// other words, it merely logs the file object's device and FCB, and
// also the access mask for this particular access. It returns a
// boolean indicating if it actually wrote information to the log.

BOOLEAN RecordInformationForFileWithCommonFcb (PFILE_OBJECT fileObject,
                                               PDEVICE_OBJECT deviceObject)
{
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Acquire the spin lock and generate the log entry.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_FCB_INFORMATION, 14);
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[1], &fileObject, 4);
    RtlCopyMemory(&eventBuffer[5], &deviceObject, 4);
    RtlCopyMemory(&eventBuffer[9], &fileObject->FsContext, 4);
    eventBuffer[13] = GetCompressedAccessMask(fileObject);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Return true only if we successfully generated the log entry.

  return (eventBuffer != NULL);
}

// NameFileIfUnopened looks in the hash of opened file objects to see if
// the file object passed is there.  If not, there is no file-open entry
// in this trace naming the file, and this function has to add such an
// entry.  The device object associated with the file object is also passed
// to this routine, since this information is needed in the log entry.  Once
// this function has added the entry, it adds the file number to the hash
// table so that a name entry does not have to be entered into this trace
// again.  If the search for an identical file object finds none, but finds
// a file object referring to the same FCB, we need merely record the FCB
// information about the file.  (However, if the parameter canUseFcb is
// FALSE, we never do this.)

void NameFileIfUnopened (PFILE_OBJECT fileObject, PDEVICE_OBJECT deviceObject,
                         BOOLEAN canUseFcb)
{
  if (sharedState->checkIfLoggingFunc &&
      (*sharedState->checkIfLoggingFunc)(ENTRY_TYPE_FILE_INFORMATION)) {
    if (AddFileObjectToHashIfAbsent(fileObject, &canUseFcb)) {
      if (canUseFcb) {
        if (!RecordInformationForFileWithCommonFcb(fileObject, deviceObject))
          RemoveFileObjectFromHash(fileObject);
      }
      else {
        if (!RecordFileInformation(fileObject, deviceObject))
          RemoveFileObjectFromHash(fileObject);
      }
    }
  }
}
