// 
// Hash.c
//
// Code for hash-table-related routines in the VTrcFS driver.
//

#include "NTDDK.h"
#include "Hash.h"

///////////////////////////////////////////////////////////////////////////
//                               MACROS
///////////////////////////////////////////////////////////////////////////

#define NUM_HASH_BUCKETS         251  /* the number of buckets is prime */
#define ENTRIES_PER_ALLOCATION   128  /* # of entries allocated at a time */

#define GetHashValue(x) ((x) % NUM_HASH_BUCKETS)  /* hash function */

// Print macro that only turns on when checked builds are on

#if DBG
#define DebugPrint(arg) DbgPrint arg
#else
#define DebugPrint(arg)
#endif

///////////////////////////////////////////////////////////////////////////
//                             STRUCTURES
///////////////////////////////////////////////////////////////////////////

// The file object entry structure is used for storing file object numbers
// in the hash table (and associated free list).

typedef struct _file_object_entry {
  PFILE_OBJECT fileObject;
  struct _file_object_entry *next;
  struct _file_object_entry *nextInFcbHash;
} FILE_OBJECT_ENTRY;

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, InitHash)
#endif

///////////////////////////////////////////////////////////////////////////
//                              GLOBALS
///////////////////////////////////////////////////////////////////////////

FILE_OBJECT_ENTRY *hashTable[NUM_HASH_BUCKETS];    // main hash table
FILE_OBJECT_ENTRY *fcbHashTable[NUM_HASH_BUCKETS]; // hash table by FCB
FILE_OBJECT_ENTRY *freeList = NULL;                // linked free list
KSPIN_LOCK hashSpinLock;                           // for accessing hash table

///////////////////////////////////////////////////////////////////////////
//                     FILE OBJECT ENTRY FUNCTIONS
///////////////////////////////////////////////////////////////////////////

// AllocateBlockOfFileObjectEntries allocates a block of file object
// entries and adds them to the free list.  It returns a boolean
// indicating whether or not it succeeded.

BOOLEAN AllocateBlockOfFileObjectEntries (void)
{
  FILE_OBJECT_ENTRY *block;
  int i;

  DebugPrint(("VTrcFS: Allocating block of file object entries.\n"));

  block = (FILE_OBJECT_ENTRY *) ExAllocatePool(NonPagedPool,
                                               sizeof(FILE_OBJECT_ENTRY) *
                                               ENTRIES_PER_ALLOCATION);
  if (block == NULL)
    return FALSE;

  for (i = 0; i < ENTRIES_PER_ALLOCATION - 1; i++)
    block[i].next = &block[i+1];
  block[ENTRIES_PER_ALLOCATION - 1].next = freeList;
  freeList = block;

  return TRUE;
}

// NewFileObjectEntry returns the address of an unused file object entry,
// obtained (and removed) from the free list.  If the free list is empty,
// this function first allocates more entries onto the free list.  This
// function returns NULL if it cannot obtain an entry (i.e. if the
// free list is empty and allocation to add to it fails).

FILE_OBJECT_ENTRY *NewFileObjectEntry (void)
{
  FILE_OBJECT_ENTRY *retVal;

  if (freeList == NULL)
    if (!AllocateBlockOfFileObjectEntries())
      return NULL;

  retVal = freeList;
  freeList = freeList->next;
  return retVal;
}

// FreeFileObjectEntry adds the given entry to the free list.

void FreeFileObjectEntry (FILE_OBJECT_ENTRY *entry)
{
  entry->next = freeList;
  freeList = entry;
}

///////////////////////////////////////////////////////////////////////////
//                              HASH FUNCTIONS
///////////////////////////////////////////////////////////////////////////

// InitHash initializes the hash table and hash table spin lock.

void InitHash (void)
{
  int i;

  KeInitializeSpinLock(&hashSpinLock);
  for (i = 0; i < NUM_HASH_BUCKETS; i++) {
    hashTable[i] = NULL;
    fcbHashTable[i] = NULL;
  }
}

// EmptyHash empties the hash table and puts its contents on the free
// list.  It can be called at general times because it mediates its
// access to the hash table and free list through the spin lock.

void EmptyHash (void)
{
  KIRQL oldirql;
  int i;
  FILE_OBJECT_ENTRY *entry;

  KeAcquireSpinLock(&hashSpinLock, &oldirql);

  // For each bucket, if that bucket is nonempty, find the last entry
  // in the bucket (which points to NULL) and have it point to the
  // free list; then, have the free list point to the first entry in
  // the bucket.  This puts everything in the bucket on the free list,
  // and we can set the bucket to NULL (empty).

  for (i = 0; i < NUM_HASH_BUCKETS; i++) {
    if (hashTable[i] != NULL) {
      for (entry = hashTable[i]; entry->next != NULL; entry = entry->next)
        ;
      entry->next = freeList;
      freeList = hashTable[i];
      hashTable[i] = NULL;
    }
    fcbHashTable[i] = NULL;
  }
  KeReleaseSpinLock(&hashSpinLock, oldirql);
}

// AddEntryToHash adds the given entry to the hash table.

void AddEntryToHash (FILE_OBJECT_ENTRY *entry)
{
  int bucket;
  PVOID fcb;

  bucket = GetHashValue((ULONG) entry->fileObject);
  entry->next = hashTable[bucket];
  hashTable[bucket] = entry;

  fcb = entry->fileObject->FsContext;
  if (fcb) {
    bucket = GetHashValue((ULONG) fcb);
    entry->nextInFcbHash = fcbHashTable[bucket];
    fcbHashTable[bucket] = entry;
  }
}

// RemoveEntryFromFcbHash removes the given entry from the hash indexed by FCB,
// if it is in it.

void RemoveEntryFromFcbHash (FILE_OBJECT_ENTRY *entry)
{
  int bucket;
  FILE_OBJECT_ENTRY *prev;
  PVOID fcb;

  // First, get the FCB pointer from the entry. If there is none, we know it
  // isn't in the hash, so we just return.

  fcb = entry->fileObject->FsContext;
  if (fcb == NULL)
    return;

  // Find which bucket of the FCB hash the entry would be in.

  bucket = GetHashValue((ULONG) fcb);

  // If it's the first entry in that bucket, remove it from the head of the
  // bucket list.

  if (fcbHashTable[bucket] == entry) {
    fcbHashTable[bucket] = entry->nextInFcbHash;
    return;
  }    

  // Otherwise, search the bucket for the entry.

  for (prev = fcbHashTable[bucket]; prev != NULL; prev = prev->nextInFcbHash) {
    if (prev->nextInFcbHash == entry) {
      prev->nextInFcbHash = entry->nextInFcbHash;
      return;
    }
  }
}

// RemoveEntryFromHash removes the given entry from the hash table but
// does NOT add it to the free list.  It returns a boolean indicating
// whether it was successful (i.e. whether it could find the entry in
// the hash table).  We assume that the passed entry is a valid pointer;
// if it were, for instance, NULL, bad things would happen.

BOOLEAN RemoveEntryFromHash (FILE_OBJECT_ENTRY *entry)
{
  int bucket;
  FILE_OBJECT_ENTRY *prev;

  bucket = GetHashValue((ULONG) entry->fileObject);

  if (hashTable[bucket] == entry) {
    hashTable[bucket] = entry->next;
    RemoveEntryFromFcbHash(entry);
    return TRUE;
  }
  for (prev = hashTable[bucket]; prev != NULL; prev = prev->next) {
    if (prev->next == entry) {
      prev->next = entry->next;
      RemoveEntryFromFcbHash(entry);
      return TRUE;
    }
  }
  return FALSE;
}

// FindFcbInHash returns a pointer to a file object entry whose file has the
// same FCB as the given parameter 'searchFcb'. If it cannot find one, it
// returns NULL.

FILE_OBJECT_ENTRY *FindFcbInHash (PVOID searchFcb)
{
  int bucket;
  FILE_OBJECT_ENTRY *entry;

  if (searchFcb == NULL)
    return NULL;

  bucket = GetHashValue((ULONG) searchFcb);
  for (entry = fcbHashTable[bucket]; entry != NULL;
       entry = entry->nextInFcbHash) {
    if (entry->fileObject->FsContext == searchFcb)
      return entry;
  }
  return NULL;
}

// FindFileObjectInHash finds the entry in the hash table with file object
// number equal to the passed one.  It returns a pointer to that entry,
// or NULL if there is no such entry.

FILE_OBJECT_ENTRY *FindFileObjectInHash (PFILE_OBJECT searchFileObject)
{
  int bucket;
  FILE_OBJECT_ENTRY *entry;

  bucket = GetHashValue((ULONG) searchFileObject);
  for (entry = hashTable[bucket]; entry != NULL; entry = entry->next) {
    if (entry->fileObject == searchFileObject)
      return entry;
  }
  return NULL;
}

// AddFileObjectToHashIfAbsent adds the file object number given to the hash
// only if it is not there already.  It can be called at general times
// because it mediates its access to the hash table and free list through
// the spin lock.  It returns a boolean indicating if it was added.  If
// *canUseFcbPtr is TRUE on entry, and this function finds that the file
// object is absent from the hash, it searches the hash for an identical
// FCB that can be used by reference; if it doesn't find one, it sets
// *canUseFcbPtr to FALSE.

BOOLEAN AddFileObjectToHashIfAbsent (PFILE_OBJECT fileObject,
                                     BOOLEAN *canUseFcbPtr)
{
  FILE_OBJECT_ENTRY *entry;
  KIRQL oldirql;

  KeAcquireSpinLock(&hashSpinLock, &oldirql);
  if (FindFileObjectInHash(fileObject) == NULL) {
    if (*canUseFcbPtr && FindFcbInHash(fileObject->FsContext) == NULL)
      *canUseFcbPtr = FALSE;
    entry = NewFileObjectEntry();
    if (entry != NULL) {
      entry->fileObject = fileObject;
      AddEntryToHash(entry);
      KeReleaseSpinLock(&hashSpinLock, oldirql);
      return TRUE;
    }
  }
  KeReleaseSpinLock(&hashSpinLock, oldirql);
  return FALSE;
}

// RemoveFileObjectFromHash removes the given file object number from
// the hash table.  It then frees the entry that contained that file
// object numer (i.e. puts it on the free list).  It can be called
// at general times because it mediates its access to the hash table
// and free list through the spin lock.

void RemoveFileObjectFromHash (PFILE_OBJECT fileObject)
{
  FILE_OBJECT_ENTRY *entry;
  KIRQL oldirql;

  KeAcquireSpinLock(&hashSpinLock, &oldirql);
  entry = FindFileObjectInHash(fileObject);
  if (entry != NULL) {
    RemoveEntryFromHash(entry);
    FreeFileObjectEntry(entry);
  }
  KeReleaseSpinLock(&hashSpinLock, oldirql);
}
