//
// VTrcFS.c
//
// Code for the VTrcFS driver, a filter driver for file systems.  It logs its
// findings by calling the LogEvent routine in the VTrcLog driver.
//
// (Uses some code by Mark Russinovich and Bryce Cogswell)
//

#include "NTDDK.h"
#include "../Common/W2Kddk.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "FastIo.h"
#include "FileName.h"
#include "GetState.h"
#include "Hash.h"

///////////////////////////////////////////////////////////////////////////
//                              MACROS
///////////////////////////////////////////////////////////////////////////

// Print macro that only turns on when checked builds are on

#if DBG
#define DebugPrint(arg) DbgPrint arg
#else
#define DebugPrint(arg)
#endif

///////////////////////////////////////////////////////////////////////////
//                             STRUCTURES
///////////////////////////////////////////////////////////////////////////

typedef struct _VTRCFS_CONTEXT {
  PIRP           origIrp;
  ULONG          seq;
  PFILE_OBJECT   openedFileObj;
  PDEVICE_OBJECT associatedDevObj;
  BOOLEAN        recordMdl;
} VTRCFS_CONTEXT;

//////////////////////////////////////////////////////////////////////////
//                          IFS KIT DEFINITIONS
//////////////////////////////////////////////////////////////////////////

typedef struct _FILE_RENAME_INFORMATION {
    BOOLEAN ReplaceIfExists;
    HANDLE RootDirectory;
    ULONG FileNameLength;
    WCHAR FileName[1];
} FILE_RENAME_INFORMATION, *PFILE_RENAME_INFORMATION;

///////////////////////////////////////////////////////////////////////////
//                        FORWARD DEFININTIONS
///////////////////////////////////////////////////////////////////////////

NTSTATUS DriverEntry (IN PDRIVER_OBJECT DriverObject,
                      IN PUNICODE_STRING RegistryPath);
NTSTATUS VTrcFSRemoveDevice (PDEVICE_OBJECT DeviceObject, PIRP Irp);
VOID VTrcFSUnload (PDRIVER_OBJECT DriverObject);

NTSTATUS VTrcFSDispatchCreate (PDEVICE_OBJECT HookDevice, IN PIRP Irp);
NTSTATUS VTrcFSDispatchClose (PDEVICE_OBJECT HookDevice, IN PIRP Irp);
NTSTATUS VTrcFSDispatchReadWrite (PDEVICE_OBJECT HookDevice, IN PIRP Irp);
NTSTATUS VTrcFSDispatchQueryInfo (PDEVICE_OBJECT HookDevice, IN PIRP Irp);
NTSTATUS VTrcFSDispatchSetInfo (PDEVICE_OBJECT HookDevice, IN PIRP Irp);
NTSTATUS VTrcFSDispatchDirectoryControl (PDEVICE_OBJECT HookDevice,
                                         IN PIRP Irp);
NTSTATUS VTrcFSDispatchFlush (PDEVICE_OBJECT HookDevice, IN PIRP Irp);
NTSTATUS VTrcFSDispatchLock (PDEVICE_OBJECT HookDevice, IN PIRP Irp);
NTSTATUS VTrcFSDispatchQuerySecurity (PDEVICE_OBJECT HookDevice, IN PIRP Irp);
NTSTATUS VTrcFSDispatchSetSecurity (PDEVICE_OBJECT HookDevice, IN PIRP Irp);
NTSTATUS VTrcFSDispatchPnp (PDEVICE_OBJECT HookDevice, IN PIRP Irp);
NTSTATUS VTrcFSDispatchPower (PDEVICE_OBJECT HookDevice, IN PIRP Irp);

NTSTATUS VTrcFSPassOnNormally (PDEVICE_OBJECT HookDevice, IN PIRP Irp);

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, DriverEntry)
#pragma alloc_text (PAGE, VTrcFSDispatchPnp)
#pragma alloc_text (PAGE, VTrcFSRemoveDevice)
#pragma alloc_text (PAGE, VTrcFSUnload)
#endif

///////////////////////////////////////////////////////////////////////////
//                              GLOBALS
///////////////////////////////////////////////////////////////////////////

// Each IRP is given a sequence number. This allows the return status
// of an IRP, which is obtained in the completion routine, to be
// associated with the IRPs parameters that were extracted in the Dispatch
// routine.

LONG sequenceNumber = 0;

///////////////////////////////////////////////////////////////////////////
//                           DRIVER ENTRY
///////////////////////////////////////////////////////////////////////////

// DriverEntry is the installable driver initialization.  Here we just
// set up the driver object and initiate communication with the VTrcLog
// driver.  The VTrcLog driver takes care of creating all our devices.

NTSTATUS DriverEntry (IN PDRIVER_OBJECT DriverObject,
                      IN PUNICODE_STRING RegistryPath)
{
  NTSTATUS status;
  UNICODE_STRING deviceNameUnicodeString;
  PDEVICE_OBJECT deviceObject;
  PFILE_OBJECT fileObject;
  PFS_HOOK_EXTENSION deviceExtension;
  int i;

  DebugPrint(("VTrcFS: Entering DriverEntry\n"));

  // Get shared state information from VTrcInit. If we can't get it, fail.

  status = GetSharedState();
  if (!NT_SUCCESS(status)) {
    DebugPrint(("VTrcFS: Did not get shared state, so exiting.\n"));
    return status;
  }

  // Initialize the file object hash table

  InitHash();

  // Create dispatch points for all routines that must be handled.
  // All entry points are registered since we might filter a
  // file system that processes all of them.

  for (i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++)
    DriverObject->MajorFunction[i] = VTrcFSPassOnNormally;

  DriverObject->MajorFunction[IRP_MJ_READ] = VTrcFSDispatchReadWrite;
  DriverObject->MajorFunction[IRP_MJ_WRITE] = VTrcFSDispatchReadWrite;
  DriverObject->MajorFunction[IRP_MJ_CREATE] = VTrcFSDispatchCreate;
  DriverObject->MajorFunction[IRP_MJ_CLOSE] = VTrcFSDispatchClose;
  DriverObject->MajorFunction[IRP_MJ_QUERY_INFORMATION] =
                                              VTrcFSDispatchQueryInfo;
  DriverObject->MajorFunction[IRP_MJ_SET_INFORMATION] = VTrcFSDispatchSetInfo;
  DriverObject->MajorFunction[IRP_MJ_DIRECTORY_CONTROL] =
                                               VTrcFSDispatchDirectoryControl;
  DriverObject->MajorFunction[IRP_MJ_FLUSH_BUFFERS] = VTrcFSDispatchFlush;
  DriverObject->MajorFunction[IRP_MJ_LOCK_CONTROL] = VTrcFSDispatchLock;
  DriverObject->MajorFunction[IRP_MJ_QUERY_SECURITY] =
                                                  VTrcFSDispatchQuerySecurity;
  DriverObject->MajorFunction[IRP_MJ_SET_SECURITY] = VTrcFSDispatchSetSecurity;
  DriverObject->MajorFunction[IRP_MJ_PNP] = VTrcFSDispatchPnp;
  DriverObject->MajorFunction[IRP_MJ_POWER] = VTrcFSDispatchPower;

  // Set up the Fast I/O dispatch table

  DriverObject->FastIoDispatch = &VTrcFSFastIoDispatchTable;

  // Note: It would be unwise to unload this driver, so we don't set an
  // unload routine.

  // Set information in the shared state that we know.

  sharedState->fsDriverObject = DriverObject;
  sharedState->emptyFSHashFunc = &EmptyHash;

  return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////
//                       REQUEST COMPLETION ROUTINES
////////////////////////////////////////////////////////////////////////////

// VTrcFSSimpleCompletionRoutine is our simple completion routine.  It logs
// the return values of a completed IRP when all we need to receive is the
// sequence number of that IRP operation.

NTSTATUS VTrcFSSimpleCompletionRoutine (PDEVICE_OBJECT DeviceObject, PIRP Irp,
                                        PVOID Context)
{
  ULONG seq = (ULONG) Context;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the return values.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_OPERATION, 14);
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[1], &seq, 4);
    RtlCopyMemory(&eventBuffer[5], &Irp->IoStatus.Status, 4);
    RtlCopyMemory(&eventBuffer[9], &Irp->IoStatus.Information, 4);
    eventBuffer[13] =
          (char) (Irp->Cancel ? FLAG_CANCELLED : FLAG_COMPLETED_SUCCESSFULLY);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Must do the following in a completion routine

  if (Irp->PendingReturned) {
    IoMarkIrpPending(Irp);
  }

  return Irp->IoStatus.Status;
}

// VTrcFSGeneralCompletionRoutine is our general completion routine.  Not
// only does it log the return values, it also does the following:
//
// 1) If context->openedFileObj is non-NULL, indicating this is the
//    completion of a file open, it logs information about the file.
// 2) If context->origIrp is non-NULL, indicating that this is an IRP we
//    created to handle this original IRP, we copy IRP fields to the
//    original IRP, complete that IRP, and free this IRP.
// 3) If context->recordMdl is TRUE, indicating that this is the completion
//    of an MDL operation, uses the appropriate entry type and copies the
//    MDL address from the IRP into the log.
//
// Thus, it is made the completion routine whenever at least one of these
// things must be done.

NTSTATUS VTrcFSGeneralCompletionRoutine (PDEVICE_OBJECT DeviceObject, PIRP Irp,
                                         PVOID Context)
{
  VTRCFS_CONTEXT *context = (VTRCFS_CONTEXT *) Context;
  PIRP origIrp;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the return values.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  if (context->recordMdl) {
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_MDL_OPERATION, 18);
    if (eventBuffer)
      RtlCopyMemory(&eventBuffer[14], &Irp->MdlAddress, 4);
  }
  else {
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_COMPLETE_OPERATION, 14);
  }
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[1], &context->seq, 4);
    RtlCopyMemory(&eventBuffer[5], &Irp->IoStatus.Status, 4);
    RtlCopyMemory(&eventBuffer[9], &Irp->IoStatus.Information, 4);
    eventBuffer[13] =
          (char) (Irp->Cancel ? FLAG_CANCELLED : FLAG_COMPLETED_SUCCESSFULLY);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // If this is the completion of a file-open IRP, add the opened file object
  // to the hash and record information about it in the log

  if (context->openedFileObj && Irp->IoStatus.Status == STATUS_SUCCESS) {
    NameFileIfUnopened(context->openedFileObj, context->associatedDevObj,
                       FALSE);
  }

  // Free the context structure after extracting the origIrp field

  origIrp = context->origIrp;
  ExFreePool(context);

  // If there is an "original" IRP that this one is derived from, handle
  // that case

  if (origIrp) {

    // Copy important possibly returned fields in this IRP to the original IRP

    origIrp->IoStatus = Irp->IoStatus;
    origIrp->PendingReturned = Irp->PendingReturned;
    origIrp->MdlAddress = Irp->MdlAddress;

    // If this IRP was pending, mark the new and original IRP's pending

    if (Irp->PendingReturned) {
      IoMarkIrpPending(Irp);
      IoMarkIrpPending(origIrp);
    }

    // Free the IRP we created

    IoFreeIrp(Irp);

    // Complete the original IRP

    IoCompleteRequest(origIrp, IO_NO_INCREMENT);

    // Return STATUS_MORE_PROCESSING_REQUIRED to indicate that we're going
    // to take control of our IRP and finish handling it.  Actually, we've
    // already handled it by freeing it.  If we didn't take control of it,
    // the I/O Manager might try to process our IRP further, which would
    // we bad since it's been freed.

    return STATUS_MORE_PROCESSING_REQUIRED;
  }

  // Must do the following in a completion routine

  if (Irp->PendingReturned) {
    IoMarkIrpPending(Irp);
  }

  return Irp->IoStatus.Status;
}

NTSTATUS VTrcFSForwardedIrpCompletion (IN PDEVICE_OBJECT DeviceObject,
                                       IN PIRP Irp, IN PVOID Context)
{
  PKEVENT Event = (PKEVENT) Context;

  UNREFERENCED_PARAMETER(DeviceObject);

  if (Irp->PendingReturned) {
    IoMarkIrpPending(Irp);
  }

  KeSetEvent(Event, IO_NO_INCREMENT, FALSE);

  return STATUS_MORE_PROCESSING_REQUIRED;
}

////////////////////////////////////////////////////////////////////////////
//              FUNCTIONS TO PASS IRP ON TO NEXT LOWER DEVICE
////////////////////////////////////////////////////////////////////////////

// VTrcFSPassOnNormally passes on the given IRP to the next lower device
// (also given).  It makes no effort to log the completion of the IRP.

NTSTATUS VTrcFSPassOnNormally (PDEVICE_OBJECT HookDevice, IN PIRP Irp)
{
  PFS_HOOK_EXTENSION hookExt = HookDevice->DeviceExtension;

  // Skip the current stack location

  IoSkipCurrentIrpStackLocation(Irp);

  // Call the next lower driver

  return IoCallDriver(hookExt->attachedDevice, Irp);
}

// VTrcFSPassOnWithNewIrp passes on the given IRP so that its completion
// can be logged.  It does this by creating a whole new IRP and passing _that_
// on.  It sets the completion routine to VTrcFSGeneralCompletionRoutine with a
// context that causes the new IRP to get freed and the original IRP to get
// completed correctly.  (Note that if this routine fails in some way, such
// as being unable to allocate memory, its failure mode is to just call
// VTrcFSPassOnNormally, which does the same thing but does not log the
// completion of the IRP and thus does not need a new IRP.)  The parameter
// openedFileObj is the file object opened by this IRP; it is NULL if this
// IRP is not an open IRP.  The parameter recordMdl indicates whether the
// operation is an MDL operation.

NTSTATUS VTrcFSPassOnWithNewIrp (PDEVICE_OBJECT HookDevice, IN PIRP Irp,
                                 ULONG seq, PFILE_OBJECT openedFileObj,
                                 BOOLEAN recordMdl)
{
  PFS_HOOK_EXTENSION  hookExt = HookDevice->DeviceExtension;
  PDEVICE_OBJECT      deviceObject = hookExt->attachedDevice;
  PIO_STACK_LOCATION  currentIrpStackLocation;
  PIO_STACK_LOCATION  newIrpFirstStackLocation;
  PIRP                newIrp;
  VTRCFS_CONTEXT      *context;

  // Allocate a new IRP with stack size sufficient for the next lower device

  context = ExAllocatePool(NonPagedPool, sizeof(VTRCFS_CONTEXT));
  if (context == NULL)
    return VTrcFSPassOnNormally(HookDevice, Irp);
  newIrp = IoAllocateIrp(deviceObject->StackSize, FALSE);
  if (newIrp == NULL) {
    ExFreePool(context);
    return VTrcFSPassOnNormally(HookDevice, Irp);
  }

  // Fill in important fields of new IRP with those of original IRP

  newIrp->MdlAddress = Irp->MdlAddress;
  newIrp->Flags = Irp->Flags;
  newIrp->AssociatedIrp = Irp->AssociatedIrp;
  newIrp->RequestorMode = Irp->RequestorMode;
  newIrp->UserIosb = Irp->UserIosb;
  newIrp->UserEvent = Irp->UserEvent;
  newIrp->Overlay = Irp->Overlay;
  newIrp->UserBuffer = Irp->UserBuffer;
  newIrp->Tail.Overlay.AuxiliaryBuffer = Irp->Tail.Overlay.AuxiliaryBuffer;
  newIrp->Tail.Overlay.OriginalFileObject =
                                        Irp->Tail.Overlay.OriginalFileObject;
  newIrp->Tail.Overlay.Thread = Irp->Tail.Overlay.Thread;

  // Copy current stack location to the first stack location in the new IRP

  currentIrpStackLocation = IoGetCurrentIrpStackLocation(Irp);
  newIrpFirstStackLocation = IoGetNextIrpStackLocation(newIrp);
  *newIrpFirstStackLocation = *currentIrpStackLocation;
  newIrpFirstStackLocation->DeviceObject = deviceObject;

  // Fill in the context structure, then set the completion routine to
  // VTrcFSGeneralCompletionRoutine with it as context

  context->origIrp = Irp;
  context->seq = seq;
  context->openedFileObj = openedFileObj;
  context->associatedDevObj = deviceObject;
  context->recordMdl = recordMdl;
  IoSetCompletionRoutine(newIrp, VTrcFSGeneralCompletionRoutine, context,
                         TRUE, TRUE, TRUE);

  // Send the new IRP to the lower-level device

  return IoCallDriver(deviceObject, newIrp);
}

// VTrcFSPassOnWithTailLogging passes on the given IRP to the next lower
// device in a way that the completion of the IRP gets logged.  The
// sequence number to use in the completion log entry is given as seq.
// The file object that is being opened, if any, is passed as openedFileObj.
// The parameter recordMdl indicates if this operation is an MDL operation.

NTSTATUS VTrcFSPassOnWithTailLogging (PDEVICE_OBJECT HookDevice, IN PIRP Irp,
                                      ULONG seq, PFILE_OBJECT openedFileObj,
                                      BOOLEAN recordMdl)
{
  PIO_STACK_LOCATION  currentIrpStackLocation, nextIrpStackLocation;
  PFS_HOOK_EXTENSION  hookExt;
  PDEVICE_OBJECT      deviceObject;
  VTRCFS_CONTEXT      *context;

  // If there isn't room in the stack for passing on the IRP, we will
  // have to make a new IRP and pass _that_ on to the next lower driver.

  if (Irp->CurrentLocation < HookDevice->StackSize)
    return VTrcFSPassOnWithNewIrp(HookDevice, Irp, seq, openedFileObj,
                                  recordMdl);

  // Get a pointer to the lower-level device object from the extension

  hookExt = HookDevice->DeviceExtension;
  deviceObject = hookExt->attachedDevice;

  // If we have to do anything special in the completion routine, such
  // as get information about a file being opened or record information
  // about an MDL, allocate a context structure for the general
  // completion routine and fill it in.  If the allocation fails, we
  // just pass this IRP on without tail logging.

  if (openedFileObj || recordMdl) {
    context = ExAllocatePool(NonPagedPool, sizeof(VTRCFS_CONTEXT));
    if (context == NULL)
      return VTrcFSPassOnNormally(HookDevice, Irp);
    context->origIrp = NULL;
    context->seq = seq;
    context->openedFileObj = openedFileObj;
    context->associatedDevObj = deviceObject;
    context->recordMdl = recordMdl;
  }

  // Copy parameters down to next level in the stack for the driver below us

  currentIrpStackLocation = IoGetCurrentIrpStackLocation(Irp);
  nextIrpStackLocation = IoGetNextIrpStackLocation(Irp);
  *nextIrpStackLocation = *currentIrpStackLocation;
  nextIrpStackLocation->DeviceObject = deviceObject;

  // Set the completion routine to the general one if something complex
  // needs to be done.  Otherwise, set it to the simple completion routine.
  // The "extra" braces below are required since IoSetCompletionRoutine is
  // a macro that expands to multiple statements.

  if (openedFileObj || recordMdl) {
    IoSetCompletionRoutine(Irp, VTrcFSGeneralCompletionRoutine, context,
                           TRUE, TRUE, TRUE);
  }
  else {
    IoSetCompletionRoutine(Irp, VTrcFSSimpleCompletionRoutine, (PVOID) seq,
                           TRUE, TRUE, TRUE);
  }

  // Call the lower-level driver

  return IoCallDriver(deviceObject, Irp);
}

NTSTATUS VTrcFSForwardIrpSynchronous (PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
  PFS_HOOK_EXTENSION deviceExtension;
  KEVENT event;
  NTSTATUS status;

  KeInitializeEvent(&event, NotificationEvent, FALSE);
  deviceExtension = (PFS_HOOK_EXTENSION) DeviceObject->DeviceExtension;

  // copy the irp stack for the next device

  IoCopyCurrentIrpStackLocationToNext(Irp);

  // set a completion routine

  IoSetCompletionRoutine(Irp, VTrcFSForwardedIrpCompletion, &event, TRUE, TRUE,
			 TRUE);

  // call the next lower device

  status = IoCallDriver(deviceExtension->attachedDevice, Irp);

  // wait for the actual completion

  if (status == STATUS_PENDING) {
    KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, NULL);
    status = Irp->IoStatus.Status;
  }

  return status;
}

///////////////////////////////////////////////////////////////////////////
//                           DISPATCH ROUTINES
///////////////////////////////////////////////////////////////////////////

// VTrcFSDispatchCreate is the dispatch routine called for any create IRP.
// It logs the parameters of the call and sets up a completion routine to
// log the return values.

NTSTATUS VTrcFSDispatchCreate (PDEVICE_OBJECT HookDevice, IN PIRP Irp)
{
  PIO_STACK_LOCATION   currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
  PFILE_OBJECT         fileObject      = currentIrpStack->FileObject;
  PFS_HOOK_EXTENSION   hookExt         = HookDevice->DeviceExtension;
  ULONG                seq;
  KIRQL                oldirql;
  PIO_SECURITY_CONTEXT securityContext;
  PCHAR                eventBuffer;
  ULONG                threadId;

  /*
  DebugPrint(("VTrcFSDispatchCreate(%x, %x), file %x\n",
              HookDevice, Irp, fileObject));
  */

  // Increment the number of create IRP's filtered.

  sharedState->filteredCreates++;

  // If this file doesn't have a name, don't bother logging this operation.

  if (fileObject->FileName.Buffer == NULL)
    return VTrcFSPassOnNormally(HookDevice, Irp);

  // Indicate that the current thread is active

  threadId = (ULONG) PsGetCurrentThreadId();
  if (threadId < sharedState->kernelUserSharedState.maxThreads)
    sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] = '\0';

  // Remove this file number from the hash since if it was open before, it's
  // different now.

  RemoveFileObjectFromHash(fileObject);

  // Log the input values.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_OPEN, 26);
  if (eventBuffer) {
    seq = InterlockedIncrement(&sequenceNumber);
    RtlCopyMemory(&eventBuffer[1], &seq, 4);
    RtlCopyMemory(&eventBuffer[5], &fileObject, 4);
    RtlCopyMemory(&eventBuffer[9], &hookExt->attachedDevice, 4);
    RtlCopyMemory(&eventBuffer[13],
                  &currentIrpStack->Parameters.Create.Options, 4);
    RtlCopyMemory(&eventBuffer[17],
                  &currentIrpStack->Parameters.Create.FileAttributes, 2);
    RtlCopyMemory(&eventBuffer[19],
                  &currentIrpStack->Parameters.Create.ShareAccess, 2);
    eventBuffer[21] = (char) currentIrpStack->Flags;
    securityContext = currentIrpStack->Parameters.Create.SecurityContext;
    if (securityContext == NULL)
      RtlZeroMemory(&eventBuffer[22], 4);
    else
      RtlCopyMemory(&eventBuffer[22],
                    &securityContext->DesiredAccess, 4);
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcFSPassOnWithTailLogging(HookDevice, Irp, seq, fileObject,
                                       FALSE);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcFSPassOnNormally(HookDevice, Irp);
  }
}

// VTrcFSDispatchClose is the dispatch routine called for any close IRP.
// It logs the parameters of the call and sets up a completion routine to
// log the return values.

NTSTATUS VTrcFSDispatchClose (PDEVICE_OBJECT HookDevice, IN PIRP Irp)
{
  PIO_STACK_LOCATION   currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
  PFILE_OBJECT         fileObject      = currentIrpStack->FileObject;
  PFS_HOOK_EXTENSION   hookExt;
  ULONG                seq;
  KIRQL                oldirql;
  PCHAR                eventBuffer;
  ULONG                threadId;

  /*
  DebugPrint(("VTrcFSDispatchClose(%x, %x), file %x\n",
              HookDevice, Irp, fileObject));
  */

  // If this file doesn't have a name, don't bother logging this operation.

  if (fileObject->FileName.Buffer == NULL)
    return VTrcFSPassOnNormally(HookDevice, Irp);

  // Indicate that the current thread is active

  threadId = (ULONG) PsGetCurrentThreadId();
  if (threadId < sharedState->kernelUserSharedState.maxThreads)
    sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] = '\0';

  // Log the file's name if it hasn't already been logged.

  hookExt = (PFS_HOOK_EXTENSION) HookDevice->DeviceExtension;
  NameFileIfUnopened(fileObject, hookExt->attachedDevice, TRUE);
  RemoveFileObjectFromHash(fileObject);

  // Log the input vaules.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_CLOSE, 9);;
  if (eventBuffer) {
    seq = InterlockedIncrement(&sequenceNumber);
    RtlCopyMemory(&eventBuffer[1], &seq, 4);
    RtlCopyMemory(&eventBuffer[5], &fileObject, 4);
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcFSPassOnWithTailLogging(HookDevice, Irp, seq, NULL, FALSE);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcFSPassOnNormally(HookDevice, Irp);
  }
}

// VTrcFSDispatchReadWrite is the dispatch routine called for any read
// or write IRP. It logs the parameters of the call and sets up a
// completion routine to log the return values.

NTSTATUS VTrcFSDispatchReadWrite (PDEVICE_OBJECT HookDevice, IN PIRP Irp)
{
  PIO_STACK_LOCATION   currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
  PFILE_OBJECT         fileObject      = currentIrpStack->FileObject;
  PFS_HOOK_EXTENSION   hookExt;
  ULONG                flags;
  ULONG                seq;
  KIRQL                oldirql;
  PCHAR                eventBuffer;
  ULONG                threadId;

  /*
  DebugPrint(("VTrcFSDispatchReadWrite(%x, %x), file %x\n",
              HookDevice, Irp, fileObject));
  */

  // If this file doesn't have a name, don't bother logging this operation.

  if (fileObject->FileName.Buffer == NULL)
    return VTrcFSPassOnNormally(HookDevice, Irp);

  // Indicate that the current thread is active

  threadId = (ULONG) PsGetCurrentThreadId();
  if (threadId < sharedState->kernelUserSharedState.maxThreads)
    sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] = '\0';

  // Log the file's name if it hasn't already been logged.

  hookExt = (PFS_HOOK_EXTENSION) HookDevice->DeviceExtension;
  NameFileIfUnopened(fileObject, hookExt->attachedDevice, TRUE);

  // Since Irp->Flags & 0x40000 has no special meaning, we take that bit
  // to indicate the mode from which this read/write call was made.  (We
  // don't modify Irp->Flags, of course; we just modify a copy of those
  // flags that we will put into the log entry.)

  flags = Irp->Flags;
  if (Irp->RequestorMode == KernelMode)
    flags |= 0x40000;

  // Log the input values.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  if (currentIrpStack->MajorFunction == IRP_MJ_READ)
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_READ, 24);
  else
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_WRITE, 24);
  if (eventBuffer) {
    seq = InterlockedIncrement(&sequenceNumber);
    RtlCopyMemory(&eventBuffer[1], &seq, 4);
    RtlCopyMemory(&eventBuffer[5], &fileObject, 4);
    RtlCopyMemory(&eventBuffer[9],
                  &currentIrpStack->Parameters.Read.ByteOffset, 5);
    RtlCopyMemory(&eventBuffer[14],
                  &currentIrpStack->Parameters.Read.Length, 4);
    RtlCopyMemory(&eventBuffer[18], &flags, 4);
    eventBuffer[22] = currentIrpStack->MinorFunction;
    eventBuffer[23] = currentIrpStack->Flags;
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

    // Call VTrcPassOnWithTailLogging to pass on the IRP and log its results.
    // Pass either TRUE or FALSE in the recordMdl parameter depending on
    // whether this is an MDL operation.

    if (currentIrpStack->MinorFunction & IRP_MN_MDL)
      return VTrcFSPassOnWithTailLogging(HookDevice, Irp, seq, NULL, TRUE);
    else
      return VTrcFSPassOnWithTailLogging(HookDevice, Irp, seq, NULL, FALSE);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcFSPassOnNormally(HookDevice, Irp);
  }
}

// VTrcFSDispatchQueryInfo is the dispatch routine called for any query-
// info IRP. It logs the parameters of the call and sets up a
// completion routine to log the return values.

NTSTATUS VTrcFSDispatchQueryInfo (PDEVICE_OBJECT HookDevice, IN PIRP Irp)
{
  PIO_STACK_LOCATION   currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
  PFILE_OBJECT         fileObject      = currentIrpStack->FileObject;
  PFS_HOOK_EXTENSION   hookExt;
  ULONG                seq;
  KIRQL                oldirql;
  PCHAR                eventBuffer;
  ULONG                threadId;

  /*
  DebugPrint(("VTrcFSDispatchQueryInfo(%x, %x), file %x\n",
              HookDevice, Irp, fileObject));
  */

  // If this file doesn't have a name, don't bother logging this operation.

  if (fileObject->FileName.Buffer == NULL)
    return VTrcFSPassOnNormally(HookDevice, Irp);

  // Indicate that the current thread is active

  threadId = (ULONG) PsGetCurrentThreadId();
  if (threadId < sharedState->kernelUserSharedState.maxThreads)
    sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] = '\0';

  // Log the file's name if it hasn't already been logged.

  hookExt = (PFS_HOOK_EXTENSION) HookDevice->DeviceExtension;
  NameFileIfUnopened(fileObject, hookExt->attachedDevice, TRUE);

  // Log the input vaules.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_QUERY_INFO, 10);
  if (eventBuffer) {
    seq = InterlockedIncrement(&sequenceNumber);
    RtlCopyMemory(&eventBuffer[1], &seq, 4);
    RtlCopyMemory(&eventBuffer[5], &fileObject, 4);
    eventBuffer[9] =
          (char) currentIrpStack->Parameters.QueryFile.FileInformationClass;
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcFSPassOnWithTailLogging(HookDevice, Irp, seq, NULL, FALSE);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcFSPassOnNormally(HookDevice, Irp);
  }
}

// VTrcFSDispatchSetInfo is the dispatch routine called for any set-
// info IRP. It logs the parameters of the call and sets up a
// completion routine to log the return values.

NTSTATUS VTrcFSDispatchSetInfo (PDEVICE_OBJECT HookDevice, IN PIRP Irp)
{
  PIO_STACK_LOCATION   currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
  PFILE_OBJECT         fileObject      = currentIrpStack->FileObject;
  PFS_HOOK_EXTENSION   hookExt         = HookDevice->DeviceExtension;
  ULONG                seq;
  KIRQL                oldirql;
  PCHAR                eventBuffer;
  ULONG                threadId;

  /*
  DebugPrint(("VTrcFSDispatchSetInfo(%x, %x), file %x\n",
              HookDevice, Irp, fileObject));
  */

  // If this file doesn't have a name, don't bother logging this operation.

  if (fileObject->FileName.Buffer == NULL)
    return VTrcFSPassOnNormally(HookDevice, Irp);

  // Indicate that the current thread is active

  threadId = (ULONG) PsGetCurrentThreadId();
  if (threadId < sharedState->kernelUserSharedState.maxThreads)
    sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] = '\0';

  // Log the file's name if it hasn't already been logged.

  hookExt = (PFS_HOOK_EXTENSION) HookDevice->DeviceExtension;
  NameFileIfUnopened(fileObject, hookExt->attachedDevice, TRUE);

  // Log the input vaules.

  switch (currentIrpStack->Parameters.SetFile.FileInformationClass) {
  case FileEndOfFileInformation :
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_SET_POSITION, 15);
    if (eventBuffer) {
      seq = InterlockedIncrement(&sequenceNumber);
      eventBuffer[1] = (char) POSITION_END_OF_FILE;
      RtlCopyMemory(&eventBuffer[2], &seq, 4);
      RtlCopyMemory(&eventBuffer[6], &fileObject, 4);
      RtlCopyMemory(&eventBuffer[10], Irp->AssociatedIrp.SystemBuffer, 5);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    break;
  case FilePositionInformation :
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_SET_POSITION, 15);
    if (eventBuffer) {
      seq = InterlockedIncrement(&sequenceNumber);
      eventBuffer[1] = (char) POSITION_FILE_POINTER;
      RtlCopyMemory(&eventBuffer[2], &seq, 4);
      RtlCopyMemory(&eventBuffer[6], &fileObject, 4);
      RtlCopyMemory(&eventBuffer[10], Irp->AssociatedIrp.SystemBuffer, 5);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    break;
  case FileAllocationInformation :
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_SET_POSITION, 15);
    if (eventBuffer) {
      seq = InterlockedIncrement(&sequenceNumber);
      eventBuffer[1] = (char) POSITION_ALLOCATION_SIZE;
      RtlCopyMemory(&eventBuffer[2], &seq, 4);
      RtlCopyMemory(&eventBuffer[6], &fileObject, 4);
      RtlCopyMemory(&eventBuffer[10], Irp->AssociatedIrp.SystemBuffer, 5);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    break;
  case FileDispositionInformation :
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_DELETE, 10);
    if (eventBuffer) {
      seq = InterlockedIncrement(&sequenceNumber);
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &fileObject, 4);
      RtlCopyMemory(&eventBuffer[9], Irp->AssociatedIrp.SystemBuffer, 1);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    break;
  case FileRenameInformation :
    {
      char                      name[240];
      ULONG                     fileNameLength;
      FILE_OBJECT               dummyFileObject;
      PFILE_RENAME_INFORMATION  renameInfo;

      renameInfo = Irp->AssociatedIrp.SystemBuffer;
      dummyFileObject.FileName.Buffer = renameInfo->FileName;
      dummyFileObject.FileName.Length = (USHORT) renameInfo->FileNameLength;
      if (renameInfo->RootDirectory != NULL &&
          ObReferenceObjectByHandle(renameInfo->RootDirectory,
                                    STANDARD_RIGHTS_REQUIRED, NULL,
                                    KernelMode,
                                    (PVOID*)&dummyFileObject.RelatedFileObject,
                                    NULL) == STATUS_SUCCESS) {
        fileNameLength = GetFullPath(&dummyFileObject, name, 240);
        ObDereferenceObject(dummyFileObject.RelatedFileObject);
      }
      else {
        dummyFileObject.RelatedFileObject = NULL;
        fileNameLength = GetFullPath(&dummyFileObject, name, 240);
      }
      KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
      eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_RENAME,
                              fileNameLength + 15);
      if (eventBuffer) {
        seq = InterlockedIncrement(&sequenceNumber);
        eventBuffer[1] = (char) (fileNameLength + 15);
        RtlCopyMemory(&eventBuffer[2], &seq, 4);
        RtlCopyMemory(&eventBuffer[6], &fileObject, 4);
        RtlCopyMemory(&eventBuffer[10],
                      &currentIrpStack->Parameters.SetFile.FileObject, 4);
        eventBuffer[14] = renameInfo->ReplaceIfExists;
        RtlCopyMemory(&eventBuffer[15], name, fileNameLength);
      }
      KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    }
    break;

  case FileLinkInformation :
    {
      char                      name[240];
      ULONG                     fileNameLength;
      FILE_OBJECT               dummyFileObject;
      PFILE_RENAME_INFORMATION  renameInfo;

      renameInfo = Irp->AssociatedIrp.SystemBuffer;
      dummyFileObject.FileName.Buffer = renameInfo->FileName;
      dummyFileObject.FileName.Length = (USHORT) renameInfo->FileNameLength;
      if (renameInfo->RootDirectory != NULL &&
          ObReferenceObjectByHandle(renameInfo->RootDirectory,
                                    STANDARD_RIGHTS_REQUIRED, NULL,
                                    KernelMode,
                                    (PVOID*)&dummyFileObject.RelatedFileObject,
                                    NULL) == STATUS_SUCCESS) {
        fileNameLength = GetFullPath(&dummyFileObject, name, 240);
        ObDereferenceObject(dummyFileObject.RelatedFileObject);
      }
      else {
        dummyFileObject.RelatedFileObject = NULL;
        fileNameLength = GetFullPath(&dummyFileObject, name, 240);
      }

      KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
      eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_LINK,
                              fileNameLength + 15);
      if (eventBuffer) {
        seq = InterlockedIncrement(&sequenceNumber);
        eventBuffer[1] = (char) (fileNameLength + 15);
        RtlCopyMemory(&eventBuffer[2], &seq, 4);
        RtlCopyMemory(&eventBuffer[6], &fileObject, 4);
        RtlCopyMemory(&eventBuffer[10],
                      &currentIrpStack->Parameters.SetFile.FileObject, 4);
        eventBuffer[14] = renameInfo->ReplaceIfExists;
        RtlCopyMemory(&eventBuffer[15], name, fileNameLength);
      }
      KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    }
    break;
  default :
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_SET_INFO, 10);
    if (eventBuffer) {
      seq = InterlockedIncrement(&sequenceNumber);
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &fileObject, 4);
      eventBuffer[9] =
             (char) currentIrpStack->Parameters.SetFile.FileInformationClass;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    break;
  }

  if (eventBuffer)
    return VTrcFSPassOnWithTailLogging(HookDevice, Irp, seq, NULL, FALSE);
  else
    return VTrcFSPassOnNormally(HookDevice, Irp);
}

// VTrcFSDispatchDirectoryControl is the dispatch routine called for any
// directory control IRP.  If it turns out to be a query-directory call,
// this function logs the parameters of the call and sets up a completion
// routine to log the return values.  Otherwise, it just passes on the IRP.

NTSTATUS VTrcFSDispatchDirectoryControl (PDEVICE_OBJECT HookDevice,
                                         IN PIRP Irp)
{
  PIO_STACK_LOCATION   currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
  PFILE_OBJECT         fileObject      = currentIrpStack->FileObject;
  PFS_HOOK_EXTENSION   hookExt;
  ULONG                seq;
  KIRQL                oldirql;
  PCHAR                eventBuffer;
  ULONG                threadId;

  /*
  DebugPrint(("VTrcFSDispatchDirectoryControl(%x, %x), file %x\n",
              HookDevice, Irp, fileObject));
  */

  // If this file doesn't have a name, or if this isn't a query-directory
  // request, don't bother logging this operation.

  if (fileObject->FileName.Buffer == NULL ||
      currentIrpStack->MinorFunction != IRP_MN_QUERY_DIRECTORY)
    return VTrcFSPassOnNormally(HookDevice, Irp);

  // Indicate that the current thread is active

  threadId = (ULONG) PsGetCurrentThreadId();
  if (threadId < sharedState->kernelUserSharedState.maxThreads)
    sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] = '\0';

  // Log the file's name if it hasn't already been logged.

  hookExt = (PFS_HOOK_EXTENSION) HookDevice->DeviceExtension;
  NameFileIfUnopened(fileObject, hookExt->attachedDevice, TRUE);

  // Log the input vaules.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_DIRECTORY_CONTROL, 11);
  if (eventBuffer) {
    seq = InterlockedIncrement(&sequenceNumber);
    RtlCopyMemory(&eventBuffer[1], &seq, 4);
    RtlCopyMemory(&eventBuffer[5], &fileObject, 4);
    eventBuffer[9] = (char) currentIrpStack->Parameters.Others.Argument3;
    eventBuffer[10] = currentIrpStack->Flags;
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcFSPassOnWithTailLogging(HookDevice, Irp, seq, NULL, FALSE);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcFSPassOnNormally(HookDevice, Irp);
  }
}

// VTrcFSDispatchFlush is the dispatch routine called for any flush IRP.
// This function logs the parameters of the call and sets up a completion
// routine to log the return value.

NTSTATUS VTrcFSDispatchFlush (PDEVICE_OBJECT HookDevice, IN PIRP Irp)
{
  PIO_STACK_LOCATION   currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
  PFILE_OBJECT         fileObject      = currentIrpStack->FileObject;
  PFS_HOOK_EXTENSION   hookExt;
  ULONG                seq;
  KIRQL                oldirql;
  PCHAR                eventBuffer;
  ULONG                threadId;

  /*
  DebugPrint(("VTrcFSDispatchFlush(%x, %x), file %x\n",
              HookDevice, Irp, fileObject));
  */

  // Log the file name if it hasn't already been logged.

  if (fileObject->FileName.Buffer != NULL) {
    hookExt = (PFS_HOOK_EXTENSION) HookDevice->DeviceExtension;
    NameFileIfUnopened(fileObject, hookExt->attachedDevice, TRUE);
  }

  // Indicate that the current thread is active

  threadId = (ULONG) PsGetCurrentThreadId();
  if (threadId < sharedState->kernelUserSharedState.maxThreads)
    sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] = '\0';

  // Log the input values.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_FLUSH, 9);
  if (eventBuffer) {
    seq = InterlockedIncrement(&sequenceNumber);
    RtlCopyMemory(&eventBuffer[1], &seq, 4);
    RtlCopyMemory(&eventBuffer[5], &fileObject, 4);
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcFSPassOnWithTailLogging(HookDevice, Irp, seq, NULL, FALSE);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcFSPassOnNormally(HookDevice, Irp);
  }
}

// VTrcFSDispatchLock is the dispatch routine called for any lock IRP.
// This function logs the parameters of the call and sets up a completion
// routine to log the return values.

NTSTATUS VTrcFSDispatchLock (PDEVICE_OBJECT HookDevice, IN PIRP Irp)
{
  PIO_STACK_LOCATION   currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
  PFILE_OBJECT         fileObject      = currentIrpStack->FileObject;
  PFS_HOOK_EXTENSION   hookExt;
  UCHAR                minorFunction;
  ULONG                seq;
  KIRQL                oldirql;
  PCHAR                eventBuffer;
  PLARGE_INTEGER       lengthPtr;
  ULONG                threadId;

  /*
  DebugPrint(("VTrcFSDispatchLock(%x, %x), file %x\n",
              HookDevice, Irp, fileObject));
  */

  // If this file doesn't have a name, don't bother logging this operation.

  if (fileObject->FileName.Buffer == NULL)
    return VTrcFSPassOnNormally(HookDevice, Irp);

  // Indicate that the current thread is active

  threadId = (ULONG) PsGetCurrentThreadId();
  if (threadId < sharedState->kernelUserSharedState.maxThreads)
    sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] = '\0';

  // Log the file's name if it hasn't already been logged.

  hookExt = (PFS_HOOK_EXTENSION) HookDevice->DeviceExtension;
  NameFileIfUnopened(fileObject, hookExt->attachedDevice, TRUE);

  // Log the input vaules.

  minorFunction = currentIrpStack->MinorFunction;
  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  switch (minorFunction) {
  case IRP_MN_LOCK :
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_LOCK, 24);
    if (eventBuffer) {
      seq = InterlockedIncrement(&sequenceNumber);
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &fileObject, 4);
      eventBuffer[9] = currentIrpStack->Flags;
      RtlCopyMemory(&eventBuffer[10],
                    currentIrpStack->Parameters.Others.Argument1, 5);
      RtlCopyMemory(&eventBuffer[15],
                    &currentIrpStack->Parameters.Others.Argument2, 9);
    }
    break;
  case IRP_MN_UNLOCK_SINGLE :
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_UNLOCK, 25);
    if (eventBuffer) {
      seq = InterlockedIncrement(&sequenceNumber);
      eventBuffer[1] = (char) 25;
      eventBuffer[2] = minorFunction;
      RtlCopyMemory(&eventBuffer[3], &seq, 4);
      RtlCopyMemory(&eventBuffer[7], &fileObject, 4);
      lengthPtr = currentIrpStack->Parameters.Others.Argument1;
      if (lengthPtr == NULL)
        RtlZeroMemory(&eventBuffer[11], 5);
      else
        RtlCopyMemory(&eventBuffer[11], lengthPtr, 5);
      RtlCopyMemory(&eventBuffer[16],
                    &currentIrpStack->Parameters.Others.Argument2, 9);
    }
    break;
  case IRP_MN_UNLOCK_ALL_BY_KEY :
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_UNLOCK, 15);
    if (eventBuffer) {
      seq = InterlockedIncrement(&sequenceNumber);
      eventBuffer[1] = (char) 15;
      eventBuffer[2] = minorFunction;
      RtlCopyMemory(&eventBuffer[3], &seq, 4);
      RtlCopyMemory(&eventBuffer[7], &fileObject, 4);
      RtlCopyMemory(&eventBuffer[11],
                    &currentIrpStack->Parameters.Others.Argument2, 4);
    }
    break;
  default :
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_UNLOCK, 11);
    if (eventBuffer) {
      seq = InterlockedIncrement(&sequenceNumber);
      eventBuffer[1] = (char) 11;
      eventBuffer[2] = minorFunction;
      RtlCopyMemory(&eventBuffer[3], &seq, 4);
      RtlCopyMemory(&eventBuffer[7], &fileObject, 4);
    }
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  if (eventBuffer)
    return VTrcFSPassOnWithTailLogging(HookDevice, Irp, seq, NULL, FALSE);
  else
    return VTrcFSPassOnNormally(HookDevice, Irp);
}

// VTrcFSDispatchQuerySecurity is the dispatch routine called for any query-
// security IRP. It logs the parameters of the call and sets up a
// completion routine to log the return values.

NTSTATUS VTrcFSDispatchQuerySecurity (PDEVICE_OBJECT HookDevice, IN PIRP Irp)
{
  PIO_STACK_LOCATION   currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
  PFILE_OBJECT         fileObject      = currentIrpStack->FileObject;
  PFS_HOOK_EXTENSION   hookExt;
  ULONG                seq;
  KIRQL                oldirql;
  PCHAR                eventBuffer;
  ULONG                threadId;

  /*
  DebugPrint(("VTrcFSDispatchQuerySecurity(%x, %x), file %x\n",
              HookDevice, Irp, fileObject));
  */

  // If this file doesn't have a name, don't bother logging this operation.

  if (fileObject->FileName.Buffer == NULL)
    return VTrcFSPassOnNormally(HookDevice, Irp);

  // Indicate that the current thread is active

  threadId = (ULONG) PsGetCurrentThreadId();
  if (threadId < sharedState->kernelUserSharedState.maxThreads)
    sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] = '\0';

  // Log the file's name if it hasn't already been logged.

  hookExt = (PFS_HOOK_EXTENSION) HookDevice->DeviceExtension;
  NameFileIfUnopened(fileObject, hookExt->attachedDevice, TRUE);

  // Log the input vaules.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_QUERY_INFO, 10);
  if (eventBuffer) {
    seq = InterlockedIncrement(&sequenceNumber);
    RtlCopyMemory(&eventBuffer[1], &seq, 4);
    RtlCopyMemory(&eventBuffer[5], &fileObject, 4);
    eventBuffer[9] = (char) FILE_INFORMATION_SECURITY_CODE;
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcFSPassOnWithTailLogging(HookDevice, Irp, seq, NULL, FALSE);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcFSPassOnNormally(HookDevice, Irp);
  }
}

// VTrcFSDispatchSetSecurity is the dispatch routine called for any set-
// security IRP. It logs the parameters of the call and sets up a
// completion routine to log the return values.

NTSTATUS VTrcFSDispatchSetSecurity (PDEVICE_OBJECT HookDevice, IN PIRP Irp)
{
  PIO_STACK_LOCATION   currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
  PFILE_OBJECT         fileObject      = currentIrpStack->FileObject;
  PFS_HOOK_EXTENSION   hookExt;
  ULONG                seq;
  KIRQL                oldirql;
  PCHAR                eventBuffer;
  ULONG                threadId;

  /*
  DebugPrint(("VTrcFSDispatchSetSecurity(%x, %x), file %x\n",
              HookDevice, Irp, fileObject));
  */

  // If this file doesn't have a name, don't bother logging this operation.

  if (fileObject->FileName.Buffer == NULL)
    return VTrcFSPassOnNormally(HookDevice, Irp);

  // Indicate that the current thread is active

  threadId = (ULONG) PsGetCurrentThreadId();
  if (threadId < sharedState->kernelUserSharedState.maxThreads)
    sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] = '\0';

  // Log the file's name if it hasn't already been logged.

  hookExt = (PFS_HOOK_EXTENSION) HookDevice->DeviceExtension;
  NameFileIfUnopened(fileObject, hookExt->attachedDevice, TRUE);

  // Log the input vaules.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_FILE_SET_INFO, 10);
  if (eventBuffer) {
    seq = InterlockedIncrement(&sequenceNumber);
    RtlCopyMemory(&eventBuffer[1], &seq, 4);
    RtlCopyMemory(&eventBuffer[5], &fileObject, 4);
    eventBuffer[9] = (char) FILE_INFORMATION_SECURITY_CODE;
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcFSPassOnWithTailLogging(HookDevice, Irp, seq, NULL, FALSE);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcFSPassOnNormally(HookDevice, Irp);
  }
}

NTSTATUS VTrcFSRemoveDevice (PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
  NTSTATUS            status;
  PFS_HOOK_EXTENSION  deviceExtension;

  PAGED_CODE();

  deviceExtension = DeviceObject->DeviceExtension;

  status = VTrcFSForwardIrpSynchronous(DeviceObject, Irp);

  IoDetachDevice(deviceExtension->attachedDevice);
  IoDeleteDevice(DeviceObject);

  // Complete the Irp

  Irp->IoStatus.Status = status;
  IoCompleteRequest(Irp, IO_NO_INCREMENT);

  return status;
}

VOID VTrcFSUnload (PDRIVER_OBJECT DriverObject)
{
  PAGED_CODE();

  // I should free the hash list here, but I don't

  if (sharedState)
    sharedState->emptyFSHashFunc = NULL;

  return;
}

NTSTATUS VTrcFSDispatchPnp (PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
  PIO_STACK_LOCATION  irpSp = IoGetCurrentIrpStackLocation(Irp);
  NTSTATUS            status;

  PAGED_CODE();

  switch(irpSp->MinorFunction) {
  case IRP_MN_REMOVE_DEVICE:
    status = VTrcFSRemoveDevice(DeviceObject, Irp);
    break;

  case IRP_MN_DEVICE_USAGE_NOTIFICATION:
    {
      PIO_STACK_LOCATION irpStack;
      ULONG count;
      BOOLEAN setPagable;
      PFS_HOOK_EXTENSION deviceExtension;

      irpStack = IoGetCurrentIrpStackLocation(Irp);

      if (irpStack->Parameters.UsageNotification.Type !=
          DeviceUsageTypePaging) {
        status = VTrcFSPassOnNormally(DeviceObject, Irp);
        break; // out of case statement
      }

      deviceExtension = DeviceObject->DeviceExtension;

      // wait on the paging path event

      status = KeWaitForSingleObject(&deviceExtension->PagingPathCountEvent,
                                     Executive, KernelMode, FALSE, NULL);

      // If removing last paging device, need to set DO_POWER_PAGABLE
      // bit here, and possible re-set it below on failure.

      setPagable = FALSE;
      if (!irpStack->Parameters.UsageNotification.InPath &&
          deviceExtension->PagingPathCount == 1) {
        // removing the last paging file so must have DO_POWER_PAGABLE bits set

        if (!(DeviceObject->Flags & DO_POWER_INRUSH)) {
          DeviceObject->Flags |= DO_POWER_PAGABLE;
          setPagable = TRUE;
        }
      }

      // Send the IRP synchronously.

      status = VTrcFSForwardIrpSynchronous(DeviceObject, Irp);

      // now deal with the failure and success cases.
      // note that we are not allowed to fail the irp
      // once it is sent to the lower drivers.

      if (NT_SUCCESS(status)) {
        IoAdjustPagingPathCount(&deviceExtension->PagingPathCount,
                                irpStack->Parameters.UsageNotification.InPath);
        if (irpStack->Parameters.UsageNotification.InPath &&
            deviceExtension->PagingPathCount == 1) {
          // first paging file addition
          DeviceObject->Flags &= ~DO_POWER_PAGABLE;
        }
      }
      else {
        // cleanup the changes done above
        if (setPagable == TRUE) {
          DeviceObject->Flags &= ~DO_POWER_PAGABLE;
          setPagable = FALSE;
        }
      }

      // Set the event so the next one can occur.

      KeSetEvent(&deviceExtension->PagingPathCountEvent,
                 IO_NO_INCREMENT, FALSE);

      // ...and complete the irp

      IoCompleteRequest(Irp, IO_NO_INCREMENT);
    } // end case IRP_MN_DEVICE_USAGE_NOTIFICATION
    break;

  default:
    // Simply forward all other Irps
    status = VTrcFSPassOnNormally(DeviceObject, Irp);
    break;
  } // end switch (irpSp->MinorFunction)

  return status;
}

NTSTATUS VTrcFSDispatchPower (PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
  PFS_HOOK_EXTENSION deviceExtension;

  PoStartNextPowerIrp(Irp);
  IoSkipCurrentIrpStackLocation(Irp);

  deviceExtension = (PFS_HOOK_EXTENSION) DeviceObject->DeviceExtension;
  return PoCallDriver(deviceExtension->attachedDevice, Irp);
}
