//
// VTrcFsrc.c
//
// Code for the VTrcFsrc driver, a file system recognizer.
//

#include "NTDDK.h"
#include "stdarg.h"
#include "stdio.h"
#include "ntdddisk.h"
#include <ntddvol.h>
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "GetState.h"

//////////////////////////////////////////////////////////////////////////
//                               MACROS
//////////////////////////////////////////////////////////////////////////

// print macro that only turns on when checked builds are on
#if DBG
#define DebugPrint(arg) DbgPrint arg
#else
#define DebugPrint(arg)
#endif

//////////////////////////////////////////////////////////////////////////
//                              GLOBALS
//////////////////////////////////////////////////////////////////////////

PDRIVER_OBJECT thisDriver;

//////////////////////////////////////////////////////////////////////////
//                            STRUCTURES
//////////////////////////////////////////////////////////////////////////

// The structure FSRC_HOOK_EXTENSION forms the device extension of any
// file system recognizer device associated with the VTrcFsrc driver. It
// contains the device attached to.

typedef struct _fsrc_hook_extension {
  PDEVICE_OBJECT attachedDevice;
} FSRC_HOOK_EXTENSION, *PFSRC_HOOK_EXTENSION;         

//////////////////////////////////////////////////////////////////////////
//                          IFS KIT DEFINITIONS
//////////////////////////////////////////////////////////////////////////

typedef
VOID
(*PDRIVER_FS_NOTIFICATION) (
    IN struct _DEVICE_OBJECT *DeviceObject,
    IN BOOLEAN FsActive
    );

NTKERNELAPI
NTSTATUS
IoRegisterFsRegistrationChange(
    IN PDRIVER_OBJECT DriverObject,
    IN PDRIVER_FS_NOTIFICATION DriverNotificationRoutine
    );

//////////////////////////////////////////////////////////////////////////
//                          FORWARD DEFINES
//////////////////////////////////////////////////////////////////////////

NTSTATUS DriverEntry (IN PDRIVER_OBJECT DriverObject,
                      IN PUNICODE_STRING RegistryPath);

NTSTATUS VTrcFsrcPassOnNormally (PDEVICE_OBJECT HookDevice, IN PIRP Irp);

VOID VTrcFsrcFileSystemNotification (PDEVICE_OBJECT DeviceObject,
                                     BOOLEAN FsActive);
NTSTATUS VTrcFsrcDispatchFSControl (IN PDEVICE_OBJECT DeviceObject,
                                      IN PIRP Irp);
NTSTATUS VTrcFsrcDispatchPnp (IN PDEVICE_OBJECT devObj, IN PIRP irp);
NTSTATUS VTrcFsrcDispatchPower (IN PDEVICE_OBJECT devObj, IN PIRP irp);

void FilterFileSystem (PVPB vpb);

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, DriverEntry)
#endif

//////////////////////////////////////////////////////////////////////////
//                          INITIALIZATION
//////////////////////////////////////////////////////////////////////////

// DriverEntry is the routine called by the system to initialize the disk
// performance driver. The driver object is set up and then the
// driver calls MyInitialize to attach to the boot devices.

NTSTATUS DriverEntry (IN PDRIVER_OBJECT DriverObject,
                      IN PUNICODE_STRING RegistryPath)
{
  NTSTATUS status;
  UNICODE_STRING deviceNameUnicodeString;
  PDEVICE_OBJECT deviceObject;
  PFILE_OBJECT fileObject;
  PFSRC_HOOK_EXTENSION deviceExtension;
  int i;

  DebugPrint(("VTrcFsrc: Entering DriverEntry\n"));

  // Get a pointer to the shared state from VTrcInit. If we can't, we fail.

  status = GetSharedState();
  if (!NT_SUCCESS(status)) {
    DebugPrint(("VTrcFsrc: Did not get shared state, so exiting.\n"));
    return status;
  }

  // Set up the device driver entry points.

  for (i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++)
    DriverObject->MajorFunction[i] = VTrcFsrcPassOnNormally;
  DriverObject->MajorFunction[IRP_MJ_FILE_SYSTEM_CONTROL] =
                                                 VTrcFsrcDispatchFSControl;
  DriverObject->MajorFunction[IRP_MJ_PNP] = VTrcFsrcDispatchPnp;
  DriverObject->MajorFunction[IRP_MJ_POWER] = VTrcFsrcDispatchPower;

  DriverObject->FastIoDispatch = NULL;

  // Copy a pointer to the driver object to the global 'thisDriver' for
  // use in later hooking.

  thisDriver = DriverObject;

  // Create a device for use in getting notified of file systems coming and
  // going.

  status = IoCreateDevice(DriverObject, sizeof(FSRC_HOOK_EXTENSION),
                          NULL, FILE_DEVICE_FILE_SYSTEM, 0, FALSE,
                          &deviceObject);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("VTrcFsrc: Failed to create main device.\n"));
    return status;
  }
  deviceExtension = (PFSRC_HOOK_EXTENSION) deviceObject->DeviceExtension;
  deviceExtension->attachedDevice = NULL;

  // Register this driver to receive notifications of file systems coming
  // and going.

  status = IoRegisterFsRegistrationChange(DriverObject,
                                          VTrcFsrcFileSystemNotification);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("VTrcFsrc Error: IoRegisterFsRegistrationChange failed, "
                "returning %x.\n", status));
    return status;
  }

  DebugPrint(("VTrcFsrc has successfully registered for file system info.\n"));

  // Attempt to open the RAW device object since this driver has already
  // been started by system initialization and will not make it through
  // the normal file system notification procedures otherwise. Then,
  // simulate being notified of this device registering itself as a
  // file system by calling VTrcFsrcFileSystemNotification.

  RtlInitUnicodeString(&deviceNameUnicodeString, L"\\Device\\RawDisk" );
  status = IoGetDeviceObjectPointer(&deviceNameUnicodeString,
                                    FILE_READ_ATTRIBUTES, &fileObject,
                                    &deviceObject);
  DebugPrint(("VTrcFsrc has discovered that \\Device\\RawDisk is at %x.\n",
              deviceObject));
  if (NT_SUCCESS(status)) {
    VTrcFsrcFileSystemNotification(deviceObject, TRUE);
    ObDereferenceObject(fileObject);
  }

  // Return success.

  return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////
//                       REQUEST COMPLETION ROUTINES
////////////////////////////////////////////////////////////////////////////

// VTrcFsrcMountCompletionRoutine is our completion routine for successful
// mount requests. It creates a new VTrcFS device to filter the file system
// that just got mounted. The context passed to this routine is the "real
// device" of the VPB involved in the mount request.

NTSTATUS VTrcFsrcMountCompletionRoutine (PDEVICE_OBJECT DeviceObject, PIRP Irp,
                                         PVOID Context)
{
  PDEVICE_OBJECT  realDevice = (PDEVICE_OBJECT) Context;

  // Do a paranoid check to make sure the IRP is returning success. If it
  // is, filter the file system.

  if (NT_SUCCESS(Irp->IoStatus.Status))
    FilterFileSystem(realDevice->Vpb);

  // Must do the following in a completion routine.

  if (Irp->PendingReturned) {
    IoMarkIrpPending(Irp);
  }

  return Irp->IoStatus.Status;
}

////////////////////////////////////////////////////////////////////////////
//              FUNCTIONS TO PASS IRP ON TO NEXT LOWER DEVICE
////////////////////////////////////////////////////////////////////////////

// VTrcFsrcPassOnNormally passes on the given IRP to the next lower device
// (also given).

NTSTATUS VTrcFsrcPassOnNormally (PDEVICE_OBJECT HookDevice, IN PIRP Irp)
{
  PFSRC_HOOK_EXTENSION hookExt = HookDevice->DeviceExtension;

  // If there is no device below us, e.g., if this is a request meant
  // for us and us alone, then just return success. This can happen if
  // a request is passed to the main device we create to get notified
  // of file systems coming and going.

  if (hookExt->attachedDevice == NULL) {
    Irp->IoStatus.Information = 0;
    Irp->IoStatus.Status = STATUS_SUCCESS;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return STATUS_SUCCESS;
  }

  // Skip the current stack location.

  Irp->CurrentLocation++;
  Irp->Tail.Overlay.CurrentStackLocation++;

  // Call the next lower driver

  return IoCallDriver(hookExt->attachedDevice, Irp);
}

// VTrcFsrcDispatchFSControl is the dispatch routine called for any file system
// control IRP. Normally, we ignore these, but if they are mount requests we
// set ourselves up to hook the new file system as soon as it becomes mounted.

NTSTATUS VTrcFsrcDispatchFSControl (PDEVICE_OBJECT HookDevice, IN PIRP Irp)
{
  PIO_STACK_LOCATION    currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
  PIO_STACK_LOCATION    nextIrpStack;
  PFSRC_HOOK_EXTENSION  hookExt;

  // If this is a mount request, and there is a device we're attached to,
  // and there's room on the stack for copying the current stack location,
  // then we set up the IRP to call our mount completion routine when it's
  // complete.

  if (currentIrpStack->MinorFunction == IRP_MN_MOUNT_VOLUME) {
    DebugPrint(("VTrcFsrc: Hook device %x received mount request: IRP %x, "
                "devobj %x, VPB %x with real device %x\n",
                HookDevice, Irp,
                currentIrpStack->Parameters.MountVolume.DeviceObject,
                currentIrpStack->Parameters.MountVolume.Vpb->RealDevice));
    hookExt = (PFSRC_HOOK_EXTENSION) HookDevice->DeviceExtension;
    if (hookExt->attachedDevice != NULL &&
        Irp->CurrentLocation >= HookDevice->StackSize) {
      nextIrpStack = IoGetNextIrpStackLocation(Irp);
      RtlMoveMemory(nextIrpStack, currentIrpStack, sizeof(IO_STACK_LOCATION));
      IoSetCompletionRoutine(Irp, VTrcFsrcMountCompletionRoutine,
                             currentIrpStack->Parameters.MountVolume.
                               Vpb->RealDevice,
                             TRUE, FALSE, FALSE);
      return IoCallDriver(hookExt->attachedDevice, Irp);
    }
  }

  return VTrcFsrcPassOnNormally(HookDevice, Irp);
}

////////////////////////////////////////////////////////////////////////////
//                    NOTIFICATION CALLBACK FUNCTIONS
////////////////////////////////////////////////////////////////////////////

// VTrcFsrcFileSystemNotification gets called whenever a device is
// registering or deregistering as a file system. The FsActive parameter
// indicates if it's registering (as opposed to deregistering).

VOID VTrcFsrcFileSystemNotification (PDEVICE_OBJECT DeviceObject,
                                     BOOLEAN FsActive)
{
  NTSTATUS status;
  PDEVICE_OBJECT hookDevice, layeredDevice;
  PFSRC_HOOK_EXTENSION deviceExtension;

  DebugPrint(("VTrcFsrcFileSystemNotification(%x, %d)\n",
              DeviceObject, FsActive ? 1 : 0));

  // We're only concerned about disks, CD's, and tapes, although it would
  // be odd for us to be notified about anything else.

  if (DeviceObject->DeviceType != FILE_DEVICE_DISK_FILE_SYSTEM &&
      DeviceObject->DeviceType != FILE_DEVICE_CD_ROM_FILE_SYSTEM &&
      DeviceObject->DeviceType != FILE_DEVICE_TAPE_FILE_SYSTEM)
    return;

  DebugPrint(("Device has type %d.\n", DeviceObject->DeviceType));

  if (FsActive) {
    // If this device is registering as a file system, we create a new
    // VTrcFsrc device to attach to it. Note that we must *not* use the
    // same device type as the filtered device, or else things like SIS
    // (the single-instance store service in Windows 2000) may assume we
    // are actually one of those file systems and treat us as such.

    status = IoCreateDevice(thisDriver, sizeof(FSRC_HOOK_EXTENSION),
                            NULL, FILE_DEVICE_FILE_SYSTEM, 0, FALSE,
                            &hookDevice);
    if (!NT_SUCCESS(status))
      return;
    DebugPrint(("VTrcFsrc about to hook onto %x with hook device %x.\n",
                DeviceObject, hookDevice));
    deviceExtension = (PFSRC_HOOK_EXTENSION) hookDevice->DeviceExtension;
    deviceExtension->attachedDevice =
      IoAttachDeviceToDeviceStack(hookDevice, DeviceObject);
    if (deviceExtension->attachedDevice) {
      hookDevice->Flags &= ~DO_DEVICE_INITIALIZING;
      DebugPrint(("VTrcFsrc successfully attached to %x.\n",
                  deviceExtension->attachedDevice));
    }
    else {
      DebugPrint(("VTrcFsrc failed to attach to %x.\n", DeviceObject));
      IoDeleteDevice(hookDevice);
    }
  }
  else {
    // If this device is deregistering as a file system, we should de-attach
    // ourselves from it. First, we need to find which device belongs to us
    // and is attached.

    for (layeredDevice = DeviceObject->AttachedDevice;
         layeredDevice != NULL;
         layeredDevice = layeredDevice->AttachedDevice) {
      for (hookDevice = thisDriver->DeviceObject;
           hookDevice != NULL;
           hookDevice = hookDevice->NextDevice) {
        if (layeredDevice == hookDevice) {
          DebugPrint(("VTrcFsrc has determined that the device going away is "
                      "attached to by VTrcFsrc device %x.\n", hookDevice));
          IoDetachDevice(DeviceObject);
          deviceExtension = hookDevice->DeviceExtension;
          deviceExtension->attachedDevice = NULL;
          // We should delete our device, but if someone's attached to it,
          // that could be unwise.
          if (!hookDevice->AttachedDevice) {
            IoDeleteDevice(hookDevice);
          }
        }
      }
    }
  }
}

////////////////////////////////////////////////////////////////////////////
//                       CREATING VTRCFS DEVICES
////////////////////////////////////////////////////////////////////////////

// LogSuccessfulHook logs the successful attachment of a new VTrcFS device
// to a file system device. The parameters are the device being attached to,
// the disk number and partition number of that device (each of which may
// be set to UNKNOWN_DEVICE_NUMBER).

VOID LogSuccessfulHook (PDEVICE_OBJECT targetDevice, unsigned long diskNumber,
                        unsigned long partitionNumber)
{
  KIRQL               oldirql;
  int                 i;
  USHORT              nameLength;
  PUNICODE_STRING     driverString;
  NTSTATUS            status;
  PCHAR               eventBuffer;

  DebugPrint(("Logging hook of %x (disk %d, part %d).\n",
              targetDevice, diskNumber, partitionNumber));

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);

  // Copy the driver name string to the event after checking to make
  // sure it's a valid address.  This copy is from Unicode to ASCII.

  driverString = &targetDevice->DriverObject->DriverName;
  if (MmIsAddressValid(driverString) &&
      MmIsAddressValid(driverString->Buffer)) {
    nameLength = driverString->Length / 2; // Divide by 2 since WCHAR=2 bytes
    if (nameLength > 128)
      nameLength = 128;
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_DEVICE_HOOK, nameLength + 12);
    if (eventBuffer) {
      eventBuffer[1] = (char) (nameLength + 12);
      for (i = 0; i < nameLength; i++)
        eventBuffer[i + 12] = * (char *) &driverString->Buffer[i];
    }
  }
  else {
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_DEVICE_HOOK, 15);
    if (eventBuffer) {
      eventBuffer[1] = (char) 15;
      RtlCopyMemory(&eventBuffer[12], "???", 3);
    }
  }
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[2], &targetDevice, 4);
    RtlCopyMemory(&eventBuffer[6], &targetDevice->DeviceType, 4);
    eventBuffer[10] = (char) diskNumber;
    eventBuffer[11] = (char) partitionNumber;
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
}

// GetDiskAndPartitionNumbers obtains the disk and partition numbers
// for the given real device and puts them in the appropriate pointer
// locations given.  If it cannot determine the disk and partition
// numbers, it tries to get the volume number and put that as the disk
// number.

void GetDiskAndPartitionNumbers (PDEVICE_OBJECT realDevice,
                                 ULONG *diskNumberPtr,
                                 ULONG *partitionNumberPtr)
{
  PIRP irp;
  STORAGE_DEVICE_NUMBER deviceNumber;
  VOLUME_NUMBER volumeNumber;
  IO_STATUS_BLOCK ioStatusBlock;
  KEVENT event;
  NTSTATUS status;

  DebugPrint(("VTrcFsrc: Getting disk and partition of %x.\n", realDevice));
  
  KeInitializeEvent(&event, NotificationEvent, FALSE);
  irp = IoBuildDeviceIoControlRequest(IOCTL_STORAGE_GET_DEVICE_NUMBER,
                                      realDevice,
                                      NULL, 0, &deviceNumber,
                                      sizeof(deviceNumber), FALSE, &event,
                                      &ioStatusBlock);
  if (irp == NULL) {
    DebugPrint(("VTrcFsrc: IoBuildDeviceIoControlRequest failed.\n"));
    return;
  }

  status = IoCallDriver(realDevice, irp);
  if (status == STATUS_PENDING) {
    KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, NULL);
    status = ioStatusBlock.Status;
  }

  if (NT_SUCCESS(status)) {
    *diskNumberPtr = deviceNumber.DeviceNumber;
    *partitionNumberPtr = deviceNumber.PartitionNumber;
    DebugPrint(("VTrcFsrc: Determined device to be disk %d, partition %d\n",
                deviceNumber.DeviceNumber, deviceNumber.PartitionNumber));
    return;
  }

  DebugPrint(("VTrcFsrc: Getting volume of %x.\n", realDevice));
  
  KeInitializeEvent(&event, NotificationEvent, FALSE);
  irp = IoBuildDeviceIoControlRequest(IOCTL_VOLUME_QUERY_VOLUME_NUMBER,
                                      realDevice,
                                      NULL, 0, &volumeNumber,
                                      sizeof(volumeNumber), FALSE, &event,
                                      &ioStatusBlock);
  if (irp == NULL) {
    DebugPrint(("VTrcFsrc: IoBuildDeviceIoControlRequest failed.\n"));
    return;
  }

  status = IoCallDriver(realDevice, irp);
  if (status == STATUS_PENDING) {
    KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, NULL);
    status = ioStatusBlock.Status;
  }

  if (!NT_SUCCESS(status)) {
    DebugPrint(("VTrcFsrc: Get-volume-number IRP returned status %x.\n",
                status));
    return;
  }

  *diskNumberPtr = volumeNumber.VolumeNumber;

  DebugPrint(("VTrcFsrc: Determined device to be volume %u\n",
              volumeNumber.VolumeNumber));
}

// FilterFileSystem creates and attaches a new VTrcFS device to the file system
// associated with the VPB vpb.

void FilterFileSystem (PVPB vpb)
{
  PDEVICE_OBJECT hookDevice, deviceToAttachTo;
  PFS_HOOK_EXTENSION deviceExtension;
  NTSTATUS status;
  ULONG diskNumber = UNKNOWN_DEVICE_NUMBER;
  ULONG partitionNumber = UNKNOWN_DEVICE_NUMBER;

  // If we don't know where the VTrcFS driver is, we can't do this.

  if (sharedState->fsDriverObject == NULL)
    return;

  // If the file system doesn't have fast I/O dispatch routines, we shouldn't
  // filter it.

  if (vpb->DeviceObject->DriverObject->FastIoDispatch == NULL) {
    DebugPrint(("VTrcFsrc: Choosing not to filter a file system with no "
                "fast dispatch routines.\n"));
    return;
  }

  // Get the disk and partition number information. If this is partition 0
  // of some disk, it's not worth filtering, since that's not a real file
  // system.

  GetDiskAndPartitionNumbers(vpb->RealDevice, &diskNumber, &partitionNumber);
  if (partitionNumber == 0) {
    DebugPrint(("VTrcFsrc: Choosing not to filter a 0th partition.\n"));
    return;
  }

  DebugPrint(("VTrcFS: Filtering file system for vpb %x.\n", vpb));

  // Create a new VTrcFS device to attach to the file system.

  status = IoCreateDevice(sharedState->fsDriverObject,
                          sizeof(FS_HOOK_EXTENSION), NULL,
                          vpb->DeviceObject->DeviceType, 0, FALSE,
                          &hookDevice);
  if (!NT_SUCCESS(status))
    return;

  // Determine the highest-level attached device and fill in the device
  // extension and flags appropriately for the new VTrcFS device.

  deviceExtension = (PFS_HOOK_EXTENSION) hookDevice->DeviceExtension;
  RtlZeroMemory(deviceExtension, sizeof(FS_HOOK_EXTENSION));
  deviceExtension->diskNumber = diskNumber;
  deviceExtension->partitionNumber = partitionNumber;
  KeInitializeEvent(&deviceExtension->PagingPathCountEvent, NotificationEvent,
                    TRUE);

  deviceExtension->attachedDevice =
    IoAttachDeviceToDeviceStack(hookDevice, vpb->DeviceObject);
  if (deviceExtension->attachedDevice) {
    DebugPrint(("VTrcFsrc: Attached new hook device %x to device %x.\n",
                hookDevice, deviceExtension->attachedDevice));
    if (deviceExtension->attachedDevice->Flags & DO_BUFFERED_IO)
      hookDevice->Flags |= DO_BUFFERED_IO;
    if (deviceExtension->attachedDevice->Flags & DO_DIRECT_IO)
      hookDevice->Flags |= DO_DIRECT_IO;
    hookDevice->Flags &= ~DO_DEVICE_INITIALIZING;
    LogSuccessfulHook(vpb->DeviceObject, diskNumber, partitionNumber);
  }
  else {
    DebugPrint(("VTrcFsrc: Failed to attach hook device %x to device %x.\n",
                hookDevice, vpb->DeviceObject));
    IoDeleteDevice(hookDevice);
  }
}

// VTrcFsrcDeviceDispatchPnp is called to dispatch PnP IRP's.

NTSTATUS VTrcFsrcDispatchPnp (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
  PFSRC_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PIO_STACK_LOCATION   currentIrpStack = IoGetCurrentIrpStackLocation(Irp);

  // If there is no device below us, e.g., if this is a request meant
  // for us and us alone, then just return success. This can happen if
  // a request is passed to the main device we create to get notified
  // of file systems coming and going.

  if (hookExt->attachedDevice == NULL) {
    Irp->IoStatus.Information = 0;
    Irp->IoStatus.Status = STATUS_SUCCESS;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return STATUS_SUCCESS;
  }

  if (currentIrpStack->MinorFunction == IRP_MN_REMOVE_DEVICE) {
    IoSkipCurrentIrpStackLocation(Irp);
    IoCallDriver(hookExt->attachedDevice, Irp);
    IoDetachDevice(hookExt->attachedDevice);
    IoDeleteDevice(DeviceObject);
    return STATUS_SUCCESS;
  }
  else {
    IoSkipCurrentIrpStackLocation(Irp);
    return IoCallDriver(hookExt->attachedDevice, Irp);
  }
}

// VTrcFsrcDeviceDispatchPower is called to dispatch power IRP's.

NTSTATUS VTrcFsrcDispatchPower (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
  PFSRC_HOOK_EXTENSION hookExt = DeviceObject->DeviceExtension;
  PIO_STACK_LOCATION   currentIrpStack = IoGetCurrentIrpStackLocation(Irp);

  // If there is no device below us, e.g., if this is a request meant
  // for us and us alone, then just return success. This can happen if
  // a request is passed to the main device we create to get notified
  // of file systems coming and going.

  if (hookExt->attachedDevice == NULL) {
    Irp->IoStatus.Information = 0;
    Irp->IoStatus.Status = STATUS_SUCCESS;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return STATUS_SUCCESS;
  }

  PoStartNextPowerIrp(Irp);
  IoSkipCurrentIrpStackLocation(Irp);
  return PoCallDriver(hookExt->attachedDevice, Irp);
}
