// 
// VTrcInit.c
//

#include "NTDDK.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "../Common/Version.h"
#include "Activate.h"

//////////////////////////////////////////////////////////////////////////
//                               MACROS
//////////////////////////////////////////////////////////////////////////

// print macro that only turns on when checked builds are on
#if DBG
#define DebugPrint(arg) DbgPrint arg
#else
#define DebugPrint(arg)
#endif

////////////////////////////////////////////////////////////////////////////
//                               GLOBALS
////////////////////////////////////////////////////////////////////////////

SHARED_STATE *sharedState;

////////////////////////////////////////////////////////////////////////////
//                           FORWARD DEFINES
////////////////////////////////////////////////////////////////////////////

NTSTATUS DriverEntry (IN PDRIVER_OBJECT  DriverObject,
                      IN PUNICODE_STRING RegistryPath);
char *DummyLogEvent (char eventType, ULONG eventLength);

NTSTATUS VTrcInitDispatchDeviceControl (IN PDEVICE_OBJECT devObj, IN PIRP irp);
NTSTATUS VTrcInitDispatchCreate (IN PDEVICE_OBJECT devObj, IN PIRP irp);
NTSTATUS VTrcInitDispatchClose (IN PDEVICE_OBJECT devObj, IN PIRP irp);
NTSTATUS VTrcInitDispatchGeneral (IN PDEVICE_OBJECT devObj, IN PIRP irp);

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, DriverEntry)
#endif

////////////////////////////////////////////////////////////////////////////
//                             DRIVER ENTRY
////////////////////////////////////////////////////////////////////////////

// DriverEntry is the installable driver initialization.  Here we set
// up the driver, create our one logger device, and do all sorts of
// miscellaneous initialization.

NTSTATUS DriverEntry (IN PDRIVER_OBJECT  DriverObject,
                      IN PUNICODE_STRING RegistryPath)
{
  NTSTATUS status = STATUS_SUCCESS;
  UNICODE_STRING deviceNameUnicodeString, linkNameUnicodeString;
  PDEVICE_OBJECT deviceObject;
  int i;

  DebugPrint(("VTrcInit: Entering DriverEntry\n"));

  // Determine if we should activate VTrace. If not, return failure.

  if (!ShouldActivateVTrace())
    return STATUS_UNSUCCESSFUL;

  // Allocate space for shared state. If we can't, we fail.

  sharedState = ExAllocatePool(NonPagedPool, sizeof(SHARED_STATE));
  if (sharedState == NULL)
    return STATUS_UNSUCCESSFUL;

  // Initialize driver-shared state fields

  KeInitializeSpinLock(&sharedState->mainMutex);
  sharedState->logEventFunc = &DummyLogEvent;
  sharedState->checkIfLoggingFunc = NULL;
  sharedState->emptyFSHashFunc = NULL;
  sharedState->userActivityFunc = NULL;
#if TARGET_NT_VERSION == 4
  sharedState->rawDriverObject = NULL;
#else
  sharedState->hibernationNotifyFunc = NULL;
#endif
  sharedState->fsDriverObject = NULL;
  sharedState->netDriverObject = NULL;
  sharedState->filteredCreates = 0;
  sharedState->kernelUserSharedState.maxThreads = MAX_ACTIVITY_THREADS;

  // Set up the device driver entry points.

  for (i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++)
    DriverObject->MajorFunction[i] = VTrcInitDispatchGeneral;
  DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] =
                                     VTrcInitDispatchDeviceControl;
  DriverObject->MajorFunction[IRP_MJ_CREATE] = VTrcInitDispatchCreate;
  DriverObject->MajorFunction[IRP_MJ_CLOSE] = VTrcInitDispatchClose;

  // Create the one and only VTrcInit device.

  RtlInitUnicodeString(&deviceNameUnicodeString, L"\\Device\\VTrcInit");
  status = IoCreateDevice(DriverObject, 0, &deviceNameUnicodeString,
                          FILE_DEVICE_VTRACE_INIT, 0, FALSE,
                          &deviceObject);
  if (!NT_SUCCESS(status))
    return status;

  // Create a link to this device so it can be accessed from user level

  RtlInitUnicodeString(&linkNameUnicodeString, L"\\DosDevices\\VTrcInit");
  status = IoCreateSymbolicLink(&linkNameUnicodeString,
                                &deviceNameUnicodeString);
  if (!NT_SUCCESS(status)) {
    IoDeleteDevice(deviceObject);
    return status;
  }

  return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////
//                          DISPATCH ROUTINES
////////////////////////////////////////////////////////////////////////////

// VTrcInitDispatchDeviceControl is the dispatch routine called to handle any
// device control IRP sent to our VTrcInit device. The main thing to
// do here is check what kind of device control request it is and handle
// it accordingly.

NTSTATUS VTrcInitDispatchDeviceControl (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  PIO_STACK_LOCATION     irpStack = IoGetCurrentIrpStackLocation(irp);
  PCHAR                  inBuf, outBuf;
  ULONG                  inBufLength, outBufLength, ioControlCode;
  NTSTATUS               status;
  KIRQL                  oldirql;
  PCHAR                  eventBuffer;

  // Set the request up as successful and returning no information, by default

  status = STATUS_SUCCESS;
  irp->IoStatus.Information = 0;

  // Get the control code (what kind of request this is)

  ioControlCode = irpStack->Parameters.DeviceIoControl.IoControlCode;

  // Get the pointer to the input/output buffer and its length.
  // These buffers are in different places depending on the method
  // used (the low two bits of ioControlCode).

  if ((ioControlCode & 0x3) == METHOD_NEITHER) {
    inBuf = irpStack->Parameters.DeviceIoControl.Type3InputBuffer;
    outBuf = irp->UserBuffer;
  }
  else {
    inBuf = irp->AssociatedIrp.SystemBuffer;
    outBuf = irp->AssociatedIrp.SystemBuffer;
  }
  inBufLength = irpStack->Parameters.DeviceIoControl.InputBufferLength;
  outBufLength = irpStack->Parameters.DeviceIoControl.OutputBufferLength;

  // Do the appropriate thing, depending on the control code

  switch (ioControlCode) {
  case VTRACE_CTRL_INDICATE_USER_ACTIVITY :
    // Indicate to the logger device that user activity just occurred, using
    // a direct function call

    if (sharedState->userActivityFunc)
      (*sharedState->userActivityFunc)();
    break;

  case VTRACE_CTRL_GET_SHARED_STATE:
    // This call is made from kernel level requesting a pointer to the
    // driver-shared state structure.  So, we just copy that pointer to the
    // output buffer

    DebugPrint(("VTrcInit: Responding to get-shared-state request.\n"));
    if (outBufLength < sizeof(SHARED_STATE *))
      status = STATUS_INVALID_PARAMETER;
    else {
      * (SHARED_STATE **) outBuf = sharedState;
      irp->IoStatus.Information = sizeof(SHARED_STATE *);
    }
    break;

  case VTRACE_CTRL_GET_KERNEL_USER_SHARED_STATE :
    if (outBufLength < sizeof(PVOID))
      status = STATUS_INVALID_PARAMETER;
    else {
      PVOID address;
      PMDL mdl;
      PFILE_OBJECT fileObject;

      mdl = IoAllocateMdl(&sharedState->kernelUserSharedState,
                          sizeof(KERNEL_USER_SHARED_STATE),
                          FALSE, FALSE, NULL);
      if (mdl == NULL)
        status = STATUS_INSUFFICIENT_RESOURCES;
      else {
        MmBuildMdlForNonPagedPool(mdl);
        address = MmMapLockedPages(mdl, UserMode);
        if (address == NULL) {
          IoFreeMdl(mdl);
          status = STATUS_UNSUCCESSFUL;
        }
        else {
          fileObject = irpStack->FileObject;
          fileObject->FsContext = address;
          fileObject->FsContext2 = mdl;
          address = (PVOID) ((ULONG)PAGE_ALIGN(address) +
                             MmGetMdlByteOffset(mdl));
          * (PVOID *) outBuf = address;
          irp->IoStatus.Information = sizeof(address);
        }
      }
    }
    break;

  case VTRACE_CTRL_LOG_EVENT :
    // To log an event, copy it to the event buffer and call LogEvent.

    if (inBufLength > MAX_EVENT_LENGTH) {
      DebugPrint(("VTrcInit: Received too-long event from user level\n"));
      status = STATUS_INVALID_PARAMETER;
    }
    else {
      KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
      eventBuffer = LOG_EVENT(inBuf[0], inBufLength);
      if (eventBuffer)
        RtlCopyMemory(&eventBuffer[1], &inBuf[1], inBufLength-1);
      KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    }
    break;
  }

  // Complete the IRP

  irp->IoStatus.Status = status;
  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return status;
}

// VTrcInitDispatchCreate is the dispatch routine called to handle any create
// IRP sent to our VTrcInit device. The main thing to do here is initialize
// the FsContext and FsContext2 fields to NULL so that we can tell at close
// if we've used them.

NTSTATUS VTrcInitDispatchCreate (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  PIO_STACK_LOCATION currentIrpStack = IoGetCurrentIrpStackLocation(irp);
  PFILE_OBJECT fileObject = currentIrpStack->FileObject;

  // Set the context values of the file to NULL so that we can recognize
  // if we've set them.

  if (fileObject != NULL) {
    fileObject->FsContext = NULL;
    fileObject->FsContext2 = NULL;  
  }

  // Set the request up as successful.

  irp->IoStatus.Status = STATUS_SUCCESS;
  irp->IoStatus.Information = 0;
    
  // Complete the IRP.

  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return STATUS_SUCCESS;
}

// VTrcInitDispatchClose is the dispatch routine called to handle any close
// IRP sent to our VTrcInit device. Here we examine the FsContext and
// FsContext2 fields of the file object to determine if we have used this
// file object to map the kernel-user shared state into user space. If we
// have, we now free any resources associated with that mapping.

NTSTATUS VTrcInitDispatchClose (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  PIO_STACK_LOCATION currentIrpStack = IoGetCurrentIrpStackLocation(irp);
  PFILE_OBJECT fileObject = currentIrpStack->FileObject;
  PMDL mdl;
  PVOID address;

  // If we created a mapping for the kernel-user shared state using
  // this file object, we free it now.

  if (fileObject != NULL) {
    address = fileObject->FsContext;
    mdl = fileObject->FsContext2;
    if (address != NULL && mdl != NULL) {
      MmUnmapLockedPages(address, mdl);
      IoFreeMdl(mdl);
    }
  }

  // Set the request up as successful

  irp->IoStatus.Status = STATUS_SUCCESS;
  irp->IoStatus.Information = 0;
    
  // Complete the IRP

  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return STATUS_SUCCESS;
}

// VTrcInitDispatchGeneral is the dispatch routine called to handle any
// miscellaneous IRP sent to our VTrcInit device. We just complete it.

NTSTATUS VTrcInitDispatchGeneral (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  // Set the request up as successful

  irp->IoStatus.Status = STATUS_SUCCESS;
  irp->IoStatus.Information = 0;
    
  // Complete the IRP

  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////
//                         DUMMY SHARED FUNCTIONS
////////////////////////////////////////////////////////////////////////////

// DummyLogEvent is a dummy routine for logging events, used until VTrcLog
// starts up and replaces this function in the shared state structure. All
// it does is return NULL, indicating the event should not be traced.

char *DummyLogEvent (char eventType, ULONG eventLength)
{
  return NULL;
}
