// 
// GetState.c
//
// Code to retrieve shared state from the VTrcInit driver.
//

#include "NTDDK.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "GetState.h"

////////////////////////////////////////////////////////////////////////////
//                          FORWARD DEFINITIONS
////////////////////////////////////////////////////////////////////////////

NTSTATUS GetSharedStateCompletionRoutine (PDEVICE_OBJECT DeviceObject,
                                          PIRP Irp, PVOID Context);

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, GetSharedState)
#pragma alloc_text (INIT, GetSharedStateCompletionRoutine)
#endif

////////////////////////////////////////////////////////////////////////////
//                              GLOBALS
////////////////////////////////////////////////////////////////////////////

SHARED_STATE *sharedState = NULL;  // points to data shared among the drivers

////////////////////////////////////////////////////////////////////////////
//                          CONNECTING TO VTRCINIT
////////////////////////////////////////////////////////////////////////////

// GetSharedState gets the driver-shared state from the VTrcInit driver by
// sending it a device control IRP.

NTSTATUS GetSharedState (void)
{
  PIRP                    irp;
  PIO_STACK_LOCATION      irpStack;
  IO_STATUS_BLOCK         ioStatusBlock;
  KEVENT                  event;
  UNICODE_STRING          deviceNameUnicodeString;
  PFILE_OBJECT            fileObject;
  PDEVICE_OBJECT          deviceObject;
  NTSTATUS                status;

  // Get a device object pointer to the VTraceSt device

  RtlInitUnicodeString(&deviceNameUnicodeString, L"\\Device\\VTrcInit");
  status = IoGetDeviceObjectPointer(&deviceNameUnicodeString,
                                    FILE_ALL_ACCESS, &fileObject,
                                    &deviceObject);
  if (!NT_SUCCESS(status))
    return status;
  ObDereferenceObject(fileObject); // We don't need the file object pointer

  // Allocate an IRP for the device request.  We only need a stack
  // size of 1, since VTraceSt isn't going to pass the IRP on.

  irp = IoAllocateIrp(1, FALSE);
  if (irp == NULL)
    return STATUS_INSUFFICIENT_RESOURCES;

  ///////////////////////////////////////
  // Initialize the fields in the IRP. //
  ///////////////////////////////////////

  // Set the system buffer parameter to the address of sharedState, since
  // that is where we want the driver-shared state put

  irp->UserBuffer = &sharedState;

  // Set the user IO status block to the address of ioStatusBlock so that
  // we can have access to the return value

  irp->UserIosb = &ioStatusBlock;

  // Initialize an event, and make it the user event of the IRP, so that
  // we can know when the IRP is completed.

  KeInitializeEvent(&event, SynchronizationEvent, FALSE);
  irp->UserEvent = &event;

  // Set other miscellaneous IRP info

  irp->Tail.Overlay.Thread = PsGetCurrentThread();
  irp->RequestorMode = KernelMode;

  // Set information in the first stack location indicating what type
  // of request this is, and other info

  irpStack = IoGetNextIrpStackLocation(irp);
  irpStack->MajorFunction = IRP_MJ_DEVICE_CONTROL;
  irpStack->MinorFunction = 0;
  irpStack->Parameters.DeviceIoControl.Type3InputBuffer = NULL;
  irpStack->Parameters.DeviceIoControl.InputBufferLength = 0;
  irpStack->Parameters.DeviceIoControl.OutputBufferLength =
                                                       sizeof(SHARED_STATE *);
  irpStack->Parameters.DeviceIoControl.IoControlCode =
                                                 VTRACE_CTRL_GET_SHARED_STATE;
  irpStack->DeviceObject = deviceObject;

  // Set our completion routine so that we're called when the IRP completes

  IoSetCompletionRoutine(irp, GetSharedStateCompletionRoutine, 0, TRUE,
                         TRUE, TRUE);

  // Pass on the IRP

  IoCallDriver(deviceObject, irp);

  // Wait for the IRP to complete

  while (KeWaitForSingleObject(&event, Executive, KernelMode, TRUE, 0) !=
         STATUS_SUCCESS)
    ;

  // Status is returned in ioStatusBlock, so check it to see how the request
  // went

  return ioStatusBlock.Status;
}

// GetSharedStateCompletionRoutine is the completion routine called
// when the GetSharedState IRP completes.  (This IRP gets the direct
// log info from VTrcInit.)  The main job of this completion routine is
// to free the IRP we allocated.

NTSTATUS GetSharedStateCompletionRoutine (PDEVICE_OBJECT DeviceObject,
                                          PIRP Irp, PVOID Context)
{
  // Copy status information to user status buffer

  *Irp->UserIosb = Irp->IoStatus;

  // Set event so that the routine waiting on this IRP can continue

  KeSetEvent(Irp->UserEvent, 0, FALSE);

  // Free the IRP

  IoFreeIrp(Irp);

  // Return STATUS_MORE_PROCESSING_REQUIRED to indicate that this driver
  // is taking control of the IRP.  This is necessary because the IRP has
  // just been freed, so we don't want anyone else trying to work on it.
  // Indeed, we have just finished processing the IRP by freeing it.

  return STATUS_MORE_PROCESSING_REQUIRED;
}
