// 
// VTrcKbd.c
//
// Code for the VTrcKbd driver, a filter driver for the keyboard.
// It logs its findings by calling the LogEvent routine in the VTrcLog
// driver.
//

#include "NTDDK.h"
#include "../Common/W2Kddk.h"
#include "../Common/NTDDKBD.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "GetState.h"

////////////////////////////////////////////////////////////////////////////
//                                MACROS
////////////////////////////////////////////////////////////////////////////

// Print macro that only turns on when checked builds are on

#if DBG
#define DebugPrint(arg) DbgPrint arg
#else
#define DebugPrint(arg)
#endif

// The following bit fields are used to keep track of which modifier
// keys are currently held down.

#define LEFT_CTRL_MODIFIER_PRESSED      1
#define RIGHT_CTRL_MODIFIER_PRESSED     2
#define LEFT_ALT_MODIFIER_PRESSED       4
#define RIGHT_ALT_MODIFIER_PRESSED      8

////////////////////////////////////////////////////////////////////////////
//                              STRUCTURES
////////////////////////////////////////////////////////////////////////////

// The structure KBD_HOOK_EXTENSION forms the device extension of any
// keyboard device associated with the VTrcKbd driver. It contains the
// device attached to and the modifiers currently pressed on the filtered
// keyboard.

typedef struct _kbd_hook_extension {
  PDEVICE_OBJECT attachedDevice;
  ULONG modifiersPressed;
} KBD_HOOK_EXTENSION, *PKBD_HOOK_EXTENSION;         

////////////////////////////////////////////////////////////////////////////
//                             FORWARD DEFINES
////////////////////////////////////////////////////////////////////////////

NTSTATUS DriverEntry (IN PDRIVER_OBJECT  DriverObject,
                      IN PUNICODE_STRING RegistryPath);

NTSTATUS VTrcKbdDispatchRead (IN PDEVICE_OBJECT devObj, IN PIRP irp);
NTSTATUS VTrcKbdDispatchPnp (IN PDEVICE_OBJECT devObj, IN PIRP irp);
NTSTATUS VTrcKbdDispatchPower (IN PDEVICE_OBJECT devObj, IN PIRP irp);
NTSTATUS VTrcKbdDispatchGeneral (IN PDEVICE_OBJECT devObj, IN PIRP irp);
USHORT Encrypt (USHORT code, USHORT flags, ULONG *modifiersPressedPtr);

NTSTATUS VTrcKbdAddDevice (IN PDRIVER_OBJECT driverObject,
                           IN PDEVICE_OBJECT physicalDeviceObject);
void VTrcKbdUnload (IN PDRIVER_OBJECT driverObject);

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, DriverEntry)
#pragma alloc_text (PAGE, VTrcKbdDispatchGeneral)
#pragma alloc_text (PAGE, VTrcKbdDispatchPower)
#pragma alloc_text (PAGE, VTrcKbdDispatchPnp)
#pragma alloc_text (PAGE, VTrcKbdAddDevice)
#pragma alloc_text (PAGE, VTrcKbdUnload)
#endif

////////////////////////////////////////////////////////////////////////////
//                             DRIVER ENTRY
////////////////////////////////////////////////////////////////////////////

// DriverEntry is the installable driver initialization.  Here we set
// up the driver and create our one device to attach to the keyboard.

NTSTATUS DriverEntry (IN PDRIVER_OBJECT  DriverObject,
                      IN PUNICODE_STRING RegistryPath)
{
  NTSTATUS  status;
  int       i;

  DebugPrint (("VTrcKbd: Entering DriverEntry\n"));

  // Get a pointer to the shared state from VTrcInit. If we can't, we
  // fail. Since returning an error would be quite bad when we're one
  // of the UpperFilters of the keyboard class, we instead just set an
  // unload routine and return success.

  status = GetSharedState();
  if (!NT_SUCCESS(status)) {
    DriverObject->DriverUnload = VTrcKbdUnload;
    return STATUS_SUCCESS;
  }

  // Create dispatch points for all the IRPs the keyboard class device
  // handles.

  for (i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++)
    DriverObject->MajorFunction[i] = VTrcKbdDispatchGeneral;

  DriverObject->MajorFunction[IRP_MJ_READ] = VTrcKbdDispatchRead;
  DriverObject->MajorFunction[IRP_MJ_PNP] = VTrcKbdDispatchPnp;
  DriverObject->MajorFunction[IRP_MJ_POWER] = VTrcKbdDispatchPower;

  DriverObject->DriverExtension->AddDevice = VTrcKbdAddDevice;
  DriverObject->DriverUnload = VTrcKbdUnload;

  // Indicate success

  return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////
//                              DEVICE ADD
////////////////////////////////////////////////////////////////////////////

NTSTATUS VTrcKbdAddDevice (IN PDRIVER_OBJECT driverObject,
                           IN PDEVICE_OBJECT physicalDeviceObject)
{
  PDEVICE_OBJECT filterDevice;
  PKBD_HOOK_EXTENSION hookExt;
  NTSTATUS status;

  DebugPrint(("VTrcKbd: Adding device on top of %x\n", physicalDeviceObject));

  // Notify VTrcLog that the power state is not hibernating.

  if (sharedState->hibernationNotifyFunc)
    (sharedState->hibernationNotifyFunc)(FALSE);

  // Create the filter device.

  status = IoCreateDevice(driverObject, sizeof(KBD_HOOK_EXTENSION), NULL,
                          FILE_DEVICE_KEYBOARD, 0, FALSE, &filterDevice);
  if (!NT_SUCCESS(status)) {
    DebugPrint(("VTrcKbd: Could not create keyboard device.\n"));
    return status;
  }

  // Keyboard uses buffered I/O, so make our device use it as well.

  filterDevice->Flags |= (DO_BUFFERED_IO | DO_POWER_PAGABLE);

  // Attach keyboard device to existing keyboard device.

  hookExt = filterDevice->DeviceExtension;
  hookExt->modifiersPressed = 0;
  hookExt->attachedDevice =
    IoAttachDeviceToDeviceStack(filterDevice, physicalDeviceObject);
  if (hookExt->attachedDevice == NULL) {
    DebugPrint(("VTrcKbd: Could not attach to keyboard device.\n"));
    IoDeleteDevice(filterDevice);
    return STATUS_UNSUCCESSFUL;
  }

  // Indicate that the keyboard device is finished initializing

  filterDevice->Flags &= ~DO_DEVICE_INITIALIZING;

  return status;
}

////////////////////////////////////////////////////////////////////////////
//                              UNLOAD
////////////////////////////////////////////////////////////////////////////

void VTrcKbdUnload (IN PDRIVER_OBJECT driverObject)
{
  DebugPrint(("VTrcKbd: Unloading\n"));
  if (driverObject->DeviceObject != NULL) {
    DebugPrint(("VTrcKbd: Unloading when a device still exists.\n"));
  }
}

////////////////////////////////////////////////////////////////////////////
//                          DISPATCH ROUTINES
////////////////////////////////////////////////////////////////////////////

// VTrcKbdReadComplete gets control after a keyboard read operation has
// completed so that we can log it.

NTSTATUS VTrcKbdReadComplete (IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp,
                              IN PVOID Context)
{
  PKEYBOARD_INPUT_DATA   keyData = Irp->AssociatedIrp.SystemBuffer;
  PKBD_HOOK_EXTENSION    hookExt = DeviceObject->DeviceExtension;
  int                    numKeys, i;
  KIRQL                  oldirql;
  USHORT                 encryptedCode;
  PCHAR                  eventBuffer;

  numKeys = Irp->IoStatus.Information / sizeof(KEYBOARD_INPUT_DATA);

  // Indicate to the logger device that user activity just occurred, using
  // a direct function call.

  if (sharedState->userActivityFunc)
    (*sharedState->userActivityFunc)();

  // There may be multiple keystrokes in this single IRP, so log them all.
  // Encrypt each keystroke before entering it into the log so that the
  // log (and eventually the trace) does not contain sensitive information
  // like passwords.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  for (i = 0; i < numKeys; i++) {
    encryptedCode = Encrypt(keyData[i].MakeCode, keyData[i].Flags,
                            &hookExt->modifiersPressed);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_KEY_PRESS, 5);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &encryptedCode, 2);
      RtlCopyMemory(&eventBuffer[3], &keyData[i].Flags, 2);
    }
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // As in (almost) any completion routine, we have to mark Irp as pending if
  // necessary.  Note the braces around the call to IoMarkIrpPending; they are
  // necessary since IoMarkIrpPending is a macro that expands to multiple
  // statements.

  if (Irp->PendingReturned) {
    IoMarkIrpPending(Irp);
  }

  return Irp->IoStatus.Status;
}

// VTrcKbdDeviceDispatchRead is called to dispatch read IRP's.  We set a
// completion routine to log the read keystrokes.

NTSTATUS VTrcKbdDispatchRead (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  PKBD_HOOK_EXTENSION hookExt = devObj->DeviceExtension;

  // Push params down for keyboard class driver.

  IoCopyCurrentIrpStackLocationToNext(irp);

  // Set the completion callback so that we can log the returned keystrokes.
  // We only log successful keystrokes, so the cancel and failure completion
  // routine fields are FALSE.

  IoSetCompletionRoutine(irp, VTrcKbdReadComplete, devObj, TRUE, FALSE,
                         FALSE);

  // Return the results of the call to the keyboard class driver.
  // In other words, pass the IRP down to the real keyboard device.

  return IoCallDriver(hookExt->attachedDevice, irp);
}

// VTrcKbdDeviceDispatchPnp is called to dispatch PnP IRP's.

NTSTATUS VTrcKbdDispatchPnp (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  PKBD_HOOK_EXTENSION hookExt = devObj->DeviceExtension;
  PIO_STACK_LOCATION  currentIrpStack = IoGetCurrentIrpStackLocation(irp);

  if (currentIrpStack->MinorFunction == IRP_MN_REMOVE_DEVICE) {
    IoSkipCurrentIrpStackLocation(irp);
    IoCallDriver(hookExt->attachedDevice, irp);
    IoDetachDevice(hookExt->attachedDevice);
    IoDeleteDevice(devObj);
    return STATUS_SUCCESS;
  }
  else {
    IoSkipCurrentIrpStackLocation(irp);
    return IoCallDriver(hookExt->attachedDevice, irp);
  }
}

// VTrcKbdDeviceDispatchPower is called to dispatch power IRP's.

NTSTATUS VTrcKbdDispatchPower (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  PKBD_HOOK_EXTENSION hookExt = devObj->DeviceExtension;
  PIO_STACK_LOCATION  currentIrpStack = IoGetCurrentIrpStackLocation(irp);
  SYSTEM_POWER_STATE  state;

  // If the system power state is changing, notify VTrcLog whether we're
  // now hibernating.

  if (currentIrpStack->MinorFunction == IRP_MN_SET_POWER &&
      currentIrpStack->Parameters.Power.Type == SystemPowerState &&
      sharedState->hibernationNotifyFunc) {
    state = currentIrpStack->Parameters.Power.State.SystemState;
    (sharedState->hibernationNotifyFunc)(state == PowerSystemHibernate);
  }

  PoStartNextPowerIrp(irp);
  IoSkipCurrentIrpStackLocation(irp);
  return PoCallDriver(hookExt->attachedDevice, irp);
}
  
// VTrcKbdDeviceDispatchGeneral is called to dispatch any type of IRP we are
// not interested in logging.

NTSTATUS VTrcKbdDispatchGeneral (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  PKBD_HOOK_EXTENSION hookExt = devObj->DeviceExtension;

  // Skip the current stack location

  IoSkipCurrentIrpStackLocation(irp);

  // Call the next lower driver

  return IoCallDriver(hookExt->attachedDevice, irp);
}

////////////////////////////////////////////////////////////////////////////
//                           ENCRYPTION ROUTINES
////////////////////////////////////////////////////////////////////////////

// Encrypt encrypts a keystroke.  The code (scan code) and flags are
// passed, and a new, encrypted code is returned.  Encryption consists
// of returning the code for 'x' instead of any letter and the code for
// '9' instead of any number, unless Ctrl or Alt is down at the time.
// This function also keeps track of which modifiers are down at any time;
// in other words, when called upon to encrypt a Ctrl or Alt keystroke,
// it doesn't actually encrypt it, but it does remember for later (through
// the variable pointed to by modifiersPressedPtr) that it was pressed or
// released.

USHORT Encrypt (USHORT code, USHORT flags, ULONG *modifiersPressedPtr)
{
  // Instead of a switch statement, which might be inefficient, we
  // use a kind of binary search for the code.  This code is pretty
  // hoary, because the letters and numbers are in dispersed clumps in
  // virtual key code list, instead of all together as in the ASCII
  // code list.

  if (code <= 38) { // code <= 38
    if (code <= 25) { // code <= 25
      if (code <= 11) { // code <= 11
        if (code <= 1 || *modifiersPressedPtr)
          return code;
        else // code represents an unmodified digit 0-9
          return 10; // representing '9'
      }
      else { // 11 < code <= 25
        if (code <= 15 || *modifiersPressedPtr)
          return code;
        else // code represents an unmodified letter among qwertyuiop
          return 45; // representing 'x'
      }
    }
    else { // 25 < code <= 38
      if (code <= 29) { // 25 < code <= 29
        if (code == 29) { // code represents the control key
          switch (flags) {
          case KEY_MAKE : // left control pressed
            *modifiersPressedPtr |= LEFT_CTRL_MODIFIER_PRESSED;
            break;
          case KEY_E0 : // right control pressed
            *modifiersPressedPtr |= RIGHT_CTRL_MODIFIER_PRESSED;
            break;
          case KEY_BREAK : // left control released
            *modifiersPressedPtr &= ~LEFT_CTRL_MODIFIER_PRESSED;
            break;
          case KEY_E0 | KEY_BREAK : // right control released
            *modifiersPressedPtr &= ~RIGHT_CTRL_MODIFIER_PRESSED;
            break;
          }
        }
        return code;
      }
      else { // 29 < code <= 38
        if (*modifiersPressedPtr)
          return code;
        else // code represents an unmodified letter among asdfghjkl 
          return 45; // representing 'x'
      }
    }
  }
  else { // code > 38
    if (code <= 56) { // 38 < code <= 56
      if (code <= 50) { // 38 < code <= 50
        if (code <= 43 || *modifiersPressedPtr)
          return code;
        else // code represents an unmodified letter among zxcvbnm
          return 45; // representing 'x'
      }
      else { // 50 < code <= 56
        if (code == 56) { // code represents the alt key
          switch (flags) {
          case KEY_MAKE : // left alt pressed
            *modifiersPressedPtr |= LEFT_ALT_MODIFIER_PRESSED;
            break;
          case KEY_E0 : // right alt pressed
            *modifiersPressedPtr |= RIGHT_ALT_MODIFIER_PRESSED;
            break;
          case KEY_BREAK : // left alt released
            *modifiersPressedPtr &= ~LEFT_ALT_MODIFIER_PRESSED;
            break;
          case KEY_E0 | KEY_BREAK : // right alt released
            *modifiersPressedPtr &= ~RIGHT_ALT_MODIFIER_PRESSED;
            break;
          }
        }
        return code;
      }
    }
    else { // 56 < code
      if (code >= 71 && code <= 82 && code != 74 && code != 78 &&
          !(flags & KEY_E0)) // code represents an unmodified keypad digit
        return 73; // representing 'keypad 9'
      else
        return code;
    }
  }
}
