//
// Activity.c
//
// Code for activity-related routines in the VTrcLog driver. Activity
// is of two types:  thread activity and user activity.
//
// The thread activity routines keep track of whether activity was
// recently performed by each thread. These routines are not perfect,
// for performance reasons. (Some of them get called really frequently,
// after all.) However, they are set up so that if they guess wrong,
// they err on the side of saying there was activity when there wasn't,
// rather than the other way around. This way, trace compaction may
// be not as effective as possible, but it never compacts away useful
// information.
//
// The user activity routines watch for ten-minute periods of user
// inactivity. They ensure that the bit BECAUSE_USER_IDLE in
// globals.whyLoggingDisabled is set if and only if the user has been
// idle (no keystrokes or mouse non-move operations) in the last ten
// minutes.
//

#include "NTDDK.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "Globals.h"
#include "Activity.h"
#include "EventMask.h"
#include "GetState.h"
#include "StartStop.h"

////////////////////////////////////////////////////////////////////////////
//                               MACROS
////////////////////////////////////////////////////////////////////////////

// The following number is -6 billion, which in 100 ns units means
// ten minutes from now.

#define TIME_BEFORE_USER_IDLE        { 0x9A5F4400, -2 }

/* The following is for debugging purposes.
// The following number is -600 million, which in 100 ns units means
// 1 minute from now.

#define TIME_BEFORE_USER_IDLE        { 0xDC3CBA00, -1 }
*/

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, InitializeThreadActivityElements)
#pragma alloc_text (INIT, InitializeUserInactivityWatch)
#endif

////////////////////////////////////////////////////////////////////////////
//                       THREAD ACTIVITY FUNCTIONS
////////////////////////////////////////////////////////////////////////////

// InitializeThreadActivityElements initializes all threads' number of message
// checks since last activity to 0.

void InitializeThreadActivityElements (void)
{
  RtlZeroMemory((PCHAR)
                sharedState->kernelUserSharedState.msgChecksSinceActivity,
                sharedState->kernelUserSharedState.maxThreads);
}

///////////////////////////////////////////////////////////////////////////
//                        USER ACTIVITY FUNCTIONS
///////////////////////////////////////////////////////////////////////////

// UserGoesIdleRoutine is the routine called when the user has been idle
// for more than 10 minutes. It simply adds the "user is idle" event mask
// bit. Its odd parameter format is so that it can be a deferred procedure
// call.

void UserGoesIdleRoutine (IN PKDPC Dpc, IN PVOID DeferredContext,
                          IN PVOID SystemArgument1, IN PVOID SystemArgument2)
{
  ChangeEventMask(EVENT_MASK_BIT_FOR_USER_IDLE_MASK, 0);
}

// IndicateUserActivity is called when the user shows a sign of activity.
// It sets the inactivity timer to 10 minutes from now.  Also, it eliminates
// the user being idle as a reason for not logging.  Also, it sets the
// variable globals.timeOfLastUserActivity to now.

void IndicateUserActivity (void)
{
  KIRQL oldirql;
  LARGE_INTEGER timeBeforeUserIdle = TIME_BEFORE_USER_IDLE;

  // Remove "the user is idle" from the list of event masks.

  ChangeEventMask(0, EVENT_MASK_BIT_FOR_USER_IDLE_MASK);

  // Set time of last user activity to now
  KeQuerySystemTime(&globals.timeOfLastUserActivity);

  // Set the user inactivity timer to 10 minutes from now
  KeSetTimer(&globals.userGoesIdleTimer, timeBeforeUserIdle,
             &globals.userGoesIdleDpc);
}

// InitializeUserInactivityWatch initializes the stuff needed to watch for
// user inactivity, namely the deferred procedure call (DPC) called when the
// user is idle, and the timer that goes off when the user is idle.  It also
// sets that timer initially to go off 10 minutes from now.  It also
// initializes globals.timeOfLastUserActivity to now.

void InitializeUserInactivityWatch (void)
{
  LARGE_INTEGER timeBeforeUserIdle = TIME_BEFORE_USER_IDLE;

  // Set time of last user activity to now
  KeQuerySystemTime(&globals.timeOfLastUserActivity);

  // Initialize DPC and timer, then set timer to go off in 10 minutes
  KeInitializeDpc(&globals.userGoesIdleDpc, &UserGoesIdleRoutine, NULL);
  KeInitializeTimer(&globals.userGoesIdleTimer);
  KeSetTimer(&globals.userGoesIdleTimer, timeBeforeUserIdle,
             &globals.userGoesIdleDpc);
  sharedState->userActivityFunc = &IndicateUserActivity;
}
