//
// EventMask.c
//
// Code dealing with the event mask for the VTrcLog driver. This file
// contains routines for initializing event masks, changing event masks,
// and getting the current event mask.
//

#include "NTDDK.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "Globals.h"
#include "EventMask.h"
#include "GetState.h"
#include "LogEvent.h"
#include "StartStop.h"

////////////////////////////////////////////////////////////////////////////
//                               MACROS
////////////////////////////////////////////////////////////////////////////

// The MaskedOut macro indicates whether the type "type" is masked out
// by the mask "mask" (in other words, whether bit number "type" in
// DWORD array "mask" is set).

#define MaskedOut(mask,type) (((mask)[(type)/8])&(1<<((type)%8)))

// The time to use mask 0 is 90 minutes, and the time to use mask 1 is
// two hours. These times are actually negative, because when you set
// a timer with KeSetTimer, negative times mean relative to the current
// time.

#define TIME_TO_USE_MASK_0         { 0x6D596400, 0xFFFFFFF3 }
#define TIME_TO_USE_MASK_1         { 0x3C773000, 0xFFFFFFEF }

/* When testing, it's useful to make event mask 0 last 90 *seconds*, and
   mask 1 last 2 *minutes*.

#define TIME_TO_USE_MASK_0         { 0xCA5B1700, 0xFFFFFFFF }
#define TIME_TO_USE_MASK_1         { 0xB8797400, 0xFFFFFFFF }
*/

///////////////////////////////////////////////////////////////////////////
//                          FORWARD DEFINITIONS
///////////////////////////////////////////////////////////////////////////

void GetEventMask (char *eventMask);

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, InitializeEventMaskChanging)
#pragma alloc_text (INIT, InitializeEntryTypeInfo)
#endif

///////////////////////////////////////////////////////////////////////////
//                             CHANGING MASK
///////////////////////////////////////////////////////////////////////////

// ChangeEventMask adds and/or removes event masks from the global set
// of event masks. It handles all consequences of such adding or removing.

void ChangeEventMask (char masksToAdd, char masksToRemove)
{
  char newMask, *eventBuffer;
  KIRQL oldirql;

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  newMask = ((globals.currentEventMasks & ~masksToRemove) | masksToAdd);
  if (newMask != globals.currentEventMasks) {
    DebugPrint(("VTrcLog: Changing event mask set from %d to %d.\n",
                globals.currentEventMasks, newMask));
    globals.currentEventMasks = newMask;

    // Log the new mask.

    eventBuffer = LogEvent((char) ENTRY_TYPE_VTRACE_SET_MASK, 33);
    if (eventBuffer) {
      GetEventMask(&eventBuffer[1]);
    }
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
}

///////////////////////////////////////////////////////////////////////////
//                          TIMED CHANGING OF MASK
///////////////////////////////////////////////////////////////////////////

// PeriodicEventMaskChangeRoutine changes the event mask to the next one.
// It needs the strange parameter list because it is used as a deferred
// procedure call. The only parameter that gets used is the DeferredContext,
// which is used to indicate which mask (0 or 1) to change to.

void PeriodicEventMaskChangeRoutine (IN PKDPC Dpc, IN PVOID WhichEventMask,
                                     IN PVOID SystemArgument1,
                                     IN PVOID SystemArgument2)
{
  KIRQL oldirql;
  LARGE_INTEGER timeToUseMask0 = TIME_TO_USE_MASK_0;
  LARGE_INTEGER timeToUseMask1 = TIME_TO_USE_MASK_1;
  ULONG whichMaskToChangeTo = (ULONG) WhichEventMask;

  // Whichever mask must be changed to, we add that mask and remove the
  // other mask. We then set the timer for when this routine will be
  // called again according to how long the new mask should last.

  switch (whichMaskToChangeTo) {
  case 0 :
    ChangeEventMask(EVENT_MASK_BIT_FOR_MASK_0, EVENT_MASK_BIT_FOR_MASK_1);
    KeInitializeDpc(&globals.changeEventMaskDpc,
                    &PeriodicEventMaskChangeRoutine, (PVOID) 1);
    KeSetTimer(&globals.changeEventMaskTimer, timeToUseMask0,
               &globals.changeEventMaskDpc);
    break;
  case 1 :
    ChangeEventMask(EVENT_MASK_BIT_FOR_MASK_1, EVENT_MASK_BIT_FOR_MASK_0);
    KeInitializeDpc(&globals.changeEventMaskDpc,
                    &PeriodicEventMaskChangeRoutine, (PVOID) 0);
    KeSetTimer(&globals.changeEventMaskTimer, timeToUseMask1,
               &globals.changeEventMaskDpc);
    break;
  }
}

// InitializeEventMaskChanging initializes the system for changing the
// event mask. It sets up the DPC and timer for the event-change routine,
// and arranges for the timer to go off for the first time in
// TIME_TO_USE_MASK_0 time from now and change to event mask 1.

void InitializeEventMaskChanging (void)
{
  LARGE_INTEGER timeToUseMask0 = TIME_TO_USE_MASK_0;

  KeInitializeDpc(&globals.changeEventMaskDpc, &PeriodicEventMaskChangeRoutine,
                  (PVOID) 1);
  KeInitializeTimer(&globals.changeEventMaskTimer);
  KeSetTimer(&globals.changeEventMaskTimer, timeToUseMask0,
             &globals.changeEventMaskDpc);
}

///////////////////////////////////////////////////////////////////////////
//                          EVENT TYPE INFORMATION
///////////////////////////////////////////////////////////////////////////

// InitializeEntryTypeInfo initializes all the information for each entry
// type. In other words, it initializes which event types are in each event
// mask.

void InitializeEntryTypeInfo (void)
{
  int i;

  // Clear all the info elements so that nothing is in any event mask.
  // presence in an event mask

  RtlZeroMemory(globals.entryTypeInfo, 256);

  // Set the proper bits for each event type, as follows:
  // EVENT_MASK_BIT_FOR_RECORD_STATE_MASK : this event type shouldn't be used
  //                                   when only recording initial state
  // EVENT_MASK_BIT_FOR_MASK_0 : this event type shouldn't be used when event
  //                             mask #0 is being used
  // EVENT_MASK_BIT_FOR_MASK_1 : this event type shouldn't be used when event
  //                             mask #1 is being used
  // EVENT_MASK_BIT_FOR_UER_IDLE_MASK : this event type shouldn't be used
  //                                    when the user is idle

  globals.entryTypeInfo[ENTRY_TYPE_PROCESS_CREATE] =
  globals.entryTypeInfo[ENTRY_TYPE_PROCESS_DESTROY] =
  globals.entryTypeInfo[ENTRY_TYPE_THREAD_CREATE] =
  globals.entryTypeInfo[ENTRY_TYPE_THREAD_DESTROY] =
  globals.entryTypeInfo[ENTRY_TYPE_PROCESS_SET_PRIORITY] =
  globals.entryTypeInfo[ENTRY_TYPE_THREAD_SET_PRIORITY] =
    (char) 0;
  globals.entryTypeInfo[ENTRY_TYPE_THREAD_SWITCH] =
    (char) (EVENT_MASK_BIT_FOR_RECORD_STATE_MASK |
            EVENT_MASK_BIT_FOR_MASK_1 |
            EVENT_MASK_BIT_FOR_USER_IDLE_MASK);

  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_GET_NOFILTER_CALL] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_GET_FILTER_CALL] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_DISPATCH_CALL] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_TRANSLATE_ACCEL_CALL] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_TRANSLATE_CALL] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_TRANSLATE_MDI_CALL] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_WAIT_CALL] =
    (char) (EVENT_MASK_BIT_FOR_RECORD_STATE_MASK |
            EVENT_MASK_BIT_FOR_MASK_1 |
            EVENT_MASK_BIT_FOR_USER_IDLE_MASK);
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_GET_RETURN] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_PEEK_REMOVE_NOFILTER_CALL] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_PEEK_REMOVE_FILTER_CALL] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_PEEK_NOREMOVE_NOFILTER_CALL] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_PEEK_NOREMOVE_FILTER_CALL] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_PEEK_RETURN] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_DISPATCH_RETURN] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_TRANSLATE_ACCEL_RETURN] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_TRANSLATE_RETURN] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_TRANSLATE_MDI_RETURN] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_SET_TIMER] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_CANCEL_TIMER] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_WAIT_RETURN] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_GET_INPUT_STATE] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_GET_QUEUE_STATUS] =
    (char) (EVENT_MASK_BIT_FOR_RECORD_STATE_MASK |
            EVENT_MASK_BIT_FOR_MASK_1 |
            EVENT_MASK_BIT_FOR_USER_IDLE_MASK);

  globals.entryTypeInfo[ENTRY_TYPE_KEY_PRESS] =
    (char) (EVENT_MASK_BIT_FOR_RECORD_STATE_MASK |
            EVENT_MASK_BIT_FOR_MASK_1 |
            EVENT_MASK_BIT_FOR_USER_IDLE_MASK);

  globals.entryTypeInfo[ENTRY_TYPE_CURSOR_LOAD] =
    (char) (EVENT_MASK_BIT_FOR_RECORD_STATE_MASK |
            EVENT_MASK_BIT_FOR_MASK_1 |
            EVENT_MASK_BIT_FOR_USER_IDLE_MASK);
  globals.entryTypeInfo[ENTRY_TYPE_CURSOR_SET] =
    (char) (EVENT_MASK_BIT_FOR_RECORD_STATE_MASK |
            EVENT_MASK_BIT_FOR_MASK_1 |
            EVENT_MASK_BIT_FOR_USER_IDLE_MASK);

  globals.entryTypeInfo[ENTRY_TYPE_WINDOW_CREATE] =
  globals.entryTypeInfo[ENTRY_TYPE_DIALOG_CREATE] =
    (char) (EVENT_MASK_BIT_FOR_RECORD_STATE_MASK |
            EVENT_MASK_BIT_FOR_MASK_1 |
            EVENT_MASK_BIT_FOR_USER_IDLE_MASK);

  globals.entryTypeInfo[ENTRY_TYPE_FILE_COMPLETE_OPERATION] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_COMPLETE_MDL_OPERATION] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_NAME] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_INFORMATION] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_FCB_INFORMATION] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_OPEN] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_READ] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_WRITE] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_CLOSE] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_QUERY_INFO] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_SET_INFO] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_DIRECTORY_CONTROL] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_OPEN_QUERY_CLOSE] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_RENAME] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_DELETE] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_FLUSH] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_LOCK] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_UNLOCK] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_SET_POSITION] =
  globals.entryTypeInfo[ENTRY_TYPE_FILE_LINK] =
    (char) EVENT_MASK_BIT_FOR_RECORD_STATE_MASK;

  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_POST_WINDOW] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_POST_THREAD] =
  globals.entryTypeInfo[ENTRY_TYPE_MESSAGE_SEND] =
    (char) (EVENT_MASK_BIT_FOR_RECORD_STATE_MASK |
            EVENT_MASK_BIT_FOR_MASK_1 |
            EVENT_MASK_BIT_FOR_USER_IDLE_MASK);

  globals.entryTypeInfo[ENTRY_TYPE_NETOBJ_COMPLETE_OPERATION] =
    (char) (EVENT_MASK_BIT_FOR_RECORD_STATE_MASK |
            EVENT_MASK_BIT_FOR_MASK_1 |
            EVENT_MASK_BIT_FOR_USER_IDLE_MASK);
  globals.entryTypeInfo[ENTRY_TYPE_NETOBJ_OPEN] =
  globals.entryTypeInfo[ENTRY_TYPE_NETOBJ_CLOSE] =
  globals.entryTypeInfo[ENTRY_TYPE_NETOBJ_CONNECT] =
  globals.entryTypeInfo[ENTRY_TYPE_NETOBJ_DISCONNECT] =
  globals.entryTypeInfo[ENTRY_TYPE_NETOBJ_SEND] =
  globals.entryTypeInfo[ENTRY_TYPE_NETOBJ_SEND_DATAGRAM] =
  globals.entryTypeInfo[ENTRY_TYPE_NETOBJ_RECEIVE] =
  globals.entryTypeInfo[ENTRY_TYPE_NETOBJ_RECEIVE_DATAGRAM] =
  globals.entryTypeInfo[ENTRY_TYPE_NETOBJ_LISTEN] =
  globals.entryTypeInfo[ENTRY_TYPE_NETOBJ_ACCEPT] =
  globals.entryTypeInfo[ENTRY_TYPE_NETOBJ_ASSOCIATE_ADDRESS] =
  globals.entryTypeInfo[ENTRY_TYPE_NETOBJ_DISASSOCIATE_ADDRESS] =
    (char) (EVENT_MASK_BIT_FOR_RECORD_STATE_MASK |
            EVENT_MASK_BIT_FOR_MASK_1 |
            EVENT_MASK_BIT_FOR_USER_IDLE_MASK);
  globals.entryTypeInfo[ENTRY_TYPE_NETOBJ_NOTIFY_CONNECT] =
  globals.entryTypeInfo[ENTRY_TYPE_NETOBJ_NOTIFY_DISCONNECT] =
  globals.entryTypeInfo[ENTRY_TYPE_NETOBJ_NOTIFY_RECEIVE] =
  globals.entryTypeInfo[ENTRY_TYPE_NETOBJ_NOTIFY_RECEIVE_DATAGRAM] =
  globals.entryTypeInfo[ENTRY_TYPE_NETOBJ_NOTIFY_RECEIVE_EXPEDITED] =
    (char) (EVENT_MASK_BIT_FOR_RECORD_STATE_MASK |
            EVENT_MASK_BIT_FOR_MASK_1 |
            EVENT_MASK_BIT_FOR_USER_IDLE_MASK);

  globals.entryTypeInfo[ENTRY_TYPE_VTRACE_VERSION] =
  globals.entryTypeInfo[ENTRY_TYPE_LOCAL_TIME] =
  globals.entryTypeInfo[ENTRY_TYPE_VTRACE_SET_MASK] =
  globals.entryTypeInfo[ENTRY_TYPE_HIGH_TIMESTAMP] =
  globals.entryTypeInfo[ENTRY_TYPE_SET_CPU_SPEED] =
  globals.entryTypeInfo[ENTRY_TYPE_VTRACE_ACTION] =
  globals.entryTypeInfo[ENTRY_TYPE_DEVICE_HOOK] =
  globals.entryTypeInfo[ENTRY_TYPE_BEAT] =
  globals.entryTypeInfo[ENTRY_TYPE_DEVICE_UNHOOK] =
  globals.entryTypeInfo[ENTRY_TYPE_USER_CHANGED] =
    (char) 0;

  globals.entryTypeInfo[ENTRY_TYPE_SYSTEM_CALL] =
    (char) (EVENT_MASK_BIT_FOR_RECORD_STATE_MASK |
            EVENT_MASK_BIT_FOR_MASK_1 |
            EVENT_MASK_BIT_FOR_USER_IDLE_MASK);

  globals.entryTypeInfo[ENTRY_TYPE_WAITOBJ_SIGNAL] =
  globals.entryTypeInfo[ENTRY_TYPE_WAITOBJ_WAIT_CALL] =
  globals.entryTypeInfo[ENTRY_TYPE_WAITOBJ_WAIT_GENERIC_CALL] =
  globals.entryTypeInfo[ENTRY_TYPE_MSG_AND_WAITOBJ_WAIT_CALL] =
    (char) (EVENT_MASK_BIT_FOR_RECORD_STATE_MASK |
            EVENT_MASK_BIT_FOR_MASK_1 |
            EVENT_MASK_BIT_FOR_USER_IDLE_MASK);
  globals.entryTypeInfo[ENTRY_TYPE_WAITOBJ_SET_TIMER] =
  globals.entryTypeInfo[ENTRY_TYPE_WAITOBJ_CANCEL_TIMER] =
  globals.entryTypeInfo[ENTRY_TYPE_WAITOBJ_WAIT_RETURN] =
  globals.entryTypeInfo[ENTRY_TYPE_WAITOBJ_WAIT_GENERIC_RETURN] =
  globals.entryTypeInfo[ENTRY_TYPE_MSG_AND_WAITOBJ_WAIT_RETURN] =
    (char) (EVENT_MASK_BIT_FOR_RECORD_STATE_MASK |
            EVENT_MASK_BIT_FOR_MASK_1 |
            EVENT_MASK_BIT_FOR_USER_IDLE_MASK);

  globals.entryTypeInfo[ENTRY_TYPE_RAWDISK_READ] =
  globals.entryTypeInfo[ENTRY_TYPE_RAWDISK_WRITE] =
    (char) EVENT_MASK_BIT_FOR_RECORD_STATE_MASK;
  globals.entryTypeInfo[ENTRY_TYPE_RAWDISK_COMPLETE_OPERATION] =
    (char) EVENT_MASK_BIT_FOR_RECORD_STATE_MASK;

  globals.entryTypeInfo[ENTRY_TYPE_SECTION_CREATE] =
  globals.entryTypeInfo[ENTRY_TYPE_SECTION_OPEN] =
  globals.entryTypeInfo[ENTRY_TYPE_SECTION_MAP_VIEW] =
  globals.entryTypeInfo[ENTRY_TYPE_SECTION_UNMAP_VIEW] =
  globals.entryTypeInfo[ENTRY_TYPE_SECTION_GET_MDL] =
  globals.entryTypeInfo[ENTRY_TYPE_TESTING] =
    (char) EVENT_MASK_BIT_FOR_RECORD_STATE_MASK;

  globals.entryTypeInfo[ENTRY_TYPE_MASKED_WHEN_USER_IDLE] =
    (char) EVENT_MASK_BIT_FOR_USER_IDLE_MASK;
}

// SetUserDefinedEventMask changes the "user" event mask to the new
// event mask given in eventMask, a pointer to a 256-bit (32-byte)
// array. Each bit in the event mask indicates whether its
// corresponding entry type should be masked out in the user event
// mask. This function also sets the current event mask to include the
// user event mask instead of any of the normal event masks.

void SetUserDefinedEventMask (char *eventMask)
{
  int i;
  KIRQL oldirql;

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  for (i = 0; i < 256; i++) {
    if (MaskedOut(eventMask, i))
      globals.entryTypeInfo[i] |= EVENT_MASK_BIT_FOR_USER_DEFINED_MASK;
    else
      globals.entryTypeInfo[i] &= ~EVENT_MASK_BIT_FOR_USER_DEFINED_MASK;
  }
  globals.currentEventMasks |= EVENT_MASK_BIT_FOR_USER_DEFINED_MASK;
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
}

// GetEventMask reads the information in the event type info array reflecting
// which events are masked out with the current event mask. It puts this
// information into a 256-bit (32-byte) array (eventMask). This function
// ignores any event mask bits related to recording the initial state.

void GetEventMask (char *eventMask)
{
  int i;

  for (i = 0; i < 256; i++) {
    if (globals.entryTypeInfo[i] & globals.currentEventMasks &
        ~EVENT_MASK_BIT_FOR_RECORD_STATE_MASK)
      eventMask[i/8] |= (1<<(i%8));
    else
      eventMask[i/8] &= ~(1<<(i%8));
  }
}
