//
// Flush.c
//
// Code for flushing log data to disk for the VTrcLog driver.  Also includes
// functions for checking when free disk space is available.
//

#include "NTDDK.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "Globals.h"
#include "Flush.h"
#include "GetState.h"
#include "LogEvent.h"
#include "StartStop.h"
#include "Unicode.h"

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, NextTraceFileExists)
#pragma alloc_text (INIT, InitializeDiskSpaceWatch)
#endif

////////////////////////////////////////////////////////////////////////////
//                                 MACROS
////////////////////////////////////////////////////////////////////////////

// The maximum number of times flushing can fail (e.g. because disk is full)
// before we stop logging

#define FAILED_FLUSHES_BEFORE_STOP_LOGGING       5

////////////////////////////////////////////////////////////////////////////
//                             STRUCTURES
////////////////////////////////////////////////////////////////////////////

// Structure documented in Rajeev Nagar's book.  It is used by
// ZwQueryVolumeInformationFile to find out how much free space
// is left in a partition.

typedef struct _FILE_FS_SIZE_INFORMATION {
  LARGE_INTEGER   TotalAllocationUnits;
  LARGE_INTEGER   AvailableAllocationUnits;
  ULONG           SectorsPerAllocationUnit;
  ULONG           BytesPerSector;
} FILE_FS_SIZE_INFORMATION, *PFILE_FS_SIZE_INFORMATION;

////////////////////////////////////////////////////////////////////////////
//                          FUNCTION PROTOTYPES
////////////////////////////////////////////////////////////////////////////

// Function documented in Rajeev Nagar's book, used here to find out
// how much free space is left in a partition.

NTSTATUS ZwQueryVolumeInformationFile
  (IN HANDLE FileHandle, OUT PIO_STATUS_BLOCK IoStatusBlock,
   OUT PVOID FsInformation, IN ULONG Length,
   IN FS_INFORMATION_CLASS FsInformationClass);
void FlushDataNow (PVOID calledFromQueue);

////////////////////////////////////////////////////////////////////////////
//                          TRACE DIRECTORY HANDLE
////////////////////////////////////////////////////////////////////////////

// OpenTraceDirectory opens the directory where traces go and puts a handle
// to that directory in traceDirectoryHandlePtr. It returns a status value.

NTSTATUS OpenTraceDirectory (HANDLE *traceDirectoryHandlePtr)
{
  NTSTATUS status;
  OBJECT_ATTRIBUTES objectAttributes;
  IO_STATUS_BLOCK ioStatusBlock;

  // Create the trace directory name by appending \Traces to the main directory
  // string, then open the trace directory.

  RtlCopyUnicodeString(&globals.traceFileUnicodeString,
                       &globals.mainDirectoryUnicodeString);
  status = RtlAppendUnicodeToString(&globals.traceFileUnicodeString,
                                    L"\\Traces");
  if (status != STATUS_SUCCESS) return status;

  objectAttributes.Length = sizeof(OBJECT_ATTRIBUTES);
  objectAttributes.RootDirectory = NULL;
  objectAttributes.ObjectName = &globals.traceFileUnicodeString;
  objectAttributes.SecurityDescriptor = NULL;
  objectAttributes.SecurityQualityOfService = NULL;
  objectAttributes.Attributes = OBJ_CASE_INSENSITIVE;
  status = ZwCreateFile(traceDirectoryHandlePtr,
                        FILE_ADD_FILE | FILE_READ_ATTRIBUTES,
                        &objectAttributes, &ioStatusBlock, 0,
                        FILE_ATTRIBUTE_DIRECTORY,
                        FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
                        FILE_OPEN,
                        FILE_DIRECTORY_FILE | FILE_SYNCHRONOUS_IO_ALERT, NULL,
                        0);
  return status;
}

////////////////////////////////////////////////////////////////////////////
//                           DISK SPACE ROUTINES
////////////////////////////////////////////////////////////////////////////

// GetDiskSpaceAvailable finds the number of bytes available on the partition
// containing the file having the given handle (fileHandle) and puts that
// number in bytesAvailablePtr.  It also puts the number of bytes in each
// sector of that partition in bytesPerSectorPtr.  It returns a status value.

NTSTATUS GetDiskSpaceAvailable (HANDLE fileHandle,
                                LARGE_INTEGER *bytesAvailablePtr,
                                ULONG *bytesPerSectorPtr)
{
  NTSTATUS status;
  IO_STATUS_BLOCK ioStatusBlock;
  FILE_FS_SIZE_INFORMATION sizeInfo;

  // Call ZwQueryVolumeInformationFile to find out how much space is left
  // on the volume the main directory is part of

  status = ZwQueryVolumeInformationFile(fileHandle, &ioStatusBlock, &sizeInfo,
                                        sizeof(sizeInfo),
                                        FileFsSizeInformation);
  if (status != STATUS_SUCCESS) return status;

  // Compute space available

  *bytesAvailablePtr = sizeInfo.AvailableAllocationUnits;
  bytesAvailablePtr->QuadPart *= sizeInfo.SectorsPerAllocationUnit;
  bytesAvailablePtr->QuadPart *= sizeInfo.BytesPerSector;

  // Fill in bytes-per-sector information

  *bytesPerSectorPtr = sizeInfo.BytesPerSector;

  // Indicate success

  return STATUS_SUCCESS;
}

// CheckDiskSpacePassiveRoutine is the routine to check for free disk space,
// run at passive level.  If it finds sufficient free space, it removes "disk
// full" and "flushing failed" as reasons for not logging, cancels the timer
// that continually checks for free disk space, and flushes the in-memory
// log.

void CheckDiskSpacePassiveRoutine (PVOID Parameter)
{
  LARGE_INTEGER bytesAvailable;
  ULONG bytesPerSector;

  // To avoid hysteresis, we make sure there is enough space to dump the
  // entire buffer contents twice

  if (GetDiskSpaceAvailable(globals.systemTraceDirectoryHandle,
                            &bytesAvailable, &bytesPerSector)
        == STATUS_SUCCESS &&
      bytesAvailable.QuadPart >= globals.freeBytesNeeded +
                                 globals.alignedLengthOfLogBuffer * 2) {
    RemoveReasonsForNotLogging(BECAUSE_DISK_FULL);
  }
}

// CheckDiskSpaceDPCRoutine is the deferred procedure call called
// periodically when the disk is full to see if free space has now appeared.
// It is called every ten minutes after the disk gets full.  The strange
// parameter list is to allow it to get called in a DPC.  Since the check
// for disk space cannot be done at the IRQL at which DPC's are made, this
// routine queues a routine to be executed by a system worker thread (which
// runs at passive level).

void CheckDiskSpaceDPCRoutine (IN PKDPC Dpc, IN PVOID DeferredContext,
                               IN PVOID SystemArgument1,
                               IN PVOID SystemArgument2)
{
  ExInitializeWorkItem(&globals.checkDiskSpaceWorkQueueItem,
                       &CheckDiskSpacePassiveRoutine, NULL);
  ExQueueWorkItem(&globals.checkDiskSpaceWorkQueueItem, DelayedWorkQueue);
}

// InitializeDiskSpaceWatch initializes the stuff needed to watch for when
// sufficient free disk space becomes available, namely the deferred procedure
// call (DPC) called every ten minutes after the disk becomes full, and the
// timer that goes off then to schedule the DPC.

void InitializeDiskSpaceWatch (void)
{
  KeInitializeDpc(&globals.checkDiskSpaceDpc, &CheckDiskSpaceDPCRoutine, NULL);
  KeInitializeTimer(&globals.checkDiskSpaceTimer);
}

////////////////////////////////////////////////////////////////////////////
//                      TRACE FILE NUMBERING ROUTINES
////////////////////////////////////////////////////////////////////////////

// ReserveCurrentTraceFileNumber reserves the current trace file number,
// i.e. enters it into the registry so that later traces will use higher
// numbers.

NTSTATUS ReserveCurrentTraceFileNumber (void)
{
  NTSTATUS status;
  OBJECT_ATTRIBUTES objectAttributes;
  HANDLE keyHandle;
  struct {
    ULONG   TitleIndex;
    ULONG   Type;
    ULONG   DataLength;
    ULONG   Data;
  } keyValueInfo;
  ULONG keyValueInfoLength;

  // If it's already reserved, our work is done

  if (globals.currentTraceFileNumberReserved) return STATUS_SUCCESS;

  // Get the registry key for VTrace

  MyClearUnicodeString(&globals.temporaryUnicodeString);
  status = RtlAppendUnicodeToString(&globals.temporaryUnicodeString,
       L"\\Registry\\Machine\\Software\\Jay Lorch\\VTrace\\CurrentVersion");
  if (status != STATUS_SUCCESS) return status;

  objectAttributes.Length = sizeof(OBJECT_ATTRIBUTES);
  objectAttributes.RootDirectory = NULL;
  objectAttributes.ObjectName = &globals.temporaryUnicodeString;
  objectAttributes.SecurityDescriptor = NULL;
  objectAttributes.SecurityQualityOfService = NULL;
  objectAttributes.Attributes = OBJ_CASE_INSENSITIVE;
  status = ZwOpenKey(&keyHandle, KEY_SET_VALUE, &objectAttributes);
  if (status != STATUS_SUCCESS) return status;

  // Set the value of the value "LastTraceNumber" to the current trace number

  MyClearUnicodeString(&globals.temporaryUnicodeString);
  RtlAppendUnicodeToString(&globals.temporaryUnicodeString,
                           L"LastTraceNumber");
  status = ZwSetValueKey(keyHandle, &globals.temporaryUnicodeString, 0,
                         REG_DWORD, &globals.currentTraceFileNumber, 4);
  ZwClose(keyHandle);

  // Indicate that we have reserved it so that we can save time on later calls
  // to this function

  if (status == STATUS_SUCCESS)
    globals.currentTraceFileNumberReserved = TRUE;

  return status;
}

// NextTraceFileExists finds out if the trace file numbered one more than
// globals.currentTraceFileNumber already exists in the VTrace traces
// directory and has nonzero length.

BOOLEAN NextTraceFileExists (void)
{
  OBJECT_ATTRIBUTES objectAttributes;
  IO_STATUS_BLOCK ioStatusBlock;
  NTSTATUS status;
  HANDLE fileHandle;
  FILE_STANDARD_INFORMATION standardInfo;

  // Create the trace file name from the main directory string and the trace
  // file number string.  For instance, if the main directory is D:\VTrace
  // and the trace file number is 3, the file will be
  // "D:\VTrace\Traces\T3.vtr.001".

  MyClearUnicodeString(&globals.traceFileUnicodeString);
  status = RtlAppendUnicodeToString(&globals.traceFileUnicodeString, L"T");
  if (status != STATUS_SUCCESS) return FALSE;
  status = RtlIntegerToUnicodeString(globals.currentTraceFileNumber + 1, 10,
                                     &globals.temporaryUnicodeString);
  if (status != STATUS_SUCCESS) return FALSE;
  status = RtlAppendUnicodeStringToString(&globals.traceFileUnicodeString,
                                          &globals.temporaryUnicodeString);
  if (status != STATUS_SUCCESS) return FALSE;
  status = RtlAppendUnicodeToString(&globals.traceFileUnicodeString,
                                    L".vtr.001");
  if (status != STATUS_SUCCESS) return FALSE;

  // Open the trace file (not creating it if it doesn't exist) for appending
  // (so that the contents are not erased)

  objectAttributes.Length = sizeof(OBJECT_ATTRIBUTES);
  objectAttributes.RootDirectory = globals.systemTraceDirectoryHandle;
  objectAttributes.ObjectName = &globals.traceFileUnicodeString;
  objectAttributes.SecurityDescriptor = NULL;
  objectAttributes.SecurityQualityOfService = NULL;
  objectAttributes.Attributes = OBJ_CASE_INSENSITIVE;
  status = ZwCreateFile(&fileHandle, FILE_APPEND_DATA | SYNCHRONIZE,
                        &objectAttributes, &ioStatusBlock, 0,
                        FILE_ATTRIBUTE_NORMAL, 0, FILE_OPEN,
                        FILE_SEQUENTIAL_ONLY | FILE_SYNCHRONOUS_IO_ALERT, NULL,
                        0);
  if (status != STATUS_SUCCESS) return FALSE;

  // Find the position of the end of file (i.e. the file size).  If this
  // is nonzero, return TRUE, otherwise return FALSE.

  status = ZwQueryInformationFile(fileHandle, &ioStatusBlock, &standardInfo,
                                  sizeof(FILE_STANDARD_INFORMATION),
                                  FileStandardInformation);
  ZwClose(fileHandle);
  if (status != STATUS_SUCCESS) return FALSE;

  return (standardInfo.EndOfFile.QuadPart != 0);
}

////////////////////////////////////////////////////////////////////////////
//                        LOW-LEVEL FLUSH ROUTINES
////////////////////////////////////////////////////////////////////////////

// FlushNBytes flushes numberOfBytes bytes from the in-memory log to
// the given output file (already open, passed by handle).  It puts
// in bytesActuallyFlushed the number of bytes actually flushed, since
// it may not be able to flush all numberOfBytes bytes.  It returns a
// status value.

NTSTATUS FlushNBytes (HANDLE outputFileHandle, ULONG numberOfBytes,
                      ULONG *bytesActuallyFlushed)
{
  KIRQL oldirql;
  IO_STATUS_BLOCK ioStatusBlock;
  NTSTATUS status;
  LARGE_INTEGER offset;

  // If this number of bytes would cause us to flush past the end of the
  // log buffer, just flush enough bytes to get us to the end of the
  // log buffer.  The caller of this function will call us again for
  // the rest.

  if (globals.positionOfLogStart + numberOfBytes >
      globals.alignedLengthOfLogBuffer)
    numberOfBytes = globals.alignedLengthOfLogBuffer -
                    globals.positionOfLogStart;

  // Write the bytes

  offset.LowPart = globals.currentTracePieceSize;
  offset.HighPart = 0;
  status = ZwWriteFile(outputFileHandle, NULL, NULL, NULL, &ioStatusBlock,
                       &globals.logBufferContents[globals.positionOfLogStart],
                       numberOfBytes, &offset, NULL);
  if (status == STATUS_SUCCESS) {
    // If successful, indicate in bytesActuallyFlushed how many bytes were
    // actually written.  Also, update the pointers of the log buffer to
    // reflect what parts of the log buffer are now unflushed.

    *bytesActuallyFlushed = ioStatusBlock.Information;
    if (*bytesActuallyFlushed == 0)
      status = STATUS_UNSUCCESSFUL;
    else {
      globals.currentTracePieceSize += *bytesActuallyFlushed;
      KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
      globals.positionOfLogStart += *bytesActuallyFlushed;
      globals.logLen -= *bytesActuallyFlushed;
      if (globals.positionOfLogStart >= globals.alignedLengthOfLogBuffer)
        globals.positionOfLogStart = 0;
      KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    }
  }
  else
    *bytesActuallyFlushed = 0;

  return status;
}

// BasicFlushRoutine flushes the data in the in-memory log to a trace file.
// Of course, this takes a while, so we first figure out how much unflushed
// data is in the log buffer, then flush that many bytes.  If the log buffer
// grows (as it almost certainly will) while we're flushing, we've already
// decided how many bytes to flush.  This function does nothing to make sure
// that it's only called once at a time, so the onus is on its caller to ensure
// that.  Also, this routine will fail if it finds the disk is too full (i.e.
// does not have enough bytes to leave the proper number free after flushing).
// The passed parameter is an open handle to the trace directory.  This
// function returns a status value. (This status value will be STATUS_SUCCESS
// even if all this function does is determine that no bytes need to be
// flushed.)

NTSTATUS BasicFlushRoutine (HANDLE traceDirectoryHandle)
{
  ULONG bytesPerSector, bytesToWrite, bytesWritten;
  KIRQL oldirql;
  OBJECT_ATTRIBUTES objectAttributes;
  IO_STATUS_BLOCK ioStatusBlock;
  NTSTATUS status;
  HANDLE fileHandle;
  LARGE_INTEGER bytesAvailable;
  BOOLEAN useCache;

  // Make sure we were passed a real directory handle.

  if (traceDirectoryHandle == NULL)
    return STATUS_UNSUCCESSFUL;

  status = GetDiskSpaceAvailable(traceDirectoryHandle, &bytesAvailable,
                                 &bytesPerSector);
  if (status != STATUS_SUCCESS) return status;

  // Decide how many bytes to write this time around by seeing how many
  // unflushed bytes are in the log buffer.  Note that we must acquire the
  // spin lock to read a valid length, or else we might catch it in the
  // middle of entering a record.  Also, while we're at it, decide whether
  // this disk access will use the file cache on the basis of whether this
  // is the last write to the file.  Cached accesses have the disadvantage
  // of making it hard to associate raw disk accesses with file accesses,
  // but the advantage of permitting writes in blocks that are not an even
  // multiple of the sector size.  So we use caching only for the last write
  // to the log file, after logging has been turned off. (We also use
  // caching if for some reason the number of bytes in the trace file is
  // not currently a multiple of the bytes-per-sector.)

  // If there isn't enough disk space to write that many bytes and still
  // have globals.freeBytesNeeded bytes left on the hard drive, add "disk
  // full" as a reason to stop logging and fail this call.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  bytesToWrite = globals.logLen;
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  if (globals.currentTracePieceSize + bytesToWrite >
      globals.maxTracePieceBytes) {
    globals.currentTracePieceNumber++;
    globals.currentTracePieceSize = 0;
  }
  useCache = (globals.whyLoggingDisabled & BECAUSE_END_NOT_FLUSHED) ||
             (globals.currentTracePieceSize % bytesPerSector);
  if (!useCache)
    bytesToWrite -= (bytesToWrite % bytesPerSector);
  if (bytesAvailable.QuadPart < globals.freeBytesNeeded + bytesToWrite) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    AddReasonsForNotLogging(BECAUSE_DISK_FULL);
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return STATUS_UNSUCCESSFUL;
  }

  // If there's nothing to write, return success before creating the file

  if (bytesToWrite == 0) return STATUS_SUCCESS;

  // Create the trace file name from the trace file number and trace piece
  // number. For instance, if the trace file number is 3 and the trace piece
  // number is 2, the file name will be T3.vtr.002 (in the trace directory).

  MyClearUnicodeString(&globals.traceFileUnicodeString);

  status = RtlAppendUnicodeToString(&globals.traceFileUnicodeString, L"T");
  if (status != STATUS_SUCCESS) return status;

  status = RtlIntegerToUnicodeString(globals.currentTraceFileNumber, 10,
                                     &globals.temporaryUnicodeString);
  if (status != STATUS_SUCCESS) return status;
  status = RtlAppendUnicodeStringToString(&globals.traceFileUnicodeString,
                                          &globals.temporaryUnicodeString);
  if (status != STATUS_SUCCESS) return status;

  if (globals.currentTracePieceNumber < 10)
    status = RtlAppendUnicodeToString(&globals.traceFileUnicodeString,
                                      L".vtr.00");
  else if (globals.currentTracePieceNumber < 100)
    status = RtlAppendUnicodeToString(&globals.traceFileUnicodeString,
                                      L".vtr.0");
  else
    status = RtlAppendUnicodeToString(&globals.traceFileUnicodeString,
                                      L".vtr.");
  if (status != STATUS_SUCCESS) return status;
  status = RtlIntegerToUnicodeString(globals.currentTracePieceNumber, 10,
                                     &globals.temporaryUnicodeString);
  if (status != STATUS_SUCCESS) return status;
  status = RtlAppendUnicodeStringToString(&globals.traceFileUnicodeString,
                                          &globals.temporaryUnicodeString);
  if (status != STATUS_SUCCESS) return status;

  // Open the trace file (creating it if it doesn't exist) for appending

  objectAttributes.Length = sizeof(OBJECT_ATTRIBUTES);
  objectAttributes.RootDirectory = traceDirectoryHandle;
  objectAttributes.ObjectName = &globals.traceFileUnicodeString;
  objectAttributes.SecurityDescriptor = NULL;
  objectAttributes.SecurityQualityOfService = NULL;
  objectAttributes.Attributes = OBJ_CASE_INSENSITIVE;
  status = ZwCreateFile(&fileHandle, FILE_WRITE_DATA | SYNCHRONIZE,
                        &objectAttributes, &ioStatusBlock, 0,
                        FILE_ATTRIBUTE_NORMAL, 0, FILE_OPEN_IF,
                        (useCache ? FILE_WRITE_THROUGH :
                                    FILE_NO_INTERMEDIATE_BUFFERING) |
                        FILE_SEQUENTIAL_ONLY | FILE_SYNCHRONOUS_IO_ALERT,
                        NULL, 0);
  if (status != STATUS_SUCCESS) return status;

  // Reserve the trace file number so that later traces won't overwrite the
  // trace on disk

  ReserveCurrentTraceFileNumber();

  // If the data goes past the end of the buffer, copy the trailing bytes
  // to the front of the buffer.

  if (globals.logBytesPastAlignedPart) {
    RtlCopyMemory(globals.logBufferContents,
                  globals.logBufferContents + globals.alignedLengthOfLogBuffer,
                  globals.logBytesPastAlignedPart);
    globals.logBytesPastAlignedPart = 0;
  }

  DebugPrint(("VTrcLog: Flushing data\n"));

  // Keep flushing as long as there is more to write and the flushing
  // does not cause an error.

  while (bytesToWrite > 0) {
    status = FlushNBytes(fileHandle, bytesToWrite, &bytesWritten);
    if (status != STATUS_SUCCESS)
      break;
    bytesToWrite -= bytesWritten;

    // If the file is opened uncached and somehow the number of bytes to be
    // written is not going to be a multiple of the sector size, fail

    if (!useCache && (bytesToWrite % bytesPerSector)) {
      status = STATUS_UNSUCCESSFUL;
      break;
    }
  }

  // Close the trace file.

  ZwClose(fileHandle);

  return status;
}

// FlushDataUntilSuccess flushes the data in the in-memory log to a trace file
// over and over again until it succeeds, or it has done it so many times in a
// row that it has had to stop logging.  It also logs the start and stop times
// of flushing.  The passed parameter is an open handle to the trace
// directory.  This function returns a status value.

NTSTATUS FlushDataUntilSuccess (HANDLE traceDirectoryHandle)
{
  NTSTATUS status;
  KIRQL oldirql;
  int attempts;
  char *eventBuffer;

  // Log the fact that we're flushing and might interfere with normal
  // operations

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LogEvent((char) ENTRY_TYPE_VTRACE_ACTION, 6);
  if (eventBuffer) {
    eventBuffer[1] = (char) ACTION_STARTED_FLUSHING;
    RtlZeroMemory(&eventBuffer[2], 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Try FAILED_FLUSHES_BEFORE_STOP_LOGGING times to flush the data.  If it
  // succeeds any time, stop.

  for (attempts = 0; attempts < FAILED_FLUSHES_BEFORE_STOP_LOGGING;
       attempts++) {
    status = BasicFlushRoutine(traceDirectoryHandle);
    if (status == STATUS_SUCCESS)
      break;
  }

  // If it didn't succeed, stop logging.  After all, the disk might be full.
  // If it did succeed, log an entry indicating that we stopped flushing now.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LogEvent((char) ENTRY_TYPE_VTRACE_ACTION, 6);
  if (eventBuffer) {
    eventBuffer[1] = (char) ACTION_STOPPED_FLUSHING;
    RtlZeroMemory(&eventBuffer[2], 4);
  }
  if (status != STATUS_SUCCESS)
    AddReasonsForNotLogging(BECAUSE_FLUSH_FAILED);
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  return status;
}

// FlushDataNow flushes data. It sets globals.flushInProgress to TRUE
// as long as it is running so that if a duplicate of it tries to run
// the duplicate fails. After this routine does the flush, it checks
// for any conditions that require flushing to be done again. If it
// finds such conditions, it repeats the flushing until these
// conditions no longer hold. The kinds of conditions are "buffer
// still dangerously full" and "end of trace not yet flushed". The
// parameter calledFromQueue indicates if this routine was called from
// a work queue; if so, it should set globals.flushQueued to FALSE and
// it can be assured that it is being called in a system thread
// context.

void FlushDataNow (PVOID calledFromQueue)
{
  KIRQL oldirql;
  BOOLEAN flushAgain, otherFlushActive;
  NTSTATUS status;
  HANDLE traceDirectoryHandle;
  ULONG reasonsToRemove;

  // Acquire the spin lock to ensure uninterrupted access to
  // globals.flushInProgress.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  otherFlushActive = globals.flushInProgress;
  globals.flushInProgress = TRUE;
  if (calledFromQueue)
    globals.flushQueued = FALSE;
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // If another copy of this routine is active, return and let it work

  if (otherFlushActive) return;

  // Get a handle to the trace directory. If we are called from the queue,
  // the system trace directory handle is fine. Otherwise, we should open
  // a handle to the trace directory for use here.

  if (calledFromQueue)
    traceDirectoryHandle = globals.systemTraceDirectoryHandle;
  else {
    status = OpenTraceDirectory(&traceDirectoryHandle);
    if (status != STATUS_SUCCESS)
      traceDirectoryHandle = NULL;
  }

  // Loop, repeatedly flushing until we don't need to any more. The reason
  // for all these checks is as follows. While we are flushing, any attempt
  // to initiate flushing must fail; however, such attempts to initiate
  // flushing may be due to valid reasons why flushing must take place
  // again, such as that the end of the trace file is not flushed or the
  // buffer has once again gotten dangerously full. So after we flush, we
  // check again to see if any of these conditions hold. Once we're sure
  // no such conditions hold, we can remove "end-not-flushed" and "buffer-
  // overflowed" as reasons not to log.

  do {
    status = FlushDataUntilSuccess(traceDirectoryHandle);
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    if (status != STATUS_SUCCESS) {
      if (globals.currentTraceFileTerminated) {

       // If the flush failed, but the current trace file has been
       // terminated, we assume the worst: that the trace file is
       // about to be or has already been uploaded. If this is the
       // case, we never want to flush the data in the buffer, so we
       // erase the log buffer. If we do this, we might as well
       // pretend the flush was successful--it will solve any
       // end-not-flushed or buffer-overflowed problems we're having.

        globals.positionOfLogStart = 0;
        globals.logLen = 0;
        globals.positionOfLogEnd = 0;
        globals.logBytesPastAlignedPart = 0;
        status = STATUS_SUCCESS;
      }
      flushAgain = FALSE;
    }
    else if (globals.logLen > globals.logDangerLen)
      flushAgain = TRUE;
    else if ((globals.whyLoggingDisabled & BECAUSE_END_NOT_FLUSHED) &&
             (globals.logLen > 0))
      flushAgain = TRUE;
    else
      flushAgain = FALSE;

    if (!flushAgain) {

      // If this is the end of the loop, record the reasons that logging
      // is disabled (but only those we can do something about). The fact
      // that we've made it this far means that if the flush was successful,
      // we can remove the reasons "end-not-flushed" and "buffer-overflowed".
      // We do this now rather than later because we're holding the spin-
      // lock now and while we were holding it we made sure that any of these
      // reasons is no longer important. Also, we're about to set
      // globals.flushInProgress to FALSE, so that if one of these conditions
      // comes up again, the flush routine will be called again.

      reasonsToRemove = globals.whyLoggingDisabled &
        (BECAUSE_END_NOT_FLUSHED | BECAUSE_BUFFER_OVERFLOWED);
      globals.flushInProgress = FALSE;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  } while (flushAgain);

  // If we successfully flushed, remove any reasons for not logging that
  // we've now done away with.

  if (status == STATUS_SUCCESS)
    RemoveReasonsForNotLogging(reasonsToRemove);

  // If we were not called from the queue, close the trace directory handle
  // we opened

  if (!calledFromQueue)
    ZwClose(traceDirectoryHandle);
}

////////////////////////////////////////////////////////////////////////////
//                        HIGH-LEVEL FLUSH ROUTINES
////////////////////////////////////////////////////////////////////////////

// FlushDataSoon ensures that the in-memory log data will be flushed some
// time later.  It is called while the mainMutex is held, so it can't do
// the flushing right now; thus, it instead queues a work item that will do
// it later.

void FlushDataSoon (void)
{
  // Only queue the work item if one is not already queued and if a flush
  // is not already ongoing

  if (!globals.flushQueued && !globals.flushInProgress) {
    globals.flushQueued = TRUE;
      // The "(PVOID) 1" here is passed to FlushDataNow as its first parameter
      // to indicate it was called from the flush queue and must therefore set
      // globals.flushQueued to FALSE.
    ExInitializeWorkItem(&globals.flushWorkQueueItem, FlushDataNow, (PVOID) 1);
    ExQueueWorkItem(&globals.flushWorkQueueItem, DelayedWorkQueue);
  }
}
