//
// Globals.h contains information about global variables in the VTrcLog
// driver. It also contains macros used throughout the driver code.
//

////////////////////////////////////////////////////////////////////////////
//                                MACROS
////////////////////////////////////////////////////////////////////////////

// print macro that only turns on when checked builds are on
#if DBG
#define DebugPrint(arg) DbgPrint arg
#else
#define DebugPrint(arg)
#endif

// A unique index for each of the Zw system calls we hook:

#define OPEN_FILE_INDEX                    0
#define CREATE_FILE_INDEX                  1
#define WAIT_FOR_SINGLE_INDEX              2
#define WAIT_FOR_MULTIPLE_INDEX            3
#define SIGNAL_AND_WAIT_INDEX              4
#define FLUSH_ICACHE_INDEX                 5
#define FLUSH_WRITE_BUFFER_INDEX           6
#define TERMINATE_PROCESS_INDEX            7
#define TERMINATE_THREAD_INDEX             8
#define WRITE_REQUEST_DATA_INDEX           9
#define WRITE_VM_INDEX                     10
#define CREATE_SECTION_INDEX               11
#define OPEN_SECTION_INDEX                 12
#define MAP_VIEW_INDEX                     13
#define UNMAP_VIEW_INDEX                   14
#define SET_EVENT_INDEX                    15
#define RELEASE_MUTANT_INDEX               16
#define RELEASE_SEMAPHORE_INDEX            17
#define SET_TIMER_INDEX                    18
#define CANCEL_TIMER_INDEX                 19
#define NUM_SYSTEM_CALLS_HOOKED            20

////////////////////////////////////////////////////////////////////////////
//                            FUNCTION TYPES
////////////////////////////////////////////////////////////////////////////

// The following are procedure types for various Zw functions.  They are
// also procedure types for the replacement Zw functions we have here.

typedef NTSTATUS (*OPEN_FILE_PROC_TYPE)
                 (PHANDLE, ACCESS_MASK, POBJECT_ATTRIBUTES, PIO_STATUS_BLOCK,
                  ULONG, ULONG);
typedef NTSTATUS (*CREATE_FILE_PROC_TYPE)
                 (PHANDLE, ACCESS_MASK, POBJECT_ATTRIBUTES, PIO_STATUS_BLOCK,
                  PLARGE_INTEGER, ULONG, ULONG, ULONG, ULONG, PVOID, ULONG);
typedef ULONG (*WAIT_FOR_SINGLE_PROC_TYPE)(HANDLE, ULONG, PLARGE_INTEGER);
typedef ULONG (*WAIT_FOR_MULTIPLE_PROC_TYPE)
                 (ULONG, PHANDLE, ULONG, ULONG, PLARGE_INTEGER);
typedef ULONG (*SIGNAL_AND_WAIT_PROC_TYPE)(HANDLE, HANDLE, ULONG,
                                           PLARGE_INTEGER);
typedef NTSTATUS (*FLUSH_ICACHE_PROC_TYPE)(PVOID, PVOID, PVOID);
typedef NTSTATUS (*FLUSH_WRITE_BUFFER_PROC_TYPE)(void);
typedef NTSTATUS (*TERMINATE_PROCESS_PROC_TYPE)(PVOID, PVOID);
typedef NTSTATUS (*TERMINATE_THREAD_PROC_TYPE)(PVOID, PVOID);
typedef NTSTATUS (*WRITE_REQUEST_DATA_PROC_TYPE)
                 (PVOID, PVOID, PVOID, PVOID, PVOID, PVOID);
typedef NTSTATUS (*WRITE_VM_PROC_TYPE)(PVOID, PVOID, PVOID, PVOID, PVOID);
typedef NTSTATUS (*CREATE_SECTION_PROC_TYPE)
                 (PHANDLE, ACCESS_MASK, POBJECT_ATTRIBUTES, PLARGE_INTEGER,
                  ULONG, ULONG, HANDLE);
typedef NTSTATUS (*OPEN_SECTION_PROC_TYPE)
                 (PHANDLE, ACCESS_MASK, POBJECT_ATTRIBUTES);
typedef NTSTATUS (*MAP_VIEW_PROC_TYPE)
                 (HANDLE, HANDLE, PVOID, ULONG, ULONG, PLARGE_INTEGER,
                  PULONG, SECTION_INHERIT, ULONG, ULONG);
typedef NTSTATUS (*UNMAP_VIEW_PROC_TYPE)(HANDLE, PVOID);
typedef NTSTATUS (*SET_EVENT_PROC_TYPE)(HANDLE, ULONG);
typedef NTSTATUS (*RELEASE_MUTANT_PROC_TYPE)(HANDLE, ULONG);
typedef NTSTATUS (*RELEASE_SEMAPHORE_PROC_TYPE)(HANDLE, ULONG, ULONG);
typedef NTSTATUS (*SET_TIMER_PROC_TYPE)(HANDLE, PLARGE_INTEGER, ULONG, ULONG,
                                        ULONG, ULONG, ULONG);
typedef NTSTATUS (*CANCEL_TIMER_PROC_TYPE)(HANDLE, ULONG);

////////////////////////////////////////////////////////////////////////////
//                             STRUCTURES
////////////////////////////////////////////////////////////////////////////

// The structure VTRACE_GLOBALS contains all the global variables used by
// this driver in one place, so it's clear what's global and what isn't.

typedef struct _vtrace_globals {
  // The following fields relate directly to event logging:

  PCHAR                        logBufferContents;
    // the log buffer itself
  ULONG                        positionOfLogStart;
    // the index into the logBufferContents buffer at which resides the
    // first byte not yet flushed to disk
  ULONG                        logLen;
    // the length of the section of the buffer not yet flushed to disk
  ULONG                        alignedLengthOfLogBuffer;
    // the size of the sector-aligned part of the log buffer
  ULONG                        positionOfLogEnd;
    // the index of the current log buffer pointer
  ULONG                        logDangerLen;
    // the number of bytes of unflushed data at which we get worried about
    // possible imminent overflow, and thus schedule data to be flushed
  ULONG                        logBytesPastAlignedPart;
    // the number of log bytes past the end of the aligned part of the buffer
  ULONG                        whyLoggingDisabled;
    // each bit indicates a reason why logging is currently disabled, so
    // logging is enabled if and only if this field has value 0

  // The following fields relate to the event mask

  CHAR                         entryTypeInfo[256];
    // an array of information about each entry type (whether it denotes
    // activity and whether it is masked off in various event masks)
  CHAR                         currentEventMasks;
    // which event masks are active
  KDPC                         changeEventMaskDpc;
    // the deferred procedure call called periodically to change the event
    // mask
  KTIMER                       changeEventMaskTimer;
    // the timer that goes off periodically to call the deferred procedure
    // call above

  // The following fields relate to flushing the in-memory log data to
  // trace files:

  BOOLEAN                      flushInProgress;
    // are we flushing data now?
  WORK_QUEUE_ITEM              flushWorkQueueItem;
    // work queue item for flushing data
  BOOLEAN                      flushQueued;
    // is the flush item queued?
  ULONG                        freeBytesNeeded;
    // minimum free disk space we must ensure always remains
  ULONG                        maxTracePieceBytes;
    // maximum size for a trace piece in bytes
  HANDLE                       systemTraceDirectoryHandle;
    // handle of the open trace directory, valid in system process only
  ULONG                        currentTraceFileNumber;
    // the current trace file number being used
  BOOLEAN                      currentTraceFileNumberReserved;
    // is the current trace file number reserved in the registry, so that it
    // will not be reused by a later run of VTrcLog?
  BOOLEAN                      currentTraceFileTerminated;
    // Is the current trace file terminated? In other words, should we
    // commit to not putting any more data in the current trace file?
  ULONG                        currentTracePieceNumber;
    // Which trace piece number is currently being output
  ULONG                        currentTracePieceSize;
    // the size, in bytes, of the current trace piece
  KDPC                         checkDiskSpaceDpc;
    // the deferred procedure call called periodically after the disk is full,
    // to check if it has become not full
  KTIMER                       checkDiskSpaceTimer;
    // the timer that goes off periodically after the disk is full, to call the
    // deferred procedure call above
  WORK_QUEUE_ITEM              checkDiskSpaceWorkQueueItem;
    // the work queue item for checking disk space when the IRQL is at passive
    // level

  // The following fields are related to the swap-context routine.  The first
  // three are global since they have to be accessed quickly from assembly
  // language.

  unsigned char *              nonOverwrittenPartOfOrigSwapRoutine;
    // the first instruction in the original context swap routine past
    // the part overwritten by the jump to the new context swap routine
  unsigned char *              origSwapRoutine;
    // original location of context swap routine

  // The following "orig..." fields are the original Zw routines before
  // hooking.  So, for instance, origOpenFile is the original ZwOpenFile.

  OPEN_FILE_PROC_TYPE          origOpenFile;
  CREATE_FILE_PROC_TYPE        origCreateFile;
  WAIT_FOR_SINGLE_PROC_TYPE    origWaitForSingleObject;
  WAIT_FOR_MULTIPLE_PROC_TYPE  origWaitForMultipleObjects;
  SIGNAL_AND_WAIT_PROC_TYPE    origSignalAndWaitForSingleObject;
  FLUSH_ICACHE_PROC_TYPE       origFlushInstructionCache;
  FLUSH_WRITE_BUFFER_PROC_TYPE origFlushWriteBuffer;
  TERMINATE_PROCESS_PROC_TYPE  origTerminateProcess;
  TERMINATE_THREAD_PROC_TYPE   origTerminateThread;
  WRITE_REQUEST_DATA_PROC_TYPE origWriteRequestData;
  WRITE_VM_PROC_TYPE           origWriteVirtualMemory;
  CREATE_SECTION_PROC_TYPE     origCreateSection;
  OPEN_SECTION_PROC_TYPE       origOpenSection;
  MAP_VIEW_PROC_TYPE           origMapViewOfSection;
  UNMAP_VIEW_PROC_TYPE         origUnmapViewOfSection;
  SET_EVENT_PROC_TYPE          origSetEvent;
  RELEASE_MUTANT_PROC_TYPE     origReleaseMutant;
  RELEASE_SEMAPHORE_PROC_TYPE  origReleaseSemaphore;
  SET_TIMER_PROC_TYPE          origSetTimer;
  CANCEL_TIMER_PROC_TYPE       origCancelTimer;

  // The following are the systemCallNumbers for the Zw routines

  ULONG                        systemCallNumbers[NUM_SYSTEM_CALLS_HOOKED];

  // The following fields relate to user activity:

  KDPC                         userGoesIdleDpc;
    // the deferred procedure called when the user has been idle for 10 minutes
  KTIMER                       userGoesIdleTimer;
    // the timer that goes off when the user has been idle for 10 minutes
  LARGE_INTEGER                timeOfLastUserActivity;
    // the time that user activity last occurred

  // The following fields are related to the beat:

  KTIMER                       beatTimer;
    // the timer that goes off every time a beat period ends
  KDPC                         beatDpc;
    // the deferred procedure called when the beat happens

  // The following fields are some Unicode strings that are allocated upon
  // driver entry and stay around forever, in part to save on later allocation
  // time:

  UNICODE_STRING               traceFileUnicodeString;
    // name of the trace file being flushed to
  UNICODE_STRING               mainDirectoryUnicodeString;
    // name of the main VTrace directory
  UNICODE_STRING               temporaryUnicodeString;
    // preallocated for various uses

  // Miscellaneous fields:

  ULONG                        CPUSpeedInMHz;
    // the CPU speed in MHz
  WORK_QUEUE_ITEM              recordStateWorkQueueItem;
    // the work queue item for recording the state after logging was started,
    // since this can only be done at passive level
  ULONG                        vtraceVersion;
    // the version number of VTrace
  PKEVENT                      userNameRequiredEvent;
    // the event used to signal VTrcSrvc that a new user name event is
    // needed
  HANDLE                       userNameRequiredEventHandle;
    // handle to the above event

} VTRACE_GLOBALS;

/////////////////////////////////////////////////////////////////////////
//                         EXTERNAL VARIABLES
/////////////////////////////////////////////////////////////////////////

extern VTRACE_GLOBALS globals;
