//
// LogEvent.c
//
// Code for adding events to the log.
//

#include "NTDDK.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "Globals.h"
#include "Flush.h"

////////////////////////////////////////////////////////////////////////
//                       IN-MEMORY EVENT LOGGING
////////////////////////////////////////////////////////////////////////

// LogEvent sets up to log an event of length eventLength and type eventType.
// It returns a pointer to which the caller should copy those eventLength
// bytes. If the logging should not be done for some reason, it returns NULL
// intead. LogEvent does several other things:
//   1) It puts the current 4-byte timestamp in the log just before the
//      location returned.
//   2) It copies the event type byte to the location returned so that the
//      caller need not log the event type.
//   3) It checks for buffer wrapping.
//   4) It checks for the buffer getting too long and arranges a flush to
//      happen soon if it finds this condition.
// The circumstances under which LogEvent will return NULL include:
//   1) Logging is turned off.
//   2) There isn't room in the buffer for this event. In this case, it
//      also adds a reason for not logging (BECAUSE_BUFFER_OVERFLOWED),
//      and adds this reason for not logging to the end of the log.
//   3) The event type requested is masked off by the current event mask.
// LogEvent makes sure that there is always room at the end of the buffer
// for the 10 bytes needed to indicate the reason logging stopped.
//
// This function assumes that the mutex IS held.

char *LogEvent (char eventType, ULONG eventLength)
{
  char *entryPtr;
  BOOLEAN bufferOverflowed;

  // Only log if logging is enabled, and if this event type is not masked off.

  if (globals.whyLoggingDisabled ||
      (globals.entryTypeInfo[(unsigned char) eventType] &
       globals.currentEventMasks))
    return NULL;

  // Read the low 4 bytes of the cycle count to the current log position.

  entryPtr = globals.logBufferContents + globals.positionOfLogEnd;
  __asm {
    _emit 0x0F       ; RDTSC - get beginning timestamp to edx:eax
    _emit 0x31
    mov ebx, entryPtr
    mov [ebx], eax   ; save low 4 bytes of timestamp
  }

  // If the log is getting too long, queue a worker thread item to flush it.
  // Also check if the log can accommodate this entry and a final 19-byte
  // entry to indicate why logging stopped. If it cannot, and this is not
  // a final 19-byte entry to indicate why logging stopped, add the reason
  // BECAUSE_BUFFER_OVERFLOWED to why logging is disabled. Note that
  // we must add the BECAUSE_BUFFER_OVERFLOWED reason *before* initiating the
  // flush, to make sure that a flush happens that will clear the
  // BECAUSE_BUFFER_OVERFLOWED reason.

  bufferOverflowed = FALSE;
  if (globals.logLen > globals.logDangerLen) {
    if (globals.logLen + eventLength + 23 > globals.alignedLengthOfLogBuffer) {
      globals.whyLoggingDisabled |= BECAUSE_BUFFER_OVERFLOWED;
      bufferOverflowed = TRUE;
      eventLength = 6;
      DebugPrint(("VTrcLog: Stopping logging because buffer overflowed\n"));
    }
    FlushDataSoon();
  }

  // Commit the log entry by incrementing the log length

  globals.logLen += (eventLength + 4);
  globals.positionOfLogEnd += (eventLength + 4);

  // If we've gone past the end of the buffer, note this in
  // globals.logBytesPastAlignedPart and wrap the log around to the
  // beginning of the buffer.

  if (globals.positionOfLogEnd >= globals.alignedLengthOfLogBuffer) {
    globals.logBytesPastAlignedPart =
      globals.positionOfLogEnd - globals.alignedLengthOfLogBuffer;
    globals.positionOfLogEnd = globals.logBytesPastAlignedPart;
  }

  // Fill in the event type and return a pointer to that place in the log

  if (bufferOverflowed) {
    entryPtr[4] = (char) ENTRY_TYPE_VTRACE_ACTION;
    entryPtr[5] = (char) ACTION_STOPPED_LOGGING;
    * (ULONG *) &entryPtr[6] = BECAUSE_BUFFER_OVERFLOWED;
    return NULL;
  }
  else {
    entryPtr[4] = eventType;
    return &entryPtr[4];
  }
}

// CheckIfLoggingEventType returns a boolean indicating whether the given
// event type is currently being logged.

BOOLEAN CheckIfLoggingEventType (char eventType)
{
  // Only log if logging is enabled, and if this event type is not masked off.

  if (globals.whyLoggingDisabled ||
      (globals.entryTypeInfo[(unsigned char) eventType] &
       globals.currentEventMasks))
    return FALSE;
  else
    return TRUE;
}

// LogHighTimestamp puts an entry in the log describing the current high
// timestamp value. This function assumes that the mutex IS held.

void LogHighTimestamp (void)
{
  char *entryPtr;

  // Only log if logging is enabled and if there's room in the log for this
  // 9-byte entry and the final 10-byte "buffer overflowed" entry.

  if (globals.whyLoggingDisabled ||
      globals.logLen + 19 > globals.alignedLengthOfLogBuffer)
    return;

  // Read the low 4 bytes of the cycle count to the current log position.
  // Read the next 4 bytes of the cycle count to 5 bytes past the current
  // log position, where the high timestamp in the entry goes.

  entryPtr = globals.logBufferContents + globals.positionOfLogEnd;
  __asm {
    _emit 0x0F       ; RDTSC - get beginning timestamp to edx:eax
    _emit 0x31
    mov ebx, entryPtr
    mov [ebx], eax   ; save low 4 bytes of timestamp
    mov [ebx+5], edx ; save high 4 bytes of timestamp
  }

  // Complete the high-timestamp entry by inserting the event type.

  globals.logBufferContents[globals.positionOfLogEnd + 4] =
    (char) ENTRY_TYPE_HIGH_TIMESTAMP;
    
  globals.logLen += 9;
  globals.positionOfLogEnd += 9;

  // If we've gone past the end of the buffer, note this in
  // globals.logBytesPastAlignedPart and wrap the log around to the
  // beginning of the buffer.

  if (globals.positionOfLogEnd >= globals.alignedLengthOfLogBuffer) {
    globals.logBytesPastAlignedPart =
      globals.positionOfLogEnd - globals.alignedLengthOfLogBuffer;
    globals.positionOfLogEnd = globals.logBytesPastAlignedPart;
  }
}
