// 
// MiscHook.c
//
// Functions related to miscellaneous hook functions set by VTrcLog
// (not Zw routine hooks).
//

#include "NTDDK.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "../Common/Version.h"
#include "Globals.h"
#include "GetState.h"
#include "LogEvent.h"
#include "MiscHook.h"
#include "ProcList.h"
#include "Unicode.h"

////////////////////////////////////////////////////////////////////////////
//                                MACROS
////////////////////////////////////////////////////////////////////////////

// IsSwapContextRoutinePointer is a boolean expression revealing whether
// the ULONG * given by p represents the beginning of the original
// swap-context routine.  It checks 28 consecutive bytes to make sure.
// The swap-context routine is different on Windows 2000 and NT 4.0.

#if TARGET_NT_VERSION == 4

#define IsSwapContextRoutinePointer(p) \
  (*(p) == 0x2D46C626 && (p)[1] == 0x8BC90A02 && (p)[2] == 0x0F519C0B && \
   (p)[3] == 0xD58BC520 && (p)[4] == 0x882C4E8A && (p)[5] == 0x89FA504B && \
   (p)[6] == 0x468B2867)

#else

#define IsSwapContextRoutinePointer(p) \
  (*(p) == 0x2D46C626 && (p)[1] == 0x0B8B9C02 && (p)[2] == 0x080CBB83 && \
   (p)[3] == 0x51000000 && (p)[4] == 0x00F4850F && (p)[5] == 0x200F0000 && \
   (p)[6] == 0x8AD58BC5)

#endif

////////////////////////////////////////////////////////////////////////////
//                           FORWARD DEFINITIONS
////////////////////////////////////////////////////////////////////////////

BOOLEAN IsUniprocessorFreeBuild (void);
NTSTATUS SetUpContextSwapHookOnMultiprocessorOrCheckedBuild (void);
unsigned char *FindOrigSwapRoutine (void);

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, IsUniprocessorFreeBuild)
#pragma alloc_text (INIT, SetUpContextSwapHookOnMultiprocessorOrCheckedBuild)
#pragma alloc_text (INIT, FindOrigSwapRoutine)
#pragma alloc_text (INIT, SetUpContextSwapHook)
#pragma alloc_text (INIT, RestoreOldContextSwapRoutine)
#endif

////////////////////////////////////////////////////////////////////////////
//                       PROCESS/THREAD CREATION HOOKS
////////////////////////////////////////////////////////////////////////////

// MyProcessNotifyRoutine is our process notification routine.  It
// is designed to be called whenever a process is created or destroyed.
// The passed parameters are the parent process's ID, the process's ID,
// and whether the process is being created (as opposed to destroyed).
// It logs the process-creation or process-destruction event.

void MyProcessNotifyRoutine (IN HANDLE ParentId, IN HANDLE ProcessId,
                             IN BOOLEAN Create)
{
  KIRQL oldirql;
  char buf[MAX_PROCESS_NAME_LENGTH];
  USHORT processNameLength;
  char *eventBuffer;

  if (Create) {
    processNameLength = GetProcessName((ULONG) ProcessId, buf,
                                       MAX_PROCESS_NAME_LENGTH);
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    // This is a variable-length entry, so its length goes in eventBuffer[1]
    eventBuffer = LogEvent((char) ENTRY_TYPE_PROCESS_CREATE,
                           processNameLength + 6);
    if (eventBuffer) {
      eventBuffer[1] = (processNameLength + 6);
      RtlCopyMemory(&eventBuffer[2], &ProcessId, 4);
      RtlCopyMemory(&eventBuffer[6], buf, processNameLength);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }
  else {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LogEvent((char) ENTRY_TYPE_PROCESS_DESTROY, 5);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &ProcessId, 4);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }
}

// MyThreadNotifyRoutine is our thread notification routine.  It
// is designed to be called whenever a thread is created or destroyed.
// The passed parameters are the thread's process's ID, the thread's ID,
// and whether the thread is being created (as opposed to destroyed).
// It logs the thread-creation or thread-destruction event.

void MyThreadNotifyRoutine (IN HANDLE ProcessId, IN HANDLE ThreadId,
                            IN BOOLEAN Create)
{
  KIRQL oldirql;
  char *eventBuffer;

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LogEvent((char) (Create ? ENTRY_TYPE_THREAD_CREATE :
                                 ENTRY_TYPE_THREAD_DESTROY),
                                 9);
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[1], &ProcessId, 4);
    RtlCopyMemory(&eventBuffer[5], &ThreadId, 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
}

////////////////////////////////////////////////////////////////////////////
//                          SWAP CONTEXT HOOK
////////////////////////////////////////////////////////////////////////////

#if TARGET_NT_VERSION == 4

NTSTATUS OverwriteKernelMemory (char *dest, char *source, int length)
{
  KSPIN_LOCK jumpProtectSpinLock;
  KIRQL oldirql;

  // It is imperative that no context switch happens during the copying
  // of the jump instruction, so we protect the jump with a spin lock.
  // (Context switches are disabled while a spin lock is held.)

  KeInitializeSpinLock(&jumpProtectSpinLock);
  KeAcquireSpinLock(&jumpProtectSpinLock, &oldirql);
  RtlCopyMemory(dest, source, length);
  KeReleaseSpinLock(&jumpProtectSpinLock, oldirql);
  return STATUS_SUCCESS;
}

#else

NTSTATUS OverwriteKernelMemory (char *dest, char *source, int length)
{
  KSPIN_LOCK jumpProtectSpinLock;
  KIRQL oldirql;
  PMDL mdl;
  PVOID writableAddress;

  mdl = IoAllocateMdl((PVOID) dest, length, FALSE, FALSE, NULL);
  if (!mdl) {
    DebugPrint(("VTrcLog: Could not allocate MDL to overwrite kernel.\n"));
    return STATUS_UNSUCCESSFUL;
  }
  MmBuildMdlForNonPagedPool(mdl);
  MmProbeAndLockPages(mdl, KernelMode, IoWriteAccess);
  writableAddress = MmMapLockedPages(mdl, KernelMode);
  if (!writableAddress) {
    MmUnlockPages(mdl);
    IoFreeMdl(mdl);
    DebugPrint(("VTrcLog: Could not map kernel memory for writing.\n"));
    return STATUS_UNSUCCESSFUL;
  }

  // It is imperative that no context switch happens during the copying
  // of the jump instruction, so we protect the jump with a spin lock.
  // (Context switches are disabled while a spin lock is held.)

  KeInitializeSpinLock(&jumpProtectSpinLock);
  KeAcquireSpinLock(&jumpProtectSpinLock, &oldirql);
  RtlCopyMemory(writableAddress, source, length);
  KeReleaseSpinLock(&jumpProtectSpinLock, oldirql);

  MmUnmapLockedPages(writableAddress, mdl);
  MmUnlockPages(mdl);
  IoFreeMdl(mdl);
  return STATUS_SUCCESS;
}

#endif

// IsUniprocessorFreeBuild returns a boolean indicating if the uniprocessor
// free build of the kernel is running. If it runs into an error, it will
// return TRUE for safety.

BOOLEAN IsUniprocessorFreeBuild (void)
{
  NTSTATUS status;
  UNICODE_STRING string;
  OBJECT_ATTRIBUTES objectAttributes;
  struct {
    ULONG   TitleIndex;
    ULONG   Type;
    ULONG   DataLength;
    WCHAR   Data[64];
  } keyValueInfo;
  HANDLE keyHandle;
  ULONG keyValueInfoLength;

  DebugPrint(("VTrcLog: Determining if it's the uniprocessor free build.\n"));

  // Initialize the Unicode string to be used in this function

  status = MyInitUnicodeString(&string, 128);
  if (status != STATUS_SUCCESS)
    return TRUE;

  // Open the Windows NT registry key.

  status = RtlAppendUnicodeToString(&string,
                                    L"\\Registry\\Machine\\Software\\Microsoft"
                                    L"\\Windows NT\\CurrentVersion");
  if (status != STATUS_SUCCESS) {
    MyFreeUnicodeString(&string);
    return TRUE;
  }

  objectAttributes.Length = sizeof(OBJECT_ATTRIBUTES);
  objectAttributes.RootDirectory = NULL;
  objectAttributes.ObjectName = &string;
  objectAttributes.SecurityDescriptor = NULL;
  objectAttributes.SecurityQualityOfService = NULL;
  objectAttributes.Attributes = OBJ_CASE_INSENSITIVE;
  status = ZwOpenKey(&keyHandle, KEY_QUERY_VALUE, &objectAttributes);
  if (status != STATUS_SUCCESS) {
    MyFreeUnicodeString(&string);
    return TRUE;
  }

  // Read the CurrentType.

  MyClearUnicodeString(&string);
  RtlAppendUnicodeToString(&string, L"CurrentType");
  status = ZwQueryValueKey(keyHandle, &string, KeyValuePartialInformation,
                           &keyValueInfo, sizeof(keyValueInfo),
                           &keyValueInfoLength);
  ZwClose(keyHandle);
  MyFreeUnicodeString(&string);
  if (status != STATUS_SUCCESS || keyValueInfo.Type != REG_SZ)
    return TRUE;

  // Determine if it starts with "Uniprocessor Free".

  if (RtlCompareMemory(keyValueInfo.Data, L"Uniprocessor Free", 34) == 34) {
    DebugPrint(("VTrcLog: It is the uniprocessor free build.\n"));
    return TRUE;
  }
  else {
    DebugPrint(("VTrcLog: It's not the uniprocessor free build.\n"));
    return FALSE;
  }
}

// VTraceSwapContextNotifyRoutine is a callback that the system can call
// whenever there is a context swap. It logs the thread switch.

VOID FASTCALL VTraceSwapContextNotifyRoutine (IN HANDLE OldThreadId,
                                              IN HANDLE NewThreadId)
{
  KIRQL oldirql;
  PCHAR eventBuffer;

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LogEvent((char) ENTRY_TYPE_THREAD_SWITCH, 5);
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[1], &NewThreadId, 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
}

// SetUpContextSwapHookOnMultiprocessorOrCheckedBuild sets up the
// context swap hook on a multiprocessor or checked build. In either case,
// the KeSetSwapContextNotifyRoutine can be used for this purpose.

NTSTATUS SetUpContextSwapHookOnMultiprocessorOrCheckedBuild (void)
{
  DebugPrint(("VtrcLog: Setting up context swap routine on multiprocessor "
              "or checked build.\n"));
  globals.origSwapRoutine = NULL;
  KeSetSwapContextNotifyRoutine(VTraceSwapContextNotifyRoutine);

  return STATUS_SUCCESS;
}

// NewSwapContext is the hook routine for SwapContext. It gets called
// instead of SwapContext, so it logs the context switch, executes the
// overwritten instruction from the old context swap routine, then
// jumps to the point in the old swap routine past that instruction.  It
// is written in assembler to make sure that any overwritten registers are
// restored, considering it isn't going to be called, but merely jumped to
// and from.

void FASTCALL NewSwapContext (void)
{
  __asm {
    // Save parameters we will overwrite
    push eax
    push ecx

    // Save interrupt mask and stop all interrupts. Since context-swaps can't
    // occur while a spinlock is held, we know no one else has the spinlock.
    pushfd
    cli

    // eventBuffer = LogEvent((char) ENTRY_TYPE_THREAD_SWITCH, 5);
    push 5
    push 0Eh
    call LogEvent

    // if (eventBuffer)
    cmp eax, 0
    je DoneLogging
    push eax  // Push eventBuffer for later use.

    //   eventBuffer[1] = PsGetCurrentThreadId();
    call PsGetCurrentThreadId
    pop ecx   // Pop eventBuffer to use it now.
    mov dword ptr [ecx+1], eax

  DoneLogging:
    // Restore interrupt mask
    popfd

    // Pop saved parameters
    pop ecx
    pop eax

    // Execute overwritten instruction from original swap routine
    mov byte ptr es:[esi+2dh],2

    // Jump to the point in the original swap routine past overwritten part
    jmp dword ptr globals.nonOverwrittenPartOfOrigSwapRoutine
  }
}

// FindOrigSwapRoutine searches in system space for the context-swap routine.
// It first looks in a few likely places (the places known to house the routine
// in a few versions of NT 4.0), then scans a likely area.  It returns the
// location of the swap routine, or NULL if it cannot find the routine.

unsigned char *FindOrigSwapRoutine (void)
{
  // The likelyLocations array contains some locations in various
  // NTOSKRNL.EXE's that the original swap context routine are located.
  // If they aren't found there, we search in the range of bytes between
  // earliestLocation and latestLocation. All these things are different
  // between NT 4.0 and Windows 2000.

#if TARGET_NT_VERSION == 4

  unsigned char *likelyLocations[6] =
      { (unsigned char *) 0x80140ca0,    // In NT 4.0 SP3
        (unsigned char *) 0x80140c50,    // In NT 4.0 SP3 after GetAdmin Fix
        (unsigned char *) 0x8013f4f0,    // In the original NT 4.0
        (unsigned char *) 0x80141ca0,    // In NT 4.0 SP4
        (unsigned char *) 0x80141e70,    // In NT 4.0 SP5
        (unsigned char *) 0x80142420 };  // In NT 4.0 SP6, SP6a
  int numLocations = 6;

  unsigned char *earliestLocation = (unsigned char *) 0x8013e000;
  unsigned char *latestLocation = (unsigned char *) 0x80150000;

#else

  unsigned char *likelyLocations[2] =
  { (unsigned char *) 0x80402584,      // In Windows 2000 Service Pack 1
    (unsigned char *) 0x80403AFA };    // In Windows 2000
  int numLocations = 2;

  unsigned char *earliestLocation = (unsigned char *) 0x80400000;
  unsigned char *latestLocation = (unsigned char *) 0x80470000;

#endif

  int i;
  unsigned long *p;
  unsigned char *searchLow, *searchHigh, *searchPos;

  // First, search through the likeliest locations, being careful not
  // to touch invalid (e.g. paged out) memory

  for (i = 0; i < numLocations; i++) {
    p = (unsigned long *) likelyLocations[i];
    if (MmIsAddressValid(p) && MmIsAddressValid(p+7) &&
        IsSwapContextRoutinePointer(p))
      return likelyLocations[i];
  }

  // Find a subrange [searchLow, searchHigh+27] of [earliestLocation,
  // latestLocation] that consists only of valid addresses

  for (searchLow = (unsigned char *) ROUND_TO_PAGES(earliestLocation);
       !MmIsAddressValid(searchLow) && searchLow <= latestLocation;
       searchLow += PAGE_SIZE)
    ;
  for (searchHigh = searchLow + PAGE_SIZE;
       MmIsAddressValid(searchHigh) && searchHigh < latestLocation;
       searchHigh += PAGE_SIZE)
    ;
  if (searchHigh > latestLocation) searchHigh = latestLocation;
  searchHigh -= 28;

  // Search from searchLow to searchHigh for the routine

  for (searchPos = searchLow; searchPos <= searchHigh; searchPos++) {
    p = (unsigned long *) searchPos;
    if (IsSwapContextRoutinePointer(p))
      return searchPos;
  }

  // If we didn't find it, return NULL

  return NULL;
}

// SetUpContextSwapHook sets up our hook routine for SwapContext.
// It constructs a jump instruction to place at the beginning of the
// original SwapContext routine that jumps to our hook routine (which
// will in turn jump back to the original swap routine once it's finished
// logging).

NTSTATUS SetUpContextSwapHook (void)
{
  unsigned char jumpInstruction[5];
  long offset;

  // If it's the multiprocessor or checked build, we can filter context
  // swaps in a simpler way, so we do that instead.

  if (!IsUniprocessorFreeBuild())
    return SetUpContextSwapHookOnMultiprocessorOrCheckedBuild();

  // Find the location of the original swap-context routine.

  globals.origSwapRoutine = FindOrigSwapRoutine();
  if (globals.origSwapRoutine == NULL) return STATUS_UNSUCCESSFUL;

  // To jump to our code (NewSwapContext) instead of the real swap routine,
  // we replace the first 5 bytes of the real swap routine with a jump to
  // the meat of NewSwapContext.  (By "meat", we mean the point past the
  // compiler-inserted stack set-up stuff that we don't need or want to
  // execute.)  We use an offset jump, so first we compute the offset between
  // the end of the 5-byte jump instruction and the beginning of the meat of
  // NewSwapContext.  We then construct the jump instruction to point to there.

  globals.nonOverwrittenPartOfOrigSwapRoutine = globals.origSwapRoutine + 5;
  offset = ((unsigned char *) &NewSwapContext) + 6
            - globals.nonOverwrittenPartOfOrigSwapRoutine;
  jumpInstruction[0] = 0xE9; // jmp
  RtlCopyMemory(&jumpInstruction[1], &offset, 4);

  return OverwriteKernelMemory(globals.origSwapRoutine, jumpInstruction, 5);
}

// RestoreOldContextSwapRoutine restores the old context swap routine after
// it has been hooked by SetUpContextSwapHook.

void RestoreOldContextSwapRoutine (void)
{
  // The following five bytes are the first five bytes of the original
  // swap routine.

  unsigned char beginningOfOrigSwapContents[5] =
                                    {0x26, 0xc6, 0x46, 0x2d, 0x02};

  // If there is no original swap routine location, this means we used
  // KeSetSwapContextNotifyRoutine instead of patching memory, so we
  // should use that routine to unhook the context-swap routine.

  if (globals.origSwapRoutine == NULL) {
    KeSetSwapContextNotifyRoutine(NULL);
    return;
  }

  // Overwrite the swap routine start with what it used to be.

  OverwriteKernelMemory(globals.origSwapRoutine, beginningOfOrigSwapContents,
                        5);
}
