// 
// ProcList.c
//
// Code for routines in the VTrcLog driver dealing with internal
// procedure lists.  These procedure lists are used to obtain the names
// and thread ID's associated with each (or a particular) running process.
//

#include "NTDDK.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "../Common/Version.h"
#include "LogEvent.h"
#include "GetState.h"
#include "Unicode.h"

////////////////////////////////////////////////////////////////////////////
//                               MACROS
////////////////////////////////////////////////////////////////////////////

#define TAG_GET_PROC_THREAD_INFO                5
#define FIRST_GUESS_AT_PROC_THREAD_INFO_SIZE    8192
#define INCREMENT_FOR_PROC_THREAD_INFO_SIZE     1024
#define MAXIMUM_PROC_THREAD_INFO_SIZE           2097152

#define MAX(a,b)  ((a)>(b)?(a):(b))

////////////////////////////////////////////////////////////////////////////
//                           FUNCTION PROTOTYPES
////////////////////////////////////////////////////////////////////////////

// This undocumented function is used to get system information.  We use it
// with tag 5 to get the process information list.

unsigned long ZwQuerySystemInformation (ULONG tag, VOID *buffer,
                                        ULONG bufferSize, ULONG *returnedSize);

////////////////////////////////////////////////////////////////////////////
//                         PROCESS LIST FUNCTIONS
////////////////////////////////////////////////////////////////////////////

// GetProcessAndThreadInfo allocates a buffer and fills it with the
// process/thread information structure list. It puts the length of
// this list in *bytesReturnedPtr. It returns a pointer to the
// buffer, or NULL if unsuccessful. It is the responsibility of the
// caller to free the space at the returned pointer if it is not NULL.

char *GetProcessAndThreadInfo (ULONG *bytesReturnedPtr)
{
  char *buf;         // buffer to hold the process and thread information
  ULONG bufSize;     // size of the buffer
  NTSTATUS status;   // status code returned by ZwQuerySystemInformation

  bufSize = FIRST_GUESS_AT_PROC_THREAD_INFO_SIZE;
  while ((buf = ExAllocatePool(NonPagedPool, bufSize)) != NULL) {
    *bytesReturnedPtr = 0;
    status = ZwQuerySystemInformation(TAG_GET_PROC_THREAD_INFO,
                                      buf, bufSize, bytesReturnedPtr);
    if (status == STATUS_SUCCESS) return buf;

    // If the buffer was the wrong size, make the buffer bigger; use the
    // value returned in bytesReturnedPtr as a hint about the needed size.

    ExFreePool(buf);
    if (status == STATUS_BUFFER_OVERFLOW ||
        status == STATUS_INFO_LENGTH_MISMATCH) {
      bufSize = MAX(*bytesReturnedPtr,
                    bufSize + INCREMENT_FOR_PROC_THREAD_INFO_SIZE);
      if (bufSize > MAXIMUM_PROC_THREAD_INFO_SIZE)
        return NULL;
    }
    else
      return NULL;
  }
  return NULL;
}

// RecordCurrentProcessThreadState records to the start of the trace the
// current state of all processes and threads in the system, including ID
// numbers and process names. It returns a status value. It assumes that
// the spin lock is NOT held, since the spin lock cannot be held while we
// call ZwQuerySystemInformation.

NTSTATUS RecordCurrentProcessThreadState (void)
{
  NTSTATUS status;
  char *processInfoPtr, *curProcessInfo, *endOfProcessInfo, *curThreadInfo;
  WCHAR *processNamePtr;
  ULONG processInfoLength, numberOfThreads, processId, threadId,
        nextEntryOffset, n;
  USHORT processNameLength, i;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Read the list of process information structures

  processInfoPtr = GetProcessAndThreadInfo(&processInfoLength);
  if (processInfoPtr == NULL)
    return STATUS_UNSUCCESSFUL;

  // Loop through each process information structure in the list.  The
  // structures are of variable length; we know when we've reached the
  // last one because we have gone past the end of the process list.

  curProcessInfo = processInfoPtr;
  endOfProcessInfo = &processInfoPtr[processInfoLength];

  while (curProcessInfo + 0x48 <= endOfProcessInfo) {

    // The following undocumented fields are in each process information
    // structure

    numberOfThreads = * (ULONG *) (curProcessInfo + 4);
    processNameLength = (* (USHORT *) (curProcessInfo + 0x38)) / 2;
    processNamePtr = * (WCHAR **) (curProcessInfo + 0x3C);
    processId = * (ULONG *) (curProcessInfo + 0x44);

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    // Copy the name of the process (but no more than
    // MAX_PROCESS_NAME_LENGTH characters) to the entry, and put
    // the length of the name copied into processNameLength.  The
    // process name is in Unicode, so the characters have to be
    // converted to ASCII as they are copied.  Before doing any of
    // this, check whether the address of the process name is valid;
    // if it isn't, call the process "???".

    if (processNameLength == 0 || MmIsAddressValid(processNamePtr)) {
      if (processNameLength > MAX_PROCESS_NAME_LENGTH)
        processNameLength = MAX_PROCESS_NAME_LENGTH;
      eventBuffer = LogEvent((char) ENTRY_TYPE_PROCESS_CREATE,
                             processNameLength + 6);
      if (eventBuffer) {
        eventBuffer[1] = (char) (processNameLength + 6);
        for (i = 0; i < processNameLength; i++)
          eventBuffer[i+6] = * (char *) processNamePtr++;
      }
    }
    else {
      eventBuffer = LogEvent((char) ENTRY_TYPE_PROCESS_CREATE, 9);
      if (eventBuffer) {
        eventBuffer[1] = (char) 9;
        RtlCopyMemory(&eventBuffer[6], "???", 3);
      }
    }
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[2], &processId, 4);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

    // The next part of the process info structure is a list of thread
    // info structures.  We loop through them all, recording the thread
    // ID's in "thread create" entries.  (However, if the loop would make
    // us access past endOfProcessInfo, we stop.)

#if TARGET_NT_VERSION == 4
    curThreadInfo = curProcessInfo + 0x88;     // Position in NT 4.0
#else
    curThreadInfo = curProcessInfo + 0xb8;     // Position in Windows 2000
#endif
    for (n = 0;
         n < numberOfThreads && curThreadInfo + 0x28 <= endOfProcessInfo;
         n++) {

      // Extract the thread ID from the undocumented thread info structure

      threadId = * (ULONG *) (curThreadInfo + 0x24);

      // Put a "thread create" entry into the log

      KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
      eventBuffer = LogEvent((char) ENTRY_TYPE_THREAD_CREATE, 9);
      if (eventBuffer) {
        RtlCopyMemory(&eventBuffer[1], &processId, 4);
        RtlCopyMemory(&eventBuffer[5], &threadId, 4);
      }
      KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

      // Advance to next thread info structure for this process

      curThreadInfo += 0x40;
    }

    // Extract the next entry offset field from the process info structure.
    // If it is 0, then this is the last process info structre; otherwise,
    // it is the length of this process info structure and thus the amount
    // we have to skip to get to the next one.

    nextEntryOffset = * (ULONG *) curProcessInfo;
    if (nextEntryOffset == 0)
      break;
    else
      curProcessInfo = curProcessInfo + nextEntryOffset;
  }

  // Finally, free the space we allocated for the process info list.

  ExFreePool(processInfoPtr);

  return STATUS_SUCCESS;
}

// GetProcessName obtains the process name for a given process ID by getting
// the process info list and searching it for that ID.  The name is put in
// processName in ASCII, but no more than maxLength characters.  This routine
// returns the length of the name copied, or 0 if no name was copied.  It
// is assumed that maxLength is at least 3, so that "???" can be made the
// process name if it cannot be found.

USHORT GetProcessName (ULONG searchProcessId, char *processName, int maxLength)
{
  NTSTATUS status;
  char *processInfoPtr, *curProcessInfo, *endOfProcessInfo;
  WCHAR *processNamePtr;
  ULONG processInfoLength, processId, nextEntryOffset;
  USHORT processNameLength, i;

  // By default, the name is set to "???"

  RtlCopyMemory(processName, "???", 3);
  processNameLength = 3;

  // Read the list of process information structures

  processInfoPtr = GetProcessAndThreadInfo(&processInfoLength);
  if (processInfoPtr == NULL)
    return processNameLength;

  // Loop through each process information structure in the list.  The
  // structures are of variable length; we know when we've reached the
  // last one because we have gone past the end of the process list.

  curProcessInfo = processInfoPtr;
  endOfProcessInfo = &processInfoPtr[processInfoLength];

  while (curProcessInfo + 0x48 <= endOfProcessInfo) {

    // Extract the process ID from the undocumented process info structure
    processId = * (ULONG *) (curProcessInfo + 0x44);

    // If it matches, get the process name and return it
    if (processId == searchProcessId) {

      // Extract a couple more undocumented fields
      processNameLength = (* (USHORT *) (curProcessInfo + 0x38)) / 2;
      processNamePtr = * (WCHAR **) (curProcessInfo + 0x3C);

      // If the process name pointer is a valid address, copy the string
      // (but not more than maxLength characters) from the Unicode source
      // to the ASCII destination; otherwise, the process stays named "???".
      // In either case, set the variable processNameLength to the number of
      // bytes in the process name.

      if (processNameLength == 0 || MmIsAddressValid(processNamePtr)) {
        if (processNameLength > maxLength)
          processNameLength = (USHORT) maxLength;
        for (i = 0; i < processNameLength; i++)
          processName[i] = * (char *) processNamePtr++;
      }
      else
        processNameLength = 3;  // Length of "???"
      break;              // Leave the loop since we found the process ID
    }

    // Extract the next entry offset field from the process info structure.
    // If it is 0, then this is the last process info structre; otherwise,
    // it is the length of this process info structure and thus the amount
    // we have to skip to get to the next one.

    nextEntryOffset = * (ULONG *) curProcessInfo;
    if (nextEntryOffset == 0)
      break;
    else
      curProcessInfo = curProcessInfo + nextEntryOffset;
  }

  // Finally, free the space we allocated for the process info list.

  ExFreePool(processInfoPtr);

  return processNameLength;
}
