//
// StartStop.c
//
// Functions related to starting and stopping logging.
//

#include "NTDDK.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "../Common/Version.h"
#include "Globals.h"
#include "LogEvent.h"
#include "Activity.h"
#include "EventMask.h"
#include "Flush.h"
#include "GetState.h"
#include "ProcList.h"
#include "Unicode.h"

////////////////////////////////////////////////////////////////////////////
//                           FORWARD DEFINITIONS
////////////////////////////////////////////////////////////////////////////

void StartDiskSpaceChecking (void);
void InitializeTrace (void);
void RecordCurrentState (PVOID reason);

////////////////////////////////////////////////////////////////////////////
//                     CHANGING REASONS FOR NOT LOGGING
////////////////////////////////////////////////////////////////////////////

// RemoveReasonsForNotLogging removes the given reasons for not logging.
// It assumes that the mutex is NOT held when this routine is called.

void RemoveReasonsForNotLogging (ULONG reasons)
{
  KIRQL oldirql;
  PCHAR eventBuffer;

  // As an optimization, if there are no reasons for not logging, don't
  // do anything.

  if (globals.whyLoggingDisabled == 0)
    return;

  // Pare down the reasons to only those that are actually in the current
  // bitmask. Then remove those reasons from the bitmask. (We pare them so
  // that later, when we do things based on the reasons removed, we only
  // consider the reasons actually removed.)

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  reasons &= globals.whyLoggingDisabled;
  globals.whyLoggingDisabled &= ~reasons;

  if (reasons) {
    DebugPrint(("VTrcLog: Removed reasons for not logging %d.\n", reasons));
  }

  if (reasons & BECAUSE_DISK_FULL) {
    DebugPrint(("VTrcLog: Cancelling disk space watch because disk-full "
                "removed as a reason.\n"));
    globals.whyLoggingDisabled &= ~BECAUSE_FLUSH_FAILED;
    KeCancelTimer(&globals.checkDiskSpaceTimer);
    FlushDataSoon();
  }

  if (globals.whyLoggingDisabled) {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }
  else {
    // If logging is no longer disabled, we must record the current state.
    // We do this by adding the ENTRY_INFO_IN_RECORD_STATE_MASK bit to the
    // event mask, and calling RecordCurrentState. However, before we call
    // RecordCurrentState, we must make sure the trace is initialized--if
    // the last trace file was terminated, we must call InitializeTrace to
    // initialize the trace.
  
    DebugPrint(("VTrcLog: Resuming logging.\n"));

    globals.currentEventMasks |= EVENT_MASK_BIT_FOR_RECORD_STATE_MASK;
    if (globals.currentTraceFileTerminated) {
      DebugPrint(("VTrcLog: Unterminating trace.\n", reasons));
      globals.currentTraceFileTerminated = FALSE;
      InitializeTrace();
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    RecordCurrentState((PVOID) reasons);

    // Notify the user-level logger that we need a user-name event to be
    // generated.

    if (globals.userNameRequiredEvent)
      KeSetEvent(globals.userNameRequiredEvent, 0, FALSE);
  }
}

// AddReasonsForNotLogging adds a given set of reasons for not logging.
// It assumes that the mutex IS held when this routine is called.

void AddReasonsForNotLogging (ULONG reasons)
{
  PCHAR eventBuffer;

  // Pare down the reasons to only those relevant, i.e., only those reasons
  // that are not currently in the bitmask.
  
  reasons &= ~globals.whyLoggingDisabled;
  DebugPrint(("VTrcLog: Adding reasons for not logging %d.\n", reasons));

  // If logging isn't stopped, add the reasons why logging stopped to the log.

  if (!globals.whyLoggingDisabled) {
    DebugPrint(("VTrcLog: Stopping logging for reasons %d.\n", reasons));
    eventBuffer = LogEvent((char) ENTRY_TYPE_VTRACE_ACTION, 6);
    if (eventBuffer) {
      eventBuffer[1] = (char) ACTION_STOPPED_LOGGING;
      RtlCopyMemory(&eventBuffer[2], &reasons, 4);
    }
  }

  // Add the reasons given to the reasons for not logging. We didn't want to
  // do this earlier because to log the reasons we stopped logging, logging
  // has to still be allowed!

  globals.whyLoggingDisabled |= reasons;

  // If these reasons terminate a trace file, and the current trace
  // file isn't already terminated, terminate the trace file.

  if ((reasons & REASONS_TO_END_TRACE) &&
      !globals.currentTraceFileTerminated) {
    DebugPrint(("VTrcLog: Terminating trace.\n"));
    globals.currentTraceFileTerminated = TRUE;

    // Add the reason "because end not flushed" to the reasons tracing is
    // disabled, since we shouldn't start logging again until the trace
    // file that has just been terminated has been flushed.

    globals.whyLoggingDisabled |= BECAUSE_END_NOT_FLUSHED;

    // Usually, we would flush now. However, there are two exceptions:
    // 1) If the system is halting, it's important to flush NOW, not soon,
    //    so we leave it to the caller of AddReasonsForNotLogging to flush
    //    immediately once the spin lock is no longer held
    // 2) If VTrace is being uninstalled, there's little point in flushing
    //    since the trace files will be wiped out. Indeed, it would be bad
    //    to initiate flushing soon because the trace directory may no longer
    //    exist by then.

    if (!(reasons & (BECAUSE_SYSTEM_HALTING | BECAUSE_UNINSTALLING)))
      FlushDataSoon();
  }

  // If this is because the disk is full, initiate periodic checking for free
  // disk space.

  if (reasons & BECAUSE_DISK_FULL) {
    StartDiskSpaceChecking();
  }

  // If we're uninstalling, clear the log to make sure it never gets flushed.

  if (reasons & BECAUSE_UNINSTALLING) {
    globals.logLen = 0;
  }
}

////////////////////////////////////////////////////////////////////////////
//                            RECORDING STATE
////////////////////////////////////////////////////////////////////////////

// RecordCurrentDeviceState records the current state of a hooked device.
// It assumes that the main mutex IS held.

void RecordCurrentDeviceState (PDEVICE_OBJECT deviceObject, ULONG diskNumber,
                               ULONG partitionNumber)
{
  USHORT nameLength;
  PUNICODE_STRING driverString;
  int i;
  PCHAR eventBuffer;

  // Copy the driver name string to the event after checking to make
  // sure it's a valid address.  This copy is from Unicode to ASCII.
  // Store the length of this string in nameLength.

  driverString = &deviceObject->DriverObject->DriverName;
  if (MmIsAddressValid(driverString) &&
      MmIsAddressValid(driverString->Buffer)) {
    nameLength = driverString->Length / 2;  // Divide by 2: WCHAR is 2 bytes
    if (nameLength > 128)
      nameLength = 128;
    eventBuffer = LogEvent((char) ENTRY_TYPE_DEVICE_HOOK, nameLength + 12);
    if (eventBuffer) {
      eventBuffer[1] = (char) (nameLength + 12);
      for (i = 0; i < nameLength; i++)
        eventBuffer[i+12] = * (char *) &driverString->Buffer[i];
    }
  }
  else {
    eventBuffer = LogEvent((char) ENTRY_TYPE_DEVICE_HOOK, 15);
    if (eventBuffer) {
      eventBuffer[1] = (char) 15;
      RtlCopyMemory(&eventBuffer[12], "???", 3);
    }
  }
  if (eventBuffer) {
    // Put the entry type, device object number, and device type in the entry
    RtlCopyMemory(&eventBuffer[2], &deviceObject, 4);
    RtlCopyMemory(&eventBuffer[6], &deviceObject->DeviceType, 4);
    eventBuffer[10] = (char) diskNumber;
    eventBuffer[11] = (char) partitionNumber;
  }
}

// RecordCurrentState records the current state of various things just
// before logging starts. Its strange parameter format (the reason why
// logging had been stopped before is passed as a PVOID instead of a ULONG)
// is so that it can be called from a work item.

void RecordCurrentState (PVOID reason)
{
  NTSTATUS status;
  LARGE_INTEGER systemTime, localTime;
  PDEVICE_OBJECT deviceObject;
  PFS_HOOK_EXTENSION fsHookExt;
  PNET_HOOK_EXTENSION netHookExt;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Make sure this routine doesn't get called at an IRQL above passive level.
  // If it has been, queue a work item to get called later and execute this
  // routine.

  if (KeGetCurrentIrql() > PASSIVE_LEVEL) {
    ExInitializeWorkItem(&globals.recordStateWorkQueueItem, RecordCurrentState,
                         reason);
    ExQueueWorkItem(&globals.recordStateWorkQueueItem, CriticalWorkQueue);
    return;
  }

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  LogHighTimestamp();
  eventBuffer = LogEvent((char) ENTRY_TYPE_VTRACE_SET_MASK, 33);
  if (eventBuffer) {
    GetEventMask(&eventBuffer[1]);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Record the processes and threads that are currently active

  RecordCurrentProcessThreadState();

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);

  // Record all the devices hooked by VTrcFS and VTrcNet.  For this, we
  // use the linked list of devices that is found in each driver object.

  if (sharedState->fsDriverObject != NULL)
    for (deviceObject = sharedState->fsDriverObject->DeviceObject;
         deviceObject != NULL; deviceObject = deviceObject->NextDevice) {
      fsHookExt = deviceObject->DeviceExtension;
      RecordCurrentDeviceState(fsHookExt->attachedDevice,
                               fsHookExt->diskNumber,
                               fsHookExt->partitionNumber);
    }

  if (sharedState->netDriverObject != NULL)
    for (deviceObject = sharedState->netDriverObject->DeviceObject;
         deviceObject != NULL; deviceObject = deviceObject->NextDevice) {
      netHookExt = deviceObject->DeviceExtension;
      RecordCurrentDeviceState(netHookExt->attachedDevice, 255, 255);
    }

  // Record the reason we had stopped logging, that has gone away now.

  eventBuffer = LogEvent((char) ENTRY_TYPE_VTRACE_ACTION, 6);
  if (eventBuffer) {
    eventBuffer[1] = (char) ACTION_STARTED_LOGGING;
    RtlCopyMemory(&eventBuffer[2], &reason, 4);
  }

  // Now that we've recorded the initial state, there's no further need
  // for the ENTRY_INFO_IN_RECORD_STATE_MASK bit in the event mask.

  globals.currentEventMasks &= ~EVENT_MASK_BIT_FOR_RECORD_STATE_MASK;

  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
}

////////////////////////////////////////////////////////////////////////////
//                           INITIALIZING TRACE
////////////////////////////////////////////////////////////////////////////

// InitializeTrace initializes the trace by incrementing the current
// trace file number and logging the trace preamble including such things
// as the version number. It is called with the main mutex held.

void InitializeTrace (void)
{
  KIRQL oldirql;
  LARGE_INTEGER systemTime, localTime;
  LARGE_INTEGER cycleCount, *cycleCountPtr;

  DebugPrint(("VTrcLog: Initializing trace.\n"));

  // Advance to the next trace file number.

  globals.currentTraceFileNumber++;
  globals.currentTraceFileNumberReserved = FALSE;
  globals.currentTracePieceNumber = 1;
  globals.currentTracePieceSize = 0;

  // Get the cycle count and local time of the beginning of the trace.
  
  cycleCountPtr = &cycleCount;
  __asm {
    _emit 0x0F       ;RDTSC - get beginning timestamp to edx:eax
    _emit 0x31
    mov ebx, cycleCountPtr
    mov [ebx], eax   ;save low 4 bytes of timestamp
    mov [ebx+4], edx ;save high 4 bytes of timestamp
  }
  KeQuerySystemTime(&systemTime);
  ExSystemTimeToLocalTime(&systemTime, &localTime);

  // Put the trace preamble directly at the beginning of the log buffer.

  RtlCopyMemory(&globals.logBufferContents[0], &cycleCount.LowPart, 4);
  globals.logBufferContents[4] = (char) ENTRY_TYPE_HIGH_TIMESTAMP;
  RtlCopyMemory(&globals.logBufferContents[5], &cycleCount.HighPart, 4);

  RtlCopyMemory(&globals.logBufferContents[9], &cycleCount.LowPart, 4);
  globals.logBufferContents[13] = (char) ENTRY_TYPE_SET_CPU_SPEED;
  RtlCopyMemory(&globals.logBufferContents[14], &globals.CPUSpeedInMHz, 4);

  RtlCopyMemory(&globals.logBufferContents[18], &cycleCount.LowPart, 4);
  globals.logBufferContents[22] = (char) ENTRY_TYPE_VTRACE_VERSION;
  RtlCopyMemory(&globals.logBufferContents[23], &globals.vtraceVersion, 4);

  RtlCopyMemory(&globals.logBufferContents[27], &cycleCount.LowPart, 4);
  globals.logBufferContents[31] = (char) ENTRY_TYPE_LOCAL_TIME;
  RtlCopyMemory(&globals.logBufferContents[32], &localTime, 8);

  // Set the log buffer information to reflect the fact that there are
  // now only 40 bytes in the trace, starting at position 0 in the log
  // buffer.

  globals.positionOfLogStart = 0;
  globals.logLen = 40;
  globals.positionOfLogEnd = 40;
  globals.logBytesPastAlignedPart = 0;

  // Empty the VTrcFS hash table since no file names are in this log yet.

  if (sharedState->emptyFSHashFunc != NULL)
    (*sharedState->emptyFSHashFunc)();
}

////////////////////////////////////////////////////////////////////////////
//                           DISK SPACE ROUTINES
////////////////////////////////////////////////////////////////////////////

// StartDiskSpaceChecking starts the periodic checking for free disk space
// every ten minutes.

void StartDiskSpaceChecking (void)
{
  LARGE_INTEGER tenMinutesFromNow = {0x9A5F4400, -2};
                                              // -6 billion (100 ns units)

  KeSetTimerEx(&globals.checkDiskSpaceTimer, tenMinutesFromNow,
               600000, &globals.checkDiskSpaceDpc);
}

////////////////////////////////////////////////////////////////////////////
//                             HIBERNATION
////////////////////////////////////////////////////////////////////////////

#if TARGET_NT_VERSION == 5

void HibernationNotifyRoutine (BOOLEAN hibernating)
{
  KIRQL oldirql;

  if (hibernating) {
    DebugPrint(("VTrcLog: Notified that the system is hibernating.\n"));
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    AddReasonsForNotLogging(BECAUSE_HIBERNATING);
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }
  else {
    DebugPrint(("VTrcLog: Notified that the system is not hibernating.\n"));
    RemoveReasonsForNotLogging(BECAUSE_HIBERNATING);
  }
}

#endif
