//
// SysCall.c
//
// Code to obtain system call numbers for Zw functions by parsing the
// system file NTDLL.DLL.
//

#include "NTDDK.h"
#include "SysCall.h"

////////////////////////////////////////////////////////////////////////////
//                                 MACROS
////////////////////////////////////////////////////////////////////////////

// MAX_NAME_LENGTH must be at least the length of the longest function
// name being searched for, including the trailing null and initial Zw

#define MAX_NAME_LENGTH                   32

// Definitions from WINNT.H (and other sources) describing the length
// of PE file format structures

#define SIZE_OF_NT_SIGNATURE              4
#define IMAGE_SIZEOF_FILE_HEADER          20
#define IMAGE_SIZEOF_NT_OPTIONAL_HEADER   224
#define IMAGE_SIZEOF_SHORT_NAME           8

// MIN returns the smaller of a and b

#define MIN(a,b)                          (((a)>(b))?(b):(a))

////////////////////////////////////////////////////////////////////////////
//                               STRUCTURES
////////////////////////////////////////////////////////////////////////////

// SECTION_INFO contains the important information we want to save
// about each section in the file.  It's a subset of the
// IMAGE_SECTION_HEADER fields.

typedef struct _section_info {
  LONG virtualAddress;
  LONG sizeOfRawData;
  LONG pointerToRawData;
} SECTION_INFO;

// IMAGE_EXPORT_DIRECTORY is lifted from WINNT.H.  It describes the
// data directory describing the exports section of a file in PE format.

typedef struct _IMAGE_EXPORT_DIRECTORY {
  ULONG   Characteristics;
  ULONG   TimeDateStamp;
  USHORT  MajorVersion;
  USHORT  MinorVersion;
  ULONG   Name;
  ULONG   Base;
  ULONG   NumberOfFunctions;
  ULONG   NumberOfNames;
  PULONG  *AddressOfFunctions;
  PULONG  *AddressOfNames;
  PUSHORT *AddressOfNameOrdinals;
} IMAGE_EXPORT_DIRECTORY, *PIMAGE_EXPORT_DIRECTORY;

// IMAGE_SECTION_HEADER is lifted from WINNT.H.  It describes the
// section header for each section in a file in PE format.

typedef struct _IMAGE_SECTION_HEADER {
  char    Name[IMAGE_SIZEOF_SHORT_NAME];
  union {
    ULONG  PhysicalAddress;
    ULONG  VirtualSize;
  } Misc;
  ULONG   VirtualAddress;
  ULONG   SizeOfRawData;
  ULONG   PointerToRawData;
  ULONG   PointerToRelocations;
  ULONG   PointerToLinenumbers;
  USHORT  NumberOfRelocations;
  USHORT  NumberOfLinenumbers;
  ULONG   Characteristics;
} IMAGE_SECTION_HEADER, *PIMAGE_SECTION_HEADER;

////////////////////////////////////////////////////////////////////////////
//                        EXTERNALLY DEFINED OBJECTS
////////////////////////////////////////////////////////////////////////////

// KeServiceDescriptorTable is the not-so-well-documented service table
// structure

typedef struct _SRVTABLE {
  PVOID *ServiceTable;
  ULONG LowCall;
  ULONG HiCall;
  PVOID *ArgTable;
} SRVTABLE, *PSRVTABLE;
extern PSRVTABLE KeServiceDescriptorTable;

////////////////////////////////////////////////////////////////////////////
//                          FORWARD DEFINITIONS
////////////////////////////////////////////////////////////////////////////

NTSTATUS OpenDLLFile (HANDLE *fileHandle);
NTSTATUS ReadDLLFile (HANDLE fileHandle, char *buf, ULONG offset,
                      ULONG length);

NTSTATUS GetVirtualToPositionConversion (SECTION_INFO *sectionInfo,
                                         USHORT numSections,
                                         LONG virtualAddress,
                                         LONG *conversion);
NTSTATUS GetOffsetToPEFileHeader (HANDLE fileHandle,
                                  LONG *offsetToPEFileHeader);
NTSTATUS GetNumberOfSections (HANDLE fileHandle, LONG offsetToPEFileHeader,
                              USHORT *numSections);
NTSTATUS ReadSectionInfo (HANDLE fileHandle, LONG offsetToPEFileHeader,
                          SECTION_INFO *sectionInfo, USHORT numSections);
NTSTATUS GetImageExportDirectory
  (HANDLE fileHandle, SECTION_INFO *sectionInfo, USHORT numSections,
   LONG offsetToPEFileHeader, IMAGE_EXPORT_DIRECTORY *imageExportDirectory);
NTSTATUS ReadFunctionNamePositions (HANDLE fileHandle,
                                    LONG namesVirtualAddress,
                                    SECTION_INFO *sectionInfo,
                                    USHORT numSections,
                                    LONG *functionNamePositions,
                                    ULONG numFunctions);
NTSTATUS ReadFunctionPositions (HANDLE fileHandle, LONG fnPtrsVirtualAddress,
                                SECTION_INFO *sectionInfo, USHORT numSections,
                                LONG *functionPositions, ULONG numFunctions);
ULONG GetSystemCallNumberFromFunction (HANDLE fileHandle,
                                       LONG functionPosition);

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, GetSystemCallNumbers)
#pragma alloc_text (INIT, OpenDLLFile)
#pragma alloc_text (INIT, ReadDLLFile)
#pragma alloc_text (INIT, GetVirtualToPositionConversion)
#pragma alloc_text (INIT, GetOffsetToPEFileHeader)
#pragma alloc_text (INIT, GetNumberOfSections)
#pragma alloc_text (INIT, ReadSectionInfo)
#pragma alloc_text (INIT, GetImageExportDirectory)
#pragma alloc_text (INIT, ReadFunctionNamePositions)
#pragma alloc_text (INIT, ReadFunctionPositions)
#pragma alloc_text (INIT, GetSystemCallNumberFromFunction)
#endif

////////////////////////////////////////////////////////////////////////////
//                          SYSTEM CALL ROUTINE
////////////////////////////////////////////////////////////////////////////

// GetSystemCallNumbers searches the NTDLL.DLL file for the system call
// numbers corresponding to certain Zw functions.  The names of the
// functions are passed in systemCallNames (without the Zw prefixes),
// the pointers to where the system call numbers should be placed are
// passed in the array systemCallNumbers, and the size of these arrays
// is passed in numSystemCalls.

NTSTATUS GetSystemCallNumbers (char **systemCallNames,
                               ULONG *systemCallNumbers, int numSystemCalls)
{
  HANDLE fileHandle;
  SECTION_INFO *sectionInfo;
  LONG *functionNamePositions, *functionPositions;
  int i, systemCallsFound, numFunctions, functionNumber;
  USHORT numSections;
  IMAGE_EXPORT_DIRECTORY imageExportDirectory;
  char foundFunctionName[MAX_NAME_LENGTH];
  NTSTATUS status;
  LONG offsetToPEFileHeader;

  // Initialize all the system call numbers to 0 so we can tell which ones
  // we haven't found yet

  systemCallsFound = 0;
  for (i = 0; i < numSystemCalls; i++)
    systemCallNumbers[i] = 0;

  // Open NTDLL.DLL for reading

  status = OpenDLLFile(&fileHandle);
  if (status != STATUS_SUCCESS)
    return status;

  // Find the offset to the PE file header and put that in offsetToPEFileHeader

  status = GetOffsetToPEFileHeader(fileHandle, &offsetToPEFileHeader);
  if (status != STATUS_SUCCESS) {
    ZwClose(fileHandle);
    return status;
  }

  // Find the number of sections in the file

  status = GetNumberOfSections(fileHandle, offsetToPEFileHeader, &numSections);
  if (numSections < 1 || numSections > 100) status = STATUS_UNSUCCESSFUL;
  if (status != STATUS_SUCCESS) {
    ZwClose(fileHandle);
    return status;
  }

  // Allocate a SECTION_INFO structure for each section of the file

  sectionInfo = ExAllocatePool(NonPagedPool,
                               sizeof(SECTION_INFO) * numSections);
  if (sectionInfo == NULL) {
    ZwClose(fileHandle);
    return STATUS_INSUFFICIENT_RESOURCES;
  }

  // Fill in the array sectionInfo with the info for each section of the
  // file, for later use in converting virtual addresses to raw positions
  // in the file

  status = ReadSectionInfo(fileHandle, offsetToPEFileHeader, sectionInfo,
                           numSections);
  if (status != STATUS_SUCCESS) {
    ExFreePool(sectionInfo);
    ZwClose(fileHandle);
    return status;
  }

  // Read the directory for the export section, which contains, among
  // other things, the number of exported functions, which we call numFunctions

  status = GetImageExportDirectory(fileHandle, sectionInfo, numSections,
                                   offsetToPEFileHeader,
                                   &imageExportDirectory);
  numFunctions = MIN(imageExportDirectory.NumberOfFunctions,
                     imageExportDirectory.NumberOfNames);
  if (numFunctions < numSystemCalls || numFunctions > 25000)
    status = STATUS_UNSUCCESSFUL;
  if (status != STATUS_SUCCESS) {
    ExFreePool(sectionInfo);
    ZwClose(fileHandle);
    return status;
  }

  // Allocate an array of LONG's to hold the positions of all the function
  // name positions (the positions in the file of each function name)

  functionNamePositions = ExAllocatePool(NonPagedPool,
                                         sizeof(LONG) * numFunctions);
  if (functionNamePositions == NULL) {
    ExFreePool(sectionInfo);
    ZwClose(fileHandle);
    return status;
  }

  // Read the function name positions

  status = ReadFunctionNamePositions(fileHandle,
                                     (LONG)imageExportDirectory.AddressOfNames,
                                     sectionInfo, numSections,
                                     functionNamePositions, numFunctions);
  if (status != STATUS_SUCCESS) {
    ExFreePool(functionNamePositions);
    ExFreePool(sectionInfo);
    ZwClose(fileHandle);
    return status;
  }

  // Allocate an array of LONG's to hold the positions of all the functions

  functionPositions = ExAllocatePool(NonPagedPool,
                                     sizeof(LONG) * numFunctions);
  if (functionPositions == NULL) {
    ExFreePool(functionNamePositions);
    ExFreePool(sectionInfo);
    ZwClose(fileHandle);
    return status;
  }

  // Read the function positions

  status = ReadFunctionPositions(fileHandle,
                                 (LONG)imageExportDirectory.AddressOfFunctions,
                                 sectionInfo, numSections,
                                 functionPositions, numFunctions);
  if (status != STATUS_SUCCESS) {
    ExFreePool(functionPositions);
    ExFreePool(functionNamePositions);
    ExFreePool(sectionInfo);
    ZwClose(fileHandle);
    return status;
  }

  // Loop through all the functions.  For each function, read its name.
  // If that name starts with Zw, check to see if it matches one of the
  // system call names we're looking for.  If it does, read the corresponding
  // function to deduce the system call number.

  for (functionNumber = 0;
       functionNumber < numFunctions && systemCallsFound < numSystemCalls;
       functionNumber++) {
    status = ReadDLLFile(fileHandle, foundFunctionName,
                         functionNamePositions[functionNumber],
                         MAX_NAME_LENGTH);
    if (status != STATUS_SUCCESS) break;

    if (foundFunctionName[0] == 'Z' && foundFunctionName[1] == 'w') {
      for (i = 0; i < numSystemCalls; i++) {
        // Only check for a name match if this system call number hasn't
        // been found yet
        if (systemCallNumbers[i] == 0 &&
            !strcmp(systemCallNames[i], &foundFunctionName[2])) {
          systemCallNumbers[i] =
            GetSystemCallNumberFromFunction(fileHandle,
                                            functionPositions[functionNumber]);
          if (systemCallNumbers[i] != 0)
            systemCallsFound++;
          break;
        }
      }
    }
  }

  // If we didn't find all the system calls, consider it unsuccessful

  if (status == STATUS_SUCCESS && systemCallsFound != numSystemCalls)
    status = STATUS_UNSUCCESSFUL;

  // Release resources we allocated, and close the file (NTDLL.DLL) we opened

  ExFreePool(functionPositions);
  ExFreePool(functionNamePositions);
  ExFreePool(sectionInfo);
  ZwClose(fileHandle);

  return status;
}

////////////////////////////////////////////////////////////////////////////
//                            FILE OPERATIONS
////////////////////////////////////////////////////////////////////////////

// OpenDLLFile opens the NTDLL.DLL system file and returns the handle
// in fileHandle.  It returns a status indicating whether it succeeded.

NTSTATUS OpenDLLFile (HANDLE *fileHandle)
{
  NTSTATUS status;
  OBJECT_ATTRIBUTES objectAttributes;
  IO_STATUS_BLOCK ioStatusBlock;
  UNICODE_STRING fileNameUnicodeString;

  RtlInitUnicodeString(&fileNameUnicodeString,
                       L"\\SystemRoot\\system32\\ntdll.dll");
  objectAttributes.Length = sizeof(OBJECT_ATTRIBUTES);
  objectAttributes.RootDirectory = NULL;
  objectAttributes.ObjectName = &fileNameUnicodeString;
  objectAttributes.SecurityDescriptor = NULL;
  objectAttributes.SecurityQualityOfService = NULL;
  objectAttributes.Attributes = OBJ_CASE_INSENSITIVE;
  status = ZwCreateFile(fileHandle, FILE_READ_DATA | SYNCHRONIZE,
                        &objectAttributes, &ioStatusBlock, 0,
                        FILE_ATTRIBUTE_DIRECTORY,
                        FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
                        FILE_OPEN,
                        FILE_NON_DIRECTORY_FILE | FILE_SYNCHRONOUS_IO_ALERT,
                        NULL, 0);
  return status;
}

// ReadDLLFile reads a certain number of bytes (length) at a certain offset
// (offset) in an open file (with handle fileHandle) into a buffer (buf).
// It returns a status value.  It may return STATUS_SUCCESS even if it
// reaches end-of-file and thus does not read the full number of requested
// bytes.

NTSTATUS ReadDLLFile (HANDLE fileHandle, char *buf, ULONG offset, ULONG length)
{
  LARGE_INTEGER offsetAsLargeInteger;
  IO_STATUS_BLOCK ioStatusBlock;
  NTSTATUS status;

  offsetAsLargeInteger.HighPart = 0;
  offsetAsLargeInteger.LowPart = offset;
  status = ZwReadFile(fileHandle, NULL, NULL, NULL, &ioStatusBlock,
                      buf, length, &offsetAsLargeInteger, NULL);
  return status;
}

////////////////////////////////////////////////////////////////////////////
//                     PE FILE FORMAT PARSING ROUTINES
////////////////////////////////////////////////////////////////////////////

// GetVirtualToPositionConversion gets the offset you have to add to
// a virtual address to convert it into a file position.  The way this
// works is you pass one virtual address in the section you're interested
// in (virtualAddress), and a pointer to where you want the offset put
// (conversion).  You must also pass an array of section information
// structures (sectionInfo) and the number of structures in that array
// (numSections).  The section info contains information that allows
// virtual addresses in a certain range to be translated into file
// positions; this function finds the section that contains the passed
// virtual address and uses the information for that section.  It returns
// a status value.

NTSTATUS GetVirtualToPositionConversion (SECTION_INFO *sectionInfo,
                                         USHORT numSections,
                                         LONG virtualAddress,
                                         LONG *conversion)
{
  int i;

  // Loop through all the sections until we find one that contains
  // virtualAddress in it, then compute the conversion from the
  // information for that section

  for (i = 0; i < numSections; i++) {
    if (sectionInfo[i].virtualAddress <= virtualAddress &&
        sectionInfo[i].virtualAddress + sectionInfo[i].sizeOfRawData >
          virtualAddress) {
      *conversion = sectionInfo[i].pointerToRawData -
                    sectionInfo[i].virtualAddress;
      return STATUS_SUCCESS;
    }
  }

  // If none of the sections contained the address, return failure

  return STATUS_UNSUCCESSFUL;
}

// GetOffsetToPEFileHeader finds the offset to the PE file header in a file.
// It is passed an open file handle (fileHandle) and a pointer in which to
// put the offset (offsetToPEFileHeader).  It returns a status value.

NTSTATUS GetOffsetToPEFileHeader (HANDLE fileHandle,
                                  LONG *offsetToPEFileHeader)
{
  LONG offsetToLFAHeader;
  NTSTATUS status;
  ULONG signature;

  // Read the offset to the LFA header, which is at the end of the DOS
  // file header at the beginning of the file

  status = ReadDLLFile(fileHandle, (char *) &offsetToLFAHeader, 60, 4);
  if (status == STATUS_SUCCESS) {
    // Read the PE file header signature, which is at the beginning of the
    // LFA header
    status = ReadDLLFile(fileHandle, (char *) &signature, offsetToLFAHeader,
                         SIZE_OF_NT_SIGNATURE);
    if (status == STATUS_SUCCESS) {
      // Check that the signature is correct.  If so, return the position
      // of the PE file header, which is just past the signature.
      if (signature == 0x4550)
        *offsetToPEFileHeader = offsetToLFAHeader + SIZE_OF_NT_SIGNATURE;
      else
        status = STATUS_UNSUCCESSFUL;
    }
  }

  return status;
}

// GetNumberOfSections finds the number of sections in a PE format file.
// It needs the offset to the PE file header (offsetToPEFileHeader),
// since the number of sections is the 3rd and 4th bytes of that header.
// It also needs a handle to the open file (fileHandle) and a place to
// put the number of sections (numSections).  It returns a status value.

NTSTATUS GetNumberOfSections (HANDLE fileHandle, LONG offsetToPEFileHeader,
                              USHORT *numSections)
{
  return ReadDLLFile(fileHandle, (char *) numSections,
                     offsetToPEFileHeader + 2, 2);
}

// ReadSectionInfo reads the information we need about each section in a
// file to an array of SECTION_INFO structures (sectionInfo) of size
// numSections.  It requires the offset to the PE file header
// (offsetToPEFileHeader) and an open file handle (fileHandle).
// It returns a status value.

NTSTATUS ReadSectionInfo (HANDLE fileHandle, LONG offsetToPEFileHeader,
                          SECTION_INFO *sectionInfo, USHORT numSections)
{
  IMAGE_SECTION_HEADER imageSectionHeader;
  LONG sectionHeaderOffset;
  NTSTATUS status;
  int i;

  // The offset to the first section header is just past the optional
  // header, which is just past the PE file header

  sectionHeaderOffset = offsetToPEFileHeader + IMAGE_SIZEOF_FILE_HEADER +
                        IMAGE_SIZEOF_NT_OPTIONAL_HEADER;

  // Loop through all the sections, reading the image section header
  // and extracting its relevant information into sectionInfo[i]

  for (i = 0; i < numSections; i++) {
    status = ReadDLLFile(fileHandle, (char *) &imageSectionHeader,
                         sectionHeaderOffset, sizeof(IMAGE_SECTION_HEADER));
    if (status != STATUS_SUCCESS) return status;
    sectionInfo[i].virtualAddress = imageSectionHeader.VirtualAddress;
    sectionInfo[i].sizeOfRawData = imageSectionHeader.SizeOfRawData;
    sectionInfo[i].pointerToRawData = imageSectionHeader.PointerToRawData;
    // The next section header is just past the one just read
    sectionHeaderOffset += sizeof(IMAGE_SECTION_HEADER);
  }

  return STATUS_SUCCESS;
}

// GetImageExportDirectory gets the directory for the export section
// (which is not necessarily in the exports section) and puts it in
// imageExportDirectory.  It requires an open file handle (fileHandle)
// and the offset to the PE file header (offsetToPEFileHeader).  Also,
// since it needs to translate virtual addresses to file positions, it
// requires the array of section information (sectionInfo, and its
// length numSections).  It returns a status value.

NTSTATUS GetImageExportDirectory (HANDLE fileHandle, SECTION_INFO *sectionInfo,
                                  USHORT numSections,
                                  LONG offsetToPEFileHeader,
                                  IMAGE_EXPORT_DIRECTORY *imageExportDirectory)
{
  LONG exportDirectoryVirtualAddress, exportDirectoryPosition, conversion;
  NTSTATUS status;

  // Read the export directory virtual address, the 4-byte value that is
  // 96 bytes into the PE file header (the first 4-byte value of the
  // first data directory)

  status = ReadDLLFile(fileHandle, (char *) &exportDirectoryVirtualAddress,
                       offsetToPEFileHeader + IMAGE_SIZEOF_FILE_HEADER + 96,
                       4);
  if (status != STATUS_SUCCESS) return status;

  // Convert the export directory virtual address into a file position

  status = GetVirtualToPositionConversion(sectionInfo, numSections,
                                          exportDirectoryVirtualAddress,
                                          &conversion);
  if (status != STATUS_SUCCESS) return status;
  exportDirectoryPosition = exportDirectoryVirtualAddress + conversion;

  // Read the export directory from the file position

  status = ReadDLLFile(fileHandle, (char *)imageExportDirectory,
                       exportDirectoryPosition,
                       sizeof(IMAGE_EXPORT_DIRECTORY));
  return status;
}

// ReadFunctionNamePositions reads the list of function name positions
// into a passed array (functionNamePositions, with length numFunctions).
// It requires an open file handle (fileHandle) and the virtual address
// of the list of name virtual addresses (namesVirtualAddress).  It also
// requires an array of section information (sectionInfo, length numSections)
// to convert virtual addresses to file positions.  It returns a status value.

NTSTATUS ReadFunctionNamePositions (HANDLE fileHandle,
                                    LONG namesVirtualAddress,
                                    SECTION_INFO *sectionInfo,
                                    USHORT numSections,
                                    LONG *functionNamePositions,
                                    ULONG numFunctions)
{
  LONG conversion, namesPosition;
  ULONG i;
  NTSTATUS status;

  // Convert the virtual address where the name virtual addresses are to a
  // file position

  status = GetVirtualToPositionConversion(sectionInfo, numSections,
                                          namesVirtualAddress, &conversion);
  if (status != STATUS_SUCCESS) return status;
  namesPosition = namesVirtualAddress + conversion;

  // Read the list of name virtual addresses from that position

  status = ReadDLLFile(fileHandle, (char *) functionNamePositions,
                       namesPosition, numFunctions * 4);
  if (status != STATUS_SUCCESS) return status;

  // Now we want to convert all those virtual addresses we read into
  // functionNamePositions[] to file positions.  To do this, we obtain
  // the necessary conversion for the first of them, and add that value
  // to all the function name positions.  (All the function names should
  // be in the same section, so the same conversion will work for all
  // of them.)

  status = GetVirtualToPositionConversion(sectionInfo, numSections,
                                          functionNamePositions[0],
                                          &conversion);
  if (status != STATUS_SUCCESS) return status;

  if (conversion != 0)
    for (i = 0; i < numFunctions; i++)
      functionNamePositions[i] += conversion;

  return status;
}

// ReadFunctionPositions reads the list of function positions into a
// passed array (functionPositions, with length numFunctions).  It requires
// an open file handle (fileHandle) and the virtual address of the list of
// function virtual addresses (fnPtrsVirtualAddress).  It also requires
// an array of section information (sectionInfo, length numSections) to
// convert virtual addresses to file positions.  It returns a status value.

NTSTATUS ReadFunctionPositions (HANDLE fileHandle, LONG fnPtrsVirtualAddress,
                                SECTION_INFO *sectionInfo, USHORT numSections,
                                LONG *functionPositions, ULONG numFunctions)
{
  LONG conversion, fnPtrsPosition;
  ULONG i;
  NTSTATUS status;

  // Convert the virtual address where the function pointers are to a file
  // position

  status = GetVirtualToPositionConversion(sectionInfo, numSections,
                                          fnPtrsVirtualAddress, &conversion);
  if (status != STATUS_SUCCESS) return status;
  fnPtrsPosition = fnPtrsVirtualAddress + conversion;

  // Read the function virtual addresses from that position

  status = ReadDLLFile(fileHandle, (char *) functionPositions, fnPtrsPosition,
                       numFunctions * 4);
  if (status != STATUS_SUCCESS) return status;

  // Now we want to convert all those function virtual addresses we just
  // read into functionPositions[] to file positions.  To do this, we find
  // the appropriate virtual address to file position conversion for the
  // first function virtual address, and add it to all the function virtual
  // addresses.  (All the functions should be in the same section, so the
  // same conversion will work for all of them.)

  status = GetVirtualToPositionConversion(sectionInfo, numSections,
                                          functionPositions[0], &conversion);
  if (status != STATUS_SUCCESS) return status;

  if (conversion != 0)
    for (i = 0; i < numFunctions; i++)
      functionPositions[i] += conversion;

  return status;
}

// SkipPreambleIfPresent checks to see if a function begins with the
// strange 25-byte preamble seen in Windows NT Terminal Services for
// certain functions. If it does, it skips over this 25 bytes and puts
// the next 11 bytes in 'function'. It assumes that 'function' is an
// 11-byte array that contains the first 11 bytes of the function. It
// assumes that these function contents were read from position
// 'functionPosition' in the file whose handle is 'fileHandle'.

void SkipPreambleIfPresent (HANDLE fileHandle, char *function,
                            LONG functionPosition)
{
  NTSTATUS status;

  // First, see if the first 11 bytes are 64A11800000083B8AC0100.

  if (function[0] != '\144') return;
  if (function[1] != '\241') return;
  if (function[2] != '\030') return;
  if (function[3] != '\000') return;
  if (function[4] != '\000') return;
  if (function[5] != '\000') return;
  if (function[6] != '\203') return;
  if (function[7] != '\270') return;
  if (function[8] != '\254') return;
  if (function[9] != '\001') return;
  if (function[10] != '\000') return;

  // If they match, read the next 10 bytes and make sure they are
  // 0000750A8D44240450E8.

  status = ReadDLLFile(fileHandle, function, functionPosition + 11, 10);
  if (status != STATUS_SUCCESS) return;
  if (function[0] != '\000') return;
  if (function[1] != '\000') return;
  if (function[2] != '\165') return;
  if (function[3] != '\012') return;
  if (function[4] != '\215') return;
  if (function[5] != '\104') return;
  if (function[6] != '\044') return;
  if (function[7] != '\004') return;
  if (function[8] != '\120') return;
  if (function[9] != '\350') return;

  // If they match, skip 25 bytes past the original function position
  // and read the 11 bytes that follow. This is the "true" beginning
  // of the function as far as we're concerned. Note that we never
  // check the last four bytes of the preamble; this is because they
  // can vary, depending on where the final call is to.

  status = ReadDLLFile(fileHandle, function, functionPosition + 25, 11);
}

// GetSystemCallNumberFromFunction reads enough of a function to extract
// the system call number it invokes, and returns that system call number.
// It is passed an open file handle (fileHandle) containing the function
// at a given position (functionPosition).  It returns 0 on failure, such
// as if the function read does not correspond to the invoking of a system
// call.

ULONG GetSystemCallNumberFromFunction (HANDLE fileHandle,
                                       LONG functionPosition)
{
  char function[11];
  ULONG retVal;
  NTSTATUS status;

  // Read the first 11 bytes of the function into 'function'.

  status = ReadDLLFile(fileHandle, function, functionPosition, 11);
  if (status != STATUS_SUCCESS) return 0;

  // If the function has a strange Terminal Services-style preamble,
  // skip over it and replace it with the "real" first 11 bytes.
  
  SkipPreambleIfPresent(fileHandle, function, functionPosition);

  // Check to make sure the 11 bytes are consistent with the way
  // system call invocations are made

  if (function[0] != '\270') return 0;    // 0xB8 -- mov eax, ...
  if (function[5] != '\215') return 0;    // 0x8D `
  if (function[6] != '\124') return 0;    // 0x54 |__ lea edx,
  if (function[7] != '\044') return 0;    // 0x24 |       [esp+4]
  if (function[8] != '\004') return 0;    // 0x04 /
  if (function[9] != '\315') return 0;    // 0xCD `__ int 2Eh
  if (function[10] != '\056') return 0;   // 0x2E /

  // If that check passed, bytes 1-4 should be the system call number,
  // so copy those into retVal

  RtlCopyMemory(&retVal, &function[1], 4);

  // Do one more check, to make sure that the system call number is in
  // the range of valid system call numbers

  if (retVal < KeServiceDescriptorTable->LowCall) return 0;
  if (retVal > KeServiceDescriptorTable->HiCall) return 0;

  // OK, everything's fine, so return the system call number

  return retVal;
}
