// 
// VTrcLog.c
//
// Code for the VTrcLog driver, the driver for the VTrace logger device.
// The logger device's main responsibilities are to process requests to add
// events to the log, and to periodically flush the log to disk.  However,
// it does several other things, including the following:
//   * It traces some events of its own and adds those to the log.
//   * It creates file-system and network filter devices for every file
//     system or network device in use.
//   * It processes other types of requests, e.g. to change the event mask.
//

#include "NTDDK.h"
#include "../Common/NTDDKBD.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "../Common/RegConstraints.h"
#include "../Common/Version.h"
#include "Globals.h"
#include "Activity.h"
#include "Beat.h"
#include "EventMask.h"
#include "Flush.h"
#include "GetState.h"
#include "LogEvent.h"
#include "MiscHook.h"
#include "ProcList.h"
#include "StartStop.h"
#include "Unicode.h"
#include "ZwHooks.h"

////////////////////////////////////////////////////////////////////////////
//                               MACROS
////////////////////////////////////////////////////////////////////////////

// The following flags are used to keep track of which resources have
// been allocated so far for the VTrcLog driver.  Thus, in case of
// fatal error, if we have to exit, we can free these resources.

#define TEMPORARY_UNICODE_RESOURCE_FLAG          1
#define MAIN_DIRECTORY_UNICODE_RESOURCE_FLAG     2
#define TRACE_FILE_UNICODE_RESOURCE_FLAG         4
#define LOG_DEVICE_RESOURCE_FLAG                 8
#define LOG_BUFFER_RESOURCE_FLAG                 16
#define HOOK_ZW_FUNCTIONS_RESOURCE_FLAG          32
#define SYMBOLIC_LINK_RESOURCE_FLAG              64
#define PROCESS_NOTIFY_ROUTINE_FLAG              128
#define THREAD_NOTIFY_ROUTINE_FLAG               256
#define CONTEXT_SWAP_HOOK_FLAG                   512
#define SET_SHARED_STATE_FIELDS_FLAG             1024
#define USER_NAME_REQUIRED_EVENT_FLAG            2048
#define TRACE_DIRECTORY_HANDLE_RESOURCE_FLAG     4096

// This macro checks whether the status variable (presumably, set by
// the last operation) is STATUS_SUCCESS.  If it isn't, it releases
// any resources so far allocated and returns the status.  It's a
// macro because it's used really often, and I didn't want to clutter
// the code with it over and over agin.

#define CHECK \
  if (status != STATUS_SUCCESS) { \
    VTraceReleaseResources(resourcesObtained, logDeviceObject, \
                           dummyLogEventFunc); \
    return status; \
  }

////////////////////////////////////////////////////////////////////////////
//                             STRUCTURES
////////////////////////////////////////////////////////////////////////////

// Structure documented in Rajeev Nagar's book.  It is used by
// ZwQueryVolumeInformationFile to find out how much free space
// is left in a partition.

typedef struct _FILE_FS_SIZE_INFORMATION {
  LARGE_INTEGER   TotalAllocationUnits;
  LARGE_INTEGER   AvailableAllocationUnits;
  ULONG           SectorsPerAllocationUnit;
  ULONG           BytesPerSector;
} FILE_FS_SIZE_INFORMATION, *PFILE_FS_SIZE_INFORMATION;

////////////////////////////////////////////////////////////////////////////
//                               GLOBALS
////////////////////////////////////////////////////////////////////////////

// All the global variables are in one neat structure:

VTRACE_GLOBALS globals;

////////////////////////////////////////////////////////////////////////////
//                           FORWARD DEFINES
////////////////////////////////////////////////////////////////////////////

NTSTATUS DriverEntry (IN PDRIVER_OBJECT  DriverObject,
                      IN PUNICODE_STRING RegistryPath);
void VTraceReleaseResources (ULONG resourcesObtained,
                             PDEVICE_OBJECT logDeviceObject,
                             LOG_FUNC_TYPE dummyLogEventFunc);

NTSTATUS GetKeyboardState (KEYBOARD_INDICATOR_PARAMETERS *keyboardStatePtr);
BOOLEAN IsUserOverridingLoad (void);

NTSTATUS VTrcLogDispatchControl (IN PDEVICE_OBJECT devObj, IN PIRP irp);
NTSTATUS VTrcLogDispatchShutdown (IN PDEVICE_OBJECT devObj, IN PIRP irp);
NTSTATUS VTrcLogDispatchGeneral (IN PDEVICE_OBJECT devObj, IN PIRP irp);

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, DriverEntry)
#pragma alloc_text (INIT, VTraceReleaseResources)
#pragma alloc_text (INIT, GetKeyboardState)
#pragma alloc_text (INIT, IsUserOverridingLoad)
#endif

////////////////////////////////////////////////////////////////////////////
//                          FUNCTION PROTOTYPES
////////////////////////////////////////////////////////////////////////////

// Function documented in Rajeev Nagar's book, used here to find out
// how much free space is left in a partition.

NTSTATUS ZwQueryVolumeInformationFile
  (IN HANDLE FileHandle, OUT PIO_STATUS_BLOCK IoStatusBlock,
   OUT PVOID FsInformation, IN ULONG Length,
   IN FS_INFORMATION_CLASS FsInformationClass);

////////////////////////////////////////////////////////////////////////////
//                             DRIVER ENTRY
////////////////////////////////////////////////////////////////////////////

// DriverEntry is the installable driver initialization.  Here we set
// up the driver, create our one logger device, and do all sorts of
// miscellaneous initialization.

NTSTATUS DriverEntry (IN PDRIVER_OBJECT  DriverObject,
                      IN PUNICODE_STRING RegistryPath)
{
  NTSTATUS status = STATUS_SUCCESS;
  UNICODE_STRING logDeviceNameUnicodeString;
  UNICODE_STRING logDeviceLinkUnicodeString;
  OBJECT_ATTRIBUTES objectAttributes;
  IO_STATUS_BLOCK ioStatusBlock;
  FILE_FS_SIZE_INFORMATION sizeInfo;
  struct {
    ULONG   TitleIndex;
    ULONG   Type;
    ULONG   DataLength;
    WCHAR   Data[256];
  } keyValueInfo;
  HANDLE keyHandle;
  ULONG keyValueInfoLength;
  KIRQL oldirql;
  ULONG resourcesObtained;
  LOG_FUNC_TYPE dummyLogEventFunc = NULL;
  PDEVICE_OBJECT logDeviceObject = NULL;

  DebugPrint(("VTrcLog: Entering DriverEntry\n"));

  // Check to make sure the user isn't forcing us to not load by holding
  // the caps- or scroll-lock light on.

  if (IsUserOverridingLoad()) {
    DebugPrint(("VTrcLog: Not loading because LED light is on.\n"));
    return STATUS_UNSUCCESSFUL;
  }

  // Get a pointer to the shared state from VTrcInit. If we can't get it,
  // VTrace is deactivated this boot, and we should exit. (Or something
  // even more wrong is going on, in which case we should also exit.)

  status = GetSharedState();
  if (!NT_SUCCESS(status)) {
    DebugPrint(("VTrcLog: Couldn't read shared state.\n"));
    return status;
  }

  // Initialize resources-obtained field so that we can know what resources
  // have ben obtained if we need to free them in case of error

  resourcesObtained = 0;

  // Initialize logging to "disabled because VTrcLog has not yet loaded".

  globals.whyLoggingDisabled = BECAUSE_VTRCLOG_UNLOADED;

  // Initialize flush-in-progress indicator

  globals.flushInProgress = FALSE;

  // Initialize activity indicators for all thread ID's

  InitializeThreadActivityElements();

  // Initialize the info for each entry type

  InitializeEntryTypeInfo();

  // Allocate temporary Unicode string

  status = MyInitUnicodeString(&globals.temporaryUnicodeString, 128); CHECK;
  resourcesObtained |= TEMPORARY_UNICODE_RESOURCE_FLAG;

  // Read registry parameters: free bytes of disk space needed to continue
  // logging, initial event mask, main directory, size of log buffer, and
  // version of VTrace

  status = RtlAppendUnicodeToString(&globals.temporaryUnicodeString,
       L"\\Registry\\Machine\\Software\\Jay Lorch\\VTrace\\CurrentVersion");
  CHECK;

  objectAttributes.Length = sizeof(OBJECT_ATTRIBUTES);
  objectAttributes.RootDirectory = NULL;
  objectAttributes.ObjectName = &globals.temporaryUnicodeString;
  objectAttributes.SecurityDescriptor = NULL;
  objectAttributes.SecurityQualityOfService = NULL;
  objectAttributes.Attributes = OBJ_CASE_INSENSITIVE;
  status = ZwOpenKey(&keyHandle, KEY_QUERY_VALUE, &objectAttributes); CHECK;

  MyClearUnicodeString(&globals.temporaryUnicodeString);
  RtlAppendUnicodeToString(&globals.temporaryUnicodeString, L"Version");
  status = ZwQueryValueKey(keyHandle, &globals.temporaryUnicodeString,
                           KeyValuePartialInformation, &keyValueInfo,
                           sizeof(keyValueInfo), &keyValueInfoLength);
  if (status == STATUS_SUCCESS && keyValueInfo.Type == REG_DWORD)
    globals.vtraceVersion = * (ULONG *) keyValueInfo.Data;
  else
    globals.vtraceVersion = 0;

  MyClearUnicodeString(&globals.temporaryUnicodeString);
  RtlAppendUnicodeToString(&globals.temporaryUnicodeString, L"FreeMBNeeded");
  status = ZwQueryValueKey(keyHandle, &globals.temporaryUnicodeString,
                           KeyValuePartialInformation, &keyValueInfo,
                           sizeof(keyValueInfo), &keyValueInfoLength);
  if (status == STATUS_SUCCESS && keyValueInfo.Type == REG_DWORD) {
    globals.freeBytesNeeded = * (ULONG *) keyValueInfo.Data;
    if (globals.freeBytesNeeded < MIN_FREE_SPACE ||
        globals.freeBytesNeeded > MAX_FREE_SPACE)
      globals.freeBytesNeeded = DEFAULT_FREE_SPACE;
    globals.freeBytesNeeded *= 1048576;     // Convert MB to bytes
  }
  else
    globals.freeBytesNeeded = DEFAULT_FREE_SPACE * 1048576;

  MyClearUnicodeString(&globals.temporaryUnicodeString);
  RtlAppendUnicodeToString(&globals.temporaryUnicodeString,
                           L"MaxTracePieceMB");
  status = ZwQueryValueKey(keyHandle, &globals.temporaryUnicodeString,
                           KeyValuePartialInformation, &keyValueInfo,
                           sizeof(keyValueInfo), &keyValueInfoLength);
  if (status == STATUS_SUCCESS && keyValueInfo.Type == REG_DWORD) {
    globals.maxTracePieceBytes = * (ULONG *) keyValueInfo.Data;
    if (globals.maxTracePieceBytes < MIN_TRACE_PIECE_SIZE_LIMIT ||
        globals.maxTracePieceBytes > MAX_TRACE_PIECE_SIZE_LIMIT)
      globals.maxTracePieceBytes = DEFAULT_TRACE_PIECE_SIZE_LIMIT;
    globals.maxTracePieceBytes *= 1048576;     // Convert MB to bytes
  }
  else
    globals.maxTracePieceBytes = DEFAULT_TRACE_PIECE_SIZE_LIMIT * 1048576;

  globals.currentEventMasks = EVENT_MASK_BIT_FOR_MASK_0;
  MyClearUnicodeString(&globals.temporaryUnicodeString);
  RtlAppendUnicodeToString(&globals.temporaryUnicodeString, L"EventMask");
  status = ZwQueryValueKey(keyHandle, &globals.temporaryUnicodeString,
                           KeyValuePartialInformation, &keyValueInfo,
                           sizeof(keyValueInfo), &keyValueInfoLength);
  if (status == STATUS_SUCCESS && keyValueInfo.Type == REG_BINARY &&
      keyValueInfo.DataLength >= 32)
    SetUserDefinedEventMask((char *) keyValueInfo.Data);

  MyClearUnicodeString(&globals.temporaryUnicodeString);
  RtlAppendUnicodeToString(&globals.temporaryUnicodeString, L"LogBufferKB");
  status = ZwQueryValueKey(keyHandle, &globals.temporaryUnicodeString,
                           KeyValuePartialInformation, &keyValueInfo,
                           sizeof(keyValueInfo), &keyValueInfoLength);

  // Multiply by 1024 in order to convert from KB to bytes.

  if (status == STATUS_SUCCESS && keyValueInfo.Type == REG_DWORD) {
    globals.alignedLengthOfLogBuffer = * (ULONG *) keyValueInfo.Data;
    if (globals.alignedLengthOfLogBuffer < MIN_LOG_BUFFER_SIZE ||
        globals.alignedLengthOfLogBuffer > MAX_LOG_BUFFER_SIZE)
      globals.alignedLengthOfLogBuffer = DEFAULT_LOG_BUFFER_SIZE;
    globals.alignedLengthOfLogBuffer *= 1024;
  }
  else
    globals.alignedLengthOfLogBuffer = DEFAULT_LOG_BUFFER_SIZE * 1024;

  MyClearUnicodeString(&globals.temporaryUnicodeString);
  RtlAppendUnicodeToString(&globals.temporaryUnicodeString, L"MainDirectory");
  status = ZwQueryValueKey(keyHandle, &globals.temporaryUnicodeString,
                           KeyValuePartialInformation, &keyValueInfo,
                           sizeof(keyValueInfo), &keyValueInfoLength); CHECK;
  ZwClose(keyHandle);
  if (keyValueInfo.Type != REG_SZ) {
    VTraceReleaseResources(resourcesObtained, logDeviceObject,
                           dummyLogEventFunc);
    return STATUS_INVALID_PARAMETER;
  }

  // Allocate main directory Unicode string and put the just-read value in it
  // (preceded by \??\)

  status = MyInitUnicodeString(&globals.mainDirectoryUnicodeString,
                               keyValueInfo.DataLength / 2 + 4); CHECK;
  resourcesObtained |= MAIN_DIRECTORY_UNICODE_RESOURCE_FLAG;

  status = RtlAppendUnicodeToString(&globals.mainDirectoryUnicodeString,
                                    L"\\??\\"); CHECK;
  status = RtlAppendUnicodeToString(&globals.mainDirectoryUnicodeString,
                                    keyValueInfo.Data); CHECK;

  // Allocate trace file Unicode string

  status = MyInitUnicodeString(&globals.traceFileUnicodeString,
                               keyValueInfo.DataLength / 2 + 24); CHECK;
  resourcesObtained |= TRACE_FILE_UNICODE_RESOURCE_FLAG;

  // Open the trace directory

  RtlCopyUnicodeString(&globals.traceFileUnicodeString,
                       &globals.mainDirectoryUnicodeString);
  status = RtlAppendUnicodeToString(&globals.traceFileUnicodeString,
                                    L"\\Traces");
  CHECK;

  objectAttributes.Length = sizeof(OBJECT_ATTRIBUTES);
  objectAttributes.RootDirectory = NULL;
  objectAttributes.ObjectName = &globals.traceFileUnicodeString;
  objectAttributes.SecurityDescriptor = NULL;
  objectAttributes.SecurityQualityOfService = NULL;
  objectAttributes.Attributes = OBJ_CASE_INSENSITIVE;
  status = ZwCreateFile(&globals.systemTraceDirectoryHandle,
                        FILE_ADD_FILE | FILE_READ_ATTRIBUTES,
                        &objectAttributes, &ioStatusBlock, 0,
                        FILE_ATTRIBUTE_DIRECTORY,
                        FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
                        FILE_OPEN,
                        FILE_DIRECTORY_FILE | FILE_SYNCHRONOUS_IO_ALERT, NULL,
                        0);
  CHECK;
  resourcesObtained |= TRACE_DIRECTORY_HANDLE_RESOURCE_FLAG;

  // Obtain the number of bytes per sector on the disk containing the
  // trace directory.  Make sure the log buffer size is an integer multiple
  // of this number.

  status = ZwQueryVolumeInformationFile(globals.systemTraceDirectoryHandle,
                                        &ioStatusBlock, &sizeInfo,
                                        sizeof(sizeInfo),
                                        FileFsSizeInformation); CHECK;
  if (globals.alignedLengthOfLogBuffer % sizeInfo.BytesPerSector) {
    globals.alignedLengthOfLogBuffer += sizeInfo.BytesPerSector;
    globals.alignedLengthOfLogBuffer -=
      (globals.alignedLengthOfLogBuffer % sizeInfo.BytesPerSector);
  }

  // Get from the registry the number of the last trace file written

  MyClearUnicodeString(&globals.temporaryUnicodeString);
  status = RtlAppendUnicodeToString(&globals.temporaryUnicodeString,
       L"\\Registry\\Machine\\Software\\Jay Lorch\\VTrace\\CurrentVersion");
  CHECK;

  objectAttributes.Length = sizeof(OBJECT_ATTRIBUTES);
  objectAttributes.RootDirectory = NULL;
  objectAttributes.ObjectName = &globals.temporaryUnicodeString;
  objectAttributes.SecurityDescriptor = NULL;
  objectAttributes.SecurityQualityOfService = NULL;
  objectAttributes.Attributes = OBJ_CASE_INSENSITIVE;

  status = ZwOpenKey(&keyHandle, KEY_QUERY_VALUE | KEY_SET_VALUE,
                     &objectAttributes); CHECK;

  MyClearUnicodeString(&globals.temporaryUnicodeString);
  RtlAppendUnicodeToString(&globals.temporaryUnicodeString,
                           L"LastTraceNumber");
  status = ZwQueryValueKey(keyHandle, &globals.temporaryUnicodeString,
                           KeyValuePartialInformation, &keyValueInfo,
                           sizeof(keyValueInfo), &keyValueInfoLength);
  if (status == STATUS_SUCCESS && keyValueInfo.Type == REG_DWORD)
    globals.currentTraceFileNumber = (* (ULONG *) keyValueInfo.Data);
  else
    globals.currentTraceFileNumber = 0;
  globals.currentTraceFileNumberReserved = TRUE;
  globals.currentTraceFileTerminated = TRUE;

  // It can happen that, because the machine was rebooted after the last
  // trace file was created but before the registry information reflecting
  // that was flushed, a trace file numbered higher than the "last trace
  // number" in the registry already exists.  If this is the case, we
  // increment the trace number and make the registry reflect this.

  while (NextTraceFileExists()) {
    globals.currentTraceFileNumber++;
    MyClearUnicodeString(&globals.temporaryUnicodeString);
    RtlAppendUnicodeToString(&globals.temporaryUnicodeString,
                             L"LastTraceNumber");
    status = ZwSetValueKey(keyHandle, &globals.temporaryUnicodeString, 0,
                           REG_DWORD, &globals.currentTraceFileNumber, 4);
    if (status != STATUS_SUCCESS) break;
  }

  ZwClose(keyHandle);

  // Read CPU speed from registry

  MyClearUnicodeString(&globals.temporaryUnicodeString);
  status = RtlAppendUnicodeToString(&globals.temporaryUnicodeString,
   L"\\Registry\\Machine\\Hardware\\Description\\System\\CentralProcessor\\0");
  CHECK;

  objectAttributes.Length = sizeof(OBJECT_ATTRIBUTES);
  objectAttributes.RootDirectory = NULL;
  objectAttributes.ObjectName = &globals.temporaryUnicodeString;
  objectAttributes.SecurityDescriptor = NULL;
  objectAttributes.SecurityQualityOfService = NULL;
  objectAttributes.Attributes = OBJ_CASE_INSENSITIVE;

  status = ZwOpenKey(&keyHandle, KEY_QUERY_VALUE, &objectAttributes); CHECK;

  MyClearUnicodeString(&globals.temporaryUnicodeString);
  RtlAppendUnicodeToString(&globals.temporaryUnicodeString, L"~MHz");
  status = ZwQueryValueKey(keyHandle, &globals.temporaryUnicodeString,
                           KeyValuePartialInformation, &keyValueInfo,
                           sizeof(keyValueInfo), &keyValueInfoLength);
  ZwClose(keyHandle);
  if (status == STATUS_SUCCESS && keyValueInfo.Type == REG_DWORD)
    globals.CPUSpeedInMHz = (* (ULONG *) keyValueInfo.Data);
  else {
    VTraceReleaseResources(resourcesObtained, logDeviceObject,
                           dummyLogEventFunc);
    return STATUS_INVALID_PARAMETER;
  }

  // Initialize log buffer. Make sure to leave room for the largest
  // possible entry to wrap around the end.

  globals.logBufferContents = ExAllocatePool(NonPagedPool,
                                             globals.alignedLengthOfLogBuffer +
                                             MAX_LOG_ENTRY_SIZE);
  if (globals.logBufferContents == NULL) {
    DebugPrint(("VTrcLog: Could not allocate log buffer.\n"));
    VTraceReleaseResources(resourcesObtained, logDeviceObject,
                           dummyLogEventFunc);
    return STATUS_INSUFFICIENT_RESOURCES;
  }
  else {
    resourcesObtained |= LOG_BUFFER_RESOURCE_FLAG;
    globals.positionOfLogStart = 0;
    globals.positionOfLogEnd = 0;
    globals.logLen = 0;
    globals.logBytesPastAlignedPart = 0;
    globals.logDangerLen = globals.alignedLengthOfLogBuffer / 4 * 3;
    dummyLogEventFunc = sharedState->logEventFunc;
    sharedState->logEventFunc = &LogEvent;
    sharedState->checkIfLoggingFunc = &CheckIfLoggingEventType;
#if TARGET_NT_VERSION == 5
    sharedState->hibernationNotifyFunc = &HibernationNotifyRoutine;
#endif
    resourcesObtained |= SET_SHARED_STATE_FIELDS_FLAG;
  }

  // Create dispatch points for all the IRPs we handle.

  DriverObject->MajorFunction[IRP_MJ_CREATE] =
  DriverObject->MajorFunction[IRP_MJ_CLOSE] =
  DriverObject->MajorFunction[IRP_MJ_CLEANUP] = VTrcLogDispatchGeneral;

  DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = VTrcLogDispatchControl;
  DriverObject->MajorFunction[IRP_MJ_SHUTDOWN] = VTrcLogDispatchShutdown;

  // Note: we don't set DriverUnload since we can't remove the
  // thread-notify routine.

  // Create the one and only VTrcLog device

  RtlInitUnicodeString(&logDeviceNameUnicodeString, L"\\Device\\VTrcLog");
  status = IoCreateDevice(DriverObject, 0, &logDeviceNameUnicodeString,
                          FILE_DEVICE_VTRACE_LOG, 0, FALSE,
                          &logDeviceObject); CHECK;
  resourcesObtained |= LOG_DEVICE_RESOURCE_FLAG;

  // Create a symbolic link for the VTrcLog device so that it can be accessed
  // from user level.

  RtlInitUnicodeString(&logDeviceLinkUnicodeString, L"\\DosDevices\\VTrcLog");
  status = IoCreateSymbolicLink(&logDeviceLinkUnicodeString,
                                &logDeviceNameUnicodeString); CHECK;
  resourcesObtained |= SYMBOLIC_LINK_RESOURCE_FLAG;

  // Register the VTrcLog device to be notified of system shutdown
  // (via an IRP) so it can stop logging then

  IoRegisterShutdownNotification(logDeviceObject);

  // Create synchronization event to communicate with VTrcSrvc

  MyClearUnicodeString(&globals.temporaryUnicodeString);
  RtlAppendUnicodeToString(&globals.temporaryUnicodeString,
              L"\\BaseNamedObjects\\VTraceUserNameRequiredEvent");
  globals.userNameRequiredEvent =
    IoCreateSynchronizationEvent(&globals.temporaryUnicodeString,
                                 &globals.userNameRequiredEventHandle);
  if (globals.userNameRequiredEvent != NULL)
    resourcesObtained |= USER_NAME_REQUIRED_EVENT_FLAG;

  // Establish hooks

  status = InstallZwHooks(); CHECK;
  resourcesObtained |= HOOK_ZW_FUNCTIONS_RESOURCE_FLAG;

  status = PsSetCreateProcessNotifyRoutine(MyProcessNotifyRoutine, FALSE);
  CHECK;
  resourcesObtained |= PROCESS_NOTIFY_ROUTINE_FLAG;

  status = SetUpContextSwapHook();
  if (status == STATUS_SUCCESS)
    resourcesObtained |= CONTEXT_SWAP_HOOK_FLAG;
  else {
    DebugPrint(("VTrcLog: Could not set up swap-context hook.\n"));
  }

  status = PsSetCreateThreadNotifyRoutine(MyThreadNotifyRoutine); CHECK;
  resourcesObtained |= THREAD_NOTIFY_ROUTINE_FLAG;

  // Now that the thread-notify routine has been installed, we must return
  // STATUS_SUCCESS.  (This is because we don't have a way of uninstalling a
  // thread-notify routine.)

  // Initialize the timers and DPC's needed to watch for user
  // inactivity and available disk space, and to log a beat and change
  // the event mask periodically.

  InitializeDiskSpaceWatch();
  InitializeUserInactivityWatch();
  InitializeBeat();
  InitializeEventMaskChanging();

  // Start logging

  RemoveReasonsForNotLogging(BECAUSE_VTRCLOG_UNLOADED);

  return status;
}

////////////////////////////////////////////////////////////////////////////
//                           RESOURCE MANAGEMENT
////////////////////////////////////////////////////////////////////////////

// VTraceReleaseResources releases any resources we have allocated for
// VTrcLog so far.  The flags in resourcesObtained lists those resources
// that have been allocated.

void VTraceReleaseResources (ULONG resourcesObtained,
                             PDEVICE_OBJECT logDeviceObject,
                             LOG_FUNC_TYPE dummyLogEventFunc)
{
  UNICODE_STRING deviceLinkUnicodeString;
  KIRQL oldirql;

  if (resourcesObtained & SET_SHARED_STATE_FIELDS_FLAG) {
    sharedState->logEventFunc = dummyLogEventFunc;
    sharedState->checkIfLoggingFunc = NULL;
#if TARGET_NT_VERSION == 5
    sharedState->hibernationNotifyFunc = NULL;
#endif
  }
  if (resourcesObtained & PROCESS_NOTIFY_ROUTINE_FLAG)
    PsSetCreateProcessNotifyRoutine(MyProcessNotifyRoutine, TRUE /* remove */);
  if (resourcesObtained & HOOK_ZW_FUNCTIONS_RESOURCE_FLAG)
    UninstallZwHooks();
  if (resourcesObtained & CONTEXT_SWAP_HOOK_FLAG)
    RestoreOldContextSwapRoutine();
  if (resourcesObtained & SYMBOLIC_LINK_RESOURCE_FLAG) {
    RtlInitUnicodeString(&deviceLinkUnicodeString, L"\\DosDevices\\VTrcLog");
    IoDeleteSymbolicLink(&deviceLinkUnicodeString);
  }
  if (resourcesObtained & LOG_DEVICE_RESOURCE_FLAG)
    IoDeleteDevice(logDeviceObject);
  if (resourcesObtained & TEMPORARY_UNICODE_RESOURCE_FLAG)
    MyFreeUnicodeString(&globals.temporaryUnicodeString);
  if (resourcesObtained & MAIN_DIRECTORY_UNICODE_RESOURCE_FLAG)
    MyFreeUnicodeString(&globals.mainDirectoryUnicodeString);
  if (resourcesObtained & TRACE_FILE_UNICODE_RESOURCE_FLAG)
    MyFreeUnicodeString(&globals.traceFileUnicodeString);
  if (resourcesObtained & LOG_BUFFER_RESOURCE_FLAG)
    ExFreePool(globals.logBufferContents);
  if (resourcesObtained & USER_NAME_REQUIRED_EVENT_FLAG)
    ZwClose(globals.userNameRequiredEventHandle);
  if (resourcesObtained & TRACE_DIRECTORY_HANDLE_RESOURCE_FLAG)
    ZwClose(globals.systemTraceDirectoryHandle);

  // There's not much we can do about THREAD_NOTIFY_ROUTINE_FLAG,
  // so just ignore it
}

////////////////////////////////////////////////////////////////////////////
//                         REGISTRY RELOADING
////////////////////////////////////////////////////////////////////////////

// ReloadRegistry is called to reload parameters from the registry. At
// the moment, the only such parameter worth reloading is the free MB
// needed parameter. We also read the current version from the registry,
// because if it isn't the same as it was when VTrcLog started up, it
// could be unwise to update our parameter based on a new version's idea
// of what the registry parameters mean.

NTSTATUS ReloadRegistry (void)
{
  NTSTATUS status;
  UNICODE_STRING string;
  OBJECT_ATTRIBUTES objectAttributes;
  struct {
    ULONG   TitleIndex;
    ULONG   Type;
    ULONG   DataLength;
    WCHAR   Data[2];
  } keyValueInfo;
  HANDLE keyHandle;
  ULONG keyValueInfoLength;

  DebugPrint(("VTrcLog: Reloading registry\n"));

  // Initialize the Unicode string to be used in this function

  status = MyInitUnicodeString(&string, 128);
  if (status != STATUS_SUCCESS)
    return status;

  // Open the VTrace registry key

  status = RtlAppendUnicodeToString(&string,
       L"\\Registry\\Machine\\Software\\Jay Lorch\\VTrace\\CurrentVersion");
  if (status != STATUS_SUCCESS) {
    MyFreeUnicodeString(&string);
    return status;
  }

  objectAttributes.Length = sizeof(OBJECT_ATTRIBUTES);
  objectAttributes.RootDirectory = NULL;
  objectAttributes.ObjectName = &string;
  objectAttributes.SecurityDescriptor = NULL;
  objectAttributes.SecurityQualityOfService = NULL;
  objectAttributes.Attributes = OBJ_CASE_INSENSITIVE;
  status = ZwOpenKey(&keyHandle, KEY_QUERY_VALUE, &objectAttributes);
  if (status != STATUS_SUCCESS) {
    MyFreeUnicodeString(&string);
    return status;
  }

  // Read the version number.  If it isn't the same as when VTrcLog started
  // up, exit without reading any further parameters.

  MyClearUnicodeString(&string);
  RtlAppendUnicodeToString(&string, L"Version");
  status = ZwQueryValueKey(keyHandle, &string, KeyValuePartialInformation,
                           &keyValueInfo, sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (status != STATUS_SUCCESS || keyValueInfo.Type != REG_DWORD ||
      globals.vtraceVersion != * (ULONG *) keyValueInfo.Data) {
    ZwClose(keyHandle);
    MyFreeUnicodeString(&string);
    if (status == STATUS_SUCCESS)
      return STATUS_UNSUCCESSFUL;
    else
      return status;
  }

  // Read the free MB needed parameter and update the global field
  // reflecting it.

  MyClearUnicodeString(&string);
  RtlAppendUnicodeToString(&string, L"FreeMBNeeded");
  status = ZwQueryValueKey(keyHandle, &string, KeyValuePartialInformation,
                           &keyValueInfo, sizeof(keyValueInfo),
                           &keyValueInfoLength);
  if (status == STATUS_SUCCESS && keyValueInfo.Type == REG_DWORD) {
    globals.freeBytesNeeded = * (ULONG *) keyValueInfo.Data;
    if (globals.freeBytesNeeded < MIN_FREE_SPACE ||
        globals.freeBytesNeeded > MAX_FREE_SPACE)
      globals.freeBytesNeeded = DEFAULT_FREE_SPACE;
    globals.freeBytesNeeded *= 1048576;     // Convert MB to bytes
  }
  else
    globals.freeBytesNeeded = DEFAULT_FREE_SPACE * 1048576;

  MyClearUnicodeString(&globals.temporaryUnicodeString);
  RtlAppendUnicodeToString(&globals.temporaryUnicodeString,
                           L"MaxTracePieceMB");
  status = ZwQueryValueKey(keyHandle, &globals.temporaryUnicodeString,
                           KeyValuePartialInformation, &keyValueInfo,
                           sizeof(keyValueInfo), &keyValueInfoLength);
  if (status == STATUS_SUCCESS && keyValueInfo.Type == REG_DWORD) {
    globals.maxTracePieceBytes = * (ULONG *) keyValueInfo.Data;
    if (globals.maxTracePieceBytes < MIN_TRACE_PIECE_SIZE_LIMIT ||
        globals.maxTracePieceBytes > MAX_TRACE_PIECE_SIZE_LIMIT)
      globals.maxTracePieceBytes = DEFAULT_TRACE_PIECE_SIZE_LIMIT;
    globals.maxTracePieceBytes *= 1048576;     // Convert MB to bytes
  }
  else
    globals.maxTracePieceBytes = DEFAULT_TRACE_PIECE_SIZE_LIMIT * 1048576;

  ZwClose(keyHandle);
  MyFreeUnicodeString(&string);
  return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////
//                          DISPATCH ROUTINES
////////////////////////////////////////////////////////////////////////////

// VTrcLogDispatchControl is the dispatch routine called to handle any
// device control IRP sent to our VTrcLog device.  The main thing to
// do here is check what kind of device control request it is and handle
// it accordingly.

NTSTATUS VTrcLogDispatchControl (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  PIO_STACK_LOCATION  irpStack = IoGetCurrentIrpStackLocation(irp);
  PCHAR               inBuf, outBuf, eventBuffer;
  ULONG               inBufLength, outBufLength, ioControlCode, reasons,
                      time1, time2;
  KIRQL               oldirql;
  NTSTATUS            status;

  // Set the request up as successful, by default.

  status = STATUS_SUCCESS;
  irp->IoStatus.Information = 0;

  // Get the control code (what kind of request this is)

  ioControlCode = irpStack->Parameters.DeviceIoControl.IoControlCode;

  // Get the pointer to the input/output buffer and its length.
  // These buffers are in different places depending on the method
  // used (the low two bits of ioControlCode).

  if ((ioControlCode & 0x3) == METHOD_NEITHER) {
    inBuf = irpStack->Parameters.DeviceIoControl.Type3InputBuffer;
    outBuf = irp->UserBuffer;
  }
  else {
    inBuf = irp->AssociatedIrp.SystemBuffer;
    outBuf = irp->AssociatedIrp.SystemBuffer;
  }
  inBufLength = irpStack->Parameters.DeviceIoControl.InputBufferLength;
  outBufLength = irpStack->Parameters.DeviceIoControl.OutputBufferLength;

  // Do the appropriate thing, depending on the control code

  switch (ioControlCode) {
  case VTRACE_CTRL_REMOVE_REASONS_TO_STOP :
    // A request to remove reasons to stop logging.  The input buffer should
    // contain four bytes indicating this reason.
    DebugPrint(("VTrcLog: Received request to remove reasons for "
                "stopping.\n"));
    if (inBufLength >= 4)
      RemoveReasonsForNotLogging(* (ULONG *) inBuf);
    else
      status = STATUS_INVALID_PARAMETER;
    break;

  case VTRACE_CTRL_ADD_REASONS_TO_STOP :
    // A request to add reasons to stop logging.  The first 4 bytes contain
    // this reason.  If this reason includes "because user-level access to the
    // trace files is required", and that access has already been granted
    // to another user-level process, return failure

    DebugPrint(("VTrcLog: Received a request to add reasons to stop.\n"));
    if (inBufLength < 4)
      status = STATUS_INVALID_PARAMETER;
    else {
      reasons = * (ULONG *) inBuf;
      KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
      if ((reasons & BECAUSE_ACCESS_REQUIRED) &&
          (globals.whyLoggingDisabled & BECAUSE_ACCESS_REQUIRED))
        status = STATUS_UNSUCCESSFUL;
      else
        AddReasonsForNotLogging(reasons);
      KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    }
    break;

  case VTRACE_CTRL_LOG_EVENT :
    // To log an event, copy it to the event buffer and call LogEvent.

    if (inBufLength > MAX_EVENT_LENGTH) {
      DebugPrint(("VTrcLog: Received too-long event from user level\n"));
      status = STATUS_INVALID_PARAMETER;
    }
    else {
      KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
      eventBuffer = LogEvent(inBuf[0], inBufLength);
      if (eventBuffer)
        RtlCopyMemory(&eventBuffer[1], &inBuf[1], inBufLength-1);
      KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    }
    break;

  case VTRACE_CTRL_FLUSH :
    // A request to flush in-memory log data to a trace file

    FlushDataNow((PVOID) 0 /* not called from queue */);
    break;

  case VTRACE_CTRL_SET_USER_DEFINED_MASK :
    // A request to set the user-defined event mask to something new.
    // The event mask must be 32 bytes (at least) to be valid.

    if (inBufLength >= 32)
      SetUserDefinedEventMask(inBuf);
    else
      status = STATUS_INVALID_PARAMETER;
    break;

  case VTRACE_CTRL_GET_MASK :
    // A request to get the current event mask. The buffer size needed
    // for this mask is at least 32 bytes.

    if (outBufLength >= 32) {
      GetEventMask(outBuf);
      irp->IoStatus.Information = 32;
    }
    else
      status = STATUS_INVALID_PARAMETER;
    break;

  case VTRACE_CTRL_INDICATE_USER_ACTIVITY :
    // An indication that user activity occurred
    IndicateUserActivity();
    break;

  case VTRACE_CTRL_GET_IDLE_TIME :
    // A request for the length of time the user has been idle.  We return
    // this time as an 8-byte 100-ns count.

    if (outBufLength < 8)
      status = STATUS_INVALID_PARAMETER;
    else {
      LARGE_INTEGER currentTime;
      KeQuerySystemTime(&currentTime);
      currentTime.QuadPart -= globals.timeOfLastUserActivity.QuadPart;
      RtlCopyMemory(outBuf, &currentTime, 8);
      irp->IoStatus.Information = 8;
    }
    break;

  case VTRACE_CTRL_TIME_LOG_EVENT :
    // To log an event, copy it to the event buffer and call LogEvent.

    if (inBufLength > MAX_EVENT_LENGTH || outBufLength < 4)
      status = STATUS_INVALID_PARAMETER;
    else {
      __asm {
        _emit 0x0F       ;RDTSC - get beginning timestamp to edx:eax
        _emit 0x31
        mov time1, eax   ;save low byte of timestamp
      }
      KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
      eventBuffer = LogEvent(inBuf[0], inBufLength);
      if (eventBuffer)
        RtlCopyMemory(&eventBuffer[1], &inBuf[1], inBufLength-1);
      KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
      __asm {
        _emit 0x0F       ;RDTSC - get beginning timestamp to edx:eax
        _emit 0x31
        mov time2, eax   ;save low byte of timestamp
      }
      if (eventBuffer) {
        * (ULONG *) outBuf = time2 - time1;
        irp->IoStatus.Information = 4;
      }
      else
        status = STATUS_UNSUCCESSFUL;
    }
    break;

  case VTRACE_CTRL_RELOAD_REGISTRY :
    // An indication that the registry has changed and so the parameters
    // from there should be reloaded.

    status = ReloadRegistry();
    break;

  case VTRACE_CTRL_CHANGE_TO_EVENT_MASK_NUMBER :
    if (inBufLength < 4)
      status = STATUS_INVALID_PARAMETER;
    else
      PeriodicEventMaskChangeRoutine(NULL, * (PVOID *) inBuf, NULL, NULL);
    break;
  }

  // Complete the IRP

  irp->IoStatus.Status = status;
  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return status;
}

// VTrcLogDispatchShutdown is the dispatch routine called to handle
// shutdown IRP's sent to our VTrcLog device.  The main thing to do
// here is to stop logging.  (We handle shutdown IRP's so that we
// get notified of shutdowns so that we can stop logging then.)

NTSTATUS VTrcLogDispatchShutdown (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  KIRQL oldirql;

  // Stop logging
    
  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  AddReasonsForNotLogging(BECAUSE_SYSTEM_HALTING);
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Commence flushing of the end of the trace file immediately

  FlushDataNow(NULL);

  // Once the data is safely flushed, complete the IRP

  irp->IoStatus.Status = STATUS_SUCCESS;
  irp->IoStatus.Information = 0;
  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return STATUS_SUCCESS;
}

// VTrcLogDispatchGeneral is the dispatch routine called to handle most
// IRP types sent to our VTrcLog device.  The main thing to do here is
// to simply complete the IRP.

NTSTATUS VTrcLogDispatchGeneral (IN PDEVICE_OBJECT devObj, IN PIRP irp)
{
  // Set the request up as successful

  irp->IoStatus.Status = STATUS_SUCCESS;
  irp->IoStatus.Information = 0;
    
  // Complete the IRP

  IoCompleteRequest(irp, IO_NO_INCREMENT);
  return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////
//                        CHECK FOR LOAD OVERRIDE
////////////////////////////////////////////////////////////////////////

// GetKeyboardState gets the current state of the keyboard indicator lights,
// putting that at the location pointed to by keyboardStatePtr.  This function
// works by building a device control IRP and passing it to the keyboard
// device.

NTSTATUS GetKeyboardState (KEYBOARD_INDICATOR_PARAMETERS *keyboardStatePtr)
{
  PIRP                    irp;
  IO_STATUS_BLOCK         ioStatusBlock;
  KEVENT                  event;
  UNICODE_STRING          kbdDeviceNameUnicodeString;
  PFILE_OBJECT            kbdFileObject;
  PDEVICE_OBJECT          kbdDeviceObject;
  NTSTATUS                status;

  // Get a device object pointer to the keyboard device

  RtlInitUnicodeString(&kbdDeviceNameUnicodeString,
                       L"\\Device\\KeyboardClass0");
  status = IoGetDeviceObjectPointer(&kbdDeviceNameUnicodeString,
                                    FILE_ALL_ACCESS, &kbdFileObject,
                                    &kbdDeviceObject);
  if (!NT_SUCCESS(status))
    return status;
  ObDereferenceObject(kbdFileObject);  // We don't need the file object pointer

  // Initialize the synchronization event to be used by the IRP

  KeInitializeEvent(&event, SynchronizationEvent, FALSE);

  // Create a device-control IRP to get the indicator information

  irp = IoBuildDeviceIoControlRequest(IOCTL_KEYBOARD_QUERY_INDICATORS,
                                      kbdDeviceObject, NULL, 0,
                                      keyboardStatePtr,
                                      sizeof(KEYBOARD_INDICATOR_PARAMETERS),
                                      FALSE, &event, &ioStatusBlock);
  if (irp == NULL)
    return STATUS_INSUFFICIENT_RESOURCES;

  // Call the driver

  status = IoCallDriver(kbdDeviceObject, irp);

  // If the call is pending, just wait for the event to be set

  if (status == STATUS_PENDING) {
    while (KeWaitForSingleObject(&event, Executive, KernelMode, TRUE, 0) !=
           STATUS_SUCCESS)
      ;
    status = ioStatusBlock.Status;
  }

  return status;
}

// IsUserOverridingLoad returns a boolean indicating whether the user is
// indicating that VTrcLog should not be loaded, i.e. by having the caps-lock
// or scroll-lock light on during system initialization.

BOOLEAN IsUserOverridingLoad (void)
{
  KEYBOARD_INDICATOR_PARAMETERS keyboardState;
  NTSTATUS status;

  status = GetKeyboardState(&keyboardState);
  if (status == STATUS_SUCCESS &&
      (keyboardState.LedFlags &
       (KEYBOARD_CAPS_LOCK_ON | KEYBOARD_SCROLL_LOCK_ON)))
    return TRUE;
  else
    return FALSE;
}
