//
// ZwHooks.c
//
// Code for Zw routine hooks in the VTrcLog driver.
//

#include "NTDDK.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "Globals.h"
#include "LogEvent.h"
#include "GetState.h"
#include "SysCall.h"
#include "StartStop.h"
#include "ZwHooks.h"

NTSYSAPI
NTSTATUS
NTAPI
ZwQueryInformationProcess(
    IN HANDLE ProcessHandle,
    IN PROCESSINFOCLASS ProcessInformationClass,
    OUT PVOID ProcessInformation,
    IN ULONG ProcessInformationLength,
    OUT PULONG ReturnLength OPTIONAL
    );

////////////////////////////////////////////////////////////////////////////
//                                 MACROS
////////////////////////////////////////////////////////////////////////////

#define MIN(a,b)    (((a)>(b))?(b):(a))

// Device flags propagated from file system device to filtering device

#define DEVICE_FLAGS_PROPAGATED            0

////////////////////////////////////////////////////////////////////////////
//                       EXTERNALLY DEFINED OBJECTS
////////////////////////////////////////////////////////////////////////////

// KeServiceDescriptorTable is the not-so-well-documented service table
// structure

typedef struct _SRVTABLE {
  PVOID *ServiceTable;
  ULONG LowCall;
  ULONG HiCall;
  PVOID *ArgTable;
} SRVTABLE, *PSRVTABLE;
extern PSRVTABLE KeServiceDescriptorTable;

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, InstallZwHooks)
#pragma alloc_text (INIT, UninstallZwHooks)
#endif

////////////////////////////////////////////////////////////////////////////
//                            DEVICE ATTACHING
////////////////////////////////////////////////////////////////////////////

// AttachToFSDevice creates a new VTrcFS device and attaches it to the
// given device object.

void AttachToFSDevice (IN PDEVICE_OBJECT deviceObject)
{
  PDEVICE_OBJECT       hookDevice = NULL, nextDeviceObject = NULL;
  NTSTATUS             status;
  PFS_HOOK_EXTENSION   hookExtension;
  PUNICODE_STRING      driverString;
  KIRQL                oldirql;
  DEVICE_TYPE          type;
  USHORT               nameLength;
  int                  i;
  ULONG                propagateFlags;
  PCHAR                eventBuffer;

  // Make sure there's a VTrcFS driver and that we've been told where it is

  if (sharedState->fsDriverObject == NULL) return;

  // Create a new device of the same type as the given device, but
  // associated with the VTrcFS driver

  type = deviceObject->DeviceType;
  status = IoCreateDevice(sharedState->fsDriverObject,
                          sizeof(FS_HOOK_EXTENSION), NULL, type, 0, FALSE,
                          &hookDevice);
  if (!NT_SUCCESS(status) || hookDevice == NULL) {
    DebugPrint(("VTrcLog: Failed to create associated FS device.\n"));
    return;
  }

  // Clear the device's init flag as per the NT DDK KB article on creating
  // device objects from a dispatch routine

  hookDevice->Flags &= ~DO_DEVICE_INITIALIZING;

  // Tentatively set the attachedDevice field in the hook extension to
  // be the hooked device, so that in the unlikely event that the hook
  // device gets sent an IRP in the few instructions between the
  // attach function and the real setting of the attachedDevice field,
  // we don't pass the IRP to a random device pointer. While we're at
  // it, set the other fields in the hook extension.

  hookExtension = hookDevice->DeviceExtension;
  RtlZeroMemory(hookExtension, sizeof(FS_HOOK_EXTENSION));
  hookExtension->attachedDevice = deviceObject;
  hookExtension->diskNumber = (char) UNKNOWN_DEVICE_NUMBER;
  hookExtension->partitionNumber = (char) UNKNOWN_DEVICE_NUMBER;
  KeInitializeEvent(&hookExtension->PagingPathCountEvent, NotificationEvent,
                    TRUE);

  // Attach to the device, getting the highest-level device already attached
  // to it in nextDeviceObject.  We need this in order to know what device to
  // call down to when we get a request; so, we put it in the attachedDevice
  // field of the hook extension.  Make one last check to make sure the device
  // hasn't been attached to in the time since we first decided to attach
  // to it, so that VTrace doesn't attach to it twice.  (Technically, we
  // should put this in a spinlock-protected section, but since it isn't
  // the end of the world if we attach twice, and since we must run
  // IoAttachDeviceToDeviceStack at passive level, we don't.)

  if (deviceObject->AttachedDevice == NULL)
    nextDeviceObject = IoAttachDeviceToDeviceStack(hookDevice, deviceObject);
  else
    nextDeviceObject = NULL;

  // If the attach failed, delete the device we just created and return

  if (nextDeviceObject == NULL) {
    DebugPrint(("VTrcLog: Connect with file system device %x of type %d "
                "failed\n", deviceObject, type));
    IoDeleteDevice(hookDevice);
    return;
  }
  hookExtension->attachedDevice = nextDeviceObject;

  // Synchronize various fields in the filter device with that of the
  // device being filtered, then reset the DO_DEVICE_INITIALIZING flag in
  // the filter device.

  propagateFlags = hookExtension->attachedDevice->Flags &
                   DEVICE_FLAGS_PROPAGATED;
  hookDevice->Flags |= propagateFlags;

  propagateFlags = hookExtension->attachedDevice->Characteristics &
                   FILE_CHARACTERISTICS_PROPAGATED;
  hookDevice->Characteristics |= propagateFlags;

  hookDevice->Flags |= DO_POWER_PAGABLE;
  hookDevice->Flags &= ~DO_DEVICE_INITIALIZING;

  // Log the fact that we're starting to hook this device

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  RecordCurrentDeviceState(deviceObject, (char) UNKNOWN_DEVICE_NUMBER,
                           (char) UNKNOWN_DEVICE_NUMBER);
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  DebugPrint(("VTrcLog: Successfully connected to file system device %x\n",
              deviceObject));
}

// AttachToNetDevice creates a new VTrcNet device and attaches it to the
// given device object.

void AttachToNetDevice (IN PDEVICE_OBJECT deviceObject)
{
  PDEVICE_OBJECT      hookDevice = NULL, nextDeviceObject = NULL;
  NTSTATUS            status;
  PNET_HOOK_EXTENSION hookExtension;
  PUNICODE_STRING     driverString;
  KIRQL               oldirql;
  DEVICE_TYPE         type;
  USHORT              nameLength;
  int                 i;
  ULONG               propagateFlags;
  PCHAR               eventBuffer;

  // Make sure the VTrcNet driver exists and we've been told where it is

  if (sharedState->netDriverObject == NULL) return;

  // Create a new device of the same type as the given device, but
  // associated with the VTrcNet driver

  type = deviceObject->DeviceType;
  status = IoCreateDevice(sharedState->netDriverObject,
                          sizeof(NET_HOOK_EXTENSION), NULL, type, 0, FALSE,
                          &hookDevice);
  if (!NT_SUCCESS(status) || hookDevice == NULL) {
    DebugPrint(("VTrcLog: Failed to create associated net device.\n"));
    return;
  }

  // Clear the device's init flag as per NT DDK KB article on creating device
  // objects from a dispatch routine

  hookDevice->Flags &= ~DO_DEVICE_INITIALIZING;

  // Set up the device extension by filling in the information needed about
  // the device we have to send requests down to and initializing the file
  // object info list to empty.  We tentatively assign to the attachedDevice
  // field the value of the device object being hooked to, just in case
  // this new hook device gets sent an IRP in the few instructions between
  // the attachment and the setting of this field to its final correct
  // value.  While we're at it, we fill in other fields of the hook extension.

  hookExtension = hookDevice->DeviceExtension;
  hookExtension->attachedDevice = deviceObject;
  hookExtension->fileObjectInfoList = NULL;

  // Attach to the device, getting the highest-level device already attached
  // to it in nextDeviceObject.  We need this in order to know what device to
  // call down to when we get a request; so, we put it in the attachedDevice
  // field of the hook extension.  Make one last check to make sure the device
  // hasn't been attached to in the time since we first decided to attach
  // to it, so that VTrace doesn't attach to it twice.  (Technically, we
  // should put this in a spinlock-protected section, but since it isn't
  // the end of the world if we attach twice, and since we must run
  // IoAttachDeviceToDeviceStack at passive level, we don't.)

  if (deviceObject->AttachedDevice == NULL)
    nextDeviceObject = IoAttachDeviceToDeviceStack(hookDevice, deviceObject);
  else
    nextDeviceObject = NULL;

  // If the attach failed, delete the device we just created and return

  if (nextDeviceObject == NULL) {
    DebugPrint(("VTrcLog: Connect with net device %x failed (type %d)\n",
                deviceObject, type));
    IoDeleteDevice(hookDevice);
    return;
  }
  hookExtension->attachedDevice = nextDeviceObject;

  // Synchronize various fields in the filter device with that of the
  // device being filtered, then reset the DO_DEVICE_INITIALIZING flag in
  // the filter device.

  propagateFlags = hookExtension->attachedDevice->Flags &
                   DEVICE_FLAGS_PROPAGATED;
  hookDevice->Flags |= propagateFlags;

  propagateFlags = hookExtension->attachedDevice->Characteristics &
                   FILE_CHARACTERISTICS_PROPAGATED;
  hookDevice->Characteristics |= propagateFlags;

  hookDevice->Flags &= ~DO_DEVICE_INITIALIZING;

  // Log the fact that we're starting to hook this device

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  RecordCurrentDeviceState(deviceObject, (char) UNKNOWN_DEVICE_NUMBER,
                           (char) UNKNOWN_DEVICE_NUMBER);
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  DebugPrint(("VTrcLog: Successfully connected to net device %x\n",
              deviceObject));
}

////////////////////////////////////////////////////////////////////////////
//                           ZW ROUTINE HOOKS
////////////////////////////////////////////////////////////////////////////

// HookOpenFile is the ZwOpenFile hook routine.  It calls the original
// ZwOpenFile to open a file, then checks whether the device the file
// belongs to is the type of device we would like to hook.  If so, it
// hooks the device.

NTSTATUS HookOpenFile (OUT PHANDLE FileHandle,
                       IN ACCESS_MASK DesiredAccess,
                       IN POBJECT_ATTRIBUTES ObjectAttributes,
                       OUT PIO_STATUS_BLOCK IoStatusBlock,
                       IN ULONG ShareAccess,
                       IN ULONG OpenOptions)
{
  NTSTATUS retVal;
  PFILE_OBJECT fileObject;
  PDEVICE_OBJECT deviceObject;
  DEVICE_TYPE type;
  ULONG previousFilteredCreates;

  // Record the number of creates VTrace has filtered before we call
  // the open-file routine

  previousFilteredCreates = sharedState->filteredCreates;

  // Call the real open-file routine, recording the return value

  retVal = (*globals.origOpenFile)(FileHandle, DesiredAccess, ObjectAttributes,
                                   IoStatusBlock, ShareAccess, OpenOptions);

  // If VTrace has filtered a create IRP in the meantime, we have probably
  // already hooked this device, so just return.  Also, if the open wasn't
  // successful, return the error code.

  if (previousFilteredCreates != sharedState->filteredCreates ||
      retVal != STATUS_SUCCESS)
    return retVal;

  // Reference the object to get the corresponding file object pointer

  if (ObReferenceObjectByHandle(*FileHandle, FILE_READ_DATA, NULL,
                                KernelMode, &fileObject, NULL) !=
      STATUS_SUCCESS)
    return retVal;

  // If we successfully referenced it, all we need out of it is the related
  // device object.  We must eventually dereference it after we're done
  // using it.

  deviceObject = IoGetRelatedDeviceObject(fileObject);

  // Check if this is a file system type we're interested in hooking.
  // If it is, then close the file handle we got (since it won't be
  // pointing to the VTrcFS hook device and its accesses won't get
  // logged), hook the device, then reopen the file.  If it isn't,
  // check if it's a network type we're interested in hooking and
  // if so treat it similary.  In any event, remember to dereference
  // the file object.

  // In either case, we also have to check that the fast I/O dispatch
  // table of the hooked device exists or does not exist in the same
  // way as the device we are going to attach to it.  In other words,
  // we cannot hook the VTrcFS device to a device that does not have
  // a fast I/O dispatch table, and we cannot hook the VTrcNet device
  // to a device that has one.  Fortunately, this check is likely
  // just paranoia, since fast I/O dispatch is supposed to be
  // present for and only for file system devices.

  // It is also important to check, for the case of file systems,
  // whether the file object has a name.  This is to prevent hooking
  // the same file system twice, for the following reason.  Unnamed
  // file objects indicate low-level objects such as \Device\Floppy
  // or \??\C:.  These objects might get opened with a create IRP
  // that VTrace never sees, because they go directly to the lower-
  // level device.  So, even though VTrace has already hooked this
  // device, globals.filteredCreates might not have gotten
  // incremented, and we would reach this point.  Besides, we don't
  // trace no-name file objects anyway, so there's no loss if we
  // don't hook this device right now even if it hasn't yet been
  // hooked.

  // As a paranoid check, after other checks, we should make sure
  // we aren't directly filtering a VTrcFS or VTrcNet device.

  type = deviceObject->DeviceType;
  if (deviceObject->DriverObject->FastIoDispatch != NULL) {
    if (fileObject->FileName.Buffer != NULL &&
        type == FILE_DEVICE_NETWORK_FILE_SYSTEM &&
        deviceObject->DriverObject != sharedState->fsDriverObject) {
      ObDereferenceObject(fileObject);
      ZwClose(*FileHandle);
      AttachToFSDevice(deviceObject);
      return (*globals.origOpenFile)(FileHandle, DesiredAccess,
                                     ObjectAttributes, IoStatusBlock,
                                     ShareAccess, OpenOptions);
    }
  }
  else {
    if ((type == FILE_DEVICE_NETWORK || type == FILE_DEVICE_TRANSPORT) &&
        deviceObject->DriverObject != sharedState->netDriverObject) {
      ObDereferenceObject(fileObject);
      ZwClose(*FileHandle);
      AttachToNetDevice(deviceObject);
      return (*globals.origOpenFile)(FileHandle, DesiredAccess,
                                     ObjectAttributes, IoStatusBlock,
                                     ShareAccess, OpenOptions);
    }
  }

  ObDereferenceObject(fileObject);
  return retVal;
}

// HookCreateFile is the ZwCreateFile hook routine.  It calls the original
// ZwCreateFile to create a file, then checks whether the device the file
// belongs to is the type of device we would like to hook.  If so, it hooks
// the device.

NTSTATUS HookCreateFile (OUT PHANDLE FileHandle,
                         IN ACCESS_MASK DesiredAccess,
                         IN POBJECT_ATTRIBUTES ObjectAttributes,
                         OUT PIO_STATUS_BLOCK IoStatusBlock,
                         IN PLARGE_INTEGER AllocationSize OPTIONAL,
                         IN ULONG FileAttributes,
                         IN ULONG ShareAccess,
                         IN ULONG CreateDisposition,
                         IN ULONG CreateOptions,
                         IN PVOID EaBuffer OPTIONAL,
                         IN ULONG EaLength)
{
  NTSTATUS retVal;
  PFILE_OBJECT fileObject;
  PDEVICE_OBJECT deviceObject;
  DEVICE_TYPE type;
  ULONG previousFilteredCreates;

  // Record the number of creates VTrace has filtered before we call
  // the create-file routine

  previousFilteredCreates = sharedState->filteredCreates;

  // Call the real create-file routine, recording the return value

  retVal = (*globals.origCreateFile)(FileHandle, DesiredAccess,
                                     ObjectAttributes, IoStatusBlock,
                                     AllocationSize, FileAttributes,
                                     ShareAccess, CreateDisposition,
                                     CreateOptions, EaBuffer, EaLength);

  // If VTrace has filtered a create IRP in the meantime, we have probably
  // already hooked this device, so just return.  Also, if the open wasn't
  // successful, return the error code.

  if (previousFilteredCreates != sharedState->filteredCreates ||
      retVal != STATUS_SUCCESS)
    return retVal;

  // Reference the object to get the corresponding file object pointer

  if (ObReferenceObjectByHandle(*FileHandle, FILE_READ_DATA, NULL,
                                KernelMode, &fileObject, NULL) !=
      STATUS_SUCCESS)
    return retVal;

  // If we successfully referenced it, all we need out of it is the related
  // device object.  We must eventually dereference it when we don't need it
  // any more.

  deviceObject = IoGetRelatedDeviceObject(fileObject);

  // Check if this is a file system type we're interested in hooking.
  // If it is, then close the file handle we got (since it won't be
  // pointing to the VTrcFS hook device and its accesses won't get
  // logged), hook the device, then recreate the file.  If it isn't,
  // check if it's a network type we're interested in hooking and
  // if so treat it similary.  In any case, remember to dereference
  // the file object.

  // In either case, we also have to check that the fast I/O dispatch
  // table of the hooked device exists or does not exist in the same
  // way as the device we are going to attach to it.  In other words,
  // we cannot hook the VTrcFS device to a device that does not have
  // a fast I/O dispatch table, and we cannot hook the VTrcNet device
  // to a device that has one.  Fortunately, this check is likely
  // just paranoia, since fast I/O dispatch is supposed to be
  // present for and only for file system devices.

  // It is also important to check, for the case of file systems,
  // whether the file object has a name.  This is to prevent hooking
  // the same file system twice, for the following reason.  Unnamed
  // file objects indicate low-level objects such as \Device\Floppy
  // or \??\C:.  These objects might get opened with a create IRP
  // that VTrace never sees, because they go directly to the lower-
  // level device.  So, even though VTrace has already hooked this
  // device, globals.filteredCreates might not have gotten
  // incremented, and we would reach this point.  Besides, we don't
  // trace no-name file objects anyway, so there's no loss if we
  // don't hook this device right now even if it hasn't yet been
  // hooked.

  // As a paranoid check, after other checks, we should make sure
  // we aren't directly filtering a VTrcFS or VTrcNet device.

  type = deviceObject->DeviceType;
  if (deviceObject->DriverObject->FastIoDispatch != NULL) {
    if (fileObject->FileName.Buffer != NULL &&
        type == FILE_DEVICE_NETWORK_FILE_SYSTEM &&
        deviceObject->DriverObject != sharedState->fsDriverObject) {
      ObDereferenceObject(fileObject);
      ZwClose(*FileHandle);
      AttachToFSDevice(deviceObject);
      return (*globals.origCreateFile)(FileHandle, DesiredAccess,
                                       ObjectAttributes, IoStatusBlock,
                                       AllocationSize, FileAttributes,
                                       ShareAccess, CreateDisposition,
                                       CreateOptions, EaBuffer, EaLength);
    }
  }
  else {
    if ((type == FILE_DEVICE_NETWORK || type == FILE_DEVICE_TRANSPORT) &&
        deviceObject->DriverObject != sharedState->netDriverObject) {
      ObDereferenceObject(fileObject);
      ZwClose(*FileHandle);
      AttachToNetDevice(deviceObject);
      return (*globals.origCreateFile)(FileHandle, DesiredAccess,
                                       ObjectAttributes, IoStatusBlock,
                                       AllocationSize, FileAttributes,
                                       ShareAccess, CreateDisposition,
                                       CreateOptions, EaBuffer, EaLength);
    }
  }

  ObDereferenceObject(fileObject);
  return retVal;
}

// HookFlushInstructionCache is the ZwFlushInstructionCache hook routine.
// It logs the call, then calls the original ZwFlushInstructionCache.

NTSTATUS HookFlushInstructionCache (IN PVOID arg1, IN PVOID arg2,
                                    IN PVOID arg3)
{
  KIRQL oldirql;
  char *eventBuffer;

  // Log the call

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LogEvent((char) ENTRY_TYPE_SYSTEM_CALL, 2);
  if (eventBuffer)
    eventBuffer[1] = (char) FLUSH_ICACHE_SYSTEM_CALL_CODE;
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Call the original Zw routine

  return (*globals.origFlushInstructionCache)(arg1, arg2, arg3);
}

// HookFlushWriteBuffer is the ZwFlushWriteBuffer hook routine.  It logs
// the call, then calls the original ZwFlushWriteBuffer.

NTSTATUS HookFlushWriteBuffer (void)
{
  KIRQL oldirql;
  char *eventBuffer;

  // Log the call

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LogEvent((char) ENTRY_TYPE_SYSTEM_CALL, 2);
  if (eventBuffer)
    eventBuffer[1] = (char) FLUSH_WRITE_BUFFER_SYSTEM_CALL_CODE;
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Call the original Zw routine

  return (*globals.origFlushWriteBuffer)();
}

// HookTerminateProcess is the ZwTerminateProcess hook routine. It logs
// the call, then calls the original ZwTerminateProcess.

NTSTATUS HookTerminateProcess (IN PVOID arg1, IN PVOID arg2)
{
  KIRQL oldirql;
  char *eventBuffer;

  // Log the call

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LogEvent((char) ENTRY_TYPE_SYSTEM_CALL, 2);
  if (eventBuffer)
    eventBuffer[1] = (char) TERMINATE_PROCESS_SYSTEM_CALL_CODE;
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Call the original Zw routine

  return (*globals.origTerminateProcess)(arg1, arg2);
}

// HookTerminateThread is the ZwTerminateThread hook routine. It logs the
// call, then calls the original ZwTerminateThread.

NTSTATUS HookTerminateThread (IN PVOID arg1, IN PVOID arg2)
{
  KIRQL oldirql;
  char *eventBuffer;

  // Log the call

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LogEvent((char) ENTRY_TYPE_SYSTEM_CALL, 2);
  if (eventBuffer)
    eventBuffer[1] = (char) TERMINATE_THREAD_SYSTEM_CALL_CODE;
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Call the original Zw routine

  return (*globals.origTerminateThread)(arg1, arg2);
}

// HookWriteRequestData is the ZwWriteRequestData hook routine. It logs
// the call, then calls the original ZwWriteRequestData.

NTSTATUS HookWriteRequestData (IN PVOID arg1, IN PVOID arg2, IN PVOID arg3,
                               IN PVOID arg4, IN PVOID arg5, IN PVOID arg6)
{
  KIRQL oldirql;
  char *eventBuffer;

  // Log the call

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LogEvent((char) ENTRY_TYPE_SYSTEM_CALL, 2);
  if (eventBuffer)
    eventBuffer[1] = (char) WRITE_REQUEST_DATA_SYSTEM_CALL_CODE;
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Call the original Zw routine

  return (*globals.origWriteRequestData)(arg1, arg2, arg3, arg4, arg5, arg6);
}

// HookWriteVirtualMemory is the ZwWriteVirtualMemory hook routine.  It
// logs the call, then calls the original ZwWriteVirtualMemory.

NTSTATUS HookWriteVirtualMemory (IN PVOID arg1, IN PVOID arg2, IN PVOID arg3,
                                 IN PVOID arg4, IN PVOID arg5)
{
  KIRQL oldirql;
  char *eventBuffer;

  // Log the call

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LogEvent((char) ENTRY_TYPE_SYSTEM_CALL, 2);
  if (eventBuffer)
    eventBuffer[1] = (char) WRITE_VM_SYSTEM_CALL_CODE;
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Call the original Zw routine

  return (*globals.origWriteVirtualMemory)(arg1, arg2, arg3, arg4, arg5);
}

// HookCreateSection is the ZwCreateSection hook routine. It calls the original
// ZwCreateSection, then logs the return values.

NTSTATUS HookCreateSection (OUT PHANDLE SectionHandle,
                            IN ACCESS_MASK DesiredAccess,
                            IN POBJECT_ATTRIBUTES ObjectAttributes OPTIONAL,
                            IN PLARGE_INTEGER MaximumSize OPTIONAL,
                            IN ULONG SectionPageProtection,
                            IN ULONG AllocationAttributes,
                            IN HANDLE FileHandle OPTIONAL)
{
  KIRQL oldirql;
  NTSTATUS retVal;
  PVOID sectionObject, fileObject;
  char *eventBuffer;

  // Call the original routine, saving the return value

  retVal = (*globals.origCreateSection)(SectionHandle, DesiredAccess,
                                        ObjectAttributes, MaximumSize,
                                        SectionPageProtection,
                                        AllocationAttributes, FileHandle);

  // If it was successful, and we can reference the section handle as a
  // pointer, log the call.  Don't forget to dereference any objects we
  // successfully reference!

  if (retVal == STATUS_SUCCESS &&
      ObReferenceObjectByHandle(*SectionHandle, STANDARD_RIGHTS_REQUIRED, NULL,
                                KernelMode, &sectionObject, NULL)
        == STATUS_SUCCESS) {
    if (FileHandle == NULL ||
        ObReferenceObjectByHandle(FileHandle, STANDARD_RIGHTS_REQUIRED, NULL,
                                  KernelMode, &fileObject, NULL)
        != STATUS_SUCCESS)
      fileObject = NULL;
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LogEvent((char) ENTRY_TYPE_SECTION_CREATE, 24);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &sectionObject, 4);
      RtlCopyMemory(&eventBuffer[5], &fileObject, 4);
      eventBuffer[9] = (char) DesiredAccess;
      if (MmIsAddressValid(MaximumSize))
        RtlCopyMemory(&eventBuffer[10], MaximumSize, 8);
      else
        RtlZeroMemory(&eventBuffer[10], 8);
      RtlCopyMemory(&eventBuffer[18], &SectionPageProtection, 2);
      RtlCopyMemory(&eventBuffer[20], &AllocationAttributes, 4);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    if (fileObject != NULL)
      ObDereferenceObject(fileObject);
    ObDereferenceObject(sectionObject);
  }

  // Return the saved return value
  
  return retVal;
}

// HookOpenSection is the ZwOpenSection hook routine. It calls the original
// ZwOpenSection, then logs the return values.

NTSTATUS HookOpenSection (OUT PHANDLE SectionHandle,
                          IN ACCESS_MASK DesiredAccess,
                          IN POBJECT_ATTRIBUTES ObjectAttributes)
{
  NTSTATUS retVal;
  KIRQL oldirql;
  PVOID sectionObject;
  char *eventBuffer;

  // Call the original routine, saving the return value

  retVal = (*globals.origOpenSection)(SectionHandle, DesiredAccess,
                                      ObjectAttributes);

  // If it was successful, and we can reference the section handle as a
  // pointer, log the call.  Don't forget to dereference the section object
  // if we successfully reference it!

  if (retVal == STATUS_SUCCESS &&
      ObReferenceObjectByHandle(*SectionHandle, STANDARD_RIGHTS_REQUIRED, NULL,
                                KernelMode, &sectionObject, NULL)
        == STATUS_SUCCESS) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LogEvent((char) ENTRY_TYPE_SECTION_OPEN, 6);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &sectionObject, 4);
      eventBuffer[5] = (char) DesiredAccess;
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    ObDereferenceObject(sectionObject);
  }

  // Return the saved return value

  return retVal;
}

// HookMapViewOfSection is the ZwMapViewOfSection hook routine. It calls the
// original ZwMapViewOfSection, then logs the return values.

NTSTATUS HookMapViewOfSection (IN HANDLE SectionHandle,
                               IN HANDLE ProcessHandle,
                               IN OUT PVOID *BaseAddress,
                               IN ULONG ZeroBits,
                               IN ULONG CommitSize,
                               IN OUT PLARGE_INTEGER SectionOffset OPTIONAL,
                               IN OUT PULONG ViewSize,
                               IN SECTION_INHERIT InheritDisposition,
                               IN ULONG AllocationType,
                               IN ULONG Protect)
{
  NTSTATUS retVal;
  KIRQL oldirql;
  PVOID sectionObject;
  PROCESS_BASIC_INFORMATION processInfo;
  NTSTATUS status;
  ULONG returnLength;
  char *eventBuffer;

  // Call the original routine, saving the return value

  retVal = (*globals.origMapViewOfSection)(SectionHandle, ProcessHandle,
                                           BaseAddress, ZeroBits, CommitSize,
                                           SectionOffset, ViewSize,
                                           InheritDisposition, AllocationType,
                                           Protect);

  // If it was successful, and we can reference the section handle as a
  // pointer, log the call.  Don't forget to dereference the section object
  // if we successfully reference it!

  if (retVal == STATUS_SUCCESS &&
      ObReferenceObjectByHandle(SectionHandle, STANDARD_RIGHTS_REQUIRED, NULL,
                                KernelMode, &sectionObject, NULL)
      == STATUS_SUCCESS) {
    status = ZwQueryInformationProcess(ProcessHandle, ProcessBasicInformation,
                                       &processInfo, sizeof(processInfo),
                                       &returnLength);
    if (status != STATUS_SUCCESS)
      processInfo.UniqueProcessId = 0xFFFFFFFF;
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LogEvent((char) ENTRY_TYPE_SECTION_MAP_VIEW, 32);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &sectionObject, 4);
      RtlCopyMemory(&eventBuffer[5], &processInfo.UniqueProcessId, 4);
      if (MmIsAddressValid(BaseAddress))
        RtlCopyMemory(&eventBuffer[9], BaseAddress, 4);
      else
        RtlZeroMemory(&eventBuffer[9], 4);
      if (MmIsAddressValid(SectionOffset))
        RtlCopyMemory(&eventBuffer[13], SectionOffset, 8);
      else
        RtlZeroMemory(&eventBuffer[13], 8);
      if (MmIsAddressValid(ViewSize))
        RtlCopyMemory(&eventBuffer[21], ViewSize, 4);
      else
        RtlZeroMemory(&eventBuffer[21], 4);
      eventBuffer[25] = (char) InheritDisposition;
      RtlCopyMemory(&eventBuffer[26], &AllocationType, 4);
      RtlCopyMemory(&eventBuffer[30], &Protect, 2);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    ObDereferenceObject(sectionObject);
  }

  // Return the saved return value

  return retVal;
}

// HookUnmapViewOfSection is the ZwUnmapViewOfSection hook routine. It logs the
// call, then calls the original ZwUnmapViewOfSection.

NTSTATUS HookUnmapViewOfSection (IN HANDLE ProcessHandle,
                                 IN PVOID BaseAddress)
{
  KIRQL oldirql;
  PROCESS_BASIC_INFORMATION processInfo;
  NTSTATUS status;
  ULONG returnLength;
  char *eventBuffer;

  // Log the call

  status = ZwQueryInformationProcess(ProcessHandle, ProcessBasicInformation,
                                     &processInfo, sizeof(processInfo),
                                     &returnLength);
  if (status != STATUS_SUCCESS)
    processInfo.UniqueProcessId = 0xFFFFFFFF;
  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LogEvent((char) ENTRY_TYPE_SECTION_UNMAP_VIEW, 9);
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[1], &processInfo.UniqueProcessId, 4);
    RtlCopyMemory(&eventBuffer[5], &BaseAddress, 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Call the original function, and return its return value

  return (*globals.origUnmapViewOfSection)(ProcessHandle, BaseAddress);
}

// HookWaitForSingleObject is the ZwWaitForSingleObject hook routine.
// It logs the call, then calls the original ZwWaitForSingleObject,
// then logs the return.

ULONG HookWaitForSingleObject (HANDLE hHandle, ULONG flags,
                               PLARGE_INTEGER timeIn100Ns)
{
  KIRQL oldirql;
  ULONG retVal, threadId;
  PCHAR eventBuffer;

  // Log the call

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LogEvent((char) ENTRY_TYPE_WAITOBJ_WAIT_CALL, 14);
  if (eventBuffer) {
    eventBuffer[1] = (char) 14;
    if (MmIsAddressValid(timeIn100Ns))
      RtlCopyMemory(&eventBuffer[2], timeIn100Ns, 8);
    else
      RtlCopyMemory(&eventBuffer[2], "Infinite", 8);
    RtlCopyMemory(&eventBuffer[10], &hHandle, 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Indicate that the current thread is active.

  threadId = (ULONG) PsGetCurrentThreadId();
  if (threadId < sharedState->kernelUserSharedState.maxThreads)
    sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] = '\0';

  // Call the original Zw routine

  retVal = (*globals.origWaitForSingleObject)(hHandle, flags, timeIn100Ns);

  // Log the return

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LogEvent((char) ENTRY_TYPE_WAITOBJ_WAIT_RETURN, 5);
  if (eventBuffer)
    RtlCopyMemory(&eventBuffer[1], &retVal, 4);
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Return the original Zw routine's return

  return retVal;
}

// HookWaitForMultipleObjects is the ZwWaitForMultipleObjects hook routine.
// It logs the call, then calls the original ZwWaitForMultipleObjects,
// then logs the return.

ULONG HookWaitForMultipleObjects (ULONG numHandles, PHANDLE hHandles,
                                  ULONG flags1, ULONG flags2,
                                  PLARGE_INTEGER timeIn100Ns)
{
  KIRQL oldirql;
  ULONG retVal, objectsToRecord, i, threadId;
  PCHAR eventBuffer;

  // Log the call

  if (numHandles == 0 || MmIsAddressValid(hHandles)) {
    objectsToRecord = MIN(numHandles, MAX_WAITABLE_OBJECTS_RECORDED);
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LogEvent((char) ENTRY_TYPE_WAITOBJ_WAIT_CALL,
                           objectsToRecord * 4 + 10);
    if (eventBuffer) {
      eventBuffer[1] = (char) (objectsToRecord * 4 + 10);
      if (MmIsAddressValid(timeIn100Ns))
        RtlCopyMemory(&eventBuffer[2], timeIn100Ns, 8);
      else
        RtlCopyMemory(&eventBuffer[2], "Infinite", 8);
      for (i = 0; i < objectsToRecord; i++)
        RtlCopyMemory(&eventBuffer[i*4+10], &hHandles[i], 4);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  // Indicate that the current thread is active.

  threadId = (ULONG) PsGetCurrentThreadId();
  if (threadId < sharedState->kernelUserSharedState.maxThreads)
    sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] = '\0';

  // Call the original Zw routine

  retVal = (*globals.origWaitForMultipleObjects)(numHandles, hHandles, flags1,
                                                 flags2, timeIn100Ns);

  // Log the return

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LogEvent((char) ENTRY_TYPE_WAITOBJ_WAIT_RETURN, 5);
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[1], &retVal, 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Return the original Zw routine's return

  return retVal;
}

// HookSignalAndWaitForSingleObject is the ZwSignalAndWaitForSingleObject
// hook routine. It logs the call, then calls the original
// ZwSignalAndWaitForSingleObject, then logs the return.

ULONG HookSignalAndWaitForSingleObject (HANDLE objectToSignal,
                                        HANDLE objectToWaitFor, ULONG flags,
                                        PLARGE_INTEGER timeIn100Ns)
{
  KIRQL oldirql;
  ULONG retVal, threadId;
  PCHAR eventBuffer;

  // Log the call

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LogEvent((char) ENTRY_TYPE_WAITOBJ_SIGNAL, 5);
  if (eventBuffer)
    RtlCopyMemory(&eventBuffer[1], &objectToSignal, 4);
  eventBuffer = LogEvent((char) ENTRY_TYPE_WAITOBJ_WAIT_CALL, 14);
  if (eventBuffer) {
    eventBuffer[1] = (char) 14;
    if (MmIsAddressValid(timeIn100Ns))
      RtlCopyMemory(&eventBuffer[2], timeIn100Ns, 8);
    else
      RtlCopyMemory(&eventBuffer[2], "Infinite", 8);
    RtlCopyMemory(&eventBuffer[10], &objectToWaitFor, 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Indicate that the current thread is active.

  threadId = (ULONG) PsGetCurrentThreadId();
  if (threadId < sharedState->kernelUserSharedState.maxThreads)
    sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] = '\0';

  // Call the original Zw routine

  retVal = (*globals.origSignalAndWaitForSingleObject)(objectToSignal,
                                                       objectToWaitFor, flags,
                                                       timeIn100Ns);

  // Log the return

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LogEvent((char) ENTRY_TYPE_WAITOBJ_WAIT_RETURN, 5);
  if (eventBuffer)
    RtlCopyMemory(&eventBuffer[1], &retVal, 4);
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Return the original Zw routine's return

  return retVal;
}

// HookSetEvent is the ZwSetEvent hook routine. It calls the original
// ZwSetEvent, then logs the return if it was successful.

NTSTATUS HookSetEvent (HANDLE eventHandle, ULONG flags)
{
  NTSTATUS retVal;
  PCHAR eventBuffer;
  KIRQL oldirql;

  retVal = (*globals.origSetEvent)(eventHandle, flags);

  if (retVal == STATUS_SUCCESS) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LogEvent((char) ENTRY_TYPE_WAITOBJ_SIGNAL, 5);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &eventHandle, 4);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  return retVal;
}

// HookReleaseMutant is the ZwReleaseMutant hook routine. It calls the original
// ZwReleaseMutant, then logs the return if it was successful.

NTSTATUS HookReleaseMutant (HANDLE mutantHandle, ULONG flags)
{
  NTSTATUS retVal;
  PCHAR eventBuffer;
  KIRQL oldirql;

  retVal = (*globals.origReleaseMutant)(mutantHandle, flags);

  if (retVal == STATUS_SUCCESS) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LogEvent((char) ENTRY_TYPE_WAITOBJ_SIGNAL, 5);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &mutantHandle, 4);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  return retVal;
}

// HookReleaseSemaphore is the ZwReleaseSemaphore hook routine. It calls the
// original ZwReleaseSemaphore, then logs the return if it was successful.

NTSTATUS HookReleaseSemaphore (HANDLE semaphoreHandle, ULONG countIncrease,
                               ULONG flags)
{
  NTSTATUS retVal;
  PCHAR eventBuffer;
  KIRQL oldirql;

  retVal = (*globals.origReleaseSemaphore)(semaphoreHandle, countIncrease,
                                           flags);

  // Only log the release if the count increase was nonzero. If it was
  // zero, this "release" is just a null operation.

  if (retVal == STATUS_SUCCESS && countIncrease) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LogEvent((char) ENTRY_TYPE_WAITOBJ_SIGNAL, 5);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &semaphoreHandle, 4);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  return retVal;
}

// HookSetTimer is the ZwSetTimer hook routine. It calls the original
// ZwSetTimer, then logs the return if it was successful.

NTSTATUS HookSetTimer (HANDLE timerHandle, PLARGE_INTEGER timeIn100Ns,
                       ULONG flags1, ULONG flags2, ULONG flags3,
                       ULONG periodInMs, ULONG flags4)
{
  NTSTATUS retVal;
  PCHAR eventBuffer;
  KIRQL oldirql;

  retVal = (*globals.origSetTimer)(timerHandle, timeIn100Ns, flags1,
                                   flags2, flags3, periodInMs, flags4);

  if (retVal == STATUS_SUCCESS) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LogEvent((char) ENTRY_TYPE_WAITOBJ_SET_TIMER, 17);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &timerHandle, 4);
      RtlCopyMemory(&eventBuffer[5], &periodInMs, 4);
      if (MmIsAddressValid(timeIn100Ns))
        RtlCopyMemory(&eventBuffer[9], timeIn100Ns, 8);
      else
        RtlZeroMemory(&eventBuffer[9], 8);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  return retVal;
}

// HookCancelTimer is the ZwCancelTimer hook routine. It calls the original
// ZwCancelTimer, then logs the return if it was successful.

NTSTATUS HookCancelTimer (HANDLE timerHandle, ULONG flags)
{
  NTSTATUS retVal;
  PCHAR eventBuffer;
  KIRQL oldirql;

  retVal = (*globals.origCancelTimer)(timerHandle, flags);

  if (retVal == STATUS_SUCCESS) {
    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LogEvent((char) ENTRY_TYPE_WAITOBJ_CANCEL_TIMER, 5);
    if (eventBuffer) {
      RtlCopyMemory(&eventBuffer[1], &timerHandle, 4);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
  }

  return retVal;
}

////////////////////////////////////////////////////////////////////////////
//                    INSTALLATION AND UNINSTALLATION
////////////////////////////////////////////////////////////////////////////

// InstallZwHooks installs all our Zw routine hooks, saving the original Zw
// routine locations first.

NTSTATUS InstallZwHooks (void)
{
  NTSTATUS status;
  char *systemCallNames[NUM_SYSTEM_CALLS_HOOKED];

  // Initialize the system call names (without Zw's) so that we can find
  // their corresponding system call numbers
  
  systemCallNames[OPEN_FILE_INDEX] = "OpenFile";
  systemCallNames[CREATE_FILE_INDEX] = "CreateFile";
  systemCallNames[WAIT_FOR_SINGLE_INDEX] = "WaitForSingleObject";
  systemCallNames[WAIT_FOR_MULTIPLE_INDEX] = "WaitForMultipleObjects",
  systemCallNames[SIGNAL_AND_WAIT_INDEX] = "SignalAndWaitForSingleObject",
  systemCallNames[FLUSH_ICACHE_INDEX] = "FlushInstructionCache";
  systemCallNames[FLUSH_WRITE_BUFFER_INDEX] = "FlushWriteBuffer";
  systemCallNames[TERMINATE_PROCESS_INDEX] = "TerminateProcess";
  systemCallNames[TERMINATE_THREAD_INDEX] = "TerminateThread";
  systemCallNames[WRITE_REQUEST_DATA_INDEX] = "WriteRequestData";
  systemCallNames[WRITE_VM_INDEX] = "WriteVirtualMemory";
  systemCallNames[CREATE_SECTION_INDEX] = "CreateSection";
  systemCallNames[OPEN_SECTION_INDEX] = "OpenSection";
  systemCallNames[MAP_VIEW_INDEX] = "MapViewOfSection";
  systemCallNames[UNMAP_VIEW_INDEX] = "UnmapViewOfSection";
  systemCallNames[SET_EVENT_INDEX] = "SetEvent";
  systemCallNames[RELEASE_MUTANT_INDEX] = "ReleaseMutant";
  systemCallNames[RELEASE_SEMAPHORE_INDEX] = "ReleaseSemaphore";
  systemCallNames[SET_TIMER_INDEX] = "SetTimer";
  systemCallNames[CANCEL_TIMER_INDEX] = "CancelTimer";

  // Search NTDLL.DLL for the system call numbers and put them in the
  // array globals.systemCallNumbers

  status = GetSystemCallNumbers(systemCallNames, globals.systemCallNumbers,
                                NUM_SYSTEM_CALLS_HOOKED);
  if (status != STATUS_SUCCESS)
    return status;

  // For each system call ZwXXX, save the old function pointer in
  // globals.origXXX, then make the service table now point to our
  // HookXXX function

  globals.origOpenFile =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[OPEN_FILE_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[OPEN_FILE_INDEX]] =
      (PVOID) HookOpenFile;

  globals.origCreateFile =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[CREATE_FILE_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[CREATE_FILE_INDEX]] =
      (PVOID) HookCreateFile;

  globals.origWaitForSingleObject =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[WAIT_FOR_SINGLE_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[WAIT_FOR_SINGLE_INDEX]] =
      (PVOID) HookWaitForSingleObject;

  globals.origWaitForMultipleObjects =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[WAIT_FOR_MULTIPLE_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[WAIT_FOR_MULTIPLE_INDEX]] =
      (PVOID) HookWaitForMultipleObjects;

  globals.origSignalAndWaitForSingleObject =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[SIGNAL_AND_WAIT_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[SIGNAL_AND_WAIT_INDEX]] =
      (PVOID) HookSignalAndWaitForSingleObject;

  globals.origFlushInstructionCache =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[FLUSH_ICACHE_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[FLUSH_ICACHE_INDEX]] =
      (PVOID) HookFlushInstructionCache;

  globals.origFlushWriteBuffer =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[FLUSH_WRITE_BUFFER_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[FLUSH_WRITE_BUFFER_INDEX]] =
      (PVOID) HookFlushWriteBuffer;

  globals.origTerminateProcess =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[TERMINATE_PROCESS_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[TERMINATE_PROCESS_INDEX]] =
      (PVOID) HookTerminateProcess;

  globals.origTerminateThread =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[TERMINATE_THREAD_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[TERMINATE_THREAD_INDEX]] =
      (PVOID) HookTerminateThread;

  globals.origWriteRequestData =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[WRITE_REQUEST_DATA_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[WRITE_REQUEST_DATA_INDEX]] =
      (PVOID) HookWriteRequestData;

  globals.origWriteVirtualMemory =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[WRITE_VM_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[WRITE_VM_INDEX]] =
      (PVOID) HookWriteVirtualMemory;

  globals.origCreateSection =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[CREATE_SECTION_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[CREATE_SECTION_INDEX]] =
      (PVOID) HookCreateSection;

  globals.origOpenSection =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[OPEN_SECTION_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[OPEN_SECTION_INDEX]] =
      (PVOID) HookOpenSection;

  globals.origMapViewOfSection =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[MAP_VIEW_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[MAP_VIEW_INDEX]] =
      (PVOID) HookMapViewOfSection;

  globals.origUnmapViewOfSection =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[UNMAP_VIEW_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[UNMAP_VIEW_INDEX]] =
      (PVOID) HookUnmapViewOfSection;

  globals.origSetEvent =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[SET_EVENT_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[SET_EVENT_INDEX]] =
      (PVOID) HookSetEvent;

  globals.origReleaseMutant =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[RELEASE_MUTANT_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[RELEASE_MUTANT_INDEX]] =
      (PVOID) HookReleaseMutant;

  globals.origReleaseSemaphore =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[RELEASE_SEMAPHORE_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[RELEASE_SEMAPHORE_INDEX]] =
      (PVOID) HookReleaseSemaphore;

  globals.origSetTimer =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[SET_TIMER_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[SET_TIMER_INDEX]] =
      (PVOID) HookSetTimer;

  globals.origCancelTimer =
    KeServiceDescriptorTable->ServiceTable
      [globals.systemCallNumbers[CANCEL_TIMER_INDEX]];
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[CANCEL_TIMER_INDEX]] =
      (PVOID) HookCancelTimer;

  return status;
}

// UninstallZwHooks uninstalls all our Zw routine hooks by replacing them
// with the original routines.

void UninstallZwHooks (void)
{
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[OPEN_FILE_INDEX]] =
      globals.origOpenFile;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[CREATE_FILE_INDEX]] =
      globals.origCreateFile;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[WAIT_FOR_SINGLE_INDEX]] =
      globals.origWaitForSingleObject;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[WAIT_FOR_MULTIPLE_INDEX]] =
      globals.origWaitForMultipleObjects;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[SIGNAL_AND_WAIT_INDEX]] =
      globals.origSignalAndWaitForSingleObject;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[FLUSH_ICACHE_INDEX]] =
      globals.origFlushInstructionCache;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[FLUSH_WRITE_BUFFER_INDEX]] =
      globals.origFlushWriteBuffer;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[TERMINATE_PROCESS_INDEX]] =
      globals.origTerminateProcess;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[TERMINATE_THREAD_INDEX]] =
      globals.origTerminateThread;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[WRITE_REQUEST_DATA_INDEX]] =
      globals.origWriteRequestData;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[WRITE_VM_INDEX]] =
      globals.origWriteVirtualMemory;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[CREATE_SECTION_INDEX]] =
      globals.origCreateSection;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[OPEN_SECTION_INDEX]] =
      globals.origOpenSection;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[MAP_VIEW_INDEX]] =
      globals.origMapViewOfSection;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[UNMAP_VIEW_INDEX]] =
      globals.origUnmapViewOfSection;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[SET_EVENT_INDEX]] =
      globals.origSetEvent;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[RELEASE_MUTANT_INDEX]] =
      globals.origReleaseMutant;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[RELEASE_SEMAPHORE_INDEX]] =
      globals.origReleaseSemaphore;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[SET_TIMER_INDEX]] =
      globals.origSetTimer;
  KeServiceDescriptorTable->ServiceTable
    [globals.systemCallNumbers[CANCEL_TIMER_INDEX]] =
      globals.origCancelTimer;
}
