// 
// Handlers.c
//
// Code for intercepting calls to network event handlers in VTrcNet.
//

#include "NTDDK.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "TdiKrnl.h"
#include "GetState.h"

////////////////////////////////////////////////////////////////////////////
//                           EVENT HANDLERS
////////////////////////////////////////////////////////////////////////////

// VTrcNetClientEventConnect handles a connection event by calling the
// "real" event handler.  Then, it logs the call and computed values.  Then,
// it returns what the real event handler returned.

NTSTATUS VTrcNetClientEventConnect (IN PVOID TdiEventContext,
                                    IN LONG RemoteAddressLength,
                                    IN PVOID RemoteAddress,
                                    IN LONG UserDataLength,
                                    IN PVOID UserData,
                                    IN LONG OptionsLength,
                                    IN PVOID Options,
                                    OUT CONNECTION_CONTEXT *ConnectionContext,
                                    OUT PIRP *AcceptIrp)
{
  // The context is a pointer to a VTRCNET_FILE_OBJECT_INFO structure, from
  // which we can derive the real event handler and its context

  VTRCNET_FILE_OBJECT_INFO *info = TdiEventContext;
  PTDI_IND_CONNECT *fn = (PTDI_IND_CONNECT *) &info->connectEventHandler;
  PFILE_OBJECT fileObject = info->fileObject;
  int addressLength;
  KIRQL oldirql;
  NTSTATUS retVal;
  PCHAR eventBuffer;

  // Call the real event handler, obtained from the info structure

  retVal = (*fn)(info->connectEventContext, RemoteAddressLength, RemoteAddress,
                 UserDataLength, UserData, OptionsLength, Options,
                 ConnectionContext, AcceptIrp);

  // Log the important parameters, including remote address

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  if (RemoteAddressLength == 0 || MmIsAddressValid(RemoteAddress)) {
    addressLength = (RemoteAddressLength > 128) ? 128 : RemoteAddressLength;
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_NOTIFY_CONNECT,
                            addressLength + 10);
    if (eventBuffer) {
      eventBuffer[1] = (char) (addressLength + 10);
      RtlCopyMemory(&eventBuffer[10], RemoteAddress, addressLength);
    }
  }
  else {
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_NOTIFY_CONNECT, 13);
    if (eventBuffer) {
      eventBuffer[1] = (char) 13;
      RtlCopyMemory(&eventBuffer[10], "???", 3);
    }
  }
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[2], &fileObject, 4);
    if (MmIsAddressValid(ConnectionContext))
      RtlCopyMemory(&eventBuffer[6], ConnectionContext, 4);
    else
      RtlZeroMemory(&eventBuffer[6], 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  return retVal;
}

// VTrcNetClientEventDisconnect handles a disconnection event by calling the
// "real" event handler.  But first, it logs the call.

NTSTATUS VTrcNetClientEventDisconnect (IN PVOID TdiEventContext,
                                       IN CONNECTION_CONTEXT ConnectionContext,
                                       IN LONG DisconnectDataLength,
                                       IN PVOID DisconnectData,
                                       IN LONG DisconnectInformationLength,
                                       IN PVOID DisconnectInformation,
                                       IN ULONG DisconnectFlags)
{
  // The context is a pointer to a VTRCNET_FILE_OBJECT_INFO structure, from
  // which we can derive the real event handler and its context

  VTRCNET_FILE_OBJECT_INFO *info = TdiEventContext;
  PTDI_IND_DISCONNECT *fn =
                      (PTDI_IND_DISCONNECT *) &info->disconnectEventHandler;
  PFILE_OBJECT fileObject = info->fileObject;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the important parameters

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_NOTIFY_DISCONNECT, 9);
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[1], &fileObject, 4);
    RtlCopyMemory(&eventBuffer[5], &ConnectionContext, 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Call the "real" event handler

  return (*fn)(info->disconnectEventContext, ConnectionContext,
               DisconnectDataLength, DisconnectData,
               DisconnectInformationLength, DisconnectInformation,
               DisconnectFlags);
}

// VTrcNetClientEventReceive handles a receive event by calling the
// "real" event handler.  But first, it logs the call.

NTSTATUS VTrcNetClientEventReceive (IN PVOID TdiEventContext,
                                    IN CONNECTION_CONTEXT ConnectionContext,
                                    IN ULONG ReceiveFlags,
                                    IN ULONG BytesIndicated,
                                    IN ULONG BytesAvailable,
                                    OUT ULONG *BytesTaken,
                                    IN PVOID Tsdu,
                                    OUT PIRP *IoRequestPacket)
{
  // The context is a pointer to a VTRCNET_FILE_OBJECT_INFO structure, from
  // which we can derive the real event handler and its context

  VTRCNET_FILE_OBJECT_INFO *info = TdiEventContext;
  PTDI_IND_RECEIVE *fn = (PTDI_IND_RECEIVE *) &info->receiveEventHandler;
  PFILE_OBJECT fileObject = info->fileObject;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the important parameters

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_NOTIFY_RECEIVE, 13);
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[1], &fileObject, 4);
    RtlCopyMemory(&eventBuffer[5], &BytesIndicated, 4);
    RtlCopyMemory(&eventBuffer[9], &ConnectionContext, 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Call the "real" event handler

  return (*fn)(info->receiveEventContext, ConnectionContext, ReceiveFlags,
               BytesIndicated, BytesAvailable, BytesTaken, Tsdu,
               IoRequestPacket);
}

// VTrcNetClientEventReceiveDatagram handles a receive-datagram event by
// calling the "real" event handler.  But first, it logs the call.

NTSTATUS VTrcNetClientEventReceiveDatagram (IN PVOID TdiEventContext,
                                            IN LONG SourceAddressLength,
                                            IN PVOID SourceAddress,
                                            IN LONG OptionsLength,
                                            IN PVOID Options,
                                            IN ULONG ReceiveDatagramFlags,
                                            IN ULONG BytesIndicated,
                                            IN ULONG BytesAvailable,
                                            OUT ULONG *BytesTaken,
                                            IN PVOID Tsdu,
                                            OUT PIRP *IoRequestPacket)
{
  // The context is a pointer to a VTRCNET_FILE_OBJECT_INFO structure, from
  // which we can derive the real event handler and its context

  VTRCNET_FILE_OBJECT_INFO *info = TdiEventContext;
  PTDI_IND_RECEIVE_DATAGRAM *fn =
             (PTDI_IND_RECEIVE_DATAGRAM *) &info->receiveDatagramEventHandler;
  PFILE_OBJECT fileObject = info->fileObject;
  int addressLength;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the relevant parameters, including the address of where this
  // datagram came from

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);

  if (SourceAddressLength == 0 || MmIsAddressValid(SourceAddress)) {
    addressLength = (SourceAddressLength > 128) ? 128 : SourceAddressLength;
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_NOTIFY_RECEIVE_DATAGRAM,
                            addressLength + 10);
    if (eventBuffer) {
      eventBuffer[1] = (char) (addressLength + 10);
      RtlCopyMemory(&eventBuffer[10], SourceAddress, addressLength);
    }
  }
  else {
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_NOTIFY_RECEIVE_DATAGRAM,
                            13);
    if (eventBuffer) {
      eventBuffer[1] = (char) 13;
      RtlCopyMemory(&eventBuffer[10], "???", 3);
    }
  }
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[2], &fileObject, 4);
    RtlCopyMemory(&eventBuffer[6], &BytesIndicated, 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Call the "real" event handler

  return (*fn)(info->receiveDatagramEventContext, SourceAddressLength,
               SourceAddress, OptionsLength, Options, ReceiveDatagramFlags,
               BytesIndicated, BytesAvailable, BytesTaken, Tsdu,
               IoRequestPacket);
}


// VTrcNetClientEventReceiveExpedited handles a receive-expedited event
// by calling the "real" event handler.  But first, it logs the call.

NTSTATUS VTrcNetClientEventReceiveExpedited
  (IN PVOID TdiEventContext, IN CONNECTION_CONTEXT ConnectionContext,
   IN ULONG ReceiveFlags, IN ULONG BytesIndicated, IN ULONG BytesAvailable,
   OUT ULONG *BytesTaken, IN PVOID Tsdu, OUT PIRP *IoRequestPacket)
{
  // The context is a pointer to a VTRCNET_FILE_OBJECT_INFO structure, from
  // which we can derive the real event handler and its context

  VTRCNET_FILE_OBJECT_INFO *info = TdiEventContext;
  PTDI_IND_RECEIVE_EXPEDITED *fn =
        (PTDI_IND_RECEIVE_EXPEDITED *) &info->receiveExpeditedEventHandler;
  PFILE_OBJECT fileObject = info->fileObject;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the relevant parameters

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_NOTIFY_RECEIVE_EXPEDITED,
                          13);
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[1], &fileObject, 4);
    RtlCopyMemory(&eventBuffer[5], &BytesIndicated, 4);
    RtlCopyMemory(&eventBuffer[9], &ConnectionContext, 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Call the "real" event handler

  return (*fn)(info->receiveExpeditedEventContext, ConnectionContext,
               ReceiveFlags, BytesIndicated, BytesAvailable, BytesTaken,
               Tsdu, IoRequestPacket);
}

// VTrcNetClientEventChainedReceive handles a chained-receive event by
// calling the "real" event handler.  But first, it logs the call.

NTSTATUS
VTrcNetClientEventChainedReceive (
    IN PVOID TdiEventContext,
    IN CONNECTION_CONTEXT ConnectionContext,
    IN ULONG ReceiveFlags,
    IN ULONG ReceiveLength,        // length of client data in TSDU
    IN ULONG StartingOffset,       // offset of start of client data in TSDU
    IN PMDL  Tsdu,                 // TSDU data chain
    IN PVOID TsduDescriptor        // for call to TdiReturnChainedReceives
    )
{
  // The context is a pointer to a VTRCNET_FILE_OBJECT_INFO structure, from
  // which we can derive the real event handler and its context

  VTRCNET_FILE_OBJECT_INFO *info = TdiEventContext;
  PTDI_IND_CHAINED_RECEIVE *fn = (PTDI_IND_CHAINED_RECEIVE *)
                                        &info->chainedReceiveEventHandler;
  PFILE_OBJECT fileObject = info->fileObject;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the important parameters

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_NOTIFY_RECEIVE, 13);
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[1], &fileObject, 4);
    RtlCopyMemory(&eventBuffer[5], &ReceiveLength, 4);
    RtlCopyMemory(&eventBuffer[9], &ConnectionContext, 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Call the "real" event handler

  return (*fn)(info->chainedReceiveEventContext, ConnectionContext,
               ReceiveFlags, ReceiveLength, StartingOffset, Tsdu,
               TsduDescriptor);
}

// VTrcNetClientEventChainedReceiveDatagram handles a chained-receive-
// datagram event by calling the "real" event handler.  But first, it logs
// the call.

NTSTATUS
VTrcNetClientEventChainedReceiveDatagram (
    IN PVOID TdiEventContext,       // the event context
    IN LONG SourceAddressLength,    // length of the originator of the datagram
    IN PVOID SourceAddress,         // string describing the originator of the
                                    // datagram
    IN LONG OptionsLength,          // options for the receive
    IN PVOID Options,               //
    IN ULONG ReceiveDatagramFlags,  //
    IN ULONG ReceiveDatagramLength, // length of client data in TSDU
    IN ULONG StartingOffset,        // offset of start of client data in TSDU
    IN PMDL  Tsdu,                  // TSDU data chain
    IN PVOID TsduDescriptor         // for call to TdiReturnChainedReceives
    )
{
  // The context is a pointer to a VTRCNET_FILE_OBJECT_INFO structure, from
  // which we can derive the real event handler and its context

  VTRCNET_FILE_OBJECT_INFO *info = TdiEventContext;
  PTDI_IND_CHAINED_RECEIVE_DATAGRAM *fn = (PTDI_IND_CHAINED_RECEIVE_DATAGRAM *)
                                     &info->chainedReceiveDatagramEventHandler;
  PFILE_OBJECT fileObject = info->fileObject;
  int addressLength;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the relevant parameters, including the address of where this
  // datagram came from

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  if (SourceAddressLength == 0 || MmIsAddressValid(SourceAddress)) {
    addressLength = (SourceAddressLength > 128) ? 128 : SourceAddressLength;
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_NOTIFY_RECEIVE_DATAGRAM,
                            addressLength + 10);
    if (eventBuffer) {
      eventBuffer[1] = (char) (addressLength + 10);
      RtlCopyMemory(&eventBuffer[10], SourceAddress, addressLength);
    }
  }
  else {
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_NOTIFY_RECEIVE_DATAGRAM,
                            13);
    if (eventBuffer) {
      eventBuffer[1] = (char) 13;
      RtlCopyMemory(&eventBuffer[10], "???", 3);
    }
  }
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[2], &fileObject, 4);
    RtlCopyMemory(&eventBuffer[6], &ReceiveDatagramLength, 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Call the "real" event handler

  return (*fn)(info->chainedReceiveDatagramEventContext, SourceAddressLength,
               SourceAddress, OptionsLength, Options, ReceiveDatagramFlags,
               ReceiveDatagramLength, StartingOffset, Tsdu, TsduDescriptor);
}

// VTrcNetClientEventChainedReceiveExpedited handles a chained-receive-
// expedited event by calling the "real" event handler.  But first, it logs
// the call.

NTSTATUS
VTrcNetClientEventChainedReceiveExpedited (
    IN PVOID TdiEventContext,
    IN CONNECTION_CONTEXT ConnectionContext,
    IN ULONG ReceiveFlags,
    IN ULONG ReceiveLength,      // length of client data in TSDU
    IN ULONG StartingOffset,     // offset of start of client data in TSDU
    IN PMDL  Tsdu,               // TSDU data chain
    IN PVOID TsduDescriptor      // for call to TdiReturnChainedReceives
    )
{
  // The context is a pointer to a VTRCNET_FILE_OBJECT_INFO structure, from
  // which we can derive the real event handler and its context

  VTRCNET_FILE_OBJECT_INFO *info = TdiEventContext;
  PTDI_IND_CHAINED_RECEIVE_EXPEDITED *fn =
                            (PTDI_IND_CHAINED_RECEIVE_EXPEDITED *)
                            &info->chainedReceiveExpeditedEventHandler;
  PFILE_OBJECT fileObject = info->fileObject;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the relevant parameters

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_NOTIFY_RECEIVE_EXPEDITED,
                          13);
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[1], &fileObject, 4);
    RtlCopyMemory(&eventBuffer[5], &ReceiveLength, 4);
    RtlCopyMemory(&eventBuffer[9], &ConnectionContext, 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Call the "real" event handler

  return (*fn)(info->chainedReceiveExpeditedEventContext, ConnectionContext,
               ReceiveFlags, ReceiveLength, StartingOffset, Tsdu,
               TsduDescriptor);
}
