// 
// ObjInfo.c
//
// Code for handling file object information structures and the lists of
// such structures.  Each such structure describes the information VTrcNet
// needs to keep around about a single open file object.
//

#include "NTDDK.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "ObjInfo.h"

////////////////////////////////////////////////////////////////////////////
//                               MACROS
////////////////////////////////////////////////////////////////////////////

// The number of VTRCNET_FILE_OBJECT_INFO structures to allocate at
// a time:

#define FILE_OBJECT_INFO_BLOCK_SIZE           8

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, InitializeFileObjectInfoStructures)
#endif

////////////////////////////////////////////////////////////////////////////
//                              GLOBALS
////////////////////////////////////////////////////////////////////////////

// The file object info structures are linked on the following free list

VTRCNET_FILE_OBJECT_INFO *fileObjectInfoFreeList;

// Access to any file object info list (including the free list) is
// protected by the following spin lock

KSPIN_LOCK fileObjectInfoLock;

////////////////////////////////////////////////////////////////////////////
//                      FILE OBJECT INFO LIST FUNCTIONS
////////////////////////////////////////////////////////////////////////////

// InitializeFileObjectInfoStructures initializes global variables we need
// to do operations on file object information structures.

void InitializeFileObjectInfoStructures (void)
{
  // Initialize the file object info free list and spin lock

  KeInitializeSpinLock(&fileObjectInfoLock);
  fileObjectInfoFreeList = NULL;
}

// AllocateBlockOfFileObjectInfos allocates a block of file object
// info structures and puts them on the free list.  It returns a
// boolean indicating whether it successfully completed.  It assumes
// that the caller has acquired the file object info spin lock.

BOOLEAN AllocateBlockOfFileObjectInfos (void)
{
  VTRCNET_FILE_OBJECT_INFO *info;
  int i;

  // Allocate the block of memory

  info = ExAllocatePool(NonPagedPool, sizeof(VTRCNET_FILE_OBJECT_INFO) *
                                      FILE_OBJECT_INFO_BLOCK_SIZE);
  if (info == NULL)
    return FALSE;

  // Link each structure to the next one, then put the entire list
  // on the free list by pointing the last one to the free list and
  // the free list to the first one.

  for (i = 0; i < FILE_OBJECT_INFO_BLOCK_SIZE - 1; i++)
    info[i].nextInfo = &info[i+1];

  info[FILE_OBJECT_INFO_BLOCK_SIZE - 1].nextInfo = fileObjectInfoFreeList;
  fileObjectInfoFreeList = info;

  return TRUE;
}

// NewFileObjectInfo gets a new file object info structure from the
// free list.  (If the free list is empty, it extends it; if it cannot
// be extended, it returns NULL.)  It assumes that the caller has acquired
// the file object info spin lock

VTRCNET_FILE_OBJECT_INFO *NewFileObjectInfo (void)
{
  VTRCNET_FILE_OBJECT_INFO *info;

  // If the free list is empty, extend it.  (If that extension fails,
  // return NULL.)

  if (fileObjectInfoFreeList == NULL)
    if (!AllocateBlockOfFileObjectInfos())
      return NULL;

  // Return the first element of the free list, pointing the free list
  // to the next element

  info = fileObjectInfoFreeList;
  fileObjectInfoFreeList = info->nextInfo;
  return info;
}

// FreeFileObjectInfo adds the file object info structure pointed to
// by info to the free list.  It assumes that the caller has acquired
// the file object info spin lock.

void FreeFileObjectInfo (VTRCNET_FILE_OBJECT_INFO *info)
{
  info->nextInfo = fileObjectInfoFreeList;
  fileObjectInfoFreeList = info;
}

// FindFileObjectInfo finds the file object information for a given
// file object in the given hook extension.  If one doesn't exist, it
// creates one and adds it to the list.  It returns the file object
// info pointer, or NULL if it can't find one and can't create one.

VTRCNET_FILE_OBJECT_INFO *FindFileObjectInfo (PNET_HOOK_EXTENSION hookExt,
                                              PFILE_OBJECT fileObject)
{
  VTRCNET_FILE_OBJECT_INFO *info;
  KIRQL oldirql;

  // Acquire spin lock for the file object info list

  KeAcquireSpinLock(&fileObjectInfoLock, &oldirql);

  // Search through the list for an info structure with the given file
  // object.  If one exists, return it... but only after releasing the
  // spin lock we're holding!

  for (info = hookExt->fileObjectInfoList; info != NULL;
       info = info->nextInfo) {
    if (info->fileObject == fileObject) {
      KeReleaseSpinLock(&fileObjectInfoLock, oldirql);
      return info;
    }
  }

  // At this point we know it isn't on the list, so we create a new
  // file object info structure and add it to the list.

  // First, try to allocate a structure.  If we fail, release the spin
  // lock and return.

  info = NewFileObjectInfo();
  if (info == NULL) {
    KeReleaseSpinLock(&fileObjectInfoLock, oldirql);
    return NULL;
  }

  // Then, fill in the structure and add it to the front of the list

  info->fileObject = fileObject;
  info->nextInfo = hookExt->fileObjectInfoList;
  hookExt->fileObjectInfoList = info;

  // Release the spin lock and return

  KeReleaseSpinLock(&fileObjectInfoLock, oldirql);
  return info;
}

// FreeInfoOfFileObject frees the file object info structure corresponding
// to a given file object in a hook extension, and removes it from that file
// object info list.

void FreeInfoOfFileObject (PNET_HOOK_EXTENSION hookExt,
                           PFILE_OBJECT fileObject)
{
  VTRCNET_FILE_OBJECT_INFO *cur, *next;
  KIRQL oldirql;

  // Acquire the spin lock for the file object info structure.

  KeAcquireSpinLock(&fileObjectInfoLock, &oldirql);

  // If the list is empty, we're done.

  if (hookExt->fileObjectInfoList == NULL) {
    KeReleaseSpinLock(&fileObjectInfoLock, oldirql);
    return;
  }

  // If the first element of the list matches our file object, remove
  // it from there and leave

  if (hookExt->fileObjectInfoList->fileObject == fileObject) {
    next = hookExt->fileObjectInfoList->nextInfo;
    FreeFileObjectInfo(hookExt->fileObjectInfoList);
    hookExt->fileObjectInfoList = next;
    KeReleaseSpinLock(&fileObjectInfoLock, oldirql);
    return;
  }

  // Otherwise, look through the list for it.  If and when we find it,
  // excise it and leave.

  for (cur = hookExt->fileObjectInfoList; cur->nextInfo != NULL;
       cur = cur->nextInfo) {
    if (cur->nextInfo->fileObject == fileObject) {
      next = cur->nextInfo->nextInfo;
      FreeFileObjectInfo(cur->nextInfo);
      cur->nextInfo = next;
      KeReleaseSpinLock(&fileObjectInfoLock, oldirql);
      return;
    }
  }

  // We didn't find it, so just leave (after releasing spin lock)

  KeReleaseSpinLock(&fileObjectInfoLock, oldirql);
}
