// 
// TdiDisp.c
//
// Dispatch routines for TDI internal device control requests.
//

#include "NTDDK.h"
#include "TdiKrnl.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "GetState.h"
#include "Handlers.h"
#include "ObjInfo.h"
#include "VTrcNet.h"

////////////////////////////////////////////////////////////////////////////
//                      TDI REQUEST DISPATCH ROUTINES
////////////////////////////////////////////////////////////////////////////

// VTrcNetTdiDispatchConnect handles an IRP representing a TDI request
// to connect, by tracing any relevant parameters and passing on the IRP
// in a way that ensures that its completion can be logged as well.
// An additional parameter is provided that points to important information
// about this TDI request passed in the current IRP stack of the IRP.

NTSTATUS VTrcNetTdiDispatchConnect (PDEVICE_OBJECT HookDevice, PIRP Irp,
                                    PFILE_OBJECT fileObject,
                                    PTDI_REQUEST_KERNEL requestKernel)
{
  ULONG seq, addressLength;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the event

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  if (MmIsAddressValid(requestKernel) &&
      MmIsAddressValid(requestKernel->RequestConnectionInformation) &&
      MmIsAddressValid(requestKernel->RequestConnectionInformation->
                       RemoteAddress)) {
    addressLength =
            requestKernel->RequestConnectionInformation->RemoteAddressLength;
    if (addressLength > 128) addressLength = 128;
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_CONNECT,
                            addressLength + 18);
    if (eventBuffer) {
      eventBuffer[1] = (char) (addressLength + 18);
      RtlCopyMemory(&eventBuffer[18],
                    requestKernel->RequestConnectionInformation->RemoteAddress,
                    addressLength);
    }
  }
  else {
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_CONNECT, 21);
    if (eventBuffer) {
      eventBuffer[1] = (char) 21;
      RtlCopyMemory(&eventBuffer[18], "???", 3);
    }
  }
  if (eventBuffer) {
    seq = InterlockedIncrement(&sequenceNumber);    // advance sequence number
    RtlCopyMemory(&eventBuffer[2], &seq, 4);
    RtlCopyMemory(&eventBuffer[6], &fileObject, 4);

    if (MmIsAddressValid(requestKernel) &&
        MmIsAddressValid(requestKernel->RequestSpecific))
      RtlCopyMemory(&eventBuffer[10], requestKernel->RequestSpecific, 8);
    else
      RtlCopyMemory(&eventBuffer[10], "????????", 8);
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcNetPassOnWithTailLogging(HookDevice, Irp, seq);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcNetPassOnNormally(HookDevice, Irp);
  }
}

// VTrcNetTdiDispatchDisconnect handles an IRP representing a TDI request
// to disconnect, by tracing any relevant parameters and passing on the IRP
// in a way that ensures that its completion can be logged as well.
// An additional parameter is provided that points to important information
// about this TDI request passed in the current IRP stack of the IRP.

NTSTATUS VTrcNetTdiDispatchDisconnect
  (PDEVICE_OBJECT HookDevice, PIRP Irp, PFILE_OBJECT fileObject,
   PTDI_REQUEST_KERNEL_DISCONNECT requestKernelDisconnect)
{
  ULONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the event

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_DISCONNECT, 17);
  if (eventBuffer) {
    seq = InterlockedIncrement(&sequenceNumber);    // advance sequence number
    RtlCopyMemory(&eventBuffer[1], &seq, 4);
    RtlCopyMemory(&eventBuffer[5], &fileObject, 4);

    if (MmIsAddressValid(requestKernelDisconnect) &&
        MmIsAddressValid(requestKernelDisconnect->RequestSpecific))
      RtlCopyMemory(&eventBuffer[9], requestKernelDisconnect->RequestSpecific,
                    8);
    else
      RtlCopyMemory(&eventBuffer[9], "????????", 8);
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcNetPassOnWithTailLogging(HookDevice, Irp, seq);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcNetPassOnNormally(HookDevice, Irp);
  }
}

// VTrcNetTdiDispatchAssociate handles an IRP representing a TDI request
// to associate an address, by tracing any relevant parameters and passing
// on the IRP in a way that ensures that its completion can be logged as
// well.  An additional parameter is provided that points to important
// information about this TDI request passed in the current IRP stack of
// the IRP.

NTSTATUS VTrcNetTdiDispatchAssociate
  (PDEVICE_OBJECT HookDevice, PIRP Irp, PFILE_OBJECT fileObject,
   PTDI_REQUEST_KERNEL_ASSOCIATE requestKernelAssociate)
{
  ULONG seq;
  NTSTATUS status;
  PFILE_OBJECT associatedFileObject;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Get a pointer to the associated file object

  if (MmIsAddressValid(requestKernelAssociate))
    status = ObReferenceObjectByHandle(requestKernelAssociate->AddressHandle,
                                       0L, 0, KernelMode,
                                       &associatedFileObject, NULL);
  else
    status = STATUS_UNSUCCESSFUL;

  // If we successfully obtained the associated file object, use it to log
  // the call

  if (NT_SUCCESS(status)) {
    // Log the event

    KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_ASSOCIATE_ADDRESS, 13);
    if (eventBuffer) {
      seq = InterlockedIncrement(&sequenceNumber);  // advance sequence number
      RtlCopyMemory(&eventBuffer[1], &seq, 4);
      RtlCopyMemory(&eventBuffer[5], &fileObject, 4);
      RtlCopyMemory(&eventBuffer[9], &associatedFileObject, 4);
      KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
      // Dereference the associated file object -- we don't need it any more
      ObDereferenceObject(associatedFileObject);
      // Pass on the IRP to log the return
      return VTrcNetPassOnWithTailLogging(HookDevice, Irp, seq);
    }
    else {
      KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
      // Dereference the associated file object -- we don't need it any more
      ObDereferenceObject(associatedFileObject);
      // Pass on the IRP; no need to log the return
      return VTrcNetPassOnNormally(HookDevice, Irp);
    }
  }
  else {
    // Don't log the completion of the IRP, but complete it
    return VTrcNetPassOnNormally(HookDevice, Irp);
  }
}

// VTrcNetTdiDispatchDisassociate handles an IRP representing a TDI request
// to disassociate an address, by tracing any relevant parameters and passing
// on the IRP in a way that ensures that its completion can be logged as
// well.  An additional parameter is provided that points to important
// information about this TDI request passed in the current IRP stack of
// the IRP.

NTSTATUS VTrcNetTdiDispatchDisassociate
  (PDEVICE_OBJECT HookDevice, PIRP Irp, PFILE_OBJECT fileObject,
   PTDI_REQUEST_KERNEL_DISASSOCIATE requestKernelDisassociate)
{
  ULONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the event
  
  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_DISASSOCIATE_ADDRESS, 9);
  if (eventBuffer) {
    seq = InterlockedIncrement(&sequenceNumber);    // advance sequence number
    RtlCopyMemory(&eventBuffer[1], &seq, 4);
    RtlCopyMemory(&eventBuffer[5], &fileObject, 4);
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcNetPassOnWithTailLogging(HookDevice, Irp, seq);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcNetPassOnNormally(HookDevice, Irp);
  }
}

// VTrcNetTdiDispatchSend handles an IRP representing a TDI request
// to send data, by tracing any relevant parameters and passing
// on the IRP in a way that ensures that its completion can be logged as
// well.  An additional parameter is provided that points to important
// information about this TDI request passed in the current IRP stack of
// the IRP.

NTSTATUS VTrcNetTdiDispatchSend (PDEVICE_OBJECT HookDevice, PIRP Irp,
                                 PFILE_OBJECT fileObject,
                                 PTDI_REQUEST_KERNEL_SEND requestKernelSend)
{
  ULONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer;

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_SEND, 17);
  if (eventBuffer) {
    seq = InterlockedIncrement(&sequenceNumber);    // advance sequence number
    RtlCopyMemory(&eventBuffer[1], &seq, 4);
    RtlCopyMemory(&eventBuffer[5], &fileObject, 4);
    if (MmIsAddressValid(requestKernelSend)) {
      RtlCopyMemory(&eventBuffer[9], &requestKernelSend->SendLength, 4);
      RtlCopyMemory(&eventBuffer[13], &requestKernelSend->SendFlags, 4);
    }
    else
      RtlZeroMemory(&eventBuffer[9], 8);
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcNetPassOnWithTailLogging(HookDevice, Irp, seq);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcNetPassOnNormally(HookDevice, Irp);
  }
}

// VTrcNetTdiDispatchSendDatagram handles an IRP representing a TDI request
// to send a datagram, by tracing any relevant parameters and passing
// on the IRP in a way that ensures that its completion can be logged as
// well.  An additional parameter is provided that points to important
// information about this TDI request passed in the current IRP stack of
// the IRP.

NTSTATUS VTrcNetTdiDispatchSendDatagram
  (PDEVICE_OBJECT HookDevice, PIRP Irp, PFILE_OBJECT fileObject,
   PTDI_REQUEST_KERNEL_SENDDG requestKernelSendDg)
{
  ULONG seq, addressLength;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the event

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  if (MmIsAddressValid(requestKernelSendDg) &&
      MmIsAddressValid(requestKernelSendDg->SendDatagramInformation) &&
      MmIsAddressValid(requestKernelSendDg->SendDatagramInformation->
                       RemoteAddress)) {
    addressLength =
            requestKernelSendDg->SendDatagramInformation->RemoteAddressLength;
    if (addressLength > 128) addressLength = 128;
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_SEND_DATAGRAM,
                            addressLength + 14);
    if (eventBuffer) {
      eventBuffer[1] = (char) (addressLength + 14);
      RtlCopyMemory(&eventBuffer[14], requestKernelSendDg->
                                      SendDatagramInformation->RemoteAddress,
                    addressLength);
    }
  }
  else {
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_SEND_DATAGRAM, 17);
    if (eventBuffer) {
      eventBuffer[1] = (char) 17;
      RtlCopyMemory(&eventBuffer[14], "???", 3);
    }
  }
  if (eventBuffer) {
    seq = InterlockedIncrement(&sequenceNumber);    // advance sequence number
    RtlCopyMemory(&eventBuffer[2], &seq, 4);
    RtlCopyMemory(&eventBuffer[6], &fileObject, 4);
    if (MmIsAddressValid(requestKernelSendDg))
      RtlCopyMemory(&eventBuffer[10], &requestKernelSendDg->SendLength, 4);
    else
      RtlZeroMemory(&eventBuffer[10], 4);
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcNetPassOnWithTailLogging(HookDevice, Irp, seq);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcNetPassOnNormally(HookDevice, Irp);
  }
}

// VTrcNetTdiDispatchReceive handles an IRP representing a TDI request
// to receive data, by tracing any relevant parameters and passing
// on the IRP in a way that ensures that its completion can be logged as
// well.  An additional parameter is provided that points to important
// information about this TDI request passed in the current IRP stack of
// the IRP.

NTSTATUS VTrcNetTdiDispatchReceive
  (PDEVICE_OBJECT HookDevice, PIRP Irp, PFILE_OBJECT fileObject,
   PTDI_REQUEST_KERNEL_RECEIVE requestKernelReceive)
{
  ULONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the event

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_RECEIVE, 17);
  if (eventBuffer) {
    seq = InterlockedIncrement(&sequenceNumber);    // advance sequence number
    RtlCopyMemory(&eventBuffer[1], &seq, 4);
    RtlCopyMemory(&eventBuffer[5], &fileObject, 4);
    if (MmIsAddressValid(requestKernelReceive)) {
      RtlCopyMemory(&eventBuffer[9], &requestKernelReceive->ReceiveLength, 4);
      RtlCopyMemory(&eventBuffer[13], &requestKernelReceive->ReceiveFlags, 4);
    }
    else {
      RtlZeroMemory(&eventBuffer[9], 8);
    }
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcNetPassOnWithTailLogging(HookDevice, Irp, seq);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcNetPassOnNormally(HookDevice, Irp);
  }
}

// VTrcNetTdiDispatchReceiveDatagram handles an IRP representing a TDI
// request to receive a datagram, by tracing any relevant parameters and
// passing on the IRP in a way that ensures that its completion can be
// logged as well.  An additional parameter is provided that points to
// important information about this TDI request passed in the current IRP
// stack of the IRP.

NTSTATUS VTrcNetTdiDispatchReceiveDatagram
  (PDEVICE_OBJECT HookDevice, PIRP Irp, PFILE_OBJECT fileObject,
   PTDI_REQUEST_KERNEL_RECEIVEDG requestKernelReceiveDg)
{
  ULONG seq, addressLength;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the event

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  if (MmIsAddressValid(requestKernelReceiveDg) &&
      MmIsAddressValid(requestKernelReceiveDg->ReceiveDatagramInformation) &&
      MmIsAddressValid(requestKernelReceiveDg->ReceiveDatagramInformation->
                       RemoteAddress)) {
    addressLength =
      requestKernelReceiveDg->ReceiveDatagramInformation->RemoteAddressLength;
    if (addressLength > 128) addressLength = 128;
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_RECEIVE_DATAGRAM,
                            addressLength + 14);
    if (eventBuffer) {
      eventBuffer[1] = (char) (addressLength + 14);
      RtlCopyMemory(&eventBuffer[14],
                    requestKernelReceiveDg->ReceiveDatagramInformation->
                    RemoteAddress, addressLength);
    }
  }
  else {
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_RECEIVE_DATAGRAM, 17);
    if (eventBuffer) {
      eventBuffer[1] = (char) 17;
      RtlCopyMemory(&eventBuffer[14], "???", 3);
    }
  }
  if (eventBuffer) {
    seq = InterlockedIncrement(&sequenceNumber);    // advance sequence number
    RtlCopyMemory(&eventBuffer[2], &seq, 4);
    RtlCopyMemory(&eventBuffer[6], &fileObject, 4);
    if (MmIsAddressValid(requestKernelReceiveDg))
      RtlCopyMemory(&eventBuffer[10], &requestKernelReceiveDg->ReceiveLength,
                    4);
    else
      RtlZeroMemory(&eventBuffer[10], 4);
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcNetPassOnWithTailLogging(HookDevice, Irp, seq);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcNetPassOnNormally(HookDevice, Irp);
  }
}

// VTrcNetTdiDispatchListen handles an IRP representing a TDI request
// to listen for a connection, by tracing any relevant parameters and passing
// on the IRP in a way that ensures that its completion can be logged as
// well.  An additional parameter is provided that points to important
// information about this TDI request passed in the current IRP stack of
// the IRP.

NTSTATUS VTrcNetTdiDispatchListen
  (PDEVICE_OBJECT HookDevice, PIRP Irp, PFILE_OBJECT fileObject,
   PTDI_REQUEST_KERNEL_LISTEN requestKernelListen)
{
  ULONG seq, addressLength;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the event

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  if (MmIsAddressValid(requestKernelListen) &&
      MmIsAddressValid(requestKernelListen->RequestConnectionInformation) &&
      MmIsAddressValid(requestKernelListen->RequestConnectionInformation->
                       RemoteAddress)) {
    addressLength =
      requestKernelListen->RequestConnectionInformation->RemoteAddressLength;
    if (addressLength > 128) addressLength = 128;
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_LISTEN,
                            addressLength + 11);
    if (eventBuffer) {
      eventBuffer[1] = (char) addressLength + 11;
      RtlCopyMemory(&eventBuffer[11],
                    requestKernelListen->RequestConnectionInformation->
                    RemoteAddress, addressLength);
    }
  }
  else {
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_LISTEN, 14);
    if (eventBuffer) {
      eventBuffer[1] = (char) 14;
      RtlCopyMemory(&eventBuffer[11], "???", 3);
    }
  }
  if (eventBuffer) {
    seq = InterlockedIncrement(&sequenceNumber);   // advance sequence number
    RtlCopyMemory(&eventBuffer[2], &seq, 4);
    RtlCopyMemory(&eventBuffer[6], &fileObject, 4);
    if (MmIsAddressValid(requestKernelListen)) {
      eventBuffer[10] =
        (requestKernelListen->RequestFlags & TDI_QUERY_ACCEPT) ? '\1' : '\0';
    }
    else
      eventBuffer[10] = '?';
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcNetPassOnWithTailLogging(HookDevice, Irp, seq);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcNetPassOnNormally(HookDevice, Irp);
  }
}

// VTrcNetTdiDispatchAccept handles an IRP representing a TDI request
// to accept a connection, by tracing any relevant parameters and passing
// on the IRP in a way that ensures that its completion can be logged as
// well.  An additional parameter is provided that points to important
// information about this TDI request passed in the current IRP stack of
// the IRP.

NTSTATUS VTrcNetTdiDispatchAccept
  (PDEVICE_OBJECT HookDevice, PIRP Irp, PFILE_OBJECT fileObject,
   PTDI_REQUEST_KERNEL_ACCEPT requestKernelAccept)
{
  ULONG seq, addressLength;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the event

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  if (MmIsAddressValid(requestKernelAccept) &&
      MmIsAddressValid(requestKernelAccept->RequestConnectionInformation) &&
      MmIsAddressValid(requestKernelAccept->RequestConnectionInformation->
                       RemoteAddress)) {
    addressLength =
      requestKernelAccept->RequestConnectionInformation->RemoteAddressLength;
    if (addressLength > 128) addressLength = 128;
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_ACCEPT,
                            addressLength + 10);
    if (eventBuffer) {
      eventBuffer[1] = (char) (addressLength + 10);
      RtlCopyMemory(&eventBuffer[10],
                    requestKernelAccept->RequestConnectionInformation->
                    RemoteAddress, addressLength);
    }
  }
  else {
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_ACCEPT, 13);
    if (eventBuffer) {
      eventBuffer[1] = (char) 13;
      RtlCopyMemory(&eventBuffer[10], "???", 3);
    }
  }
  if (eventBuffer) {
    seq = InterlockedIncrement(&sequenceNumber);   // advance sequence number
    RtlCopyMemory(&eventBuffer[2], &seq, 4);
    RtlCopyMemory(&eventBuffer[6], &fileObject, 4);
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcNetPassOnWithTailLogging(HookDevice, Irp, seq);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    return VTrcNetPassOnNormally(HookDevice, Irp);
  }
}

// VTrcNetTdiDispatchSetHandler handles an IRP representing a TDI request
// to set an event handler, by substituting our own special event handler
// to log those calls, then passing on the IRP.  An additional parameter
// is provided that points to important information about this TDI request
// passed in the current IRP stack of the IRP.

NTSTATUS VTrcNetTdiDispatchSetHandler
  (PDEVICE_OBJECT HookDevice, PIRP Irp, PFILE_OBJECT fileObject,
   PTDI_REQUEST_KERNEL_SET_EVENT requestKernelSetEvent)
{
  PNET_HOOK_EXTENSION hookExt;
  VTRCNET_FILE_OBJECT_INFO *info;
  KIRQL oldirql;
  PCHAR eventBuffer;

  if (MmIsAddressValid(requestKernelSetEvent) &&
      requestKernelSetEvent->EventHandler != NULL) {
    hookExt = HookDevice->DeviceExtension;
    info = FindFileObjectInfo(hookExt, fileObject);

    // If we find object info for the file object, we change the event
    // handler and event context request so that the callback routine
    // is our callback routine.  We also set fields in info so that the
    // callback routine will know the original callback routine and its
    // context and be able to call it with that context.

    if (info != NULL) {
      switch (requestKernelSetEvent->EventType) {
      case TDI_EVENT_CONNECT :
        info->connectEventHandler = requestKernelSetEvent->EventHandler;
        info->connectEventContext = requestKernelSetEvent->EventContext;
        requestKernelSetEvent->EventHandler = VTrcNetClientEventConnect;
        requestKernelSetEvent->EventContext = info;
        break;

      case TDI_EVENT_DISCONNECT :
        info->disconnectEventHandler = requestKernelSetEvent->EventHandler;
        info->disconnectEventContext = requestKernelSetEvent->EventContext;
        requestKernelSetEvent->EventHandler = VTrcNetClientEventDisconnect;
        requestKernelSetEvent->EventContext = info;
        break;

      case TDI_EVENT_RECEIVE :
        info->receiveEventHandler = requestKernelSetEvent->EventHandler;
        info->receiveEventContext = requestKernelSetEvent->EventContext;
        requestKernelSetEvent->EventHandler = VTrcNetClientEventReceive;
        requestKernelSetEvent->EventContext = info;
        break;

      case TDI_EVENT_RECEIVE_DATAGRAM :
        info->receiveDatagramEventHandler =
                                       requestKernelSetEvent->EventHandler;
        info->receiveDatagramEventContext =
                                       requestKernelSetEvent->EventContext;
        requestKernelSetEvent->EventHandler =
                                       VTrcNetClientEventReceiveDatagram;
        requestKernelSetEvent->EventContext = info;
        break;

      case TDI_EVENT_RECEIVE_EXPEDITED :
        info->receiveExpeditedEventHandler =
                                       requestKernelSetEvent->EventHandler;
        info->receiveExpeditedEventContext =
                                       requestKernelSetEvent->EventContext;
        requestKernelSetEvent->EventHandler =
                                       VTrcNetClientEventReceiveExpedited;
        requestKernelSetEvent->EventContext = info;
        break;

      case TDI_EVENT_CHAINED_RECEIVE :
        info->chainedReceiveEventHandler = requestKernelSetEvent->EventHandler;
        info->chainedReceiveEventContext = requestKernelSetEvent->EventContext;
        requestKernelSetEvent->EventHandler = VTrcNetClientEventChainedReceive;
        requestKernelSetEvent->EventContext = info;
        break;

      case TDI_EVENT_CHAINED_RECEIVE_DATAGRAM :
        info->chainedReceiveDatagramEventHandler =
                                       requestKernelSetEvent->EventHandler;
        info->chainedReceiveDatagramEventContext =
                                       requestKernelSetEvent->EventContext;
        requestKernelSetEvent->EventHandler =
                                    VTrcNetClientEventChainedReceiveDatagram;
        requestKernelSetEvent->EventContext = info;
        break;

      case TDI_EVENT_CHAINED_RECEIVE_EXPEDITED :
        info->chainedReceiveExpeditedEventHandler =
                                       requestKernelSetEvent->EventHandler;
        info->chainedReceiveExpeditedEventContext =
                                       requestKernelSetEvent->EventContext;
        requestKernelSetEvent->EventHandler =
                                    VTrcNetClientEventChainedReceiveExpedited;
        requestKernelSetEvent->EventContext = info;
        break;
      }
    }
  }

  return VTrcNetPassOnNormally(HookDevice, Irp);
}
