// 
// VTrcNet.c
//
// Code for the VTrcNet driver, a filter driver for network transport layers.
// It logs its findings by calling the LogEvent routine in the VTrcLog
// driver.
//
// (Uses some code by Mark Russinovich and Bryce Cogswell)
//

#include "NTDDK.h"
#include "TdiKrnl.h"
#include "../Common/W2Kddk.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include "GetState.h"
#include "ObjInfo.h"
#include "TdiDisp.h"

////////////////////////////////////////////////////////////////////////////
//                                MACROS
////////////////////////////////////////////////////////////////////////////

// print macro that only turns on when checked builds are on
#if DBG
#define DebugPrint(arg) DbgPrint arg
#else
#define DebugPrint(arg)
#endif

////////////////////////////////////////////////////////////////////////////
//                          FORWARD DEFINITIONS
////////////////////////////////////////////////////////////////////////////

NTSTATUS DriverEntry (IN PDRIVER_OBJECT DriverObject,
                      IN PUNICODE_STRING RegistryPath);

NTSTATUS VTrcNetPassOnNormally (PDEVICE_OBJECT HookDevice, IN PIRP Irp);

NTSTATUS VTrcNetDispatchCreate (PDEVICE_OBJECT HookDevice, IN PIRP Irp);
NTSTATUS VTrcNetDispatchClose (PDEVICE_OBJECT HookDevice, IN PIRP Irp);
NTSTATUS VTrcNetDispatchDeviceControl (PDEVICE_OBJECT HookDevice, IN PIRP Irp);
NTSTATUS VTrcNetDispatchPnp (PDEVICE_OBJECT HookDevice, IN PIRP Irp);
NTSTATUS VTrcNetDispatchPower (PDEVICE_OBJECT HookDevice, IN PIRP Irp);
VOID VTrcNetUnload (PDRIVER_OBJECT DriverObject);

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, DriverEntry)
#pragma alloc_text (PAGE, VTrcNetDispatchPnp)
#pragma alloc_text (PAGE, VTrcNetUnload)
#endif

////////////////////////////////////////////////////////////////////////////
//                               GLOBALS
////////////////////////////////////////////////////////////////////////////

// Each IRP is given a sequence number. This allows the return status
// of an IRP, which is obtained in the completion routine, to be 
// associated with the IRPs parameters that were extracted in the Dispatch
// routine.

LONG sequenceNumber = 0;

////////////////////////////////////////////////////////////////////////////
//                            DRIVER ENTRY
////////////////////////////////////////////////////////////////////////////

// DriverEntry is the installable driver initialization.  Here we just
// set up the driver object and initiate communication with the VTrcLog
// driver.  The VTrcLog driver takes care of creating all our devices.

NTSTATUS DriverEntry (IN PDRIVER_OBJECT DriverObject,
                      IN PUNICODE_STRING RegistryPath)
{
  NTSTATUS status;
  int i;

  DebugPrint(("VTrcNet: Entering DriverEntry\n"));

  // Read driver-shared state from VTrcLog driver.

  status = GetSharedState();
  if (!NT_SUCCESS(status))
    return status;

  // Initialize file object info-related stuff

  InitializeFileObjectInfoStructures();

  // Create dispatch points for all routines that must be handled. 
  // All entry points are registered since we might filter a
  // network driver that processes all of them.

  for (i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++)
    DriverObject->MajorFunction[i] = VTrcNetPassOnNormally;

  DriverObject->MajorFunction[IRP_MJ_CREATE] = VTrcNetDispatchCreate;
  DriverObject->MajorFunction[IRP_MJ_CLOSE] = VTrcNetDispatchClose;

  DriverObject->MajorFunction[IRP_MJ_INTERNAL_DEVICE_CONTROL] =
  DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] =
                                              VTrcNetDispatchDeviceControl;
  DriverObject->MajorFunction[IRP_MJ_PNP] = VTrcNetDispatchPnp;
  DriverObject->MajorFunction[IRP_MJ_POWER] = VTrcNetDispatchPower;

//  DriverObject->DriverUnload = VTrcNetUnload;

  // Save driver object in shared state structure now that we're ready to
  // receive requests
  
  sharedState->netDriverObject = DriverObject;

  return STATUS_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////
//                       REQUEST COMPLETION ROUTINES
////////////////////////////////////////////////////////////////////////////

// VTrcNetNormalCompletionRoutine is the completion routine set for
// a typical IRP that needs to have its completion logged.

NTSTATUS VTrcNetNormalCompletionRoutine (IN PDEVICE_OBJECT DeviceObject,
                                         IN PIRP Irp, IN PVOID Context)
{
  LONG  seq = (LONG) Context;   // the context is the sequence number
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the return values.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_COMPLETE_OPERATION, 13);
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[1], &seq, 4);
    RtlCopyMemory(&eventBuffer[5], &Irp->IoStatus.Status, 4);
    RtlCopyMemory(&eventBuffer[9], &Irp->IoStatus.Information, 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // As in (almost) any completion routine, we have to mark Irp as pending if
  // necessary.

  if (Irp->PendingReturned) {
    IoMarkIrpPending(Irp);
  }

  return Irp->IoStatus.Status;
}

// VTrcNetOtherIRPCompletionRoutine is the completion routine set for
// an IRP that we created to handle an IRP sent to us.  Completing this
// IRP consists of freeing it, then completing the original IRP.  To
// enable this, the context contains the original IRP, which contains
// the sequence number in its UserEvent field.

NTSTATUS VTrcNetOtherIRPCompletionRoutine (PDEVICE_OBJECT DeviceObject,
                                           PIRP Irp, PVOID Context)
{
  PIRP origIrp = (PIRP) Context;
  ULONG seq;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Extract the sequence number from the original IRP's UserEvent field,
  // then restore that field's proper value from the new IRP's field

  seq = (ULONG) origIrp->UserEvent;
  origIrp->UserEvent = Irp->UserEvent;

  // Log the return values.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_COMPLETE_OPERATION, 13);
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[1], &seq, 4);
    RtlCopyMemory(&eventBuffer[5], &Irp->IoStatus.Status, 4);
    RtlCopyMemory(&eventBuffer[9], &Irp->IoStatus.Information, 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Copy important possibly returned fields in this IRP to the original IRP

  origIrp->IoStatus = Irp->IoStatus;
  origIrp->PendingReturned = Irp->PendingReturned;
  origIrp->MdlAddress = Irp->MdlAddress;

  // If this IRP was pending, mark the new and original IRP's pending

  if (Irp->PendingReturned) {
    IoMarkIrpPending(Irp);
    IoMarkIrpPending(origIrp);
  }

  // Free the IRP we created
 
  IoFreeIrp(Irp);

  // Complete the original IRP

  IoCompleteRequest(origIrp, IO_NO_INCREMENT);

  // Return STATUS_MORE_PROCESSING_REQUIRED to indicate that we're going
  // to take control of our IRP and finish handling it.  Actually, we've
  // already handled it by freeing it.  If we didn't take control of it,
  // the I/O Manager might try to process our IRP further, which would
  // we bad since it's been freed.

  return STATUS_MORE_PROCESSING_REQUIRED;
}

NTSTATUS VTrcNetForwardedIrpCompletion (IN PDEVICE_OBJECT DeviceObject,
                                        IN PIRP Irp, IN PVOID Context)
{
  PKEVENT Event = (PKEVENT) Context;

  UNREFERENCED_PARAMETER(DeviceObject);

  if (Irp->PendingReturned) {
    IoMarkIrpPending(Irp);
  }

  KeSetEvent(Event, IO_NO_INCREMENT, FALSE);

  return STATUS_MORE_PROCESSING_REQUIRED;
}

////////////////////////////////////////////////////////////////////////////
//              FUNCTIONS TO PASS IRP ON TO NEXT LOWER DEVICE
////////////////////////////////////////////////////////////////////////////

// VTrcNetPassOnNormally passes on the given IRP to the next lower device
// (also given).  It makes no effort to log the completion of the IRP.

NTSTATUS VTrcNetPassOnNormally (PDEVICE_OBJECT HookDevice, IN PIRP Irp)
{
  PNET_HOOK_EXTENSION hookExt = HookDevice->DeviceExtension;

  // Back the stack up one

  IoSkipCurrentIrpStackLocation(Irp);

  // Call the next lower driver

  return IoCallDriver(hookExt->attachedDevice, Irp);
}

// VTrcNetPassOnWithNewIrp passes on the given IRP so that its completion
// can be logged.  It does this by creating a whole new IRP and passing _that_
// on.  It sets the completion routine to VTrcNetOtherIRPCompletionRoutine so
// that the new IRP's completion causes the new IRP to get freed and the
// original IRP to get completed correctly.  (Note that if this routine
// fails in some way, such as being unable to allocate memory, its
// failure mode is to just call VTrcNetPassOnNormally, which
// does the same thing but does not log the completion of the IRP and
// thus does not need a new IRP.)

NTSTATUS VTrcNetPassOnWithNewIrp (PDEVICE_OBJECT HookDevice, IN PIRP Irp,
                                  ULONG seq)
{
  PNET_HOOK_EXTENSION hookExt = HookDevice->DeviceExtension;
  PIRP                newIrp;
  PIO_STACK_LOCATION  currentIrpStackLocation;
  PIO_STACK_LOCATION  newIrpFirstStackLocation;

  // Allocate a new IRP with stack size sufficient for the next lower device

  newIrp = IoAllocateIrp(hookExt->attachedDevice->StackSize, FALSE);
  if (newIrp == NULL)
    return VTrcNetPassOnNormally(HookDevice, Irp);

  // Fill in important fields of new IRP with those of original IRP

  newIrp->MdlAddress = Irp->MdlAddress;
  newIrp->Flags = Irp->Flags;
  newIrp->AssociatedIrp = Irp->AssociatedIrp;
  newIrp->RequestorMode = Irp->RequestorMode;
  newIrp->UserIosb = Irp->UserIosb;
  newIrp->UserEvent = Irp->UserEvent;
  newIrp->Overlay = Irp->Overlay;
  newIrp->UserBuffer = Irp->UserBuffer;
  newIrp->Tail.Overlay.AuxiliaryBuffer = Irp->Tail.Overlay.AuxiliaryBuffer;
  newIrp->Tail.Overlay.OriginalFileObject =
                                        Irp->Tail.Overlay.OriginalFileObject;
  newIrp->Tail.Overlay.Thread = Irp->Tail.Overlay.Thread;

  // Copy current stack location to the first stack location in the new IRP

  currentIrpStackLocation = IoGetCurrentIrpStackLocation(Irp);
  newIrpFirstStackLocation = IoGetNextIrpStackLocation(newIrp);
  *newIrpFirstStackLocation = *currentIrpStackLocation;

  // Set the completion routine to VTrcNetOtherIRPCompletionRoutine, with
  // the original IRP as context.  Save the sequence number in the UserEvent
  // field so that it can be used in the completion routine.  (The completion
  // routine will restore the original UserEvent field from the new IRP.)

  Irp->UserEvent = (PKEVENT) seq;
  IoSetCompletionRoutine(newIrp, VTrcNetOtherIRPCompletionRoutine,
                         (PVOID) Irp, TRUE, TRUE, TRUE);

  // Send the new IRP to the lower-level device

  return IoCallDriver(hookExt->attachedDevice, newIrp);
}

// VTrcNetPassOnWithTailLogging passes on the given IRP to the
// next lower device (also given) in a way that the completion of the
// IRP gets logged.

NTSTATUS VTrcNetPassOnWithTailLogging (PDEVICE_OBJECT HookDevice, IN PIRP Irp,
                                       ULONG seq)
{
  PIO_STACK_LOCATION  currentIrpStackLocation, nextIrpStackLocation;
  PNET_HOOK_EXTENSION hookExt = HookDevice->DeviceExtension;

  // If there isn't room in the stack for passing on the IRP, we will
  // have to make a new IRP and pass _that_ on to the next lower driver.
  // Otherwise, we just copy the current stack location to the next
  // stack location and pass on the IRP with a tail-logging completion
  // routine, VTrcNetNormalCompletionRoutine.

  if (Irp->CurrentLocation < HookDevice->StackSize) {
    return VTrcNetPassOnWithNewIrp(HookDevice, Irp, seq);
  }
  else {
    currentIrpStackLocation = IoGetCurrentIrpStackLocation(Irp);
    nextIrpStackLocation = IoGetNextIrpStackLocation(Irp);
    // Copy parameters down to next level in the stack for the driver below us
    *nextIrpStackLocation = *currentIrpStackLocation;
    // But make sure not to copy the completion routine!
    IoSetCompletionRoutine(Irp, VTrcNetNormalCompletionRoutine, (PVOID) seq,
                           TRUE, TRUE, TRUE);
    return IoCallDriver(hookExt->attachedDevice, Irp);
  }
}

NTSTATUS VTrcNetForwardIrpSynchronous (PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
  PNET_HOOK_EXTENSION deviceExtension;
  KEVENT event;
  NTSTATUS status;

  KeInitializeEvent(&event, NotificationEvent, FALSE);
  deviceExtension = (PNET_HOOK_EXTENSION) DeviceObject->DeviceExtension;

  // copy the irp stack for the next device

  IoCopyCurrentIrpStackLocationToNext(Irp);

  // set a completion routine

  IoSetCompletionRoutine(Irp, VTrcNetForwardedIrpCompletion, &event, TRUE,
                         TRUE, TRUE);

  // call the next lower device

  status = IoCallDriver(deviceExtension->attachedDevice, Irp);

  // wait for the actual completion

  if (status == STATUS_PENDING) {
    KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, NULL);
    status = Irp->IoStatus.Status;
  }

  return status;
}

////////////////////////////////////////////////////////////////////////////
//                          DISPATCH ROUTINES
////////////////////////////////////////////////////////////////////////////

// VTrcNetDispatchCreate is the dispatch routine called for any create
// IRP. It logs the parameters of the call and calls a pass-on routine
// to pass on the IRP and log its completion.

NTSTATUS VTrcNetDispatchCreate (PDEVICE_OBJECT HookDevice, IN PIRP Irp)
{
  PIO_STACK_LOCATION currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
  PFILE_OBJECT fileObject = currentIrpStack->FileObject;
  PNET_HOOK_EXTENSION hookExt = HookDevice->DeviceExtension;
  ULONG seq = InterlockedIncrement(&sequenceNumber);
  KIRQL oldirql;
  PFILE_FULL_EA_INFORMATION eaInfo;
  ULONG nameLength, addressLength, truncatedAddressLength, connectionContext;
  char objectType;
  PCHAR eventBuffer;
  ULONG threadId;

  // Increment the number of create IRP's filtered

  sharedState->filteredCreates++;

  // The information about this creation are in the system buffer of the
  // IRP, which is a FILE_FULL_EA_INFORMATION structure.

  eaInfo = Irp->AssociatedIrp.SystemBuffer;

  // First, figure out what kind of network object is being opened from eaInfo.
  // If eaInfo is NULL, it's a control channel.  Otherwise, look at the EaName
  // field to see if it matches TdiTransportAddress or TdiConnectionContext.
  // While we're at it, we get the address length if it's a transport address
  // and the connection context number if it's a connection endpoint.

  objectType = UNKNOWN_OBJ_TYPE;
  if (eaInfo == NULL)
    objectType = CONTROL_CHANNEL_OBJ_TYPE;
  else if (MmIsAddressValid(eaInfo)) {
    nameLength = eaInfo->EaNameLength;
    if (MmIsAddressValid(eaInfo->EaName)) {
      if (nameLength == 16 &&
          !strncmp(eaInfo->EaName, TdiTransportAddress, 16)) {
        objectType = TRANSPORT_ADDRESS_OBJ_TYPE;
        addressLength = eaInfo->EaValueLength;
        if (addressLength > 128) addressLength = 128;
      }
      else if (nameLength == 17 &&
               !strncmp(eaInfo->EaName, TdiConnectionContext, 17)) {
        objectType = CONNECTION_ENDPOINT_OBJ_TYPE;
        if (eaInfo->EaValueLength == 4)
          connectionContext = * (ULONG *) &eaInfo->EaName[nameLength+1];
        else
          connectionContext = 0;
      }
    }
  }

  // Indicate that the current thread is active.

  threadId = (ULONG) PsGetCurrentThreadId();
  if (threadId < sharedState->kernelUserSharedState.maxThreads)
    sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] = '\0';

  // Put the information relevant to this type of object into the log entry
  // buffer and log it

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  // This is a variable-length entry.  Put the length in
  // eventBuffer[1].  Put different information at
  // eventBuffer[15...] depending on the object type:
  // the connection context for a connection endpoint, the address for
  // a transport address, and nothing for anything else.

  if (objectType == CONNECTION_ENDPOINT_OBJ_TYPE) {
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_OPEN, 19);
    if (eventBuffer) {
      eventBuffer[1] = (char) 19;
      RtlCopyMemory(&eventBuffer[15], &connectionContext, 4);
    }
  }
  else if (objectType == TRANSPORT_ADDRESS_OBJ_TYPE) {
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_OPEN, addressLength + 15);
    if (eventBuffer) {
      eventBuffer[1] = (char) (addressLength + 15);
      RtlCopyMemory(&eventBuffer[15], &eaInfo->EaName[nameLength+1],
                    addressLength);
    }
  }
  else {
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_OPEN, 15);
    if (eventBuffer)
      eventBuffer[1] = (char) 15;
  }
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[2], &seq, 4);
    RtlCopyMemory(&eventBuffer[6], &fileObject, 4);
    RtlCopyMemory(&eventBuffer[10], &hookExt->attachedDevice, 4);
    eventBuffer[14] = objectType;
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    // Pass on the IRP, tracing its completion
    return VTrcNetPassOnWithTailLogging(HookDevice, Irp, seq);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    // Pass on the IRP, ignoring its completion
    return VTrcNetPassOnNormally(HookDevice, Irp);
  }
}

// VTrcNetDispatchClose is the dispatch routine called for any close
// IRP. It logs the parameters of the call and calls a pass-on routine
// to pass on the IRP and log its completion.  It also frees any
// VTRCNET_FILE_OBJECT_INFO structures allocated for this file object.

NTSTATUS VTrcNetDispatchClose (PDEVICE_OBJECT HookDevice, IN PIRP Irp)
{
  PIO_STACK_LOCATION currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
  PFILE_OBJECT fileObject = currentIrpStack->FileObject;
  PNET_HOOK_EXTENSION hookExt = HookDevice->DeviceExtension;
  ULONG seq = InterlockedIncrement(&sequenceNumber);
  KIRQL oldirql;
  PCHAR eventBuffer;
  ULONG threadId;

  // Indicate that the current thread is active.

  threadId = (ULONG) PsGetCurrentThreadId();
  if (threadId < sharedState->kernelUserSharedState.maxThreads)
    sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] = '\0';

  // Log the parameters of this IRP

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_NETOBJ_CLOSE, 9);
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[1], &seq, 4);
    RtlCopyMemory(&eventBuffer[5], &fileObject, 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // Free any file object info structure allocated for this file object

  FreeInfoOfFileObject(hookExt, fileObject);

  // Pass on the IRP so that the completion gets logged

  return VTrcNetPassOnWithTailLogging(HookDevice, Irp, seq);
}

// VTrcNetDispatchDeviceControl is the dispatch routine called for device
// control (and internal device control) IRP's.  This function may log the
// parameters and completion of the IRP if appropriate.  Also, if this IRP
// is setting up a callback routine, we substitute our own callback routine
// so that we can log calls to those routines.  (For example, a callback
// routine that is called whenever data is received.)

NTSTATUS VTrcNetDispatchDeviceControl (PDEVICE_OBJECT HookDevice, IN PIRP Irp)
{
  PIO_STACK_LOCATION currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
  PFILE_OBJECT fileObject = currentIrpStack->FileObject;
  PVOID parameters = &currentIrpStack->Parameters.Others;
  PNET_HOOK_EXTENSION hookExt = HookDevice->DeviceExtension;
  NTSTATUS status;
  ULONG threadId;

  // If this is a device control IRP, it might be able to be converted to an
  // internal device control IRP by calling TdiMapUserRequest (this is what
  // the next lower-level device would do anyway).  If so, we convert it here.

  if (currentIrpStack->MajorFunction == IRP_MJ_DEVICE_CONTROL) {
    status = TdiMapUserRequest(hookExt->attachedDevice, Irp, currentIrpStack);
    if (status != STATUS_SUCCESS)
      return VTrcNetPassOnNormally(HookDevice, Irp);
  }

  // At this point, the major function should have been converted to
  // IRP_MJ_INTERNAL_DEVICE_CONTROL, but we do a sanity check just to
  // make sure.

  if (currentIrpStack->MajorFunction != IRP_MJ_INTERNAL_DEVICE_CONTROL)
    return VTrcNetPassOnNormally(HookDevice, Irp);

  // Indicate that the current thread is active.

  threadId = (ULONG) PsGetCurrentThreadId();
  if (threadId < sharedState->kernelUserSharedState.maxThreads)
    sharedState->kernelUserSharedState.msgChecksSinceActivity[threadId] = '\0';

  // Depending on the minor function, it could be any type of TDI request.
  // The parameters for the TDI request will be in the parameters element
  // of the current IRP stack location, in a format depending on what kind
  // of request it is.

  switch (currentIrpStack->MinorFunction) {
  case TDI_CONNECT :
    return VTrcNetTdiDispatchConnect(HookDevice, Irp, fileObject, parameters);
  case TDI_DISCONNECT :
    return VTrcNetTdiDispatchDisconnect(HookDevice, Irp, fileObject,
                                        parameters);
  case TDI_ASSOCIATE_ADDRESS :
    return VTrcNetTdiDispatchAssociate(HookDevice, Irp, fileObject,
                                       parameters);
  case TDI_DISASSOCIATE_ADDRESS :
    return VTrcNetTdiDispatchDisassociate(HookDevice, Irp, fileObject,
                                          parameters);
  case TDI_SEND :
    return VTrcNetTdiDispatchSend(HookDevice, Irp, fileObject, parameters);
  case TDI_SEND_DATAGRAM :
    return VTrcNetTdiDispatchSendDatagram(HookDevice, Irp, fileObject,
                                          parameters);
  case TDI_RECEIVE :
    return VTrcNetTdiDispatchReceive(HookDevice, Irp, fileObject, parameters);
  case TDI_RECEIVE_DATAGRAM :
    return VTrcNetTdiDispatchReceiveDatagram(HookDevice, Irp, fileObject,
                                             parameters);
  case TDI_LISTEN :
    return VTrcNetTdiDispatchListen(HookDevice, Irp, fileObject, parameters);
  case TDI_ACCEPT :
    return VTrcNetTdiDispatchAccept(HookDevice, Irp, fileObject, parameters);
  case TDI_SET_EVENT_HANDLER :
    return VTrcNetTdiDispatchSetHandler(HookDevice, Irp, fileObject,
                                        parameters);
  default :
    return VTrcNetPassOnNormally(HookDevice, Irp);
  }
}

NTSTATUS VTrcNetDispatchPnp (PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
  PIO_STACK_LOCATION   irpSp = IoGetCurrentIrpStackLocation(Irp);
  PNET_HOOK_EXTENSION  deviceExtension = DeviceObject->DeviceExtension;
  NTSTATUS             status;

  PAGED_CODE();

  switch(irpSp->MinorFunction) {
  case IRP_MN_REMOVE_DEVICE:
    status = VTrcNetForwardIrpSynchronous(DeviceObject, Irp);
    IoDetachDevice(deviceExtension->attachedDevice);
    IoDeleteDevice(DeviceObject);

    // Complete the Irp

    Irp->IoStatus.Status = status;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    break;

  default:
    // Simply forward all other Irps
    status = VTrcNetPassOnNormally(DeviceObject, Irp);
    break;
  } // end switch (irpSp->MinorFunction)

  return status;
}

NTSTATUS VTrcNetDispatchPower (PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
  PFS_HOOK_EXTENSION deviceExtension;

  PoStartNextPowerIrp(Irp);
  IoSkipCurrentIrpStackLocation(Irp);

  deviceExtension = (PFS_HOOK_EXTENSION) DeviceObject->DeviceExtension;
  return PoCallDriver(deviceExtension->attachedDevice, Irp);
}

VOID VTrcNetUnload (PDRIVER_OBJECT DriverObject)
{
  PAGED_CODE();

  // I should free the file object info structures here, but I don't

  return;
}
