//
// VTrcRawD.c
//
// Code for the VTrcRawD driver, a filter driver for raw disks.  It logs its
// findings by calling the LogEvent routine in the VTrcLog driver.
//
// (Uses a lot of code by Microsoft Corporation, by Mike Glass and Bob
// Rinne, obtained from the diskperf code included with the DDK.)
//

#include "ntddk.h"
#include "ntdddisk.h"
#include "stdarg.h"
#include "stdio.h"
#include "../Common/VTrace.h"
#include "../Common/VTrcDrv.h"
#include <ntddvol.h>
#include <mountdev.h>
#include "GetState.h"

////////////////////////////////////////////////////////////////////////////
//                                GLOBALS
////////////////////////////////////////////////////////////////////////////

// Each IRP is given a sequence number. This allows the return status
// of an IRP, which is obtained in the completion routine, to be
// associated with the IRP's parameters that were extracted in the dispatch
// routine.

LONG sequenceNumber = 0;

////////////////////////////////////////////////////////////////////////////
//                                 MACROS
////////////////////////////////////////////////////////////////////////////

#define FILTER_DEVICE_PROPAGATE_FLAGS            DO_POWER_PAGABLE
#define FILTER_DEVICE_PROPAGATE_CHARACTERISTICS (FILE_REMOVABLE_MEDIA |  \
                                                 FILE_READ_ONLY_DEVICE | \
                                                 FILE_FLOPPY_DISKETTE)

////////////////////////////////////////////////////////////////////////////
//                              STRUCTURES
////////////////////////////////////////////////////////////////////////////

// The structure RAW_HOOK_EXTENSION forms the device extension of
// any raw device associated with the VTrcRawD driver.

typedef struct _raw_hook_extension {
  PDEVICE_OBJECT TargetDeviceObject;
  ULONG DiskNumber;
  ULONG PartitionNumber;
  ULONG LastPartitionFiltered;
  // must synchronize paging path notifications
  KEVENT PagingPathCountEvent;
  ULONG PagingPathCount;
} RAW_HOOK_EXTENSION, *PRAW_HOOK_EXTENSION;

////////////////////////////////////////////////////////////////////////////
//                          FUNCTION DECLARATIONS
////////////////////////////////////////////////////////////////////////////

NTSTATUS
DriverEntry(
  IN PDRIVER_OBJECT DriverObject,
  IN PUNICODE_STRING RegistryPath
  );

NTSTATUS
VTrcRawDAddDevice(
  IN PDRIVER_OBJECT driverObject,
  IN PDEVICE_OBJECT physicalDeviceObject
  );

VOID
VTrcRawDSyncFilterWithTarget(
  IN PDEVICE_OBJECT FilterDevice,
  IN PDEVICE_OBJECT TargetDevice
  );

NTSTATUS
VTrcRawDDispatchPnp(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp
  );

NTSTATUS
VTrcRawDDispatchPnpDeviceUsageNotification(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp
  );

NTSTATUS
VTrcRawDDispatchPower(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp
  );

NTSTATUS
VTrcRawDCreate(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp
  );

NTSTATUS
VTrcRawDReadWrite(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp
  );

NTSTATUS
VTrcRawDDispatchControl(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp
  );

NTSTATUS
VTrcRawDDispatchGeneral(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp
  );

VOID
VTrcRawDUnload(
  IN PDRIVER_OBJECT DriverObject
  );

NTSTATUS
VTrcRawDReadWriteCompletion(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP           Irp,
  IN PVOID          Context
  );

NTSTATUS
VTrcRawDSynchronousIrpCompletion(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp,
  IN PVOID Context
  );

NTSTATUS
VTrcRawDPnpCompletionRoutine(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp,
  IN PVOID Context
  );

NTSTATUS
VTrcRawDForwardIrpSynchronous(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp
  );

void GetDiskAndPartitionNumbers(
  PDEVICE_OBJECT filterDevice
  );

////////////////////////////////////////////////////////////////////////////
//                          DEBUGGER DECLARATIONS
////////////////////////////////////////////////////////////////////////////

#if DBG

#define DEBUG_BUFFER_LENGTH 256

ULONG VTrcRawDDebug = 3;
UCHAR VTrcRawDDebugBuffer[DEBUG_BUFFER_LENGTH];

// VTrcRawDDebugPrint prints a debugging message DebugMessage if the print
// level given in DebugPrintLevel is >= the current debug level, VTrcRawDDebug.

VOID
VTrcRawDDebugPrint(
  ULONG DebugPrintLevel,
  PCCHAR DebugMessage,
  ...
  )
{
  va_list ap;
  va_start(ap, DebugMessage);

  if (DebugPrintLevel <= VTrcRawDDebug) {
    _vsnprintf(VTrcRawDDebugBuffer, DEBUG_BUFFER_LENGTH, DebugMessage, ap);
    DbgPrint(VTrcRawDDebugBuffer);
  }

  va_end(ap);
}

#define DebugPrint(x)   VTrcRawDDebugPrint x

#else

#define DebugPrint(x)

#endif

////////////////////////////////////////////////////////////////////////////
//                                SECTIONS
////////////////////////////////////////////////////////////////////////////

// Define the sections that allow for discarding (i.e. paging) some of
// the code.

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, DriverEntry)
#pragma alloc_text (PAGE, VTrcRawDAddDevice)
#pragma alloc_text (PAGE, VTrcRawDSyncFilterWithTarget)
#pragma alloc_text (PAGE, VTrcRawDDispatchPnp)
#pragma alloc_text (PAGE, VTrcRawDDispatchPnpDeviceUsageNotification)
#pragma alloc_text (PAGE, VTrcRawDUnload)
#endif

////////////////////////////////////////////////////////////////////////////
//                        INITIALIZATION ROUTINES
////////////////////////////////////////////////////////////////////////////

// DriverEntry is the routine called by the system to initialize the disk
// filter driver.

NTSTATUS
DriverEntry(
  IN PDRIVER_OBJECT DriverObject,
  IN PUNICODE_STRING RegistryPath
  )
{
  PRAW_HOOK_EXTENSION deviceExtension;
  NTSTATUS status;
  ULONG ulIndex;

  DebugPrint((2, "VTrcRawD: Entering DriverEntry.\n"));

  // Get a pointer to the shared state from VTrcInit. If we can't, we
  // fail. Since returning an error would be catastrophic when we're
  // one of the UpperFilters of the storage volumes device class, we
  // instead just set an unload routine and return success.

  status = GetSharedState();
  if (!NT_SUCCESS(status)) {
    DebugPrint((2, "VTrcRawD: Could not get shared state.\n"));
    DriverObject->DriverUnload = VTrcRawDUnload;
    return STATUS_SUCCESS;
  }

  DebugPrint((2, "VTrcRawD: Got shared state OK.\n"));

  // Create dispatch points

  for (ulIndex = 0; ulIndex <= IRP_MJ_MAXIMUM_FUNCTION; ulIndex++)
    DriverObject->MajorFunction[ulIndex] = VTrcRawDDispatchGeneral;

  // Set up the device driver entry points.

  DriverObject->MajorFunction[IRP_MJ_CREATE]         = VTrcRawDCreate;
  DriverObject->MajorFunction[IRP_MJ_READ]           = VTrcRawDReadWrite;
  DriverObject->MajorFunction[IRP_MJ_WRITE]          = VTrcRawDReadWrite;
  DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = VTrcRawDDispatchControl;
  DriverObject->MajorFunction[IRP_MJ_PNP]            = VTrcRawDDispatchPnp;
  DriverObject->MajorFunction[IRP_MJ_POWER]          = VTrcRawDDispatchPower;

  DriverObject->DriverExtension->AddDevice           = VTrcRawDAddDevice;
  DriverObject->DriverUnload                         = VTrcRawDUnload;

  return STATUS_SUCCESS;
} // end DriverEntry()

////////////////////////////////////////////////////////////////////////////
//                        FILTER ESTABLISHMENT
////////////////////////////////////////////////////////////////////////////

NTSTATUS VTrcRawDAddDevice (IN PDRIVER_OBJECT driverObject,
                            IN PDEVICE_OBJECT physicalDeviceObject)
{
  NTSTATUS             status;
  PDEVICE_OBJECT       filterDeviceObject;
  PRAW_HOOK_EXTENSION  deviceExtension;

  PAGED_CODE();

  DebugPrint((1, "VTrcRawD: Filtering device %x\n", physicalDeviceObject));

  status = IoCreateDevice(driverObject, sizeof(RAW_HOOK_EXTENSION),
                          NULL, FILE_DEVICE_DISK, 0, FALSE,
                          &filterDeviceObject);
  if (!NT_SUCCESS(status)) {
    DebugPrint((1, "VTrcRawDAddDevice: Cannot create filter device\n"));
    return status;
  }

  // Set the DO_DIRECT_IO flag in the filter device
  
  filterDeviceObject->Flags |= DO_DIRECT_IO;

  // Initialize the device extension of the filter device

  deviceExtension = (PRAW_HOOK_EXTENSION) filterDeviceObject->DeviceExtension;
  RtlZeroMemory(deviceExtension, sizeof(RAW_HOOK_EXTENSION));
  deviceExtension->DiskNumber = UNKNOWN_DEVICE_NUMBER;
  deviceExtension->PartitionNumber = UNKNOWN_DEVICE_NUMBER;

  // Attach the device object to the highest device object in the chain and
  // return the previously highest device object, which is passed to
  // IoCallDriver when pass IRPs down the device stack

  DebugPrint((1, "VTrcRawD: Attaching filter dev %X to physical dev %X.\n",
              filterDeviceObject, physicalDeviceObject));
  deviceExtension->TargetDeviceObject =
    IoAttachDeviceToDeviceStack(filterDeviceObject, physicalDeviceObject);

  // If the attachment failed, delete the filter device and return failure.

  if (deviceExtension->TargetDeviceObject == NULL) {
    IoDeleteDevice(filterDeviceObject);
    DebugPrint((1, "VTrcRawD: Unable to attach %x to target %x\n",
                filterDeviceObject, physicalDeviceObject));
    return STATUS_NO_SUCH_DEVICE;
  }

  // Initialize the event for handling paging path counting

  KeInitializeEvent(&deviceExtension->PagingPathCountEvent,
                    NotificationEvent, TRUE);

  // Synchronize the fields of the filter device with the filtered device

  DebugPrint((3, "VTrcRawD: Synchronizing filter dev with physical dev\n"));
  VTrcRawDSyncFilterWithTarget(filterDeviceObject,
                               deviceExtension->TargetDeviceObject);

  // Clear the DO_DEVICE_INITIALIZING flag

  filterDeviceObject->Flags &= ~DO_DEVICE_INITIALIZING;

  return STATUS_SUCCESS;
}

// VTrcRawDSyncFilterWithTarget makes sure that the given filter device
// has similar flags/characteristics/alignment as the device it is attached
// to.

VOID
VTrcRawDSyncFilterWithTarget(
  IN PDEVICE_OBJECT FilterDevice,
  IN PDEVICE_OBJECT TargetDevice
  )
{
  ULONG propFlags;
  PRAW_HOOK_EXTENSION deviceExtension;

  PAGED_CODE();

  // Propagate all useful flags from target to VTrcRawD. MountMgr will look
  // at the VTrcRawD object capabilities to figure out if the disk is
  // a removable and perhaps other things.

  propFlags = TargetDevice->Flags & FILTER_DEVICE_PROPAGATE_FLAGS;
  FilterDevice->Flags |= propFlags;
  if (!(TargetDevice->Flags & DO_POWER_PAGABLE)) {
    deviceExtension = FilterDevice->DeviceExtension;
    deviceExtension->PagingPathCount = 1;
  }

  propFlags = TargetDevice->Characteristics &
              FILTER_DEVICE_PROPAGATE_CHARACTERISTICS;
  FilterDevice->Characteristics |= propFlags;

  // Propagate driver's alignment requirements

  FilterDevice->AlignmentRequirement = TargetDevice->AlignmentRequirement;
}

////////////////////////////////////////////////////////////////////////////
//                           DISPATCH ROUTINES
////////////////////////////////////////////////////////////////////////////

// VTrcRawDDispatchPnp dispatches PnP IRP's for VTrcRawD.

NTSTATUS
VTrcRawDDispatchPnp(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp
  )
{
  PRAW_HOOK_EXTENSION  deviceExtension = DeviceObject->DeviceExtension;
  PIO_STACK_LOCATION   irpSp = IoGetCurrentIrpStackLocation(Irp);
  NTSTATUS             status;

  PAGED_CODE();

  DebugPrint((2, "VTrcRawDDispatchPnp: dev %X irp %X func %d\n",
             DeviceObject, Irp, irpSp->MinorFunction));

  switch(irpSp->MinorFunction) {
  case IRP_MN_START_DEVICE:
    status = VTrcRawDForwardIrpSynchronous(DeviceObject, Irp);
    VTrcRawDSyncFilterWithTarget(DeviceObject,
                                 deviceExtension->TargetDeviceObject);
    if (NT_SUCCESS(status))
      GetDiskAndPartitionNumbers(DeviceObject);
    Irp->IoStatus.Status = status;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return status;

  case IRP_MN_DEVICE_USAGE_NOTIFICATION:
    return VTrcRawDDispatchPnpDeviceUsageNotification(DeviceObject, Irp);

  case IRP_MN_REMOVE_DEVICE:
    // Call the Remove Routine handler to schedule a completion routine

    DebugPrint((2, "VTrcRawD: Handling IRP_MN_REMOVE_DEVICE\n"));
    status = VTrcRawDForwardIrpSynchronous(DeviceObject, Irp);
    IoDetachDevice(deviceExtension->TargetDeviceObject);
    IoDeleteDevice(DeviceObject);
    Irp->IoStatus.Status = status;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return status;

  default:
    // Check for disk and partition numbers if we don't know them already.
    // This check is not really necessary here, because the other calls
    // to GetDiskAndPartitionNumbers should get them as soon as they're
    // available, but we do it here just in case the other calls aren't
    // successful when we think they'll be.

    GetDiskAndPartitionNumbers(DeviceObject);
    IoCopyCurrentIrpStackLocationToNext(Irp);
    IoSetCompletionRoutine(Irp, VTrcRawDPnpCompletionRoutine, NULL, TRUE,
                           TRUE, TRUE);
    return IoCallDriver(deviceExtension->TargetDeviceObject, Irp);
  } // end switch (irpSp->MinorFunction)
} // end VTrcRawDDispatchPnp()


// VTrcRawDDispatchPnpDeviceUsageNotification dispatches PnP IRP's with
// a minor function code of IRP_MN_DEVICE_USAGE_NOTIFICATION for VTrcRawD.

NTSTATUS
VTrcRawDDispatchPnpDeviceUsageNotification(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp
  )
{
  PRAW_HOOK_EXTENSION  deviceExtension = DeviceObject->DeviceExtension;
  PIO_STACK_LOCATION   irpStack = IoGetCurrentIrpStackLocation(Irp);
  ULONG                count;
  BOOLEAN              setPagable;
  NTSTATUS             status;

  DebugPrint((3, "VTrcRawD: Handling PnP device usage notification req\n"));

  if (irpStack->Parameters.UsageNotification.Type != DeviceUsageTypePaging) {
    IoSkipCurrentIrpStackLocation(Irp);
    return IoCallDriver(deviceExtension->TargetDeviceObject, Irp);
  }

  // wait on the paging path event

  status = KeWaitForSingleObject(&deviceExtension->PagingPathCountEvent,
                                 Executive, KernelMode, FALSE, NULL);

  // If removing last paging device, need to set DO_POWER_PAGABLE
  // bit here, and possible re-set it below on failure.

  DebugPrint((3, "*** BEFORE PAGING: DeviceObject=%X, InPath=%s, "
              "PagingPathCount=%d, ThisPagable=%s LowerPagable=%s ***\n",
              DeviceObject,
              irpStack->Parameters.UsageNotification.InPath ? "yes" : "no",
              deviceExtension->PagingPathCount,
              (DeviceObject->Flags & DO_POWER_PAGABLE) ? "yes" : "no",
              (deviceExtension->TargetDeviceObject->Flags &
               DO_POWER_PAGABLE) ? "yes" : "no"));

  setPagable = FALSE;
  if (!irpStack->Parameters.UsageNotification.InPath &&
      deviceExtension->PagingPathCount == 1) {
    // removing the last paging file so must have DO_POWER_PAGABLE bits set

    if (DeviceObject->Flags & DO_POWER_INRUSH) {
      DebugPrint((3, "VTrcRawDDispatchPnp: last paging file "
                 "removed but DO_POWER_INRUSH set, so not "
                 "setting PAGABLE bit for DO %p\n", DeviceObject));
    }
    else {
      DebugPrint((2, "VTrcRawDDispatchPnp: Setting PAGABLE "
                 "bit for DO %p\n", DeviceObject));
      DeviceObject->Flags |= DO_POWER_PAGABLE;
      setPagable = TRUE;
    }
  }

  // Send the IRP synchronously.

  status = VTrcRawDForwardIrpSynchronous(DeviceObject, Irp);

  // Now deal with the failure and success cases. Note that we are not
  // allowed to fail the IRP once it is sent to the lower drivers.

  if (NT_SUCCESS(status)) {
    IoAdjustPagingPathCount(&deviceExtension->PagingPathCount,
                            irpStack->Parameters.UsageNotification.InPath);
    if (irpStack->Parameters.UsageNotification.InPath &&
        deviceExtension->PagingPathCount == 1) {
      // first paging file addition
      DebugPrint((3, "VTrcRawDDispatchPnp: Clearing PAGABLE bit "
                     "for DO %p\n", DeviceObject));
      DeviceObject->Flags &= ~DO_POWER_PAGABLE;
    }
  }
  else {
    // cleanup the changes done above
    if (setPagable == TRUE) {
      DeviceObject->Flags &= ~DO_POWER_PAGABLE;
      setPagable = FALSE;
    }
  }

  DebugPrint((3, "*** AFTER PAGING: DeviceObject=%X, InPath=%s, "
              "PagingPathCount=%d, ThisPagable=%s LowerPagable=%s "
              "setPagable = %s ***\n",
              DeviceObject,
              irpStack->Parameters.UsageNotification.InPath ? "yes" : "no",
              deviceExtension->PagingPathCount,
              (DeviceObject->Flags & DO_POWER_PAGABLE) ? "yes" : "no",
              (deviceExtension->TargetDeviceObject->Flags &
               DO_POWER_PAGABLE) ? "yes" : "no",
              setPagable ? "yes" : "no"));

  // I don't entirely trust the above code to keep the DO_POWER_PAGABLE
  // flag of the filter device consistent with that of the filtered device,
  // so this code ensures is here to ensure it.

  if (deviceExtension->TargetDeviceObject->Flags & DO_POWER_PAGABLE)
    DeviceObject->Flags |= DO_POWER_PAGABLE;
  else
    DeviceObject->Flags &= ~DO_POWER_PAGABLE;

  // Set the event so the next one can occur.

  KeSetEvent(&deviceExtension->PagingPathCountEvent, IO_NO_INCREMENT, FALSE);

  // ...and complete the irp

  DebugPrint((3, "VTrcRawD: Completing PnP device usage notification req\n"));
  IoCompleteRequest(Irp, IO_NO_INCREMENT);
  return status;
}

// VTrcRawDDispatchPower dispatches power IRP's for VTrcRawD.

NTSTATUS
VTrcRawDDispatchPower(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp
  )
{
  PRAW_HOOK_EXTENSION deviceExtension;

  DebugPrint((1, "VTrcRawD: Handling power IRP %x, dev %x, func %d\n", Irp,
              DeviceObject, IoGetCurrentIrpStackLocation(Irp)->MinorFunction));

  PoStartNextPowerIrp(Irp);
  IoSkipCurrentIrpStackLocation(Irp);

  deviceExtension = (PRAW_HOOK_EXTENSION) DeviceObject->DeviceExtension;
  return PoCallDriver(deviceExtension->TargetDeviceObject, Irp);
} // end VTrcRawDDispatchPower

// VTrcRawDCreate services open IRP's. It establishes the driver's existence
// by returning STATUS_SUCCESS.

NTSTATUS
VTrcRawDCreate(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp
  )
{
  UNREFERENCED_PARAMETER(DeviceObject);

  Irp->IoStatus.Status = STATUS_SUCCESS;

  IoCompleteRequest(Irp, IO_NO_INCREMENT);
  return STATUS_SUCCESS;
}

// VTrcRawDDispatchReadWrite is the driver entry point for read and write
// requests to disks to which the VTrcRawD driver has attached.  This driver
// does logging and then sets a completion routine so that it can log
// additional information when the request completes. Then it calls the
// next driver below it.

NTSTATUS
VTrcRawDReadWrite(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp
  )
{
  PRAW_HOOK_EXTENSION deviceExtension = DeviceObject->DeviceExtension;
  PIO_STACK_LOCATION currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
  ULONG flags = Irp->Flags;
  KIRQL oldirql;
  ULONG seq;
  PCHAR eventBuffer;

  // Log the read/write

  // Log the input vaules
  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  if (currentIrpStack->MajorFunction == IRP_MJ_READ)
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_RAWDISK_READ, 24);
  else
    eventBuffer = LOG_EVENT((char) ENTRY_TYPE_RAWDISK_WRITE, 24);
  if (eventBuffer) {
    seq = InterlockedIncrement(&sequenceNumber);
    RtlCopyMemory(&eventBuffer[1], &seq, 4);
    if ((flags & 8) && (Irp->AssociatedIrp.MasterIrp != NULL))
      RtlCopyMemory(&eventBuffer[5],
                    &Irp->AssociatedIrp.MasterIrp->
                    Tail.Overlay.OriginalFileObject, 4);
    else
      RtlCopyMemory(&eventBuffer[5], &Irp->Tail.Overlay.OriginalFileObject, 4);
    RtlCopyMemory(&eventBuffer[9],
                  &currentIrpStack->Parameters.Read.ByteOffset, 5);
    RtlCopyMemory(&eventBuffer[14],
                  &currentIrpStack->Parameters.Read.Length, 4);
    RtlCopyMemory(&eventBuffer[18], &flags, 4);
    eventBuffer[22] = (char) deviceExtension->DiskNumber;
    eventBuffer[23] = (char) deviceExtension->PartitionNumber;
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

    // Set completion routine so that we can log the completion of the IRP

    IoCopyCurrentIrpStackLocationToNext(Irp);
    IoSetCompletionRoutine(Irp, VTrcRawDReadWriteCompletion, (void *)seq,
                           TRUE, TRUE, TRUE);
  }
  else {
    KeReleaseSpinLock(&sharedState->mainMutex, oldirql);
    // Set current stack back one
    IoSkipCurrentIrpStackLocation(Irp);
  }

  // Return the results of the call to the disk driver.

  return IoCallDriver(deviceExtension->TargetDeviceObject, Irp);
} // end VTrcRawDReadWrite()

// VTrcRawDDispatchControl dispatches device control IRP's. The main reason
// we filter these is so we can see IOCTL_VOLUME_ONLINE calls. Once such a
// call completes, we can obtain the disk and partition numbers.

NTSTATUS
VTrcRawDDispatchControl(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp
  )
{
  PRAW_HOOK_EXTENSION deviceExtension = DeviceObject->DeviceExtension;
  NTSTATUS status;

  // If it's the "put volume online" I/O control code, and we haven't
  // yet determined the disk and partition numbers, then dispatch it
  // and get the disk and partition numbers. We do this because the
  // disk and partition numbers don't seem to become valid until after
  // this IOCTL completes successfully.

  if (deviceExtension->PartitionNumber == UNKNOWN_DEVICE_NUMBER &&
      IoGetCurrentIrpStackLocation(Irp)->Parameters.DeviceIoControl.
      IoControlCode == IOCTL_VOLUME_ONLINE) {
    status = VTrcRawDForwardIrpSynchronous(DeviceObject, Irp);
    if (NT_SUCCESS(status)) {
      DebugPrint((3, "VTrcRawD: Volume successfully put online.\n"));
      GetDiskAndPartitionNumbers(DeviceObject);
    }
    Irp->IoStatus.Status = status;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return status;
  }

  // Otherwise, just dispatch it normally.

  IoSkipCurrentIrpStackLocation(Irp);
  return IoCallDriver(deviceExtension->TargetDeviceObject, Irp);
} // end VTrcRawDDispatchGeneral()

// VTrcRawDDispatchGeneral dispatches general IRP's that require no processing
// by VTrcRawD.

NTSTATUS
VTrcRawDDispatchGeneral(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp
  )
{
  PRAW_HOOK_EXTENSION deviceExtension = DeviceObject->DeviceExtension;
  IoSkipCurrentIrpStackLocation(Irp);
  return IoCallDriver(deviceExtension->TargetDeviceObject, Irp);
} // end VTrcRawDDispatchGeneral()

// VTrcRawDUnload prepares for unloading the driver.

VOID
VTrcRawDUnload(
  IN PDRIVER_OBJECT DriverObject
  )
{
  PAGED_CODE();

  DebugPrint((1, "VTrcRawD: Unloading VTrcRawD\n"));
}

////////////////////////////////////////////////////////////////////////////
//                         COMPLETION ROUTINES
////////////////////////////////////////////////////////////////////////////

// VTrcRawDReadWriteCompletion will get control from the system at the
// completion of a read or write IRP. It logs that completion.

NTSTATUS
VTrcRawDReadWriteCompletion(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP           Irp,
  IN PVOID          Context
  )
{
  PRAW_HOOK_EXTENSION deviceExtension = DeviceObject->DeviceExtension;
  ULONG seq = (ULONG) Context;
  KIRQL oldirql;
  PCHAR eventBuffer;

  // Log the return and its return values.

  KeAcquireSpinLock(&sharedState->mainMutex, &oldirql);
  eventBuffer = LOG_EVENT((char) ENTRY_TYPE_RAWDISK_COMPLETE_OPERATION, 13);
  if (eventBuffer) {
    RtlCopyMemory(&eventBuffer[1], &seq, 4);
    RtlCopyMemory(&eventBuffer[5], &Irp->IoStatus.Status, 4);
    RtlCopyMemory(&eventBuffer[9], &Irp->IoStatus.Information, 4);
  }
  KeReleaseSpinLock(&sharedState->mainMutex, oldirql);

  // As in (almost) any completion routine, we have to mark Irp as pending if
  // necessary.  Note the braces around the call to IoMarkIrpPending --- they
  // are important because IoMarkIrpPending is a macro that expands to multiple
  // statements.

  if (Irp->PendingReturned) {
    IoMarkIrpPending(Irp);
  }

  return STATUS_SUCCESS;
} // VTrcRawDReadWriteCompletion

// VTrcRawDSynchronousIrpCompletion completes an IRP that is being performed
// synchronously. To aid in this, this routine sets a notification event,
// which is passed in the Context parameter.

NTSTATUS
VTrcRawDSynchronousIrpCompletion(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp,
  IN PVOID Context
  )
{
  PKEVENT Event = (PKEVENT) Context;

  UNREFERENCED_PARAMETER(DeviceObject);
  UNREFERENCED_PARAMETER(Irp);

  KeSetEvent(Event, IO_NO_INCREMENT, FALSE);

  return STATUS_MORE_PROCESSING_REQUIRED;
} // end VTrcRawDSynchronousIrpCompletion()

// VTrcRawDPnpCompletionRoutine is the completion routine for miscellaneous
// PNP requests. It just makes sure that after the PNP request is completed,
// the DO_POWER_PAGABLE bit of the filter device is consistent with that of
// the filtered device.

NTSTATUS
VTrcRawDPnpCompletionRoutine(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp,
  IN PVOID Context
  )
{
  PRAW_HOOK_EXTENSION deviceExtension = DeviceObject->DeviceExtension;

  UNREFERENCED_PARAMETER(Context);

  if (deviceExtension->TargetDeviceObject->Flags & DO_POWER_PAGABLE)
    DeviceObject->Flags |= DO_POWER_PAGABLE;
  else
    DeviceObject->Flags &= ~DO_POWER_PAGABLE;

  return Irp->IoStatus.Status;
}

////////////////////////////////////////////////////////////////////////////
//                               HELPERS
////////////////////////////////////////////////////////////////////////////

// VTrcRawDForwardIrpSynchronous dispatches the given IRP to the filtered
// device synchronously. The parameter DeviceObject refers to the filter
// device.

NTSTATUS
VTrcRawDForwardIrpSynchronous(
  IN PDEVICE_OBJECT DeviceObject,
  IN PIRP Irp
  )
{
  PRAW_HOOK_EXTENSION deviceExtension = DeviceObject->DeviceExtension;
  KEVENT event;
  NTSTATUS status;

  KeInitializeEvent(&event, NotificationEvent, FALSE);

  // Copy the stack down, set a completion routine that will set the
  // notification event when the IRP is completed, then call the lower driver.

  IoCopyCurrentIrpStackLocationToNext(Irp);
  IoSetCompletionRoutine(Irp, VTrcRawDSynchronousIrpCompletion, &event,
                         TRUE, TRUE, TRUE);
  status = IoCallDriver(deviceExtension->TargetDeviceObject, Irp);

  // Wait for the actual completion.

  if (status == STATUS_PENDING) {
    KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, NULL);
    status = Irp->IoStatus.Status;
  }

  return status;
} // end VTrcRawDForwardIrpSynchronous()

// GetDiskAndPartitionNumbers obtains the disk and partition numbers
// for the given filter device if they have not already been
// determined. If it cannot determine the disk and partition numbers,
// it tries to get the volume number and put that as the disk number.

void GetDiskAndPartitionNumbers (PDEVICE_OBJECT filterDevice)
{
  PRAW_HOOK_EXTENSION filterExtension = filterDevice->DeviceExtension;
  PIRP irp;
  STORAGE_DEVICE_NUMBER deviceNumber;
  VOLUME_NUMBER volumeNumber;
  IO_STATUS_BLOCK ioStatusBlock;
  KEVENT event;
  NTSTATUS status;

  if (filterExtension->PartitionNumber != UNKNOWN_DEVICE_NUMBER)
    return;

  DebugPrint((3, "VTrcRawD: Getting disk and partition of %x.\n",
              filterDevice));
  
  KeInitializeEvent(&event, NotificationEvent, FALSE);
  irp = IoBuildDeviceIoControlRequest(IOCTL_STORAGE_GET_DEVICE_NUMBER,
                                      filterExtension->TargetDeviceObject,
                                      NULL, 0, &deviceNumber,
                                      sizeof(deviceNumber), FALSE, &event,
                                      &ioStatusBlock);
  if (irp == NULL) {
    DebugPrint((1, "VTrcRawD: IoBuildDeviceIoControlRequest failed.\n"));
    return;
  }

  status = IoCallDriver(filterExtension->TargetDeviceObject, irp);
  if (status == STATUS_PENDING) {
    KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, NULL);
    status = ioStatusBlock.Status;
  }

  if (NT_SUCCESS(status)) {
    filterExtension->DiskNumber = deviceNumber.DeviceNumber;
    filterExtension->PartitionNumber = deviceNumber.PartitionNumber;
    DebugPrint((1, "VTrcRawD: Determined device to be disk %d, partition %d\n",
                deviceNumber.DeviceNumber, deviceNumber.PartitionNumber));
    return;
  }

  if (filterExtension->DiskNumber != UNKNOWN_DEVICE_NUMBER)
    return;

  DebugPrint((3, "VTrcRawD: Getting volume of %x.\n", filterDevice));
  
  KeInitializeEvent(&event, NotificationEvent, FALSE);
  irp = IoBuildDeviceIoControlRequest(IOCTL_VOLUME_QUERY_VOLUME_NUMBER,
                                      filterExtension->TargetDeviceObject,
                                      NULL, 0, &volumeNumber,
                                      sizeof(volumeNumber), FALSE, &event,
                                      &ioStatusBlock);
  if (irp == NULL) {
    DebugPrint((1, "VTrcRawD: IoBuildDeviceIoControlRequest failed.\n"));
    return;
  }

  status = IoCallDriver(filterExtension->TargetDeviceObject, irp);
  if (status == STATUS_PENDING) {
    KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, NULL);
    status = ioStatusBlock.Status;
  }

  if (!NT_SUCCESS(status)) {
    DebugPrint((1, "VTrcRawD: Get-volume-number IRP returned status %x.\n",
                status));
    return;
  }

  filterExtension->DiskNumber = volumeNumber.VolumeNumber;

  DebugPrint((3, "VTrcRawD: Determined device to be volume %u\n",
              volumeNumber.VolumeNumber));
}
