//
// Compression.cpp contains routines related to file compression.
//

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <winsvc.h>
#include <winioctl.h>
#include "../Common/VTrace.h"
#include "Debug.h"
#include "Structs.h"
#include "Helpers.h"
#include "Web.h"

////////////////////////////////////////////////////////////////////////////
//                                GLOBALS
////////////////////////////////////////////////////////////////////////////

extern REGISTRY_INFO regInfo;
extern HANDLE hVTrcLogDevice;

////////////////////////////////////////////////////////////////////////////
//                        FILE SELECTION ROUTINES
////////////////////////////////////////////////////////////////////////////

// WorthCompressing takes as input a data structure describing a file
// (findData) and returns a boolean indicating whether the file is
// worth compressing.  It also places the number of bytes to compress
// in bytesToCompress if it returns TRUE.

BOOL WorthCompressing (WIN32_FIND_DATA& findData, DWORD& bytesToCompress)
{
  DWORD length;

  if (findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) return FALSE;
  length = strlen(findData.cFileName);
  if (length < 3) return TRUE;
  if (!strcmp(&findData.cFileName[length-3], ".gz")) return FALSE;
  if (!strcmp(&findData.cFileName[length-3], "txt")) return FALSE;

  if (findData.nFileSizeHigh)
    bytesToCompress = 0xFFFFFFFFUL;
  else
    bytesToCompress = findData.nFileSizeLow;

  return TRUE;
}

// TraceDirectorySizable returns a boolean indicating if the trace directory
// is getting full enough to warrant a compression even though we had a
// compression already recently.

BOOL TraceDirectorySizable (void)
{
  DWORD totalSize = 0;
  char searchPattern[MAX_PATH];
  HANDLE findHandle;
  WIN32_FIND_DATA findData;

  // Loop through all the files, adding up their sizes. If the size ever
  // passes MAX_BYTES_BEFORE_COMPRESS, return TRUE.

  sprintf(searchPattern, "%s\\Traces\\*", regInfo.mainDirectory);
  findHandle = FindFirstFile(searchPattern, &findData);
  if (findHandle != INVALID_HANDLE_VALUE) {
    do {
      totalSize += findData.nFileSizeLow;
      if (totalSize >= MAX_BYTES_BEFORE_COMPRESS) {
        FindClose(findHandle);
        return TRUE;
      }
    } while (FindNextFile(findHandle, &findData));
    FindClose(findHandle);
  }

  // We didn't find enough bytes, so return FALSE.

  return FALSE;
}

////////////////////////////////////////////////////////////////////////////
//                           FILE COMPRESSION
////////////////////////////////////////////////////////////////////////////

// CompressFile compresses the file described in fileToCompressData.
// The other parameter it is passed is the regInfo structure containing
// information from the registry, including the name of the main directory,
// where GZIP.EXE will be found, and the name of the trace directory, where
// the described file is.

void CompressFile (WIN32_FIND_DATA& fileToCompressData)
{
  DWORD fileSize;
  char gzipFullPath[MAX_PATH], commandLine[MAX_PATH+8];
  STARTUPINFO startupInfo;
  PROCESS_INFORMATION processInfo;

  if (!WorthCompressing(fileToCompressData, fileSize) || fileSize == 0)
    return;

  // Launch a GZIP process to compress the file

  sprintf(gzipFullPath, "%s\\GZIP.EXE", regInfo.mainDirectory);
  sprintf(commandLine, "GZIP \"%s\\Traces\\%s\"", regInfo.mainDirectory,
          fileToCompressData.cFileName);
  GetStartupInfo(&startupInfo);
  if (!CreateProcess(gzipFullPath, commandLine, NULL, NULL, FALSE,
                     DETACHED_PROCESS, NULL, NULL, &startupInfo, &processInfo))
    return;

  // Wait for the compression to be finished

  WaitForSingleObject(processInfo.hProcess, INFINITE);
}

// CompressAll compresses all the uncompressed files in the Traces directory.

void CompressAll (void)
{
  char searchPattern[MAX_PATH];
  HANDLE findHandle;
  WIN32_FIND_DATA findData;

  // First, make sure the user is idle.

  if (GetUserIdleTime() < USER_IDLE_MS_BEFORE_COMPRESS)
    return;

  // Compress the files. After each compression, make sure the user is
  // still idle.

  sprintf(searchPattern, "%s\\Traces\\*", regInfo.mainDirectory);
  findHandle = FindFirstFile(searchPattern, &findData);
  if (findHandle != INVALID_HANDLE_VALUE) {
    do {
      CompressFile(findData);
    } while (GetUserIdleTime() >= USER_IDLE_MS_BEFORE_COMPRESS &&
             FindNextFile(findHandle, &findData));
    FindClose(findHandle);
  }
}
