//
// Helpers.cpp contains useful routines for VTrcSrvc.
//
// For 79-character width, truncate the window after this sentence is finished.

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <winioctl.h>
#include <winsvc.h>
#include <lmcons.h>
#include <lmerr.h>
#include <lmapibuf.h>
#include <lmwksta.h>
#include <time.h>
#include "../Common/VTrace.h"
#include "Structs.h"

///////////////////////////////////////////////////////////////////////////
//                          EXTERNAL GLOBALS
///////////////////////////////////////////////////////////////////////////

extern HANDLE hVTrcLogDevice;

///////////////////////////////////////////////////////////////////////////
//                                MATH
///////////////////////////////////////////////////////////////////////////

// MillisecondsBetweenTimes computes the number of milliseconds that
// passed between the beginning time and the ending time, and returns that.
// If there is an overflow, it returns 0x7FFFFFFF. (It could return
// 0xFFFFFFFF, but for safety it returns something that wouldn't be considered
// negative even if used in a signed variable.)

DWORD MillisecondsBetweenTimes (FILETIME beginTime, FILETIME endTime)
{
  LARGE_INTEGER beginLargeInteger, endLargeInteger;

  beginLargeInteger.LowPart = beginTime.dwLowDateTime;
  beginLargeInteger.HighPart = beginTime.dwHighDateTime;
  endLargeInteger.LowPart = endTime.dwLowDateTime;
  endLargeInteger.HighPart = endTime.dwHighDateTime;

  endLargeInteger.QuadPart -= beginLargeInteger.QuadPart;
  endLargeInteger.QuadPart /= 10000;

  if (endLargeInteger.HighPart)
    return 0x7FFFFFFF;
  else
    return endLargeInteger.LowPart;
}

///////////////////////////////////////////////////////////////////////////
//                          REGISTRY ACCESS
///////////////////////////////////////////////////////////////////////////

// ReadRegistry reads information from the registry into the
// REGISTRY_INFO structure, returning FALSE if it is unsuccessful and
// returning TRUE if it is successful.  The "reloading" parameter
// indicates whether we have already read the registry earlier and are
// just reloading it now; if so, we should read the version number
// first, and fail if that version number does not match the one
// already in regInfoPtr. Also, we shouldn't reread the main directory
// or user number if we're reloading.

BOOL ReadRegistry (REGISTRY_INFO *regInfoPtr, BOOL reloading)
{
  HKEY hKey;
  DWORD type, length, n;
  LONG status;

  // Open registry key

  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                        "Software\\Jay Lorch\\VTrace\\CurrentVersion", 0,
                        KEY_QUERY_VALUE, &hKey);
  if (status != ERROR_SUCCESS)
    return FALSE;

  // Read version number

  length = sizeof(DWORD);
  status = RegQueryValueEx(hKey, "Version", 0, &type, (BYTE *) &n, &length);
  if (status != ERROR_SUCCESS || type != REG_DWORD || length != 4 ||
      (reloading && n != regInfoPtr->vtraceVersion)) {
    RegCloseKey(hKey);
    return FALSE;
  }
  regInfoPtr->vtraceVersion = n;

  // Only reread the main directory and user number if we're not reloading.

  if (!reloading) {
    // Read main directory

    length = MAX_PATH;
    status = RegQueryValueEx(hKey, "MainDirectory", 0, &type,
                             (BYTE *) regInfoPtr->mainDirectory, &length);
    if (status != ERROR_SUCCESS || type != REG_SZ) {
      RegCloseKey(hKey);
      return FALSE;
    }

    // Read user number

    length = sizeof(DWORD);
    status = RegQueryValueEx(hKey, "UserNumber", 0, &type,
                             (BYTE *) &regInfoPtr->userNumber, &length);
    if (status != ERROR_SUCCESS || type != REG_DWORD || length != 4) {
      RegCloseKey(hKey);
      return FALSE;
    }
  }

  // Read automatic upload value.  If it's not there, pretend it's true.

  length = sizeof(DWORD);
  status = RegQueryValueEx(hKey, "AlwaysConnected", 0, &type,
                           (BYTE *) &n, &length);
  if (status != ERROR_SUCCESS || type != REG_DWORD || length != 4)
    regInfoPtr->automaticUpload = TRUE;
  else
    regInfoPtr->automaticUpload = (n != 0);

  // Read proxy server name.  If it's not there, pretend it's blank.

  length = MAX_SERVER_LENGTH;
  status = RegQueryValueEx(hKey, "ProxyServer", 0, &type,
                           (BYTE *) regInfoPtr->proxyServer, &length);
  if (status != ERROR_SUCCESS || type != REG_SZ)
    regInfoPtr->proxyServer[0] = '\0';

  // Read proxy port number.  If it's not there, pretend it's 0.

  length = 4;
  status = RegQueryValueEx(hKey, "ProxyPort", 0, &type,
                           (BYTE *) &regInfoPtr->proxyPort, &length);
  if (status != ERROR_SUCCESS || type != REG_DWORD || length != 4)
    regInfoPtr->proxyPort = DEFAULT_PROXY_PORT;

  // Read transfer size limit.  If it's not there, pretend it's the default.
  // Make sure the transfer size is in the valid range.

  length = 4;
  status = RegQueryValueEx(hKey, "MaxTransferKB", 0, &type,
                           (BYTE *) &regInfoPtr->maxTransferSize, &length);
  if (status != ERROR_SUCCESS || type != REG_DWORD || length != 4 ||
      regInfoPtr->maxTransferSize < MIN_TRANSFER_SIZE_LIMIT ||
      regInfoPtr->maxTransferSize > MAX_TRANSFER_SIZE_LIMIT)
    regInfoPtr->maxTransferSize = DEFAULT_TRANSFER_SIZE_LIMIT * 1024;
  else
    regInfoPtr->maxTransferSize *= 1024;

  // Read free MB needed.  If it's not there, pretend it's the default.
  // Make sure the free MB is in the valid range.

  length = 4;
  status = RegQueryValueEx(hKey, "FreeMBNeeded", 0, &type,
                           (BYTE *) &regInfoPtr->freeBytesNeeded, &length);
  if (status != ERROR_SUCCESS || type != REG_DWORD || length != 4 ||
      regInfoPtr->freeBytesNeeded < MIN_FREE_SPACE ||
      regInfoPtr->freeBytesNeeded > MAX_FREE_SPACE)
    regInfoPtr->freeBytesNeeded = DEFAULT_FREE_SPACE * 1048576;
  else
    regInfoPtr->freeBytesNeeded *= 1048576;

  // Read max trace piece MB.  If it's not there, pretend it's the default.
  // Make sure the trace piece MB is in the valid range.

  length = 4;
  status = RegQueryValueEx(hKey, "MaxTracePieceMB", 0, &type,
                           (BYTE *) &regInfoPtr->maxTracePieceBytes, &length);
  if (status != ERROR_SUCCESS || type != REG_DWORD || length != 4 ||
      regInfoPtr->maxTracePieceBytes < MIN_TRACE_PIECE_SIZE_LIMIT ||
      regInfoPtr->maxTracePieceBytes > MAX_TRACE_PIECE_SIZE_LIMIT)
    regInfoPtr->maxTracePieceBytes = DEFAULT_TRACE_PIECE_SIZE_LIMIT * 1048576;
  else
    regInfoPtr->maxTracePieceBytes *= 1048576;

  // Read upload URL list.  If it's not there, pretend it's the default.

  length = MAX_UPLOAD_URL_LIST_LENGTH;
  status = RegQueryValueEx(hKey, "UploadUrlList", 0, &type,
                           (BYTE *) regInfoPtr->uploadUrlList, &length);
  if (status != ERROR_SUCCESS || type != REG_SZ)
    strcpy(regInfoPtr->uploadUrlList, DEFAULT_UPLOAD_URL_LIST);

  // Close the registry key and return success indicator

  RegCloseKey(hKey);
  return TRUE;
}

// GetCurrentTraceFileNumber obtains the current trace file number and puts
// it in currentTraceFileNumberPtr. It returns a boolean indicating if it
// was successful.

BOOL GetCurrentTraceFileNumber (DWORD *currentTraceFileNumberPtr)
{
  HKEY hKey;
  LONG status;
  DWORD length, type;

  // Open registry key

  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                        "Software\\Jay Lorch\\VTrace\\CurrentVersion", 0,
                        KEY_QUERY_VALUE, &hKey);
  if (status != ERROR_SUCCESS)
    return FALSE;

  // Read various parameters from the registry that we must fail if we
  // don't find

  length = 4;
  status = RegQueryValueEx(hKey, "LastTraceNumber", 0, &type,
                           (BYTE *) currentTraceFileNumberPtr, &length);
  if (status != ERROR_SUCCESS || type != REG_DWORD || length != 4) {
    RegCloseKey(hKey);
    return FALSE;
  }

  RegCloseKey(hKey);
  return TRUE;
}

// GetMetadataAge returns the number of milliseconds it has been since the
// metadata was last dumped.

DWORD GetMetadataAge (void)
{
  LONG status;
  HKEY hKey;
  DWORD type, length;
  FILETIME currentTimeFT;
  SYSTEMTIME currentTimeST;
  LARGE_INTEGER timeMetadataRecorded, currentTime;

  status = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                        "Software\\Jay Lorch\\VTrace\\CurrentVersion", 0,
                        KEY_QUERY_VALUE, &hKey);
  if (status != ERROR_SUCCESS) return 0;

  length = 8;
  status = RegQueryValueEx(hKey, "TimeMetadataRecorded", 0, &type,
                           (BYTE *) &timeMetadataRecorded, &length);
  RegCloseKey(hKey);

  if (status != ERROR_SUCCESS || type != REG_BINARY || length != 8)
    return 0;

  GetLocalTime(&currentTimeST);
  if (!SystemTimeToFileTime(&currentTimeST, &currentTimeFT))
    return 0;
  currentTime.HighPart = currentTimeFT.dwHighDateTime;
  currentTime.LowPart = currentTimeFT.dwLowDateTime;

  // To convert 100-ns units to 1-ms units, divide by 10,000

  return (DWORD) ((currentTime.QuadPart - timeMetadataRecorded.QuadPart) /
                  10000);
}

// GetUserName finds the current logged-in user name and domain.  It
// takes as input the open Winlogon registry key (hKey), a buffer in
// which to put the user name and domain (userName), and the length of
// the buffer (bufferSize).  It puts a null-terminated string there
// and returns the length of the string not including the terminating
// null.

DWORD GetUserName (HKEY hKey, char *userName, DWORD bufferSize)
{
  DWORD nameLength, domainLength;

  // Read the user name into the buffer.  If this fails, return the
  // empty string

  nameLength = bufferSize;
  if (RegQueryValueEx(hKey, "DefaultUserName", 0, NULL,
                      (BYTE *) userName, &nameLength) != ERROR_SUCCESS) {
    *userName = '\0';
    return 0;
  }

  // If there is no room for a domain name, return just the name

  if (nameLength >= bufferSize) {
    userName[bufferSize-1] = '\0';
    return bufferSize - 1;
  }

  // Read the domain name.  If this succeeds, put the @ between the name
  // and domain and return the full string; if it fails, just return the
  // name.

  domainLength = bufferSize - nameLength;
  if (RegQueryValueEx(hKey, "DefaultDomainName", 0, NULL,
                      (BYTE *) &userName[nameLength], &domainLength)
                      == ERROR_SUCCESS) {
    userName[nameLength - 1] = '@';
    return nameLength + domainLength - 1;
  }
  else
    return nameLength - 1;
}

///////////////////////////////////////////////////////////////////////////
//                        COMMUNICATION WITH VTRCLOG
///////////////////////////////////////////////////////////////////////////

// GetUserIdleTime returns the number of milliseconds the user has been idle.

DWORD GetUserIdleTime (void)
{
  LARGE_INTEGER idleTime;
  DWORD returnSize;

  if (!DeviceIoControl(hVTrcLogDevice, VTRACE_CTRL_GET_IDLE_TIME, NULL, 0,
                       &idleTime, 8, &returnSize, NULL)) return 0;
  if (returnSize != 8) return 0;

  // To convert 100-ns units to 1-ms units, divide by 10,000

  return (DWORD) (idleTime.QuadPart / 10000);
}
