// UserName.cpp
//
// This file contains routines for entering a user-name-changed event
// whenever it is needed. This file exports UserNameThreadRoutine, which
// is launched by the VTrcSrvc service to handle this.
//
// For 79-character width, truncate the window after this sentence is finished.

#include <windows.h>
#include <winioctl.h>
#include <winsvc.h>
#include <stdio.h>
#include "../Common/VTrace.h"
#include "../Common/RegConstraints.h"
#include "Debug.h"
#include "Structs.h"
#include "Helpers.h"

////////////////////////////////////////////////////////////////////////////
//                                TYPES
////////////////////////////////////////////////////////////////////////////

typedef struct _user_name_globals {
  HANDLE userNameRequiredEventHandle;
  HANDLE userNameChangedEventHandle;
  HKEY currentUserNameRegistryKey;
  char lastUserNameLogged[MAX_USER_NAME_LENGTH];
} USER_NAME_GLOBALS;

////////////////////////////////////////////////////////////////////////////
//                             GLOBALS
////////////////////////////////////////////////////////////////////////////

USER_NAME_GLOBALS globals;
extern HANDLE hVTrcLogDevice;

////////////////////////////////////////////////////////////////////////////
//                          FORWARD DEFINITIONS
////////////////////////////////////////////////////////////////////////////

void RecordUserName (BOOL evenIfUnchanged);

////////////////////////////////////////////////////////////////////////////
//                      INITIALIZING / TERMINATING
////////////////////////////////////////////////////////////////////////////

// InitializeUserNameHandling opens handles to events related to handling
// user-name events.

BOOL InitializeUserNameHandling (void)
{
  LONG err;

  // Open a handle to the user-name required event.

  globals.userNameRequiredEventHandle = 
    OpenEvent(SYNCHRONIZE, FALSE, "VTraceUserNameRequiredEvent");
  if (globals.userNameRequiredEventHandle == NULL) {
    DisplayError("Could not open user-name-required event");
    return FALSE;
  }

  // Create a new event with which to receive notifications of the current
  // user name changed.

  globals.userNameChangedEventHandle = CreateEvent(NULL, FALSE, FALSE, NULL);
  if (globals.userNameChangedEventHandle == NULL) {
    DisplayError("Could not create event to be notified of name changes");
    CloseHandle(globals.userNameRequiredEventHandle);
    return FALSE;
  }

  // Open the registry key that contains the current user name.

  err = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                   "Software\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon",
                     0, KEY_NOTIFY | KEY_READ,
                     &globals.currentUserNameRegistryKey);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not open user name registry key");
    CloseHandle(globals.userNameChangedEventHandle);
    CloseHandle(globals.userNameRequiredEventHandle);
    return FALSE;
  }

  // Arrange to be notified via globals.userNameChangedEventHandle of any
  // changes to the current user name.

  err = RegNotifyChangeKeyValue(globals.currentUserNameRegistryKey, FALSE,
                                REG_NOTIFY_CHANGE_LAST_SET,
                                globals.userNameChangedEventHandle, TRUE);
  if (err != ERROR_SUCCESS) {
    SetLastError(err);
    DisplayError("Could not arrange to be notified of key value change");
    RegCloseKey(globals.currentUserNameRegistryKey);
    CloseHandle(globals.userNameChangedEventHandle);
    CloseHandle(globals.userNameRequiredEventHandle);
    return FALSE;
  }

  // Set the "last user name" to a dummy value so that the first time we
  // look at the user name it appears to have changed.

  strcpy(globals.lastUserNameLogged, "Dummy Name XQI");

  // Return an indication of success.

  return TRUE;
}

// TerminateUserNameHandling releases any resources obtained by a successful
// call to InitializeUserNameHandling.

void TerminateUserNameHandling (void)
{
  RegCloseKey(globals.currentUserNameRegistryKey);
  CloseHandle(globals.userNameChangedEventHandle);
  CloseHandle(globals.userNameRequiredEventHandle);
}

////////////////////////////////////////////////////////////////////////////
//                               USER NAME
////////////////////////////////////////////////////////////////////////////

// RecordUserName records the current user name in the log. However, if
// evenIfUnchanged is FALSE and the name hasn't changed since the last time
// this routine was called, it does nothing.

void RecordUserName (BOOL evenIfUnchanged)
{
  char entry[MAX_USER_NAME_LENGTH+2];
  DWORD entryLength, returnSize;

  entryLength = GetUserName(globals.currentUserNameRegistryKey, &entry[2],
                            MAX_USER_NAME_LENGTH) + 2;
  if (evenIfUnchanged || strcmp(&entry[2], globals.lastUserNameLogged)) {
    entry[0] = (char) ENTRY_TYPE_USER_CHANGED;
    entry[1] = (char) entryLength;
    DeviceIoControl(hVTrcLogDevice, VTRACE_CTRL_LOG_EVENT,
                    entry, entryLength, NULL, 0, &returnSize, NULL);
    strcpy(globals.lastUserNameLogged, &entry[2]);
  }
}

////////////////////////////////////////////////////////////////////////////
//                            MAIN ROUTINES
////////////////////////////////////////////////////////////////////////////

// UserNameLoop loops, waiting for an event each time, and handles the
// events as they occur.

void UserNameLoop (void)
{
  HANDLE waitHandles[2];
  LONG err;

  // Initialize the array of handles on which to wait.

  waitHandles[0] = globals.userNameRequiredEventHandle;
  waitHandles[1] = globals.userNameChangedEventHandle;

  // Loop, handling events as they occur.

  while (1) {
    PrintDebugInfo("Waiting for next user-name event.\n", 0, 0);
    switch (WaitForMultipleObjects(2, waitHandles, FALSE, INFINITE)) {
    case WAIT_OBJECT_0 :
      PrintDebugInfo("Was notified that a user-name event is required.\n",
                     0, 0);
      RecordUserName(TRUE);
      break;
    case WAIT_OBJECT_0 + 1 :
      PrintDebugInfo("User name changed.\n", 0, 0);
      RecordUserName(FALSE);

      // Set ourselves up again to be notified of a change to this registry
      // key, since RegNotifyChangeKeyValue only notifies you once, and we
      // just got notified.
      
      err = RegNotifyChangeKeyValue(globals.currentUserNameRegistryKey, FALSE,
                                    REG_NOTIFY_CHANGE_LAST_SET,
                                    globals.userNameChangedEventHandle, TRUE);
      if (err != ERROR_SUCCESS) {
        SetLastError(err);
        DisplayError("Could not reestablish notification of user changing");
      }
      break;
    }
  }
}

// UserNameThreadRoutine is the thread routine for performing user-name
// logging.

void UserNameThreadRoutine (void)
{
  if (InitializeUserNameHandling()) {
    PrintDebugInfo("Starting user-name loop.\n", 0, 0);
    UserNameLoop();
    TerminateUserNameHandling();
  }
}
