//
// VTrcSrvc.cpp contains the code for VTrcSrvc, the VTrace service.
// This service does the following periodic tasks:  record metadata,
// compress and upload trace files, and change the event mask.
//
// The service-related code in this file is extensively borrowed from
// Paula Tomlinson's article, "How to Write an NT Service", in the Feb
// 1996 issue of the Windows Developer's Journal.
//

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <winioctl.h>
#include <winsvc.h>
#include "../Common/VTrace.h"
#include "../Common/VTrcSrvc.h"
#include "Structs.h"
#include "Debug.h"
#include "Helpers.h"
#include "Compression.h"
#include "Web.h"
#include "UserName.h"

///////////////////////////////////////////////////////////////////////////
//                              GLOBALS
///////////////////////////////////////////////////////////////////////////

SERVICE_STATUS_HANDLE hService;
HANDLE hUserNameThread, hCompressionThread;
HANDLE hVTrcLogDevice;
HANDLE hDoneEvent;
SERVICE_STATUS serviceStatus;
REGISTRY_INFO regInfo;

///////////////////////////////////////////////////////////////////////////
//                        FORWARD DEFINITIONS
///////////////////////////////////////////////////////////////////////////

void ServiceMain (DWORD argc, LPTSTR *argv);
void ServiceHandler (DWORD control);
void CompressionThreadRoutine (void);
void InitServiceStatus (void);
BOOL NotifySCM (DWORD state, DWORD progress);

void RecordMetadata (void);

///////////////////////////////////////////////////////////////////////////
//                               MAIN
///////////////////////////////////////////////////////////////////////////

// The main routine contacts the service control dispatcher with the
// name and address of the service main routine ServiceMain.  It is
// the responsibility of the dispatcher to start the ServiceMain routine.

void main (void)
{
  SERVICE_TABLE_ENTRY ServiceTable[] = {
    { "VTrcSrvc", (LPSERVICE_MAIN_FUNCTION) &ServiceMain },
    { NULL, NULL }
  };

  StartServiceCtrlDispatcher(ServiceTable);
}

///////////////////////////////////////////////////////////////////////////
//                          SERVICE ROUTINES
///////////////////////////////////////////////////////////////////////////

// ServiceMain is the main routine for this service.  It registers the
// service control handler, initializes the "done" event, and creates the
// thread that will do the main work of the service.  It runs until that
// done event is signalled.

void ServiceMain (DWORD argc, LPTSTR *argv)
{
  DWORD threadId;

  // Register the service control handler ServiceHandler

  hService = RegisterServiceCtrlHandler("VTrcSrvc",
                                        (LPHANDLER_FUNCTION) &ServiceHandler);
  if (hService == NULL) return;

  // Initialize the serviceStatus structure so we can send status information

  InitServiceStatus();

  // Get information from the registry

  NotifySCM(SERVICE_START_PENDING, 1);           // Step 1 of starting up
  if (!ReadRegistry(&regInfo, FALSE)) return;

  // Create the "done" event, initialized to unsignalled

  NotifySCM(SERVICE_START_PENDING, 2);           // Step 2 of starting up
  hDoneEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
  if (hDoneEvent == NULL) return;

  // Open a handle to the VTrcLog device, needed for various reasons

  NotifySCM(SERVICE_START_PENDING, 3);           // Step 3 of starting up
  hVTrcLogDevice = CreateFile("\\\\.\\VTrcLog", GENERIC_READ | GENERIC_WRITE,
                              0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,
                              NULL);
  if (hVTrcLogDevice == INVALID_HANDLE_VALUE) {
    CloseHandle(hDoneEvent);
    return;
  }

  // Create the thread that performs the user-logging control work of this
  // service.

  NotifySCM(SERVICE_START_PENDING, 4);           // Step 4 of starting up
  hUserNameThread =
    CreateThread(0, 0, (LPTHREAD_START_ROUTINE) &UserNameThreadRoutine, 0, 0,
                 &threadId);

  // Create the thread that performs the compression-related work of this
  // service

  NotifySCM(SERVICE_START_PENDING, 5);           // Step 5 of starting up
  hCompressionThread =
    CreateThread(0, 0, (LPTHREAD_START_ROUTINE) &CompressionThreadRoutine, 0,
                 0, &threadId);

  NotifySCM(SERVICE_RUNNING, 0);                 // We are now started up

  // Wait for the signal to stop, then terminate the work threads, close
  // handles, and leave

  WaitForSingleObject(hDoneEvent, INFINITE);

  if (hCompressionThread != NULL) {
    TerminateThread(hCompressionThread, 0);
    CloseHandle(hCompressionThread);
    hCompressionThread = NULL;
  }
  if (hUserNameThread != NULL) {
    TerminateThread(hUserNameThread, 0);
    CloseHandle(hUserNameThread);
    hUserNameThread = NULL;
  }

  CloseHandle(hDoneEvent);
  CloseHandle(hVTrcLogDevice);
}

// ServiceHandler is the routine that is called by the service control
// dispatcher to execute various control commands.

void ServiceHandler (DWORD control)
{
  switch (control) {
  case SERVICE_CONTROL_STOP :
    NotifySCM(SERVICE_STOP_PENDING, 1);
    Sleep(PAUSE_FOR_FLUSHING_MS);
    SetEvent(hDoneEvent);             // Set event to terminate ServiceMain
    NotifySCM(SERVICE_STOPPED, 0);
    break;

  case SERVICE_CONTROL_PAUSE :
    NotifySCM(SERVICE_PAUSE_PENDING, 1);
    if (hUserNameThread != NULL) SuspendThread(hUserNameThread);
    if (hCompressionThread != NULL) SuspendThread(hCompressionThread);
    NotifySCM(SERVICE_PAUSED, 0);
    break;

  case SERVICE_CONTROL_CONTINUE :
    NotifySCM(SERVICE_CONTINUE_PENDING, 1);
    if (hUserNameThread != NULL) ResumeThread(hUserNameThread);
    if (hCompressionThread != NULL) ResumeThread(hCompressionThread);
    NotifySCM(SERVICE_RUNNING, 0);
    break;

  case SERVICE_CONTROL_INTERROGATE :
    NotifySCM(serviceStatus.dwCurrentState, 0);  // Send status information
    break;

  case SERVICE_CONTROL_RELOAD_REGISTRY :
    ReadRegistry(&regInfo, TRUE);
    break;
  }
}

// CompressionThreadRoutine does the work of this service related to
// compression. After a period of idleness, it starts compressing trace files.
// If automatic uploading is set, this routine will do more than just compress
// files: it will also upload them. This routine will not compress files if
// it just did so recently unless either the trace directory is getting full
// or the user has been active between the two compressions.

void CompressionThreadRoutine (void)
{
  DWORD idleTime, returnSize, reasons, millisecondsSinceLastCompression;
  FILETIME timeOfLastCompression = {0, 0}, currentTime;

  while (1) {
    idleTime = GetUserIdleTime();
    if (idleTime < USER_IDLE_MS_BEFORE_COMPRESS)
      Sleep(USER_IDLE_MS_BEFORE_COMPRESS - idleTime);
    else {

      // Now we know the user has been idle long enough. Now we see whether
      // one of the following conditions hold:
      //
      // 1) It has been a while since the last compression
      // 2) The user has been active since the last compression
      // 3) The trace directory has reached a sizable size
      //
      // If any of these hold, we will compress now, otherwise we will wait
      // a while and check these conditions again.
      
      GetSystemTimeAsFileTime(&currentTime);
      millisecondsSinceLastCompression =
        MillisecondsBetweenTimes(timeOfLastCompression, currentTime);
      if (idleTime < millisecondsSinceLastCompression ||
          millisecondsSinceLastCompression > MAX_MS_BETWEEN_COMPRESSIONS ||
          (regInfo.automaticUpload && TraceDirectorySizable())) {

        // OK, it's time to compress. First, get access to the trace files.
        
        reasons = BECAUSE_ACCESS_REQUIRED;
        if (!DeviceIoControl(hVTrcLogDevice, VTRACE_CTRL_ADD_REASONS_TO_STOP,
                             &reasons, 4, NULL, 0, &returnSize, NULL)) {

          // We didn't get access to the trace files, so we wait for
          // our turn.

          PrintDebugInfo("Could not obtain access to compress files. "
                         "Waiting.\n", 0, 0);
          Sleep(PAUSE_FOR_ACCESS_MS);
        }
        else {

          // OK, we now have access to the trace files and we're ready to
          // start compressing.
          
          PrintDebugInfo("Compressing now. User has been idle for %ld ms "
                         "and last compression was %ld ms ago.\n",
                         idleTime, millisecondsSinceLastCompression);
          timeOfLastCompression = currentTime;
          Sleep(PAUSE_FOR_FLUSHING_MS);

          // Before we compress, record metadata.

          RecordMetadata();
          CompressAll();
          if (regInfo.automaticUpload)
            SendAll(regInfo);

          // Now that we're done compressing, release access to the trace
          // files by telling VTrcLog that "access required" is no longer a
          // reason for not logging. While we're at it, also tell VTrcLog
          // that "disk full" is also no longer a reason for not logging,
          // since we've compressed and possibly uploaded some files to make
          // room. (If we didn't make enough room, there's no harm in telling
          // VTrcLog we did; it'll find out soon enough that there isn't
          // enough room.)
          
          reasons = BECAUSE_ACCESS_REQUIRED | BECAUSE_DISK_FULL;
          DeviceIoControl(hVTrcLogDevice, VTRACE_CTRL_REMOVE_REASONS_TO_STOP,
                          &reasons, 4, NULL, 0, &returnSize, NULL);
          Sleep(MAX_MS_BEFORE_TRY_ANOTHER_COMPRESS);
        }
      }
      else {

        // Even though the user has been idle a while, it's not been long
        // enough since the last compression. Wait a while and check again
        // after then.
        
        PrintDebugInfo("Not compressing because user has been idle for %ld ms "
                       "but last compression was %ld ms ago\n",
                       idleTime, millisecondsSinceLastCompression);
        Sleep(MAX_MS_BEFORE_TRY_ANOTHER_COMPRESS);
      }
    }
  }
}

// InitServiceStatus initializes the global structure serviceStatus with
// status information to be sent to the service control dispatcher at
// appropriate times.

void InitServiceStatus (void)
{
  serviceStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
  serviceStatus.dwCurrentState = SERVICE_START_PENDING;
  serviceStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP |
                                     SERVICE_ACCEPT_PAUSE_CONTINUE |
                                     SERVICE_ACCEPT_SHUTDOWN;
  serviceStatus.dwWin32ExitCode = 0;
  serviceStatus.dwServiceSpecificExitCode = 0;
  serviceStatus.dwCheckPoint = 0;
  serviceStatus.dwWaitHint = 5000;
}

// NotifySCM notifies the service control manager of the current state of
// this service (state) as well as its progress within that state (progress).

BOOL NotifySCM (DWORD state, DWORD progress)
{
  serviceStatus.dwCurrentState = state;
  serviceStatus.dwCheckPoint = progress;

  return SetServiceStatus(hService, &serviceStatus);
}

///////////////////////////////////////////////////////////////////////////
//                           RECORDING METADATA
///////////////////////////////////////////////////////////////////////////

// RecordMetadata creates a process from the Metadata.exe file in the main
// VTrace directory.  It waits for that process, which will record current
// metadata information, to complete.

void RecordMetadata (void)
{
  STARTUPINFO startupInfo;
  PROCESS_INFORMATION processInfo;
  char fileName[MAX_PATH];

  // First, make sure we haven't recorded metadata too recently.

  if (GetMetadataAge() < METADATA_VERY_OLD_MS)
    return;

  // Launch the process to record the metadata

  sprintf(fileName, "%s\\Metadata.exe", regInfo.mainDirectory);
  GetStartupInfo(&startupInfo);
  CreateProcess(fileName, NULL, NULL, NULL, FALSE, DETACHED_PROCESS,
                NULL, NULL, &startupInfo, &processInfo);
  WaitForSingleObject(processInfo.hProcess, INFINITE);
}
