//
// Web.cpp contains routines related to interacting with the Web site.
//

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include "Structs.h"
#include "Helpers.h"
#include "JSocket.h"
#include "../Common/WebConstraints.h"

////////////////////////////////////////////////////////////////////////////
//                               MACROS
////////////////////////////////////////////////////////////////////////////

#define MIN(a,b)               (((a)>(b))?(b):(a))
#define DATA_BUFFER_SIZE       4096

////////////////////////////////////////////////////////////////////////////
//                          FORWARD DEFINITIONS
////////////////////////////////////////////////////////////////////////////

MY_ERROR SendAndDeleteFile (WIN32_FIND_DATA& findData, SERVER_INFO& serverInfo,
                            REGISTRY_INFO& regInfo);
MY_ERROR SendFile (FILE_INFO& fileInfo, SERVER_INFO& serverInfo,
                   REGISTRY_INFO& regInfo);
MY_ERROR SendFileToServer (FILE_INFO& fileInfo, SERVER_INFO& serverInfo,
                           REGISTRY_INFO& regInfo);
MY_ERROR SendFilePiece (FILE_INFO& fileInfo, SERVER_INFO& serverInfo,
                        REGISTRY_INFO& regInfo, DWORD bytesToTransfer);

MY_ERROR GetRemoteFileSize (FILE_INFO& fileInfo, SERVER_INFO& serverInfo,
                            REGISTRY_INFO& regInfo);

MY_ERROR SendLastTraceUploadedMessage (SERVER_INFO& serverInfo,
                                       REGISTRY_INFO& regInfo);

////////////////////////////////////////////////////////////////////////////
//                        FILE SELECTION ROUTINES
////////////////////////////////////////////////////////////////////////////

// WorthUploading takes as input a data structure describing a file
// (findData) and returns a boolean indicating whether the file is
// worth uploading.  It also places the number of bytes to upload
// in bytesToUpload if it returns TRUE.

BOOL WorthUploading (WIN32_FIND_DATA& findData, DWORD& bytesToUpload)
{
  if (findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) return FALSE;

  if (findData.nFileSizeHigh ||
      findData.nFileSizeLow > MAX_FILE_SIZE_TO_UPLOAD)
    bytesToUpload = MAX_FILE_SIZE_TO_UPLOAD;
  else
    bytesToUpload = findData.nFileSizeLow;

  return TRUE;
}

////////////////////////////////////////////////////////////////////////////
//                      CHOOSING AMONG WEB SERVERS
////////////////////////////////////////////////////////////////////////////

// ChangeWebServer changes to the next web server in the list given in
// the registry information (regInfo).  It sets the server information in
// serverInfo accordingly.  It returns a status value.

MY_ERROR ChangeWebServer (SERVER_INFO& serverInfo, REGISTRY_INFO& regInfo)
{
  int i;
  char *p, *dest;

  // Increment the server index so that we can try the next one

  serverInfo.currentUrlListIndex++;
  if (serverInfo.currentUrlListIndex >= serverInfo.urlListSize)
    serverInfo.currentUrlListIndex = 0;

  // Increment the number of URL's that have failed in a row without
  // success.  If this is at least the total number of URL's available, and
  // at least 3, fail completely.

  serverInfo.numUrlsFailedConsecutively++;
  if (serverInfo.numUrlsFailedConsecutively >= serverInfo.urlListSize &&
      serverInfo.numUrlsFailedConsecutively >= 3)
    return MY_ERROR_TOTAL;

  // Search for the URL numbered serverInfo.currentUrlListIndex in the
  // URL list in the registry parameters.  The URL's are separated by
  // semicolons.

  for (i = 0, p = regInfo.uploadUrlList; i < serverInfo.currentUrlListIndex;
       i++) {
    p = strchr(p, ';');
    if (p == NULL) {
      p = regInfo.uploadUrlList;
      serverInfo.currentUrlListIndex = 0;
      serverInfo.urlListSize = i + 1;
      break;
    }
    p++;
  }

  // If there is a proxy server, then setting up the server info is
  // straightforward:  the server is the proxy server, the port is the
  // proxy port, and the partial URL is the URL from the list (pointed to
  // by p).  If the URL pointed to by p is empty, though, fail this URL.

  if (regInfo.proxyServer[0] != '\0') {
    if (*p == ';' || *p == '\0')
      return ChangeWebServer(serverInfo, regInfo);
    strcpy(serverInfo.currentServer, regInfo.proxyServer);
    serverInfo.currentPort = regInfo.proxyPort;
    dest = serverInfo.currentPartialUrl;
    while (*p != ';' && *p != '\0')
      *dest++ = *p++;
    *dest = '\0';
    return MY_ERROR_NONE;
  }

  // If there is no proxy server, we have to parse the URL from the list
  // (pointed to by p)

  // First, make sure it starts with "http://", otherwise fail this URL

  if (strncmp(p, "http://", 7))
    return ChangeWebServer(serverInfo, regInfo);

  // Then, copy the server name from the next bytes

  p += 7;
  dest = serverInfo.currentServer;
  while (*p != ':' && *p != '/' && *p != '\0' && *p != ';')
    *dest++ = *p++;
  *dest = '\0';

  // If there's a colon, read the port number, otherwise make the port
  // equal to the default of 80

  if (*p == ':') {
    serverInfo.currentPort = 0;
    for (p++; *p >= '0' && *p <= '9'; p++)
      serverInfo.currentPort = serverInfo.currentPort * 10 + (*p - '0');
  }
  else
    serverInfo.currentPort = 80;

  // Copy the partial URL remaining.  If it is empty, fail this URL

  if (*p == ';' || *p == '\0')
    return ChangeWebServer(serverInfo, regInfo);
  dest = serverInfo.currentPartialUrl;
  while (*p != ';' && *p != '\0')
    *dest++ = *p++;
  *dest = '\0';

  // Return successful status value

  return MY_ERROR_NONE;
}

// InitServerInfo initializes the serverInfo structure by setting it to
// reflect the first valid URL in the list given in regInfo.  In other words,
// it sets the current server to the first in the list of possible servers.

void InitServerInfo (SERVER_INFO& serverInfo, REGISTRY_INFO& regInfo)
{
  serverInfo.currentUrlListIndex = -1;
  serverInfo.urlListSize = MAX_UPLOAD_URL_LIST_LENGTH;
  serverInfo.numUrlsFailedConsecutively = -1;
  ChangeWebServer(serverInfo, regInfo);
}

// IndicateServerIsWorking is called when a Success or Error response is
// obtained.  It indicates that the server is responding appropriately and
// should not be considered to have failed.

inline void IndicateServerIsWorking (SERVER_INFO& serverInfo)
{
  serverInfo.numUrlsFailedConsecutively = 0;
}

////////////////////////////////////////////////////////////////////////////
//                        LOW-LEVEL SEND ROUTINES
////////////////////////////////////////////////////////////////////////////

// SendBytes sends the contents of a given buffer (buf) of a certain length
// (length) to the terminus of a given socket (sock).  It returns a status
// value.

MY_ERROR SendBytes (JSocket& sock, char *buf, int length)
{
  int n;

  while (length > 0) {
    n = sock.Send(buf, length);
    if (n == 0 || n == SOCKET_ERROR)
      return MY_ERROR_SERVER;
    else {
      buf += n;
      length -= n;
    }
  }

  return MY_ERROR_NONE;
}

// SendString sends a NULL-terminated string (buf) to the terminus of a socket
// (sock).  It returns a status value.

MY_ERROR SendString (JSocket& sock, char *buf)
{
  return SendBytes(sock, buf, strlen(buf));
}

// SendCrc sends the given CRC crc to the given socket sock.  It only sends
// the least-significant 3 bytes of the CRC, in little endian order.  It
// returns a status value.

MY_ERROR SendCrc (JSocket& sock, DWORD crc)
{
  char crcAsStr[3];

  crcAsStr[0] = (unsigned char) (crc % 256);
  crcAsStr[1] = (unsigned char) ((crc >> 8) % 256);
  crcAsStr[2] = (unsigned char) ((crc >> 16) % 256);
  return SendBytes(sock, crcAsStr, 3);
}

// ConnectToWebsite connects the given socket (sock) to the current server.
// If it fails with a certain server, it tries changing servers until it
// succeeds.  It returns a status value.  Parameters:
//
//   sock -- the socket to create and connect
//   serverInfo -- information about the current server (can be changed)
//   regInfo -- information about the registry parameters read

MY_ERROR ConnectToWebsite (JSocket& sock, SERVER_INFO& serverInfo,
                           REGISTRY_INFO& regInfo)
{
  MY_ERROR err;

  // Create a socket

  if (!sock.Create())
    return MY_ERROR_TOTAL;

  while (!sock.Connect(serverInfo.currentServer, serverInfo.currentPort)) {
    err = ChangeWebServer(serverInfo, regInfo);
    if (err != MY_ERROR_NONE) return err;
  }

  return MY_ERROR_NONE;
}

////////////////////////////////////////////////////////////////////////////
//                       PROTOCOL INTERPRETATION
////////////////////////////////////////////////////////////////////////////

// ReceiveReply receives an HTTP reply from the given socket (sock).
// It puts the entire reply in the given buffer buf (assumed to be of
// size DATA_BUFFER_SIZE).  It returns a pointer to the main body of the
// reply (after the headers, such as the one giving the content length).
// If no reply is received, or it cannot be determined where the headers
// end, it returns NULL.

char *ReceiveReply (JSocket& sock, char *buf)
{
  LONG n, length;
  char *p;

  // Read as much as we can, up to DATA_BUFFER_SIZE bytes, into buf, then
  // add a null character at the end

  length = 0;
  while ((n = sock.Receive(&buf[length], DATA_BUFFER_SIZE - 1 - length)) > 0)
    length += n;
  buf[length++] = '\0';

  // Search for two newlines in a row, or separated by a single carriage-
  // return, since the reply comes right after that

  p = buf;
  while ((p = strchr(p, '\n')) != NULL) {
    if (p[1] == '\n') return &p[2];
    if (p[1] == '\r' && p[2] == '\n') return &p[3];
    p++;   // Not p+=2, since p+1 might be '\0' and we don't want to skip that
  }

  // We did not find two newlines, so return NULL

  return NULL;
}


////////////////////////////////////////////////////////////////////////////
//                            UPLOAD ROUTINES
////////////////////////////////////////////////////////////////////////////

// SendAll sends all the local files in the VTrace traces directory to the
// remote Web server.  It is passed as a parameter some useful registry
// information (regInfo).

void SendAll (REGISTRY_INFO& regInfo)
{
  char searchPattern[MAX_PATH];
  HANDLE findHandle;
  WIN32_FIND_DATA findData;
  SERVER_INFO serverInfo;
  int numIterationsDone;
  BOOL anyFileErrorsThisIteration;
  WSADATA wsaData;

  // Initialize sockets access

  if (WSAStartup(MAKEWORD(1, 1), &wsaData) != 0)
    return;

  // Initialize server info

  InitServerInfo(serverInfo, regInfo);

  // Do several iterations, each time trying to successfully send and
  // delete all the files in the directory.  Stop after 3 iterations (thus
  // giving each file 3 chances) or when no files failed in an iteration.
  // If any file fails with MY_ERROR_TOTAL, stop the iterations entirely.

  sprintf(searchPattern, "%s\\Traces\\*.*", regInfo.mainDirectory);
  numIterationsDone = 0;
  do {
    anyFileErrorsThisIteration = FALSE;
    findHandle = FindFirstFile(searchPattern, &findData);
    if (findHandle == INVALID_HANDLE_VALUE)
      break;
    do {
      switch (SendAndDeleteFile(findData, serverInfo, regInfo)) {
      case MY_ERROR_TOTAL :
        FindClose(findHandle);
        WSACleanup();
        return;
      case MY_ERROR_FILE :
        anyFileErrorsThisIteration = TRUE;
        break;
      }
    } while (FindNextFile(findHandle, &findData));
    FindClose(findHandle);
    numIterationsDone++;
  } while (anyFileErrorsThisIteration && numIterationsDone < 3);

  // If everything uploaded successfully, send a message indicating so.

  if (!anyFileErrorsThisIteration)
    SendLastTraceUploadedMessage(serverInfo, regInfo);

  // Terminate sockets access

  WSACleanup();
}

// SendAndDeleteFile tries sending a local file to any remote Web server to be
// created as a file there.  If it succeeds, it deletes the local file.
// It takes as input information about the file (findData), a structure to use
// for updating the current server (serverInfo), and the registry information
// (regInfo).  This routine returns a status value.

MY_ERROR SendAndDeleteFile (WIN32_FIND_DATA& findData, SERVER_INFO& serverInfo,
                            REGISTRY_INFO& regInfo)
{
  FILE_INFO fileInfo;
  MY_ERROR err;

  if (!WorthUploading(findData, fileInfo.fileSize)) return MY_ERROR_NONE;

  sprintf(fileInfo.fileNameHere, "%s\\Traces\\%s", regInfo.mainDirectory,
          findData.cFileName);
  sprintf(fileInfo.fileNameThere, "u%u/%s", regInfo.userNumber,
          findData.cFileName);

  err = SendFile(fileInfo, serverInfo, regInfo);
  if (err != MY_ERROR_NONE) return err;

  if (!DeleteFile(fileInfo.fileNameHere))
    return MY_ERROR_FILE;

  return MY_ERROR_NONE;
}

// SendFile tries sending a local file to any remote Web server to be
// created as a file there.  It takes as input information about the file
// (fileInfo), a structure to use for updating the current server (serverInfo),
// and the registry information (regInfo).  This routine returns a status
// value.

MY_ERROR SendFile (FILE_INFO& fileInfo, SERVER_INFO& serverInfo,
                   REGISTRY_INFO& regInfo)
{
  MY_ERROR err;

  if (fileInfo.fileSize == 0) return MY_ERROR_NONE;

  // Loop, each time calling SendFileToServer and changing web servers
  // if it fails due to the current server

  while (1) {
    err = SendFileToServer(fileInfo, serverInfo, regInfo);
    if (err != MY_ERROR_SERVER)
      return err;
    err = ChangeWebServer(serverInfo, regInfo);
    if (err != MY_ERROR_NONE)
      return err;
  }
}

// SendFileToServer sends a local file to the current remote Web server to
// be created as a file there.  (It is permitted to change Web servers if
// a server fails to connect at all, but otherwise it must return an error
// if a server fails.)  It takes as input information about the file
// (fileInfo), the current server (serverInfo), and the registry information
// (regInfo).  This routine returns a status value.

MY_ERROR SendFileToServer (FILE_INFO& fileInfo, SERVER_INFO& serverInfo,
                           REGISTRY_INFO& regInfo)
{
  MY_ERROR err;
  DWORD bytesToTransfer;

  // Find out how many bytes were already transferred.  If all the bytes
  // were transferred, return success.  If less than 1 KB or more than the
  // actual file size were transferred, pretend no data has been sent yet.

  err = GetRemoteFileSize(fileInfo, serverInfo, regInfo);
  if (err != MY_ERROR_NONE) return err;
  if (fileInfo.remoteFileSize == fileInfo.fileSize) return MY_ERROR_NONE;
  if (fileInfo.remoteFileSize < 1024 ||
      fileInfo.remoteFileSize > fileInfo.fileSize)
    fileInfo.remoteFileSize = 0;

  // Open the local file

  fileInfo.fileHandle = CreateFile(fileInfo.fileNameHere, GENERIC_READ, 0,
                                   NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,
                                   NULL);
  if (fileInfo.fileHandle == INVALID_HANDLE_VALUE)
    return MY_ERROR_FILE;

  // Position the file pointer just after the number of bytes that are
  // already at the remote site

  if (fileInfo.remoteFileSize &&
      SetFilePointer(fileInfo.fileHandle, fileInfo.remoteFileSize, NULL,
                     FILE_BEGIN) == 0xFFFFFFFF)
    return MY_ERROR_FILE;

  // Transfer a chunk of bytes at a time, never more than
  // regInfo.maxTransferSize

  do {
    bytesToTransfer = MIN(fileInfo.fileSize - fileInfo.remoteFileSize,
                          regInfo.maxTransferSize);
    err = SendFilePiece(fileInfo, serverInfo, regInfo, bytesToTransfer);
    if (err == MY_ERROR_NONE)
      fileInfo.remoteFileSize += bytesToTransfer;
    else
      break;
  } while (fileInfo.remoteFileSize < fileInfo.fileSize);

  CloseHandle(fileInfo.fileHandle);

  return err;
}

// SendFilePiece sends the next bytesToTransfer bytes from a local file
// to the current remote Web server to be appended to a remote file there
// at position fileInfo.remoteFileSize.  (It is permitted to change Web
// servers if a server fails to connect at all, but otherwise it must
// return an error if a server fails.)  It takes as input information
// about the file (fileInfo) including the file handle open to the
// position where the bytes begin, the current server (serverInfo), and
// the registry (regInfo).  This routine returns a status value.

MY_ERROR SendFilePiece (FILE_INFO& fileInfo, SERVER_INFO& serverInfo,
                        REGISTRY_INFO& regInfo, DWORD bytesToTransfer)
{
  JSocket sock;
  DWORD bytesRead, crc, i;
  char buf[DATA_BUFFER_SIZE], *reply;
  MY_ERROR err;

  // First make sure that the user has not become non-idle

  if (GetUserIdleTime() < USER_IDLE_MS_BEFORE_COMPRESS)
    return MY_ERROR_TOTAL;

  // Connect to the remote web site

  err = ConnectToWebsite(sock, serverInfo, regInfo);
  if (err != MY_ERROR_NONE) return err;

  // Send HTTP headers and preamble of content

  sprintf(buf,
          "POST %s HTTP/1.0\n"
          "Accept: text/html\n"
          "User-Agent: VTrace\n"
          "Content-type: application/x-www-form-urlencoded\n"
          "Content-length: %d\n"
          "\n"                     // Content begins:
          "%d.%d.%d\n"             // Version number is first line
          "%c\n"                   // Action type (append or create)
          "%s\n",                  // File name is third line
          serverInfo.currentPartialUrl,
          bytesToTransfer + strlen(fileInfo.fileNameThere) + 12,
            // The content length includes an additional 12 bytes: 6 for the
            // version number and newline after, 2 for the action and newline
            // after, 1 for the newline after the file name, and 3 for the CRC
          regInfo.vtraceVersion >> 16,
          (regInfo.vtraceVersion >> 8) % 256,
          regInfo.vtraceVersion % 256,
          fileInfo.remoteFileSize ? 'A' : 'C',  // Action type
          fileInfo.fileNameThere);
  err = SendString(sock, buf);
  if (err != MY_ERROR_NONE) return err;

  // Loop, each time sending up to DATA_BUFFER_SIZE bytes of the file, until
  // all the remaining bytes of the file have been sent.  Compute the CRC
  // as we go along.

  crc = 0;
  while (bytesToTransfer > 0) {
    if (!ReadFile(fileInfo.fileHandle, buf,
                  MIN(bytesToTransfer, DATA_BUFFER_SIZE), &bytesRead, NULL) ||
        bytesRead < 1)
      return MY_ERROR_FILE;
    err = SendBytes(sock, buf, bytesRead);
    if (err != MY_ERROR_NONE) return err;
    for (i = 0; i < bytesRead; i++)
      crc += (DWORD) (unsigned char) buf[i];
    bytesToTransfer -= bytesRead;
  }

  // If there are any bytes left over, fail

  if (bytesToTransfer != 0)
    return MY_ERROR_FILE;

  // Send the CRC as a check of all the bytes sent

  err = SendCrc(sock, crc);
  if (err != MY_ERROR_NONE) return err;

  // Try to receive the HTTP reply, then return MY_ERROR_NONE if and only if
  // we got a successful one.  If we got a reply starting with something other
  // than "Success" or "Error", this suggests a web server misconfiguration
  // and so we return a server error status code.  Otherwise, we indicate
  // that the server responded appropriately and is thus working.

  reply = ReceiveReply(sock, buf);
  if (reply == NULL) return MY_ERROR_SERVER;
  if (!strncmp(reply, "Success", 7)) {
    IndicateServerIsWorking(serverInfo);
    return MY_ERROR_NONE;
  }
  if (!strncmp(reply, "Error", 5)) {
    IndicateServerIsWorking(serverInfo);
    return MY_ERROR_FILE;
  }
  return MY_ERROR_SERVER;
}

// GetRemoteFileSize queries a Web server (given in serverInfo) for the length
// of the remote file of a certain name (given in fileInfo), and puts that
// length in fileInfo.remoteFileLength.  It also takes a parameter giving
// certain registry values (regInfo), of which it needs the version of VTrace.
// It returns a status value.

MY_ERROR GetRemoteFileSize (FILE_INFO& fileInfo, SERVER_INFO& serverInfo,
                            REGISTRY_INFO& regInfo)
{
  JSocket sock;
  char buf[DATA_BUFFER_SIZE], *p, *reply;
  MY_ERROR err;

  // Connect to the remote web site

  err = ConnectToWebsite(sock, serverInfo, regInfo);
  if (err != MY_ERROR_NONE) return err;

  // Send HTTP headers and content

  sprintf(buf,
          "POST %s HTTP/1.0\n"
          "Accept: text/html\n"
          "User-Agent: VTrace\n"
          "Content-type: application/x-www-form-urlencoded\n"
          "Content-length: %d\n"
          "\n"                     // Content begins:
          "%d.%d.%d\n"             // Version number is first line
          "S\n"                    // Action type "size" is second line
          "%s\n",                  // File name is third line
          serverInfo.currentPartialUrl,
          strlen(fileInfo.fileNameThere) + 9,
            // The content length includes an additional 9 bytes: 6 for the
            // version number and newline after, 2 for the action and newline
            // after, and 1 for the newline after the file name
          regInfo.vtraceVersion >> 16,
          (regInfo.vtraceVersion >> 8) % 256,
          regInfo.vtraceVersion % 256,
          fileInfo.fileNameThere);
  err = SendString(sock, buf);
  if (err != MY_ERROR_NONE) return err;

  // Get the reply, and look for the numeric file size in it.  If we get
  // a reply starting with something other than "Success" or "Error", this
  // suggests a web server misconfiguration and so we return a server error
  // status code.  Otherwise, we indicate that the server responded
  // appropriately and is thus working.  If we get an "Error" code, we
  // assume the remote file size is 0 and return success (MY_ERROR_NONE).

  reply = ReceiveReply(sock, buf);
  if (reply == NULL) return MY_ERROR_SERVER;
  if (!strncmp(reply, "Success", 7)) {
    IndicateServerIsWorking(serverInfo);
    fileInfo.remoteFileSize = 0;
    for (p = &reply[7]; *p; p++)
      if ((*p >= '0') && (*p <= '9'))
        fileInfo.remoteFileSize = fileInfo.remoteFileSize * 10 + (*p - '0');
    return MY_ERROR_NONE;
  }
  if (!strncmp(reply, "Error", 5)) {
    IndicateServerIsWorking(serverInfo);
    fileInfo.remoteFileSize = 0;
    return MY_ERROR_NONE;
  }
  return MY_ERROR_SERVER;
}

// SendLastTraceUploadedMessage creates a remote file named
// LastTraceUploaded.txt that contains the number of the last trace file
// created. Since this function is meant to be called after all files in
// the trace directory are gone, this indicates to the server that all
// files through that trace number have been created and uploaded.

MY_ERROR SendLastTraceUploadedMessage (SERVER_INFO& serverInfo,
                                       REGISTRY_INFO& regInfo)
{
  JSocket sock;
  DWORD crc, i, traceFileNumber, bytesToTransfer;
  char fileNameThere[MAX_PATH], buf[DATA_BUFFER_SIZE], contents[20], *reply;
  MY_ERROR err;

  // Obtain the number of the last trace file created, and put it in the
  // buffer to be sent as the file.

  if (!GetCurrentTraceFileNumber(&traceFileNumber))
    return MY_ERROR_FILE;
  sprintf(contents, "%d\n", traceFileNumber);
  bytesToTransfer = strlen(contents);

  // Get the name of the remote file that will be created.

  sprintf(fileNameThere, "u%u/LastTraceUploaded.txt", regInfo.userNumber);

  // Connect to the remote web site

  err = ConnectToWebsite(sock, serverInfo, regInfo);
  if (err != MY_ERROR_NONE) return err;

  // Send HTTP headers and preamble of content

  sprintf(buf,
          "POST %s HTTP/1.0\n"
          "Accept: text/html\n"
          "User-Agent: VTrace\n"
          "Content-type: application/x-www-form-urlencoded\n"
          "Content-length: %d\n"
          "\n"                     // Content begins:
          "%d.%d.%d\n"             // Version number is first line
          "C\n"                    // Action type (append or create)
          "%s\n"                   // File name is third line
          "%s",                    // Followed by file contents
          serverInfo.currentPartialUrl,
          bytesToTransfer + strlen(fileNameThere) + 12,
            // The content length includes an additional 12 bytes: 6 for the
            // version number and newline after, 2 for the action and newline
            // after, 1 for the newline after the file name, and 3 for the CRC
          regInfo.vtraceVersion >> 16,
          (regInfo.vtraceVersion >> 8) % 256,
          regInfo.vtraceVersion % 256,
          fileNameThere, contents);
  err = SendString(sock, buf);
  if (err != MY_ERROR_NONE) return err;

  // Compute and send the CRC.

  crc = 0;
  for (i = 0; i < bytesToTransfer; i++)
    crc += (DWORD) (unsigned char) contents[i];

  // Send the CRC as a check of all the bytes sent

  err = SendCrc(sock, crc);
  if (err != MY_ERROR_NONE) return err;

  // Try to receive the HTTP reply, then return MY_ERROR_NONE if and only if
  // we got a successful one.  If we got a reply starting with something other
  // than "Success" or "Error", this suggests a web server misconfiguration
  // and so we return a server error status code.  Otherwise, we indicate
  // that the server responded appropriately and is thus working.

  reply = ReceiveReply(sock, buf);
  if (reply == NULL) return MY_ERROR_SERVER;
  if (!strncmp(reply, "Success", 7)) {
    IndicateServerIsWorking(serverInfo);
    return MY_ERROR_NONE;
  }
  if (!strncmp(reply, "Error", 5)) {
    IndicateServerIsWorking(serverInfo);
    return MY_ERROR_FILE;
  }
  return MY_ERROR_SERVER;
}
